
(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: abstraction.ml,v 1.58 2003/08/20 18:44:42 rupak Exp $

 *
 * This file is part of the BLAST Project.
 *)


(**
 * This module implements the abstraction interface connecting the model
 * checker to the concrete transition system.  It defines a Region module
 * based on Predicate Abstraction and BDDs, the symbolic pre and post,
 * and routines for refinment used by the model checker.
 *) 


open BlastArch

module Symbol = Ast.Symbol
module Constant = Ast.Constant
module Expression = Ast.Expression
module Predicate = Ast.Predicate

(* [Greg] Very dirty hack because we renamed Stats to VampyreStats in
   vampyre...  We should clean this stats business anyway. *)
module Stats = VampyreStats


module Make_C_Abstraction =
  functor(C_SD : SYSTEM_DESCRIPTION with module Command = BlastCSystemDescr.C_Command) ->
struct

  let arg_subst _ = failwith "Not implememted"

  let check_races = ref false

  let assumed_invariants = ref []

  (** Code relating to initializing the BDD package.
      Defines the manager, true, and false -BDDs.
      *)
  let bddManager = CaddieBdd.init 0 32 256 512 (* Find out what these numbers mean  *)
  let bddZero = CaddieBdd.bddFalse bddManager
  let bddOne = CaddieBdd.bddTrue bddManager

  (* KILL VARIABLES *)
  let killPredList = ref [] 
  let killLocation = ref None 
  let killPredTable = Hashtbl.create 37
  let killCounter = ref 0
  let killCounterCovered = ref 0
  let killCounterCovered_force = ref 0
  let killCounterCovered_seen = ref 0


  (* Split predicate Table *)
  let post_split_table = Hashtbl.create 109
  let loc_pred_table = Hashtbl.create 1009
  (* this will map locations to the predicates that are useful there 
      which is info gleaned from the interpolants *)
  let globally_useful_pred_list = ref []
  (* predicates who's value we want to know pretty much everywhere ... 
     e.g. spec predicates *)
 
   (******************************************************************)
  (* Code relating to the management of global list of predicates *)

  (* Global table storing atomic predicate, index, and bdd corresponding
     to this predicate *)
  
  let maxPossiblePreds = 1300 (* ho ho ho ... parport makes this crash! *)
  let tableOfPreds = ref ([] : (int * Predicate.predicate * CaddieBdd.bdd) list)
  
  let hashTableOfPreds = Hashtbl.create 109
  let arrayOfPreds = Array.create maxPossiblePreds (Predicate.True,bddOne) 
  let listOfConstants = ref [] 
  let predIndex = ref (-1)
  let getNextIndex () = predIndex := !predIndex + 1; !predIndex
  let getCurrentIndex () = !predIndex

  let resetIndex () = predIndex := -1

  (** Stats and debug routine *)
  let printTableOfPreds () =
    let rec loop lst i =
      match lst with
        [] -> i
      | (a,b,_) :: rest ->
         begin
           Message.msg_string Message.Normal ("//Index = "^(string_of_int a)^" Predicate: \n"^(Predicate.toString b)^";\n");
           loop rest (i+1)
	 end
    in
    let len = loop (!tableOfPreds) 0 in
    Message.msg_string Message.Normal ("\nNumber of predicates ="^(string_of_int len)^"\n")


 
        (* Given an index, find predicate corresponding to that index;
           raises Not_found if no such index 
        *)
  let getPred index = 
    let rec _searchPred lst i =
      match lst with
          [] -> raise Not_found
        | (j,p,_)::rest -> if (j=i) then p else _searchPred rest i
    in
      if
        (index > !predIndex)
      then
	begin
	  Message.msg_string Message.Debug "Looking for unknown predicate!";
	  raise Not_found
	end
      else
	fst (arrayOfPreds.(index))
        (* _searchPred (!tableOfPreds) index *)

  (* Given a predicate, find the index in the global table corresponding to that predicate;
     raises Not_found if no such index 
  *)
  let getPredIndex pred = 
    let rec _searchPred lst pr =
      match lst with
          [] -> raise Not_found
        | (j,p,_)::rest -> if (p=pr) then j else _searchPred rest pr
    in
      (Hashtbl.find hashTableOfPreds pred)
      (* _searchPred (!tableOfPreds) pred *)

  (* Given a predicate, find the BDD node corresponding to that predicate *)
  let getPredBdd pred = 
    let rec _searchPred lst pr =
      match lst with
          [] -> raise Not_found
        | (_,p,b)::rest -> if (p=pr) then b else _searchPred rest pr
    in
    let i =  getPredIndex pred in
      snd (arrayOfPreds.(i))
      (*_searchPred (!tableOfPreds) pred*)

  (* Given an index, find bdd corresponding to that index;
     raises Not_found if no such index 
  *)
  let getPredIndexBdd index = 
    let rec _searchPred lst i =
      match lst with
          [] -> raise Not_found
        | (j,_,b)::rest -> if (j=i) then b else _searchPred rest i
    in
      if
        (index > !predIndex)
      then
        raise Not_found
      else
	snd (arrayOfPreds.(index))
        (*_searchPred (!tableOfPreds) index *)



(* the following function scans the code and fills up a the constant table which is used in predH > 0 *)
  
let isConstant e = 
  match e with 
      Expression.Constant c -> true
    | _ -> false

let lval_constant_table = Hashtbl.create 1009


let make_const_table () = 
  (* we trawl over the whole CFA and learn for every lval expression all the constant exp *)
  let module LocationSet =
    Set.Make (struct
                type t = int 
                let compare x y =  compare x  y
              end) in
    
  let add_to_table (x,y) = 
    if (isConstant y && (not (isConstant x))) then
      begin
	let l = try Hashtbl.find lval_constant_table x with Not_found -> [] in
	  if (List.mem y l) then () else Hashtbl.replace lval_constant_table x (y::l)
	end
    else ()
  in
  let process_statement s = 
    match s with
	C_SD.Command.Expr e -> 
	  List.iter (fun p -> add_to_table p) (Expression.assignments_of_expression e)
      | _ -> ()
  in
  let sprint_edge e = 
    let process_atom p = 
      match p with
	  Predicate.Atom a ->
	    begin
	      match a with
		  Expression.Binary (op,e1,e2) -> (List.iter add_to_table [(e1,e2);(e2,e1)])
		| _ -> ()
	    end
	| _ -> ()
    in
    let (l1,c,l2) = (C_SD.get_source e, C_SD.get_command e, C_SD.get_target e) in 
      match c.C_SD.Command.code with
	  C_SD.Command.Block l -> List.iter process_statement l
	| C_SD.Command.Pred p ->  
	    if (Options.getValueOfInt "predH" > 1) 
	    then 
	      List.iter (process_atom) (Predicate.getAtoms p)
	    else ()
	| _ -> ()
  in 
  let sprint_function fname =
    let already_visited_source_locations = ref LocationSet.empty in
    let rec sprint_location_and_edges loc =
      let (_,loc_id) = C_SD.location_coords loc in
        if (LocationSet.mem loc_id !already_visited_source_locations) 
	  then
            ()
        else
          begin
            let outgoing_edges = List.rev (C_SD.get_outgoing_edges loc) in
	    let _ = List.iter sprint_edge outgoing_edges in
	      already_visited_source_locations := (LocationSet.add (snd (C_SD.location_coords loc)) !already_visited_source_locations) ;
	      (List.iter (fun e -> sprint_location_and_edges (C_SD.get_target e)) outgoing_edges)          end 
    in
      sprint_location_and_edges (C_SD.lookup_entry_location fname)
  in
    List.iter sprint_function (C_SD.list_of_functions ());
    let print_entry lv constlist = 
      Message.msg_string Message.Debug (Printf.sprintf " %s : " (Expression.toString lv));
      List.iter (fun x -> Message.msg_string Message.Debug (Printf.sprintf " %s,"(Expression.toString x))) constlist;
      Message.msg_string Message.Debug "\n"; ()
    in
      Hashtbl.iter print_entry lval_constant_table 
(* total hack -- this should be replaced with the 
   function that pulls lvals out of bigger expressions from the psrc directory ! *)
      

let rec getConstants p = 
  let rec getConstants_exp e = 
    match e with
	Expression.Binary(_,e1,e2) -> List.flatten (List.map getConstants_exp [e1;e2])
      | Expression.Unary(Expression.UnaryMinus,Expression.Constant (Constant.Int i)) -> 
	  [e]
      | Expression.Constant(Constant.Int i) -> [e]
      | _ -> []
  in
    match p with
	Predicate.True -> []
      | Predicate.False -> []
      | Predicate.And l -> List.flatten (List.map getConstants l)
      | Predicate.Or l -> List.flatten (List.map getConstants l)
      | Predicate.Atom e -> getConstants_exp e
      | _ -> []

let constant_preds p = 
  let _ = Message.msg_string Message.Debug ("Finding Constant Preds for: "^(Predicate.toString p)^"\n") 
  in
  let lvals_of_predicate p = 
    match p with
	Predicate.Atom e -> 
	  begin
	    match e with
		Expression.Binary(op,e1,e2) -> [e1;e2]
	      | _ -> []
	  end
      | _ -> []
  in
  let current_pred_lvals = (lvals_of_predicate p) in
  let make_predicate v c = Predicate.Atom (Expression.Binary (Expression.Eq,v,c)) in
  let process_lval v = 
    try 
      let rv =  List.map (make_predicate v) (Hashtbl.find lval_constant_table v) in
	List.iter (fun x -> Message.msg_string Message.Debug ((Predicate.toString x)^"\n")) rv;
	rv
    with
	Not_found -> 
	  Message.msg_string Message.Debug ("lval not in lval_constant_table: "^(Expression.toString v));
	  []
  in
  let rv1 =  List.flatten (List.map process_lval (lvals_of_predicate p )) in
    (* in more generality *)
  (*let rv2 =
    let mk_pred c c' = Predicate.substitute [(c,c')] p in
    let pred_lv = getVarExps p in
      match pred_lv with
	  [h] -> 
	    begin
	      let pred_consts = getConstants p in
	      if (List.length pred_consts <> 1) then []
	      else 
		List.map (mk_pred (List.head pred_consts)) (Hashtbl.find lval_constant_table h)
	    end
  in    *)
    rv1(* @rv2 *)

 (** Add a new atomic predicate to the global table of
    predicates and introduce a new BDD variable corresponding
    to the index
  *)
                        (* It returns the index of the pred  added *)
  let actual_addPred pred =
    let reflexcheck e = 
      match e with
	  Expression.Binary(_,e1,e2) -> e1 = e2 
	| _ -> false
    in
      try 
	Hashtbl.find hashTableOfPreds pred 
      with Not_found ->        
	begin
          (* Check if negation of the predicate exists already *)
          let _negPred = Predicate.negateAtom pred in
            try  
	      Hashtbl.find hashTableOfPreds _negPred
            with Not_found -> 
              begin
                (* check that there is at least one variable --
                   otherwise the predicate is Predicate.True or
                   Predicate.False : no point adding it.
                   *)
            match pred with
              Predicate.Atom e ->
                begin
		  if reflexcheck e then -1 
		  else
		    match (Expression.allVarExps e) with
			[] -> -1
                      | _  ->  
			  begin
                            Message.msg_string Message.Debug ("//addPred: adding predicate \n"^(Predicate.toString pred)^";");
			    let ps = Predicate.toString pred in
                              Message.msg_string Message.Error ("addPred: (gui) adding predicate "^(ps)^" to the system.\n");
			      Misc.append_to_file ".newpred" (ps^"\n");
			  
                          match pred with
			      Predicate.Atom _ ->
				let i = getNextIndex () in
				let bddnd = CaddieBdd.bddNewVar bddManager in                                                             tableOfPreds := (i, pred, bddnd) :: (!tableOfPreds); 
				  Array.set arrayOfPreds i (pred,bddnd);
				  Hashtbl.add hashTableOfPreds pred i;
				    i
                              | _ -> failwith "addPred: Attempt to add non-atomic predicate to predicate table"
			  end
                end
              | _ -> failwith "addPred: Unexpected predicate type."
          end
      end

 

let addPred p = 
  let pred = Predicate.canonicize p in
  (* so for 1 2 3 we do this hack *)
  if (Options.getValueOfInt "predH" = 0) then actual_addPred pred
  else (* the above value should be 1 or 2 but we shall not check it here *)
    begin
      let const_pred_list = constant_preds pred
      in 
      let _ = List.map actual_addPred const_pred_list in
      actual_addPred pred
    end
    
let just_addPred p = 
  let pred = Predicate.canonicize p in
    actual_addPred pred

      
  let getPredsContainingVar str =
    let containsVar p str =
      let e_str = Expression.Lval (Expression.Symbol str) in
      match p with 
        Predicate.Atom expr -> (List.mem e_str (Expression.allVarExps expr))
      |	_ -> false
    in
    let rec _go_through_list lst str accum =
      match lst with
        [] -> accum
      |	(_, p, _) :: rest -> if (containsVar p str) then _go_through_list rest str (p::accum) else
        _go_through_list rest str accum
    in
    _go_through_list (!tableOfPreds) str []
       
  
     
let dump_abs () = 
  let plist = List.map Misc.snd3 !tableOfPreds in
     let loc_info = 
       match Options.getValueOfInt "craig" >= 2 with
	 true -> 
	   begin
	     let loc_l = Misc.hashtbl_keys loc_pred_table in
	     let get_pred l = try Hashtbl.find loc_pred_table l with Not_found -> [] in

	       List.map (fun loc -> (loc, get_pred loc)) loc_l
	   end
	 | false -> []
     in
     let fname = (Options.getValueOfString "mainsourcename")^".abs" in
     let _ = Message.msg_string Message.Debug ("Writing out .abs file: "^fname) in     
     let oc = open_out fname in
       Marshal.to_channel oc (plist,loc_info) []; close_out oc; 
let _ = Message.msg_string Message.Debug ("Done writing .abs file") in     
     ()


  (******************************************************************)
  (* Code to convert Predicates to BDD's *)
  let rec convertPredToBdd pred =
    match pred with
      Predicate.Atom a -> 
	begin
	  try
	    getPredBdd pred
	  with
	    Not_found -> 
	      Message.msg_string Message.Debug ("Did not find predicate "^(Predicate.toString pred)) ;
	      CaddieBdd.bddNot (getPredBdd (Predicate.negateAtom pred))
	end
    | Predicate.And plist -> List.fold_left (fun b -> fun p -> CaddieBdd.bddAnd b (convertPredToBdd p)) bddOne plist
    | Predicate.Or plist ->  List.fold_left (fun b -> fun p -> CaddieBdd.bddOr b (convertPredToBdd p)) bddZero plist
    | Predicate.Not p -> CaddieBdd.bddNot (convertPredToBdd p)
    | Predicate.Implies (p1, p2) -> CaddieBdd.bddImp (convertPredToBdd p1) (convertPredToBdd p2)
    | Predicate.True -> bddOne
    | Predicate.False -> bddZero
    | _ -> failwith "convertPredToBdd: cannot handle quantified predicates"
	  
  (** 
    Code to convert a cube from a BDD (given as an int array) to
    the corresponding predicate
  *)
  let convertCubeToPred cube =
    let _convertCubeWorker cube = 
      (let predlst = ref [Predicate.True] 
       in
         for i = 0 to (Array.length cube -1) do
           if
             (cube.(i) == 0)
           then
             predlst := (Predicate.negate (getPred i))::!predlst
           else
             begin
               if
                 (cube.(i)==1)
               then
                 predlst := (getPred i)::!predlst
               else (* cube.(i) == 2 *)
                 ()
             end
         done;
         !predlst)
    in
    let pl = (_convertCubeWorker cube)
    in
      if (Array.length cube = 1) then
        (List.hd pl) else
          Predicate.And pl


  (* Code to convert a BDD to a Predicate (sum-of-products form) *)
  let convertBddToPred bdd =
    let list_of_minterms = ref []
    in
    let _worker cube =
      list_of_minterms := (convertCubeToPred cube)::(!list_of_minterms)
    in
      if (CaddieBdd.bddEqual bddOne bdd) then
        Predicate.True
      else 
        begin
          if (CaddieBdd.bddEqual bddZero bdd) then
            Predicate.False
          else
            begin
              CaddieBdd.bddForeachCube bdd _worker;
              if (List.length (!list_of_minterms) == 1) 
              then (List.hd !list_of_minterms)
              else Predicate.disjoinL (!list_of_minterms)
            end
        end


  (**************************************************************************)
        

  (* Code to ask theorem prover queries and check satisfiability *)        
  let theoremProverCache = Hashtbl.create 49033
                             
  (* Statistics: Number of queries to the Theorem Prover *)
  let statTotalNumberOfQueries = ref 1
  let statCoveredQueries = ref 1
  let statTotalCached = ref 0
  let statWorstCaseQueriesForPost = ref 0
  let statActualQueriesForPost = ref 0
  let statAssumeQueriesForPost = ref 0
  let statNumPost = ref 0
  let statNumAssumePost = ref 0
  let statReachedCubes = ref 0
  let statVisitedFunctionTable = Hashtbl.create 109

  let reset_abs_stats () = 
    statTotalCached := 0;
    statTotalNumberOfQueries := 1;
    statWorstCaseQueriesForPost := 0;
    statReachedCubes := 0;
    statActualQueriesForPost := 0;
    statAssumeQueriesForPost := 0;
    statNumPost := 0;
    statNumAssumePost := 0;
    statCoveredQueries := 0; 
    ()
  
  (****************************************************************************)
      
      
  exception SatisfiableException
    
  (* This is the main interface to the theorem prover. This takes
     a BDD and a list of predicates, and checks, for each cube of the BDD if
     the conjunction of the cube and the list of predicates is satisfiable.
  *)

  let isSatisfiable (bdd, plist) =
    try 
    begin
    Message.msg_string Message.Debug "In isSatisfiable..\n";
    Message.msg_string Message.Debug ("Bdd = "^(Predicate.toString (convertBddToPred bdd)));
    Message.msg_string Message.Debug "\nList of predicates = ";
    let rec loop lst = 
      match lst with [] -> () | a::rest -> (ignore (Message.msg_string Message.Debug (Predicate.toString a)); loop rest) 
    in
      loop plist;
      let rec _syntaxCheck cube alist =
        (*
        let rec _DBGprintAlist lst = 
	  match lst with [] -> () 
          | (a,b)::rest -> (Message.msg_string Message.Debug ("Arity = "^(string_of_int a)^"\tPredicate = ") ;
	                    Message.msg_string Message.Debug (Predicate.toString (getPred b)); _DBGprintAlist rest)
	in
        _DBGprintAlist alist;
        *)
        begin
	  try
            (match alist with
              [] -> false
            | (i, j) :: rest ->
                if ((cube.(j) <> 2) && (cube.(j) <> i)) then
                  true (* Unsatisfiable -- Easy win! *)
                else
                  _syntaxCheck cube rest)
with Invalid_argument s -> failwith ("Invalid_argument "^s^" here")
        end
      in
        
      let modeledPredIndexList = (* List of modeled predicates or their negations from plist *)
        let rec _origPreds plist orig = (* List of atomic predicates *)
          match plist with
              [] -> orig
            | a::rest ->
                begin
                  try
                    let i = getPredIndex a
                    in
                      _origPreds rest ((1,i) :: orig)
                  with
                      Not_found -> 
                        begin
                          try
                            let i = getPredIndex (Predicate.negate a)
                            in
                              _origPreds rest ((0,i) :: orig)
                          with 
                              Not_found ->
                                _origPreds rest orig
                        end
                end
        in
          _origPreds plist []
      in
      let _satworker plist cube =
        (* check if this cube conjoined with plist is satisfiable *)
        (* Note: The "cube" may be fake, so don't print it or try to convert
	   it to a predicate. Ugly hack, but works. So far.
	   *)
        (* Syntactic heuristic *)
        if (_syntaxCheck cube modeledPredIndexList) then
          ()
        else
          begin
            let _make_Pred (cube : int array) (plist : Predicate.predicate list) =
              let _ = Message.msg_string Message.Debug "HITHER" in
	      
	      let pref = ref [] in
                begin
                  for i = 0 to (Array.length cube - 1) do
                    if (cube.(i)==2) 
                    then () 
                    else 
                      begin
                         if (cube.(i)==1) then 
                           pref := (getPred i)::(!pref)
                         else
                           pref := (Predicate.negate (getPred i))::(!pref)
                      end
                   done;
                   Predicate.implies (Predicate.conjoinL ((!pref) @ plist)) (Predicate.False)
                end 
            in
            begin
                (*(ignore (Printf.printf "Asking query...%d\n" !statTotalNumberOfQueries));*)
              statTotalNumberOfQueries := !statTotalNumberOfQueries + 1;
              statCoveredQueries := !statCoveredQueries + 1;
		let eToCheck = _make_Pred cube plist
                in
                let _ = Message.msg_string Message.Debug "query-start" in
                    let ans = 
		      if (Options.getValueOfBool "nocache") then 
			TheoremProver.queryExp eToCheck
		      else
			begin
			  statTotalCached := !statTotalCached + 1;
			  try
			    Hashtbl.find theoremProverCache eToCheck
			  with Not_found -> 
			    let _ans = TheoremProver.queryExp eToCheck in
			    Hashtbl.add theoremProverCache eToCheck _ans;
			    statTotalCached := !statTotalCached - 1;
			    _ans
			end
		    in
		    let _ = Message.msg_string Message.Debug "query-end" in
		    if ans = false
                    then
                      raise SatisfiableException
                    else
                      ()
            end
          end
      in
        try
          if (CaddieBdd.bddEqual bdd bddZero) then
            false
          else 
            begin
              if (CaddieBdd.bddEqual bdd bddOne) 
              then
                let _tmpcube = (* build a temporary cube with nothing in it *)
                  Array.make (*size *)(getCurrentIndex ()+2) (* content *)2 (* Note: The size
									     is certainly a vast
									     overestimate, but this
									     is the easiest way 
									     The "+2" removes additional
									       complications relating
									       to index being -1 initially.
									       Note that this is a fake
									       cube -- we should neve
									       try to print it or convert it
									       to a predicate!

									       Ugly hack, but works. So far.
									       *)
                in
                  _satworker plist _tmpcube;
                  false
              else
                begin
                  ignore (CaddieBdd.bddForeachCube bdd (_satworker plist));
                  false
                end
            end
        with SatisfiableException ->
          true
       end
       with Out_of_memory -> failwith "isSatisfiable raises out of mem" 
(* JHALA: SLAM_POST post_bdd *)
(* bdd: current region,
   pred_list : (pred, p1, p2) list where pred: modelled predicate, p1: wp(pred wrt whatever), p2 : ( wp(\not pred wrt whatever)) *)
let askTheoremProver pred =
    (*Message.msg_string Message.Debug ("Asking query..."^(string_of_int !statTotalNumberOfQueries)^"\n");*)
  statTotalNumberOfQueries := !statTotalNumberOfQueries + 1;
  
  if (Options.getValueOfBool "nocache") then 
    TheoremProver.queryExp pred
  else
    begin

      if (Hashtbl.mem theoremProverCache pred) 
      then 
	begin
          
          statTotalCached := !statTotalCached + 1;
          (Hashtbl.find theoremProverCache pred)
      end
      else 
	begin
          let t2 = TheoremProver.queryExp pred in
          Hashtbl.add theoremProverCache pred t2;
          t2
	end
    end


(* Same as ask thm prover but assumes that you have told the thm prover your assumptions pred is the total predicate ie asm -> foo and query is the foo ... theoretically you could pass just one arg and pattern match but ...*)


let askTheoremProverContext pred query =
 (*Message.msg_string Message.Debug ("Asking query..."^(string_of_int !statTotalNumberOfQueries)^"\n");*)
  statTotalNumberOfQueries := !statTotalNumberOfQueries + 1;
  
  if (Options.getValueOfBool "nocache") then 
    begin
      Message.msg_string Message.Debug (Predicate.toString query);
      TheoremProver.queryExpContext query
    end
    else
    begin
      if (Stats.time "lookup_cache_mem" (Hashtbl.mem theoremProverCache) pred) 
      then 
	begin
          statTotalCached := !statTotalCached + 1;
          (Stats.time "lookup_cache_find"  (Hashtbl.find theoremProverCache) pred)
	end
      else 
	begin
        let t2 = TheoremProver.queryExpContext query in
        Stats.time "lookup_cache_add" (Hashtbl.add theoremProverCache pred) t2;
        t2
	end
    end

  (**************************************************************************)

  (*
   * This module defines a symbolic representation for C programs, equipped
   * with boolean operations.
   *)
  module Region =
  struct

    (* list of integers corresponding to the predicates used in this region *)
    type predList = int list

    type abstract_data_region = {
      bdd   : CaddieBdd.bdd ;
      preds : predList ;
    }

    type concrete_data_region = {
      pred : Predicate.predicate ;
    }

    (* the type of data regions *)
    type data_region =
        Abstract of abstract_data_region
      | Concrete of concrete_data_region

    (* the type of atomic regions *)
    type stack_contents =
        CallStack of C_SD.location list
      | EveryStack

    let list_of_stack stack = 
      match stack with
	EveryStack -> []
      |	CallStack s -> s

    type atomic_region = {
      location : C_SD.location ;
      data     : data_region ;
      stack    : stack_contents;
    }

    (* davare: added code to define a control region *)
    type control_region = {
	p_location : C_SD.location;
	p_stack    : stack_contents;
      }
	  
    (* this module allows to manipulate maps of pairs (location, stack) to
       data regions *)
    module AtomicRegionTable =
      Hashtbl.Make (
      struct
        type t = C_SD.location * stack_contents 
	let hash (loc,stack) = 
	  let stack_to_string s =
	    match s with
	      EveryStack -> "EveryStack@@"
	    |	CallStack(_s) -> List.fold_right (function x -> function y -> ((C_SD.location_to_string x)^"@@"^y)) _s ""
	  in
	  Hashtbl.hash ((C_SD.location_to_string loc)^"@@@"^(stack_to_string stack))
	    
        let equal (l1, s1) (l2, s2) = 
	  let l1s = C_SD.location_to_string l1 in
          let l2s = C_SD.location_to_string l2 in
          let stack_eq stack1 stack2 = 
	    match (stack1,stack2) with
	      (CallStack(_s1),CallStack(_s2)) ->
		if ((List.map C_SD.location_to_string _s1) = (List.map C_SD.location_to_string _s2)) 
		then true else false
	    | (EveryStack,EveryStack) -> true
	    | _ -> false
	  in
	  (* Message.msg_string Message.Debug ("COMPARE args: "^l1s^" "^l2s); *)
	  if ((l1s = l2s) && (stack_eq s1 s2)) then true else false 
      end)

    (* davare: this module allows to manipulate maps of pairs (location list, stack list) to
       data regions *)
    module ParallelAtomicRegionTable =
      Hashtbl.Make (
      struct
        type t = control_region list  
	let hash (c_r_list) = 
	  let stack_to_string s =
	    match s with
	      EveryStack -> "EveryStack@@"
	    |	CallStack(_s) -> List.fold_right (function x -> function y -> ((C_SD.location_to_string x)^"@@"^y)) _s ""
	  in
	  let control_region_to_string c = 
	    (C_SD.location_to_string c.p_location) ^ "@@@" ^ (stack_to_string c.p_stack)
	  in
	  let hash_string = ref ""
	  in
	  List.iter(fun c_r -> hash_string := ((!hash_string) ^ "&&&&" ^ (control_region_to_string c_r)) ) c_r_list; 
	  Hashtbl.hash (!hash_string)
	    
        let equal (control_region_list_1) (control_region_list_2) = 
          let stack_eq stack1 stack2 = 
	    match (stack1,stack2) with
	      (CallStack(_s1),CallStack(_s2)) ->
		if ((List.map C_SD.location_to_string _s1) = (List.map C_SD.location_to_string _s2)) 
		then true else false
	    | (EveryStack,EveryStack) -> true
	    | _ -> false
	  in
	  let flag = ref true
	  in
	   List.iter2(fun c_r_1 c_r_2 -> 
		if ( (C_SD.location_to_string c_r_1.p_location) <> (C_SD.location_to_string c_r_2.p_location)
		|| (not (stack_eq c_r_1.p_stack c_r_2.p_stack)) ) then
		flag := false
	   ) control_region_list_1 control_region_list_2;
	   !flag;	
      end)
    
    type parallel_atomic_region = {
      p_control    : control_region list;
      p_data       : data_region ;
    }
	
    (* the type of regions *)
    type t =
        Empty
      | Atomic of atomic_region
      | Union of data_region AtomicRegionTable.t
      | ParallelAtomic of parallel_atomic_region
      | ParallelUnion of data_region ParallelAtomicRegionTable.t

    let atomic_region_parameter = 37
	(* the size of the Hashtbl for a union region *)
    (* the pretty-printer for abstract data regions *)
    
    let print_abstract_data_region fmt abst_data_reg =
      Format.fprintf fmt "@[Abstract(@[bdd=" ;
      Predicate.print fmt (convertBddToPred abst_data_reg.bdd) ;
      Format.fprintf fmt ";@ preds= [...]" ; (* Jhala *)
(*
      Misc.list_printer_from_printer Predicate.print fmt (List.map getPred abst_data_reg.preds) ;*)
      Format.fprintf fmt "@])@]"

    (* the pretty-printer for concrete data regions *)
    let print_concrete_data_region fmt conc_data_reg =
      Format.fprintf fmt "@[Concrete(@[pred=" ;
      Predicate.print fmt conc_data_reg.pred ;
      Format.fprintf fmt "@])@]"

    (* the pretty-printer for data regions *)
    let print_data_region fmt = function
        Abstract abst_data_reg ->
          print_abstract_data_region fmt abst_data_reg
      | Concrete conc_data_reg ->
          print_concrete_data_region fmt conc_data_reg

    
 
    (* the pretty-printer for atomic regions *)
    let print_atomic_region fmt atomic_reg =
      Format.fprintf fmt "@[Atomic(@[location=" ;
      C_SD.print_location fmt atomic_reg.location ;
      Format.fprintf fmt ";@ data=" ;
      print_data_region fmt atomic_reg.data ;
      Format.fprintf fmt ";@ stack=" ;
      (match atomic_reg.stack with
	EveryStack -> Format.fprintf fmt "Everystack"
      |	_ -> Misc.list_printer_from_printer C_SD.print_location fmt (list_of_stack atomic_reg.stack));
        Format.fprintf fmt "@])@]"

    (* the pretty-printer for parallel atomic regions *)
    let print_parallel_atomic_region fmt p_atomic_reg =
      Format.fprintf fmt "@[ParallelAtomic";
      List.iter(fun ctrl ->
         Format.fprintf fmt "(@[location=" ;
         C_SD.print_location fmt ctrl.p_location ;
         Format.fprintf fmt ";@ stack=" ;         
	 (match ctrl.p_stack with
	 EveryStack -> Format.fprintf fmt "Everystack"
         | _ -> Misc.list_printer_from_printer C_SD.print_location fmt (list_of_stack ctrl.p_stack));
      ) p_atomic_reg.p_control;
      Format.fprintf fmt ";@ data=" ;
      print_data_region fmt p_atomic_reg.p_data ;
      Format.fprintf fmt "@])@]"

    (* the pretty-printer for regions *)
    let rec print fmt reg =
      match reg with
          Empty ->
            Format.fprintf fmt "@[Bottom@]"
        | Atomic atomic_reg ->
            print_atomic_region fmt atomic_reg ;
        | Union atomic_reg_table ->
	    let reglist = ref [] in
	    let tbl_to_list (l, s) d =
	      reglist := ( { location = l ; data = d ;stack = s ;})::(!reglist);
            in
              Format.fprintf fmt "@[Union " ;
              (AtomicRegionTable.iter tbl_to_list atomic_reg_table) ; (*JHALA this ensures all the reachable region is not printed...*)
              Misc.list_printer_from_printer print_atomic_region fmt (!reglist);
              Format.fprintf fmt "@]"
	| ParallelAtomic p_atomic_reg ->
	    Format.fprintf fmt "@[Parallel Atomic " ;
	    List.iter (fun i -> 
		let this_atomic = Atomic { location = i.p_location; data = p_atomic_reg.p_data; stack = i.p_stack;}  
		in print fmt this_atomic;
	    ) p_atomic_reg.p_control;
	    Format.fprintf fmt "@]"
	| ParallelUnion p_atomic_reg_table ->
	    let reglist = ref [] in
	    let tbl_to_list (c_r_list) d =
	      reglist := ( { p_control = c_r_list ; p_data = d ;})::(!reglist);
            in
              Format.fprintf fmt "@[ParallelUnion " ;
              (ParallelAtomicRegionTable.iter tbl_to_list p_atomic_reg_table) ;
              Misc.list_printer_from_printer print_parallel_atomic_region fmt (!reglist);
              Format.fprintf fmt "@]"


    (* the string converter for regions *)
    let to_string = Misc.to_string_from_printer print

			 
    (* the empty region *)
    let bot = Empty


    (* make a region precise *)
    (* This is the function written at Berkeley *)
    let concretize reg =
      Message.msg_string Message.Debug "In Region.concretize -- argument is:" ;
      Message.msg_printer Message.Debug print reg ;

      let concretize_data_region = function
          Abstract abst_data_reg ->
            Concrete { pred = convertBddToPred abst_data_reg.bdd ; }
        | Concrete conc_data_reg ->
            Concrete conc_data_reg in
      let concretize_atomic_region atomic_reg =
        { location = atomic_reg.location ;
          data = concretize_data_region atomic_reg.data ;
          stack = atomic_reg.stack ; }
      in
      let concretize_p_atomic_region p_atomic_reg =
        { p_control = p_atomic_reg.p_control;
          p_data = concretize_data_region p_atomic_reg.p_data ;}
      in
        match reg with
            Empty ->
              Empty
          | Atomic atomic_reg ->
              Atomic (concretize_atomic_region atomic_reg)
          | Union atomic_reg_table ->
              failwith "Region.concretize: argument is a union: unimplemented"
	  | ParallelAtomic p_atomic_reg ->
		ParallelAtomic (concretize_p_atomic_region p_atomic_reg)
          | ParallelUnion p_atomic_reg_table ->
              failwith "Region.concretize: argument is a parallel union: unimplemented"

		
		
    (* This is the function written by Gilles *)
    let concretize_map f_location f_stack f_predicate reg =
      Message.msg_string Message.Debug "In Region.concretize -- argument is:" ;
      Message.msg_printer Message.Debug print reg ;
      let concretize_data_region = function
	  Abstract abst_data_reg ->
            Concrete { pred = f_predicate (convertBddToPred abst_data_reg.bdd) }
        | Concrete conc_data_reg ->
            Concrete { pred = f_predicate conc_data_reg.pred } in
      let concretize_control_region p_control =
	{ p_location = f_location p_control.p_location ;
	  p_stack = f_stack p_control.p_stack;
	} 
      in
      let concretize_atomic_region atomic_reg =
        { location = f_location atomic_reg.location;
          data = concretize_data_region atomic_reg.data;
          stack = f_stack atomic_reg.stack; }
      in
      let concretize_p_atomic_region p_atomic_reg =
        { p_control = List.map concretize_control_region p_atomic_reg.p_control;
          p_data = concretize_data_region p_atomic_reg.p_data ;}
      in
        match reg with
            Empty ->
              Empty
          | Atomic atomic_reg ->
              Atomic (concretize_atomic_region atomic_reg)
          | Union atomic_reg_table ->
              failwith "Region.concretize: argument is a union: unimplemented"
	  | ParallelAtomic p_atomic_reg ->
	      ParallelAtomic (concretize_p_atomic_region p_atomic_reg)
	  | ParallelUnion _ ->
	      failwith "Region.concretize_map: parallel union not implemented"

    (* returns true if reg denotes an empty set, false otherwise *)
    let is_empty reg =
      Message.msg_string Message.Debug "In Region.is_empty -- argument is:" ;
      Message.msg_printer Message.Debug print reg ;
(*      try *)
      match reg with
          Empty ->
            true
        | Atomic atomic_reg ->
            begin
              match atomic_reg.data with
                  Abstract abst_data_reg ->
                    not (isSatisfiable (abst_data_reg.bdd, []))
                | Concrete conc_data_reg ->
                    not (isSatisfiable (bddOne, [conc_data_reg.pred]))
            end
        | ParallelAtomic p_atomic_reg ->
	    begin
	      match p_atomic_reg.p_data with 
                  Abstract abst_data_reg ->
                    not (isSatisfiable (abst_data_reg.bdd, []))
                | Concrete conc_data_reg ->
                    not (isSatisfiable (bddOne, [conc_data_reg.pred]))
            end
	| Union _ ->
            failwith "Region.is_empty: argument is a union: unimplemented"
	| ParallelUnion _ ->
            failwith "Region.is_empty: argument is a parallel union: unimplemented"
(*      with a -> (print_string (Printexc.to_string a)); failwith "isempty is the culprit" *)


    (* returns true if reg1 is included in reg2, false otherwise *)
    let rec leq reg1 reg2 =
      try 
	Message.msg_string Message.Debug "In Region.leq -- arguments are: (suppressed-RJ)" ;
	(*Message.msg_printer Message.Debug print reg1 ;
	Message.msg_printer Message.Debug print reg2 ;*) 
	
	let leq_data_region data_reg1 data_reg2 =
	  
          match (data_reg1, data_reg2) with
            (Abstract abst_data_reg1, Abstract abst_data_reg2) ->
	      if (Options.getValueOfBool "bddcov") 
	      then
		begin
		  let union_bdd = CaddieBdd.bddOr abst_data_reg1.bdd abst_data_reg2.bdd in
		  CaddieBdd.bddEqual abst_data_reg2.bdd union_bdd
		end
	      else
		let b = CaddieBdd.bddAnd abst_data_reg1.bdd (CaddieBdd.bddNot abst_data_reg2.bdd)
		in
		not (isSatisfiable (b,[]))
          | (Concrete conc_data_reg1, Concrete conc_data_reg2) ->
              let notreg2 = Predicate.negate conc_data_reg2.pred in
              let pr = Predicate.conjoinL [conc_data_reg1.pred; notreg2]
              in
              not (isSatisfiable (bddOne, [pr]))
          | (Abstract abst_data_reg, Concrete conc_data_reg) ->
	      let conc_of_abst_data_reg = concretize reg1 in
	      leq conc_of_abst_data_reg reg2
          | (Concrete conc_data_reg, Abstract abst_data_reg) ->
	      let conc_of_abst_data_reg = concretize reg2 in
	      leq reg1 conc_of_abst_data_reg

(*              failwith "Region.leq: concrete and abstract region types mixed: unimplemented" *)
	in
        match (reg1, reg2) with
          (Empty, _) ->
            true
        | (_, Empty) ->
              (* note that (Empty, Empty) is caught above *)
            false
        | (Atomic atomic_reg1, Atomic atomic_reg2) ->
            if ((atomic_reg1.location == atomic_reg2.location || atomic_reg2.location = C_SD.fake_location) &&
                (atomic_reg1.stack = atomic_reg2.stack || atomic_reg2.stack = EveryStack))
              then
              leq_data_region atomic_reg1.data atomic_reg2.data
            else
              false
	| (ParallelAtomic p_atomic_reg1, ParallelAtomic p_atomic_reg2) ->
	    if (leq_data_region p_atomic_reg1.p_data p_atomic_reg2.p_data) == true
	    then
	      begin
		List.for_all2 ( fun i j -> 
		  ( (i.p_location == j.p_location) && (i.p_stack = j.p_stack) )
		    ) p_atomic_reg1.p_control p_atomic_reg2.p_control;
	      end
	    else false
        | (Atomic atomic_reg, Union atomic_reg_table) ->
	    let location_is_kill_location =
	      let rv = 
		match (!killLocation) with 
		    None -> false
		| Some(l) -> (atomic_reg.location == l)
	      in
	      rv
	    in
	    begin
	      try
                  let retval = leq_data_region atomic_reg.data
				 (* JHALA: ARM.find *)
				 (AtomicRegionTable.find atomic_reg_table (atomic_reg.location,atomic_reg.stack))
		  in
		  let _ = if (retval && location_is_kill_location) then killCounterCovered := !killCounterCovered + 1
		  in
		    if (location_is_kill_location && (!killCounter > 1)) 
		    then 
		      (killCounterCovered_force := !killCounterCovered_force + 1; true) 
		    else retval
		with 
		    Out_of_memory -> failwith "ahaha"
		  |   Not_found -> (Message.msg_string Message.Debug "Atregtable: failed to hit table"; is_empty reg1) 
			
              end
        | (ParallelAtomic p_atomic_reg, ParallelUnion p_atomic_reg_table) ->
	    let location_is_kill_location =
	      let rv = 
		match (!killLocation) with 
		    None -> false
		  | Some(l) -> List.for_all (fun x -> x.p_location == l) p_atomic_reg.p_control 
	      in
		rv
	    in
	      begin
		try
                  let retval = leq_data_region p_atomic_reg.p_data
				 (* JHALA: ARM.find *)
				 (ParallelAtomicRegionTable.find p_atomic_reg_table p_atomic_reg.p_control)
		  in
		  let _ = if (retval && location_is_kill_location) then killCounterCovered := !killCounterCovered + 1
		  in
		    if (location_is_kill_location && (!killCounter > 1)) 
		    then 
		      (killCounterCovered_force := !killCounterCovered_force + 1; true) 
		    else retval
		with 
		    Out_of_memory -> failwith "ahaha"
		  |   Not_found -> (Message.msg_string Message.Debug "Atregtable: failed to hit table"; is_empty reg1) 
			
              end 
        | (Union _, _) ->
            failwith "Region.leq: first argument is a union: unimplemented"
        | (ParallelUnion _, _) ->
            failwith "Region.leq: first argument is a parallel union: unimplemented"
	| _ -> failwith "Region.leq: Cannot mix single and parallel regions"
	with Not_found -> failwith "out of mem here"

    (* returns true if r1 is equivalent to r2, false if not *)
    let eq reg1 reg2 = (leq reg1 reg2) && (leq reg2 reg1)


    (* union of two regions *)
    let cup reg1 reg2 =
      Message.msg_string Message.Debug "In Region.cup -- arguments are: args suppressed" ;
      (*Message.msg_printer Message.Debug print reg1 ;
      Message.msg_printer Message.Debug print reg2 ;*)

      let min_stack stack1 stack2 = 
	match stack1 with
	  EveryStack -> stack2
	| foo -> (match stack2 with EveryStack -> foo | bar -> failwith "Incomparable args to min_stack")
      in
      let cup_data_region data_reg1 data_reg2 =
        match (data_reg1, data_reg2) with
          (Abstract abst_data_reg1, Abstract abst_data_reg2) ->
            Abstract { bdd = CaddieBdd.bddOr abst_data_reg1.bdd abst_data_reg2.bdd ;
                       preds = Misc.union abst_data_reg1.preds abst_data_reg2.preds ; }
        | (Concrete conc_data_reg1, Concrete conc_data_reg2) ->
            Concrete { pred = Predicate.disjoinL [conc_data_reg1.pred;
                                                   conc_data_reg2.pred] ; }
        | (Abstract abst_data_reg, Concrete conc_data_reg)
        | (Concrete conc_data_reg, Abstract abst_data_reg) ->
            failwith "Region.cup: concrete and abstract region types mixed: unimplemented"
      in
      match (reg1, reg2) with
        (Empty, _) ->
          reg2
      | (_, Empty) ->
          reg1
      | (Atomic atomic_reg1, Atomic atomic_reg2) ->
          if (atomic_reg1.location == atomic_reg2.location)
          then
	    begin
	      if (atomic_reg1.stack = atomic_reg2.stack)
	      then Atomic { location = atomic_reg1.location ;
			    data = cup_data_region atomic_reg1.data atomic_reg2.data ;
			    stack = atomic_reg1.stack; (* same as the other stack *)}
	      else 
		if (atomic_reg1.stack = EveryStack || atomic_reg1.stack = EveryStack)
		then Atomic { location = atomic_reg1.location ;
			      data = cup_data_region atomic_reg1.data atomic_reg2.data ;
                              stack = min_stack atomic_reg1.stack atomic_reg2.stack; }
		    (* This should not be EveryStack as there is a 
		       connection between the data and stack -- 
		       so for correctness, it should be the min of the two stacks *)
		else
		  begin
		    let newTable = (AtomicRegionTable.create atomic_region_parameter) in
		    AtomicRegionTable.add newTable (atomic_reg1.location, atomic_reg1.stack) atomic_reg1.data;
		    AtomicRegionTable.add newTable (atomic_reg2.location, atomic_reg2.stack) atomic_reg2.data;
		    Union newTable
		  end
		    
	    end
	  else 
	    begin
	      let newTable = (AtomicRegionTable.create atomic_region_parameter) in
	      AtomicRegionTable.add newTable (atomic_reg1.location, atomic_reg1.stack) atomic_reg1.data;
	      AtomicRegionTable.add newTable (atomic_reg2.location, atomic_reg2.stack) atomic_reg2.data;
	      Union newTable
	    end
      | (ParallelAtomic p_atomic_reg1, ParallelAtomic p_atomic_reg2) ->
          let flag_equal_locations = ref true in
	  let flag_equal_stacks = ref true in
	  begin
	  List.iter2( fun ctr1 ctr2 ->
		if (ctr1.p_location <> ctr2.p_location)
		then flag_equal_locations := false;
		
		if (ctr1.p_stack <> ctr2.p_stack)
		then flag_equal_stacks := false;

	  ) p_atomic_reg1.p_control p_atomic_reg2.p_control;

	  if ( !flag_equal_locations = true && !flag_equal_stacks = true) (* we can just return a ParallelAtomicRegion *)
	  then 
	  begin
		ParallelAtomic { p_control = p_atomic_reg1.p_control; (* both are same *)
			   p_data    = cup_data_region p_atomic_reg1.p_data p_atomic_reg2.p_data; }
	  end
	  else
	  begin
		if ( !flag_equal_locations = true && !flag_equal_stacks = false)
		then
		begin
		let return_parallel_atomic = 
			{ 
				p_control = List.map2( 
				fun ctr_1 ctr_2 ->
					{ p_location = ctr_1.p_location; (* both are same *)
					  p_stack    = min_stack ctr_1.p_stack ctr_2.p_stack; 
					}
				) p_atomic_reg1.p_control p_atomic_reg2.p_control;
				p_data    = cup_data_region p_atomic_reg1.p_data p_atomic_reg2.p_data;
			}
		in
		ParallelAtomic return_parallel_atomic
		end
		else
		begin
		  let newTable = (ParallelAtomicRegionTable.create atomic_region_parameter) in
		  ParallelAtomicRegionTable.add newTable p_atomic_reg1.p_control p_atomic_reg1.p_data;
	          ParallelAtomicRegionTable.add newTable p_atomic_reg2.p_control p_atomic_reg2.p_data;
	          ParallelUnion newTable			
		end
	  end
	end
      | (Atomic atomic_reg, Union atomic_reg_table)
      | (Union atomic_reg_table, Atomic atomic_reg) ->
          begin
            try (* JHALA: ARM.find *)
              let data_reg_in_table = AtomicRegionTable.find atomic_reg_table (atomic_reg.location,atomic_reg.stack)
              in
              begin
		(AtomicRegionTable.replace atomic_reg_table (atomic_reg.location, atomic_reg.stack)
                   (cup_data_region data_reg_in_table atomic_reg.data));
		Union atomic_reg_table
	      end
            with Not_found ->
              begin
		(AtomicRegionTable.add atomic_reg_table (atomic_reg.location, atomic_reg.stack) atomic_reg.data);
		Union atomic_reg_table
	      end
          end
      | (ParallelAtomic p_atomic_reg, ParallelUnion p_atomic_reg_table)
      | (ParallelUnion p_atomic_reg_table, ParallelAtomic p_atomic_reg) ->
          begin
            try (* JHALA: ARM.find *)
              let data_reg_in_table = ParallelAtomicRegionTable.find p_atomic_reg_table p_atomic_reg.p_control
              in
              begin
		(ParallelAtomicRegionTable.replace p_atomic_reg_table p_atomic_reg.p_control
                   (cup_data_region data_reg_in_table p_atomic_reg.p_data));
		ParallelUnion p_atomic_reg_table
	      end
            with Not_found ->
              begin
		(ParallelAtomicRegionTable.add p_atomic_reg_table (p_atomic_reg.p_control) p_atomic_reg.p_data);
		ParallelUnion p_atomic_reg_table
	      end
          end
      | (Union _, Union _) ->
          failwith "Region.cup: both arguments are unions: unimplemented"
      | (ParallelUnion _, ParallelUnion _) ->
          failwith "Region.cup: both arguments are parallel unions: unimplemented"
      | ( (ParallelAtomic _ | ParallelUnion _), (Atomic _ | Union _ ))
      | ( (Atomic _ | Union _ ), (ParallelAtomic _ | ParallelUnion _)) ->
	  failwith "Region.cup: single and parallel regions cannot be mixed" 
   
    exception EmptyStack
	
    (* intersection of two regions *)
    let cap reg1 reg2 =
      Message.msg_string Message.Debug "In Region.cap -- arguments are:" ;
      Message.msg_printer Message.Debug print reg1 ;
      Message.msg_printer Message.Debug print reg2 ;

      let cap_data_region data_reg1 data_reg2 =
        match (data_reg1, data_reg2) with
            (Abstract abst_data_reg1, Abstract abst_data_reg2) ->
              Abstract { bdd = CaddieBdd.bddAnd abst_data_reg1.bdd abst_data_reg2.bdd ;
                         preds = Misc.union abst_data_reg1.preds abst_data_reg2.preds ; }
          | (Concrete conc_data_reg1, Concrete conc_data_reg2) ->
              Concrete { pred = Predicate.conjoinL [conc_data_reg1.pred;
                                                    conc_data_reg2.pred] ; }
          | (Abstract abst_data_reg, Concrete conc_data_reg)
          | (Concrete conc_data_reg, Abstract abst_data_reg) ->
              Concrete { pred = Predicate.conjoinL [conc_data_reg.pred;
                                                    convertBddToPred abst_data_reg.bdd] ; }
      in
      let stack_cap stack1 stack2 = 
	match stack1 with 
	  CallStack s -> if (stack2 = stack1 || stack2 = EveryStack) then stack1 else raise EmptyStack
	| EveryStack -> stack2
      in
      let location_cap loc1 loc2 = 
	(* This sometimes misbehaves: by raising Out_of_memory.
	   I suppose this has to do with circular structures.
	   Therefore, we check location equality based on location ids.
	if (loc1 = C_SD.fake_location) then loc2 
	else if (loc2 = C_SD.fake_location) then loc1
	else if (loc1 = loc2) then loc1 
	else raise EmptyStack
	   *)
	let (i1, i2) = C_SD.location_coords loc1 and
	    (j1, j2) = C_SD.location_coords loc2 in
	let (fake1, fake2) = C_SD.location_coords C_SD.fake_location in
	if ((i1 = fake1) && (i2 = fake2)) then loc2
	else if ((j1 = fake1) && (j2 = fake2)) then loc1 
	else if ((i1 = j1) && (i2 = j2)) then loc1 
        else raise EmptyStack
      in
        match (reg1, reg2) with
            (Empty, _) ->
              Empty
          | (_, Empty) ->
              Empty
          | (Atomic atomic_reg1, Atomic atomic_reg2) ->
              begin
		try 
                  Atomic { location = (location_cap atomic_reg1.location atomic_reg2.location);
                           data = (cap_data_region atomic_reg1.data atomic_reg2.data);
                           stack = (stack_cap atomic_reg1.stack atomic_reg2.stack); 
			   }
		with EmptyStack -> Empty
	      end
	  | (ParallelAtomic p_atomic_reg1, ParallelAtomic p_atomic_reg2) ->
              (* add fake location stuff  here : rupak *)
	      if ((List.for_all2 (fun x y -> x.p_location = y.p_location) p_atomic_reg1.p_control p_atomic_reg2.p_control))
              then
		begin
		try 
                  ParallelAtomic { p_control = List.map2 (fun ctr1 ctr2 ->
						{ p_location = ctr1.p_location ; (* both same *)
						  p_stack = stack_cap ctr1.p_stack ctr2.p_stack;}
					) p_atomic_reg1.p_control p_atomic_reg2.p_control;  
				   p_data = cap_data_region p_atomic_reg1.p_data p_atomic_reg2.p_data ;
                           }
		with EmptyStack -> Empty
		end
              else
	      begin
                Empty
	      end
	  | (Atomic atomic_reg, Union atomic_reg_table)
          | (Union atomic_reg_table, Atomic atomic_reg) ->
              begin
                try (* JHALA : ARM.find *)
                  let data_reg_in_table = AtomicRegionTable.find atomic_reg_table (atomic_reg.location,atomic_reg.stack)
                  in
                    Atomic { location = atomic_reg.location ;
                             data = cap_data_region data_reg_in_table atomic_reg.data ;
                             stack = atomic_reg.stack ; }
                with Not_found ->
                  Empty
              end
          | (ParallelAtomic p_atomic_reg, ParallelUnion p_atomic_reg_table)
          | (ParallelUnion p_atomic_reg_table, ParallelAtomic p_atomic_reg) ->
              begin
                try (* JHALA : ARM.find *)
                  let data_reg_in_table = ParallelAtomicRegionTable.find p_atomic_reg_table p_atomic_reg.p_control
                  in
                    ParallelAtomic { p_control = p_atomic_reg.p_control ;
				     p_data = cap_data_region data_reg_in_table p_atomic_reg.p_data ; }
                with Not_found ->
                  Empty
              end
          | (Union _, Union _) ->
              failwith "Region.cap: both arguments are unions: unimplemented"
          | (ParallelUnion _, ParallelUnion _) ->
              failwith "Region.cap: both arguments are unions: unimplemented"
          | ( (ParallelAtomic _ | ParallelUnion _), (Atomic _ | Union _ ))
          | ( (Atomic _ | Union _ ), (ParallelAtomic _ | ParallelUnion _)) ->
	      failwith "Region.cap: single and parallel regions cannot be mixed"

    let rename_vars r new_name = 
      let do_predicate p = Predicate.alpha_convert new_name p 
      in
      let do_atomic_region = function
	| Concrete c ->
            Concrete {pred = do_predicate c.pred}
	| Abstract a ->
            Concrete ({pred = do_predicate (convertBddToPred a.bdd)})
      in
      match r with
      | Empty -> Empty
      | Atomic ar ->
          Atomic {ar with data = do_atomic_region ar.data}
      | Union h as r ->
          AtomicRegionTable.iter (fun k data -> AtomicRegionTable.replace h k (do_atomic_region data)) h;
          r	   
      |	ParallelAtomic par ->
	  ParallelAtomic {par with p_data = do_atomic_region par.p_data}
      |	ParallelUnion p ->
	  ParallelAtomicRegionTable.iter (fun k data -> ParallelAtomicRegionTable.replace p k (do_atomic_region data)) p;
	  r


	    (* region to predicate takes a region and returns a predicate corresponding to the data region
	       of the region 
	       *)
    let region_to_predicate r = 
      let do_atomic_region = function
	| Concrete c -> c.pred
	| Abstract a -> convertBddToPred a.bdd
      in
      match r with
      | Empty -> Predicate.False
      | Atomic ar -> do_atomic_region ar.data
      | ParallelAtomic par -> do_atomic_region par.p_data
      | Union atomic_region_table -> 
	  begin
	    AtomicRegionTable.fold
	      ( fun control data_region curr_pred -> Predicate.disjoinL [ curr_pred ; (do_atomic_region data_region) ] )
	      atomic_region_table 
	      Predicate.False
	  end
      | ParallelUnion p_atomic_region_table -> 
	  begin
	    ParallelAtomicRegionTable.fold
	      ( fun control data_region curr_pred -> Predicate.disjoinL [ curr_pred ; (do_atomic_region data_region) ] )
	      p_atomic_region_table 
	      Predicate.False
	  end

    let change_location_and_stack r1 r2 =
      match (r1, r2) with 
	(Empty, _) -> Empty
      | (Atomic a1, Empty) -> Atomic a1
      | (Atomic a1, Atomic a2) -> Atomic { a1 with stack = a2.stack;
					     location = a2.location; }
      | (_,_) -> failwith "change_location_and_stack: unimplemented"

	
    (* Our coverage lookup structure will map pc's to regions that have them.
     * This will only work with atomic regions!!
     * A faster method would be nice if we can come up with one. *)
    module LocationTbl = Hashtbl.Make(struct
      type t = C_SD.location

      let equal l1 l2 = C_SD.location_to_string l1 = C_SD.location_to_string l2

      let hash l = Hashtbl.hash (C_SD.location_to_string l)
    end)
	   
    type 'a coverers = (t * 'a) list LocationTbl.t

    exception Empty_region
    exception Early_stop

    let getPc = function
	Atomic a -> a.location
      |	Empty -> raise Empty_region
      |	Union _ -> failwith "Atomic region expected in region coverage code; got Union"
      |	ParallelAtomic _ -> failwith "Atomic region expected in region coverage code; got ParallelAtomic"
      |	ParallelUnion _ -> failwith "Atomic region expected in region coverage code; got ParallelUnion"

    let emptyCoverers () = LocationTbl.create 100

    let clearCoverers = LocationTbl.clear

    let getAnyFromCoverer cov =
      let reg = ref None in
      try
	LocationTbl.fold (fun _ l _ ->
	  match l with
	    (_, data)::_ -> reg := Some data; raise Early_stop
	  | _ -> ()) cov ();
	None
      with Early_stop -> !reg

    let addCoverer cov reg data =
      try
	let pc = getPc reg in
	try
	  let current = LocationTbl.find cov pc in
	  LocationTbl.add cov pc ((reg, data)::current)
	with Not_found -> LocationTbl.add cov pc [(reg, data)]
      with Empty_region -> ()

    let findCoverer cov reg =
      try
	let pc = getPc reg in
	try
	  let current = LocationTbl.find cov pc in
	  let (_, data) = List.find (fun (reg', _) -> leq reg reg') current in
	  Some data
	with Not_found -> None
      with Empty_region -> getAnyFromCoverer cov

    type restriction = unit
    let negate_restriction _ = ()
    let only_locals _ = ()
    let only_globals = ()
    let locals_and_globals _ = ()
    let restrict () r = r
  end

(*
  module AR = AccessRegion.Make(Region)
  type 'a phi_table = (Ast.Symbol.symbol, 'a AR.t ref) Hashtbl.t 
*)
  type phi_table = (Ast.Expression.lval, Region.t ref) Hashtbl.t 

  let assumed_invariants_region = 
    ref (Region.Atomic { Region.location = C_SD.fake_location ;
			 Region.data = Region.Concrete { Region.pred = Predicate.True; } ;
			 Region.stack = Region.EveryStack ; })

  let assumed_invariants_list_to_region () =
    let data = Region.Concrete { Region.pred = Predicate.conjoinL !assumed_invariants ; } in
       Region.Atomic { Region.location = C_SD.fake_location ;
                       Region.data = data ;
                       Region.stack = Region.EveryStack ; }

  (**************************************************************************)

  (**
   * This module defines the operations of this abstraction.
   *)
  module Operation =
  struct
    type t = C_SD.edge
    let print = C_SD.print_edge
    let to_string = Misc.to_string_from_printer print

    type exp = Ast.Expression.expression
    type lval = Ast.Expression.lval

    type info =
      | Normal
      | Call of lval option * string * exp list
      | Ret of exp

    open BlastCSystemDescr.C_Command
    let get_info edge =
      match (C_SD.edge_to_command edge).code with
	Block [Return e] -> Ret e
      |	FunctionCall e ->
	  (match e with
	    Expression.FunctionCall (Expression.Lval (Expression.Symbol n), el) ->
	      Call (None, n, el)
	  | Expression.Assignment (Expression.Assign, targ, Expression.FunctionCall
				     (Expression.Lval (Expression.Symbol n), el)) ->
				       Call (Some targ, n, el)
	  | _ -> raise (Failure "Function call edge has non-call expression"))
      |	_ -> Normal
  end

  (**************************************************************************)
  (******** Functions required by the GUI for reading data structures *******)
  (**************************************************************************)

let stack_to_string stack = 
  match stack with
      Region.EveryStack -> "[_]"
    | Region.CallStack stack_list -> 
	begin
	  let stack_string = 
	    List.fold_left 
	      (fun a -> fun b -> (a^(C_SD.get_location_fname b)^";\n"))
	      "" stack_list
	  in
	    "[ "^stack_string^" ]"
	end
  
let data_to_string_list data =
  let pred_is_atomic p = 
    match p with
      Predicate.Atom _ -> true
    | Predicate.Not (Predicate.Atom _) -> true
    | Predicate.True | Predicate.False -> true 
    | _ -> false
  in
    match data with
	Region.Concrete p ->
	  begin
	    match p.Region.pred with
		Predicate.And l -> 
		  let _ = 
		    if (List.exists (fun x -> not (pred_is_atomic x)) l) 
		    then ( 
		      Message.msg_string 
			Message.Normal 
			("Warning: region has non-atomic conjuncts "^(Predicate.toString p.Region.pred)))
		  in
		    List.map Predicate.toString l
	      | Predicate.True -> []
	      |	Predicate.Atom a -> [ Predicate.toString p.Region.pred ]
	      |	Predicate.Not (Predicate.Atom _) -> [ Predicate.toString p.Region.pred ]
	      | _ -> failwith ("Bad trace -- pred not in And-l format:"^(Predicate.toString p.Region.pred))
	end
    | Region.Abstract _ -> failwith "Bad trace -- abstract data region"


let reg_to_strings reg = 
  match reg with
      Region.Union _ -> failwith "Bad trace -- union region"
    | Region.Atomic r -> 
	begin
	  let loc = r.Region.location in
	  let data = r.Region.data in
	  let stack = r.Region.stack in
	   (data_to_string_list data,stack_to_string stack)
	end
    | Region.ParallelAtomic _
    | Region.ParallelUnion _ 
    | Region.Empty ->
	failwith "reg_to_strings: bad region"
  
let reg_function_name reg = 
  match reg with
      Region.Union _ -> failwith "Bad trace -- union region"
    | Region.Atomic r -> 
	begin
	  let loc = r.Region.location in
	  C_SD.get_location_fname loc
	end
    | Region.ParallelAtomic _
    | Region.ParallelUnion _ 
    | Region.Empty ->
	failwith "reg_function_name: bad region"

let reg_to_string reg = 
  match reg with
      Region.Union _ -> failwith "Bad trace -- union region"
    | Region.Atomic r -> 
	begin
	  let loc = r.Region.location in
	  C_SD.location_to_string loc
	end
    | Region.ParallelAtomic _
    | Region.ParallelUnion _ 
    | Region.Empty ->
	failwith "reg_function_name: bad region"

let loc_to_string loc =
  match C_SD.get_source_position loc with
      Some(file,line,col) -> (string_of_int line^":")
    | None -> ""

let loc_to_int loc =
  match C_SD.get_source_position loc with
      Some(file,line,col) -> line (* (string_of_int line^":")*)
    | None -> -1

let op_to_string op =
  let s_loc = C_SD.get_source op in
  let x = C_SD.get_command op in
  let s1 = C_SD.get_target op in
    (loc_to_int s_loc, loc_to_int s1, (Printf.sprintf "%s \t %s" (loc_to_string s_loc) (C_SD.Command.to_string x)))

  (**********************End GUI functions *********************************)
  (*************************************************************************)


  (**************************************************************************)

  (**
   * This modules is the abstract interface to the model checker for C System
   * Descriptions.  It uses Region module as abstract region, and implements
   * C Command operations on regions.
   *)


  (** Code to generate new temporary variable names **)
  let newTmpVarIndex = ref (-1)
  let getNextTmpVarIndex () =
    newTmpVarIndex := !newTmpVarIndex +1; !newTmpVarIndex
	
  (* Check if a symbol is a temporary *)
  (* a string is temporary if it starts with a number (obtained from getNewTmpVar)
     or contains the character '%' (obtained from brand_new_suffix)
     *)
  let isTemporary sym =
    if (Misc.is_digit (String.get sym 0) || String.contains sym '%') then
      true
    else
      false

    
  (* This is used by the strongest postcondition to give names
     to the existentiallly quantified variables.
     This is also used by the wp when substitutions are explicit.
  *)
	 (* jhala -- making this go inside recursively -- ask rupak if this is ok *)
  let getNewTmpVar (t : Expression.lval) : Expression.expression =
    let _makeStr s = (string_of_int (getNextTmpVarIndex ()))^"_"^s 
    in
(*
    let rec _convertLval exp =
      match exp with
      |	Expression.Symbol s -> Expression.Symbol (_makeStr s)
      |	Expression.Dereference e -> 
	  Expression.Dereference (_convertExp e)
      |	Expression.Access (op, e1, fld) -> Expression.Access(op, _convertExp e1, fld)
      |	Expression.Indexed (e1, e2) -> Expression.Indexed (_convertExp e1, _convertExp e2)
    and
	_convertExp exp =
	match exp with
      |	Expression.Lval l -> Expression.Lval (_convertLval l)
      |	Expression.Binary (Expression.Plus, e, ind) -> Expression.Binary(Expression.Plus, _convertExp e, ind)
      |	_ -> failwith ("getNewTmpVar : Strange exp :"^(Expression.toString exp))
    in
    try _convertLval t with
      _ -> failwith ("getNewTmpVar : Strange expression "^(Expression.lvalToString t))
  *)
    Expression.Chlval (t, _makeStr "")


  let get_new_unknown () = 
    getNewTmpVar (Expression.Symbol "BLAST_UNKNOWN")

  (***************************************************************************)
  (*         weakest preconditions and strongest postconditions              *)
  (***************************************************************************)

  (****************************************************************************************)
  (* Code to handle the aliasing between variables in the wp and sp computations. *)

  (* Code to manage hashtable mapping each predicate to its list of variables of the form
     x, x.f, *x, &x, and ( *x ).f *)
  let predTbl = Hashtbl.create 11
  let predTbl_deep = Hashtbl.create 11



  (* get the set of variable expressions in a predicate.  we ask this both for a predicate and its
     negation, so we have to deal with both cases 
     The returned list contains either lvalues or Unary(AddressOf, lvalue).
     *)
  let getVarExps pred  = 
    try Hashtbl.find predTbl pred
    with Not_found -> 
      (try Hashtbl.find predTbl (Predicate.negate pred)
       with Not_found -> 
         let f = Predicate.allVarExps pred
         in
           Hashtbl.add predTbl pred f;
           f)

  let getVarExps_deep pred  = 
      try Hashtbl.find predTbl_deep pred
      with Not_found -> 
	(try Hashtbl.find predTbl_deep (Predicate.negate pred)
	with Not_found -> 
	  let f = Predicate.allVarExps_deep pred in
	  let f' = List.filter 
		     (fun x -> 
			match x with
			    (Expression.Symbol _) -> true 
			  | _ -> false
		     ) f 
	  in
	  let f''= List.map (fun x ->
			       match x with
				  (Expression.Symbol y) -> y 
				 | _ -> failwith "should only receive symbols here !"
			    ) f'
	  in
	  Hashtbl.add predTbl_deep pred f'';
	  f'')	   

  (* like getVarExps, but it filters out the expressions of the form &x.  Such expressions are
     unnecessary for alias queries, since they are never aliased to anything.  *)
  let getVarsAndDerefs (exp : Predicate.predicate) : Expression.expression list =
    let filter_addrofs = 
      List.filter 
	(function e -> 
          match e with
            Expression.Unary(Expression.Address, _) -> false
          |        _ -> true)
	(getVarExps exp)
    in
    List.map 
      (fun e -> 
	match e with
	  Expression.Lval _ -> e
	| Expression.Chlval (_,_) -> e
	| _ -> failwith ("getVarsAndDerefs: Non lval expression! "^(Expression.toString e)))
      filter_addrofs
      
(* a cheap hack: I want the constituent varsandderefs of an expression -- but dont want to rewrite the function above so... *)
  let getVarsAndDerefs_exp (e : Expression.expression) : Expression.expression list =
    let _pred = Predicate.Atom (Expression.Binary (Expression.Eq, e, Expression.Constant (Constant.Int 1))) 
    in
      getVarsAndDerefs _pred 

  (* this function returns a list of expressions (of the form x, *x, x.f, or ( *x ).f) in 
     pred that may be aliased to target. 
     we assume that the given C program 
     is in a normal form such that we only need to do deal with one level of 
     pointers.  for example, a statement of the form **p = 3 will be 
     represented as (tmp = *p; *tmp = 3). *)

	(* this table caches results to getPossibleAliases *)
  let queryAlias_table = Hashtbl.create 1009
	
  let queryAlias_cache x y =
    try Hashtbl.find queryAlias_table (x,y)
    with Not_found ->
      begin
	let rv = AliasAnalyzer.queryAlias x y in
	  Hashtbl.add queryAlias_table (x,y) rv;
	  rv
      end
      
  let getPossibleAliases target pred =
    if (Options.getValueOfString "alias" = "") then []
    else 
      begin
	let predVars = getVarsAndDerefs pred in
	  (* we only need to check alias pairs where at least one is a deref or a field access of a deref *)
	let candidates : Expression.expression list =
	  if Expression.isDeref (Expression.Lval target) then predVars
	  else List.filter Expression.isDeref predVars in
	  (* remove any candidates that are definitely not aliased to target.  
	     such candidates are determined by consulting the results of a pointer 
	     analysis. *)
	  List.filter (queryAlias_cache (Expression.Lval target)) candidates
	    (* JHALA: this should not return ANY chlvals.... *)   
      end
        

  (**********************************************************************************)

  (************************************************************************************)

  (* Jhala: An evil hack to solve the problem of who got the retval in a funcall ... *)
      
  (* retTarget always points to the last target -- it is set when doing a 
     post from the stack info of the region 
     [Rupak:] Changing type of junkRet to Expression.lval
     *)
  let junkRet = Expression.Dereference (Expression.Sizeof 0)

  let __NotImplementedFunctionName = "NOT_IMPLEMENTED_FUNCTION" (* is there an #define for ocaml ? *)
  let __BLAST_DummyFunctionName = "__BLAST_DUMMY_FUNCTION"
				    
  let retTarget = ref junkRet
  let wpCapRegion = ref Predicate.True


  let sat_filter cubes_disjl cap_reg =
    let print_info_flag = (cap_reg <> Predicate.True) in
    let _ = if print_info_flag then Message.msg_string Message.Debug ("Cap Region: "^(Predicate.toString cap_reg)) in
    let _candidate_terms = Predicate.to_dnf (cubes_disjl) in
    let is_term_satisfiable term = 
      if (print_info_flag) 
      then
	begin
	  Printf.printf (".");
	  not ((Options.getValueOfBool "wpsat") && (askTheoremProver (Predicate.implies (Predicate.conjoinL [term;!wpCapRegion]) Predicate.False)))
	end
      else
	true
    in
    let _surviving_terms = List.filter is_term_satisfiable _candidate_terms in
    if (true) then
      begin
	Message.msg_string Message.Debug ("Original cubes: "^(string_of_int (List.length _candidate_terms)));
	Message.msg_string Message.Debug ("Remaining cubes: "^(string_of_int (List.length _surviving_terms)))
      end;
    Predicate.disjoinL _surviving_terms



  (* compute WP for expressions involving pointers.  in particular, we have
     to consider aliasing if we are given an assignment statement.
     otherwise, we do the standard WP computation. 
     possibleAliases is the set of alias pairs that need to be considered for an
     assignment statement *)
    
 (* It does explicit substitutions when the flag is true.
    *)
(* RJ: my head hangs in shame re: the compress-flag *)
let compress_flag = ref false
(* note that when wp is called with compress_flag set to true, it must be that
   explicit_flag is also true *)
(* this is really only to be called when doing the interpolant thing *)

(* pred_l is a list of predicates -- the first of which is the one where we will add stuff 
while the others correspond to constraints further down the trace and so we merely substitute
and so on in them. It returns a pred_list as well ... *)
let make_exp_lval e = 
  match e with
      Expression.Lval l -> l
    | _ -> failwith ("bad call to make_exp_lval: "^(Expression.toString e))

let lval_map_table = Hashtbl.create 1009

(* VVVVIMP : TBD make this trace check faster ! *)
let trace_lv_table = Hashtbl.create 109  (* this is actually a table of expressions *)

(* lval list -> unit *)
let add_to_lv_table candidates = 
   (*let candidates = getVarsAndDerefs pred in*)
   Message.msg_string Message.Debug "Adding to lv_table:";
     List.iter (fun x -> 
		  Message.msg_string Message.Debug (Expression.toString x);
		  Hashtbl.replace trace_lv_table x true) candidates
       


let lv_of_expr e =
  match e with
      Expression.Lval l -> l 
    | _ -> failwith ("bad arg to lv_of_expr"^(Expression.toString e))
	
let get_lval_aliases lval = (* JHALA: VVV IMP ! TBD *)
  if (Options.getValueOfString "alias" = "" || Options.getValueOfBool "incref")  then [lval]
  else
    begin
      (* first a rather slow way to do things *)
      let aliases =
	List.filter 
	  (fun x -> AliasAnalyzer.queryAlias (Expression.Lval lval) x) 
	  (Misc.hashtbl_keys trace_lv_table)
      in
	lval::(List.map lv_of_expr aliases)
    end

(* we assume that this function is called in a specific order because it changes and uses the lval_map_table *)
    
(* this returns an Expression.expression *)
let update_map lval = 
  let new_lval = getNewTmpVar lval in
    Message.msg_string Message.Debug (Printf.sprintf "update_map: %s -> %s" (Expression.lvalToString lval) (Expression.toString new_lval));
    Hashtbl.add lval_map_table lval new_lval
	
let _lval_map_a map_table lval = 
  try
    Hashtbl.find map_table lval
  with Not_found -> 
    begin (* we need this to be deterministic *)
      let new_lval = Expression.Chlval (lval,"0") (* getNewTmpVar lval*) in
	Hashtbl.add map_table lval new_lval;
	new_lval
    end

let lval_map_a map_table e = 
  Expression.deep_alpha_convert (_lval_map_a map_table) e

let pred_lval_map_a map_table pred =
  Predicate.deep_alpha_convert (_lval_map_a map_table) pred
	
let _lval_map = _lval_map_a lval_map_table 
(*
  lval  =
  try
  Hashtbl.find lval_map_table lval
  with Not_found -> 
  begin
  let new_lval = getNewTmpVar lval in
  Hashtbl.add lval_map_table lval new_lval;
  new_lval
  end
*)

let lval_map = lval_map_a lval_map_table 
(* Expression.deep_alpha_convert _lval_map exp *)
    
let pred_lval_map = pred_lval_map_a lval_map_table
(* Predicate.deep_alpha_convert _lval_map pred *)

(* let copy_lval_map_table () = Hashtbl.copy lval_map_table *)


let ite (cond,t1,t2) = 
  Predicate.disjoinL 
    [ Predicate.conjoinL [ cond;t1] ; 
      Predicate.conjoinL [Predicate.Not cond; t2]]

(* unlike in the paper, this requires the "chlvals" to be supplied *)
let equate elist e'list =
  (* requires that the lengths of the lists are the same *)
  let ee'list = List.combine elist e'list in
    Predicate.conjoinL
      (List.map 
	 (fun (e,e') -> Predicate.Atom (Expression.Binary(Expression.Eq,e,e')))
	 ee'list)

(* RJ: THERE CAN BE BUGS WITH THE LVAL_MAP INTERFACE !! PLEASE CHECK IF YOU GET WIERD ANSWERS THAT THE CONSTRAINTS GENERATED 
ARE GOOD ! -- better still PROVE this interface correct. *)

(* instead of pred_l we shall pass this just the "present" block *)
let rec block_wp e pred =
  let _ = assert (pred = Predicate.True) in
    match e with
	C_SD.Command.Expr (Expression.Assignment(Expression.Assign, target, valu)) ->
	  begin
	    (* i'm doing this again for no good reason *)
	    let target_closure = C_SD.expression_closure (Expression.Lval target) in
	    let valu_closure = 
	      List.map 
		(
		  fun x -> 
		    let vx = 
		      (Expression.substituteExpression [(Expression.Lval target, valu)] x)
		    in 
		      try Expression.Lval (Expression.push_deref (lv_of_expr vx))
		      with _ -> vx
		)
		target_closure
	    in
	      (* shouldnt really need to do this latter operation as its subsumed by the one later but ... *)
	    (* done up front: let _ = add_to_lv_table (List.flatten (List.map getVarsAndDerefs_exp (target_closure@valu_closure))) in *)
	    let possibleAliases = get_lval_aliases target in
	    let _ = 
	      Printf.printf "blk: assign lhs: %s     assign rhs: %s \n" 
		(Expression.lvalToString target) (Expression.toString valu);
	      Printf.printf "Possible aliases = %s\n\n" 
		(List.fold_left (fun a -> fun b -> a^(Expression.toString (Expression.Lval b))) "" possibleAliases)
	    in
	    let lv_triple_list = 
	      let make_triple lv = 
		let eclos = C_SD.expression_closure (Expression.Lval lv) in
		  (lv,eclos,List.map lval_map eclos)
	      in
		List.map make_triple possibleAliases
	    in
	      (* implicit inv that the possibleAliases have the same closure set as the target *)
	    (* now updating the map so that all the things that might possibly have changed have new values for 
 	       "before" the operation *)
	    let relevant_lvals = List.flatten (List.map Misc.snd3 lv_triple_list) in
	       (*let old_map = copy_lval_map_table () in*)	
	    let _ = List.iter (fun x -> update_map (lv_of_expr x)) relevant_lvals in
	       (* after this we are using the updated lval_map *)
	    (* done up front: let _ = add_to_lv_table relevant_lvals in *)

	    let valu_closure_vals = List.map lval_map valu_closure in 
	      (* note: the above is done AFTER updating the map *)
	      (* implicit invariant, the closures come out in the right order ... *)
	    let target' = lval_map (Expression.Lval target) in
	    let target_addr = Expression.addressOf target' in

	    let process_alias (lv,lv_closure,new_vals) = 
	      (*let new_vals = List.map (lval_map_a old_map) lv_closure in *)	      
	      if lv = target then
		equate new_vals valu_closure_vals
	      else
		begin
		  let old_vals = List.map lval_map lv_closure in
		  let eq_new = equate new_vals valu_closure_vals in
		  let eq_old = equate new_vals old_vals in
		  let lv_addr = lval_map (Expression.addressOf (Expression.Lval lv)) in
		  let cond = 
		    Predicate.Atom (Expression.Binary(Expression.Eq,target_addr,lv_addr))
		  in
		    ite (cond,eq_new,eq_old)
		end
	    in
	    let new_cons = 
	      List.map process_alias lv_triple_list 
	    in 
	      Message.msg_string Message.Debug ("bwp: "^(Predicate.toString (Predicate.conjoinL new_cons)));
	      Predicate.conjoinL (new_cons@[pred])
end
      | C_SD.Command.Return exp ->
          Message.msg_string Message.Debug "WP of return";
	  if ((!retTarget) <> junkRet) then
            begin
              Message.msg_printer 
		Message.Debug C_SD.Command.print 
		{ C_SD.Command.code = C_SD.Command.Block 
					[(C_SD.Command.Expr(Expression.Assignment(Expression.Assign, !retTarget, exp)))];C_SD.Command.read = []; (* printing ignores .read and .written *)
C_SD.Command.written = []}; (* ditto *)
            (block_wp (C_SD.Command.Expr(Expression.Assignment(Expression.Assign, !retTarget, exp))) pred)
          end
        else pred

    | _ -> failwith("bad call to compressed_wp!")
	

let rec wp explicit_flag sat_flag stmt pred (edge,i) =
  (*
    if (not explicit_flag) then Message.msg_string Message.Normal ("Computing wp of predicate "^(Predicate.toString pred)) ;
  *)
  match stmt with
      C_SD.Command.Expr e->
        begin
          match e with
	      (* Jhala : am restructuring the wp code *)
              Expression.Assignment(Expression.Assign, target, valu) -> (* operation target = value *)
		let assigned_value = 
		  if explicit_flag then (* EXPLICIT *)
		    begin
		      Message.msg_string Message.Debug "explicit wp \n";
		      let newvar = (getNewTmpVar target) in
			newvar
		    end
		  else (* IMPLICIT *)
		    begin
		      Message.msg_string Message.Debug "implicit wp \n";
		      valu
		    end
		in
		let _print_info_flag = (!wpCapRegion = Predicate.True) in
		let possibleAliases = Stats.time "getPossibleAliases" (getPossibleAliases target) pred in
		  if (_print_info_flag) then
		    begin
		      Message.msg_string Message.Debug (Printf.sprintf "assign lhs: %s     assign rhs: %s \n" (Expression.lvalToString target) (Expression.toString valu));
		      Message.msg_string Message.Debug 
			(Printf.sprintf "Possible aliases = %s\n\n" 
			   (List.fold_left (fun a -> fun b -> a^(Expression.toString b)) "" possibleAliases));
		    end;
                  (* compute all possible alias scenarios.  we can remove the case when the target
                     appears in the pred, because these are always aliased (and there is no point of
                     doing the usual wp substitution in this case). *)
		  let aliasSets = Misc.powerSet (List.filter (function al -> al <> (Expression.Lval target)) possibleAliases)
		  in
		  let _ = if _print_info_flag then
		    Message.msg_string Message.Debug ("aliasSets size:"^(string_of_int (List.length aliasSets))^"\n") in
		    
		  let wp_one aliasSet =
		    (* In the implicit version, we would construct substitutions here.
		       But here we should not substitute: the aliasExps will take care of that.
		       Jhala: how exactly will that happen ?? CHECK !!
		    *)
		    let substSet = List.map (function (_,e) -> (e,assigned_value))
				     (List.filter (function (b,e) -> b=true) aliasSet) in
		      (*
			if (not explicit_flag) then 
			List.iter (fun (e1, e2) -> 
			Message.msg_string Message.Normal 
			("Substitute: "^(Expression.toString e1)^(Expression.toString e2))) substSet ;
		      *)
                      (* create expressions representing the alias pairs *)
	      	    let aliasExps = List.map 
				      (function (b,e) ->
					 let targetAddr = Expression.addressOf (Expression.Lval target)
					 and eAddr = Expression.addressOf e 
					 in
					   match b with 
					       true -> Predicate.Atom (Expression.Binary(Expression.Eq,targetAddr,eAddr))
					     | false -> Predicate.Atom (Expression.Binary(Expression.Ne,targetAddr,eAddr))) 
				      aliasSet 
		    in
                      (* compute the weakest precondition *)
                      (* without pointers, the following line would be the entire
                         weakest precondition computation, with an empty substSet *)
		    let _s1 = Predicate.substitute ((Expression.Lval target, assigned_value) :: substSet) pred in
		      (*	if (not explicit_flag) then Message.msg_string Message.Normal ("_s1 = "^(Predicate.toString _s1)) ; *)
                      (* conjoin the weakest precondition result to the expressions
                         representing this particular aliasing scenario *)
                      Predicate.conjoinL (List.append (_s1:: aliasExps) 
					    (if explicit_flag then
					       [Predicate.Atom (Expression.Binary
								  (Expression.Eq,
								   assigned_value, valu))]
					     else [])
					 )
		  in
		  let actual_pre = (Predicate.disjoinL (List.map wp_one aliasSets)) in
		    (*
		      if (not explicit_flag) then 
		      Message.msg_string Message.Normal ("Actual pre = "^(Predicate.toString actual_pre)) ;
		    *)
		    if sat_flag then 
		      sat_filter (actual_pre) Predicate.True
		    else 
		      actual_pre
			
	    |        Expression.Unary(Expression.Assume, e) -> Predicate.conjoinL [(Predicate.Atom e); pred]
	    |        _ -> failwith "wp: not implemented"
	end
    | C_SD.Command.Return exp ->
        Message.msg_string Message.Debug "WP of return";
	if ((!retTarget) <> junkRet) then
          begin
            Message.msg_printer Message.Debug C_SD.Command.print { C_SD.Command.code = C_SD.Command.Block [(C_SD.Command.Expr(Expression.Assignment(Expression.Assign, !retTarget, exp)))];
								   C_SD.Command.read = []; (* printing ignores .read and .written *)
								   C_SD.Command.written = []}; (* ditto *)
            (wp explicit_flag sat_flag (C_SD.Command.Expr(Expression.Assignment(Expression.Assign, !retTarget, exp))) pred (edge,i))
          end
        else pred
	  (* Jhala: wp RETURN, here. Devilish hack. Sorry ! *)

  (*
    Strongest postcondition
  *)
	  (* Jhala: Need to fold in rupaks changes to SP *)
  (*
    Strongest postcondition
    We do not need an explicit flag because sp is always explicit
  *)
  let rec sp sat_flag stmt pred =
    match stmt with
        C_SD.Command.Expr e ->
          begin
            match e with  
                Expression.Assignment(Expression.Assign, target, value) ->
                  begin 
		    let targetExp = Expression.Lval target 
		    in
                    let possibleAliases = getPossibleAliases target pred in
                      (* compute all possible alias scenarios.  we can remove the case when the target
                         appears in the pred, because these are always aliased (and there is no point of
                         doing the usual wp substitution in this case). *)
                    let aliasSets = Misc.powerSet (List.filter (function al -> al <> (Expression.Lval target)) possibleAliases) in
                    let newvar = (getNewTmpVar target) in
                    let sp_one aliasSet =
                      (* create the pairs for substituting expressions in pred for
                         value, when pred and target are aliased *)
                      let substSet = List.map (function (_,e) -> (e,newvar))
                                       (List.filter (function (b,e) -> b=true) aliasSet) in
                        (* create expressions representing the alias pairs *)
                      let aliasExps = List.map 
                                        (function (b,e) ->
                                           let targetAddr = Expression.addressOf (Expression.Lval target)
                                           and eAddr = Expression.addressOf e in
                                             match b with 
                                                 true -> Predicate.Atom (Expression.Binary(Expression.Eq,targetAddr,eAddr))
                                               | false -> Predicate.Atom (Expression.Binary(Expression.Ne,targetAddr,eAddr))) aliasSet 
                      in
                        (* compute the strongest postcondition *)
                        (* without pointers, the following line would be the entire
                           strongest postcondition computation, with an empty substSet *)
                      let s1 = Predicate.substitute ((targetExp, newvar) :: substSet) pred in
                      let s2 = Expression.substituteExpression [(targetExp, newvar)] value in
                        Predicate.conjoinL (List.append 
                                              [s1; Predicate.Atom (Expression.Binary (Expression.Eq, targetExp, s2))] 
                                              aliasExps)
                    in
                    let actual_sp =  Predicate.disjoinL (List.map sp_one aliasSets) in
		    if sat_flag then
		      sat_filter (actual_sp) Predicate.True else actual_sp
                  end
              |        Expression.Unary (Expression.Assume, e) -> (* Predicate.normalize *)
		    (Predicate.And [(Predicate.Atom e); pred])
              |        _ -> failwith "sp: not implemented"
          end
      |  C_SD.Command.Return exp ->
	  if ((!retTarget) <> junkRet) then
	    begin
	      sp sat_flag (C_SD.Command.Expr (Expression.Assignment (Expression.Assign, !retTarget, exp))) pred
	    end
	  else
	    pred

  (*******************************************************************************************)


  let can_escape lv =
    C_SD.can_escape lv
	
  let is_lock v =
      C_SD.is_a_lock v

  let is_local v =
    Message.msg_string Message.Debug ("[Abstraction.is_local " ^(Expression.lvalToString v) ^ "]");
    let rec _is_local lv =
      match lv with
	Expression.Symbol s ->
	  String.contains s '@'
      | Expression.Access (_, e, _) -> _is_local_e e
      | Expression.Dereference e -> _is_local_e e
      | Expression.Indexed (e1, e2) -> _is_local_e e1 || _is_local_e e2
    and
	_is_local_e e =
      match e with
	Expression.Lval l -> _is_local l
      |	Expression.Chlval (l, _) -> _is_local l
      |	Expression.Binary (_, e1, e2) -> _is_local_e e1 || _is_local_e e2
      |	Expression.Unary (_, e) -> _is_local_e e
      |	Expression.Constant _ -> false
      |	_ -> failwith ("is_local : Strange expression "^(Expression.toString e))
    in
    _is_local v

  (*******************************************************************************************)
      (*******************************************************************************************)

  (* Pre and Post *)

(* concrete Pre for a function call *)
(* take the region, and replace all the variables that are the formals
   of the function you just called with the expressions that were used
   to call it *)	

(* Jhala: Misc functions required for the pre of function calls *)

(* Misc functions to play with parameters etc. *)

let _get_param exp = match exp with
  Expression.Lval (Expression.Symbol(x)) -> x
| _ -> failwith ("Invalid function Parameter " ^ (Expression.toString exp))
(* All function call params should be variables ... or you need
   unification ... ! *)

let listToMap l = function x -> (try List.assoc x l with _ -> x)

let real_fun_name fcexp = 
   match fcexp with 
     Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), elist) -> fname
   | Expression.Assignment (Expression.Assign,_ , 
			   Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), _)) ->
			     fname
   | _ -> failwith "bad call to real_fun_name"

let deconstructFunCall exp = 
   match exp with 
    Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), elist) -> 
      begin
	try
	  let formalList = C_SD.lookup_formals fname in
      (* check for vararg: in case there is a Variable formal argument, take only the first n+1 arguments,
         and print a warning message. vararg functions are usually things like printk which we ignore
         anyway, so correctness should not be affected.
	 *)
	  match formalList with
	    C_SD.Variable formals ->
	      begin
		
		Message.msg_string Message.Minor (* Message.Error generates too many messages*)
		  ("WARNING: Function "^fname^" has variable arguments. May discard arguments in this call.");
		let _len = List.length formals and _lenactuals = List.length elist in
		if (_len == _lenactuals) then (* this is the easy else case *)
		  let nameAssocList = try (List.combine formals elist) with _ ->
		    failwith ("deconstructFunCall: FunCall with too many/few args "^(Expression.toString exp))
		  in
		  (fname,nameAssocList, junkRet)
		else (* more difficult : no arg, or more *)
		  begin
		    if (_len < _lenactuals) then
		      let nameAssocList = try (List.combine formals (Misc.truncateList elist _len)) with _ ->
			failwith ("deconstructFunCall: FunCall with too many/few args "^(Expression.toString exp))
		      in
		      (fname,nameAssocList, junkRet)
		    else
		      let nameAssocList = try (List.combine (Misc.truncateList formals _lenactuals) elist) with _ ->
			failwith ("deconstructFunCall: FunCall with too many/few args "^(Expression.toString exp))
		      in
		      (fname,nameAssocList, junkRet)
			
		  end
	      end
	  |	C_SD.Fixed formals ->
	      let nameAssocList = try (List.combine formals elist) with _ ->
		failwith ("deconstructFunCall: FunCall with too many/few args "^(Expression.toString exp))
	      in
	      (fname,nameAssocList, junkRet)
	with C_SD.NoSuchFunctionException f -> (__NotImplementedFunctionName,[],junkRet)
      end
  | Expression.Assignment (Expression.Assign,target, 
			   Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), elist)) ->
     begin   
      try
      let formalList = C_SD.lookup_formals fname in
      (* let actuals = List.map _get_param elist in *)
      (* check for vararg: in case there is a $vararg argument, take only the first n+1 arguments,
         and print a warning message. vararg functions are usually things like printk which we ignore
         anyway, so correctness should not be affected.
      *)
      match formalList with
	C_SD.Variable formals ->
	  begin
	    Message.msg_string Message.Error ("WARNING: Function "^fname^" has variable arguments. May discard arguments in this call.");
	    let _len = List.length formals and _lenactuals = List.length elist in
	    if (_len == _lenactuals) then (* this is the easy else case *)
              let nameAssocList = try (List.combine formals elist) with _ ->
		failwith ("deconstructFunCall: FunCall with too many/few args "^(Expression.toString exp))
              in
              (fname,nameAssocList, target)
            else (* more difficult : no arg, or more *)
              begin
		if (_len < _lenactuals) then
		  let nameAssocList = try (List.combine formals (Misc.truncateList elist _len)) with _ ->
	            failwith ("deconstructFunCall: FunCall with too many/few args "^(Expression.toString exp))
		  in
		  (fname,nameAssocList, target)
		else
		  let nameAssocList = try (List.combine (Misc.truncateList formals _lenactuals) elist) with _ ->
	            failwith ("deconstructFunCall: FunCall with too many/few args "^(Expression.toString exp))
		  in
		  (fname,nameAssocList, target)
		    
              end
	  end
      |	C_SD.Fixed formals ->
	  let nameAssocList = try (List.combine formals elist) with _ ->
	    failwith ("deconstructFunCall: FunCall with too many/few args "^(Expression.toString exp))
	  in
	  (fname,nameAssocList,target)
      with
	C_SD.NoSuchFunctionException f -> (f,[],target)
     end
  | _ -> 
      if(Options.getValueOfBool "nofp") then
        begin
	  Message.msg_string Message.Debug ("Bad Call to deconstruct function call ! "^(Expression.toString exp));
	    (__NotImplementedFunctionName, [], junkRet)
	end
      else 
        failwith ("Bad Call to deconstruct function call ! "^(Expression.toString exp))
	
	
  let funCallSubstitute l = Predicate.substitute 
      (List.map (function (x,y) -> (Expression.Lval (Expression.Symbol(x)), Expression.Lval (Expression.Symbol(y)))) l)
      
(*  removeFunPreds fname predList = (preds that remain ,delPreds) 
    the ones that remain are the ones that do not contain any
    variables  local to that function *)

  let removeFunPreds fname predList = 
    let formalList = C_SD.lookup_formals fname in
    let formals = 
      match formalList with
	C_SD.Variable l -> l
      |	C_SD.Fixed l -> l
    in
    let _keepPred i = 
      let pred_i = getPred i in
      let vars_pred_i = getVarExps_deep pred_i in
      if ( List.exists (function elt -> List.mem elt formals) vars_pred_i) then false else true 
    in
    let remain_preds = List.filter _keepPred predList in
    let delPreds = List.filter (function x -> (not (_keepPred x))) predList in
    (remain_preds,delPreds) 
      
(* lastTarget looks at the stack of call points and returns the
   following info: (target := if the call was x = f(y), f, the
   successor location) *)

let last_loc_pointer = ref (None)
(* Jhala: Another evil hack as greg doesnt let  me look at the top of thestack !! *)

let lastTarget () = 
  match (!last_loc_pointer) with
      None -> (Message.msg_string Message.Debug "lastTarget called with errorLocation  as top of stack !"); failwith "Empty Call Stack"
    | Some(callLoc) ->
	let outEdges = C_SD.get_outgoing_edges callLoc in
	let _ = if (List.length outEdges <> 1) then (Message.msg_string Message.Error "Call point with bad successor !!") in
	let callEdge = (List.hd outEdges) in
	let fcallExp = 
	  match (C_SD.get_command callEdge) with	
	      {C_SD.Command.code = C_SD.Command.FunctionCall fcexp} -> fcexp 
	    | _ -> failwith "bad command at funcall point !"
	in
	let (fname,_,target) = deconstructFunCall fcallExp in
	let retLoc = C_SD.get_target (List.hd outEdges) in
	(target,fname,retLoc)
	 
let isReturn op = 
   let cmd = C_SD.get_command op in
   match cmd.C_SD.Command.code with
       C_SD.Command.Block l -> (List.exists (function x -> match x with C_SD.Command.Return e -> true | _ -> false) l)
   | _ -> false

let returnExp op = 
  let cmd = C_SD.get_command op in
    match cmd.C_SD.Command.code with
	C_SD.Command.Block l -> 
	  begin
	    match List.hd (List.rev l) with 
		C_SD.Command.Return e -> e 
	      | _ -> failwith "Bad arg to returnExp 1"
	  end
| _ -> failwith "Bad arg to returnExp"

let isBlock op =
   let cmd = C_SD.get_command op in
   match cmd.C_SD.Command.code with
          C_SD.Command.Block _ -> true
      | _ -> false

let isPredicate op =
   let cmd = C_SD.get_command op in
   match cmd.C_SD.Command.code with
          C_SD.Command.Pred _ -> true
      | _ -> false

let isFunCall op = 
   let cmd = C_SD.get_command op in
   match cmd.C_SD.Command.code with
       C_SD.Command.FunctionCall _ -> true 
   | _ -> false

let opFunName op = 
   let cmd = C_SD.get_command op in
   match cmd.C_SD.Command.code with 
   C_SD.Command.FunctionCall fcallExp -> Misc.fst3 (deconstructFunCall fcallExp)
   | _ -> Message.msg_string Message.Error "opFunName called with bad arg !"; failwith "opFunName called with bad arg !"


  (* function to substitute tid -> notTid and locals with fresh copies.
     Uses the global phi_com_counter to get new names for locals 
     *)
  (* this is used to give new names to all the local variables inside the phi *)
 
  let new_name suffix (lv : Expression.lval) = 
    let rec _new_name lv =
      match lv with
	Expression.Symbol v ->
	  Expression.Symbol 
	    (if (Symbol.compare v "tid") = 0 
	    then "notTid"
	    else
	      if (Symbol.compare v "notTid") = 0 then "tid" else
	      if (Symbol.compare v "self") = 0 then "self" else
	      if String.contains v '@' 
	      then (v^suffix)
	      else v)
      | Expression.Access (aop, e, fld) -> 
	  Expression.Access(aop, _new_name_e e, fld)
      |	Expression.Dereference e -> Expression.Dereference (_new_name_e e)
      |	Expression.Indexed (e1, e2) -> Expression.Indexed (_new_name_e e1, _new_name_e e2)
    and
	_new_name_e e =
      match e with
	Expression.Lval l -> Expression.Lval (_new_name l)
      |	Expression.Chlval (l, suff) -> 
	  begin
	    let l' = 
	      match l with
		Expression.Symbol _ -> l
	      | Expression.Dereference e -> Expression.Dereference (_new_name_e e)
	      | Expression.Access (op, e1, fld) -> Expression.Access (op, _new_name_e e1, fld)
	      | Expression.Indexed (e1, e2) -> failwith "_new_name : array not handled"
	    in
	    Expression.Chlval (l', suff)
	  end
      |	Expression.Binary (op, e1, e2) -> Expression.Binary (op, _new_name_e e1, _new_name_e e2)
      |	Expression.Unary (op, e) -> Expression.Unary (op, _new_name_e e)
      |	_ -> e
    in
    _new_name lv
      


  (******************************************************************************)
	    
      (* every so often we need fresh names for variables. The
	 suffix_counter variable, and the brand_new_suffix
	 function provides the utility: essentially these
	 provide a "brand new suffix" that is tagged on to the current
	 id / lvalue.

	 Sometimes, we need to know the current suffix. get_current_suffix returns
	 the current suffix.
	 *)
  let suffix_counter = ref 0 
  let brand_new_suffix () = 
    suffix_counter := !suffix_counter + 1;
    "%_bsuf_"^(string_of_int !suffix_counter)
		 
  let get_current_suffix () =
    "%_bsuf_"^(string_of_int !suffix_counter)

  (******************************************************************************)
  (* new : with aliasing *)
  let phi_command_to_pred sym flag phi_table =
    assert (flag == true) ;
    
    let phi_pred newname sym1 sym2 =
      let phi = 
	try
	  Region.rename_vars 
	    !(Hashtbl.find phi_table sym2) 
	    (newname) 
	with Not_found ->
	  
	  begin
	    (* RUPAK: check that this is sound.
	       If the name sym2 is not in the phi_table, this means 
	       that the syntactic lvalue name sym2 is never used in
	       the program to read an address or to update an address.
	       Hence we can assume that the phi is false.
	       *)
	    Message.msg_string Message.Error ("phi_command_to_pred raised not found on lvalue "^(Expression.lvalToString sym2)) ;
	    Region.bot
	  end
      in
      begin
	match phi with
	  Region.Empty -> Predicate.False
	|	_ ->
	    begin
	      if (sym1 = sym2) then Region.region_to_predicate phi
	      else
		begin
		  Predicate.And 
		    [ Predicate.Atom (Expression.Binary(Expression.Eq, 
							Expression.addressOf (Expression.Lval sym1),
							Expression.addressOf (Expression.Lval sym2)));
		      Region.region_to_predicate phi ] 
		end
	    end
      end
    in
    if can_escape sym then
      begin
	if flag then 
	  begin
	    let aliases = AliasAnalyzer.getAllAliases sym in
	    let newname = new_name (brand_new_suffix ()) in
	    let p = Predicate.normalize (Predicate.Or (List.map (phi_pred newname sym) (sym :: aliases)))
	    in
	    Message.msg_string Message.Normal ("Phi for "^(Expression.lvalToString sym)^" is "^(Predicate.toString p)) ;
	    p
	  end
	else 
	  begin
	    failwith "phi_command_to_pred : flag should not be false "
	  end
      end
    else
      Predicate.False

(* Old, without aliasing
  let phi_command_to_pred sym flag phi_table =
    let phi = 
      try
	Region.rename_vars 
	  !(Hashtbl.find phi_table sym) 
	  (new_name (brand_new_suffix ())) 
      with Not_found -> failwith ("phi_command_to_pred raised not found on lvalue "^(Expression.lvalToString sym))
    in
    assert (flag == true) ;
    let phi_pred = Region.region_to_predicate phi in
    if can_escape sym then
      begin
	if flag then phi_pred 
	else 
	  begin
	    failwith "phi_command_to_pred : flag should not be false " ;
	    Predicate.Not phi_pred
	  end
      end
    else
      Predicate.False
*)

let asgn_of_funcall fcallExp = 
  let (fname,_subs,target) = (deconstructFunCall fcallExp) in
  let exists_fun = 
    try 
      let fml = C_SD.lookup_formals fname 
      in true 
    with C_SD.NoSuchFunctionException _ -> false
  in
  let asgn_list = 
    if (exists_fun || (target = junkRet)) then
      List.map (fun (x,e) -> C_SD.Command.Expr( Expression.Assignment (Expression.Assign,Expression.Symbol(x),e))) _subs 
    else
      [C_SD.Command.Expr( Expression.Assignment (Expression.Assign, target, get_new_unknown ()))]
  in         
    (* PREV: List.fold_right (wp explicit_flag) asgn_list pred *)
    (* WITH ANNOT: NOT SURE THIS IS CORRECT: check with Jhala *)
    asgn_list


(* Jhala: End of misc functions ... *)

  (* Concrete predecessor: Given a pair (C_SD location, predicate), and a C_SD command operation,
     compute the set of concrete predecessor states.
     
     This is used in the counterexample-based refinement process.
  *)
      (* Type : Region.concrete_data_region -> C_SD.Command.t -> C_SD.edge -> Region.concrete_data_region *)
  let rec concrete_data_pre explicit_flag sat_flag conc_data_reg cmd edge phi_table =
    Message.msg_string Message.Debug "In concrete_data_pre -- arguments are:" ;
    if explicit_flag then (Message.msg_string Message.Debug " EXPLICIT WP ");
    Message.msg_printer Message.Debug Region.print_concrete_data_region conc_data_reg ;
    Message.msg_printer Message.Debug C_SD.Command.print cmd ;

    let pred = conc_data_reg.Region.pred in
    let newpred =
      let rec _doBlock stmtlst pred i =
                match stmtlst with
                  [] -> if (Options.getValueOfBool "wpsat" && sat_flag) then sat_filter pred (!wpCapRegion) else pred
                | stmt::rest -> 
                    begin
                      let wp_pred = wp explicit_flag sat_flag stmt pred (edge,i) in
                      if (rest = []) then 
                         if (Options.getValueOfBool "wpsat" && sat_flag) then 
                                 sat_filter wp_pred (!wpCapRegion)
                              else 
                                 wp_pred
                      else
                         _doBlock rest wp_pred (i+1)
                    end
      in
      match cmd.C_SD.Command.code with
          C_SD.Command.Block el ->
            begin         
              (* PREV: List.fold_right (fun a -> fun b -> wp false a b (edge,0)) el pred *)
              _doBlock (List.rev el) (* predicate *)pred (* end of block *)0 
            end
        | C_SD.Command.Pred e -> Predicate.conjoinL  [e; pred]
        | C_SD.Command.FunctionCall fcallExp -> 
        let asgn_list = asgn_of_funcall fcallExp in
              _doBlock asgn_list pred 0
        | C_SD.Command.Skip -> pred
	| C_SD.Command.Havoc lv ->
	    let new_var = 
	      match getNewTmpVar (Expression.Symbol "__foo_havoc") with 
		Expression.Chlval (Expression.Symbol x, s) -> s^x
	      | _ -> failwith "Abstraction.concrete_data_pre: Havoc: TBD"
	    in
	    let new_name v = 
	      if (v = lv) 
	      then Expression.Symbol new_var
	      else v
	    in 
	    Predicate.alpha_convert new_name pred 
	| C_SD.Command.HavocAll -> failwith "Abstraction.concrete_data_pre: FunctionCall: TBD"
	| C_SD.Command.Phi (sym, flag) -> 
	    let e = phi_command_to_pred sym flag phi_table in
	    let cmd' = {cmd with C_SD.Command.code = C_SD.Command.Pred e} in
	    let res = concrete_data_pre explicit_flag sat_flag conc_data_reg cmd' edge phi_table in
	    res.Region.pred
    in
    { Region.pred = newpred ; }

(*
	    let new_var = 
	      match getNewTmpVar (Expression.Symbol sym) with 
		Expression.Symbol x -> x
	      | _ -> failwith "Abstraction.concrete_data_pre: FunctionCall: TBD"
	    in
	    let new_name v = 
	      if (Symbol.compare v sym) = 0 
	      then new_var
	      else v
	    in 
	    let temp_pred = Predicate.alpha_convert new_name pred in 
	    let pred_region = Region.Atomic { Region.location = C_SD.fake_location ;
					      Region.data = Region.Concrete { Region.pred = temp_pred ; } ;
					      Region.stack = Region.EveryStack } in
	    let new_name v = 
	      if Symbol.compare v C_SD.tid = 0 then C_SD.notTid else v
	    in
	    let phi = Region.rename_vars (AR.concrete !(Hashtbl.find phi_table sym)) new_name in
	    if Region.is_empty (Region.cap !assumed_invariants_region (Region.cap pred_region phi))
	    then pred
	    else temp_pred
*)

(** block-concrete_data_pre *)


(* to do block_concrete_data_pre we need to know all the lvals that appear in a trace -- so we define load_lv_table here *)

    let load_lv_table op_l = 
  let add_exp e = 
    let clos  = C_SD.expression_closure e in
      add_to_lv_table (List.flatten (List.map getVarsAndDerefs_exp clos))
  in   
  let proc_asgn asgn = 
      match asgn with
	C_SD.Command.Expr (Expression.Assignment(Expression.Assign, target, valu)) ->
	  let target_closure = C_SD.expression_closure (Expression.Lval target) in
	  let valu_closure = 
	    List.map 
	      (
		fun x -> 
		  let vx = 
		    (Expression.substituteExpression [(Expression.Lval target, valu)] x)
		  in 
		    try Expression.Lval (Expression.push_deref (lv_of_expr vx))
		    with _ -> vx
	      )
	      target_closure
      	  in
	    add_to_lv_table (List.flatten (List.map getVarsAndDerefs_exp (target_closure@valu_closure))) 
     | C_SD.Command.Return e -> add_exp e (* already dealt with *)   
     | _ -> ()
  in
  let process_op op = 
    match (C_SD.get_command op).C_SD.Command.code with
	C_SD.Command.Skip -> ()
      | C_SD.Command.Pred e ->  (add_to_lv_table (getVarsAndDerefs e))
      | C_SD.Command.Block el -> List.iter proc_asgn el
      | C_SD.Command.FunctionCall fcallExp -> 
	  let (_,_,target) = deconstructFunCall fcallExp in
	  let asgn_list = asgn_of_funcall fcallExp in
	  let _ =  if (target <> junkRet) then add_exp (Expression.Lval target) in
	    List.iter proc_asgn asgn_list
      | _ -> ()
  in
    Hashtbl.clear trace_lv_table;
    List.iter process_op op_l

let blocks_of_pred p = 
  match p with 
      Predicate.And l -> l
    | _ -> failwith "bad call to blocks_of_pred"
      
(*let atomic_reg_of reg = 
  match reg with
      Region.Atomic adr -> adr
    | _ -> failwith "atomic_r !"
*)

(* this function expects the conc_data_reg to look like And ( pred_list) and returns something with that
   form as well *)

(* this function pretends he's called with the predicate true -- its the callers job to collect 
   all the constraints! 
*)


let rec block_concrete_data_pre atomic_reg cmd edge (* phi_table *) =
  Message.msg_string Message.Debug "In block_concrete_data_pre -- arguments are:" ;
  Message.msg_printer Message.Debug Region.print_atomic_region atomic_reg ;
  Message.msg_printer Message.Debug C_SD.Command.print cmd ;
  let _ = 
    if (isReturn edge) then
      begin
        Message.msg_string Message.Debug "block c_ Pre: Hit isReturn\n";
        (* recall that we put the call points in the stack *)
        (* we assume that a function call edge is the only edge
           going to its target location *)
	Message.msg_string Message.Debug "Ingoing edges : ";
        List.iter (fun a -> Message.msg_string Message.Debug (Operation.to_string a)) (C_SD.get_ingoing_edges atomic_reg.Region.location);
        Message.msg_string Message.Debug "\nNow matching...\n";
        let rec get_fn_call_edge lst =
          match lst with
	      a::rest -> 
                begin
                  if (isFunCall a) then
                    (* check there are no other function calls *)
                    if (List.exists isFunCall rest) then
		      failwith "bc pre: More than one function call in incoming edges!"
                    else a
                  else
                    get_fn_call_edge rest
                end
            | [] -> failwith ("pre: No corresponding function call found")
        in
        let callLoc = (C_SD.get_source (get_fn_call_edge (C_SD.get_ingoing_edges atomic_reg.Region.location))) 
	in
            last_loc_pointer := Some(callLoc); 
	  retTarget := Misc.fst3 (lastTarget ());
	  last_loc_pointer := None;
      end
  in
    (* this should be a list of blocks of constraints *)
  let newpred =
    let rec _doBlock stmtlst pred i =
      match stmtlst with
          [] -> (*if (Options.getValueOfBool "wpsat" && sat_flag) then sat_filter pred (!wpCapRegion) else*) 
	    pred
        | stmt::rest -> 
            begin
              let wp_pred = block_wp (* explicit_flag sat_flag*) stmt pred (* (edge,i)*) in
                (*if (rest = []) then 
                  if (Options.getValueOfBool "wpsat" && sat_flag) then 
                  sat_filter wp_pred (!wpCapRegion)
                  else 
                  wp_pred
                  else*)
                _doBlock rest wp_pred (i+1)
            end
    in
      match cmd.C_SD.Command.code with
          C_SD.Command.Block el ->
	    begin         
	      (* PREV: List.fold_right (fun a -> fun b -> wp false a b (edge,0)) el pred *)
	      let block_pred = _doBlock (List.rev el) (* predicate *) (Predicate.True) (* end of block *)0 
	       in
		block_pred
		(* Predicate.And (block_pred :: pred_l)*)
	    end
        | C_SD.Command.Pred e -> 
	    begin 
	      let rv =  (* should take a boolean closure here -- see popl submission *)
		add_to_lv_table (getVarsAndDerefs e); 
	        pred_lval_map e (* Predicate.And (e::pred_l) *)
	      in
	         Message.msg_string Message.Debug ("bwp pred: "^(Predicate.toString rv));
	         rv
            end
        | C_SD.Command.FunctionCall fcallExp -> 
	    let (fname,_subs,target) = (deconstructFunCall fcallExp) in
	    let exists_fun = 
	      try 
		let fml = C_SD.lookup_formals fname 
		in true 
	      with C_SD.NoSuchFunctionException _ -> false
	    in
	    let asgn_list = 
	      if (exists_fun || (target = junkRet)) then
		List.map (fun (x,e) -> C_SD.Command.Expr( Expression.Assignment (Expression.Assign,Expression.Symbol(x),e))) _subs 
	      else
		[C_SD.Command.Expr( Expression.Assignment (Expression.Assign, target, get_new_unknown ()))]
	    in         
	      (* PREV: List.fold_right (wp explicit_flag) asgn_list pred *)
	      (* WITH ANNOT: NOT SURE THIS IS CORRECT: check with Jhala *)
	      _doBlock asgn_list (Predicate.True) 0
        | C_SD.Command.Skip -> Predicate.True
	| C_SD.Command.Havoc lv -> 
	    failwith ("Havoc unimplemented in block_concrete_data_pre")
	    (*
		  let new_var = 
	      match getNewTmpVar (Expression.Symbol "__foo_havoc") with 
	      Expression.Chlval (Expression.Symbol x, s) -> s^x
	      | _ -> failwith "Abstraction.concrete_data_pre: Havoc: TBD"
	      in
	      let new_name v = 
	      if (v = lv) 
	      then Expression.Symbol new_var
	      else v
	      in 
	      Predicate.alpha_convert new_name pred 
	    *)
	| C_SD.Command.HavocAll -> failwith "Abstraction.block_concrete_data_pre: FunctionCall: TBD"
	| C_SD.Command.Phi (sym, flag) ->
	    failwith "block_cdp: Phi unimplemented"
	    (*
	      let e = phi_command_to_pred sym flag phi_table in
	      let cmd' = {cmd with C_SD.Command.code = C_SD.Command.Pred e} in
	      let res = concrete_data_pre explicit_flag sat_flag conc_data_reg cmd' edge phi_table in
	      res.Region.pred
	    *)
  in
  let _ = (retTarget := junkRet) in
    newpred
    (* { Region.pred = newpred ; } *)
    
(** block-wp end *)


  let rec concrete_data_post conc_data_reg cmd edge phi_table =
    Message.msg_string Message.Debug "In concrete_data_post -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print_concrete_data_region conc_data_reg ;
    Message.msg_printer Message.Debug C_SD.Command.print cmd ;

    let pred = conc_data_reg.Region.pred in
    let flag = Options.getValueOfBool "wpsat" in
    let newpred =
      match cmd.C_SD.Command.code with
          C_SD.Command.Block el -> List.fold_left (fun a -> fun b -> sp flag b a) pred el
        | C_SD.Command.Pred e -> Predicate.normalize (Predicate.And [e; pred])
        | C_SD.Command.FunctionCall fcallExp -> 
            begin
              let (_, _subs, _) = deconstructFunCall fcallExp in
              let asgn_list = List.map (fun (x,e) -> 
                C_SD.Command.Expr(Expression.Assignment 
                                    (Expression.Assign,Expression.Symbol(x),e))) _subs 
              in         
              List.fold_left (fun a -> fun b -> sp flag b a) pred asgn_list
            end
        | C_SD.Command.Skip -> pred
	| C_SD.Command.Havoc sym -> 	    
	    let new_var = (* sym should be assigned a fresh variable *)
	      match getNewTmpVar (Expression.Symbol "__foo") with 
		Expression.Chlval (Expression.Symbol x, s) -> Expression.Lval (Expression.Symbol (s^x))
	      | _ -> failwith "Abstraction.concrete_data_post: Havoc of lval: TBD"
	    in
	    let assign = Expression.Assignment(Expression.Assign, sym, new_var) in
	    let stmt = C_SD.Command.Expr assign in
	    sp flag stmt pred
        | C_SD.Command.HavocAll -> failwith "Abstraction.concrete_data_post: HavocAll: TBD"
	| C_SD.Command.Phi (sym, flag) -> 
	    let e = phi_command_to_pred sym flag phi_table in
	    Predicate.normalize (Predicate.And [e; pred])
    in
    { Region.pred = newpred ; }


  (* Abstract pre operator. Keep track of predicates mentioned 
     in reg.preds
  *)
      (* Type: Region.abstract_data_region -> C_SD.Command.t -> Region.abstract_data_region *)
  let abstract_data_pre abst_data_reg cmd edge phi_table =
    Message.msg_string Message.Debug "In abstract_data_pre -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;
    Message.msg_printer Message.Debug C_SD.Command.print cmd ;

    (* Implementation of the H function from DasDillPark-CAV99 *)
    let rec _H (bdd, p) predlist =
      (* optimization: check if (abst_data_reg, p) is satisfiable? *)
      if (isSatisfiable (bdd, p)) then
        begin
          match predlist with
              [] -> if (isSatisfiable (bdd, p)) then bddOne else bddZero
            | a:: rest ->
                let pr = getPred a in
                let cr = concrete_data_post { Region.pred = pr ; } cmd edge phi_table in
                let wppos = cr.Region.pred in
                let wpneg = Predicate.negate wppos in
                  CaddieBdd.bddOr 
                    (CaddieBdd.bddAnd (getPredIndexBdd a) (_H (bdd, wppos :: p) rest))
                    (CaddieBdd.bddAnd (CaddieBdd.bddNot (getPredIndexBdd a)) (_H (bdd, wpneg :: p) rest))
        end
      else
        bddZero
    in
      { Region.bdd = _H (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds ;
        Region.preds = abst_data_reg.Region.preds ; }


   
let abstract_data_post_das_dill abst_data_reg cmd edge post_location phi_table =
   Message.msg_string Message.Debug "In abstract_data_post -- arguments are:" ;
   Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;
   Message.msg_printer Message.Debug C_SD.Command.print cmd ;
   
    (* Implementation of the H function from DasDillPark-CAV99 *)
   let rec _H_cmd _cmd (bdd,p) predlist =
      Message.msg_string Message.Debug ("In _H of post, bdd = "^(Predicate.toString (convertBddToPred bdd)));
      Message.msg_string Message.Debug ("pred = "^(match p with [] -> "" | _ -> Predicate.toString (List.hd p)));
      (* optimization: check if (abst_data_reg, p) is satisfiable? *)
      if (isSatisfiable (bdd,p)) then
        begin
          match predlist with
            [] -> if (isSatisfiable (bdd,p)) then bddOne else bddZero
            | a:: rest ->
                let pr = getPred a in
                  (*
                    let cr = concretePre {Region.conclocation = reg.Region.abslocation; Region.pred = pr} op in
                  *)
                let cr = concrete_data_pre false false { Region.pred = pr ; } _cmd edge phi_table in
                let wppos = cr.Region.pred in
                let ncr = concrete_data_pre false false { Region.pred = Predicate.negate pr ; } _cmd edge phi_table in
                let wpneg = ncr.Region.pred
                in
                  CaddieBdd.bddOr 
                    (CaddieBdd.bddAnd (getPredIndexBdd a) (_H_cmd _cmd (bdd, wppos :: p) rest))
                    (CaddieBdd.bddAnd (CaddieBdd.bddNot (getPredIndexBdd a)) (_H_cmd _cmd (bdd, wpneg :: p) rest))
        end
   else
   bddZero
   in
   let _H = _H_cmd cmd in

   match cmd.C_SD.Command.code with 
   	C_SD.Command.FunctionCall fcallExp ->
	  let (fname,_subs,_) = (deconstructFunCall fcallExp) in 
          let asgn_block =
            { cmd with C_SD.Command.code = C_SD.Command.Block (List.map (fun (x,e) -> C_SD.Command.Expr( Expression.Assignment (Expression.Assign,Expression.Symbol(x), e))) _subs) }
          in
              { 
      	      Region.preds = abst_data_reg.Region.preds;
      	      Region.bdd =   _H_cmd asgn_block (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds;
	     }
	  (*let subs = List.map Misc.swap _subs in
	  let rec _make_copy_list lst c_lst pBdd = 
	    match lst with
	      [] -> (c_lst,pBdd)
	    | h::t -> 
		let pred_h = getPred h in
		let cpred_h = funCallSubstitute subs pred_h in
		if (cpred_h = pred_h) 
		then (_make_copy_list t c_lst pBdd) (* no copy made *)
		else (
		  let cpred_i = addPred cpred_h in
		  let nextPBdd = CaddieBdd.bddAnd (CaddieBdd.bddBiimp
					       (getPredIndexBdd h)
					       (getPredIndexBdd
						  cpred_i)) pBdd in
		  _make_copy_list t (cpred_i::c_lst) nextPBdd) (* copy made *)
	  in 
	  let (subsPreds, newBdd) = _make_copy_list abst_data_reg.Region.preds [] abst_data_reg.Region.bdd
	  in 
   { Region.bdd = newBdd ;
   Region.preds = abst_data_reg.Region.preds} *)
   	  (* What was just done for the function call ... *) 
	  (* Push stuff onto the stack: ... in the rest of post 
	     1. edge, on return, the location will be
	     the other end of the edge ...
	     2. if edge was a y = f(x) then post on return reflects this *)
	  (* Go through list of predicates and compute "copies" *)
	  (* Compute BDD of reg ^_i (pred_i = copy_pred_i) *)
	  (* Start location = start location of function *)
      |	C_SD.Command.Block l  -> 
	  if ( not (List.exists (function x -> match x with C_SD.Command.Return e -> true | _ -> false) l))
	  then 
	    let _ = (retTarget := junkRet) in
	    { 
      	      Region.preds = abst_data_reg.Region.preds;
      	      Region.bdd =   _H  (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds;
	     }
	  else 
	    let (targ, fname, _) = lastTarget () in
 (* Jhala: the t.o.s. is passed using last_loc_pointer  reg.Region.stack *) 
	    let _ = (retTarget := targ) in
	    let (newPreds,delPreds) = removeFunPreds fname abst_data_reg.Region.preds in
            let _ = Message.msg_string Message.Debug "Return: Calling _H" in
	    let newBdd = (_H  (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds) in
            let _ = Message.msg_string Message.Debug "Return: Back from _H" in
	    let _predToVar i =  -1 in (* JHALA: To be done ... ??? *)
	    let ex_newBdd = newBdd in
	      (*try CaddieBdd.exists (CaddieBdd.fromList bddManager (List.map
							_predToVar
							delPreds)) newBdd 
	      with _ -> newBdd (* Jhala: TBD... how to go from pred to
				  bdd variable (int) ? wrapper needs
				  work *)*)
	    
	    let _ = (retTarget := junkRet) in
	    { 
      	      Region.preds = newPreds@delPreds;
      	      Region.bdd =  ex_newBdd;
	    } 
   | _ -> { 
             Region.bdd = _H (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds ;
             Region.preds = abst_data_reg.Region.preds ;
          }



(* -----  JHALA: The new abstract_data_post : SLAM_POST *)
   let assume_pred_ref = ref None 
(* We use this to check if a given predicate is unchanged in this post ie if its wp = p 
   ... which also means <?> wpnp = np this is cleared every time postBdd is called *)

let unchangedPredTable = Hashtbl.create 37
 
let dontcareArray = Array.create maxPossiblePreds false
let outOfScopeArray = Array.create maxPossiblePreds false

let extract_scope var = 
  let names = Misc.chop var "@" in
    try List.nth names 1 with _ -> "__BLAST_GLOBAL" 
    

let postBdd bdd _pred_list _post_split_preds =
  Message.msg_string Message.Debug "In postBdd..\n";
  Message.msg_string Message.Debug ("post of assume"^(match !assume_pred_ref with Some(e) -> ("YES:"^(Predicate.toString e)) | _ -> "NO")^"\n");
  Message.msg_string Message.Debug ("Bdd = "^(Predicate.toString (convertBddToPred bdd)));
  Message.msg_string Message.Debug "\nList of predicates = ";
  let pred_list = List.map (fun (a,b,c,d) -> (a,b,c)) _pred_list in
  let post_split_preds = List.map (fun (a,b,c,d) -> (a,b,c)) _post_split_preds in
  let craig_flag = (Options.getValueOfInt "craig" = 2) in
  let craig_flag1 = (Options.getValueOfInt "craig" = 1) in
  let dc_flag = (Options.getValueOfBool "dc") in
  let _ = 
    if (Options.getValueOfBool "scope") then 
      begin
	Misc.array_reset outOfScopeArray false;
	List.iter (fun (a,_,_,d) -> Array.set outOfScopeArray a d) _pred_list 
      end
  in
  let rec loop lst = 
    match lst with 
	[] ->  () 
      | (a,b,c,d)::rest -> 
	  begin
	    ignore (Message.msg_string Message.Debug (Predicate.toString (getPred a)));
	    ignore (Message.msg_string Message.Debug (Predicate.toString b));
	    ignore (Message.msg_string Message.Debug (Predicate.toString c)); 
	    ignore (Message.msg_string Message.Debug (string_of_bool d));
	    loop rest
	  end
  in
  let _ = loop _pred_list in
  let _ =  Message.msg_string Message.Debug ("Will split on: \n");
    loop  _post_split_preds
  in

  (* assuming that if p = wp, then np = wpnp ...*)
  let wp_different_filter (p,wp,_) = 
    if craig_flag then true 
    else
      match Options.getValueOfString "post" with
	  "slam" -> true
	| "slam-dumb" -> not (getPred p = wp)
	| "slam1" -> not ((getPred p = wp) && (!assume_pred_ref = None))
	| "slam2" -> let p_pred = getPred p in 
	    (not (p_pred = wp) || 
	     (match !assume_pred_ref with 
		  None -> false
		|	Some(asm) -> not (Misc.intersection (getVarExps_deep p_pred) (getVarExps_deep asm) = [])))  
            (* JHALA: Instead of this intersection check I would like to say 
               if the two preds are textually the same or imply 
               each other or some such ...*)
	|  a -> failwith ("Unknown post algorithm "^a)
  in
  (* I dont have a good idea of when an "if" affects/doesnt affect a predicate, so for now, leave them all in ... there should be some cone-check here ...*)
  let (_changed_pred_list, unchanged_pred_list) = Stats.time "wp_diff_filter" (Misc.filter_cut wp_different_filter) pred_list in
  let cpl_addon = if craig_flag1 then post_split_preds else [] in
  let changed_pred_list = _changed_pred_list @ cpl_addon in
    
  let _ = Hashtbl.clear unchangedPredTable in
  let _ = List.iter (function x -> Hashtbl.add unchangedPredTable (Misc.fst3 x) true) unchanged_pred_list in
(* all the unchanged preds are dumped into this table *)
  let sumofcubesBdd = ref bddZero in
  
  let conjunctL = ref [] in
	    (* conjunct that the present cube produces post-command c *)
  let _ = Misc.array_reset dontcareArray false in
  
   (* START TIME HOLE *)
  let _check_Pred_list_simplify assumptions pred_triple_list conjunct_pointer =
   let asm_conj = Predicate.conjoinL assumptions in
   let _ = Message.msg_string Message.Debug ("Assuming pre-state: "^(Predicate.toString asm_conj)) in
   let _ = Stats.time "simplify_assume" (TheoremProver.simplify_assume ()) asm_conj in 
   let _check_pred (*_asm_conj*) (p,wpp,wpnp) = 
     if ( dontcareArray.(p) && dc_flag) then false 
       (* this tuple will never be called if I dont care about it if craig is true *)
     else 
       
          let eToCheck1 = Predicate.implies asm_conj wpp in
          let eToCheck2 = Predicate.implies asm_conj wpnp in
	  let _ = Message.msg_string Message.Debug ("Post pred: "^(Predicate.toString (getPred p))) in
	  let _ = Message.msg_string Message.Debug ("Post query+: "^(Predicate.toString wpp)) in
	  let _ = Message.msg_string Message.Debug ("Post query-: "^(Predicate.toString wpnp)) in
	  let ans1 = Stats.time "simp-a1:" (askTheoremProverContext eToCheck1) wpp in
          let ans2 = Stats.time "simp-a2:" (askTheoremProverContext eToCheck2) wpnp in
          (*let ans1 = (* Stats.time "simp-a1:"*) askTheoremProver eToCheck1 in
          let ans2 = (* Stats.time "simp-a2:"*) askTheoremProver eToCheck2 in*)
          match ans1 with
          true ->  
            if  (ans2 = true) then 
              begin
              Message.msg_string Message.Debug ("Cube implies both wp(p) and wp(neg p): "^(Predicate.toString (getPred p))); 
              conjunct_pointer := (p,true)::(p,false)::(!conjunct_pointer);
              true
              end
            else
              begin 
	        conjunct_pointer := (p,true)::(!conjunct_pointer);
	        true               
              end
         | false ->  
	     if (ans2 = true) then  
               begin
	         conjunct_pointer := (p,false)::(! conjunct_pointer);
	         true
               end
	     else 
	       false
   in
   let _ = Stats.time "_check_pred" (List.map (_check_pred (* asm_conj *))) pred_triple_list in
     statWorstCaseQueriesForPost :=  2*(List.length pred_list) + !statWorstCaseQueriesForPost;
     statActualQueriesForPost := 2*(List.length pred_triple_list) + (!statActualQueriesForPost);
     statNumPost := !statNumPost + 1;
     (match (!assume_pred_ref) with
	  Some _ -> 
	    (statAssumeQueriesForPost := 2*(List.length pred_triple_list) + (!statAssumeQueriesForPost); statNumAssumePost := !statNumAssumePost + 1)
	| None -> ())
	   ;
     TheoremProver.simplify_pop ();
     true
   in
   (* END TIME HOLE *)

   let _check_Pred_list_vampyre assumptions pred_triple_list conjunct_pointer = 
    let is_cube_consistent  = TheoremProver.assume assumptions in
    let asm_conj = Predicate.conjoinL assumptions in
    let _check_pred assumptions (p,wpp,wpnp) = 
      if ((dontcareArray.(p)) && dc_flag) then false
      else 
	let eToCheck1 = Predicate.implies asm_conj wpp in
	let eToCheck2 = Predicate.implies asm_conj wpnp in
	let ans1 = (*Stats.time "ask1:"*) (askTheoremProverContext eToCheck1) wpp in
	let ans2 = (*Stats.time "ask2:"*) (askTheoremProverContext eToCheck2) wpnp in
	  match ans1 with
              true ->  
		if  (ans2 = true) then 
		  begin
		    Message.msg_string Message.Debug ("Cube implies both wp(p) and wp(neg p): "^(Predicate.toString (getPred p))); 
		    conjunct_pointer := (p,true)::(p,false)::(!conjunct_pointer);
		    true
		  end
		  
		else
		  begin 
		    conjunct_pointer := (p,true)::(!conjunct_pointer);
		    true               
		  end
	    | false -> 
	 if (ans2 = true) then  
           begin
	     conjunct_pointer := (p,false)::(!conjunct_pointer);
	     true
           end
	 else 
	   false
    in
      
      if (is_cube_consistent) then 
	begin 
          let _ = List.map (_check_pred asm_conj) pred_triple_list in
            true
	end 
      else false
   in
 
  let _check_Pred_list = 
   match (Options.getValueOfString "s" ) with
   "vampyre" -> _check_Pred_list_vampyre
   | _ -> _check_Pred_list_simplify
  in
  (* the preds below are the integer identifiers *)
  let rec make_conjunct_bdd conjl = 
    match conjl with
      [] -> bddOne 
    | (p,true)::tl ->  CaddieBdd.bddAnd (getPredIndexBdd p) (make_conjunct_bdd tl) 
    | (p,false)::tl -> CaddieBdd.bddAnd (CaddieBdd.bddNot (getPredIndexBdd p))  (make_conjunct_bdd tl) 
  in
  
  let rec refine_cube pre_reg cube_list split_preds = 
    (* Jhala: this next line is very slow and should be optimized *)
    let has_no_value (x,wp,wpn) = List.for_all (fun y -> x <> fst y) cube_list in
    let preds_to_split_on = List.filter (has_no_value) split_preds in 
    match preds_to_split_on with
      [] -> make_conjunct_bdd cube_list
    | (h,wp,wpneg)::t -> 
	begin
	  let cube_plus = (h,true)::cube_list in
	  let pre_reg_plus = wp::pre_reg in
	  let conj_plus_ptr = ref cube_plus in
	  let _ = _check_Pred_list pre_reg_plus preds_to_split_on conj_plus_ptr in
          let plus_bdd = refine_cube pre_reg_plus !conj_plus_ptr preds_to_split_on in
  
          let cube_minus = (h,false)::cube_list in
	  let pre_reg_minus = wpneg::pre_reg in
	  let conj_minus_ptr = ref cube_minus in
	  let _ = _check_Pred_list pre_reg_minus preds_to_split_on conj_minus_ptr in
          let minus_bdd = refine_cube pre_reg_minus !conj_minus_ptr preds_to_split_on in
          
          CaddieBdd.bddOr plus_bdd minus_bdd
         end
  in
  (* let custom_post_cube assumptions  =
    Message.msg_string Message.Debug "In custom_post_cube";
    let asm_conj = Predicate.conjoinL assumptions in
    let _ = Stats.time "simplify_assume" (TheoremProver.simplify_assume ()) asm_conj in
    let is_sat (pi,wp,wpnp) = 
    (* check if this pred is satifiable in the post - returns a boolean *)
       let query = Predicate.implies (Predicate.conjoinL (wp::assumptions)) Predicate.False in
       let ac_query = Predicate.implies wp Predicate.False in 
         not (askTheoremProverContext query ac_query) 
         (* if the predicate is NOT valid iff fmla is satisfiable *) 
    in
    let _process_var_preds l = 
      List.fold_left (fun x y -> CaddieBdd.bddOr x (getPredIndexBdd (Misc.fst3 y))) bddZero (List.filter is_sat l)
    in
     let rv =  List.fold_left (fun x y -> CaddieBdd.bddAnd (_process_var_preds y) x) bddOne (!custom_preds_wp)
    in
       TheoremProver.simplify_pop (); 
       Message.msg_string Message.Debug "Leaving custom_post_cube";
       rv
    (* Jhala: ideally i should only fire queries for those variables that are affected by the cmd *)
  in
*)

  let _check_Cube _pred_list (cube : int array)  =
    (Message.msg_string Message.Debug (" In _check_Cube"));
    statReachedCubes := !statReachedCubes + 1;
     conjunctL := [];   
     let _ = 
       if (dc_flag) then Misc.array_reset dontcareArray false;
       if (craig_flag) then 
	 begin
	   Misc.array_reset dontcareArray true; (* RJ: whoa! *)
	   List.iter (fun (i,_,_) -> Array.set dontcareArray i false) pred_list
	   (* make sure we care about the predicates that we care about *)
	 end
     in
       (* if (_syntaxCheck cube modeledPredIndexList) then 
	  () else --syntactic heuristic *)
     let cubePostIsEmpty = 
       match (!assume_pred_ref) with 
           Some(e) -> (askTheoremProver (Predicate.implies (Predicate.conjoinL [(convertCubeToPred cube);e]) Predicate.False))
         | None -> false
     in 
       if (cubePostIsEmpty) then ()
       else 
         begin   
           let pref = match (!assume_pred_ref) with None -> ref [] | Some(e) -> ref [e] in
           let _plist = ref _pred_list in
	begin
          for i = 0 to (Array.length cube - 1) do
            if (cube.(i)==2) 
              then 
		begin
		  try
		    if ((dc_flag) && (outOfScopeArray.(i)))  then
		      begin
			Message.msg_string Message.Debug ("DC about: "^(Predicate.toString (getPred i))^" \n"); 
			Array.set dontcareArray i true
		      end
		  with _ -> ()
		end
	      else 
		begin
		  let i_pred = getPred i in
		(* Assume the invariant that i_pred : Predicate.Atom *)
                  let i_unchanged = (Hashtbl.mem unchangedPredTable i)  
		  in
		    if (cube.(i)==1) then
		      begin
			pref := i_pred::(!pref);
			if (i_unchanged && not(dontcareArray.(i))) 
			then conjunctL := (i,true)::(!conjunctL);
		      end
		    else
		      begin
			pref := (Predicate.negate i_pred)::(!pref);
			if (i_unchanged && not(dontcareArray.(i))) 
			then conjunctL := (i,false)::(!conjunctL);
		      end
		end
          done;
	 
	 (* Now ask the thmprover if pref -> each fellow in predlist *)
          if (Stats.time "_check_Pred_list" (_check_Pred_list !pref)  changed_pred_list conjunctL) 
          then 
            begin
	      (* Jhala: I'm being paranoid about the soundness and aliasing. something to assuage me *)
	      let _ =
		match (Options.getValueOfBool "scheck") with
		    false -> ()
		  | true ->
		      begin
			let cube_l = List.map 
				       (fun (i,b) -> if b then (getPred i) else Predicate.Not(getPred i))
				       !conjunctL 
			in
			  if (TheoremProver.queryExp (Predicate.Implies ((Predicate.And cube_l),(Predicate.False))) && (!assume_pred_ref = None)) 
			  then 
			    begin
			      Message.msg_string Message.Debug "SCHECK: Post of block is FALSE !!";
			      Message.msg_string Message.Normal "SCHECK: Post of block is FALSE !!";

			      Message.msg_string Message.Error  "SCHECK: Post of block is FALSE !!";
			      ()
			    end
			  else 
			    ()
		      end
	      in
               (*  this is where we plug in the post_split_cube *)
	       let splitup_bdd = 
		 if (false (*Options.getValueOfBool "cref"*)) then (refine_cube !pref !conjunctL post_split_preds)
		 else (make_conjunct_bdd !conjunctL) 
	       in
		  sumofcubesBdd := CaddieBdd.bddOr (splitup_bdd) (!sumofcubesBdd)                
		    (*
		      let os_succ_cube = (make_conjunct_bdd !conjunctL) in
		      if (Options.getValueOfString "custom" = "") 
		      then 
		      sumofcubesBdd := CaddieBdd.bddOr (make_conjunct_bdd !conjunctL) (!sumofcubesBdd)                
		      else
		      begin
                      let custom_succ_bdd = (custom_post_cube !pref) in
		      sumofcubesBdd := CaddieBdd.bddOr 
                      (CaddieBdd.bddAnd custom_succ_bdd (make_conjunct_bdd !conjunctL)) 
                      (!sumofcubesBdd); 
                      end
		      *)
  	     end; () 
  (* !conjunctL *)
        end;
      end
  in
  begin
    
    if (CaddieBdd.bddEqual bdd bddZero) then
      sumofcubesBdd := bddZero
    else 
      begin
	if (CaddieBdd.bddEqual bdd bddOne) 
	then
          let _tmpcube = (* build a temporary cube with nothing in it *)
            Array.make (*size *)(getCurrentIndex ()+2) (* content *)2 
	      
(* Note: The size is certainly a vast	overestimate, but this
   is the easiest way. The "+2" removes additional complications relating
   to index being -1 initially. Note that this is a fake cube -- we should
   never try to print it or convert it to a predicate!Ugly hack, but 
   works. So far *)
          in
	  _check_Cube changed_pred_list _tmpcube
	else
          ignore (Stats.time "bddForEachCube-post" (CaddieBdd.bddForeachCube bdd) (_check_Cube changed_pred_list));
      end;
    (Message.msg_string Message.Debug ("Returning from Post-Bdd = "^(Predicate.toString (convertBddToPred !sumofcubesBdd))));
    !sumofcubesBdd 
  end

let wp_table = Hashtbl.create 1009 
let stackInfo = ref None
let my_loc_print l = 
  let (i,j) = C_SD.location_coords l in
    Message.msg_string Message.Debug (Printf.sprintf "location (%d,%d)" i j)

	  
let rec abstract_data_post_graf_saidi abst_data_reg cmd edge post_location phi_table =
  Message.msg_string Message.Debug "In abstract_data_post GRAF_SAIDI_POST -- arguments are:" ;
  Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;
  
  Message.msg_printer Message.Debug C_SD.Command.print cmd ;
  let src,dst = (C_SD.get_source edge,C_SD.get_target edge) in
  let _ = my_loc_print src in
  (*
    let optional_pred = match cmd.C_SD.Command.code with
    C_SD.Command.Havoc sym -> 
    let new_name v = 
    if Symbol.compare v C_SD.tid = 0 then C_SD.notTid else v
    in
    let phi = Region.rename_vars !(Hashtbl.find phi_table sym) new_name in
    let phi_pred = Region.region_to_predicate phi in
    let inv_pred = Region.region_to_predicate !assumed_invariants_region in 
    let p = Predicate.conjoinL [phi_pred; inv_pred] in
    Some (p)
    | _ -> None 
    in
  *)
  let (target_loc (*,_*)) = C_SD.location_coords post_location (* (C_SD.get_target edge) *) in
  let _ = Message.msg_string Message.Debug (Printf.sprintf "absdatapost postloc: (%d,%d)" (fst target_loc)  (snd target_loc)) 
  in
  let craig_flag = Options.getValueOfInt "craig" = 2 in
  let craig_flag1 = Options.getValueOfInt "craig" = 1 in
  let craig_preds () = 
    try
      Hashtbl.find loc_pred_table target_loc (* doesnt work if this isnt the real successor! *)
    with
	Not_found -> []
  in
  let predlist =      
    if (Options.getValueOfBool "forget") 
    then abst_data_reg.Region.preds 
    else 
      if (craig_flag) 
      then 
	craig_preds ()
      else (List.map Misc.fst3 !tableOfPreds) 
  in
  let _H_cmd _cmd bdd actual_predlist =
    Message.msg_string Message.Debug ("In _H of abstract_data_post, bdd = "^(Predicate.toString (convertBddToPred bdd)));

    (* optimization: check if (abst_data_reg, p) is satisfiable? *)
    (* so here is where we shall use stackInfo -- the format is (top of stack,2nd top of stack, call op) *)

    let (tos,tos2,call) = match (!stackInfo) with
	Some (t,t2,c) -> (t,t2,c) 
      | None -> failwith "scope info NOT supplied !"
    in
    let _ = Message.msg_string Message.Debug (Printf.sprintf "tos=%s , tos2=%s , call=%s" tos tos2 (string_of_bool call)) in
    let out_of_scope pred = 
      if (Options.getValueOfBool "scope") then
	let _stay_ff x = 
	  let scope_x = (extract_scope x) in
	    ((tos = scope_x) || (call && tos2 = scope_x) || (scope_x = "__BLAST_GLOBAL"))
	in
	let predVars = getVarExps_deep pred in
	  List.exists (fun x -> not (_stay_ff x)) predVars
      else 
	true
    in
    let make_pred_triple () a = 
      let pr = getPred a in
        match (!assume_pred_ref) with 
	    None ->
              let cr = concrete_data_pre false false { Region.pred = pr ; } _cmd edge phi_table in
              let wppos = cr.Region.pred in
              let ncr = concrete_data_pre false false { Region.pred = Predicate.negate pr ; } _cmd edge phi_table in
              let wpneg = ncr.Region.pred in
		
                (*if (wppos,wpneg) = (_check_wppos,_check_wpneg) then*)
                (a,wppos,wpneg, (out_of_scope pr))
                (* else failwith ("bad wp_table")*)
          | Some(e) -> (a,pr,Predicate.negate pr,(out_of_scope pr) )
    in
      
    (*let _ = if not (Options.getValueOfString "custom" = "") then  
      custom_preds_wp := List.map ( function x -> 
      (List.map (function y -> 
							 let (a,b,c,d) = (make_pred_triple () (Misc.thd3 y)) in (a,b,c) ) x)
      ) !custom_predicate_list
      in *)
      
    let post_split_preds = 
      if craig_flag1 then craig_preds ()
      else []
			     (*if (Options.getValueOfBool "cref") 
			       then (try Hashtbl.find post_split_table post_location with Not_found -> []) 
			       else [] *)
    in
      Stats.time "postBdd" (postBdd bdd  (List.map ( Stats.time "mk_pred_triple" make_pred_triple ()) actual_predlist)) (List.map (make_pred_triple ()) post_split_preds)
  in
  let _H = _H_cmd cmd in
    
    match cmd.C_SD.Command.code with 
	C_SD.Command.FunctionCall fcallExp ->
	  let (fname,_subs,_) = (deconstructFunCall fcallExp) in
	    if (fname = __NotImplementedFunctionName) then abst_data_reg
	    else 
         let asgn_block =
           { cmd with C_SD.Command.code = C_SD.Command.Block (List.map (fun (x,e) -> C_SD.Command.Expr( Expression.Assignment (Expression.Assign,Expression.Symbol(x), e ))) _subs) }
         in
           { 
      	     Region.preds = abst_data_reg.Region.preds;
      	     Region.bdd =  _H_cmd asgn_block (abst_data_reg.Region.bdd) predlist (*abst_data_reg.Region.preds*);
	   }
	   
      (* What was just done for the function call ... *) 
	   (* Push stuff onto the stack: ... in the rest of post 
	      1. edge, on return, the location will be
	      the other end of the edge ...
	      2. if edge was a y = f(x) then post on return reflects this *)
	   (* Go through list of predicates and compute "copies" *)
	   (* Compute BDD of reg ^_i (pred_i = copy_pred_i) *)
	   (* Start location = start location of function *)
      | C_SD.Command.Skip ->  abst_data_reg
	  
      | C_SD.Command.Block [] ->  abst_data_reg
	  
      | C_SD.Command.Block l  -> 
	  if ( not (List.exists (function x -> match x with C_SD.Command.Return e -> true | _ -> false) l))
	  then 
	    let _ = (retTarget := junkRet) in
	      { 
      		Region.preds = abst_data_reg.Region.preds;
      		Region.bdd =   _H  (abst_data_reg.Region.bdd) predlist (*abst_data_reg.Region.preds*);
	      }
	  else 
	    let (targ, fname, _) = lastTarget () in
	      (* Jhala: the t.o.s. is passed using last_loc_pointer  reg.Region.stack *) 
	    let _ = (retTarget := targ) in
	    let (newPreds,delPreds) = removeFunPreds fname abst_data_reg.Region.preds in
            let _ = Message.msg_string Message.Debug "Return: Calling _H" in
	    let newBdd = (_H  (abst_data_reg.Region.bdd) predlist (*abst_data_reg.Region.preds*)) in
            let _ = Message.msg_string Message.Debug "Return: Back from _H" in
	    let _predToVar i =  -1 in (* JHALA: To be done ... ??? *)
	    let ex_newBdd = newBdd in
	      (*try CaddieBdd.exists (Bdd.fromList bddManager (List.map
		_predToVar
		delPreds)) newBdd 
		with _ -> newBdd (* Jhala: TBD... how to go from pred to
		bdd variable (int) ? wrapper needs
		work *)*)
	      
	    let _ = (retTarget := junkRet) in
	      { 
      		Region.preds = newPreds@delPreds;
      		Region.bdd =  ex_newBdd;
	      } 
      | C_SD.Command.Pred e -> 
	  
	  if (e = Predicate.True) then ( Message.msg_string Message.Debug "thinks its true!";abst_data_reg)
	  else 
            begin
              assume_pred_ref := Some(e);
	      Message.msg_string Message.Debug "calling _H";
              let retReg = 
		{ 
		  Region.bdd = _H (abst_data_reg.Region.bdd) predlist (*abst_data_reg.Region.preds*) ;
		  Region.preds = abst_data_reg.Region.preds ;
		}
              in  
		assume_pred_ref := None;          
		retReg
            end
      | C_SD.Command.Phi (sym, flag) -> 
	  let e = phi_command_to_pred sym flag phi_table in
  (* Now treat just like C_SD.Command.Pred e *)
  if (e = Predicate.True) then abst_data_reg
  else 
    begin
      assume_pred_ref := Some(e);
      let retReg = 
        { 
          Region.bdd = _H (abst_data_reg.Region.bdd) predlist (* abst_data_reg.Region.preds*) ;
          Region.preds = abst_data_reg.Region.preds ;
        }
      in  
        assume_pred_ref := None;          
        retReg
    end
| C_SD.Command.Havoc sym -> 
    let new_var = 
      match getNewTmpVar (Expression.Symbol "__foo") with 
	  Expression.Chlval (Expression.Symbol x, s) -> Expression.Lval (Expression.Symbol (s^x))
	| _ -> failwith "Abstraction.concrete_data_pre: Havoc: TBD"
    in
    let assign = Expression.Assignment(Expression.Assign, sym, new_var) in
    let stmt = C_SD.Command.Expr assign in
    let cmd' = {cmd with C_SD.Command.code = C_SD.Command.Block [stmt]} in
      abstract_data_post_graf_saidi abst_data_reg cmd' edge post_location phi_table
| C_SD.Command.HavocAll -> failwith "Abstraction.abstract_data_post_graf_Saidi: HavocAll: TBD"       
    

   (* Starting the definition of PRE *)


   let setWPCapRegion reg =
     let conc_reg = Region.concretize reg in
       wpCapRegion :=
          match conc_reg with 
             Region.Empty -> Predicate.False
           | Region.Atomic atomic_reg -> 
                begin
		  match atomic_reg.Region.data with
                    Region.Concrete foo -> foo.Region.pred
                  | _ -> failwith ("bad call to setWPCapRegion")
		end
	   | Region.ParallelAtomic p_atomic_reg ->
	       begin
		 match p_atomic_reg.Region.p_data with
		   Region.Concrete foo -> foo.Region.pred
		 | _ -> failwith ("bad call to setWPCapRegion")
	       end
	   | _ -> failwith ("bad call to setWPCapRegion: Strange region type")

  let resetWPCapRegion () = 
     wpCapRegion := Predicate.True
 
  (* the symbolic pre function: (approximate) weakest pre-condition *)
   let pre_template subs_flag sat_flag reg op phi_table =
     Message.msg_string Message.Debug "In pre -- arguments are:" ;
     Message.msg_printer Message.Debug Region.print reg ;
     Message.msg_printer Message.Debug Operation.print op ;
     Message.msg_string Message.Debug (if subs_flag then "Substitute \n" else "Explicit (no substitute) \n");
     Message.msg_string Message.Debug (if sat_flag then "sat-filter-on \n" else "sat-filter-off \n");
    match reg with
        Region.Empty ->
          Region.bot
      | Region.Atomic atomic_reg ->
          if ((atomic_reg.Region.location <> (C_SD.get_target op)) && (not (isReturn op)) && (not (isFunCall op))) then
            (* Jhala: This is only if this is not a return op *)
            begin
              Message.msg_string Message.Error "WARNING: pre: blast internal error suspected" ;
              Message.msg_string Message.Error "              operation do not match this region's location" ;
	      failwith "location mismatch";
              Region.bot
            end
          else
            begin
              let new_stack =
                if (isFunCall op) then
                  begin 
                    match (C_SD.get_command op).C_SD.Command.code with
                   
                    C_SD.Command.FunctionCall fcexp -> 
                    let function_is_defined fcexp = 
                       match fcexp with
                         Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), _) -> 
                           C_SD.is_defined fname
                       | Expression.Assignment (_, _, Expression.FunctionCall (Expression.Lval (Expression.Symbol fname), _)) -> 
                           C_SD.is_defined fname
                       | _ -> 
                               (* Rupak: this case happens if there is a function pointer call.
				  Previously we failed here with:
				  failwith ("Bad function call :"^(Expression.toString fcexp))
				  However we can punt this by assuming that the function pointer call
				  does not call a known function, and continue.
				  
				  Do this only if nofp is set.
				  *)
			   if (Options.getValueOfBool "nofp") then
			     false
			   else
			       failwith ("pre_template: Bad function call (Possible function pointer?):"^(Expression.toString fcexp))
                    in
                    if (not (function_is_defined fcexp)) then
                       atomic_reg.Region.stack
                    else 
                    begin
                      let (_fname,_,target) = deconstructFunCall fcexp in
                      if (_fname = __NotImplementedFunctionName) 
                      then 
                         atomic_reg.Region.stack
                      else   
                            if (_fname <> __BLAST_DummyFunctionName) then 
                               try (Region.CallStack (List.tl (Region.list_of_stack atomic_reg.Region.stack))) 
                               with e -> (print_string (Printexc.to_string e);failwith "Incorrect callstack type !!")
                            else 
                               atomic_reg.Region.stack (* was a skip *)
                               (*begin
                               let vars_in_support = List.flatten (List.map getVarExps_deep (preds_of_region reg)) in   
                               if (not (List.mem target vars_in_support)) then 
                                  atomic_reg.Region.stack 
                               else failwith ("Undeclared important function"^_fname)
                               end*)
                    end
                   |    _ -> failwith "This can never happen: isFuncall checked that this was a function call!"
                   end
                 else
                  if (isReturn op)
                  then
                    begin
                    Message.msg_string Message.Debug "Pre: Hit isReturn\n";
                    (* recall that we put the call points in the stack *)
                    (* we assume that a function call edge is the only edge
                       going to its target location *)
                    Message.msg_string Message.Debug "Ingoing edges : ";
                    List.iter (fun a -> Message.msg_string Message.Debug (Operation.to_string a)) (C_SD.get_ingoing_edges atomic_reg.Region.location);
                    Message.msg_string Message.Debug "\nNow matching...\n";
                    let rec get_fn_call_edge lst =
                       match lst with
                         a::rest -> 
                           begin
                             if (isFunCall a) then
                               (* check there are no other function calls *)
                               if (List.exists isFunCall rest) then
                                  failwith "pre: More than one function call in incoming edges!"
                               else a
                             else
                               get_fn_call_edge rest
                           end
                       | [] -> failwith ("pre: No corresponding function call found")
                    in
                    let callLoc = (C_SD.get_source (get_fn_call_edge (C_SD.get_ingoing_edges atomic_reg.Region.location))) in
                    last_loc_pointer := Some(callLoc); retTarget := Misc.fst3 (lastTarget ()); last_loc_pointer := None;
                            
                    ( Region.CallStack (callLoc::(Region.list_of_stack atomic_reg.Region.stack)))
                    
                    end
                  else
                    atomic_reg.Region.stack
              in
              let new_location = (C_SD.get_source op)
              and new_data =
                match atomic_reg.Region.data with
                    Region.Abstract abst_data_reg ->
                      Region.Abstract (abstract_data_pre abst_data_reg (C_SD.get_command op) op phi_table)
                  | Region.Concrete conc_data_reg ->
                      Region.Concrete (concrete_data_pre subs_flag sat_flag conc_data_reg (C_SD.get_command op) (op) phi_table)
              in
                retTarget := junkRet;
                Region.Atomic { Region.location = new_location ;
                                Region.data = new_data ;
                                Region.stack = new_stack ; }
              
            end
      |	Region.ParallelAtomic p_atomic_reg ->
	  Region.bot
	    (* i am here *)
      | Region.Union _ ->
          failwith "pre: argument is a union: unimplemented"
      | Region.ParallelUnion _ ->
          failwith "pre: argument is a parallel union: unimplemented"


(* Jhala: am setting sat flag to false for now ... *)
   let pre (* reg op phi_table *) = pre_template true false (* reg op phi_table *)
   let spre (* reg op phi_table *) = pre_template false false (* reg op phi_table *)
   
   
   let abstract_data_post abst_data_reg  cmd edge post_location phi_table = 
     match (Options.getValueOfString "post") with
         "H" -> (* Stats.time "Post H: " *) (abstract_data_post_das_dill abst_data_reg cmd) edge post_location phi_table
       | _  -> (* Stats.time "Post Graf-Saidi"*) (abstract_data_post_graf_saidi abst_data_reg  cmd) edge post_location phi_table 
	   
   let kill_filter_fun x  = Hashtbl.mem killPredTable (getPred x) 
			      
   let hg_exists kill_ff bdd = 
     let soc_bdd = ref bddZero in
     let make_cube_bdd cube = 
       print_string "in make_cube";
       let tmp_cube_bdd = ref bddOne in
	 for i = 0 to (Array.length cube - 1) do
	   if (cube.(i)==2 || (kill_ff i) ) 
           then 
	     begin
	       if (not (cube.(i)==2)) then Message.msg_string Message.Debug ("killing pred: "^(Predicate.toString (getPred i))^"\n")
             end
           else 
	     begin
   	       if (cube.(i)==1) then 
		 tmp_cube_bdd := CaddieBdd.bddAnd (!tmp_cube_bdd) (getPredIndexBdd i)
	       else
		 tmp_cube_bdd := CaddieBdd.bddAnd (!tmp_cube_bdd) (CaddieBdd.bddNot (getPredIndexBdd i))
	     end
	 done;
	 soc_bdd := CaddieBdd.bddOr (!soc_bdd) (!tmp_cube_bdd)
     in
       (Message.msg_string Message.Debug ("Calling hg_exists = "^(Predicate.toString (convertBddToPred bdd))));
       CaddieBdd.bddForeachCube bdd make_cube_bdd;
       (Message.msg_string Message.Debug ("Bye bye hg_exists = "^(Predicate.toString (convertBddToPred !soc_bdd))));
       !soc_bdd

   let scope_kill data_region fname = 
     Message.msg_string Message.Debug "Inside scope_kill:\n";
     match data_region with
	 Region.Abstract(abst_data_reg) ->
	   begin
	     Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg;
	     let _kill_ff x =  
	       let predVars = getVarExps_deep (getPred x) in
		 List.exists (fun x -> Misc.is_substring x ("@"^fname)) predVars
	     in
	     let new_bdd = hg_exists _kill_ff abst_data_reg.Region.bdd in
	     let ret_abst_data_reg  = { Region.bdd = new_bdd; Region.preds = abst_data_reg.Region.preds} in
	       Message.msg_string Message.Debug "Returning from kill_data_region with :\n";
	       Message.msg_printer Message.Debug Region.print_abstract_data_region ret_abst_data_reg;
	       Region.Abstract(ret_abst_data_reg)
           end
       | _ -> 
	   begin
	     Message.msg_string Message.Debug "bad call to kill"; 
	     data_region
	   end
	   	 
   let kill_data_region data_region = 
     Message.msg_string Message.Debug "Inside kill_data_region with :\n";
     
     match data_region with
	 Region.Abstract(abst_data_reg) ->
	   begin
	     Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;	   
	     (*	     let _preds_in_bdd =  (CaddieBdd.bddSupportSet abst_data_reg.Region.bdd) in
 	     
		   let _preds_to_quantify = List.filter kill_filter_fun _preds_in_bdd in
		   let _ = Message.msg_string Message.Debug "Will quantify";  
		   List.iter (function i -> Message.msg_string Message.Debug  (Predicate.toString (getPred i))) _preds_to_quantify 
		   in
		   
		   let _varset = CaddieBdd.fromList bddManager _preds_to_quantify in 
		   let new_bdd = CaddieBdd.exists _varset abst_data_reg.Region.bdd in 
   *)
	     let new_bdd = hg_exists kill_filter_fun abst_data_reg.Region.bdd  in
             let ret_abst_data_reg  = { Region.bdd = new_bdd; Region.preds = abst_data_reg.Region.preds} in
	       Message.msg_string Message.Debug "Returning from kill_data_region with :\n";
	       Message.msg_printer Message.Debug Region.print_abstract_data_region ret_abst_data_reg;
	       Region.Abstract(ret_abst_data_reg)
           end
         | _ -> 
	     begin
	       Message.msg_string Message.Debug "bad call to kill"; 
	       data_region
	     end
	 
            
   (* the symbolic post function: (approximate) strongest post-condition *)
   (* TODO: if checking races, add havocs *)
   let rec post reg op phi_table =
     Message.msg_string Message.Debug "In post -- arguments are:" ;
     Message.msg_printer Message.Debug Region.print reg ;
     Message.msg_printer Message.Debug Operation.print op ;
     let _ = Hashtbl.replace statVisitedFunctionTable (C_SD.get_location_fname (C_SD.get_target op)) true 
     in
     let preds_of_region reg = 
       match reg with
	   Region.Atomic atr -> 
             begin     
               match atr.Region.data with 
		   Region.Abstract abst_data_reg -> List.map getPred abst_data_reg.Region.preds
		 | _ -> failwith "Bad call to preds_of_region"
             end
	 | Region.ParallelAtomic p_atr ->
             begin     
               match p_atr.Region.p_data with 
		   Region.Abstract abst_data_reg -> List.map getPred abst_data_reg.Region.preds
		 | _ -> failwith "Bad call to preds_of_region"
             end
	 | _ -> failwith "Bad call to preds_of_region"
     in
     match reg with
       Region.Empty -> Region.bot
     | Region.Atomic atomic_reg ->
	 begin
           try 
	     if (atomic_reg.Region.location <> (C_SD.get_source op)) then
	       begin
		 Message.msg_string Message.Error "WARNING: post: blast internal error suspected" ;
		 Message.msg_string Message.Error "               operation do not match this region's location" ;
		 
		 Region.bot
	       end
	     else
	       begin
                 try 

		       let callStack = Region.list_of_stack atomic_reg.Region.stack in
	               let tos_fname = 
			 if (not (callStack = [])) 
			 then (last_loc_pointer := Some (List.hd callStack);C_SD.get_location_fname (C_SD.get_target op)) 
			 else ( last_loc_pointer := None; __BLAST_DummyFunctionName) 
		       in
		       let tos_2_fname = 
			 if (List.length callStack >= 1) then (C_SD.get_location_fname (List.hd callStack)) else __BLAST_DummyFunctionName 
		       in
		       let _ = Message.msg_string Message.Debug (Printf.sprintf "tos_fname: %s, tos_2_fname: %s" tos_fname tos_2_fname) in

 		       (* So last_loc_pointer points to the top of the stack at all times or to errorloc Rif the stack is empty *)
                 let (new_location, new_stack) = 
                    
                    if (isFunCall op) then
                          begin
                           match (C_SD.get_command op).C_SD.Command.code with
                            C_SD.Command.FunctionCall fcexp -> 
                          
                               let (_fname,_,target) = deconstructFunCall fcexp in
(*RUPAK*)			       let _ = stackInfo := Some (_fname,tos_fname,true) in(*endRUPAK*)
                               if (_fname = __NotImplementedFunctionName) 
                               then 
				 ((C_SD.get_target op), atomic_reg.Region.stack)
                               else   
				 begin
                                   if (_fname <> __BLAST_DummyFunctionName) then
				     (let newStack =
				       (C_SD.get_source op) :: callStack
				     in
                                     ((C_SD.lookup_entry_location _fname),
				      Region.CallStack newStack))
                                   else 
                                     begin
                                       let _wrap_getvars _pred = List.map (function x -> Expression.Symbol x) (getVarExps_deep _pred) in 
                                       let vars_in_support = List.flatten (List.map _wrap_getvars (preds_of_region reg)) in   
                                       if (not (List.mem target vars_in_support)) then 
                                         ((C_SD.get_target op),atomic_reg.Region.stack) (* is a skip *)
                                       else failwith ("Undeclared important function"^_fname)
                                     end
				 end
                           | _ -> failwith "what the (post) !"
				 
                          end
                    else 
                      try 
			if (isReturn op) then 
			  begin
			    let _ = stackInfo := Some (tos_fname,tos_2_fname,true) in
			   (*RUPAK*)let (ret, _, nextLoc) = lastTarget () in
			   retTarget := ret ;
			   let newStack =
			       List.tl callStack
			   in
			   (nextLoc,(Region.CallStack newStack) ) 
			  end
			else 
			  begin
			    let _ = stackInfo := Some (tos_fname,tos_2_fname,false) in
			    (C_SD.get_target op,atomic_reg.Region.stack) 
			  end
                      with Failure ("tl") -> 
			begin
			  let _ = stackInfo := Some (tos_fname,tos_2_fname,false) in
			  (C_SD.get_target op, atomic_reg.Region.stack ) 
		      end
   in 
   let get_new_data () () =
     let _new_data =
       try 
         match atomic_reg.Region.data with
             Region.Abstract abst_data_reg ->
               Region.Abstract (abstract_data_post abst_data_reg (C_SD.get_command op) op new_location phi_table)
	   | Region.Concrete conc_data_reg ->
               Region.Concrete 
	       (concrete_data_post conc_data_reg (C_SD.get_command op) op phi_table)
         with C_SD.NoSuchFunctionException f -> failwith "Foo"
     in
       (* Adding scoping hack *)
       (* Adding kill *)
       match ((Options.getValueOfBool "scope") && (isReturn op) (*&& false*) ) with
	   true  -> 
	     begin
	       Message.msg_printer Message.Debug C_SD.print_location (C_SD.get_source op);
	       Message.msg_printer Message.Debug C_SD.print_location new_location;
	       Stats.time "scope_kill" (scope_kill _new_data) (Misc.snd3 (lastTarget ()))
	       end
	 | false -> _new_data
   in
     (*RUPAK*)
     (* set retTarget back to junk *)
     (*retTarget := junkRet ;*)(*end RUPAK *)
   let new_data = Stats.time "get_new_data" (get_new_data ()) () in
   let _ = retTarget := junkRet in
     begin
       (*Message.msg_string Message.Debug "Post: No kill location \n";*)
       Region.Atomic { Region.location = new_location ;
                       Region.data = new_data ;
                       Region.stack = new_stack ; }
     end
     (*
       | Some (l) -> 
	   begin
	     Message.msg_string Message.Debug "Post: kill location will compare:";
	     Message.msg_printer Message.Debug C_SD.print_location new_location;
	     Message.msg_printer Message.Debug C_SD.print_location l;
	     
             if (new_location == l) then 
	       (* Calling kill_data_region *)
	       let _ = killCounter := (!killCounter) + 1 in
		 ( Region.Atomic { Region.location = new_location;
				   Region.data = (* kill_data_region [Jhala: removed] *) new_data;
				   Region.stack = new_stack; })
	     else
	       Region.Atomic { Region.location = new_location ;
                               Region.data = new_data ;
                               Region.stack = new_stack ; }
	   end
*)
		   with
         Failure("Empty Call Stack") -> Region.bot
    end
   with 
            C_SD.NoSuchFunctionException f -> (* treat it as a skip *)
               Message.msg_string Message.Minor ("Unexpected NoSuchFunctionException "^f);
               (* THIS SHOULD BE COMMENTED OUT IF UNDEFINED FUNCTIONS ARE TO BE AVOIDED *)
               (* failwith ("Unexpected NoSuchFunctionException "^f); *)
               Region.Atomic
                { Region.location = C_SD.get_target op;
                 Region.data = atomic_reg.Region.data;
                 Region.stack = atomic_reg.Region.stack;
               } 
               (* COMMENT OUT UPTO HERE *)

          end
      | Region.ParallelAtomic p_atomic_reg ->
          assert ((List.length p_atomic_reg.Region.p_control) >= 0 + 1); (* make sure our operation is in range *)
	  let new_data_region = ref p_atomic_reg.Region.p_data in 
	  let new_control_region = ref [] in
		for i = 0 to List.length p_atomic_reg.Region.p_control do
		  if (i <> 0)
		  then
		    (* copy control region without any change, data doesn't change either *)
		    new_control_region := List.append !new_control_region [(List.nth p_atomic_reg.Region.p_control i); ]
		  else
		  begin
		    (* modify the control region and data region with a recursive post call *)
		    let this_control = List.nth p_atomic_reg.Region.p_control i in 
		    let fake_atomic_region = Region.Atomic { Region.location = this_control.Region.p_location;
							     Region.data     = !new_data_region;
							     Region.stack    = this_control.Region.p_stack;
							   } in
		    let result_atomic_region = 
			match (post fake_atomic_region op phi_table) 
			with 
			Region.Atomic p -> p
			| _ -> failwith "post: called post with atomic_region, but did not return atomic region"
		    
		    in
		    new_control_region := List.append !new_control_region [ { Region.p_location = result_atomic_region.Region.location;
								            Region.p_stack    = result_atomic_region.Region.stack;}; ];
		    new_data_region := result_atomic_region.Region.data;
		  end;
		done;

		Region.ParallelAtomic
		{ Region.p_control = !new_control_region;
		  Region.p_data    = !new_data_region;
		}

      | Region.Union _ ->
          failwith "post: argument is a union: unimplemented"
      | Region.ParallelUnion _ ->
          failwith "post: argument is a parallel union: unimplemented"

let print_debug_info oplist = 
   let rec get_funs ops lst = 
     match ops with 
       [] -> lst
     | op::rest ->
       begin
         match (C_SD.get_command op).C_SD.Command.code with
           C_SD.Command.FunctionCall fcexp -> 
              begin
               match fcexp with
                 Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), _) -> 
                    if (C_SD.is_defined fname) then get_funs rest lst 
                    else get_funs rest (fname::lst)
               | Expression.Assignment (_, _, Expression.FunctionCall (Expression.Lval (Expression.Symbol fname), _)) -> 
                    if (C_SD.is_defined fname) then get_funs rest lst 
                    else get_funs rest (fname::lst)
               | _ -> get_funs rest (("FnPtr::"^(Expression.toString fcexp)) :: lst)
              end
         | _ -> get_funs rest lst
       end
   in 
   let impfun = get_funs oplist [] in
   if (List.length impfun > 0) then 
     begin 
       Message.msg_string Message.Error "The following are suspect funs:\n";
       List.iter (fun x -> Message.msg_string Message.Error (x^" ,")) impfun ;
       Message.msg_string Message.Error "\n"
     end



  (*******************************************************************************************)

   let statsMaxNumberOfPredsActive = ref 0
					   (* Keeps track of the number of predicates active
                                              at any point, discounting scope
                                           *)

  (* Accelerate and focus *)

   (* the precise function required by the ABSTRACTION signature *)
   let precise = Region.concretize (* GILLES stuff : Region.concretize_map (fun l -> l) (fun s -> s) (fun p -> p) *)


   let check_emptiness op =
    if (Options.getValueOfString "refine" = "fwd") then
      isPredicate op
    else
     match (C_SD.get_command op).C_SD.Command.code with
       C_SD.Command.Pred Predicate.True 
     | C_SD.Command.Skip 
     | C_SD.Command.Block [] -> false
     | _ -> true
 

   (* check if cover check will be useful *)
   let cover_check_useful reg =
     if (Options.getValueOfString "cover" = "join") then
       (* return true if reg is a join point *)
       begin
         match reg with
           Region.Empty -> true
         | Region.Atomic ar -> 
              begin (* check coverability only if more than one edge comes into this location *)
               let incoming_edges =  C_SD.get_ingoing_edges ar.Region.location in
               (List.length (incoming_edges) > 1) ||
               List.mem ar.Region.location (List.map (C_SD.get_source) incoming_edges) (* to handle things like ERROR: goto ERROR; *)
              end
         | _ -> true
       end
     else
       true


   let coverable reg =
     match reg with
      Region.Empty -> true
      | Region.Atomic a -> C_SD.is_loopback a.Region.location
      | _ -> false

      
   (* returns true if reg1 is covered by reg2, false otherwise *)
   let covered reg1 reg2 =
     (* This is currently buggy...  Let's just call leq for now *)
     Region.leq reg1 reg2
     (*
     Message.msg_string Message.Debug "In covered -- arguments are:" ;
     Message.msg_printer Message.Debug Region.print reg1 ;
     Message.msg_printer Message.Debug Region.print reg2 ;

     let rec first_letters stack acc =
       match stack with
           []   ->
             List.rev acc
         | top::s ->
             if (List.memq top acc)
             then first_letters s acc
             else first_letters s (top::acc)
     and list_eq l1 l2 =
       match (l1, l2) with
           ([], []) -> true
         | (h1::t1, h2::t2) ->
             (h1 == h2) && (list_eq t1 t2)
         | _ -> false
     and first_letters_of_stack_contents stack_contents =
       match stack_contents with
           Region.CallStack s ->
             first_letters s []
         | Region.EveryStack ->
             failwith "covered: stack contents is surprisingly EveryStack: unimplemented"
     and bdd_of_data_region d =
       match d with
           Region.Abstract abst_data_reg ->
             abst_data_reg.Region.bdd
         | Region.Concrete _             ->
             failwith "covered: data region is surprisingly concrete: unimplemented"
     in
       match (reg1, reg2) with
           (Region.Empty, _) ->
             true
         | (_, Region.Empty) ->
             (* note that (Empty, Empty) is caught above *)
             false
         | (Region.Atomic atomic_reg1, Region.Atomic atomic_reg2) ->
             if ((atomic_reg1.Region.location == atomic_reg2.Region.location) &&
                 (list_eq (first_letters_of_stack_contents atomic_reg1.Region.stack)
                          (first_letters_of_stack_contents atomic_reg2.Region.stack)))
             then
               let bdd1 = bdd_of_data_region atomic_reg1.Region.data
               and bdd2 = bdd_of_data_region atomic_reg2.Region.data in
               let b = CaddieBdd.bddAnd bdd1 (CaddieBdd.bddNot bdd2)
               in
                 not (isSatisfiable (b,[]))
             else
               false
         | (Region.Atomic atomic_reg, Region.Union atomic_reg_table) ->
             let fl_atomic = first_letters_of_stack_contents atomic_reg.Region.stack in
             let bdd_atomic = bdd_of_data_region atomic_reg.Region.data in
             let bdd_union_with_same_fl =
               Region.AtomicRegionTable.fold
                 (fun (loc, stack_con) data_reg acc ->
                    if ((loc == atomic_reg.Region.location) &&
                        (list_eq (first_letters_of_stack_contents stack_con) fl_atomic))
                    then CaddieBdd.bddOr acc (bdd_of_data_region data_reg)
                    else acc)
                 atomic_reg_table
                 bddZero in
             let b = CaddieBdd.bddAnd bdd_atomic (CaddieBdd.bddNot bdd_union_with_same_fl)
             in
               not (isSatisfiable (b,[]))
         | (_, _) ->
             failwith "covered: argument types unexpected: unimplemented"
     *)

  (* do_focus : get new predicates to add to the current set *)
   exception AnnotationException  
   exception NoNewPredicatesException

   
   let predsFromForLoops b =
            (* Hack for 'for' loops 
               Here is what we do: for each temporary variable added, we find lower and upper bounds for
               the real variable "b" associated with the temporary. Then we add all predicates real = i
               for i in (lower, upper). This is a hack, so we do not try to be very clever about this.
               The hope is that simple cases will be caught.
            *)
     Message.msg_string Message.Error "";
     Message.msg_string Message.Error "";
     Message.msg_string Message.Error ("Warning: Adding predicates for a possible for loop involving index variable "^b);
     Message.msg_string Message.Error ("Warning: The current handling of such for loops in BLAST is extremely elementary");
     Message.msg_string Message.Error ("Warning: and is a major cause for state explosion.");
     Message.msg_string Message.Error ("Warning: We recommend that you manually remove the for loop from your program");
     Message.msg_string Message.Error ("Warning: by unrolling, or by replacing the concrete upper bound by an uninitialized variable.");
     Message.msg_string Message.Error ("Blast shall continue in the meantime.");
     Message.msg_string Message.Error "The Control key is at the bottom left of your keyboard, and C is in the third line of the alphabet, between X and V.";
     Message.msg_string Message.Error "";
     Message.msg_string Message.Error "";
     let find_bounds str = (* lower <= upper *)
       let preds_with_str = getPredsContainingVar str in
               let rec upper_lower str predList (u, l) =
                 match predList with 
                   [] -> (u,l)
                 | p :: rest ->
                    begin
                     match p with
                       Predicate.Atom (Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol s), Expression.Constant (Constant.Int a))) ->
                       if (s = str) then 
                         let newu = if (a > u) then a else u and newl = if (a < l) then a else l in
                         upper_lower str rest (newu, newl)
                       else upper_lower str rest (u,l)
                     | Predicate.Atom (Expression.Binary (Expression.Lt, Expression.Lval (Expression.Symbol s), Expression.Constant (Constant.Int a)))
                     | Predicate.Atom (Expression.Binary (Expression.Gt, Expression.Lval (Expression.Symbol s), Expression.Constant (Constant.Int a)))
                     | Predicate.Atom (Expression.Binary (Expression.Le, Expression.Lval (Expression.Symbol s), Expression.Constant (Constant.Int a)))
                     | Predicate.Atom (Expression.Binary (Expression.Ge, Expression.Lval (Expression.Symbol s), Expression.Constant (Constant.Int a))) 
                     | Predicate.Atom (Expression.Binary (Expression.Lt, Expression.Constant (Constant.Int a), Expression.Lval (Expression.Symbol s)))
                     | Predicate.Atom (Expression.Binary (Expression.Gt, Expression.Constant (Constant.Int a), Expression.Lval (Expression.Symbol s)))
                     | Predicate.Atom (Expression.Binary (Expression.Le, Expression.Constant (Constant.Int a), Expression.Lval (Expression.Symbol s)))
                     | Predicate.Atom (Expression.Binary (Expression.Ge, Expression.Constant (Constant.Int a), Expression.Lval (Expression.Symbol s))) ->
                       if (s = str) then 
                         let newu = if (a > u) then a else u and newl = if (a < l) then a else l in
                         upper_lower str rest (newu, newl)
                       else upper_lower str rest (u,l)
                     | _ -> upper_lower str rest (u,l)
                    end
                in
                upper_lower str preds_with_str (min_int, max_int)
            in
            let (u, l) = find_bounds b in
            if (l = min_int || u = max_int) then [] (* Uh oh -- trouble! *)
            else
            (let localPredList = ref [] in
            for i = l to u do
              localPredList := 
                (Predicate.Atom (Expression.Binary (Expression.Eq, 
                                                    Expression.Lval (Expression.Symbol b), 
                                                    Expression.Constant (Constant.Int i)))) :: !localPredList     
            done;
            !localPredList)
              
  
(******** extracting new predicates from **********)
(* from a cube *)

let getNewPreds orlist =
    let newp = 
      try 
	List.fold_left 
	  (function prlst -> function p -> prlst @ (Stats.time "getUsefulPreds" TheoremProver.getUsefulPredicates p)) [] orlist 
      with Failure "simplify" -> (Message.msg_string Message.Debug "Simplify exception!"; raise NoNewPredicatesException)
    in
      match newp with 
	  [] ->
            begin
              Message.msg_string Message.Error  "\n do_focus: Vampyre bug: cannot prove region to be empty!\n" ;
              Message.msg_string Message.Error "Taking all predicates!\n" ;
              Message.msg_string Message.Error ("Vampyre failed on query:\n "^(Predicate.toString (Predicate.Or (orlist))));
	      List.fold_left (function plist -> function p -> plist @ (Predicate.getAtoms p)) [] orlist
            end
	| _ -> newp

(* tweaks a given pred to drop primes and so on ... *)

let _newTemporariesAdded = ref []

let _substOriginalVars pred =
  (* substitute original variables for temporaries *)
  let _massage symbol = (* get back originl varname from temporary name *)
    let drop_initial_number =
      if (Misc.is_digit (String.get symbol 0)) then
        let _ind = String.index symbol '_' in
        let len = String.length symbol in
          (String.sub symbol (_ind+1) (len - _ind - 1))
      else symbol
    in
      if (String.contains drop_initial_number '%') then
	  begin
	    let _ind = String.index symbol '%' in
	      String.sub drop_initial_number 0 _ind
	  end
      else
	drop_initial_number
  in
  let rec massage_lval l =
    match l with
      | Expression.Symbol s -> Expression.Symbol (_massage s)
      | Expression.Access(op, e1, s) -> Expression.Access(op, massage e1, s)
      |	Expression.Dereference e -> Expression.push_deref (Expression.Dereference (massage e))
      |	Expression.Indexed (e1, e2) -> Expression.Indexed (massage e1, massage e2)
  and massage e =
    match e with
	Expression.Lval l -> Expression.Lval (massage_lval l)
      |	Expression.Chlval (l,_) -> Expression.Lval (massage_lval l)
      |	Expression.Binary(op, e1, e2) -> Expression.Binary(op, massage e1, massage e2)
      | Expression.Unary(Expression.Address, e) -> Expression.addressOf (massage e)
      | Expression.Unary(op, e) -> Expression.Unary(op, massage e)
      | Expression.Constant _   -> e
      | Expression.FunctionCall (Expression.Lval fnname, arglist) -> 
          begin
            (* worry about sel4 used to model pointer dereferences.
               We should rename sel4 to select, no? sel4 is a remnant of
               Necula's sel4 -- but there we were keeping track of bytes for memory safety
            *)
            let _getFnName fnname =
              match fnname with
                  Expression.Symbol s -> s
                | _ -> failwith ("focus: unexpected function call "^(Expression.lvalToString fnname))
            in
            let _fnNm = _getFnName fnname in
              if(_fnNm = "sel4" (* case: Arrow *)) then
                begin
                  if (List.length arglist <> 2) then
                    failwith "focus: unexpected use of sel4! More than two arguments."
                  else
                    begin
                      let _var = massage (List.hd arglist) and _deref = List.nth arglist 1 in
                        match _deref with
                            Expression.Constant i -> Expression.Lval (Expression.Dereference _var)
                          | Expression.Lval (Expression.Symbol s) -> 
				Expression.Lval (Expression.Access (Expression.Arrow, _var, _massage s))
                          | _ -> failwith "focus: unexpected use of sel4! Strange arguments."
                    end 
                end
              else
                if(_fnNm = "select" (* case Dot *)) then
                  begin
                    if (List.length arglist <> 2) then
                      failwith ("focus: unexpected use of select! More than two arguments. "^(Expression.toString e))
                    else
                      begin
                        (*
                          let _var = _massage (_getFnName (List.hd arglist)) and _deref = List.nth arglist 1 in
                        *)
                        let _var = (massage (List.hd arglist)) and _deref = List.nth arglist 1 in
                          match _deref with
                            | Expression.Lval (Expression.Symbol s) -> 
				Expression.Lval (Expression.Access (Expression.Dot, _var, _massage s))
			    | Expression.Constant (Constant.Int _) ->
				Expression.Lval (Expression.Dereference _var)
                            | _ -> failwith ("focus: unexpected use of select! Strange arguments. (deref)"^(Expression.toString e))
                      end
                  end
                else
                  if(_fnNm = "addrOf") then
                    begin
                      if (List.length arglist <> 1) then
                        failwith "focus: unexpected use of addrOf! More than one argument."
                      else
                        begin
                          let _var = massage (List.hd arglist) in
                            Expression.Unary(Expression.Address, _var)
                        end 
                    end
                  else
                    (* Other function calls can only be due to unary/binary operators 
                       that were converted to uninterpreted functions in the theoremprover
                    *)
                    begin
                      match _fnNm with
                          "BitAnd" -> Expression.Binary(Expression.BitAnd, 
							massage (List.hd arglist), 
							massage (List.nth arglist 1))
                        | "BitOr" -> Expression.Binary(Expression.BitOr, 
						       massage (List.hd arglist), 
						       massage (List.nth arglist 1))
                        | "Xor" -> Expression.Binary(Expression.Xor, 
                                                     massage (List.hd arglist), 
                                                     massage (List.nth arglist 1))
                        | "Div" -> Expression.Binary(Expression.Div, 
                                                     massage (List.hd arglist), 
                                                     massage (List.nth arglist 1))
                        | "LShift" -> Expression.Binary(Expression.LShift, 
							massage (List.hd arglist), 
							massage (List.nth arglist 1))
                        | "RShift" -> Expression.Binary(Expression.RShift, 
							massage (List.hd arglist), 
							massage (List.nth arglist 1))
                        | "Rem" -> Expression.Binary(Expression.Rem, 
                                                     massage (List.hd arglist), 
                                                     massage (List.nth arglist 1))
                        | "BitNot" -> Expression.Unary(Expression.BitNot, 
						       massage (List.hd arglist))
			| "chlval" -> massage (List.hd arglist)
			| "foffset" -> Expression.Binary(Expression.FieldOffset,
							 massage (List.hd arglist),
							 massage (List.nth arglist 1))

                        | _ -> failwith ("focus: unexpected function name "^(_getFnName fnname))
                    end
          end
      | _ -> failwith "focus: Expression type not handled.."
  in
    match pred with
        Predicate.Atom (Expression.Binary(Expression.Eq, Expression.Lval (Expression.Symbol s), e)) ->
          begin
            (*Message.msg_string Message.Error ("IN focus pred = "^(Predicate.toString pred));*)
            if (isTemporary s) then
              begin
                let orig_s = _massage s in
                let rec massage_check_lval l orig_s =
		  match l with
		    | Expression.Access(op, e1, s) -> Expression.Access(op, massage_check e1 orig_s, s)
                    | Expression.Symbol str -> 
                        begin
                          let ss = (_massage str)
                          in
                            if (ss = orig_s) then (* need annotations *)
                              begin
                                try
                                  let tempVarName = "$tmp_" in
                                    _newTemporariesAdded := 
                                    if List.mem (tempVarName^ss, ss) (!_newTemporariesAdded) then
                                      (!_newTemporariesAdded)
                                    else ("$tmp_"^ss, ss) :: (!_newTemporariesAdded);
                                    Expression.Symbol ss
                                with Not_found -> 
                                  Expression.Symbol ss
                                  
                              end
                            else (Expression.Symbol ss)
                        end
		    | Expression.Dereference e -> Expression.push_deref (Expression.Dereference (massage_check e orig_s))
		    | Expression.Indexed (e1, e2) -> Expression.Indexed (massage_check e1 orig_s, 
									 massage_check e2 orig_s)
		and massage_check e orig_s =
                  match e with
		    | Expression.Lval l -> Expression.Lval (massage_check_lval l orig_s)
		    | Expression.Chlval _ -> failwith "massage_check: Chlval not handled"
                    |   Expression.Binary(op, e1, e2) -> Expression.Binary(op, massage_check e1 orig_s, massage_check e2 orig_s)
                    | Expression.Unary(Expression.Address, e) -> Expression.addressOf (massage_check e orig_s)
                    | Expression.Unary(op, e) -> Expression.Unary(op, massage_check e orig_s)
                    | Expression.Constant _   -> e
                    | Expression.FunctionCall (Expression.Lval fnname, arglist) -> 
			begin
                          (* worry about sel4 used to model pointer dereferences.
			     We should rename sel4 to select, no? sel4 is a remnant of
			     Necula's sel4 -- but there we were keeping track of bytes for memory safety
			  *)
			  let _getFnName fnname =
			    match fnname with
				Expression.Symbol s -> s
			      | _ -> failwith ("focus: unexpected function call here: "^(Expression.lvalToString fnname))
			  in
			  let _fnNm = _getFnName fnname in
			    if(_fnNm = "sel4") then
			      begin
				if (List.length arglist <> 2) then
				  failwith "focus: unexpected use of sel4! More than two arguments."
				else
				  begin
				    let _var = (massage_check (List.hd arglist) orig_s) and _deref = List.nth arglist 1 in
				      match _deref with
					  Expression.Constant i -> Expression.Lval (Expression.Dereference  _var)
					| Expression.Lval Expression.Symbol s -> 
					    Expression.Lval (Expression.Access (Expression.Arrow, _var, _massage s))
					| _ -> failwith "focus: unexpected use of sel4! Strange arguments."
				  end 
			      end
			    else
			      if(_fnNm = "select" (* case Dot *)) then
				begin
				  if (List.length arglist <> 2) then
				    failwith ("focus: unexpected use of select! More than two arguments."^(Expression.toString e))
				  else
				    begin
				      let _var = (massage_check (List.hd arglist) orig_s) and _deref = List.nth arglist 1 in
					match _deref with
					  | Expression.Lval (Expression.Symbol s) -> Expression.Lval (Expression.Access (Expression.Dot, _var, _massage s))
					  | Expression.Constant (Constant.Int _) -> Expression.Lval (Expression.Dereference _var)
					  | _ -> failwith ("focus: unexpected use of select! Strange arguments."^(Expression.toString e))
				    end
				end
			      else
				
				if(_fnNm = "addrOf") then
				  begin
				    if (List.length arglist <> 1) then
				      failwith "focus: unexpected use of addrOf! More than one argument."
				    else
				      begin
					let _var =  (massage_check (List.hd arglist) orig_s) in
					  Expression.addressOf _var
				      end 
				  end
				else  
				  (* Other function calls can only be due to unary/binary operators 
				     that were converted to uninterpreted functions in the theoremprover
				  *)
				  begin
				    match _fnNm with
					"BitAnd" -> Expression.Binary(Expression.BitAnd, 
								      massage (List.hd arglist), 
								      massage (List.nth arglist 1))
				      | "BitOr" -> Expression.Binary(Expression.BitOr, 
								     massage (List.hd arglist), 
								     massage (List.nth arglist 1))
				      | "Xor" -> Expression.Binary(Expression.Xor, 
								   massage (List.hd arglist), 
								   massage (List.nth arglist 1))
				      | "Div" -> Expression.Binary(Expression.Div, 
								   massage (List.hd arglist), 
								   massage (List.nth arglist 1))
				      | "LShift" -> Expression.Binary(Expression.LShift, 
								      massage (List.hd arglist), 
								      massage (List.nth arglist 1))
				      | "RShift" -> Expression.Binary(Expression.RShift, 
								      massage (List.hd arglist), 
								      massage (List.nth arglist 1))
				      | "Rem" -> Expression.Binary(Expression.Rem, 
								   massage (List.hd arglist), 
								   massage (List.nth arglist 1))
				      | "BitNot" -> Expression.Unary(Expression.BitNot, 
								     massage (List.hd arglist))
				      | "chlval" -> massage (List.hd arglist)
				      | "foffset" -> Expression.Binary(Expression.FieldOffset, 
								   massage (List.hd arglist), 
								   massage (List.nth arglist 1))
				      | _ -> failwith ("focus: unexpected function name "^(_getFnName fnname))
				  end
			end
			
                    | _ -> failwith "do_focus: expression type not handled: TBD"
                in
                  Predicate.Atom (Expression.Binary(Expression.Eq, Expression.Lval (Expression.Symbol orig_s), massage_check e orig_s))
              end
            else
              Predicate.Atom (Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol s), massage e))
          end
      | Predicate.Atom e -> Predicate.Atom (massage e)
      | _  -> failwith "do_focus: obtained non-atomic predicate: unexpected"
	  

(* this is the real function that does the brunt of the hard work ....
jhala: i would like to use this elsewhere hence am sucking it out of do focus *)

let callsToFocus = ref 0

let clean_new_predicates newpreds = 
  let _ = _newTemporariesAdded := [] in
  let _substnewpreds = 
    try List.map _substOriginalVars newpreds 
    with e -> 
      begin
	Message.msg_string Message.Debug ("_substOrigVars raises :"^(Printexc.to_string e));
	Message.msg_string Message.Error ("_substOrigVars raises :"^(Printexc.to_string e));
	[]
      end
  in
  let _tempPreds = 
    (* Also need to add predicates of the form "%temp_x = x" and "P(%temp_x/x)" for each predicate
       with x but not %temp_x 
       TBD !!!! *)
    let predsFromFor = List.fold_left (fun a -> fun (t,b) -> List.append (predsFromForLoops b ) a) 
                         [] (!_newTemporariesAdded) in
      predsFromFor  
  in
  let allPreds = _substnewpreds (*@ _tempPreds*) in
  allPreds

let extractUsefulPredicates emptyPred = 
  let _ = incr callsToFocus in
    (* let newpreds = Stats.time "getUsefulPreds" TheoremProver.getUsefulPredicates emptyPred *)
    (* while the above should have been suficient, Vampyre seems to have trouble with Or's.
       So we convert to DNF, and check if each conjunct is satisfiable, collecting useful predicates. *)
  let emptyPredDNF_form = Predicate.convertDNF emptyPred in
    (* Printf.printf "Converting to DNF and printing...\n%s\n" (Predicate.toString emptyPredDNF_form);*)
  let orlist = 
     match emptyPredDNF_form with
	 Predicate.Or plist -> plist
       | a -> [a]
  in
  let newpreds = getNewPreds orlist in
  let l = clean_new_predicates newpreds in
    l
      
(* this is a highly stateful function! *)
let last_no_of_preds = ref 0
let predicate_fixpoint_reached () = 
  if (!last_no_of_preds = List.length !tableOfPreds) then 
    begin
      printTableOfPreds ();
      true
    end
  else
    begin
      last_no_of_preds := List.length !tableOfPreds;
      false
    end

  
   let do_focus abst_data_reg empty_conc_data_reg =
     Message.msg_string Message.Debug "In do_focus -- arguments are:" ;
     Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;
     Message.msg_printer Message.Debug Region.print_concrete_data_region empty_conc_data_reg ;
     let emptyPred = empty_conc_data_reg.Region.pred in
     let allPreds = extractUsefulPredicates emptyPred in
     (*let _= 
       begin
	 Message.msg_string Message.Error "Candidate preds";
	 List.iter (fun p -> Message.msg_string Message.Error (Predicate.toString p) ) allPreds
       end
     in*)
     (* if this round did not add any new predicates, then we should restart *)
 let _ = Misc.write_to_file ".newpred" " " in     
let oldnum = getCurrentIndex () in
     let added_preds = List.map addPred allPreds in
     let newnum = getCurrentIndex () in
     if (oldnum = newnum) then
	begin
        Message.msg_string Message.Debug "Restarting..." ;
	Options.setValueOfBool "localrestart" true ;
        end
     else
        begin
        Message.msg_string Message.Debug "At least one predicate added ..." ;
	Options.setValueOfBool "localrestart" false ;
        end ;
      
     let newpredlist = Misc.compact (Misc.difference (added_preds) (-1 :: abst_data_reg.Region.preds)) 
                         (* This is not correct ... as there may be a descendant along the error path 
                            that does not have a certain pred that is in this list which the 
                            ancestor ie the current data region has... if you are doing PATH that is ...*)
        in
                if (newpredlist = []) then (Message.msg_string Message.Error "No new predicates ! Unexpected ...";
                                               printTableOfPreds () ;
					       if (Options.getValueOfInt "predH" < 6) then
        					 raise NoNewPredicatesException;
							       );
                  begin  
		    let _ =  Message.msg_string Message.Debug ("NEW PREDICATES FOUND \n") in
                   let _listOfAllActivePreds = abst_data_reg.Region.preds @ newpredlist in
                   ignore (if (List.length _listOfAllActivePreds) > !statsMaxNumberOfPredsActive then 
                     begin
                      statsMaxNumberOfPredsActive := List.length _listOfAllActivePreds;
		       Message.msg_string Message.Debug ("Size of pred set: "^(string_of_int (List.length !tableOfPreds)) );
                      Message.msg_string Message.Debug "Active Predicates"; 
                      List.iter (function i -> Message.msg_string Message.Debug  (Predicate.toString (getPred i)))  
                                _listOfAllActivePreds
                      end);
                { Region.bdd = abst_data_reg.Region.bdd ;
                  Region.preds = _listOfAllActivePreds ; }
                 end


let replace_data_region reg new_data_reg =
       match reg with 
       Region.Atomic _reg ->
        Region.Atomic 
	  { Region.location = _reg.Region.location;
            Region.data = new_data_reg;
            Region.stack = _reg.Region.stack;
          }
      | _ -> failwith ("bad call to replace_data_region")


(* ******************* ERROR TRACE FILTER ****************************** *)

(* the following procedure takes a trace and spits out that subtrace of it which is used to 
   get a contradiction -- it is to help in "manual" counterexample analysis *)

let  trace_filter region_list op_list = 
  (* the above is the unsatisfiable list of regions and operations -- arranged in "past --> present" order*)
  let reg_list_r = List.rev region_list in
  let op_list_r = List.rev op_list in
  let process_op_reg (phi , map) (op, reg) = () in
    ()
    
(* ******************* find_all_preds ********************************** *)


     (* this function will look at ALL subtraces of a given trace and find all predicates coming from it *)



(* again oplist is in increasing order past --> present *)
(* r1 -> o1 -> r2 -> o2 -> ... -> on -> rn *)

let extract_pred reg = 
  match reg with
      Region.Atomic r -> 
	begin
	  match r.Region.data with 
	      Region.Concrete x -> x.Region.pred 
	    | Region.Abstract a -> convertBddToPred a.Region.bdd
		(*
		(Message.msg_string Message.Error "Bad call to extract_pred"; failwith ("Bad call to extract pred"))*)
	end
    | _ -> (Message.msg_string Message.Error "Bad call to extract_pred"; failwith ("Bad call to extract pred"))

let extract_abs_data_reg reg = 
  match reg with
      Region.Atomic atomic_reg ->
	begin
	  match atomic_reg.Region.data with
	      Region.Abstract ar -> ar
	    | _ -> failwith "conc_dr given to extract_abs_data_reg"
	end
    | _ -> failwith "union region given to extract_abs_data_reg"




let insert_data_reg reg dr = 
  match reg with
      Region.Atomic atomic_reg ->
	begin
	  Region.Atomic 
	    { Region.data = dr;
	      Region.location = atomic_reg.Region.location;
	      Region.stack = atomic_reg.Region.stack;
	    }
	end
    | _ -> failwith "bad call to insert_data_reg"


(* this is called only in "block" mode ... and with noforget *)
let focus_stamp reg =
  let abst_data_reg = extract_abs_data_reg reg in
  let cur = getCurrentIndex () in
  let present_prec = 
    try List.hd abst_data_reg.Region.preds 
    with _ -> -1 
  in
  let new_reg = 
    insert_data_reg reg 
      (Region.Abstract 
	 { Region.bdd = abst_data_reg.Region.bdd;
	   Region.preds = [cur];
	 })
  in
    (*if (Options.getValueOfInt "craig" = 2) then
      begin (* localrestart will be true only if all the locations already know about all the predicates *)
      
      if (present_prec = cur && Options.getValueOfBool "localrestart")
      then
      begin
      dump_abs ();
      raise NoNewPredicatesException;
      end
      else
      begin
      Options.setValueOfBool "localrestart" false;
      new_reg
      end
      end
      else *)
      if (present_prec = cur && not ((Options.getValueOfInt "craig" = 2))) then
	begin
	  dump_abs ();
	  raise NoNewPredicatesException;
	end
      else
	new_reg




(* Cheap hack: go through all the predicates along oplist and add them to the list of predicates. *)

let find_all_predicates_cheap oplist =
   let rec _more lst accum =
     match lst with
       [] -> accum
     | a :: rest ->
         begin
           match (C_SD.get_command a).C_SD.Command.code with
             C_SD.Command.Pred e ->
                if e = Predicate.True then _more rest accum
                else 
                  begin
                    let ee = match e with Predicate.Not e' -> e' | Predicate.Atom _ -> e | _ -> failwith "nonatomic predicate on path!" in
                    _more rest (ee :: accum)
                  end
           | _ -> _more rest accum
         end
   in
   let all_preds_in_path = _more oplist [] in
   let newpredlist = Misc.compact 
                       (Misc.difference 
                          (List.map addPred all_preds_in_path) 
                          (-1 :: ( List.map Misc.fst3 !tableOfPreds))) in 
   let number = List.length newpredlist in
   Message.msg_string Message.Debug ("find_all_predicates_cheap found "^
                                      (string_of_int number)^" predicates.") ;
   number


let find_all_predicates reglist oplist = 
  let list_of_predicates = ref [] in
  let oplist_r = List.rev oplist in
  let reglist_r = List.rev reglist in

  let rec _proc_sub seed ol = 
    (* take wp of seed wrt (ol.head,rl.head) -- query for unsat if unsat take preds... quit and break out ..
       else if sat, then take region and recur on it with ol.tail, rl.tail *)
    match ol with
	op::t -> 
	  begin 
	    let emptytable = Hashtbl.create 2 in
	    let pre_reg = pre seed op emptytable in
	      if (Region.is_empty pre_reg) then 
		begin
		  let allpreds = extractUsefulPredicates (extract_pred pre_reg) in
		    Message.msg_string Message.Debug ("Learning new predicates: ");
		    List.iter (fun x -> list_of_predicates := x::(!list_of_predicates); 
			      Message.msg_string Message.Debug ((Predicate.toString x)^" , ")) allpreds;
		    Message.msg_string Message.Debug ("From region: "^(Predicate.toString (extract_pred pre_reg)))
		      (* do we or do we not recur on pre_reg ? as of now, NO *)
		end
	      else 
		_proc_sub pre_reg t
	  end
      | [] -> ()
  in

  let rec process_subtrace ol_r rl_r = 
    match rl_r with
     h::h1::t ->
       let seed_reg = (replace_data_region h (Region.Concrete {Region.pred = Predicate.True})) in
       _proc_sub seed_reg ol_r; 
	    process_subtrace (List.tl ol_r) (h1::t)
      | _ -> ()
  in
(*
  let current_preds = !tableOfPreds in
  let current_pred_vars = 
    List.map (fun x -> Expression.Symbol x)
      (Misc.compact (List.flatten (List.map (fun x -> getVarExps_deep (Misc.snd3 x)) current_preds)))
  in
  let make_predicate v c = Predicate.Atom (Expression.Binary (Expression.Eq,v,c)) in
  let process_var v = 
    try 
      List.map (make_predicate v) (Hashtbl.find lval_constant_table v)
    with
	Not_found -> failwith ("lval not in lval_constant_table: "^(Expression.toString v))
  in
  let new_const_preds = List.flatten (List.map process_var current_pred_vars) in
  let _ = List.map addPred new_const_preds in
  let _ = 
    if (List.length !tableOfPreds = List.length current_preds) then
      failwith (" No new constant preds learnt! ")
    else 
      begin
	Message.msg_string Message.Normal "new_const_predicates found: \n";
	List.iter (fun x -> Message.msg_string Message.Normal (Predicate.toString x)) new_const_preds;
	Message.msg_string Message.Normal ("# new preds = "^(string_of_int (List.length new_const_preds )));
      end
*)  
    (* this is the point where the main function is called ! *)
(*  The stuff in this needs to be worked on ... its *very* slow ... 
  let _ = process_subtrace oplist_r reglist_r in
    let newpredlist = Misc.compact (Misc.difference (List.map addPred !list_of_predicates) (-1 :: ( List.map Misc.fst3 !tableOfPreds))) in
    Message.msg_string Message.Normal "find_all_predicates found: \n";
    List.iter (fun x -> Message.msg_string Message.Normal (Predicate.toString x)) !list_of_predicates;
    Message.msg_string Message.Normal ("# new preds = "^(string_of_int (List.length newpredlist)));
    if (newpredlist = [] ) then failwith ("find_all_predicates also fails to get new predicates !!") *)

   if find_all_predicates_cheap oplist == 0 then
    begin
     let _ = process_subtrace oplist_r reglist_r in
     let newpredlist = Misc.compact (Misc.difference (List.map addPred !list_of_predicates) (-1 :: ( List.map Misc.fst3 !tableOfPreds))) in
     Message.msg_string Message.Debug "find_all_predicates found: \n";
     List.iter (fun x -> Message.msg_string Message.Debug (Predicate.toString x)) !list_of_predicates;
     Message.msg_string Message.Debug ("# new preds = "^(string_of_int (List.length newpredlist)));
     if (newpredlist = [] ) then failwith ("find_all_predicates also fails to get new predicates !!")
    end


     (**********************************************************************************)
     (******* the procedure below generates a boolean cfa ******************************)
     (**********************************************************************************)


(* This is the main function that generates a constraint formula for an edge and a bunch of predicates *)
(* The boolean variable for a particular predicate p = __BS__i where i is the predIndex of the pred *)
(* The value of the variable at the next time is __BS__i'*)

  let state_vars_table = Hashtbl.create 1009
  let comb_vars_table = Hashtbl.create 1009
  let bs_table = Hashtbl.create 1009
  let dumpBoolFlag = ref true 

  let predicate_update_constraints (cmd,edge) pred_list = 
    let _ = Message.msg_string Message.Debug "In predicate_update_constraints" in
    let make_boolean_state_var pred = 
      let _ = Message.msg_string Message.Debug "In make_boolean_state_var" in
      let pred_index = try getPredIndex pred with Not_found -> failwith ("Unknown pred? "^(Predicate.toString pred)) in
      let _ = Message.msg_string Message.Debug "back from getPredIndex" in
      let pre_var = "b_p__BS__"^(string_of_int pred_index) in
      let post_var = "b_p__BS__"^(string_of_int pred_index)^"_n" in
      let _ = Hashtbl.replace state_vars_table pre_var true in
      (Predicate.Atom (Expression.Lval (Expression.Symbol pre_var)), 
       Predicate.Atom (Expression.Lval (Expression.Symbol post_var)))
    in
    let _assume_predicate_list = 
      match cmd.C_SD.Command.code with 
          C_SD.Command.Pred e -> [e] 
	| _ -> []
    in
    let constraint_for_pred pred =
      let _ = Message.msg_string Message.Debug "In constraint_for_pred" in
      let (pre_var,post_var) = make_boolean_state_var pred in
      (* let post_pred = rename_with_suffix pred "_n" in *)
      let (* hack *) emptytable = Hashtbl.create 2 in
      let wp_pred = (concrete_data_pre false false 
		       { Region.pred = pred ; } 
		       cmd 
		       edge emptytable).Region.pred in
	[(Predicate.iff pre_var pred);(Predicate.iff wp_pred post_var)] 
	(* I deliberately made it wp <--> x_p' *)
    in
    let _ = Message.msg_string Message.Debug "Leaving predicate_update_constraints" in
      (Predicate.conjoinL 
((List.flatten (List.map constraint_for_pred pred_list)) @ _assume_predicate_list))

(* this takes a predicate in a rich logic and then employs uclid etc to get a predicate (Predicate.predicate)
which is totally boolean ... it also adds to the comb_var_table ...*)
      	
  let pred_to_bool_counter = ref 0

let predicate_to_boolean pred = 
   let _ = Message.msg_string Message.Debug "In predicate_to_boolean" in
   let _ = (pred_to_bool_counter := !pred_to_bool_counter + 1) in
   let int_to_atom _s = 
     let _ = print_string "." in
     let neg = Misc.is_prefix "-" _s in
     let s = List.hd (Misc.chop _s "-") in
     let sym = 
       if Hashtbl.mem bs_table s 
       then Hashtbl.find bs_table s 
       else 
	 begin
	   Hashtbl.replace  comb_vars_table ("b_p_x_"^s) true;
	   ("b_p_x_"^s) 
	 end
     in 
       match neg with
	   true -> Predicate.negate (Predicate.Atom (Expression.Lval (Expression.Symbol sym)))
	 | false -> (Predicate.Atom (Expression.Lval (Expression.Symbol sym)))
   in
    
   let _ = Hashtbl.clear bs_table in
   let modelname = "tmp_"^(string_of_int !pred_to_bool_counter) in
   let _ = TheoremProver.uclid_output modelname pred "" in
   let _ = Sys.command ("rm "^modelname^".cnf") in   
   let _ = Message.msg_string Message.Debug "Now calling Uclid" in
   let uclidret = Sys.command ("runuclid "^modelname^".ucl cnf") in
   let _ = Message.msg_string Message.Debug "Back from Uclid" in

     (* now we have to read back in tmp.cnf and make sense of it ...*)
   let _ = Message.msg_string Message.Debug "Now parsing cnf file" in

   let string_clause_of_string x = List.tl (List.rev (Misc.words_of_string x)) in
   let list_of_clauses = 
     try 
       List.tl (List.map string_clause_of_string (Misc.string_list_of_file (modelname^".cnf"))) 
     with _ -> [] (* this exception happens if uclid doesnt generate blasttrace.cnf which is when either the 
		     thing is always true ? or its always false ? let us simply say [] now *)
   in
   let _ = Message.msg_string Message.Debug "Done parsing cnf file" in
   (* drop the first line which is a bunch of whooey anyways ...*)
     (* now we must make up a map so as to know which are the fun variables (ie. of interest ...*)
   let _ = Sys.command ("fgrep \"b_p\" "^modelname^".map > BS.map") in
   let _ = Message.msg_string Message.Debug "Reading in BS map" in
   let var_map = List.map 
		   (fun l -> Hashtbl.replace bs_table (List.hd (List.tl 
l)) (List.hd l))
		   (List.map Misc.words_of_string (Misc.string_list_of_file "BS.map"))
   in
   let _ = Message.msg_string Message.Debug "Done reading BS map" in
     (* now all we need is judicious use of map ! *)
   let retval = Predicate.And (List.map (fun l -> Predicate.Or (List.map int_to_atom l)) list_of_clauses)
   in
     Message.msg_string Message.Debug "Returning from predicate_to_boolean";
     retval

  let generate_boolean_cfa _predlist start_loc err_loc outfile =
    let module LocationSet =
      Set.Make (struct
                  type t = int 
                  let compare x y =  compare x  y
                end) in
    let cfa_chan = open_out ".trans.cfa" in
    let bcfa_chan = open_out ".trans.bcfa" in
    let predlist = List.map Predicate.canonicize _predlist in
    let sprint_edge e = 
      let (l1,c,l2) = (C_SD.get_source e, C_SD.get_command e, C_SD.get_target e) in 
      let (_,y1) = C_SD.location_coords l1 in
      let (_,y2) = C_SD.location_coords l2 in
      let blast_pred_e = (predicate_update_constraints (c,e) predlist) in
      let _ = Message.msg_string Message.Debug "Back from pred_upd_const" in
      let _ = flush stdout in
	match (!dumpBoolFlag) with
	    true -> 
	      begin
		output_string bcfa_chan 
		  (Printf.sprintf "[ %d ]  ->  [ %d ] : %s ; \n \n" y1 y2 
		     (Predicate.toString (predicate_to_boolean blast_pred_e)));
		Message.msg_string Message.Debug "Dumped edge to file"
	      end
	  | false -> 
	      begin
		output_string cfa_chan
		  (Printf.sprintf "[ %d ]  ->  [ %d ] : %s ; \n \n" y1 y2 
		     ( Predicate.toString blast_pred_e ));
		print_string "dumped edge to file"
	      end
    in

    let sprint_function fname =
      let already_visited_source_locations = ref LocationSet.empty in
      let rec sprint_location_and_edges loc =
	let (_,loc_id) = C_SD.location_coords loc in
        if (LocationSet.mem loc_id !already_visited_source_locations) then
          ()
        else
          begin
            let outgoing_edges = List.rev (C_SD.get_outgoing_edges loc) in
	    let _ = List.iter sprint_edge outgoing_edges in
	      already_visited_source_locations := (LocationSet.add (snd (C_SD.location_coords loc)) !already_visited_source_locations) ;
		(List.iter (fun e -> sprint_location_and_edges (C_SD.get_target e)) outgoing_edges)          end 
      in
        sprint_location_and_edges (C_SD.lookup_entry_location fname)
    in
    let _ = List.map addPred predlist in
    let _ = dumpBoolFlag := true in
      (* this thing below will have to change when we add support for function calls *)
    let (* trans_string *) _ = 
       (* Misc.string_list_cat "\n" *)
	 (List.iter sprint_function (C_SD.list_of_functions ()))
    in
    let _ = dumpBoolFlag := false in
    let _ = 
      (* Misc.string_list_cat "\n" *)
	(List.iter sprint_function (C_SD.list_of_functions ()))
    in
      
    let (_,_) = (close_out cfa_chan, close_out bcfa_chan) in
    let reach_string = Printf.sprintf "[ %d ] -> [ %d ];" (snd (C_SD.location_coords start_loc)) (snd (C_SD.location_coords err_loc)) 
    in
    let state_string = (Misc.string_list_cat "," (Misc.run_on_table (fun x _ -> x) state_vars_table))^";" in
    let comb_string = (Misc.string_list_cat "," (Misc.run_on_table (fun x _ -> x) comb_vars_table))^";" in

    let boolean_cfa_string = 
      Printf.sprintf "STATEVAR \n %s \n COMBVAR \n %s \n TRANS \n"
	state_string comb_string
    in
    let fo_cfa_string = 
      Printf.sprintf "STATEVAR \n %s \n COMBVAR \n %s \n TRANS \n"
	state_string "" (* comb_string fo_trans_string reach_string *)
    in
    
    let _ = Misc.write_to_file ".temp1" (Printf.sprintf "REACH \n %s \n" reach_string) in
      Misc.write_to_file (".temp") boolean_cfa_string;
      ignore (Sys.command ("cat .temp .trans.bcfa .temp1 > "^outfile^".bcfa"));
      ignore (Sys.command ("cat .temp .trans.cfa .temp1 > "^outfile^".cfa"));
      ignore (Sys.command ("rm .temp")); 
      ignore (Sys.command ("rm .temp1"));
      ignore (Sys.command "rm .trans.bcfa"); 
      ignore (Sys.command "rm .trans.cfa") ;
      ()
(*
 
  Misc.write_to_file (outfile^".bcfa") boolean_cfa_string;
  Misc.write_to_file (outfile^".cfa") fo_cfa_string 
*)
      
(************************* END BOOLEAN CFA DUMP routines ****************************)
(************************* START the uclidsat cref stuff ****************************)

  (* this next function will produce the huge formula that we will send off to UCLID *)
  (* region0 -> op1 -> region 1 -> ...-> op n -> region n *)
  (* as of now this just dumps the trace to a file and then *)

  let rename_with_suffix pred suffix_string = 
    let new_name l =
      match l with 
	Expression.Symbol s -> Expression.Symbol (s^suffix_string)
      |	_ -> failwith "rename_with_suffix"
    in
    Predicate.alpha_convert new_name pred 
      (*
	let vars_in_pred = 
	let _var_string_list = ref [] in
	List.iter 
        (fun s -> if (not (List.mem s (!listOfConstants))) 
        then _var_string_list := s::(!_var_string_list)
           else  ()
        )
        ( getVarExps_deep pred ); !_var_string_list 
	in
	let subs_pairs = 
	List.map (fun x -> (Expression.Symbol (x), Expression.Symbol(x^suffix_string))) vars_in_pred
	in
	Predicate.substitute  subs_pairs pred
      *)
  

  let trace_map_table = Hashtbl.create 1009 
  let comp_tm_table = Hashtbl.create 1009
 

  let got_new_split = ref false 
  let location_coords_of_region reg = 
    match reg with
	Region.Atomic _reg ->  (_reg.Region.location)
      | _ -> failwith "bad call to location_coords_of_region"


  let sign x = if x > 0 then 1 else -1

  let var_of_string x = 
    let frags = 
      List.filter (fun y -> y <> "") (Misc.chop x "_") 
    in
      (int_of_string (List.nth frags 3), int_of_string (List.nth frags 4),(List.length frags = 6))
  

  let process_block_constraints cons_list =
    let _ = Sys.command "rm t-*.cnf" in
    let clauses = ref 0 in
    let pred_array = Array.create_matrix (!predIndex + 1) (List.length cons_list + 1) 0 in
    let pred_prime_array = Array.create_matrix (!predIndex + 1) (List.length cons_list + 1) 0 in
    let _ = Hashtbl.clear comp_tm_table in

    let rec _process_block_constraints i offset cons_list =
      let replace_var s = 
	let var = int_of_string s in
	  (* the branch prediction is getting messed up here *)
	  if (var = 0) then "0" else string_of_int (var + ((sign var)*offset))
      in
      let convert_cnf_file () = 
	let h::t = List.map Misc.words_of_string (Misc.string_list_of_file "tmp.cnf") in
	let new_lines = (List.map (fun l -> ((String.concat "    " (List.map replace_var l))^"\n")  ) t) in
	  Misc.write_list_to_file ("t-"^(string_of_int i)^".cnf") new_lines;
	  clauses := !clauses + (int_of_string (List.nth h 3));
	  int_of_string (List.nth h 2) (* this returns the number of vars in this CNF file *)
      in
      let process_map () =
	let map_items = List.map Misc.words_of_string (Misc.string_list_of_file ("BS-"^(string_of_int i)^".map")) in
	let proc_item lst =
	  match lst with
	   (x::y::_) ->
	     begin
	       let (pred_id,time,prime) = var_of_string x in 
	         match prime with
		  true -> pred_prime_array.(pred_id).(time) <- (int_of_string y + offset)
	         | false -> 
		   begin
		     Hashtbl.add comp_tm_table (int_of_string y + offset) (pred_id,time);
		     pred_array.(pred_id).(time) <- (int_of_string y + offset)
		   end
             end
          | _ -> failwith "proc_iterm"
	in
	  List.iter proc_item map_items
      in
	match cons_list with
	    [] -> offset
	  | hd::tl -> 
	      begin
		TheoremProver.uclid_output "tmp" hd "";
		Sys.command "runuclid tmp.ucl";
		Sys.command ("fgrep \"b_p\" tmp.map > BS-"^(string_of_int i)^".map");
		process_map ();
		let num_new_vars = convert_cnf_file () in
		  _process_block_constraints (i+1) (offset + num_new_vars) tl
		    (* recall that we must end up recurring with (i+1) (offset + vars) tl *)	
	      end
    in
    let variables = _process_block_constraints 1 0 cons_list in
      (* now we have to cat the files together and add to it the tying together constraints *)
    let connect_constraint = 
      let s = ref "" in
	for j = 0 to (!predIndex) do
	  for k = 2 to (List.length cons_list) do
	    let v1 = pred_array.(j).(k) in
	  let v2 = pred_prime_array.(j).(k-1) in
	    s := !s^(Printf.sprintf "%d    -%d    0 \n -%d   %d   0\n" v1 v2 v1 v2);
	  done;
	done;
	clauses := !clauses + 2*(1 + !predIndex)*(List.length cons_list - 1);
	!s
    in
      Misc.write_to_file "sat-info" (Printf.sprintf "p cnf %d %d \n %s" variables !clauses connect_constraint);
      ignore (
	Sys.command "cat sat-info t-*.cnf > trace.cnf"; 
	Sys.command "rm ztrace.*";
	Sys.command "smvsat trace.cnf satout";);
      try
	let vars_in_proof =  
	  List.map (fun x -> int_of_string (List.hd x)) 
	    (List.map Misc.words_of_string (Misc.string_list_of_file "ztrace.variables")) 
	in
	  List.filter (Hashtbl.mem comp_tm_table) vars_in_proof
      with _ -> failwith "Trace is satisfiable, apparently!"
      
	
	
   let analyze_trace region_list op_list = 
     let _ = Hashtbl.clear trace_map_table in
     let pipe_counter = ref 0 in
     let next_pipe_var pred i = 
       Printf.sprintf "b_p__BS__%d_%d" (getPredIndex pred) i 
     in
     let predicate_list = List.map Misc.snd3 !tableOfPreds in
     
     
	let constraint_to_boolean_with_renaming cons i =
	let boolean_cons = predicate_to_boolean cons in
	let rename_function l = 
	match l with
	Expression.Symbol x ->
	if (Misc.is_prefix "b_p_x_" x) then Expression.Symbol (x^("_t_"^(string_of_int i))) 
	else Expression.Symbol x
	| _ -> failwith "constraint_to_boolean_with_renaming"
       in
	 Predicate.alpha_convert rename_function boolean_cons
     in
 
     let process_op (reg,op,i) =
       let rename_with_suffix pred suffix_string = 
         let vars_in_pred = 
           let _var_string_list = ref [] in
             List.iter 
               (fun s -> if (not (List.mem s (!listOfConstants))) 
                then _var_string_list := s::(!_var_string_list)
                else  ()
               )
               ( getVarExps_deep pred ); !_var_string_list 
         in
         let subs_pairs = 
           List.map 
	     (fun x -> (Expression.Lval (Expression.Symbol (x)), Expression.Lval (Expression.Symbol(x^suffix_string)))) 
	     vars_in_pred
         in
           Predicate.substitute  subs_pairs pred
       in
       let _assume_predicate_list = 
         let cmd = (C_SD.get_command op).C_SD.Command.code in
           match cmd with 
               C_SD.Command.Pred e -> [(rename_with_suffix e ("_"^(string_of_int i)))] 
	     | _ -> []
       in

       let process_pred pred =
	 let emptytable = Hashtbl.create 2 in
	 Message.msg_string Message.Debug "Creating empty table in process_pred: will probably break.";
         let _phi = spre (replace_data_region reg (Region.Concrete {Region.pred = pred})) op emptytable in
         let phi_p = 
              if (_assume_predicate_list <> []) then pred 
              else 
              begin
                match _phi with 
                Region.Atomic foo -> 
                  begin
                    match foo.Region.data with 
                        Region.Concrete x -> x.Region.pred
                      | _ -> failwith("bad return from pre !")
                  end
               |  _ -> failwith ("bad region in analyze trace")
	      end
         in
	 let p_i_minus_1_prime = rename_with_suffix pred ("_"^(string_of_int (i-1))^("_p")) in
         let p_i = rename_with_suffix pred ("_"^(string_of_int i)) in
         let phi_p_i = rename_with_suffix phi_p ("_"^(string_of_int i)) in
         let p_i_prime = rename_with_suffix pred ("_"^(string_of_int (i))^("_p")) in
          (* a simple iff *)
         let pipe = next_pipe_var pred i in
	 let pipe_prime = pipe^"_p" in
         let _pipe_pred = Predicate.Atom (Expression.Lval (Expression.Symbol pipe)) in
	 let _pipe_pred_prime = Predicate.Atom (Expression.Lval (Expression.Symbol pipe_prime)) in
	 let _last_pipe_pred_prime = Predicate.Atom (Expression.Lval (Expression.Symbol (next_pipe_var pred (i-1)))) in 
	   
	 let left_cons = [(Predicate.iff _pipe_pred p_i)] 
         in
         let right_cons = Predicate.iff phi_p_i _pipe_pred_prime in
           right_cons::left_cons (* avoiding the use of conjoinL *)
       in

     (* Jhala: for now I am NOT adding an option that does the piecewise translation
	-- that shall simply be the default 
     *)       
       let op_constraints = List.flatten (List.map process_pred predicate_list) in
	 (Predicate.And (_assume_predicate_list @ op_constraints)) 
	 (*
	   let retval = 
	   in
	   Message.msg_string Message.Normal ("Done booleanifying block"^(string_of_int i));
	   retval
	 *)
   in
       (* the above is what you get for (reg,op,i)-now make a list of those and hit it with List.map*)
     let rec make_op_reg_i_triple rl ol i = 
       match (rl,ol) with 
	   (h::t,h1::t1) -> (h,h1,i)::(make_op_reg_i_triple t t1 (i+1)) 
	 | ([],[]) -> []
	 | _ -> failwith ("poor args to make_op_reg_i_triple")
     in
     let op_reg_i_list = make_op_reg_i_triple  (List.tl region_list) op_list 1 in
       (* at this point things start to get a little messy ... *)
     

     
     let uclid_constraint_list = List.map process_op op_reg_i_list in
     let useful_vars = process_block_constraints uclid_constraint_list in
     (*	 
     let uclid_constraints = Predicate.And((List.map process_op op_reg_i_list)) in
       Message.msg_string Message.Normal "Querying Uclid";
       let useful_vars = TheoremProver.uclidSat uclid_constraints in
	 begin
	   match useful_vars with 
	       [] -> failwith "UCLID says trace is satisfiable!"
	     | l -> List.map (fun x -> print_string (x^" ")) useful_vars
	 end;
	 
	*)
	  (* toggle comment*)
     let process_useful_var var =
       try
	 let (index,i) = Hashtbl.find comp_tm_table var in (* toggle comments *)
	 (* let (index,i,_) = var_of_string var in *)            

	 let (reg,_,_) = List.find (fun (_,_,j) -> j=i-1) op_reg_i_list in
	   
	 let location_of_pred = location_coords_of_region reg in
	 let lc = C_SD.location_coords location_of_pred in
	 let l = try Hashtbl.find post_split_table lc with _ -> [] in
	 let newl = 
	   if (List.mem index l) then l else 
	     begin
	       Message.msg_string 
		 Message.Debug (Printf.sprintf "%s : (%d , %d);" 
				   (Predicate.toString (getPred index)) (fst lc) (snd lc) (* (C_SD.location_to_string location_of_pred) *));
	       
	       got_new_split := true;
	       (index::l)
	     end
	 in
	   Hashtbl.replace post_split_table lc newl
       with 
	   Not_found -> (Message.msg_string Message.Debug ("Warning: BOGUS USEFUL VAR!!Not splitting on pipe-var"^(string_of_int var));failwith "Bogus useful var!")
     in
       got_new_split := false;
       List.iter process_useful_var useful_vars;
       if (!got_new_split) then Message.msg_string Message.Debug "Learnt new split predicate"
       else failwith ("No new split information !!")
       ;
       Message.msg_string Message.Debug "Restarting MC with split information"
	 
(*
     if (trace_is_feasible) then failwith("uclid: COUNTEREXAMPLE IS VALID !") else failwith("uclid: INVALID COUNTEREXAMPLE!")
      *)
(**************************************************************************************)

  
(**** interpolation stuff begin ****)
(***********************************************************************************)
(***********************************************************************************)
(*****************     Interpolation based predicate discovery    ******************)
(***********************************************************************************)
(***********************************************************************************)

(**
 * This "module" implements the stuff done for learning predicates by interpolating
 * The main function craig_preds: trace -> list of preds for now
 * Actually we should return a list of locations where the preds should be used
 * but first we shall just return the list of predicates
 *) 

(* the general plan:
   1. set up arrays, reg_array, op_array 
   2. construct constraints using block_concrete_data_pre, put in array
   3. create Atom -> block map 
   4. find interesting block sets -- i.e. those which are unsatisfiable, corresponding to 
      unsatisfiable trace fragments (type: int list list)
   5. List.iter block_refine (4.) ...
   6. return the min of the numbers seen in (4.) as the point from which to refine.
*)

(* note that the function below changes the loc_pred_table *)


let extract_foci_preds arg = 
  let raw_preds = TheoremProver.extract_foci_preds arg in
  let _ = Message.msg_string Message.Debug "raw foci preds" in
  let _ = List.iter (fun x -> Message.msg_string Message.Debug (Predicate.toString x)) raw_preds 
  in
  let cleaned_preds =   (clean_new_predicates raw_preds) in
  let _ = Message.msg_string Message.Debug "cleaned foci preds" in
  let _ = List.iter (fun x -> Message.msg_string Message.Debug (Predicate.toString x)) cleaned_preds 
  in
  let pred_ids = List.map addPred cleaned_preds in
    List.filter (fun x -> x <> -1) pred_ids


let pred_block_table = Hashtbl.create 109

(* funky stuff you can do with functional programming! *)
let next_pos dir x = x + dir


let print_lp_table () = 
  let print_loc l = 
    Message.msg_string Message.Debug (Printf.sprintf "Location (%d,%d)" (fst l) (snd l));
    let p_list = Hashtbl.find loc_pred_table l in
      List.iter (fun p -> Message.msg_string Message.Debug (string_of_int p)) p_list
  in
    List.iter print_loc (Misc.hashtbl_keys loc_pred_table)


let toggle_next_pos npf =
  let dir = npf 0 in
    next_pos (-dir)

let next_back = next_pos (-1) 
(* end funky stuff *)

(* block_a_t ... interface to simplify *)

let simp_ctr = ref 0

let block_simplify_assert p = 
  Message.msg_string Message.Debug ("bs_assert: "^(Predicate.toString p));
  simp_ctr := !simp_ctr + 1;
  TheoremProver.simplify_assume () p

let block_simplify_pop () =
Message.msg_string Message.Debug  "bs_pop"; 
  assert (!simp_ctr > 0);
  TheoremProver.simplify_pop_noflush ();
  simp_ctr := !simp_ctr - 1
    
let block_simplify_reset () = 
  Message.msg_string Message.Debug "bs_reset";
  Misc.repeat_fn ( TheoremProver.simplify_pop) (!simp_ctr);
  simp_ctr := 0

(* end interface to simplify *)
    (* this returns an atomic_region *)
let insert_pred_in_reg reg pred =
  match reg with
      Region.Atomic atr ->
	
	{ Region.location = atr.Region.location;
	  Region.data = Region.Concrete {Region.pred = pred};
	  Region.stack = atr.Region.stack;
	}
    | _ -> failwith "bad argument to insert_pred_in_reg!"

let block_a_t_flag = ref false	
let bat_counter = ref 0	   


(* this is used by predH 7 *)
let dom_rel l1 l2 = 
  (* true if l1 > l2, false otherwise *)
  if l2 = [] then true
  else
    let (sorted_l1,sorted_l2) = (List.sort compare l1,List.sort compare l2) in
    let (min,max) = (List.hd sorted_l2, Misc.list_last sorted_l2)  in
    let rec dom_check l =
      match l with
	  h1::h2::tl -> if (h1 < min && max < h2) then true else dom_check (h2::tl)
	| [h] -> false
	| [] -> false
    in
      dom_check l1
  
let lv_temp_table = Hashtbl.create 109 
 
let block_analyze_trace reg_l op_l id_l = 
  Message.msg_string Message.Debug "In block_a_t:";
  Message.msg_string Message.Debug ("In block_a_t:"^(string_of_int (List.length op_l)));
  let _ = 
    let rec printer idl opl ctr = 
      match (idl,opl) with
	  (id::id_tl,op::op_tl) ->
	    Message.msg_string Message.Debug ("BNum: "^(string_of_int ctr));
	    Message.msg_string Message.Debug ("Id: "^(string_of_int id));
	    (Message.msg_printer Message.Debug Operation.print op);
	    printer id_tl op_tl (ctr+1)
	| ([],[]) -> ()
    in
      printer (Misc.chop_last id_l) op_l 0
  in
  let _ = (block_a_t_flag := false) in (* to check if any new info has been added *)
  let len_reg_l = List.length reg_l in
  let len_op_l = List.length op_l in
  let _ = assert (len_reg_l = len_op_l + 1) in
  let op_array = Array.of_list op_l in
  let reg_array = Array.of_list reg_l in
    (* we've set up the various arrays ... now to get the block constraints *)
  let conj_block_array = Array.create (len_op_l) Predicate.True in
  let block_array = Array.create len_op_l Predicate.True in
  let posn = ref (len_op_l - 1) in 
  let _ = Message.msg_string Message.Debug "block_a_t: building blocks start" in  
  let unsat_flag = ref false in
  let _ = block_simplify_reset () in
  
  let cons_folder (op, post_reg) pre_pred ()  = 
    Message.msg_string Message.Debug ("cons_folder !posn = "^(string_of_int !posn));
    if (not (!unsat_flag)) then
      begin
	let atomic_post_reg = 
	  match post_reg with
	      Region.Atomic atr -> atr
	    | _ -> failwith "bad region given to cons_folder!"
	in
	let block_cons = (block_concrete_data_pre atomic_post_reg (C_SD.get_command op) op) in
	let conj_block_cons = pred_lval_map pre_pred in
	let is_unsat = 
	  block_simplify_assert block_cons;
	  block_simplify_assert conj_block_cons;
	  let rv = TheoremProver.simplify_is_contra () in
	    block_simplify_pop ();
	    rv
	in
	  Array.set block_array !posn block_cons;
	  Array.set conj_block_array !posn conj_block_cons;
	  posn := !posn - 1;
	  if (is_unsat) then
	    Message.msg_string Message.Debug 
	      ("Found unsat at level"^(string_of_int (!posn +1)));
	  if (is_unsat && not (Options.getValueOfInt "predH" >= 6))
	  then unsat_flag := true;
      end
  in
    
  let conj_l = List.map (extract_pred) (Misc.chop_last reg_l) in
  let op_reg_l = List.combine op_l (List.tl reg_l) in
  let _ = if not (Options.getValueOfString "alias" = "") then load_lv_table op_l in
  let _ = List.fold_right2 cons_folder op_reg_l conj_l ();block_simplify_reset ()
  in
  let _ = Message.msg_string Message.Debug "block_a_t: building blocks end" in
    (* the various arrays are loaded! *)
  (*
  let _ = Message.msg_string Message.Debug "Blocks found:(ALLCONS)" in
  let _ = Array.iter (fun x -> Message.msg_string Message.Debug (Predicate.toString x)) block_array in *)
  
 (*   let _ = Message.msg_string Message.Normal "Conj blocks found:" in
      
let _ = List.iter (fun x -> Message.msg_string Message.Normal (Predicate.toString x)) conj_block_l in
 *)
  let suppress_array = Array.make len_op_l false in (* used to "drop assumes" *)
    
  (* block_arrays set up as well ... now the pred --> block map  -- maybe we dont really need this *)
    (* start setting up pred_block_table *) 
    (*
      let _ = Hashtbl.clear pred_block_table in
      let _ = 
      for i = 0 to (Array.length block_array - 1) do 
    (* should have the same length as conj_block_array *)
      let atom_l = (Predicate.get_atoms conj_block_array.(i)) 
      @ (Predicate.get_atoms block_array.(i)) 
      in
      List.iter (fun a -> Hashtbl.add pred_block_table a i) atom_l
      done;
      in
    (* done setting up pred_block_table *)
    *)
  let extract_block i = 
    match suppress_array.(i) with
	true -> Predicate.True
      | false -> block_array.(i)
  in
  let drop_constraint i = 
    let is_assume_constraint i = 
      match (C_SD.get_command (op_array.(i))).C_SD.Command.code with
	  C_SD.Command.Pred _ -> true
	| _ -> false
    in
      match is_assume_constraint i with
	  true -> Array.set suppress_array i true
	| false -> ()
  in
  let restore_constraint i = 
    Array.set suppress_array i false
  in (* we may never actually use this function *)
  let restore_all_constraints () = 
    Array.iteri (fun i a -> Array.set suppress_array i false) suppress_array 
  in
  let check_unsat p =
    let _ = block_simplify_assert p in
    let rv = TheoremProver.simplify_is_contra () in
    let _ = block_simplify_pop () in
      rv
  in
  let check_block_unsat block_l =
    let _ = block_simplify_reset () in
    let head = Misc.min block_l in
    let trace_pred_l = List.map (fun x -> extract_block x) block_l  in
    let all_trace_pred_l = (conj_block_array.(head))::trace_pred_l in
    let all_together = 
      let t_empty = check_unsat (Predicate.And trace_pred_l) in
      let all_empty = check_unsat (Predicate.And all_trace_pred_l) in
	(t_empty || all_empty) 
    in
    let one_by_one = 
      block_simplify_reset ();
      block_simplify_assert (conj_block_array.(head));
      List.iter (fun i -> block_simplify_assert (extract_block i)) block_l;
      TheoremProver.simplify_is_contra ()
    in
      block_simplify_reset ();
      if one_by_one && not (all_together) then
	Message.msg_string Message.Debug "simplify inconsistency!";
      all_together
  in
      (* now to find "interesting" block sets TBD *)
  let (ref_start_pos, ref_start_reg, interesting_block_sets) = 
    begin (* find_interesting_blocks *)
      (* like I said, TBD *) 
      let pos = ref (len_op_l - 1) in
      (*let contra_flag = ref false *) 
      let rec getUsefulBlocks mustSet dir =
	match mustSet with 
	    [] -> 
	      begin
		(* we dont even know the first point of infeasibility *)
		(* we should be moving backwards -- so dir is  "--" *)
		pos := len_op_l - 1;
		block_simplify_assert (conj_block_array.(!pos));
		(* let _ = contra_flag := TheoremProver.simplify_is_contra () in*)
		  while (not (TheoremProver.simplify_is_contra ())  && !pos >= 0) do
		  block_simplify_pop ();
		  block_simplify_assert (extract_block !pos);
		  block_simplify_assert conj_block_array.(!pos);
		  (* contra_flag := TheoremProver.simplify_is_contra (); *)
		  pos := dir (!pos)
		done;
		if (not (TheoremProver.simplify_is_contra ())) then
		  begin
		    block_simplify_reset ();
		    []
		  end
		else
		  begin
		    let rev_dir = toggle_next_pos dir in
		    let next_dir = 
		      match Options.getValueOfString "tsd" with
			  "zig" -> (rev_dir)
			| _ -> (next_back)
		    in
		      block_simplify_reset ();
		      getUsefulBlocks ((rev_dir !pos)::mustSet) next_dir
		  end
	      end
	  | h::t -> 
	      begin
		(* we start zigging from h -- the "head" of the trace is the last elt of h::t *)
		(* head is ONLY used for the forward region ! *)
		let head = Misc.list_last (h::t) in
		  block_simplify_assert (conj_block_array.(head));
		  List.iter (fun i -> block_simplify_assert (extract_block i)) mustSet;
		  (* contra_flag := TheoremProver.simplify_is_contra ();*)
		  if (TheoremProver.simplify_is_contra () (*!contra_flag*)) then
		    begin
		      block_simplify_reset ();
		      (* first a sanity check! *)
		   (*   I've commented out the sanity check
			let good_set = check_block_unsat mustSet in
			if not good_set then 
			  begin
			    (* and yet! *)
			    let _ = TheoremProver.set_simplify_query_flag () in
			    let contra1 = 
			      block_simplify_reset ();
			      block_simplify_assert (Predicate.And ((conj_block_array.(head)::(List.map (fun i -> (extract_block i)) mustSet))));
			      TheoremProver.simplify_is_contra ()
			    in
			      block_simplify_reset ();
			      block_simplify_assert (conj_block_array.(head));
			      List.iter (fun i -> block_simplify_assert (extract_block i)) mustSet;
			      let contra = TheoremProver.simplify_is_contra () in
			      Message.msg_string Message.Error ("contra = "^(string_of_bool contra));
				Message.msg_string Message.Debug ("contra1 = "^(string_of_bool contra1));
			      if contra && not (contra1) then
				failwith "well I'll be!"
			      else
				failwith "getUsefulBlocks returns bogus set!"
			  end
			else *)
			  mustSet
		    end
		  else
		    begin
		      let next_dir = 
			match Options.getValueOfString "tsd" with
			    "zig" -> (pos := dir h; toggle_next_pos dir) 
			      (* this is the last place we zigged to *)
			  | _ -> (pos := len_op_l - 1;next_back)
		      in
		      	while (not (TheoremProver.simplify_is_contra ())) do
   			Message.msg_string Message.Debug 
			  ("Now asserting block: "^(string_of_int !pos));
			block_simplify_assert (extract_block !pos);
			(*contra_flag := TheoremProver.simplify_is_contra ();*)
			pos := dir !pos
		      done; 
		      assert (TheoremProver.simplify_is_contra ());
		      let rev_dir = toggle_next_pos dir in
			block_simplify_reset ();
			getUsefulBlocks ((rev_dir !pos)::mustSet) next_dir
		    end
	      end
      in
	(* predH = 6 trawls the trace and drops assumes in the middle to find "better" blocks *)
	(* predH = 7 drops the blocks that are "dominated" *)
	match (Options.getValueOfInt "predH" >= 6) with
	    true ->
	      begin
		let ubl = ref [] in
		let next_useful_blocks = ref [] in
		  
		  next_useful_blocks := getUsefulBlocks [] next_back;
		  (*if (!next_useful_blocks = []) then failwith "No useful blocks found!";*)
		  while (not (!next_useful_blocks = []) && (List.length !ubl < Options.getValueOfInt "bset")) do
		    Message.msg_string Message.Debug "Another u.b. found";
		    ubl := (!next_useful_blocks)::!ubl;
		    List.iter drop_constraint !next_useful_blocks;
		    pos := 0; (* must start it off at the end! *)
		    (* this WILL do the suppress-assume trick *)
		    (* adds the assume indices to a table, when doing getUseful, skip over !pos
		       if !pos is in the table. Elementary. *)
		    next_useful_blocks := getUsefulBlocks [] next_back
		  done;
		  match (!ubl) with
		      [] -> (-1, Region.Atomic (insert_pred_in_reg (List.hd reg_l) (Predicate.And (Array.to_list block_array))),[])
		    | l -> 
			let head = Misc.min (List.flatten l) in
			let l' = 
			  if (Options.getValueOfInt "predH" = 7) then 
			    let max_l = Misc.maximal dom_rel l in
			      Message.msg_string Message.Debug ("Maximal blocks: "^(string_of_int (List.length max_l)));
			      Message.msg_string Message.Debug ("Maximal blocks: "^(string_of_int (List.length max_l)));
			      Message.msg_string Message.Debug ("Orig blocks: "^(string_of_int (List.length l)));
			      Message.msg_string Message.Debug ("Orig blocks: "^(string_of_int (List.length l)));
			      max_l
			  else l
			in
			  (head, List.nth reg_l head, l')
	      end
	  | false -> 
	      let useful_blocks = getUsefulBlocks [] next_back in
		
		match useful_blocks with
		    [] -> 
		      Message.msg_string Message.Debug "No useful blocks found!";
		      (-1, Region.Atomic (insert_pred_in_reg (List.hd reg_l) (Predicate.And (Array.to_list block_array))),[])
		  | l -> 
		      let head = Misc.min l in
			(head, List.nth reg_l head, [l])
      (* if the trace is feasible then ref_start_pos = -1 and ref_start_reg is the 
	 region for the root from which the trace happens *)
    end
  in
    (* the more vanilla option *)
    (* more vicious hacks *)
  let min_reg_id = ref (-1) in
  let min_adr = ref None in
  
  let focus_refine (block_l : int list) = 
    (* essentially extract the predicates appearing in those blocks *)
    let _ = block_a_t_flag := true in
    let head = Misc.min block_l in
    let head_reg = reg_array.(head) in
    let abst_data_reg = extract_abs_data_reg (head_reg) in
    let trace_pred_l = List.map (fun x -> extract_block x) block_l  in
    let all_trace_pred_l = (conj_block_array.(head))::trace_pred_l in
    let t_empty = check_unsat (Predicate.And trace_pred_l) in
    let _ = block_simplify_reset () in    
      (*let sanity_check = check_block_unsat block_l in     
	if not (sanity_check) then 
	failwith "bad block list in focus_refine!"
	else*)
      begin
	let old_pred_list_len = List.length (abst_data_reg.Region.preds) in 
	let new_adr = 
      do_focus abst_data_reg { Region.pred = 
				 match t_empty with
				     true -> Predicate.And trace_pred_l
				   | false -> Predicate.And (all_trace_pred_l)
			     }
  in
  let new_pred_list_len = List.length (new_adr.Region.preds) in
    if (new_pred_list_len >= old_pred_list_len) then block_a_t_flag := true;
    if (head <= !min_reg_id || !min_reg_id = -1) then
      begin
	min_reg_id := head;
	Array.set reg_array head (insert_data_reg head_reg (Region.Abstract new_adr)) 
      end;
    ()
  end
  in
  let interpolant_refine (block_l : int list) = 
    Message.msg_string Message.Debug "Inside interpolant refine";
    (* assert (not (block_l = [])); *)
    
 
    let ord_block_l = List.sort compare block_l in
      
   let start = List.hd ord_block_l in
    let cons_l = List.map extract_block  ord_block_l in
    let start_cons = 
      if check_unsat (Predicate.And cons_l) then Predicate.True
      else conj_block_array.(start) 
    in	
(* for debugging purposes, we shall dump the interesting blocks *)
    let _ = 
      begin
	Message.msg_string Message.Debug (Printf.sprintf "num blocks = %d" (List.length block_l));
	Message.msg_string Message.Debug (Printf.sprintf "start cons = \n %s" (Predicate.toString start_cons));
	Message.msg_string Message.Debug ("Constraints");
	List.iter (fun p -> Message.msg_string Message.Debug (Predicate.toString p) ) cons_l
      end
    in
(* let proof = getFociProof (start_cons::cons_l) in *)
    let rec cut before_l after_l =
      (* before_l + after_l = block_l *)
      match after_l with
	  [] -> ()
	| end_pos::tl -> 
	    let (refine_begin,refine_end) =
	      if (before_l = []) 
	      then (start,start)
	      else (List.hd before_l + 1, end_pos)
	    in
	    let cons_l_pre = start_cons::(List.map extract_block before_l) 
	    and cons_l_post = List.map extract_block after_l
	    in
	    let foci_preds = 
	      (* JHALA: temporary patch for segfaults! *)
	      match (check_unsat (Predicate.And cons_l_pre)) || (check_unsat (Predicate.And cons_l_post)) with
		  true -> []
		| false -> extract_foci_preds (cons_l_pre,cons_l_post) 
	    in 
	    let _ = Message.msg_string Message.Debug "extraction done" in
	    let _ = List.iter (fun i -> Message.msg_string Message.Debug (string_of_int i)) foci_preds
	    in
	    (*let _ = Message.msg_string Message.Normal ("Array size"^(string_of_int len_op_l)) in *)
	      (* loop begin *)
	      for i = refine_begin to refine_end do
		let _ = Message.msg_string Message.Debug ("Accessing i= "^(string_of_int i)) in
		let (l_id(*,_*)) = C_SD.location_coords (location_coords_of_region (reg_array.(i))) in
		let _ = Message.msg_string Message.Debug (Printf.sprintf "loc_id (%d, %d)" (fst l_id) (snd l_id)) 
		in
		 
		let present_p_l = try Hashtbl.find loc_pred_table l_id with Not_found -> []
		in

		let to_add = 
		List.filter
		  (fun p_i -> not( List.mem p_i present_p_l)) foci_preds 
		     in 
let _ = Message.msg_string Message.Debug "Adding pred "; List.iter (fun p_i -> block_a_t_flag := true; Message.msg_string Message.Debug (string_of_int p_i)) to_add
in
  Hashtbl.replace loc_pred_table l_id (to_add@present_p_l)
		  
              done;
              cut (end_pos::before_l) tl 
   in
     cut [] ord_block_l
 (* so the "head" of the list is the min of the list, should also be the last in 
       the list *)
 in
let pre_flag = predicate_fixpoint_reached () in
let _ = restore_all_constraints () in
let _ = Message.msg_string Message.Debug ("#block sets = "^(string_of_int (List.length interesting_block_sets)))
in
let _ = 
  (match (Options.getValueOfInt "craig" <> 0) with
      true -> List.iter interpolant_refine
    | false -> List.iter focus_refine
  )  
  interesting_block_sets
in
let _ = Message.msg_string Message.Debug "Leaving block_a_t" in
let _ = block_simplify_reset () in 
let _ = bat_counter := !bat_counter + 1 in
  match ((Options.getValueOfInt "craig" = 2) && not (ref_start_pos = -1) && not (!block_a_t_flag)) with
      true -> 
	if (Options.getValueOfBool "localrestart") then
	  begin
	    print_lp_table ();
	    printTableOfPreds (); 
	    dump_abs ();
	    failwith ("No new preds found in block_analyze_trace!")
          end 
        else 
	  begin
	    Options.setValueOfBool "localrestart" true;
	    dump_abs ();
	    (ref_start_pos, reg_array.(ref_start_pos))
	  end
    |  false -> 
	 begin
	   Options.setValueOfBool "localrestart" false;
	   dump_abs ();
	   (ref_start_pos, reg_array.(ref_start_pos))
	 end

(* TODO! Jhala: 6/13/03
   a. drop_assume/the drop_table/using drop_table in getUsefulBlock
   b. interpolant_refine 
   b1. extract_preds: interpolant --> predIndex list. i.e. this function 
       should canonicize and add the predicate to the table etc.   
   b2. getFociProof
   b3. getFociInterpolant_pre
   
*)
						  
(*** interpolation stuff end ****)    

(**************************************************************************************)

  (* the refinment operator required by the ABSTRACTION signature *)
  let focus reg_to_refine empty_reg  =
    Message.msg_string Message.Debug "In focus -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print reg_to_refine ;
    Message.msg_printer Message.Debug Region.print empty_reg ;

    match (reg_to_refine, empty_reg) with
        (Region.Atomic abst_atomic_reg, Region.Atomic empty_conc_atomic_reg) ->
          begin
            try 
              match (abst_atomic_reg.Region.data, empty_conc_atomic_reg.Region.data) with
                  (Region.Abstract abst_data_reg, Region.Concrete empty_conc_data_reg) ->
                    Region.Atomic { Region.location = abst_atomic_reg.Region.location ;
                                    Region.data = Region. Abstract (Stats.time "do_focus" (do_focus abst_data_reg) empty_conc_data_reg) ;
                                    Region.stack = abst_atomic_reg.Region.stack ; }
              | _ -> failwith "focus: input regions should be abstract and concrete in this order"
            with NoNewPredicatesException ->
                 ( Message.msg_string Message.Debug "In focus -- arguments are:" ;
                   Message.msg_printer Message.Debug Region.print reg_to_refine ;
                   Message.msg_printer Message.Debug Region.print empty_reg ;
                   Message.msg_string Message.Debug "No new predicates found. Aborting." ;
                   raise NoNewPredicatesException
                 )
          end
      | _ -> failwith "focus: arguments should be atomic regions"


  (* the following function takes a concrete data region and a list of predicates and returns an 
   abstract region corresponding to it. Cousot would call it alpha *)

     let abstract atomic_conc_reg pred_list =
           let assumption  =
              match atomic_conc_reg.Region.data with
                Region.Concrete cr -> cr.Region.pred
              | _ -> failwith ("bad call to atomic_conc_reg")
           in
           assume_pred_ref := Some(assumption);
           let abstract_bdd = postBdd bddOne (List.map (function i -> (i,getPred i,Predicate.negate (getPred i),true)) pred_list) [] 
	   in
           assume_pred_ref := None;
             Region.Abstract { Region.bdd = abstract_bdd; Region.preds = pred_list}


     let intersect_with_abstraction reg1 reg2 =
        (* reg1 : abstract, reg2 : concrete or abstract, returns reg1 cap reg2 : abstract *)
       let reg2' = precise reg2 in
        match (reg1,reg2') with
          (Region.Atomic atr1, Region.Atomic atr2) ->
           begin
	     Message.msg_string Message.Debug ("intersect_with_abstraction:\n" ^ 
					       "    reg1 = " ^ Region.to_string reg1 ^
					       "    reg2 = " ^ Region.to_string reg2);
             match (atr1.Region.data,atr2.Region.data) with
               (Region.Abstract abst_1,Region.Concrete conc_2) ->
                  let preds = abst_1.Region.preds in
                  let abst_2 = abstract atr2 preds in
                  let newreg2 = Region.Atomic { Region.location = atr2.Region.location;
                                                Region.data = abst_2;
                                                Region.stack = atr2.Region.stack; }
                  in
                  Region.cap reg1 newreg2
             | _ -> failwith "bad call to intersect_with_abstraction"
           end
        | _ -> failwith "bad call to intersect_with_abstraction 2"

  let print_abs_stats () = 
     Message.msg_string Message.Debug ("TP Queries:" ^ (string_of_int !statActualQueriesForPost));
     Message.msg_string Message.Debug ("Reached Cubes:" ^ (string_of_int !statReachedCubes));
     ()


  (* the function giving the possibly enabled operations *)
  let enabled_ops reg =
    match reg with
        Region.Empty ->
          []
      | Region.Atomic atomic_reg ->
          C_SD.get_outgoing_edges atomic_reg.Region.location
      | Region.ParallelAtomic _ -> failwith "Not implemented yet!"
	
(*      | Region.ParallelAtomic p_atomic_reg ->
	  let return_list = ref [] in
	  for i = 0 to (List.length p_atomic_reg.Region.p_control) - 1 do
		let this_control_region = (List.nth p_atomic_reg.Region.p_control i) in
		let current_successors = C_SD.get_outgoing_edges this_control_region.Region.p_location in 
			return_list := (List.append !return_list (List.map(fun x -> (x, i) ) current_successors));
	  done;
	  !return_list *)
      | Region.Union atomic_reg_map ->
          failwith "enabled_ops: argument is a union: unimplemented"
      | Region.ParallelUnion p_atomic_reg_map ->
          failwith "enabled_ops: argument is a parallel union: unimplemented"


  let initialize_abstraction () =
     let compute_max_and_average linfolist =
        let (max, sum) = List.fold_left (fun (m,s) -> fun ((lc1,lc2), lst) -> 
           let l = List.length lst in 
           ((if l > m then l else m), s+l)) 
           (0,0) linfolist in
        let avg = let l = List.length linfolist in if l = 0 then 0.0 else (float_of_int sum) /. (float_of_int l) in
        Message.msg_string Message.Normal ("Maximum predicates active = "^(string_of_int max)) ;
        Message.msg_string Message.Error ("Maximum predicates active = "^(string_of_int max)) ;
        Message.msg_string Message.Normal ("Average predicates active = "^(string_of_float avg)) ;
        Message.msg_string Message.Error ("Average predicates active = "^(string_of_float avg)) ;
	()
    in
    match (Options.getValueOfBool "loadabs") with
      true ->
	begin
	  (* load up loc_pred_table *)
	  let _ = Message.msg_string Message.Normal "initialize abstraction" in
	    let fname = Options.getValueOfString "mainsourcename" in
	    let ic = open_in (fname^".abs") in
	let (pred_list, loc_info_list) = (Marshal.from_channel ic : (Predicate.predicate list * ((int * int) * int list) list)) 
	in
        (* Find the maximum and the average lengths of lists *)
        compute_max_and_average loc_info_list ;
	let _ = List.map just_addPred (List.rev pred_list) in
	let add_tuple (l,p_list) = (Hashtbl.replace loc_pred_table l p_list) in
	  List.iter add_tuple loc_info_list;
	  Message.msg_string Message.Normal "done initializing";
	  ()
      end 
| false -> ()



  let initKill _killLocation _killPreds = 
    
    let adder_fn x = Hashtbl.add killPredTable (Predicate.canonicize x) true in
    let _ = (List.map adder_fn _killPreds) in
      killLocation := _killLocation;
      Message.msg_string Message.Debug "Kill predicates are:";
      Hashtbl.iter (function i -> function j -> Message.msg_string Message.Debug  (Predicate.toString i)) killPredTable;
      Message.msg_string Message.Debug "End kill predicates ";()
	
  (* this function will allow us to create the initial and error regions
     to give to the model checking function *)

  let create_region loclist stack_bool seedPreds initialStatePredicate =
   (* JHALA: The init and error regions have different stacks -- 
      I dont know how to navigate the module hierarchy so as to be able to export 
      this so I will say that true means EveryStack and false means emptyStack *)
    let atom_list = 
      (Predicate.getAtoms seedPreds) @ (Predicate.getAtoms initialStatePredicate)
    in
    let ind_list = List.map (fun x -> addPred x) atom_list in 
    let pred_indexes = List.filter (fun a -> if (a> -1) then true else false) ind_list in
      (*let our_bdd = convertPredToBdd p in *)
    let our_bdd =
      try
	convertPredToBdd initialStatePredicate
      with Not_found -> failwith "create_region: Some predicate in the initial state not found in the predicate table!"
    in
    if (Options.getValueOfBool "par") then
	Region.ParallelAtomic { Region.p_control = 
	  List.map 
	    (fun x -> 
	    { 
	      Region.p_location = x; 
       		Region.p_stack = if stack_bool then Region.EveryStack else Region.CallStack [] 
			      }  
				  )
			      loclist ;
			    Region.p_data = Region.Abstract { Region.bdd = our_bdd ;
							      Region.preds = pred_indexes ; } ;
			  } 
     else
	Region.Atomic {
	  Region.location = List.hd loclist ;
	  Region.data = Region.Abstract { Region.bdd = our_bdd ; Region.preds = pred_indexes ; } ;
	  Region.stack = if stack_bool then Region.EveryStack else Region.CallStack [] ;
        }


	  
  (* for an error location L, the error location is the region where
     some thread is in location L.
     We model this as 
     (L, fake_loc, fake_loc, ..., fake_loc) \/ 
     (fake_loc, L, fake_loc, ..., fake_loc) \/
     ....
     (fake_loc, ..., fake_loc, L)
  *)

  let create_error_region loc num errorStatePredicate = 
    let pred_indexes = [] in
    let our_bdd =
      try
	convertPredToBdd errorStatePredicate
      with Not_found -> failwith "create_region: Some predicate in the error state not found in the predicate table!"
    in
    let err_r = ref Region.bot in
    let e_control_loc loc num i = 
      let rec _cntrl_worker loc i curr lst =
	if (curr = 0) then lst
	else
	  begin
	    if (curr = i ) then 
	      _cntrl_worker loc i (curr - 1) (({ Region.p_location = loc ; 
						 Region.p_stack = Region.EveryStack; }):: lst)
	    else
	      _cntrl_worker loc i (curr - 1) (({Region.p_location = C_SD.fake_location ; 
						 Region.p_stack = Region.EveryStack; }):: lst)
	  end
      in
      _cntrl_worker loc i num []
    in
    for i = 1 to num do 
      let new_r = 
	Region.ParallelAtomic { Region.p_control = e_control_loc loc num i ;
			      	Region.p_data = Region.Abstract { Region.bdd = our_bdd ;
								  Region.preds = pred_indexes ; } ;
			      } 
      in
      err_r := Region.cup !err_r new_r
    done ;
    !err_r

      

   let print_stats fmt () =
     Message.msg_string Message.Normal ("Total number of queries = "^(string_of_int (!statTotalNumberOfQueries))^"\n");
 Message.msg_string Message.Normal ("Total number of non-post queries = "^(string_of_int (!statCoveredQueries))^"\n");
   Message.msg_string Message.Normal ("Worst-case number of post queries = "^(string_of_int (!statWorstCaseQueriesForPost))^"\n");
   Message.msg_string Message.Normal ("Actual number of post queries = "^(string_of_int (!statActualQueriesForPost))^"\n");
   Message.msg_string Message.Normal ("Assume post queries = "^(string_of_int (!statAssumeQueriesForPost))^"\n");
     Message.msg_string Message.Normal ("Total posts = "^(string_of_int (!statNumPost))^"\n");
Message.msg_string Message.Normal ("Total assume posts = "^(string_of_int (!statNumAssumePost))^"\n");

     Message.msg_string Message.Normal ("Total cached = "^(string_of_int (!statTotalCached))^"\n\n");

     Message.msg_string Message.Normal ("List of predicates:\n");
     printTableOfPreds ();
     Message.msg_string Message.Normal 
       ("Maximum number of predicates active together (discounting scope) = "^(string_of_int (!statsMaxNumberOfPredsActive))^"\n");
     Message.msg_string Message.Normal ("Nb. nodes of kill location: "^(string_of_int (!killCounter)));
     Message.msg_string Message.Normal ("Nb. nodes of kill location covered: "^(string_of_int (!killCounterCovered)));
     Message.msg_string Message.Normal ("Nb. nodes of kill location covered bogus: "^(string_of_int (!killCounterCovered_force)));
     Message.msg_string Message.Normal ("Functions visited:");
     Hashtbl.iter (fun x y -> print_string (x^" , ")) statVisitedFunctionTable;
     (* JHALA: for no reason but laziness, I choose to do state saving here -- i.e. dumping abstraction to files *)
     dump_abs (); ()




(* **************** CONCURRENCY **************** *)

(* this function will reset all the appropriate global data structures ... *)

  let reset () = 
    Hashtbl.clear theoremProverCache;
    Hashtbl.clear predTbl;
    Hashtbl.clear predTbl_deep; 
    Hashtbl.clear unchangedPredTable; 
    (*(*noreset*) Hashtbl.clear killPredTable; *)
    (*(*noreset*) killLocation := None; *)
    killCounter := 0;
    killCounterCovered := 0;
    killCounterCovered_force := 0;
    killCounterCovered_seen := 0;
(* [RUPAK] Do not reset the predicates
    tableOfPreds := [];
    predIndex := -1;
*)
    retTarget := junkRet;
    last_loc_pointer := (None);
    assume_pred_ref := None;
    ()


(*  let iter_global_variables = C_SD.iter_global_variables
*)
  let iter_global_variables = C_SD.iterate_all_lvals


  let expression_mentions_locals e =
    try
      let reject_local (name : Expression.lval) = if is_local name then raise Exit; name in
      let _ = Ast.Expression.alpha_convert reject_local e in false
    with Exit -> true

  let rec trim_predicate_to_globals = function
    | Predicate.True -> Predicate.True
    | Predicate.False -> Predicate.False
    | Predicate.And preds -> Predicate.And (List.map trim_predicate_to_globals preds)
    | Predicate.Or preds -> failwith "Abstraction.trim_predicate_to_globals: Predicate.Or"
    | Predicate.Not (Predicate.Atom e) as pred -> if expression_mentions_locals e then Predicate.True else pred
    | Predicate.Not _ -> failwith "Abstraction.trim_predicate_to_globals: Predicate.Not"
    | Predicate.Implies (p, q) -> failwith "Abstraction.trim_predicate_to_globals: Predicate.Implies"
    | Predicate.Iff (p, q) -> failwith "Abstraction.trim_predicate_to_globals: Predicate.Iff"
    | Predicate.All (sym, pred) -> failwith "Abstraction.trim_predicate_to_globals: Predicate.All"
    | Predicate.Exist (sym, pred) -> failwith "Abstraction.trim_predicate_to_globals: Predicate.Exists"
(*
    | Predicate.All (sym, pred) -> Predicate.All (sym, trim_predicate_to_globals pred)
    | Predicate.Exist (sym, pred) -> Predicate.Exist (sym, trim_predicate_to_globals pred)
*)
    | Predicate.Atom e as pred -> if expression_mentions_locals e then Predicate.True else pred

  let trim_region_to_globals =    
    Region.concretize_map
      (fun l -> C_SD.fake_location) (fun s -> Region.EveryStack)
      (fun p -> Message.msg_string Message.Debug ("trim_region_to_globals:" ^ 
						  Predicate.toString p); 
	        (* trim_predicate_to_globals*) p)

  let erase_data = function
    | Region.Empty -> Region.Empty
    | Region.Atomic a ->
	let new_data = match a.Region.data with
	| Region.Concrete _ -> Region.Concrete {Region.pred = Predicate.True}
	| Region.Abstract adr ->
	    Region.Abstract {adr with Region.bdd = CaddieBdd.bddTrue bddManager}
	in
	Region.Atomic {a with Region.data = new_data}
    | Region.ParallelAtomic a ->
	let new_data = match a.Region.p_data with
	| Region.Concrete _ -> Region.Concrete {Region.pred = Predicate.True}
	| Region.Abstract adr ->
	    Region.Abstract {adr with Region.bdd = CaddieBdd.bddTrue bddManager}
	in
	Region.ParallelAtomic {a with Region.p_data = new_data}
    | Region.Union _ -> failwith "Abstraction.erase_data"
    | Region.ParallelUnion _ -> failwith "Abstraction.erase_data parallel union unimplemented"

    (* this takes a predicate formula and returns a region out of it *)
  let predicate_to_region p = 
    Region.Atomic 
      { Region.location = C_SD.fake_location;
	Region.data = Region.Concrete { Region.pred = p} ;
	Region.stack = Region.EveryStack;
      }


  let create_concrete_true_region edge =
    Region.Atomic 
      { Region.location = (C_SD.get_target edge) ;
	Region.data = Region.Concrete { Region.pred = Predicate.True } ;
	Region.stack = Region.EveryStack
      } 

  let create_abstract_true_region edge =
    Region.Atomic 
      { Region.location = (C_SD.get_target edge) ;
	Region.data = Region.Abstract { Region.bdd = bddOne ; 
					Region.preds = List.map Misc.fst3 (!tableOfPreds)
				      } ;
	Region.stack = Region.EveryStack
      } 
    

  type access_type = Anone | Aread | Awrite | Aboth
  let accessing op v =
    let reads, writes = C_SD.reads_and_writes_at_edge op in
    match (List.mem v reads, List.mem v writes) with
    | false, false -> Anone
    | false, true -> Awrite
    | true, false -> Aread
    | true, true -> Aboth

  let get_all_writes op =
    snd (C_SD.reads_and_writes_at_edge op)

  let get_all_reads op =
    fst (C_SD.reads_and_writes_at_edge op)

  let reads_and_writes op = C_SD.reads_and_writes_at_edge op 

  let aliasesOf lval = if can_escape lval then ( lval::(AliasAnalyzer.getAllAliases lval)) else []

  let eq_addr_constraint_region lval1 lval2 = 
    let p = 
	Predicate.Atom 
	(Expression.Binary 
	  (Expression.Eq, 
	   Expression.addressOf (Expression.Lval lval1),
	   Expression.addressOf (Expression.Lval lval2)))
    in
      predicate_to_region p

  (* this should return the list of possible aliases in the whole program, of some lval 
     maybe we should keep a Hashtbl or something and query that every time 
   *)

  let make_multiply_instantiable_region r1 r2 lval1 lval2 = 
    Message.msg_string Message.Debug "In make_multiply instantiable\n\n" ;
    Message.msg_string Message.Debug ("r1 is"^(Region.to_string r1)) ;
    Message.msg_string Message.Debug ("r2 is"^(Region.to_string r2)) ;
    let rename_fun = new_name (brand_new_suffix ()) in
    let eq_addr_const_reg = eq_addr_constraint_region lval1 (Expression.alpha_convert_lval rename_fun lval2) in 
      Region.cap 
	(Region.cap !assumed_invariants_region eq_addr_const_reg)
	(Region.cap r1 (Region.rename_vars r2 rename_fun))
	
			 
  (* the below function should ONLY be called on two regions known to be such that:
     1. lval_1, lval_2 are such that they are accessed in region r1 r2 respectively
     2. neither of the lvals is a lock.
  *)
  let multiply_instantiable r1 r2 lval1 lval2 =
    let r = make_multiply_instantiable_region r1 r2 lval1 lval2 
    in
      if Region.is_empty r then None else Some r

  
    
	
(* **************** END concurrency **************** *)

(* ******************************************************************************)
(**
 * This module implements a procedure that when given a DAG cfa, with a single sink,
  (and a single root ?) pulls the wp back from the sink to the root  .
 *)
 
(* here we go ... *)
(*****************************************************************************************)
(************************************ MAP functions **************************************) 


(********************************************************************************)
(* Jhala: Adding blast-proof.ml stuff over here ... 
   we'll let Gregoire fuss with the modules etc *)
(***** BLAST-PROOF *****)
(* module BlastProof =
   struct *)
  (* Data Structures for Proofs *)  
   type  data_literal = (int * bool) 
   type  data_cube =  data_literal list 
   type   data_region =  data_cube list
   
   type atomic_proof_region = {
	   prog_counter             : C_SD.location;
	   call_stack               : Region.stack_contents;
	   data_reg_cubes           :  data_region;
          (* The above is the list of cubes --before-- we turned it into bdds the invariant on the proof tree is that the two data regs should in fact be identical logically. We shall check this when constructing the proof -- indeed, it should be possible to do the model checking only on these lists *)
     (*data_reg_preds           : int list;  list of relevant predicates -- not needed, due to the check that we shall perform when making the proof *)
     (* data_reg_bdd             : CaddieBdd.bdd ; should come from bddfying the data_reg_preds *)
	 }
   type  proof_region = 
       Empty_pr
     | Atomic_pr of atomic_proof_region
	 
   type lemma_index = int 
   (* The proof tree will have edges labelled with "lemmas" and a separate (lemma_index, blast_lemma) Hashtbl *)
   type vampyre_lemma_index =  int 
   type cube_post_lemma = 
       Empty_post of lemma_index
     | Nonempty_post of ( data_literal * vampyre_lemma_index) list

   type blast_lemma = 
       Vampyre_Lemma of Proof.proof 
       (* The most basic kind of proof *)
     | Post_Edge_Lemma of ( data_cube * cube_post_lemma) list 
(* [(cube,[...(l,prf)...])...] : prf is a proof of cube -> wp(l,edge) *)
(* proofs labelling the edges of the tree *)
(* Note that for post of assume, the assume pred is not going to be in the hyp: data_cube as given above -- that must come from the op labelling the edge (PROOF-CHECKER)*)
     | Covering_Lemma (* TBD: Still have to figure out the covering *)
     | Skip_Lemma (* for SKIPs and so on whenever the post is the same *)



(* Code to ask theorem prover queries and check satisfiability *)        
let theoremProverCache_proof = Hashtbl.create 37
let trivialProof = Proof.G("truei", [])
 
  (* Statistics: Number of queries to the Theorem Prover *)
let statTotalNumberOfQueries_proof = ref 1
let statTotalCached_proof = ref 0
let statTotalMismatches = ref 0
let vampyre_lemma_counter = ref 1 


(* [Rupak] Hack to print ascii lf readable proofs *)

let generateLfOutput expression proof =
   let v_exp = Input.pushNeg (TheoremProver.convertToVampyreSyntax expression)
   in
   let olf = !Input.printLFSyntax 
   in
   try
     Input.printLFSyntax := true;
   match proof with
     Proof.G("truei", []) -> () (* trivial proof. do not print *)
   | Proof.Unimplemented ->        Message.msg_string Message.Debug "Vampyre bug: raised Unimplemented!"
   | _ ->
     ignore (VampyrePretty.fprintf (!VampyreErrormsg.proofChannel)
     "pf%s : pf @[%a@] \n%% @! :: @[%a@].\n" 
     ("foo"^(string_of_int (!vampyre_lemma_counter))) Input.dp v_exp Proof.dp proof);
     Input.printLFSyntax := olf
   with _ ->
     begin
       Message.msg_string Message.Debug "Vampyre bug: raised Unimplemented!";
(* Trivial proof : dont print it *)
       Input.printLFSyntax := true;
       ignore (VampyrePretty.fprintf !VampyreErrormsg.proofChannel
       "pf%s : pf @[%a@] \n%% @! :: @[%a@].\n" 
       ("foo"^(string_of_int (!vampyre_lemma_counter))) Input.dp v_exp Proof.dp (Proof.G("truei", [])));
       Input.printLFSyntax := olf
     end

let generateLfOutput_noproof expression =
   let v_exp = Input.pushNeg (TheoremProver.convertToVampyreSyntax expression)
   in
   let olf = !Input.printLFSyntax 
   in
     Input.printLFSyntax := true;
     ignore (VampyrePretty.fprintf !VampyreErrormsg.proofChannel
	       "pf%s : pf @[%a@] \n%% @! :: @[%a@].\n" 
	       ("foo"^(string_of_int (!vampyre_lemma_counter))) Input.dp v_exp Proof.dp (Proof.G("truei", [])));
     Input.printLFSyntax := olf
	 
	 


let askTheoremProver_proof pred =
    (*Message.msg_string Message.Debug ("Asking query..."^(string_of_int !statTotalNumberOfQueries)^"\n");*)
  Message.msg_string Message.Debug "In askTheoremProver_proof";
  statTotalNumberOfQueries_proof := !statTotalNumberOfQueries_proof + 1;
  if (Hashtbl.mem theoremProverCache_proof pred) 
  then 
      begin
        statTotalCached_proof := !statTotalCached_proof + 1;
        Some (snd (Hashtbl.find theoremProverCache_proof pred))
      end
  else 
      begin
        let t2 = TheoremProver.queryExp_proof pred in
	  match t2 with
	      Some(_t2) -> 
		begin
		  Hashtbl.add theoremProverCache_proof pred (_t2,!vampyre_lemma_counter);
		  generateLfOutput pred _t2;		  
		  vampyre_lemma_counter := 1 + !vampyre_lemma_counter;
		  Some(!vampyre_lemma_counter -1)
		end
	    |_ -> None
      end

let askTheoremProver_noproof conjpred implpred =
    (*Message.msg_string Message.Debug ("Asking query..."^(string_of_int !statTotalNumberOfQueries)^"\n");*)
  Message.msg_string Message.Debug "In askTheoremProver_noproof";
  statTotalNumberOfQueries_proof := !statTotalNumberOfQueries_proof + 1;
  if (Hashtbl.mem theoremProverCache conjpred) 
  then 
    begin
      statTotalCached_proof := !statTotalCached_proof + 1;
      if (Hashtbl.find theoremProverCache conjpred) then Some(1) else None
    end
  else 
    begin
      let t2 = TheoremProver.queryExp conjpred in
        Hashtbl.add theoremProverCache conjpred t2;
	match t2 with
	    true -> 
	      begin 
		generateLfOutput_noproof implpred;
		vampyre_lemma_counter := 1 + !vampyre_lemma_counter;
		Some(1)
	      end
	  | false -> None
    end



(* This is the (lemma_index,blast_lemma) table *) 
let lemmaTable = Hashtbl.create 53
let lemma_counter = ref 0 

let insert_lemma lem = 
  (* Message.msg_string Message.Normal "Adding to lemmaTable";*)
  let lc = !lemma_counter in
    Hashtbl.add lemmaTable lc lem;
    lemma_counter := (!lemma_counter)+1;
    lc

let skip_lemma_index = insert_lemma Skip_Lemma (* 0 is the default skip lemma *)

let clear_lemma_table () = 
  Hashtbl.clear lemmaTable; lemma_counter := 0


let print_proofTable () = 
  Message.msg_string Message.Normal ("Total number of lemmas: "^(string_of_int !lemma_counter)^"\n");
  Message.msg_string Message.Normal ("Total number of vampyre obligations: "^(string_of_int !vampyre_lemma_counter)^"\n");
  Message.msg_string Message.Normal ("Total number of REGION MISMATCHES: "^(string_of_int !statTotalMismatches)^"\n");
 

(*  let table_chan = open_out_bin "foo.table" in
    Marshal.to_channel table_chan lemmaTable [] ;
    close_out table_chan; *)
    (*close_out (!VampyreErrormsg.proofChannel);*)
    ()

(* Need the following functions ... *)
   (* done ..print_proof_region : *)
   (* done ..check_reg_eq : proof_region -> Region.t -> bool *)
   (* done ..extract_proof_region : Region.t -> proof_region *)

   let print_proof_region preg_formal =
     let print_loc loc = 
       print_string ((C_SD.location_to_string loc )^";") 
     in
     let _ = print_string "Proof Region: \n" in
       match preg_formal with
	   Empty_pr -> print_string "Empty \n"
	 | Atomic_pr (preg) -> 
	     begin
	       let _ = print_string ("Location: "^(C_SD.location_to_string preg.prog_counter)^"\n"); print_string "Stack : [ " in
	       let _ = 
		 match preg.call_stack with
		     Region.EveryStack -> print_string "EveryStack"
		   | Region.CallStack s -> List.iter print_loc s 
	       in
		 print_string "] \n";
		 let print_literal (i,b) = 
		   match b with
		       true -> print_string (Predicate.toString (getPred i));print_string "."
		     | false -> print_string "~"; print_string (Predicate.toString (getPred i)); print_string "."
		 in
		 let print_cube l = print_string "+ "; List.iter print_literal l; print_string "\n" 
		 in
		   List.iter print_cube preg.data_reg_cubes;
		   print_string "End proof region \n"
	     end

   exception DifferentRegions 

   let make_data_region_bdd data_reg = 
     let _bdd = ref bddZero in
     let rec make_cube_bdd cube = 
       match cube with
	   [] -> bddOne 
	 | (p,true)::tl ->  CaddieBdd.bddAnd (getPredIndexBdd p) (make_cube_bdd tl) 
	 | (p,false)::tl -> CaddieBdd.bddAnd (CaddieBdd.bddNot (getPredIndexBdd p))  (make_cube_bdd tl) 
     in
     let add_cube c = 
       _bdd := CaddieBdd.bddOr (make_cube_bdd c) (!_bdd)
     in
       List.iter add_cube data_reg; !_bdd
       
   let check_reg_eq preg_formal reg = 
      (* We shall only check the equivalence of the data regions ... *)
     let retval = 
       try
	 match preg_formal with 
	     Empty_pr -> (reg = Region.bot)
	   | Atomic_pr preg -> 
	       begin
		 let reg_data_bdd = 
		 match reg with 
		     Region.Atomic a -> 
		       begin
			 match a.Region.data with
			     Region.Abstract ab -> ab.Region.bdd
			   | _ -> raise DifferentRegions
		       end
		   | _ -> raise DifferentRegions
	       in
		 let preg_data_bdd = make_data_region_bdd preg.data_reg_cubes 
		 in
		   CaddieBdd.bddEqual preg_data_bdd reg_data_bdd
	       end
       with
	   DifferentRegions -> false
     in
       if ( not retval) then 
	 begin
	   (* Message.msg_string Message.Error "REGION MISMATCH !"; *)
	   statTotalMismatches := !statTotalMismatches + 1
	 end;
       retval

   let extract_data_proof_region data_reg = 
     let rip_bdd b = 
       let cube_list = ref [] in
       let rip_cube (cube : int array) =
	 let conjl = ref [] in
	   for i = 0 to (Array.length cube - 1) do
             if (cube.(i)==2) 
             then () 
             else  conjl := (i,(cube.(i)==1))::(!conjl)
	   done;
	   cube_list := (!conjl)::(!cube_list)
	 in
	 if (CaddieBdd.bddEqual b bddOne) then [[]] 
	 else ( CaddieBdd.bddForeachCube b rip_cube; !cube_list)
     in
       rip_bdd data_reg.Region.bdd

   let extract_proof_region reg = 
       let rip_bdd b = 
	 let cube_list = ref [] in
	 let rip_cube (cube : int array) =
	   let conjl = ref [] in
	     for i = 0 to (Array.length cube - 1) do
               if (cube.(i)==2) 
               then () 
               else  conjl := (i,(cube.(i)==1))::(!conjl)
	     done;
	     cube_list := (!conjl)::(!cube_list)
	 in
	   if b = bddOne then [[]] 
	   else ( CaddieBdd.bddForeachCube b rip_cube; !cube_list)
       in
	 match reg with 
             Region.Atomic a -> 
	       begin
		 match a.Region.data with
		     Region.Abstract ab -> 
		       Atomic_pr ( 
			 { 
			   prog_counter = a.Region.location;
			   call_stack = a.Region.stack;
			   data_reg_cubes = Stats.time "rip_bdd" rip_bdd ab.Region.bdd;
			 })
			 
		   | _ -> failwith "extract_proof_region: bad region !"
	       end
	   | Region.Empty -> Empty_pr
	   | _ -> failwith "extract_proof_region: bad region !"
		 
		 
(* print_proofTable *)   
(* post_proof : Region.t -> Operation.t -> (proof_region * lemma_index) *)
(* the next 4 functions culminate in post_proof *)
(* postBdd_proof, abstract_data_post_proof_graf_saidi,abstract_data_post_Proof, post_proof *)

   let assume_pred_ref_proof = ref None 
   let postBdd_proof (bdd, pred_list) =
     Message.msg_string Message.Debug "In postBdd_proof..\n";
     Message.msg_string Message.Debug ("post of assume"^(match !assume_pred_ref_proof with Some(_) -> "YES" | _ -> "NO")^"\n");
     Message.msg_string Message.Debug ("Bdd = "^(Predicate.toString (convertBddToPred bdd)));
     Message.msg_string Message.Debug "\nList of predicates = ";
     let rec loop lst = 
       match lst with 
	   [] ->  () 
	 | (a,b,c)::rest -> (ignore (Message.msg_string Message.Debug (Predicate.toString (getPred a)));ignore (Message.msg_string Message.Debug (Predicate.toString b));ignore (Message.msg_string Message.Debug (Predicate.toString c));  loop rest) 
     in
     let _ = loop pred_list 
     in
  (* let (changed_pred_list , unchanged_pred_list ) =  pred_list in*) 
  (* pretend that everyone has changed *)  
  (* we shall not use wp_different_filter as its an optimization *)
     let sumofcubesList = ref [] in 
     let conjunctL = ref [] in
     let asm_cubeL = ref [] in
	    (* conjunct that the present cube produces post-command c *)
    (* deleting _check_Pred_list_simplify as here we must use vampyre *)
     let _check_Pred_list assumptions pred_triple_list = 
       Message.msg_string Message.Debug "in _check_Pred_list_pr";


       (* URGENT: VAMPYRE BUG! let is_cube_consistent  = TheoremProver.assume assumptions in *)
       let quickCheck wpx eToCheck =
         let rec quick_canonicize pred =
           match pred with
             Predicate.Atom e ->
	       begin
        	  let atm = 
	           match e with
	             Expression.Binary (Expression.Ne, Expression.Lval (Expression.Symbol sym1), Expression.Lval (Expression.Symbol sym2)) ->
		       if (sym1 < sym2) then 
		          Expression.Binary (Expression.Ne, Expression.Lval (Expression.Symbol sym1), Expression.Lval (Expression.Symbol sym2))
		       else
		          Expression.Binary (Expression.Ne, Expression.Lval (Expression.Symbol sym2), Expression.Lval (Expression.Symbol sym1))
	           | Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol sym1), Expression.Lval (Expression.Symbol sym2)) ->
		       if (sym1 < sym2) then 
		          Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol sym1), Expression.Lval (Expression.Symbol sym2))
		       else
		          Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol sym2), Expression.Lval (Expression.Symbol sym1))
		    
	           |	Expression.Binary (Expression.Ne, ee, Expression.Lval (Expression.Symbol sym2)) ->
		          Expression.Binary (Expression.Ne, Expression.Lval (Expression.Symbol sym2), ee) 
	           |	Expression.Binary (Expression.Eq, ee, Expression.Lval (Expression.Symbol sym2)) ->
		          Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol sym2), ee)
	           |	Expression.Binary (Expression.Ge, e1, e2) -> Expression.Binary(Expression.Le, e2, e1)
	           |	Expression.Binary (Expression.Gt, e1, e2) -> Expression.Binary(Expression.Lt, e2, e1)
		  
	           |	_ -> e
	       in
	       Predicate.Atom atm
	     end
          |  Predicate.Not ap -> quick_canonicize (Predicate.negateAtom ap) 
          |  Predicate.True -> Predicate.True
          |  Predicate.False -> Predicate.False
          |	_ -> failwith ("quick_canonicize: Predicate "^(Predicate.toString pred)^" is not atomic")
         in
	 let wpx_canon = quick_canonicize wpx in
	 let finder e = (quick_canonicize e) = wpx_canon in
	   if (List.exists finder assumptions) then (Some(0)) else 
	     
	     (askTheoremProver_proof eToCheck)

       in
	 (* check if wpx follows trivially from the assuptions if so return Some(trivialProof) else return (askTheoremProver_proof eToCheck) 
*)
       let _check_pred assumptions (p,wpp,wpnp) = 
	 let eToCheck1 = List.fold_right (fun a -> fun b -> Predicate.implies a b) assumptions wpp in
	 let eToCheck2 = List.fold_right (fun a -> fun b -> Predicate.implies a b) assumptions wpnp in
	 let (ans1,ans2) = 
	   if (Options.getValueOfBool "pfgen") then 
	     ((quickCheck wpp eToCheck1),(quickCheck wpnp eToCheck2))
	   else
	     begin
	       let asm =  Predicate.conjoinL assumptions in
	       let conj_e1 = Predicate.implies asm wpp in
	       let conj_e2 = Predicate.implies asm wpnp in
		 ((askTheoremProver_noproof conj_e1 eToCheck1),(askTheoremProver_noproof conj_e2 eToCheck2))
	     end
	 in
	   match (ans1,ans2) with
               ((Some(i1)),(Some(i2))) ->
		 begin
		   Message.msg_string Message.Debug ("Cube implies both wp(p) and wp(neg p): "^(Predicate.toString (getPred p))); 
		   conjunctL := ((p,true),i1)::((p,false),i2)::(!conjunctL);
		   true
		 end
	     | (Some(i1),_) -> 
		 begin 
                   conjunctL := ((p,true),i1)::(!conjunctL);
		   true               
		 end
		 
	     | (_,Some(i2)) -> 
		 begin 
		   conjunctL := ((p,false),i2)::(!conjunctL);
		   true               
		 end
	     | (_,_) -> false
       in 
       let _ = List.map (_check_pred assumptions) pred_triple_list in
         ()
	 (* URGENT: fix the is_cube_consistent in the original post -- it must be consistent as we should have checked earler *)
     in
     let _check_Cube _pred_list (cube : int array)  =
       (Message.msg_string Message.Debug (" In _check_Cube"));
       conjunctL := []; asm_cubeL := [];   
       let cubePostIsEmpty = 
	 (Message.msg_string Message.Debug (" In cubePostIsEmpty"));
         match (!assume_pred_ref_proof) with 
             Some(e) -> askTheoremProver_proof ( List.fold_right (fun a -> fun b -> Predicate.implies a b) [(convertCubeToPred cube);e] Predicate.False)
           | None -> askTheoremProver_proof ( List.fold_right (fun a -> fun b -> Predicate.implies a b) [(convertCubeToPred cube)] Predicate.False)
 (* URGENT ! Jhala: fix this in the original post! it could ? still be false ... *)
       in 
       let pref = match (!assume_pred_ref_proof) with None -> ref [] | Some(e) -> ref [e] 
       in
	 for i = 0 to (Array.length cube - 1) do
	   (Message.msg_string Message.Debug (" In do_loop"));
	   if (cube.(i)==2) 
	   then () 
	   else 
	     begin
	       let i_pred = getPred i in
(* Assume the invariant that i_pred : Predicate.Atom *)
		 if (cube.(i)==1) then
		   begin
		     pref := i_pred::(!pref);
		     asm_cubeL := (i,true)::(!asm_cubeL);
		   end
		 else
		       begin
			 pref := (Predicate.negate i_pred)::(!pref);
			 asm_cubeL := (i,false)::(!asm_cubeL);
		       end
	     end
	 done;
	 
	 match cubePostIsEmpty with 
	     Some(prf) -> (* inconsistent, empty *)
	       sumofcubesList := (!asm_cubeL ,Empty_post (prf))::(!sumofcubesList)
		 
	   | None -> (* consistent, nonempty *) 
	       begin  
	 (* Now ask the thmprover if pref -> each fellow in predlist *)
		 (_check_Pred_list !pref  pred_list); 
		 sumofcubesList := (!asm_cubeL , Nonempty_post (!conjunctL))::(!sumofcubesList) 
	       end
     in
       begin
	 if (CaddieBdd.bddEqual bdd bddZero) then
	   sumofcubesList := []
	 else 
	   begin
	     if (CaddieBdd.bddEqual bdd bddOne) 
	     then
               let _tmpcube = (* build a temporary cube with nothing in it *)
		 Array.make (*size *)(getCurrentIndex ()+2) (* content *)2 
		   
(* Note: The size is certainly a vast	overestimate, but this
   is the easiest way. The "+2" removes additional complications relating
   to index being -1 initially. Note that this is a fake cube -- we should
   never try to print it or convert it to a predicate!Ugly hack, but 
   works. So far *)
               in
		 _check_Cube pred_list _tmpcube
	     else
               ignore (CaddieBdd.bddForeachCube bdd (_check_Cube pred_list));
	   end;
	 (Message.msg_string Message.Debug ("Returning from Post-Bdd_proof"));
	 
         let extract_reg_from_pep pep = 
	   let take (dcube,cpl) = 
	    match cpl with
		Empty_post _ -> None 
	      | Nonempty_post l -> Some (List.map fst l)
	   in
	     Misc.list_gather_option (List.map take pep)
	 in
	 let post_data_cubeList = extract_reg_from_pep (!sumofcubesList) in
	 let post_lemma_index = insert_lemma ( Post_Edge_Lemma !sumofcubesList) in
	   (post_data_cubeList,post_lemma_index)
       end
       
 let abstract_data_post_proof_graf_saidi abst_data_reg cmd edge phi_table : (data_literal list list * lemma_index) =
     Message.msg_string Message.Debug "In abstract_data_post_proof SLAM_POST -- arguments are:" ;
     Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;
     Message.msg_printer Message.Debug C_SD.Command.print cmd ;
     let _H_cmd_proof _cmd bdd predlist =
       Message.msg_string Message.Debug ("In _H of abstract_data_post_Proof, bdd = "^(Predicate.toString (convertBddToPred bdd)));
       let make_pred_triple a = 
         let pr = getPred a in
           match (!assume_pred_ref_proof) with 
               None -> 
                 let cr = concrete_data_pre false false { Region.pred = pr ; } _cmd edge phi_table in
                 let wppos = cr.Region.pred in
                 let ncr = concrete_data_pre false false { Region.pred = Predicate.negate pr ; } _cmd edge phi_table in
                 let wpneg = ncr.Region.pred in
                   (a,wppos,wpneg)
             | Some(e) -> (a,pr,Predicate.negate pr)
       in
         postBdd_proof (bdd,(List.map make_pred_triple predlist))
     in
     let (_H_proof : CaddieBdd.bdd -> Region.predList -> ( data_literal list list * lemma_index)) = _H_cmd_proof cmd in
     match cmd.C_SD.Command.code with 
           C_SD.Command.FunctionCall fcallExp ->
	     let (fname,_subs,_) = (deconstructFunCall fcallExp) in
               if (fname = __NotImplementedFunctionName || not (C_SD.is_defined fname)) then (extract_data_proof_region abst_data_reg, skip_lemma_index)
               else 
		 let asgn_block =
                   { cmd with C_SD.Command.code = C_SD.Command.Block (List.map (fun (x,e) -> C_SD.Command.Expr( Expression.Assignment (Expression.Assign,Expression.Symbol(x), e ))) _subs) }
		 in
		   (* Region.preds = abst_data_reg.Region.preds;*)
      	 (_H_cmd_proof asgn_block (abst_data_reg.Region.bdd) abst_data_reg.Region.preds) (* {Region.bdd = bddZero;Region.preds = []}*)
		   
	 | C_SD.Command.Skip ->  (extract_data_proof_region abst_data_reg, skip_lemma_index)
             (* let _ = (retTarget := junkRet) in
		_H_proof  (abst_data_reg.Region.bdd) abst_data_reg.Region.preds*)
		 
     (* Stats.time "rip_bdd" rip_bdd abst_data_reg.Region.bdd *)
	 | C_SD.Command.Block [] ->  (extract_data_proof_region abst_data_reg, skip_lemma_index)
	 | C_SD.Command.Block l  -> 
             if ( not (List.exists (function x -> match x with C_SD.Command.Return e -> true | _ -> false) l))
	     then 
	       let _ = (retTarget := junkRet) in
                 _H_proof  (abst_data_reg.Region.bdd) abst_data_reg.Region.preds
		   
	     else 
	       let (targ, fname, _) = lastTarget () in
 (* Jhala: the t.o.s. is passed using last_loc_pointer  reg.Region.stack *) 
	       let _ = (retTarget := targ) in
	       let (newPreds,delPreds) = removeFunPreds fname abst_data_reg.Region.preds in
               let _ = Message.msg_string Message.Debug "Return: Calling _H_proof" in
	       let (newCubes,index) = (_H_proof  (abst_data_reg.Region.bdd) abst_data_reg.Region.preds) in
               let _ = Message.msg_string Message.Debug "Return: Back from _H_pr" in
	       let _predToVar i =  -1 in (* JHALA: To be done ... ??? *)
	       let _ = (retTarget := junkRet) in
                (* Region.preds = newPreds@delPreds;*)
      		 (newCubes,index)
		 
	 | C_SD.Command.Pred e -> 
             (* if (e = Predicate.True) then abst_data_reg else : OPT*) 
               begin
		 assume_pred_ref_proof := Some(e);
		 let (newCubes,index) = _H_proof (abst_data_reg.Region.bdd) abst_data_reg.Region.preds 
		 (* Region.preds = abst_data_reg.Region.preds ; *)
		 in  
		   assume_pred_ref_proof := None;          
		   (newCubes,index)
               end
	 | C_SD.Command.Phi (_,_) 
	 | C_SD.Command.Havoc _ 
	 | C_SD.Command.HavocAll -> failwith "abstract_data_post_proof_graf_saidi: Havoc and HavocAll : TBD"
	       
   let abstract_data_post_proof abst_data_reg  cmd edge phi_table = 
     match (Options.getValueOfString "post") with
         "H" -> 
	   begin
	     Message.msg_string Message.Error "ERROR!: Proof Generation with H-post is not yet implemented"; 
	     failwith "Unimplemented post-proof with H"
	   end
       | _  -> 
	     let (rr : (data_literal list list * lemma_index)) = 
	       (* Stats.time "Post-proof Graf-Saidi"*) (abstract_data_post_proof_graf_saidi abst_data_reg  cmd edge) phi_table 
	     in
	       rr

   let post_proof reg op phi_table =
     Message.msg_string Message.Debug "In post_proof -- arguments are:" ;
     Message.msg_printer Message.Debug Region.print reg ;
     Message.msg_printer Message.Debug Operation.print op ;
     
     let preds_of_region reg = 
       match reg with
	   Region.Atomic atr -> 
             begin     
               match atr.Region.data with 
		   Region.Abstract abst_data_reg -> List.map getPred abst_data_reg.Region.preds
		 | _ -> failwith "post_proof: Bad call to preds_of_region "
             end
	 | _ -> failwith "post_proof: Bad call to preds_of_region "
     in
       match reg with
           Region.Empty -> 
	     begin
	       Message.msg_string Message.Error "WARNING: post_proof called with empty region";
	       (Empty_pr,skip_lemma_index) (* failwith "post_proof: arg is Empty Region !" *)
	     end
	 | Region.Atomic atomic_reg ->
	     begin
               try 
		 if (atomic_reg.Region.location <> (C_SD.get_source op)) then
		   begin
                     Message.msg_string Message.Error "WARNING: post: blast internal error suspected" ;
                     Message.msg_string Message.Error "               operation do not match this region's location" ;
                     failwith "post_proof: args mismatch !"
                   (* Region.bot *) 
		   end
               else
		 begin
                   try 
                     let callStack = Region.list_of_stack atomic_reg.Region.stack in
	             let _ = 
		       if (callStack <> []) 
                       then (last_loc_pointer := Some (List.hd callStack)) 
                       else ( last_loc_pointer := None)
		     in
	          (* So last_loc_pointer points to the top of the stack at all times or to errorloc Rif the stack is empty *)
                     let (new_location, new_stack) = 
		       
			 if (isFunCall op) then
			   begin
                             match (C_SD.get_command op).C_SD.Command.code with
				 C_SD.Command.FunctionCall fcexp -> 
				   let (_fname,_,target) = deconstructFunCall fcexp in
				     if (_fname = __NotImplementedFunctionName || not (C_SD.is_defined _fname)) 
				     then 
				       ((C_SD.get_target op), atomic_reg.Region.stack)
				     else   
				       begin
					 if (_fname <> __BLAST_DummyFunctionName) then
					   ((C_SD.lookup_entry_location _fname),(Region.CallStack ((C_SD.get_source op):: callStack)))  
					 else 
					   begin
                                             let _wrap_getvars _pred = List.map (function x -> Expression.Symbol x) (getVarExps_deep _pred) in 
                                             let vars_in_support = List.flatten (List.map _wrap_getvars (preds_of_region reg)) in   
                                             if (not (List.mem target vars_in_support)) then 
                                               ((C_SD.get_target op),atomic_reg.Region.stack) (* is a skip *)
                                             else failwith ("Undeclared important function"^_fname)
				           end
			                end
                                  | _ -> failwith "what the (post) !"
		             end
                        else 
	
			   try 
			     if (isReturn op) then (Misc.thd3 (lastTarget ()),(Region.CallStack (List.tl callStack)) ) 
			     else (C_SD.get_target op,atomic_reg.Region.stack) 
			   with Failure ("tl") -> (C_SD.get_target op, atomic_reg.Region.stack ) 
	
       
                     and (new_data,p_index) =
                       try 
			 match atomic_reg.Region.data with
			     Region.Abstract abst_data_reg ->
                               ((abstract_data_post_proof abst_data_reg (C_SD.get_command op)) op phi_table)
			   | Region.Concrete conc_data_reg ->
			       begin
				 Message.msg_string Message.Error "ERROR!: post_proof given concrete region"; failwith "post_proof: arg concrete region"
			       end
                       with C_SD.NoSuchFunctionException f -> failwith ("post_proof: Unexpected NoSuchFunctionException "^f)
                     in
                      ( Atomic_pr { prog_counter = new_location ;
                                   call_stack = new_stack ; 
				   data_reg_cubes = new_data;
				 }, p_index)
		   with
		       Failure("Empty Call Stack") -> (Empty_pr,skip_lemma_index)
		 end
               with 
		   C_SD.NoSuchFunctionException f -> (* treat it as a skip *)
		     failwith ("post_proof: Unexpected NoSuchFunctionException "^f)
		     (*Message.msg_string Message.Minor ("Unexpected NoSuchFunctionException "^f);*)
		     (* THIS SHOULD BE COMMENTED OUT IF UNDEFINED FUNCTIONS ARE TO BE AVOIDED *)
		     (* failwith ("Unexpected NoSuchFunctionException "^f); *)
		     (*
		       Atomic_pr { prog_counter = C_SD.get_target op old_location;
				   call_stack = atomic_reg.Region.stack old_stack;
				   data_reg_cubes = old_data;
				 },  skip_lemma_index*)
		       
       	     end
	 | Region.Union _ ->
             failwith "post_proof: arg is a union !"   
	 | _ -> failwith "post_proof: parallel regions not supported"   

let print_stats_proof fmt () =
  Message.msg_string Message.Normal ("Proof Statistics:");
   Message.msg_string Message.Normal ("Total number of queries = "^(string_of_int (!statTotalNumberOfQueries))^"\n");
   Message.msg_string Message.Normal ("Total cached = "^(string_of_int (!statTotalCached))^"\n\n");     
   ()

(* ********** END BlastProof ************ *)



(************** BEGIN XML dumping of trace **************************)
(* the following code takes a trace and dumps it in XML format readable
   by SLAM's viewer.
*)



let dumpXml op_list =
  let filecallHashtbl = Hashtbl.create 37 in (* This hashtable maintains the number of
                                                times this function has been called in this
                                                trace, to generate appropriate names
                                                for the xml files *)
  let startFn = Options.getValueOfString "main" in
  let getIndexFromHashtbl fnam =
    try 
      let currindex = Hashtbl.find filecallHashtbl fnam
      in
      Hashtbl.replace filecallHashtbl fnam (currindex + 1);
      currindex+1
    with
      Not_found -> (Hashtbl.add filecallHashtbl fnam 1; 1) 
  in
  Hashtbl.add filecallHashtbl (startFn) 1;
  let get_procname loc = "main" in (* Fix this: how to get proc name from location? *)
  let fpStack = Stack.create () in
  let write fp str = output_string fp str in
  let rec dumpOp i fp start op_list =
    
    if (start) then 
		    (* Preamble *)
      ignore (write fp "<?xml version='1.0'?>\n";
	      write fp "<path:path name=\"";
	      write fp (string_of_int i);
	      write fp "\" xmlns:path = \"x-schema:pathschema.xml\" xmlns:env=\"x-schema:envschema.xml\" >\n\n")
    else ();
    match op_list with 
      [] ->  
	write fp "</path:path>\n";
	close_out fp; 0
    | op::rest ->
	begin
	  let target_loc = C_SD.get_source op
	  and c_command = C_SD.get_command op in
	  let sloc = C_SD.get_source_position target_loc in
	  match sloc with
	    None -> 
	      begin
		let ssloc =  C_SD.get_source_position (C_SD.get_target op) in
		match ssloc with 
		  None -> failwith "Strange edge with artificial nodes on both ends!"
		|  Some (fn, ln, _) ->
		    begin
		      write fp "<path:stmt index=\"";
		      write fp (string_of_int i);
		      write fp "\" file=\"";
		      write fp (fn);
		      write fp "\" proc=\"";
		      write fp (get_procname target_loc);
		      write fp "\" line=\"";
		      write fp (string_of_int (ln));
		      write fp "\" notentry=\"true\">\n";
		      write fp "<env:state xmlns:env=\"x-schema:envschema.xml\">\n";
		      write fp "</env:state>\n";
		      match c_command.C_SD.Command.code with
			C_SD.Command.FunctionCall ex ->
			  begin
			    let (fnam,_,_) = deconstructFunCall ex in
			   if (fnam = __NotImplementedFunctionName) then  
                             begin (* function pointer call. Treat as default code *)
                              write fp "<path:default code=\"--\"></path:default>\n";
                              write fp "</path:stmt>/n";
			      let pos = (dumpOp (i+1) fp true rest) in
			      pos + 1
                             end
                           else
			      begin
                              (* write this function call, then recursively call the dump routine for the callee *)
			      Stack.push fp fpStack;
		    		     (* open file for new function *)
			      let index = getIndexFromHashtbl fnam in
			      let srcname = ("BLAST."^fnam^"."^(string_of_int index)^".xml") in
			      let newfp = open_out srcname in
			      write fp "<path:call caller=\"";
                              write fp (get_procname target_loc);
                              write fp "\" callee=\"";
			      write fp fnam;
                              write fp "\" code=\"--\" src=\"";
                              write fp srcname; 
			      write fp "\"/>\n";
			      write fp "</path:stmt>\n";
			      let pos = (dumpOp (i+1) newfp true rest) in
			      pos + 1
                              end
                          end
		      | C_SD.Command.Block l ->
			  begin
			    write fp "<path:default code=\"--\"></path:default>\n";
			    write fp "</path:stmt>\n\n";
			    match (List.hd (List.rev l)) with
			      C_SD.Command.Return _ -> (* return *)
				write fp "</path:path>\n";
				flush fp; close_out fp;
				(try
				  dumpOp (i+1) (Stack.pop fpStack) false rest 
				with Stack.Empty -> i)

			    |	_ -> (* recursively continue *)
				let pos = dumpOp (i+1) fp false rest
				in
				pos + 1
			  end
		      |  _  -> 
			  begin
			    write fp "<path:default code=\"--\"></path:default>\n";
			    write fp "</path:stmt>\n\n";
			    let pos = dumpOp (i+1) fp false rest 
			    in
			    pos + 1
			  end
		    end
	      end
	  | Some (fn, ln, _) ->
	      begin
		write fp "<path:stmt index=\"";
		write fp (string_of_int i);
		write fp "\" file=\"";
		write fp (fn);
		write fp "\" proc=\"";
		write fp (get_procname target_loc);
		write fp "\" line=\"";
		write fp (string_of_int (ln));
		write fp "\" notentry=\"true\">\n";
		write fp "<env:state xmlns:env=\"x-schema:envschema.xml\">\n";
		write fp "</env:state>\n";
		
		match c_command.C_SD.Command.code with
		  C_SD.Command.FunctionCall ex -> 
			  begin
			    let (fnam,_,_) = deconstructFunCall ex in
			    if (fnam = __NotImplementedFunctionName) then  
                              begin
                              (* function pointer call. Treat as default code *)
			      
		              write fp "<path:default code=\"--\"></path:default>\n";
                              write fp "</path:stmt>/n";
			      let pos = (dumpOp (i+1) fp true rest) in
			      pos + 1
                              end
                            else
			      begin
                              (* write this function call, then recursively call the dump routine for the callee *)
			      Stack.push fp fpStack;
		    		     (* open file for new function *)
			      let index = getIndexFromHashtbl fnam in
			      let srcname = ("BLAST."^fnam^"."^(string_of_int index)^".xml") in
			      let newfp = open_out srcname in
			      write fp "<path:call caller=\"";
                              write fp (get_procname target_loc);
                              write fp "\" callee=\"";
			      write fp fnam;
                              write fp "\" code=\"--\" src=\"";
                              write fp srcname; 
			      write fp "\"/>\n";
			      write fp "</path:stmt>\n";
			      let pos = (dumpOp (i+1) newfp true rest) in
			      pos + 1
                              end
                          end
		| C_SD.Command.Block l ->
		    begin
		      write fp "<path:default code=\"--\"></path:default>\n";
		      write fp "</path:stmt>\n\n";
		      match (List.hd (List.rev l)) with
			C_SD.Command.Return _ -> (* return *)
			  write fp "</path:path>\n";
			  (flush fp; close_out fp; 
			  try 
			    dumpOp (i+1) (Stack.pop fpStack) false rest 
			  with Stack.Empty ->
			    i)
		      |	_ -> (* recursively continue *)
			  let pos = dumpOp (i+1) fp false rest
			  in
			  pos + 1
		    end
		| _ -> 
		    begin
		      write fp "<path:default code=\"--\"></path:default>\n";
		      write fp "</path:stmt>\n\n";
		      let pos = dumpOp (i+1) fp false rest 
		      in
		      pos + 1
		    end
	      end
	end
  in

  let i = 0 in
  let fp = open_out ("BLAST."^startFn^".xml") in
  ignore (dumpOp 0 fp true op_list);
  try
   while (true) do
     let fp = Stack.pop fpStack in
     write fp "</path:path>\n";
     close_out fp
   done
  with Stack.Empty -> ()


  type spec = unit
  type spec_result =
    | Yes
    | No
    | Maybe of Ast.Predicate.predicate
  let summary_post _ = failwith "summary_post not implemented"
  let init_spec = summary_post
  let build_spec = summary_post
  let check_spec = summary_post
end
