#!/usr/bin/perl
# annot.pl
# by jeremy condit
#
# post-processing pass on gcc output to produce annotations expected
# by pcc scripts.
#
# deficiencies:
# - we assume that all arguments consume one word on the stack,
#   which is just not so.
# - the mechanism for determining stack space required is very
#   fragile; we just remember the last time we subtracted a
#   constant from the stack pointer.
# - very few sanity checks.

use strict;

my $args = 0;
my $stack = 0;

&processFile;

###
### functions
###

# processFile
#
# the main loop echoes lines of assembly until it sees the beginning
# of an annotation.

sub processFile
{
    print <<EOF;
ANN_STARTLF
IMPORT mach.ans ccured.ans
ANN_ENDLF
EOF

    while (my $line = <STDIN>)
    {
        if ($line =~ /^\s*ann begin function (.*)/)
        {
            &processFunction($1);
        }
        elsif ($line =~ /^\s*ann begin inv (.*)/)
        {
            &processInv($1);
        }
        elsif ($line =~ /^\s*ann dom (.*)/)
        {
            &processDom($1);
        }
        elsif ($line =~ /^\s*subl \$(\d*),\%esp/)
        {
            $stack = $1 / 4;
            print $line;
        }
        else
        {
            print $line;
        }
    }
}

# processFunction
#
# process a function annotation, emitting the annotation in the
# expected format.

sub processFunction
{
    my ($name) = @_;

    my @vars = &processVars;
    $args = @vars;
    $stack += $args + 2;

    my $typestr = &makeFullTypeString("\n                     ", \@vars);

    print <<EOF;
    ANN_FUNCTION($name,
                 \%LF_$typestr\%_LF,
                 \%LF_true\%_LF,
                 RB(),
                 $args,
                 $stack)
EOF

    $stack = 0;
}

# processInv
#
# process an invariant for the current program point, emitting it
# in the expected format.

sub processInv
{
    my ($dom) = @_;
    
    my @vars = &processVars;

    my $typestr = &makeFullTypeString("\n                ", \@vars);

    $dom = ($dom eq "loop") ? "ANN_DOM_LOOP" : ("L" . $dom);

    print <<EOF;
    ANN_INV($dom,
            \%LF_$typestr\%_LF,
            RB())
EOF
}

# processDom
#
# process a dominator annotation, printing the annotation and an
# associated label.

sub processDom
{
    my ($dom) = @_;

    print <<EOF;
L$dom:
    ANN_DOMINATOR
EOF
}

# processVars
#
# collect a list of variable annotations, producing a list of hashes,
# one for each variable, where each hash contains "name", "loc", and
# "type", exactly as they appear in the annotation.
#
# used by both processFunction and processInv.

sub processVars
{
    my $done = 0;
    my @vars = ();

    while ((!$done) && (my $line = <STDIN>))
    {
        if ($line =~ /^\s*ann end/)
        {
            $done = 1;
        }
        elsif ($line =~ /^\s*ann var ([^,]*),([^,]*),(.*)/)
        {
            push @vars, {name => $1, loc => $2, type => $3};
        }
    }

    return @vars;
}

# translateLoc
#
# translate a raw gcc location (e.g. %eax, -12(%ebp)) to one in the
# format used by proof-carrying code (e.g. eax, loc5).

sub translateLoc
{
    my ($loc) = @_;
    my $result;

    if ($loc =~ /^\%([a-z]+)$/)
    {
        $result = $1;
    }
    elsif ($loc =~ /^(-?[0-9]*)\(\%ebp\)$/)
    {
        $result = "loc" . ($args + 2 - ($1 / 4));
    }
    
    return $result;
}

# makeTypeString
#
# given a hash from variable names to variable locations, a hash
# describing the variable itself, and a prefix string, produce a
# type predicate such as (of foo int).

sub makeTypeString
{
    my ($varhash, $var, $prefix) = @_;
    my $loc = &translateLoc($$var{loc});
    my $type = $$var{type};

    $type =~ s/@(\w+)/$$varhash{$1}/g;

    return "$prefix(of $loc $type)";
}

# makeFullTypeString
#
# produce a type string that is the conjunction of the type predicates
# for all of the indicated variables.  takes a whitespace string to be
# prepended to each line, plus a reference to an array of hashes
# corresponding to the variables.

sub makeFullTypeString
{
    my ($whitespace, $varsref) = @_;
    my @vars = @$varsref;
    my $result;

    if (@vars == 0)
    {
        $result = "true";
    }
    else
    {
        my %varhash;
        foreach my $var (@vars)
        {
            $varhash{$$var{name}} = &translateLoc($$var{loc});
        }

        my $last = &makeTypeString(\%varhash, pop @vars, "    ");
        my @lf = map &makeTypeString(\%varhash, $_, "(/\\ "), @vars;

        for (my $i = 0; $i < @lf; $i++)
        {
            $last .= ")";
        }

        push @lf, $last;
        $result = join $whitespace, @lf;
    }

    return $result;
}


