#!/usr/bin/perl
#
# The main driver for the CCured system
#
use strict;
use FindBin;
use Data::Dumper;

use lib "$FindBin::Bin"; # The libraries are in the same dirctory


my $curator = CCured->new(@ARGV);

if(! defined($curator->{CURETYPE})) {
    $curator->{CURETYPE} = 'none';
}
# Check the curetype
my @cures = ('none', 'wild', 'infer', 'paper');
if(! grep { $_ eq $curator->{CURETYPE} } @cures) {
    die "Don't know this cure ($curator->{CURETYPE}). Try one of: " .
        join(',', @cures) . ".\n";
}
   
if(! defined($curator->{CCUREDARGS})) {
    $curator->{CCUREDARGS} = [];
}
if($curator->{VERBOSE}) {
    push @{$curator->{CCUREDARGS}}, "--verbose";
}
$curator->{MERGE_INOBJ} = 1;

$curator->setVersion();

# print Dumper($curator);

$curator->doit();

exit(0);

#
# Now we customize a merger
#
package CCured;
use strict;

use FindBin;
use Data::Dumper;
use File::Basename;

use lib "$FindBin::Bin"; # The libraries are in the same dirctory

use Merger;

BEGIN {
    @CCured::ISA = qw(Merger);
    $CCured::dir = "$FindBin::Bin/..";
}


# We need to customize the collection of arguments
sub collectOneArgument {
    my($self, $arg, $pargs) = @_;
    if($arg eq "--releaselib") {
        $self->{RELEASELIB} = 1; return 1;
    }
    if($arg =~ m|--curetype=(.+)$|)  {
        $self->{CURETYPE} = $1; return 1;
    }
    if($arg =~ m|--usecil|) {
        $self->{USECIL} = 1; return 1;
    }
    if($arg =~ m|--emitinfer|) {
        $self->{EMITINFER} = 1; return 1;
    }
    if($arg =~ m|--optimize|) {
        $self->{OPTIMIZE} = 1; return 1;
    }
    if($arg eq '--nogc') {
        $self->{NOGC} = 1; 
        return 1;
    }
    if($arg eq '--optimonly') {
        $self->{OPTIMONLY} = 1;
        push @{$self->{CILARGS}}, $arg;
        return 1;
    }
    # See if the super class understands this
    return $self->SUPER::collectOneArgument($arg, $pargs);
}

# Customize the help message
sub usage {
    print "ccured [options] [gcc_or_mscl options]\n";
}

sub helpMessage {
    my($self) = @_;
    # Print first the original
    $self->SUPER::helpMessage();
    print <<EOF;

CCured specific options:

  --curetype=type    Selects the type of cure
                       none     No cure at all. Just compile
                       wild     All pointers are made wild
                       infer    Inference is used

  --usecil           Emit and use the CIL
  --emitinfer        Emit the INFER
  --optimize         Run the optimizer
  --stages           Show the processing stages
  --optimonly        Assumes that the cured file already exists and just
                     optimizes it
  --releaselib Link with the release version of the CCured runtime library


 The --leavealone files are not cured.

EOF
}

sub ccuredExecutable {
    my($self) = @_;
    my $base = "$CCured::dir/obj/ccured";
    if(!$self->{NATIVECAML}) {
        if($^O eq 'MSWin32') {
            # For some strange reason we have to start the bytecode like this
            return $base . ".byte.exe $base" . ".byte.exe";
        } else {
            return $base . ".byte.exe ";
        }
    } else {
        return $base . ".asm.exe";
    }
}

sub preprocess_beforecil {
    my($self, $src, $dest, $ppargs) = @_;
    my @args = @{$ppargs};
    if($self->{CURETYPE} ne 'none') {
        push @args, $self->{DEFARG} . "CCURED";
        # Force include fixup.h
        unshift @args, $self->forceIncludeArg("$CCured::dir/lib/fixup.h");
    }
    return $self->SUPER::preprocess_beforecil($src, $dest, \@args);
}

# Linking after CIL
sub link_aftercil {
    my ($self, $psrcs, $dest, $ppargs, $ccargs, $ldargs) = @_;

    # Regular linking. Add the library
    my @srcs = @{$psrcs};
    if($self->{CURETYPE} ne 'none') {
        push @srcs,
        "$CCured::dir/obj/ccured_" . $self->{MODENAME} .
            ($self->{RELEASELIB} ? "_releaselib" : "_debug") .
                ".$self->{LIBEXT}";        
    }
    return $self->SUPER::link_aftercil(\@srcs, 
                                       $dest, $ppargs, $ccargs, $ldargs);
}


#################### THE CURE 
sub applyCil {
    my ($self, $src, $ppargs) = @_;

    if($self->{VERBOSE}) { print "Curing $src\n"; }
    my ($base, $dir, $ext) = fileparse($src, "\\.[^.]+");

    my $src = "$dir$base$ext";

    my $ccured = $self->ccuredExecutable(); 
    my $cmd = "$ccured ";
    if($self->{MODENAME} eq "MSVC") {
        $cmd .= " --MSVC ";
    }
    $cmd .= (join(' ', @{$self->{CILARGS}}) . " ");
    my $result;
    my $resultExt;
    if(defined $self->{CURETYPE} && $self->{CURETYPE} ne 'none') {  
        $cmd .= " --curetype $self->{CURETYPE} ";
        if($self->{EMITINFER}) {
            $cmd .= (" --inferout $dir$base" . "infer.c ");
        }
        $cmd .= (" --curedout $dir$base" . "cured.c ");
        $resultExt = "cured.c";
        if($self->{OPTIMIZE}) {
            $cmd .= (" --optimout $dir$base" . "cured.optim.c ");
            $resultExt = "cured.optim.c";
        }
    } else {
        if($self->{USECIL} || !defined($resultExt)) {
            $resultExt = "cil.c";
            $cmd .= " --cilout $dir$base$resultExt";
        }
    }
    $result = "$dir$base$resultExt";
    $cmd .= " $src";
    $self->runShell($cmd);

    return $result;
}

sub preprocess_aftercil {
    my ($self, $src, $dest, $ppargs) = @_;
    my @args = @{$ppargs};
    push @args, ($self->{DEFARG} . 'CCURED',
                 $self->{NOGC} ? $self->{DEFARG} . 'CCURED_NO_GC' : '',
                 $self->{INCARG} . $CCured::dir . "/lib");
    return $self->SUPER::preprocess_aftercil($src, $dest, \@args);
}

1;
