/* 
 *
 * Copyright (c) 2001 by
 *  George C. Necula	necula@cs.berkeley.edu
 *  Scott McPeak        smcpeak@cs.berkeley.edu
 *  Wes Weimer          weimer@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products 
 * derived from  this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// ccuredcheck.h
// contains lots of functions and macros to do the basic CCured
// runtime checks and bookkeeping; calls to the macros here are
// inserted by the CCured translator into the instrumented C program

                       
#ifdef _MSVC
  #ifdef _GNUCC
    #error Exactly one of _MSCV and _GNUCC must be defined.
  #else
    #define FILEFUNC __FILE__
    #define LINEFUNC __LINE__
    #define CHECK_STATIC
     /* Must restrict packing on 32-bit boundary because otherwise the
      * length and tag fields are not adjacent to the data */
    #pragma pack(4)
    #define NORETURN __declspec(noreturn)
    #define NAKED    __declspec(naked)
  #endif
#else
  #ifdef _GNUCC
    #define FILEFUNC (__FILE__ ":" __FUNCTION__)
    #define LINEFUNC __LINE__
    #undef  _GNUCC
    #define _GNUCC 1
    #define CHECK_STATIC static
    #define NORETURN __attribute__((noreturn))
    #define NAKED
  #else
    #error Exactly one of _MSVC or _GNUCC must be defined.
  #endif
#endif

// a lot of CCrured's implementation is dependent on the assumption
// that pointers are represented in 32 bits..
#define I32 int
#define U32 unsigned I32

#ifndef MIN
  #define MIN(x,y) (((x) < (y)) ? (x) : (y))
#endif

// libc prototypes (we avoid pulling in entire headers; sm: I no longer
// remember why)
extern unsigned int strlen(const char *);

// each code is associated with a string that gets printed when
// we abort execution due to runtime check failure
#define FAIL_BELOW_STACK_ADDRESS 0
#define FAIL_STACK_ADDRESS       1
#define FAIL_LBOUND              2
#define FAIL_UBOUND              3
#define FAIL_FUNCTIONPOINTER     4
#define FAIL_DECR_FSEQ           5
#define FAIL_NULL                6
#define FAIL_NULL_STRING         7
#define FAIL_STRINGBOUND         8
#define FAIL_UNALIGNED           9
#define FAIL_MALLOC              10
#define FAIL_REALLOC             11
#define FAIL_FREE                12
#define FAIL_REALLOC_SHRINK      13
#define FAIL_REALLOC_GROW        14
#define FAIL_NONPOINTER          15
#define FAIL_SCALARLOG           16
#define FAIL_NULLSTRING          17
#define FAIL_VA_BADARG           18
#define FAIL_VA_NOSTART          19
#define FAIL_VA_MANY             20
#define FAIL_VA_BADTYPE          21
#define FAIL_UNRECOGNIZEDFORMAT  22
#define FAIL_OVERFLOW            23
#define NUM_FAIL_CODES           24   /* sm: keep this updated */

// sm: two modes of operation: with and w/o file/line info;
// specify FP_FAIL_IS_VERBOSE=1 in Makefile or .ccuredrc to get
// the extra info in any mode; the default is to omit this
// information if RELEASELIB is specified (run test/verify_fp_verbose
// to see if this works)
#if !defined(RELEASELIB) && !defined(FP_FAIL_IS_VERBOSE)
  #define FP_FAIL_IS_VERBOSE
#endif

#if defined(FP_FAIL_IS_VERBOSE)
  #define FP_FAIL_EXTRA_PARAMS , char *file, int line
  #define FP_FAIL_EXTRA_ARGS , file, line
  #define FILE_AND_LINE , FILEFUNC, LINEFUNC
  #define FP_FAIL_STR(s) fp_fail_str(s, __FILE__, __LINE__)
  #define FP_FAIL(code) fp_fail(code, __FILE__, __LINE__)
#else
  #define FP_FAIL_EXTRA_PARAMS
  #define FP_FAIL_EXTRA_ARGS
  #define FILE_AND_LINE
  #define FP_FAIL_STR(s) fp_fail_str(s)
  #define FP_FAIL(code) fp_fail(code)
#endif

// some shorter synonmys
//#define FPFP FP_FAIL_EXTRA_PARAMS
//#define FPFA FP_FAIL_EXTRA_ARGS
//#define FAL FILE_AND_LINE

// declaration of our failure functions
// fail with a given error message
NORETURN  void fp_fail_str(char *str  FP_FAIL_EXTRA_PARAMS);

// fail with a given code, which maps to a message
NORETURN  void fp_fail(int msgId  FP_FAIL_EXTRA_PARAMS);

// the user just tried to dereference a non-pointer (null 'b' field);
// 'l' is the 'p' field of that pointer, so we search our log of casts
// (if present) to see where this (non-)pointer arose; either way, the
// program aborts with a message about dereferencing a non-pointer
NORETURN void searchLog(unsigned long l  FP_FAIL_EXTRA_PARAMS);

// Special forms of error functions
#ifdef _MSVC
  #define searchLog_zero(b, p) \
   __asm { mov  ebx, p };\
   __asm { mov  eax, b };\
   __asm { test eax, eax };\
   __asm { je  searchLog };
#else
  #define searchLog_zero(b, p) if(!(b)) searchLog(p);
#endif


// ---------------- check function return value --------------
// CHECK_GETFRAME: return the frame pointer of the current function;
// we assume (for stacks growing towards 0) that all local variables
// (those on the stack, anyway) are at addresses less-or-equal this value
#if defined(_MSVC)
  #define CHECK_GETFRAME(putwhere) { \
     char *__bpVal; \
     _asm { mov __bpVal, ebp }; \
     (putwhere) = __bpVal; }
#elif defined(_GNUCC)
  // sm: gcc doesn't inline functions unless the optimizer is turned
  // on, but I really want to be able to test our system without the
  // optimizer; so I'm turning this into a macro, so we get the frame
  // of the *caller*, not the callee
  #define CHECK_GETFRAME(putwhere) (putwhere) = __builtin_frame_address(0);
#endif

// given a value being returned from a function call, verify it is
// not a pointer to this function's locals (so, transitively, we
// should never be able to return a pointer to a deallocated frame);
// The pointer passed to this function should be the _b or _e component
// of a pointer, not the _p component because that one can be
// adjusted through arithmetic. In each case pass the component
// that is NULL if the pointer is an integer (_e for FSEQ and
// _b for all others). 
//#define CHECK_RETURNPTR(p)  __CHECK_RETURNPTR(p  FILE_AND_LINE)

// sm: as for CHECK_GETFRAME itself, I'm changing this to a macro
// to ensure inline expansion even when optimization is turned off

/* gn. If gcc inlines the function containing the CHECK_RETURNPTR then 
 * CHECK_GETFRAME will obtain the frame address of the wrong function. To 
 * prevent this I have changed box.ml to add a special volatile temporary 
 * first in the argument list and use its address as the frame address */        
#ifdef _MSVC
  #define CHECK_RETURNPTR(p, pFirstLocal)                        \
  {                                                              \
    char* bpVal;                                                 \
    /*CHECK_GETFRAME(bpVal);*/                                       \
    bpVal = (char*)pFirstLocal;                                  \
    /* _MSVC seems to put the stack at lowest addresses */       \
    if (p && (char*)p <= bpVal) {                                \
      fp_fail(FAIL_BELOW_STACK_ADDRESS  FILE_AND_LINE);          \
    }                                                            \
  }
#else
  #define CHECK_RETURNPTR(p, pFirstLocal)                                   \
  {                                                                         \
    char* bpVal;                                                            \
    /*CHECK_GETFRAME(bpVal);*/                                              \
    bpVal = (char*)pFirstLocal;                                             \
                                                                            \
    /* Others (gcc) put the stack up. Check only the first 2M to avoid */   \
    /* rejecting pointers that are really into the heap                */   \
    if (p && ! ((bpVal - (char*)p) >> 21)) {                                \
      fp_fail(FAIL_BELOW_STACK_ADDRESS  FILE_AND_LINE);                     \
    }                                                                       \
  }
#endif


// ----------------- writing pointers through pointers or to globals --------
// given an address the user is writing through a pointer, verify
// it is not a pointer into the stack; the 'where' argument is the
// address of the pointer whose value will *change*, i.e. this is
// checking something like ``*where = p''
#define CHECK_STOREPTR(where, p)  __CHECK_STOREPTR(where, p  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_STOREPTR(void **where, void *p  FP_FAIL_EXTRA_PARAMS)
{
  #ifdef _GNUCC
  // the following check assumes the stack grows towards smaller
  // addresses; that's an ABI feature not a compiler issue, but
  // we're pretending it is anyway
    if ((unsigned)p - (unsigned)where < 0x100000U) {
      // sm: if the check below fails, it means that 'p'
      // is a stack address; however, since 'where' is only a little
      // (<1M) smaller than 'p', it too must be a stack address;
      // moreover, it's in a warmer frame, so it will go away before
      // 'p' will, so this should be ok
      //
      // I've implemented this extra check as a way to try to allow
      // EDG's exception-handling code to work
      //
      // update: now that this is in CHECK_STOREPTR, it applies to
      // safe/seq as well as wild; too bad I don't have a testcase
      // for the former...
      return;
    }
  #endif // _GNUCC

  if (p) {
    char* bpVal;
    int   delta;
    CHECK_GETFRAME(bpVal)
    delta = (bpVal - (char*)p);

    #ifndef DISABLE_HEAPIFY
      // the usual code
      delta >>= 20;  // Look +- 1MB from the frame (UNSOUND). We need a better way to do this
      if(delta == 0 || delta == -1) {  // Catch 0 or 1
        fp_fail(FAIL_STACK_ADDRESS  FP_FAIL_EXTRA_ARGS);
      }
    #else
      // some code which should have the same performance effect
      // as the above code where the 'if' condition is always false
      delta <<= 4;
      if (delta == 1 || delta == -1) {    // never true (does the compiler know this?)
        fp_fail(FAIL_STACK_ADDRESS  FP_FAIL_EXTRA_ARGS);
      }
    #endif
  }
}


// ---------------- queries on b,p fat pointers -----------------
// given a b,p fat pointer that's going to be dereferenced, retrieve
// the length field; the length is the number of 32-bit words the
// program is allowed to read/write
#define CHECK_FETCHLENGTH(p,b) __CHECK_FETCHLENGTH(p, b  FILE_AND_LINE)
__inline static unsigned int __CHECK_FETCHLENGTH(void *p, void *b  FP_FAIL_EXTRA_PARAMS)
{
  // if the 'b' field is null, we fail
  if (!b) {
    searchLog((unsigned long)p  FP_FAIL_EXTRA_ARGS);
  }

  // length is stored just before the first word of user-accessible area
  return (unsigned int)(* (((U32*)b) - 1));
}

// given a b,p fat pointer, return a pointer to the word just beyond the
// last one the user is allowed to read/write
#define CHECK_FETCHEND(p,b) __CHECK_FETCHEND(p, b  FILE_AND_LINE)
__inline static void * __CHECK_FETCHEND(void *p, void *b  FP_FAIL_EXTRA_PARAMS)
{
  unsigned int len = __CHECK_FETCHLENGTH(p, b  FP_FAIL_EXTRA_ARGS);
  return (void*)((char*)b + (len << 2));
}



// This is the type of a function descriptor
struct __functionDescriptor {
  unsigned int _len ; // Always 0
  void (* _pfun)() ; // Pointer to a function
  unsigned int _nrargs ; // The number of arguments
};

// And a macro to initialize the function descriptor
#define DEFINE_CCURED_FUNCTIONDESCRIPTOR(f, nrargs) \
 struct __functionDescriptor f ## __descriptor={ 0, (void (*)())(f), (nrargs)};

// ---------------- usage checks for b,p fat pointers ---------------
// given a b,p fat pointer to a function, and 'n' the number of arguments
// actually passed, verify that b,p is actually a function pointer, and that
// 'n' is enough arguments
#define CHECK_FUNCTIONPOINTER(p,b,n) __CHECK_FUNCTIONPOINTER(p, b, n  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_FUNCTIONPOINTER(void *p, void *b, int nrActualArgs  FP_FAIL_EXTRA_PARAMS)
{
  // non-pointer check
  if (!b)
    searchLog((unsigned long)p  FP_FAIL_EXTRA_ARGS);

  // the check below allows 'nrActualArgs' to exceed the number of args
  // the function accepts, because the usual calling convention will
  // harmlessly ignore extra arguments (and our type system has no notion
  // of obligated use)
  if( ((U32*)b)[-1] != 0         ||      // length field is 0 for fn pointers
      ((U32*)b)[ 0] != (U32)p    ||      // first word is actual code pointer
      ((U32*)b)[+1] > nrActualArgs ) {   // second word is # args fn requires
    fp_fail(FAIL_FUNCTIONPOINTER  FP_FAIL_EXTRA_ARGS);
  }
}


// make this forward-declared to discourage gcc from inlining it
NORETURN static void ubound_fail(void *bend  FP_FAIL_EXTRA_PARAMS);

// Check the upper bound:
//   bend - pointer to the first byte the user is *not* allowed to read/write
//   p    - pointer the user is using to read/write
//   plen - length in bytes of the user's data being read/written
#define CHECK_UBOUND(bend,p,plen) __CHECK_UBOUND(bend,p,plen  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_UBOUND(void *bend, void *p, unsigned int plen  FP_FAIL_EXTRA_PARAMS)
{
  // sm: Escalating the +1 war again: under the interpretation specified
  // in George's 11/10/2001 email, and consistent with the way box.ml
  // generates code, any access to the 'bend' byte is illegal.  I know this
  // will cause problems, so I want to find where they are and fix them
  // there.

  // the user has full control over both 'p' and 'plen'; we can't get
  // away with fewer than two tests; test for overflow first -- we
  // might at least get a common subexpression from 'p+plen'
  if ((unsigned)p + plen < (unsigned)plen) {
    fp_fail(FAIL_OVERFLOW  FP_FAIL_EXTRA_ARGS);
  }

  // if we're writing just to the legal end, then p+plen will equal 'bend';
  if ((unsigned)p + plen > (unsigned)bend) {
    // failure
    ubound_fail(bend  FP_FAIL_EXTRA_ARGS);
  }
  else {
    // no problem
  }
}

// out-of-line ubound failure analysis; we don't want this in the
// main check because that gets inlined everywhere
NORETURN static void ubound_fail(void *bend  FP_FAIL_EXTRA_PARAMS)
{
  if (bend == 0) {
    // sm: George wants this to be classified as non-pointer (which is
    // a better description), so we'll special-case it
    fp_fail(FAIL_NONPOINTER  FP_FAIL_EXTRA_ARGS);
  }
  else {
    // valid 'end' field means it's an upper-bounds failure
    fp_fail(FAIL_UBOUND  FP_FAIL_EXTRA_ARGS);
  }
}


// like above, but we allow 'p' to be null (in which case plen is ignored)
#define CHECK_UBOUNDNULL(bend,p,plen) __CHECK_UBOUNDNULL(bend,p,plen  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_UBOUNDNULL(void *bend, void *p,
                        unsigned int plen  FP_FAIL_EXTRA_PARAMS)
{
  if (!p) {
    // ok
  }
  else {
    __CHECK_UBOUND(bend, p, plen  FP_FAIL_EXTRA_ARGS);
  }
}

// Check the lower bound
//   base - pointer to the first byte the user is allowed to read/write
//   p    - pointer the user is using to read/write
#define CHECK_LBOUND(bend,p) __CHECK_LBOUND(bend,p  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_LBOUND(void *base, void *p  FP_FAIL_EXTRA_PARAMS)
{
  if (p < base) {
    fp_fail(FAIL_LBOUND  FP_FAIL_EXTRA_ARGS);
  } else {
    // ok
  }
}

// check both upper and lower bounds; interpretations of params are
// as in the respective check functions above:
//   bend - pointer to the first byte the user is *not* allowed to read/write
//   base - pointer to the first byte the user is allowed to read/write
//   p    - pointer the user is using to read/write
//   plen - length in bytes of the user's data being read/written
#define CHECK_BOUNDS(base,bend,p,plen) __CHECK_BOUNDS(base,bend,p,plen  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_BOUNDS(void *base, void *bend,
                    void *p, unsigned int plen  FP_FAIL_EXTRA_PARAMS) {
  __CHECK_LBOUND(base, p  FP_FAIL_EXTRA_ARGS);
  __CHECK_UBOUND(bend, p, plen  FP_FAIL_EXTRA_ARGS);
}

// same as above, but it's ok if p is null
#define CHECK_BOUNDSNULL(base,bend,p,plen) __CHECK_BOUNDSNULL(base,bend,p,plen  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_BOUNDSNULL(void *base, void *bend,
                        void *p, unsigned int plen  FP_FAIL_EXTRA_PARAMS) {
  // sm: a (my?) comment here used to say: "lbound already succeeds on null";
  // that is not true
  if (p) {
    __CHECK_LBOUND(base, p  FP_FAIL_EXTRA_ARGS);
    __CHECK_UBOUND(bend, p, plen  FP_FAIL_EXTRA_ARGS);
  }
}

// like CHECK_BOUNDS except we're given the length of the accessible
// area in words, instead of a pointer to just beyond that area
#define CHECK_BOUNDS_LEN(base,bw,p,plen) __CHECK_BOUNDS_LEN(base,bw,p,plen  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_BOUNDS_LEN(void *base, unsigned int bwords,
                        void *p, unsigned int plen  FP_FAIL_EXTRA_PARAMS)
{
  __CHECK_BOUNDS(base, (char*)base + (bwords << 2),      // compute end
                 p, plen  FP_FAIL_EXTRA_ARGS);
}

// like CHECK_BOUNDS except we're given a 'base' field that is known to
// have the length stored at base[-1]
#define CHECK_BOUNDS_INDEX(base,p,plen) __CHECK_BOUNDS_INDEX(base,p,plen  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_BOUNDS_INDEX(void *base, void *p, unsigned int plen  FP_FAIL_EXTRA_PARAMS)
{
  __CHECK_BOUNDS(base, __CHECK_FETCHEND(p, base  FP_FAIL_EXTRA_ARGS),
                 p, plen  FP_FAIL_EXTRA_ARGS);
}


// ----------------------- misc? ------------------------
// checks associated with casting a seq to an fseq
#define CHECK_SEQTOFSEQ(b,p) __CHECK_SEQTOFSEQ(b, p  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_SEQTOFSEQ(void *b, void *p FP_FAIL_EXTRA_PARAMS)
{
  __CHECK_LBOUND(b, p FP_FAIL_EXTRA_ARGS);
}

// verify that 'x' is nonnegative
#define CHECK_POSITIVE(x) __CHECK_POSITIVE(x  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_POSITIVE(int x  FP_FAIL_EXTRA_PARAMS)
{
  if (x < 0) {
    fp_fail(FAIL_DECR_FSEQ  FP_FAIL_EXTRA_ARGS);
  }
}

// verify that computing 'p+x' does not overflow
#define CHECK_ADVANCE(p, x) __CHECK_ADVANCE(p, x  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_ADVANCE(char *p, int x  FP_FAIL_EXTRA_PARAMS)
{
  if (p + x < p) {
    fp_fail(FAIL_DECR_FSEQ  FP_FAIL_EXTRA_ARGS);
  }
}

// we can turn off the null check to improve performance;
// sm: for reasons still unclear to me, George wants to explicitly check,
// so this remains commented-out..
// #define OMIT_NULL_CHECK

#define CHECK_NULL(p) __CHECK_NULL(p  FILE_AND_LINE)
#ifdef OMIT_NULL_CHECK
  #define __CHECK_NULL(p)
#else
  __inline CHECK_STATIC
  void __CHECK_NULL(void *p  FP_FAIL_EXTRA_PARAMS)
  {
    if (!p) {
      // sm: we used to call searchLog here but that makes no sense
      // since 'p' is 0
      fp_fail(FAIL_NULL  FP_FAIL_EXTRA_ARGS);
    }
  }
#endif


// --------------------- strings -------------------
// given a string pointer, return a pointer to the 0 byte which
// terminates it; the RWSTRING-qualified pointers are *not* allowed
// to overwrite this 0 byte
#define CHECK_FETCHSTRINGEND(p) __CHECK_FETCHSTRINGEND(p  FILE_AND_LINE)
__inline static void * __CHECK_FETCHSTRINGEND(char *p  FP_FAIL_EXTRA_PARAMS)
{
  unsigned int len;
  /* WEIMER: mod_gzip doesn't seem to like this one, so I'm
   * temporarily changing the semantics to return NULL
   * Wed Oct 24 10:44:47 PDT 2001 */
  if (!p) {
    // fp_fail(FAIL_NULLSTRING  FP_FAIL_EXTRA_ARGS);
    return (void*)0;
  }

  len = strlen(p);

  // another chapter in the +1 saga:
    // Do not include the terminating 0 !!!
    // weimer: do include it (for reading) ! XXX! hack for now ...

    // sm: 2/02/02  Now that string changes have been made elsewhere,
    // I'm removing the +1.  This makes "make testrun/string1
    // INFERBOX=infer" properly fail with Ubound.  Wes, I need to know
    // which testcase I potentially break.

  //len += 1;

  return p + len;
}


// context: box.ml
//   This function is used by box.ml when casting from wild or index to
//   String.  I think we should disallow mixing of wild and string, but
//   that's somewhat of a different issue.  In any event, in box.ml, the
//   expected semantics of this are:
//     Given a b,p pointer, verify that p points to one of the bytes
//     in the memory area denoted by b (failing with STRINGBOUND
//     instead of {U,L}BOUND on failure..).
// context: ccuredlib.c
//   I have now removed all uses of this function from ccuredlib.c, preferring
//   instead fatp_verify_nul, which has simpler semantics and usage.
#define CHECK_STRINGMAX(p,b) __CHECK_STRINGMAX(p, b  FILE_AND_LINE)
__inline CHECK_STATIC
unsigned int __CHECK_STRINGMAX(void *p, void* b  FP_FAIL_EXTRA_PARAMS)
{
  int nrWords;
  int max;
  //  U32 *addr = (U32*)p;
  if(! b) {
    fp_fail(FAIL_NONPOINTER  FP_FAIL_EXTRA_ARGS);
  }
  if(p < b) {
    fp_fail(FAIL_STRINGBOUND  FP_FAIL_EXTRA_ARGS);
  }
  nrWords = *(((U32*)b) - 1);
  max     =  ((char*)b - (char*)p) + (nrWords << 2);
  if(max <= 0) {
    fp_fail(FAIL_STRINGBOUND  FP_FAIL_EXTRA_ARGS);
  }
  return (unsigned int)max;
}


// this function is no longer used anywhere (2/02/02); marked for deletion
#if 0
  #define CHECK_STRINGLEN(p,b) __CHECK_STRINGLEN(p, b  FILE_AND_LINE)
  __inline CHECK_STATIC
  unsigned int __CHECK_STRINGLEN(const void *p, unsigned int max  FP_FAIL_EXTRA_PARAMS)
  {
    int i = 0;
    const char *x = (const char*)p;
    for(; max; max--, i++, x++) {
      if(! *x) {
        return i;
      }
    }
    fp_fail(FAIL_STRINGBOUND  FP_FAIL_EXTRA_ARGS);
    return 0;
  }
#endif // 0


/************** TAG MANIPULATION ***************/

// TODO: modify tag scheme so the 1s are associated with the 'b' fields

/* We have one tag bit per aligned data word. The tag bit value is 1 if the
 * data word contains a pointer (i.e. the _p first-half of a fat pointer)
 * and 0 otherwise. A very important invariant is that a tag bit sequence
 * "10" denotes two data words that contain a fat pointer (with the fields
 * _p and _b in that order). To maintain this invariant we have to clear the
 * tag bit for word "w - 1" when we write anything in the word "w". An
 * exception is when we write a whole fat pointer in which case we write the
 * tags "10".  
 *
 * Also note that we're counting bits starting at the right edge and moving
 * left.  So, if I have a block of memory containing one fat pointer, with
 * a valid base and ptr, it looks like:
 *
 *  <-- 32 --> <-- 32 --> <-- 32 --> <--------------- 32 -------------------->
 *  +---------+----------+----------+-----------------------------------------+
 *  | size: 2 |    ptr   |   base   | 0000 0000 0000 0000 0000 0000 0000 0001 |
 *  +---------+----------+----------+-----------------------------------------+
 *                                                                         ^^
 * The last "1" and "0" are the "10" combination referred to above.
 *
 * (However, we work left to right among the *words* of the tag area.  Also,
 * if the machine is little-endian, the byte containing the value 1 will actually
 * be first in the tag area, since they are written using U32 pointers.)
 */

// Note:  In this section, several chunks of code are controlled by the NO_TAGS
// #define.  The purpose of NO_TAGS is to measure the performance impact of
// tag manipulations -- the system is unsound when NO_TAGS is defined.

// The number of tags bits per data word
#define TAGBITS          1
#define TAGBITSMASK      ((1 << TAGBITS) - 1)    // 1

// How many tags fit into a 32-bit word
#define TAGSPERWORD_LOG  (5 - TAGBITS + 1)       //  5
#define TAGSPERWORD      (1 << TAGSPERWORD_LOG)  //  0x20

// The tags for a fat pointer
#define WILDPTR_TAG       1  // a 1 followed by a 0
#define WILDPTR_MASK      (TAGBITSMASK | (TAGBITSMASK << TAGBITS))   // 3


typedef struct tagAddr
{
  U32 *word;                  /* This is the address of the word in which
                               * the tag lives */
  int bit;                    /* This is the index of the bit where the tag
                               * starts. 0 means the LSB. Will always be
                               * less or equal 32 - TAGBITS (at least one
                               * tag still fits)  */
} TAGADDR;


// given a TAGADDR, retrieve either a 0 or 1
#define READ_TAG(addr) ((*((addr).word) >> (addr).bit) & 1)
   

// first, we give definitions for the tag functions in the normal
// case where tags are in effect
#ifndef NO_TAGS

// given the number of data words, return the number of tag words that
// must be allocated to track the types in the data
__inline CHECK_STATIC
int CHECK_NRTAGWORDS(unsigned int nrDataWords)
{
  return (nrDataWords + TAGSPERWORD - 1) >> TAGSPERWORD_LOG;
}


/* Compute the address of the tag for a given word. Both the base and p must
 * be aligned on word boundary */
#define CHECK_FETCHTAGADDR(b,bw,p) __CHECK_FETCHTAGADDR(b,bw,p FILE_AND_LINE)
__inline CHECK_STATIC
TAGADDR __CHECK_FETCHTAGADDR(void *base,            // Base
                             unsigned int bwords,   // words in area
                             void *p                // the pointer
                             FP_FAIL_EXTRA_PARAMS)
{
  TAGADDR res;
  int wrdIdx    = ((char*)p - (char*)base) >> 2;
  res.word      = (U32*)base + bwords + (wrdIdx >> TAGSPERWORD_LOG);
  res.bit       = TAGBITS * (wrdIdx & (TAGSPERWORD - 1));
  return res;
}

// read the tag bit for the given word; if 'p' is not aligned, we get the
// tag for the word whose interior it's pointing at (as if we aligned it first)
__inline CHECK_STATIC
unsigned CHECK_FETCHTAGBIT(void *base,            // start of memory area
                           unsigned int bwords,   // # of words in the memory area
                           void *p)               // pointer to word we're interested in
{
  // composition of READ_TAG and CHECK_FETCHTAGADDR, above
  int wrdIdx    = ((char*)p - (char*)base) >> 2;
  U32 *word     = (U32*)base + bwords + (wrdIdx >> TAGSPERWORD_LOG);
  int bit       = TAGBITS * (wrdIdx & (TAGSPERWORD - 1));
  return (*word >> bit) & 1;
}


/* Compute the number of bits that correspond to the number of words covered
 * by the range given. nextAddr is 1 past the last address in the range */
__inline CHECK_STATIC
int CHECK_NRTAGBITS(char* startAddr, char* nextAddr)
{
  int startWord = ((int)startAddr) >> 2;
  int nextWord  = ((int)nextAddr + 3) >> 2;
  return (nextWord - startWord) * TAGBITS;
}

/* Zero out the tags for a written memory area. Later we'll come and fix the
 * tags for the words that contain pointers. We must clear also the tag
 * for the word that preceedes the first written word. To avoid unnecessary
 * clearing of tag bits make sure that you pass in offset the offset to the
 * first scalar field written. This means what we write the tags for the
 * memory area "where + offset - 4" to "where + size - 1". */
__inline CHECK_STATIC
void CHECK_ZEROTAGS(void *base,
                    unsigned int nrbwords,
                    void *where,
                    unsigned int size /* in bytes */,
                    unsigned int offset /* in bytes */) 
{
  char *wherec = (char*)where;
    
  // sm: there used to be a "return" at this point, effectively disabling
  // this function.. I'm re-enabling

  // Adjust for the offset
  wherec += offset;
  size  -= offset;
  if((void*)wherec > base) { // Adjust for the preceeding word
    wherec -= 4;
    size  += 4;
  }

  {
    TAGADDR stag    = CHECK_FETCHTAGADDR(base, nrbwords, wherec);
    int nrBits      = CHECK_NRTAGBITS(wherec, wherec + size);

    if(stag.bit > 0) {
      if(stag.bit + nrBits < 32) { // Don't zero to the end
        *stag.word ++ &= (~ (((1 << nrBits) - 1) << stag.bit));
        nrBits = 0;
      } else {
        *stag.word ++ &= (~ ((0 - 1) << stag.bit));
        nrBits -= (32 - stag.bit);
      }
    }
    while(nrBits >= 32) {
      *stag.word ++ = (U32)0;
      nrBits -= 32;
    }
    if(nrBits > 0) { // Some leftover in the last word
      *stag.word &= (~ ((1 << nrBits) - 1));
    }
  }
}

/* Copy the tags going forward. Call this in conjuction with any memory
 * copy. The psrc and pdst and size should be _exactly_ as passed to memcpy.
 * NOTE THAT THIS MIGHT NOT WORK PROPERLY IN CASE OF OVERLAPPING AREAS
 * (since the bits are read first and then written as a block. To model
 * overlapping correctly we need to copy one tag at a time.) Note that this
 * is used in a somewhat unusual way in realloc.  */
void CHECK_COPYTAGSFORW(void *bsrc, /* base of src */
                        unsigned int lensrc, /* len of src (in words) */
                        char* psrc, /* pointer of src */
                        void *bdst, /* base of dest */
                        unsigned int lendst, /* words in dest*/
                        char* pdst, /* pointer of dest */
                        unsigned int size /* Size in bytes
                                           * of the memory
                                           * being copied */);

/* Copy the tags going backward. The psrc, pdst and size arguments should be
 * exactly as passed to memcpy. The copying is done going backwards from the
 * last word. NOTE THAT THIS MIGHT NOT WORK PROPERLY IN CASE OF OVERLAPPING
 * AREAS (since the bits are read first and then written as a block. To
 * model overlapping correctly we need to copy one tag at a time.) Note that
 * this is used in a somewhat unusual way in realloc.  */
void CHECK_COPYTAGSBACK(void *bsrc, /* base of src */
                        unsigned int lensrc, /* len of src (in words) */
                        char* psrc, /* pointer of src */
                        void *bdst, /* base of dest */
                        unsigned int lendst, /* words in dest*/
                        char* pdst, /* pointer of dest */
                        unsigned int size);

#define CHECK_WILDPOINTERREAD(b,n,w) __CHECK_WILDPOINTERREAD(b,n,w  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_WILDPOINTERREAD(void *base, unsigned int nrWords,
                            void *where  FP_FAIL_EXTRA_PARAMS)
{
  TAGADDR tag = CHECK_FETCHTAGADDR(base, nrWords, where);
  U32 diffTag, memTag;

  if((U32)where & 0x3) {
    fp_fail(FAIL_UNALIGNED  FP_FAIL_EXTRA_ARGS);
  }

  memTag  = (* tag.word++) >> tag.bit;

  if(tag.bit == (32 - TAGBITS)) { // Only the tag for the first half
    // sm: I'm *un*able to verify that the following line matters,
    // despite considerable attempts.  I think when this is because
    // when the writing code (below) is functioning correctly, the
    // next tag is always 0 if the first tag is 1, hence the |= operation
    // does not change the value of memTag.  Nevertheless, I think
    // the code is correct in intent.

    memTag |= ((* tag.word) << TAGBITS);
  }

  diffTag = (memTag & WILDPTR_MASK) ^ WILDPTR_TAG;
  if(diffTag == 0)
    return;
  // We give it one more chance. Maybe the word that would be read for base
  // is 0, in which case everything is still safe. This will handle the case
  // when an area has been initialized with 0 and the tags are all 0

  /* Note that this is not quite right. We should not be writing in the
   * memory. Instead we should be just pretending to load a 0 instead of the
   * base */
  *(((U32*)where) + 1) = (U32)0;
  return;
}


/* We are writing (changing the value of) a fat pointer
 * which, itself, lives inside a tagged block of memory */
// this first version does not check for writing stack pointers
#define CHECK_WILDPOINTERWRITE_NOSTACKCHECK(b,n,w,x,y) \
        __CHECK_WILDPOINTERWRITE_NOSTACKCHECK(b,n,w,x,y  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_WILDPOINTERWRITE_NOSTACKCHECK(
  // the base of the block in which the fat pointer lives
  // (the byte just after the length)
  void *base,

  // the number of words in the block containing
  // the fat pointer
  unsigned int nrWords,

  // the address of the fat pointer itself; this will
  // point to the '_p' field, since it is first
  void **where,

  // 'whatbase' is the value we want the fat pointer's
  // '_b' (base) field to become
  void *whatbase,

  // and this is what we want its '_p' field to become
  void *whatp

  FP_FAIL_EXTRA_PARAMS)
{
  // compute bit address of tag bit
  TAGADDR tag  = CHECK_FETCHTAGADDR(base, nrWords, where);

  // sm: maybe this alignment check should go first?
  if((U32)where & 0x3) {
    fp_fail(FAIL_UNALIGNED  FP_FAIL_EXTRA_ARGS);
  }

  // mask out the existing tag bits, and OR in the "10"
  // for a valid fat pointer
  *tag.word = (* tag.word & (~ (WILDPTR_MASK << tag.bit))) |
                  (WILDPTR_TAG << tag.bit);

  #if 0     // testing (forgot gdb was showing me values in decimal!)
  {
    U32 data = *(tag.word);                // read tags word
    data &= ~(WILDPTR_MASK << tag.bit);     // strip existing tag pair
    data |= WILDPTR_TAG << tag.bit;         // add "01" bitpattern
    *(tag.word) = data;                    // write it back
  }
  #endif // 0

  if (tag.bit < 32 - TAGBITS) {
    // Both fit in the same word
  }
  else {
    // sm: I'm able to verify that the following lines do have
    // a good effect; scott/arraytags will fail if these lines
    // are removed

    // the "10" straddles a word boundary, so the "0" didn't
    // actually get written above (left-shift just discards
    // extra bits); so we now write that last "0" into the
    // next tag word
    tag.word ++;
    *tag.word = (*tag.word & (~ (WILDPTR_MASK >> TAGBITS))) |
                   (WILDPTR_TAG >> TAGBITS);
  }
}


// same as the above check, but also check for writing stack pointers
#define CHECK_WILDPOINTERWRITE(b,n,w,x,y) __CHECK_WILDPOINTERWRITE(b,n,w,x,y  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_WILDPOINTERWRITE(
  void *base, unsigned int nrWords, void **where,
  void *whatbase, void *whatp   FP_FAIL_EXTRA_PARAMS)
{
  __CHECK_WILDPOINTERWRITE_NOSTACKCHECK(base, nrWords, where, whatbase,
                                       whatp  FP_FAIL_EXTRA_ARGS);

  // this is the stack pointer check
  if (whatbase) {
    CHECK_STOREPTR(where, whatp);
  }
}


// now, we give stub definitions for when tags are disabled
#else     // NO_TAGS
  // for query functions, sometimes we can answer sensibly, but sometimes
  // the query is invalid
  #define CHECK_NRTAGWORDS(datawords) 0
  #define CHECK_FETCHTAGADDR(b,bw,p) ERROR_NO_TAGS_IS_DEFINED
  #define CHECK_FETCHTAGBIT(b,bw,p) ERROR_NO_TAGS_IS_DEFINED
  #define CHECK_NRTAGBITS(start, next) 0

  // for tag manipulations, make them no-ops
  #define CHECK_ZEROTAGS(b,bw,w,s,o)
  #define CHECK_COPYTAGSFORW(bs,ls,ps,bd,ld,pd,sz)
  #define CHECK_COPYTAGSBACK(bs,ls,ps,bd,ld,pd,sz)
  #define CHECK_WILDPOINTERREAD(b,n,w)

  // even without tags there's a small check to perform on pointer write
  #define CHECK_WILDPOINTERWRITE(b,n,w,x,y) __CHECK_WILDPOINTERWRITE(b,n,w,x,y  FILE_AND_LINE)
  __inline CHECK_STATIC
  void __CHECK_WILDPOINTERWRITE(
    // see above for interpretation of these fields
    void *base, unsigned int nrWords, void **where,
    void *whatbase, void *whatp
    FP_FAIL_EXTRA_PARAMS)
  {
    if (whatbase) {
      CHECK_STOREPTR(where, whatp);
    }
  }
#endif    // NO_TAGS


// sm: I need this for the _setjmp_w macro; it does the same thing
// as fatp_write_atleast() in ccuredlib.c
#define CHECK_WILDP_WRITE_ATLEAST(p, b, n) \
        __CHECK_WILDP_WRITE_ATLEAST(p, b, n  FILE_AND_LINE)
__inline CHECK_STATIC
void __CHECK_WILDP_WRITE_ATLEAST(
  void *p,    // pointer to first byte to write
  void *b,    // base of the home area to be written
  int n       // number of bytes to write
  FP_FAIL_EXTRA_PARAMS)
{                         
  // words in the home area
  unsigned nwords = CHECK_FETCHLENGTH(p, b);
  
  // bytes available between 'p' and the end
  int nbytes = (int)(nwords << 2) - ((char*)p - (char*)b);

  if (n < 0) {
    FP_FAIL(FAIL_LBOUND);
  }
  if (n > nbytes) {
    FP_FAIL(FAIL_UBOUND);
  }

  CHECK_ZEROTAGS(b, nwords, p, n, 0);
}


// ---------------- "table" pointers and splay tree access -----------
// sm: this stuff is largely defunct now, mainly because the
// splay tree turned out to be *much* too slow a data structure

/* weimer: our hooks into the splay tree data structure */
#include "splay.h"

// here, 'end' is the last+1 byte
__inline CHECK_STATIC
void CHECK_REGISTERAREA(int kind, void *base, void *end)
{
  if (kind == registerTagged || kind == registerSized) {
      end = CHECK_FETCHEND(base,base);
  }
  splayInsertInterval(kind, base, (char*)end - (char*)base);
  //splayPrintTree(&__global_tree[kind]);
  return;
}


// sm: changed this to a macro, so that unregistering works in gcc
// even when optimization is turned off
#define CHECK_UNREGISTERFRAME()                                \
{                                                              \
  char* thisf;                                                 \
  CHECK_GETFRAME(thisf);                                       \
  splayRemoveKindBelow(MR_STACK, registerTagged, thisf);       \
  splayRemoveKindBelow(MR_STACK, registerSized, thisf);        \
  splayRemoveKindBelow(MR_STACK, registerSeq, thisf);          \
}


__inline CHECK_STATIC
void * CHECK_FINDHOME(int kind, void *p)
{
  return splayFindHome(kind, p);
}

__inline CHECK_STATIC
void * CHECK_FINDHOMEEND(int kind, void *p, void **e)
{
  return splayFindHomeEnd(kind, p, e);
}


__inline CHECK_STATIC
void* trusted_cast(void *x) {
  return x;
}
