/* 
 *
 * Copyright (c) 2001 by
 *  George C. Necula	necula@cs.berkeley.edu
 *  Scott McPeak        smcpeak@cs.berkeley.edu
 *  Wes Weimer          weimer@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products 
 * derived from  this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// ccuredlib.c
// runtime library for CCured programs; mostly libc wrappers

// The security model for these wrappers is that the programmer is the
// adversary, trying to break memory safety by any clever means,
// despite the program having been passed through the CCured
// translator.  However, this is (for the moment) an ideal; there are
// still a few holes in the wrappers.  We fix them when we notice them
// and the fix is not too onerous; we note below where we choose not
// to fix, with "UNSOUND".  Any unsoundness that is not marked as such
// is a bug.
//
// The security model we believe we enforce everywhere is that the
// programmer is not malicious, but perhaps forgetful.  Of course
// there is no formal model for this, but it means roughly we try to
// be very careful about off-by-one and other typical overflow
// scenarios, but do not (for example) guard against a
// carefully-constructed bogus argument to a FILE*-accepting function.
//
// One global unsoundness: threading.  This library is not
// thread-safe.  It only uses globals in a few places; that might be
// solvable.  But it extensively uses invariants which tie certain
// pieces of data together, and those invariants are, at times, false.
// In a threaded environment, another thread might access the data
// while the invariants are false (e.g. the tags are out of synch with
// data values).  Fixing this would require acquiring some kind of
// lock in almost all the code which manipulates tags...


#include <stdio.h>
#include <stdlib.h>      // malloc/free (?)
#include <setjmp.h>      // setjmp, jmp_buf
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>       // __ctype_b
#include <time.h>        // time, time_t
#include <stdarg.h>      // various varargs-related

#ifdef _GNUCC
  #include <unistd.h>    /* struct stat, fd_set */
  #include <fcntl.h>     /* open; pulled up here to avoid implicit decl */
  #include <signal.h>    // signal
  #include <syslog.h>    // syslog, LOG_ERR(=3)
  #include <sys/times.h> // times, clock_t
  #include <sys/time.h>  // gettimeofday, struct timeval, struct timezone
  #include <sys/types.h> // size_t?
  #include <sys/socket.h>// struct sockaddr, sendto
  #ifndef __CYGWIN__
    #define O_BINARY 0     /* doesn't exist in linux */
    #include <sys/mman.h>  // MAP_FAILED
    #include <glob.h>      // glob, glob_t, GLOB_xxx codes
  #endif
  #define _MAX_PATH 255  /* wouldn't begin to know where to properly find this */
#endif

// some external prototypes which used to be hanging about in random places;
// I doubt we need most (any?) of them, but I'll keep some of them
extern char* strerror(int);

#ifdef _MSVC
  char *strcpy(char *, char const *);
  unsigned int strlen(char const *);
  extern int write(int fid, void /* const */ *buff, unsigned int size);
  extern int read(int fid, void *buff, unsigned int size);
#endif


// sanity check
#if defined(_DEBUG) && defined(RELEASELIB)
  #error _DEBUG and RELEASELIB are inconsistent
#endif

// some random stuff..; declares fp_fail, and has many checking macros/functions
#include "ccured.h"

// debugging print
#if 0
  #define DP(arg) printf arg
#else
  #define DP(arg)
#endif

// round a byte count up to the next word boundary
#define ALIGNBYTES(n) (((n) + sizeof(int) - 1) & ~(sizeof(int) - 1))

// round a byte count up to the next word boundary, and return
// the count in *words*
#define BYTESTOWORDS(n) (((n) + 3) >> 2)

// multiply two unsigned 32-bit values but check for overflow;
// this is to guard against the case where our wrapper only checks
// the product but libc does iteration internally; yes division is
// somewhat expensive, but this is usally called in a situation where
// a system call or even some I/O might happen, which certainly
// dominates the cost of integer division
static unsigned mult_u32(unsigned x, unsigned y)
{
  if (x==1 || y==1 || x==0 || y==0) {
    // can't overflow
  }
  else if (x > 0xFFFFFFFF / y) {
    fp_fail(FAIL_OVERFLOW  FILE_AND_LINE);
  }

  return x*y;
}


// --------------- garbage collector interface ---------------
// pull in declarations for garbage collector
#ifndef RELEASELIB
  #define GC_DEBUG
#endif
#include "gc/include/gc.h"

// 10/28/01: changed from defining 'malloc' etc to defining
// 'ccured_malloc', with a #define in ccured.h, because there
// was a problem with the former under cygwin: when libc calls
// malloc (twice) before main runs, it the call to GC_MALLOC
// below would always segfault.  we tried some hackery where
// our malloc would try to avoid handling the first 2 calls,
// but that didn't help.  so we decided to avoid linker tricks
// altogether and rely in the C preprocessor instead to intercept
// malloc/free

// cause references below to malloc/free/realloc to use
// the garbage collector's versions instead
void *ccured_malloc(unsigned sz)
{
  void *ret;

  // sm: I need one more byte allocated, since that's where I stash
  // the final, untouchable null; clearly we don't need this for *every*
  // allocation, only those to which char* pointers point.. but I will
  // let someone else be smarter about choosing when to do this (of
  // course, that decision must accompany the decision not to *write*
  // that final null!)
  sz++;

  DP(("malloc(%d)\n", sz));
  ret = GC_MALLOC(sz);
  DP(("malloc(%d) yielded %p\n", sz, ret));

  // trying to track down bug in health
  //memset(ret, 0xAA, sz);

  return ret;
}

// sm 3/07/02: added '+1' to calloc and realloc following the
// same reasoning as above in malloc

void *ccured_calloc(size_t x, size_t y)
{
  // GC_MALLOC always zeroes its result
  void *ret = GC_MALLOC(x*y+1);
  DP(("calloc(%d,%d) yielded %p\n", x, y, ret));
  return ret;
}

void ccured_free(void *p)
{
  DP(("free(%p)\n", p));
  GC_FREE(p);
}

void *ccured_realloc(void *p, unsigned int sz)
{
  void *ret = GC_REALLOC(p, sz==0? 0 : sz+1);
  DP(("realloc(%p, %d) yielded %p\n", p, sz, ret));
  return ret;
}

// I intend to call this from user code to invoke
// the garbage collector, and verify it has accomplished
// some useful work; returns # of bytes freed
int ccured_explicit_gc()
{
  int before = GC_get_free_bytes();
  GC_gcollect();
  return GC_get_free_bytes() - before;
}

// the GC has been instrumented (optionally) to call this function
// every time a block is noticed to be reachable; the idea is I will
// add code to detect if the programmer freed this block, in which
// case I might take some additional action (like warn the user?)
void noticedReachable(void *block, size_t size)
{
  //printf("noticedReachable: %p, sz=%d\n", block, size);
}

void noticedUnreachable(void *block, size_t size)
{
  //printf("noticedUnreachable: %p, sz=%d\n", block, size);
}

void noticedContiguousUnreachable(void *block, size_t size, size_t arraySize)
{
  //printf("noticedContiguousUnreachable: %p, sz=%d, arraySize=%d (ratio: %d)\n",
  //       block, size, arraySize, arraySize/size);
}



// ----------------- multi-word pointer templates -------------------
// the instrumented program will contain its own declarations of various
// multi-word pointers; but this library doesn't see that code, so we
// also declare these, and then the functions below use the names
// declared here; it's important to make sure that the pointer types
// used are consistent with the types implied by the mangled names,
// because *nothing* systematically checks this for consistency

// name mangling algorithm:
//   1. write down the type of the function on one line,
//      including return type; e.g.
//        int *foo(char *q, struct Bar *b, int y);
//
//   2. add safe/fseq/seq/wild qualifiers to pointer types, e.g.:
//        int * safe foo(char * fseq q, struct Bar * wild b, int y);
//
//   3. append an underscore to the function's name, e.g.:
//        foo_
//
//   4. read the qualifiers left to right, appending one character
//      for each qualifier, according to the following map:
//        safe    s
//        seq     q
//        fseq    f
//        wild    w
//      e.g.,
//        foo_sfw
//      Exception: nested qualifiers are read right to left; for example,
//        int getopt(int argc, char * SAFE * FSEQ argv, char *opts)
//      which accepts an FSEQ pointer to an array of SAFE pointers, is
//        getopt_fss
//
//   5. for the wrappers below, use the declared structures instead
//      of actual syntactic qualifiers, e.g.:
//        int *foo(fseqp_char q, fatp_void b, int y);
//      in this case, if the wrapper needed to access b's fields,
//      it would be necessary to declare a fatp_Bar (and possibly
//      also struct Bar), or at least cast the _p field on use
//
//   Notes:
//     - we currently do not dig inside structures and mangle names
//       according to qualifiers in there; this may be a mistake, but
//       it's how it currently (2/01/02) works
//     - I do not know how we mangle names where pointers to functions,
//       pointers to arrays, etc., are involved
//     - for historical reasons, wild pointers are typically called "fatp"
//       when naming the types..


// there seems to be little rhyme or reason as to why some of the fields
// are declared void* and some are declared char* ...

typedef struct fatp_char {
  char * _p ;
  void * _b ;
} fatp_char;

typedef struct fseqp_char {
  char *_p;
  char *_e;
} fseqp_char;

typedef struct fatp_char1 {
  char const  * _p ;
  void * _b ;
} fatp_char1;

typedef struct fatp_void {
  void *_p ;
  void *_b ;
} fatp_void;

typedef struct fseqp_void {  // for realloc_ff
  void *_p;
  void *_e;
} fseqp_void;

typedef struct fatp_void1 {
  void /*const*/ * _p ;      // sm: removed 'const' because triggers warnings -- why was it present?
  void * _b ;
} fatp_void1;

typedef struct seqp_void {
  void /*const*/ * _p ;
  void * _b ;
  void * _e ;
} seqp_void;

typedef struct seqp_char {
  char * _p ;
  char * _b ;
  char * _e ;
} seqp_char;

typedef struct seqp_int {
  int * _p ;
  int * _b ;
  int * _e ;
} seqp_int;

typedef struct fatp_fatp_char {
  fatp_char * _p ;
  void * _b ;
} fatp_fatp_char;

typedef struct fseqp_fatp_char {
  fatp_char * _p ;
  void * _e ;
} fseqp_fatp_char;

typedef struct seq_fatp_char {
  fatp_char * _p ;
  void * _b ;
  void * _e;
} seq_fatp_char;

typedef struct fatp_FILE {
  FILE * _p ;
  void * _b ;
} fatp_FILE;

typedef struct fatp_struct__stat {
  struct _stat * _p ;
  void * _b ;
} fatp_struct__stat;

typedef struct fatp_time_t1 {
  time_t * _p ;
  void * _b ;
} fatp_time_t1;


// sm: these duplicate things in ccuredcheck.h, which gets #included above
// via ccured.h; however, #defining NO_CHECKS would disable that inclusion,
// and I don't understand when NO_CHECKS is used, so I leave this for now..
#ifdef _MSVC
  #define I32 int
#endif
#ifdef _GNUCC
  #define I32 int
#endif
#define U32 unsigned I32

#ifndef MIN
  #define MIN(x,y) (((x) < (y)) ? (x) : (y))
#endif


// --------------------- error reporting / aborting ---------------
// line these up with ccuredcheck.h codes to verify correspondence
char* failMessages[] = {
  "Below stack address",              // 0
  "Stack address",
  "Lbound",
  "Ubound",
  "Function pointer",
  "Decrement FSEQ",                   // 5
  "Null pointer",
  "Null string",
  "Non-terminated string",
  "Unaligned pointer",
  "Malloc",                           // 10
  "Realloc",
  "Free",
  "Realloc (shrink)",
  "Realloc (grow)",
  "Non-pointer",                      // 15
  "Scalar log",
  "Null string in strlen",
  "variable argument",
  "call to va_arg without call to va_start",
  "too many calls to va_arg",         // 20
  "type mismatch between caller and use of va_arg function",
  "unrecognized format character",
  "Integer arithmetic overflow"
};

NORETURN  void fp_fail_str(char *str  FP_FAIL_EXTRA_PARAMS)
{
  #ifdef FP_FAIL_IS_VERBOSE
    #define MSGFMT "Failure at %s:%d: %s"
  #else
    #define MSGFMT "Failure: %s"
  #endif

  // obscure C: juxtaposed string constants are implicitly concatenated
  fprintf(stderr, MSGFMT "\n"  FP_FAIL_EXTRA_ARGS, str);
  fflush(stderr);

  #if defined(_GNUCC) && !defined(NO_FPFAIL_SYSLOG)
    syslog(LOG_ERR, MSGFMT  FP_FAIL_EXTRA_ARGS, str);
  #endif

  #if defined(_MSVC) && defined(_DEBUG)
    // on windows a debug trap is desired
    _asm { int 3 }
  #else
    // on unix an abort signal does nicely
    abort();
  #endif

  #undef MSGFMT
}

NORETURN  void fp_fail(int msgId  FP_FAIL_EXTRA_PARAMS)
{
  fp_fail_str(failMessages[msgId]  FP_FAIL_EXTRA_ARGS);
}


// ------------------- general-purpose: fatp ---------------------
// this section has some general purpose queries and checkers for
// various kinds of pointers, which are then shared by the wrappers

// possible TODO: these don't do the best thing with FP_FAIL_IS_VERBOSE

// given a fat pointer, return the # of bytes legal to read/write
// starting from its _p field
__inline static int fatp_length(fatp_char ptr)
{
  unsigned nwords = CHECK_FETCHLENGTH(ptr._p, ptr._b);

  // the code that was in memset_ww had a -1, but I think that's
  // wrong now that strings are handled orthogonally
  return (int)(nwords << 2) - (ptr._p - (char*)ptr._b);
}

// fail unless there are >=n bytes read/writable after ptr
__inline static void fatp_verify_atleast(fatp_char ptr, int n)
{
  if (n < 0) {
    FP_FAIL(FAIL_LBOUND);
  }
  if (n > fatp_length(ptr)) {
    FP_FAIL(FAIL_UBOUND);
  }
}


// clear the tags for the 'n' bytes starting at ptr
__inline static void fatp_clear_tags(fatp_char ptr, int n)
{
  // no-op when NO_TAGS
  CHECK_ZEROTAGS(ptr._b, CHECK_FETCHLENGTH(ptr._p, ptr._b),
                 ptr._p, n, 0);
}

// copy the tags for the 'n' bytes starting at 'src' to the
// tags for the 'n' bytes starting at 'dest'; tries to handle
// overlapping memory areas properly, because we do not trust
// the programmer to respect e.g. memcpy's requirement of non-
// overlapping areas
__inline static void fatp_copy_tags(fatp_char dest, fatp_char src, int n)
{
  unsigned int srclen = CHECK_FETCHLENGTH(src._p, src._b);
  unsigned int destlen = CHECK_FETCHLENGTH(dest._p, dest._b);

  if (dest._p < src._p) {
    // in the old memmove_www, the forward copy was used when it
    // appeared that src and dest don't overlap; that check was
    // wrong because the *tags* might have overlapped anway; but it
    // suggests forward may be preferred for performance reasons, so
    // a possible micro-optimization is to do a more careful check
    // for nonoverlapping and use forward in that case
    CHECK_COPYTAGSFORW(src._b, srclen, src._p,
                       dest._b, destlen, dest._p, n);
  }
  else if (dest._p > src._p) {
    CHECK_COPYTAGSBACK(src._b, srclen, src._p,
                       dest._b, destlen, dest._p, n);
  }
  else {
    // dest and src are the same place; nothing to do
  }
}


// verify the memory is writable, and then clear the tags
__inline static void fatp_write_atleast(fatp_char ptr, int n)
{
  fatp_verify_atleast(ptr, n);
  fatp_clear_tags(ptr, n);
}

// verify there is a nul value between ptr and the end of its buffer;
// typically the context is we're about to pass a char* to libc which
// will interpret it as a nul-terminated string, so we want to make
// sure it will only access (read) bytes within the valid memory area;
// returns 'len'
// (terminology: nul is ASCII 0; null is a pointer value)
__inline static int verify_nul(char const *p, int len)
{
  int origlen = len;
  while (len) {
    if (*p == 0) {
      // ok
      return origlen;
    }
    len--;
    p++;
  }

  // didn't find a nul in the accessible area
  fp_fail(FAIL_STRINGBOUND  FILE_AND_LINE);
  return 0;     // silence warning
}

// verify_nul, for wild pointers
__inline static int fatp_verify_nul(fatp_char ptr)
{
  return verify_nul(ptr._p, fatp_length(ptr));
}


// ------------------- general-purpose: seqp ---------------------
// similar section to the above, but for sequence pointers

// given a seq pointer, return the # of bytes legal to read/write
// starting from its _p field
__inline static int seqp_length(seqp_char ptr)
{
  return (int)ptr._e - (int)ptr._p;
}

// fail unless there are >=n bytes read/writable after ptr
__inline static void seqp_verify_atleast(seqp_char ptr, int n)
{
  if (n < 0) {
    FP_FAIL(FAIL_LBOUND);
  }
  if (n > seqp_length(ptr)) {
    FP_FAIL(FAIL_UBOUND);
  }
}

// I'm not consistent is using this for seq/fseq, but I may as well
// keep the ones I have
#define seqp_write_atleast(ptr, n) seqp_verify_atleast(ptr, n)

// verify there is a nul value between ptr and the end of its buffer
__inline static int seqp_verify_nul(seqp_char ptr)
{
  return verify_nul(ptr._p, seqp_length(ptr));
}



// ------------------- general-purpose: fseqp ---------------------
// similar section to the above, but for forward-sequence pointers

// given an fseq pointer, return the # of bytes legal to read/write
// starting from its _p field
__inline static int fseqp_length(fseqp_char ptr)
{
  return (int)ptr._e - (int)ptr._p;
}

// fail unless there are >=n bytes read/writable after ptr
__inline static void fseqp_verify_atleast(fseqp_char ptr, int n)
{
  if (n < 0) {
    FP_FAIL(FAIL_LBOUND);
  }
  if (n > fseqp_length(ptr)) {
    FP_FAIL(FAIL_UBOUND);
  }
}

#define fseqp_write_atleast(ptr, n) fseqp_verify_atleast(ptr, n)

// verify there is a nul value between ptr and the end of its buffer
__inline static int fseqp_verify_nul(fseqp_char ptr)
{
  return verify_nul(ptr._p, fseqp_length(ptr));
}



// ---------------- tag manipulation ------------------
#ifndef NO_TAGS
/* Copy the tags going forward. Call this in conjuction with any memory
 * copy. The psrc and pdst and size should be _exactly_ as passed to memcpy.
 * NOTE THAT THIS MIGHT NOT WORK PROPERLY IN CASE OF OVERLAPPING AREAS
 * (since the bits are read first and then written as a block. To model
 * overlapping correctly we need to copy one tag at a time.) Note that this
 * is used in a somewhat unusual way in realloc.  */
void CHECK_COPYTAGSFORW(void *bsrc, /* base of src */
                        unsigned int lensrc, /* len of src */
                        char* psrc, /* pointer of src */
                        void *bdst, /* base of dest */
                        unsigned int lendst, /* words in dest*/
                        char* pdst, /* pointer of dest */
                        unsigned int size /* Size in bytes
                                           * of the memory
                                           * being copied */)
{
  TAGADDR stag, dtag;
  // Word starting address
  char    *alignSrc      = (char*)((U32)psrc & (~ 3));
  char    *alignDest     = (char*)((U32)pdst & (~ 3));
  char    *nextDestByte  = pdst + size; // Byte after the last copied
  int     bitsToCopy     = CHECK_NRTAGBITS(psrc, psrc + size);
  U32     srcTags   = 0;  // Keep some srcTags in here (always at LSB)
  int     nrSrcTags = 0;  // How many live bits we have in srcTags
  int     first = 1;      // A marker that says that we are copying the first

  /* If the alignment of the destination is different from that of the
   * source then we just zero out all of the destination tags  */
  if(((U32)psrc ^ (U32)pdst) & 3) {
    CHECK_ZEROTAGS(bdst, lendst, pdst, size, 0);
    return;
  }
  // Assume for now the we are copying whole words always
  stag  = CHECK_FETCHTAGADDR(bsrc, lensrc, alignSrc);
  dtag  = CHECK_FETCHTAGADDR(bdst, lendst, alignDest);
  // While there are some more bits in the source
  while(bitsToCopy) {
    // See if we need to refill srcTags
    if(nrSrcTags == 0) {
      srcTags = (*stag.word) >> stag.bit;
      nrSrcTags = MIN(32 - stag.bit, bitsToCopy);
      if(first) { // The first batch of bits
        first = 0;
        // If we are doing partial word read and write
        if(((U32)pdst & 3)) {
          // We know that source and destination have the same alignment
          srcTags &= (~ TAGBITSMASK); // Zero out the tag for the first word
        }
        if(! (srcTags & TAGBITSMASK) || (void*)alignDest > bdst) {
           /* We are starting the write with a tag 0. Must zero the previous
            * tag as well */
          // Pretend we have started behind with src
          stag.bit -= TAGBITS;
          if(stag.bit == - TAGBITS) {
            stag.word --;
            stag.bit  += 32;
          }
          dtag.bit -= TAGBITS;
          if(dtag.bit == - TAGBITS) {
            dtag.word --;
            dtag.bit += 32;
          }
          srcTags <<= TAGBITS; // Insert a 0 tag
          bitsToCopy += TAGBITS;
        }
      }
      if(bitsToCopy == nrSrcTags) { // This is the last batch
        if(((U32)nextDestByte & 3)  ||
           // We are ending the copy with the first half of a fat pointer
           ((srcTags >> (nrSrcTags - TAGBITS)) & TAGBITSMASK)) {
          // Zero out the tag for the last word
          srcTags &= (~ (TAGBITSMASK << (nrSrcTags - TAGBITS))); 
        }
      }
      stag.bit  += nrSrcTags;
      if(stag.bit >= 32) {
        stag.word ++;
        stag.bit -= 32;
      }
    }
    // Put some tags in the destination 
    {
      int tocopy  = MIN(nrSrcTags, 32 - dtag.bit);
      U32 mask    = tocopy < 32 ? (1 << tocopy) - 1 : (U32)-1;
      U32 towrite = (srcTags & mask) << dtag.bit;
      *dtag.word  = (*dtag.word & (~ (mask << dtag.bit))) | towrite;
      // Advance bit pointers
      nrSrcTags   -= tocopy;
      bitsToCopy  -= tocopy;
      srcTags     >>= tocopy;
      dtag.bit    += tocopy;
      if(dtag.bit == 32) { // Move on to the next destination pointer
        dtag.word ++;
        dtag.bit  = 0;
      }
    }
  }
}

/* Copy the tags going backward. The psrc, pdst and size arguments should be
 * exactly as passed to memcpy. The copying is done going backwards from the
 * last word. NOTE THAT THIS MIGHT NOT WORK PROPERLY IN CASE OF OVERLAPPING
 * AREAS (since the bits are read first and then written as a block. To
 * model overlapping correctly we need to copy one tag at a time.) Note that
 * this is used in a somewhat unusual way in realloc.  */
void CHECK_COPYTAGSBACK(void *bsrc, /* base of src */
                        unsigned int lensrc, /* len of src */
                        char* psrc, /* pointer of src */
                        void *bdst, /* base of dest */
                        unsigned int lendst, /* words in dest*/
                        char* pdst, /* pointer of dest */
                        unsigned int size)
{
  TAGADDR stag, dtag;
  // Word starting address
  //  char    *alignSrc      = (char*)((U32)psrc & (~ 3)); 
  char    *nextSrcByte   = psrc + size; // Byte after the last copied
  char    *nextSrcWord   = (char*)((U32)(nextSrcByte + 3) & (~ 3));
  char    *alignDest     = (char*)((U32)pdst & (~ 3));
  char    *nextDestByte  = pdst + size; // Byte after the last copied
  char    *nextDestWord  = (char*)((U32)(nextDestByte + 3) & (~ 3));
  int     bitsToCopy     = CHECK_NRTAGBITS(psrc, psrc + size);
  U32     srcTags   = 0;  // Keep some srcTags in here (always at LSB)
  int     nrSrcTags = 0;  // How many live bits we have in srcTags
  int     first = 1;      // A marker that says that we are copying the first

  /* If the alignment of the destination is different from that of the 
   * source then we just zero out all of the destination tags  */
  if(((U32)psrc ^ (U32)pdst) & 3) {
    CHECK_ZEROTAGS(bdst, lendst, pdst, size, 0);
    return;
  }
  // Assume for now the we are copying whole words always
  stag  = CHECK_FETCHTAGADDR(bsrc, lensrc, nextSrcWord - 4);
  dtag  = CHECK_FETCHTAGADDR(bdst, lendst, nextDestWord - 4);
  // While there are some more bits in the source
  while(bitsToCopy > 0) {
    // See if we need to refill srcTags
    if(nrSrcTags == 0) {
      srcTags = (*stag.word);
      // Maybe all the source tags fit into one word
      nrSrcTags = stag.bit + TAGBITS;
      if(nrSrcTags > bitsToCopy) {
        srcTags >>= (nrSrcTags - bitsToCopy);
        nrSrcTags = bitsToCopy;
      }
      if(first) { // The first batch. Meaning the bits at highest addresses
        first = 0;
        // If partial word write
        if(((U32)nextDestByte & 3) ||
           // We end the copy with the first half of a fat pointer
           ((srcTags >> (nrSrcTags - TAGBITS)) & TAGBITSMASK)) {
          // Zero out the last tag
          srcTags &= (~ (TAGBITSMASK << (nrSrcTags - TAGBITS)));
        }
      }
      if(nrSrcTags == bitsToCopy) {
        // Last batch. Meaning the bits at lowest addr
        // If partial word write, zero out the tag
        if(((U32)pdst & 3)) {
          srcTags &= (~ TAGBITSMASK);
        }
      }
      stag.bit  = 32 - TAGBITS;
      stag.word --;
    }      
    // Put some tags in the destination
    {
      int tocopy;
      int doff;
      U32 mask, towrite;
    WriteToDest:
      if(nrSrcTags < dtag.bit + TAGBITS) {
        tocopy = nrSrcTags;
        doff   = dtag.bit + TAGBITS - nrSrcTags;
      } else {
        tocopy = dtag.bit + TAGBITS;
        doff   = 0;
      }
      mask    = tocopy < 32 ? (1 << tocopy) - 1 : (U32)-1;
      towrite = ((srcTags >> (nrSrcTags - tocopy)) & mask) << doff;
      *dtag.word  = (*dtag.word & (~ (mask << doff))) | towrite;
      // Advance bit pointers
      nrSrcTags   -= tocopy;
      dtag.bit   -= tocopy;
      bitsToCopy  -= tocopy;
      if(dtag.bit < 0) { // Move on to the next destination pointer
        dtag.word --;
        dtag.bit  = 32 - TAGBITS;
      }
      // If we have just copied the tag at the lowest address
      // And the first tag is 0, then we must zero the previous tag as well
      if(bitsToCopy == 0 && (! (srcTags & TAGBITSMASK)) &&
         (void*)alignDest > bdst) {
        tocopy = TAGBITS;
        srcTags = 0;
        nrSrcTags = TAGBITS;
        goto WriteToDest;
      }
    }
  }
}
#endif // !NO_TAGS


// ------------------- malloc -------------------------
// malloc, returning a wild pointer; has to allocate and initialize
// the extra bookkeeping info for wild areas
fatp_void malloc_w(unsigned int size)
{
  fatp_void res;
  int  nrWords    = BYTESTOWORDS(size);       // align the size
  int  nrTagWords = CHECK_NRTAGWORDS(nrWords);

  // allocate: one word for the length, nrWords for data, nrTagWords for tags
  int  newSize    = ((1 + nrWords + nrTagWords) << 2);

  U32  *alloc     = (U32*)malloc(newSize);
  if(! alloc) {
    FP_FAIL(FAIL_MALLOC);
  }

  *(alloc++)   = nrWords;   // Write the length.
  res._p = (void*)alloc;    // This is the start of the data
  res._b = (void*)alloc;
  alloc += nrWords;         // Skip the data section

  for (; nrTagWords; nrTagWords--) { // Clear the tags
    *(alloc++) = 0;
  }

  return res;
}

// allocate, yield fseq
// (sm: why isn't this called malloc_f?)
fseqp_void malloc_fseq_f(unsigned size)
{
  fseqp_void res;

  // align size; fseq's e field needs to be aligned (?) so it can only
  // express boundaries that are aligned (theoretical bugfix: we used
  // to align 'e' without aligning size, leaving potentially three
  // bytes unprotected)
  size = ALIGNBYTES(size);

  res._p = malloc(size);
  res._e = ((char*)res._p) + size;      // first inaccessible word

  return res;
}


// ----------------------- free ----------------------
// note: when the GC is enabled, calls to free are no-ops anyway;
// the below is for the case where we disable the GC and just
// trust the programmer to call free correctly, as a way to measure
// the GC's performance impact

void free_safe(void *x)
{
  free(x);
}

// sm: this little hack is needed because George calls
// this free on __HEAPIFY'd areas, but the value passed
// is 4 greater than what should be passed to free
// sm: is this still true?  I thought it got fixed
void free_minus4(void *x)
{
  free((void*)( (int)x - 4 ));
}

void free_w(fatp_void x)
{
  U32 *p = (U32*)x._p;
  if(! x._b || x._b != p) {
    FP_FAIL(FAIL_FREE);
  }

  // Now find the real start
  // (sm: there's been some debate about whether this -1
  // should be here.. but for now I think it should..)
  free((void*)(p-1));
}

void free_f(fseqp_void x)
{
  // Wes had a comment here basically pointing out the unsoundness of
  // believing the program's use of free(); as I've (now) noted above,
  // these wrappers are effectively no-ops in the "official" CCured
  free(x._p);
}
void free_F(fseqp_void x) { free_f(x); }

// sm: I find it strange that we have free_q but not malloc_q ...
void free_q(seqp_void x)
{
  if (!x._b || x._b != x._p) {
    FP_FAIL(FAIL_FREE);
  }
  free(x._b);
}
void free_Q(seqp_void x) { free_q(x); }


// -------------------- calloc ----------------------
fatp_void calloc_w(unsigned int count, unsigned int osize)
{
  // the unchecked multiply here is ok, because our bookkeeping info
  // only ever stores the product, so if that ends up smaller than
  // what the programmer thought we'll catch it later in a bounds check
  unsigned size = ALIGNBYTES(osize * count);

  // let's just call malloc; it simplifies the code and fixes a bug I noticed
  // (holdover from the two-bits tags days)
  fatp_void res = malloc_w(size);

  // ccured_malloc clears the area, but in case we're not using the gc..
  // (malloc_w set length and zeroed tags already; we just zero data)
  memset(res._p, 0, size);

  return res;
}

// sm: why isn't this called calloc_f?
fseqp_void calloc_fseq_f(unsigned int count,
                        unsigned int osize)
{
  unsigned size = ALIGNBYTES(osize * count);
  fseqp_void res = malloc_fseq_f(size);
  memset(res._p, 0, size);
  return res;
}


// sm: where did this **bizarre** nonsense come from?
void *calloc_rbnode(unsigned int nrelem, unsigned int osize)
{
  char *res = (char*)calloc(nrelem, osize);
  char *where = res + 5 * 4;
  int i;
  //  printf("Allocating %d RBNode's, each of size %d (24 + %d)\n",
  //         nrelem, osize, osize - 24);
  for(i=0; i<(int)nrelem; i++, where += osize) {
    *((int*)where) = osize - 5 * 4;
  }
  return (void*)res;
}


// ------------------------- realloc ------------------------
fatp_void realloc_ww(fatp_void orig, unsigned int newsz) 
{
  int  nrWords    = BYTESTOWORDS(newsz);
  int  nrTagWords = CHECK_NRTAGWORDS(nrWords);            // 0 if NO_TAGS #defined
  int  newSize    = ((1 + nrWords + nrTagWords) << 2);    // length/data/tags
  int  origNrWords;
  //U32  *origbase  = (U32*)orig._b;
  U32  *alloc;

  if (orig._b != orig._p) {
    FP_FAIL(FAIL_REALLOC);
  }

  if (!orig._p) {        // NULL argument. A straight malloc
    return malloc_w(newsz);
  }

  if (newsz == 0) {      // A free operation. Pass it along.
    free_w(orig);
    orig._p = NULL;
    orig._b = NULL;
    return orig;
  }

  // The old number of words
  origNrWords = CHECK_FETCHLENGTH(orig._p, orig._b);
  if(origNrWords == nrWords) {
    return orig; // No change
  }

  if (nrWords < origNrWords) {
    // Shrinking. Move the tags to lower addresses before we realloc.
    // copy left to right for proper overlapping behavior
    CHECK_COPYTAGSFORW(orig._b, origNrWords, (char*)orig._b,
                       orig._b, nrWords, (char*)orig._b,
                       nrWords << 2);

    // Now realloc. The (old) length, data and the surviving tags will be copied
    alloc = realloc((U32*)orig._b - 1, newSize);
    if (!alloc) {
      FP_FAIL(FAIL_REALLOC_SHRINK);
    }
  }
  else {
    // Growing. Realloc first and then move the tags to higher addresses
    alloc = realloc((U32*)orig._b - 1, newSize);
    if (!alloc) {
      FP_FAIL(FAIL_REALLOC_GROW);
    }
    // Move the tags from last to first for proper overlapping behavior
    CHECK_COPYTAGSBACK(alloc + 1, origNrWords, (char*)(alloc + 1),
                       alloc + 1, nrWords, (char*)(alloc + 1),
                       origNrWords << 2);

    // sm: bugfix: need to zero out the cleared space, otherwise the tags
    // left behind lead to bad stuff; alternatively, clearing the tags for
    // the left-behind memory should also be adequate
    memset(alloc+1+origNrWords, 0, (nrWords-origNrWords) * 4);
  }

  // Set the new length
  *alloc = nrWords;
  alloc ++;  // Skip the length
  orig._p = (void*)alloc;
  orig._b = (void*)alloc;
  return orig;
}


// sm: our current implementation actually can't hit this, because we don't
// treat realloc polymorphically; this was from experimenting with James'
// options to do inference unsoundly
fseqp_void realloc_ff(fseqp_void ptr, size_t newsize)
{
  // I can't be sure 'ptr' really is the start of a memory block, so I'll
  // play it safe; GC should clean up after me
  fseqp_void ret;

  size_t oldsize = (int)ptr._e - (int)ptr._p;
  if (oldsize >= newsize) {
    // just give them back the same pointer, and don't release any memory
    ret._p = ptr._p;
    ret._e = (void*)( (int)(ptr._p) + newsize );
    return ret;
  }

  // allocate some new space
  ret._p = malloc(newsize);
  if (!ret._p) {
    FP_FAIL(FAIL_MALLOC);
  }
  ret._e = (void*)( (int)(ret._p) + newsize );

  // copy the original data into this new buffer
  memcpy(ret._p, ptr._p, oldsize);

  // zero out the remainder
  memset((void*)( (int)(ret._p) + oldsize ), 0, newsize-oldsize);

  // finally give back this pointer
  return ret;
}


// sm: I don't know why I wrote the above wrapper to be so complicated,
// when I think this simple one will do ..
fseqp_char realloc_fs(char *ptr, size_t newsize)
{
  fseqp_char ret;
  ret._p = ccured_realloc(ptr, newsize);
  ret._e = ret._p + newsize;
  return ret;
}

// weimer, copied from scott's realloc_fs
seqp_void realloc_qq(seqp_void ptr, size_t newsize)
{
  seqp_void ret;

  ret._p = ccured_realloc(ptr._p,newsize);
  ret._e = (char *)ret._p + newsize;
  ret._b = ret._p; 

  return ret;
}

// ------------------------ fopen ------------------------
fatp_FILE fopen_www(fatp_char1 fname, fatp_char1 mode)
{
  fatp_FILE res;
  res._p = fopen(fname._p, mode._p);

  // this little trick helps ensure the program treats FILE* opaquely
  res._b = (void*)0;       // We should not be reading from here !!!

  return res;
}

fatp_FILE fopen_wws(fatp_char1 fname, char * mode)
{
  fatp_FILE res;
  res._p = fopen(fname._p, mode);

  // this little trick helps ensure the program treats FILE* opaquely
  res._b = (void*)0;       // We should not be reading from here !!!

  return res;
}



fatp_FILE fopen_wss(const char* fname, const  char * mode)
{
  fatp_FILE res;
  res._p = fopen(fname, mode);
  res._b = (void*)0;
  return res;
}


// --------------- accesses to stdin/stdout/stderr ------------------
 //#ifdef _MSVC
 //// A major hack to intercept accesses to _iob on MSVC
 //void * get_iob(int n) {
 //  return (void*)& _iob[n];
 //}
 //fatp_void get_iob_w(int n) {
 //  fatp_void res;
 //  res._p = (void*)& _iob[n];
 //  res._b = (void*)0;
 //  return res;
 //}
 //

// all uses of stdin/stdout/stderr in the program are translated
// into get_stdin()/get_stdout()/get_stderr()
fatp_void get_stdin_w()
{
  fatp_void res;
  res._p = (void*)stdin;
  res._b = 0;
  return res;
}
fatp_void get_stdout_w()
{
  fatp_void res;
  res._p = (void*)stdout;
  res._b = 0;
  return res;
}
fatp_void get_stderr_w()
{
  fatp_void res;
  res._p = (void*)stderr;
  res._b = 0;
  return res;
}


FILE *get_stdin() {
  return stdin;
}
FILE *get_stdout() {
  return stdout;
}
FILE *get_stderr() {
  return stderr;
}


// --------------- misc FILE* functions ----------------
// sm: UNSOUND: I think these are a security hole for a malicious
// program: we let the programmer pass any value at all to fflush
// (etc.), so if the programmer *does* know how libc behaves and what
// the structure of FILE* is, a carefully-crafted bogus FILE could
// break memory safety

int fflush_w(fatp_FILE f)
{
  return fflush(f._p);
}

int fclose_w(fatp_FILE f)
{
  return fclose(f._p);
}

#ifdef _GNUCC
  /* Appears in putc for GCC */
  #ifdef __CYGWIN__
    int __swbuf_w(int c, fatp_void fl) {
      return __swbuf(c, (FILE*)fl._p);
    }
    /* Appears in getc for GCC */
    int __srget_w(fatp_void fl) {
      return __srget((FILE*)fl._p);
    }
  #endif // __CYGWIN__

  // sm: internally, my putc uses _IO_putc
  // see /usr/include/features.h to find glibc version numbers
  #if defined(__GLIBC__) && __GLIBC__ == 2
    int _IO_putc_w(int c, fatp_void fl) {
      return _IO_putc(c, (FILE*)fl._p);
    }

    int _IO_getc_w(fatp_void fl) {
      return _IO_getc((FILE*)fl._p);
    }
  #endif
#endif // _GNUCC

int putc_w(int c, fatp_void fl) {
  return putc(c, (FILE*)fl._p);
}

int getc_w(fatp_void fl) {
  return getc((FILE*)fl._p);
}

// sm: similar (no error checking..) wrappers for scott/putc
int fputc_w(int c, fatp_void fl) {
  return fputc(c, (FILE*)fl._p);
}

int fgetc_w(fatp_void fl) {
  return fgetc((FILE*)fl._p);
}

int puts_w(fatp_char s) {
  return puts((char*)s._p);
}

int fileno_w(fatp_FILE f) {
  return fileno(f._p);
}

int ferror_w(fatp_FILE fp) {
  return ferror(fp._p);
}

int fseek_w(fatp_FILE fp, long offset, int whence)
{
  return fseek(fp._p, offset, whence);
}

void clearerr_w(fatp_FILE fp)
{
  clearerr(fp._p);
}


// ---------------------- fread ------------------------
size_t fread_ws(fatp_char buff, size_t size, size_t count, FILE *fl)
{
  size_t requested = mult_u32(size, count);
  size_t res;
  fatp_write_atleast(buff, requested);
  res = fread(buff._p, size, count, fl);
  return res;
}

size_t fread_ww(fatp_char buff, size_t size, size_t count, fatp_FILE fl)
{
  return fread_ws(buff, size, count, fl._p);
}


size_t fread_fs(fseqp_char buff, size_t size, size_t count, FILE *fl)
{
  size_t requested = mult_u32(size, count);

  CHECK_NULL(fl);
  CHECK_NULL(buff._p);      // sm: I think this is redundant
  CHECK_UBOUND(buff._e, (void*)buff._p, requested);
  return fread(buff._p, size, count, fl);
}
size_t fread_Fs(fseqp_char buff, size_t size, size_t count, FILE *fl) {
  return fread_fs(buff, size, count, fl);
}


// ------------------------- fwrite -----------------------
size_t fwrite_ws(fatp_char buff, size_t size, size_t count, FILE *fl)
{
  size_t requested = mult_u32(size, count);
  size_t res;
  fatp_verify_atleast(buff, requested);
  res = fwrite(buff._p, size, count, fl);
  return res;
}

size_t fwrite_ww(fatp_char buff, size_t size, size_t count, fatp_FILE fl)
{
  return fwrite_ws(buff, size, count, fl._p);
}

size_t fwrite_qs(seqp_char buff, size_t size, size_t count, FILE *fl)
{
  size_t requested = mult_u32(size, count);

  CHECK_NULL(fl);
  CHECK_BOUNDS(buff._b, buff._e, (void*)buff._p, requested);
  return fwrite(buff._p, size, count, fl);
}
size_t fwrite_Qs(seqp_char buff, size_t size, size_t count, FILE *fl) {
  return fwrite_qs(buff, size, count, fl);
}

size_t fwrite_fs(fseqp_char buff, size_t size, size_t count, FILE *fl)
{
  size_t requested = mult_u32(size, count);

  CHECK_NULL(fl);
  CHECK_NULL(buff._p);
  CHECK_UBOUND(buff._e, (void*)buff._p, requested);
  return fwrite(buff._p, size, count, fl);
}
size_t fwrite_Fs(fseqp_char buff, size_t size, size_t count, FILE *fl) {
  return fwrite_fs(buff, size, count, fl);
}


// -------------------------- fgets -----------------------
fatp_char fgets_wws(fatp_char buffer, int size, FILE *stream)
{
  fatp_char res;
  fatp_write_atleast(buffer, size);
  res._p = fgets(buffer._p, size, stream);
  res._b = buffer._b;
  return res;
}

/* fgets and gets have different behavior concerning \n. Since we want gets
 * to use fgets, we write a separate version of fgets which replaces the \n
 * with \0, as gets would do */
fatp_char fgets_nonewline_wws(fatp_char buffer, int size, FILE *stream)
{
  fatp_char res = fgets_wws(buffer, size, stream);

  char *tmp = res._p;
  while (*tmp != '\n' && *tmp != '\0') tmp++;     // find newline or nul
  *tmp ='\0';

  return res;
}

/* It is unsound to use gets, so we turn it into a fgets */
fatp_char gets_ww(fatp_char buffer)
{
  return fgets_nonewline_wws(buffer, fatp_length(buffer), stdin);
}

fatp_char fgets_www(fatp_char buffer, int size, fatp_FILE stream)
{
  return fgets_wws(buffer, size, stream._p);
}


fseqp_char fgets_ffw(fseqp_char buffer, int len, fatp_FILE fp)
{
  fseqp_write_atleast(buffer, len);

  if (!fgets(buffer._p, len, fp._p)) {
    fseqp_char res;
    res._p = 0;
    res._e = 0;
    return res;
  }
  else {
    return buffer;
  }
}
fseqp_char fgets_FFw(fseqp_char buffer, int len, fatp_FILE fp) {
  return fgets_ffw(buffer, len, fp);
}

// another version for seq
seqp_char fgets_qqw(seqp_char buffer, int len, fatp_FILE fp)
{
  seqp_write_atleast(buffer, len);

  if (!fgets(buffer._p, len, fp._p)) {
    seqp_char res;
    res._b = 0;
    res._p = 0;
    res._e = 0;
    return res;
  }
  else {
    return buffer;
  }
}


// -------------------- read ----------------------
int read_w(int fid, fatp_char buff, unsigned int size)
{
  fatp_write_atleast(buff, size);
  return read(fid, buff._p, size);
}

int read_f(int fid, fseqp_char buff, unsigned int size)
{
  fseqp_write_atleast(buff, size);
  return read(fid, buff._p, size);
}
int read_F(int fid, fseqp_char buff, unsigned int size) {
  return read_f(fid, buff, size);
}

int read_q(int fid, seqp_char buff, unsigned int size)
{
  seqp_write_atleast(buff, size);
  return read(fid, buff._p, size);
}
int read_Q(int fid, seqp_char buff, unsigned int size) {
  return read_q(fid, buff, size);
}


// --------------------- write -------------------
int write_w(int fid, fatp_char buff, unsigned int size)
{
  fatp_verify_atleast(buff, size);
  return write(fid, buff._p, size);
}

int write_f(int fid, fseqp_char buff, unsigned int size)
{
  fseqp_verify_atleast(buff, size);
  return write(fid, buff._p, size);
}
int write_F(int fid, fseqp_char buff, unsigned int size) {
  return write_f(fid, buff, size);
}

int write_q(int fid, seqp_char buff, unsigned int size)
{
  seqp_verify_atleast(buff, size);
  return write(fid, buff._p, size);
}
int write_Q(int fid, seqp_char buff, unsigned int size) {
  return write_q(fid, buff, size);
}


// ---------------- misc file-descriptor funcs -----------------
int fstat_w(int fid, fatp_char buff)
{
  fatp_write_atleast(buff, sizeof(struct _stat));
  return fstat(fid, (struct _stat*)buff._p);
}

int stat_ww(fatp_char path, fatp_char buff)
{
  fatp_write_atleast(buff, sizeof(struct _stat));
  fatp_verify_nul(path);
  return stat(path._p, (struct _stat*)buff._p);
}

int open_w(fatp_char path, int oflag, int pmode)
{
  fatp_verify_nul(path);
  return open(path._p, oflag, pmode);
}

int unlink_w(fatp_char path)
{
  fatp_verify_nul(path);
  return unlink(path._p);
}


// ----------------- setjmp/longjmp --------------------
// allowing the programmer to call longjmp is UNSOUND because
// the jmp_buf could be filled with arbitrary trash, which
// then gets copied into the process registers; for the safe/seq
// world one possible solution is to use the patcher to make
// the jmp_buf type opaque, so the programmer can't touch it;
// but then there's still the problem of calling longjmp after
// the setjmp's caller has returned, which I don't know how to
// solve

// I want to cast, but jmp_buf is defined as an array type..
struct jmp_buf_wrapper {
  jmp_buf buf;
};

// given a fat pointer, make it look like a jmp_buf
#define AS_JMP_BUF(ptr) ((struct jmp_buf_wrapper*)ptr._p)->buf

#if 0   // no wrappers for setjmp
  // TODO: you can't wrap setjmp!  once the wrapper returns, that
  // means setjmp's caller has returned, so the context to which
  // longjmp wants to return is gone!

  int setjmp_w(fatp_char jbuf)
  {
    fatp_write_atleast(jbuf, sizeof(jmp_buf));
    return setjmp(AS_JMP_BUF(jbuf));
  }

  // (glibc-ism) the leading underscore means it does not save
  // the signal context
  int _setjmp_w(fatp_char jbuf)
  {
    return setjmp_w(jbuf);
  }

  #if defined(_GNUCC) && !defined(__CYGWIN__)
    // sm: isn't this wrong?  I mean, sigsetjmp takes a buffer
    // which is larger than a simple jmp_buf, right?
    // update: no, at least on glibc 2, the two structures are the same,
    // and it contains a flag inside it to tell whether the signal
    // mask was saved
    int __sigsetjmp_w(fatp_char jbuf, int __savemask)
    {
      fatp_write_atleast(jbuf, sizeof(jmp_buf));
      return __sigsetjmp(AS_JMP_BUF(jbuf), __savemask);
    }
  #endif

  // sm 4/01/02: in the seq/fseq case, box.ml inserts the checks
  // directly (!), so there seems to be no way to provoke a need for
  // setjmp_q
  int setjmp_q(seqp_char jbuf)
  {
    seqp_verify_atleast(jbuf, sizeof(jmp_buf));
    return setjmp(AS_JMP_BUF(jbuf));
  }

  int _setjmp_q(seqp_char jbuf)
  {
    return setjmp_q(jbuf);
  }

  #if defined(_GNUCC) && !defined(__CYGWIN__)
    int __sigsetjmp_q(seqp_char jbuf, int __savemask)
    {
      seqp_verify_atleast(jbuf, sizeof(jmp_buf));
      return __sigsetjmp(AS_JMP_BUF(jbuf), __savemask);
    }
  #endif
#endif // 0


void longjmp_w(fatp_char jbuf, int res)
{
  fatp_verify_atleast(jbuf, sizeof(jmp_buf));
  longjmp(AS_JMP_BUF(jbuf), res);
}
void longjmp_q(seqp_char jbuf, int res)
{
  seqp_verify_atleast(jbuf, sizeof(jmp_buf));
  longjmp(AS_JMP_BUF(jbuf), res);
}


// ---------------- string-to-number conversions ---------------
double atof_w(fatp_char str)
{
  fatp_verify_nul(str);
  return atof(str._p);
}

int atoi_w(fatp_char str)
{
  fatp_verify_nul(str);
  return atoi(str._p);
}

long atol_w(fatp_char str)
{
  fatp_verify_nul(str);
  return atol(str._p);
}


// --------------------- qsort -----------------
#ifndef NO_TAGS
// return 1 if:
//   - 'base' is aligned to a word boundary
//   - 'size' is aligned to a word boundary
//   - the tags corresponding to each element of size 'size' are
//     identical to every other element, in the range 'base' up to
//     'base + nelts*size'
// otherwise return 0
int qsort_regular_tags(fatp_char base, size_t nelts, size_t size)
{
  unsigned word, elt;     // iterators for loops below
  unsigned areaSize = CHECK_FETCHLENGTH(base._b, base._p);

  // pointer to the first word of the first element
  U32 *firstElt = (U32*)base._p;

  // do this check after declaring the variables...
  if (((int)base._p & 3) || (size & 3)) {
    return 0;   // unaligned
  }

  // now size is in WORDS (was in bytes)
  size >>= 2;

  // iterate over the words in an element
  for (word=0; word<size; word++) {
    // what is the first element's tag for this word?
    unsigned firstTag = CHECK_FETCHTAGBIT(base._b, areaSize, firstElt);

    // iterate over all elements beyond the first
    U32 *otherElt = firstElt + size;
    for (elt=1; elt<nelts; elt++) {
      // is the tag for this elt/word equal to the corresponding tag
      // for the first elt?
      if (firstTag != CHECK_FETCHTAGBIT(base._b, areaSize, otherElt)) {
        // they differ; we fail
        return 0;
      }

      // advance to next element
      otherElt += size;
    }

    // advance to next word in the first element
    firstElt++;
  }

  // every element matched the first
  return 1;
}
#else
  #define qsort_regular_tags(b,n,s) 1
#endif


void qsort_zero_tags(fatp_char base, size_t nelts, size_t size)
{
  // zero the tags, *unless* they happen to be such that sorting
  // cannot change them
  if (qsort_regular_tags(base, nelts, size)) {
    // every element's tags are the same!  sorting cannot affect them
    fatp_verify_atleast(base, mult_u32(nelts, size));
  }
  else {
    // tags not the same, conservatively clear them
    fatp_write_atleast(base, mult_u32(nelts, size));
  }
}


// sm: UNSOUND: this function is bad, right?  it lets the user
// derive a safe pointer from a wild pointer, right?  (our
// model is incomplete -- this should be easy to fix..)
void qsort_wsss(fatp_char base, size_t nmemb,
                size_t size,
                int (*compar)(const void *, const void *))
{
  qsort_zero_tags(base, nmemb, size);
  qsort(base._p, nmemb, size, compar);
}

/* weimer : Wed Dec 12 13:12:34 PST 2001 */
void qsort_fsss(fseqp_char base,
                unsigned int nmemb,
                unsigned int size,
                int (*compar)(const void *arg0, const void *arg1))
{
  fseqp_verify_atleast(base, mult_u32(nmemb, size));
  qsort(base._p, nmemb, size, compar);
}


// ------- qsort_wsww ------
// I need these to thunk back to user's sort function
void *qsort_wsww_base;
int (*qsort_wsww_compare)(fatp_void, fatp_void);

int qsort_wsww_thunk(void const *left, void const *right)
{
  // map the 'left' and 'right' lean pointers to the
  // fat pointers we need
  fatp_void fatleft, fatright;
  fatleft._p = (void*)left;      // cast away constness since we have no runtime enforcement of it..
  fatleft._b = qsort_wsww_base;
  fatright._p = (void*)right;
  fatright._b = qsort_wsww_base;

  // and call the user-supplied sorting function, which
  // expects fat pointers
  return qsort_wsww_compare(fatleft, fatright);
}

// this function's test case is test/small2/qsort.c
void qsort_wsww(fatp_char base, size_t nmemb,
                size_t size,
                int (*compare)(fatp_void, fatp_void))
{
  qsort_zero_tags(base, nmemb, size);

  // save the pertinent values
  qsort_wsww_base = base._b;
  qsort_wsww_compare = compare;

  qsort(base._p, nmemb, size, qsort_wsww_thunk);
}


// -------- qsort_fsff -----------
// thunking variables
char *qsort_fsff_end;
int (*qsort_fsff_compare)(fseqp_char, fseqp_char);

int qsort_fsff_thunk(void const *left, void const *right)
{
  // wrap lean args as fseqp's
  fseqp_char fatleft, fatright;
  fatleft._p = (char*)left;
  fatleft._e = qsort_fsff_end;
  fatright._p = (char*)right;
  fatright._e = qsort_fsff_end;

  // call user's comparison function
  return qsort_fsff_compare(fatleft, fatright);
}

void qsort_fsff(fseqp_char base, size_t nmemb,
                size_t size,
                int (*compare)(fseqp_char, fseqp_char))
{
  fseqp_verify_atleast(base, mult_u32(nmemb, size));

  // save values for thunk
  qsort_fsff_end = base._e;
  qsort_fsff_compare = compare;

  qsort(base._p, nmemb, size, qsort_fsff_thunk);
}


// ---------------- str* functions -------------------
// throughout this section, I make the simplifying assumption that
// there are no pointers stored in regions pointed-to by the
// arguments; it would be a really strange program which had pointers
// in these areas but somehow knew their representations did not
// contain 0 bytes

//matth:  leave these here for INFERBOX=wild test cases
int strncmp_ww(fatp_char s1,
               fatp_char s2,  unsigned int size)
{
  // sm: we used to have some code here which worried about
  // nul-termination, but that's irrelevant since strncmp
  // itself agrees not to look beyond 'size' bytes under
  // any circumstances

  fatp_verify_atleast(s1, size);
  fatp_verify_atleast(s2, size);
  return strncmp(s1._p, s2._p, size);
}

int strcmp_ww(fatp_char s1, fatp_char s2)
{
  // sm: then this code called strncmp_ww with a large length..
  // that seems to not be what we want..

  fatp_verify_nul(s1);
  fatp_verify_nul(s2);
  return strcmp(s1._p, s2._p);
}
/*
char * strcpy_sww(fatp_char dest, fatp_char src)
{
  // the old code copied the tags; I'm going to assume if you're
  // using strcpy there aren't any pointers in this area, and
  // just clear the tags (if I'm wrong, pull from CVS around 2/01/02
  // to get the tag-copying version)
  fatp_verify_nul(src);
  fatp_write_atleast(dest, strlen(src._p)+1);
  strcpy(dest._p, src._p);
  return NULL;
}
   
char * strcpy_sqq(seqp_char dest, seqp_char src) // weimer
{
  // the old code copied the tags; I'm going to assume if you're
  // using strcpy there aren't any pinters in this area, and
  // just clear the tags (if I'm wrong, pull from CVS around 2/01/02
  // to get the tag-copying version)
  verify_nul(src._p, src._e - src._p);
  CHECK_UBOUND(dest._e, (void*)dest._p, strlen(src._p)+1); 
  strcpy(dest._p, src._p);
  return NULL;
}

fseqp_char strcpy_fqq(seqp_char dest, seqp_char src) // weimer
{
  fseqp_char result;
  // the old code copied the tags; I'm going to assume if you're
  // using strcpy there aren't any pointers in this area, and
  // just clear the tags (if I'm wrong, pull from CVS around 2/01/02
  // to get the tag-copying version)
  verify_nul(src._p, src._e - src._p);
  CHECK_UBOUND(dest._e, (void*)dest._p, strlen(src._p)+1);
  strcpy(dest._p, src._p);
  result._p = dest._p;
  result._e = dest._e; 
  return result;
}

fseqp_char strcpy_fff(fseqp_char dest, fseqp_char src) // weimer
{
  // the old code copied the tags; I'm going to assume if you're
  // using strcpy there aren't any pointers in this area, and
  // just clear the tags (if I'm wrong, pull from CVS around 2/01/02
  // to get the tag-copying version)
  verify_nul(src._p, src._e - src._p);
  CHECK_UBOUND(dest._e, (void*)dest._p, strlen(src._p)+1);
  strcpy(dest._p, src._p);
  return dest;
}
fseqp_char strcpy_FFF(fseqp_char dest, fseqp_char src) {
  return strcpy_fff(dest, src); 
}

char * strcpy_sff(fseqp_char dest, fseqp_char src) // weimer
{
  strcpy_fff(dest, src);
  return NULL;
}
char * strcpy_sFF(fseqp_char dest, fseqp_char src) {
  return strcpy_sff(dest, src);
}

fatp_char strcpy_www(fatp_char dest, fatp_char src)
{
  // the old code copied the tags; I'm going to assume if you're
  // using strcpy there aren't any pointers in this area, and
  // just clear the tags (if I'm wrong, pull from CVS around 2/01/02
  // to get the tag-copying version)
  fatp_verify_nul(src);
  fatp_write_atleast(dest, strlen(src._p)+1);
  strcpy(dest._p, src._p);
  return dest;
}
*/
//matth:  leave these here for INFERBOX=wild test cases
unsigned int strlen_w(fatp_char s)
{
  fatp_verify_nul(s);
  return strlen(s._p);
}

fatp_char strchr_ww(fatp_char s, int chr)
{
  fatp_char res;
  fatp_verify_nul(s);
  res._b = s._b;
  res._p = strchr(s._p, chr);
  return res;
}
/*
// sm: hmm.. must be missing a model, this function's signature
// doesn't exactly make sense.. but I doubt that makes it unsound
fseqp_char strrchr_fs(char *str, int chr)
{
  fseqp_char res;
  res._p = strrchr(str, chr);
  res._e = res._p? str + strlen(str) : NULL;     // bugfix -- unsound to have null p but not null e!
  return res;
}
fseqp_char strrchr_Fs(char *str, int chr) {
  return strrchr_fs(str, chr);
}

// sm: for m88k
fseqp_char strchr_ff(fseqp_char str, int chr)
{
  fseqp_char res;
  fseqp_verify_nul(str);

  res._p = strchr(str._p, chr);
  res._e = res._p? str._e : NULL;
  return res;
}
fseqp_char strchr_FF(fseqp_char str, int chr) {
  return strchr_ff(str, chr);
}

// sm: for m88k, now it wants to use sequence pointers..
seqp_char strchr_qq(seqp_char str, int chr)
{
  seqp_char res;
  seqp_verify_nul(str);

  res._p = strchr(str._p, chr);
  if (res._p) {
    res._b = str._b;
    res._e = str._e;
  }
  else {
    res._b = NULL;
    res._e = NULL;
  }
  return res;
}
seqp_char strchr_QQ(seqp_char str, int chr) {
  return strchr_qq(str, chr);
}

fatp_char strcat_www(fatp_char dest, fatp_char src)
{
  fatp_verify_nul(src);
  fatp_verify_nul(dest);
  fatp_write_atleast(dest, strlen(dest._p) + strlen(src._p) + 1);

  strcat(dest._p, src._p);
  return dest;
}

char * strcat_sqq(seqp_char dest, seqp_char src) // weimer
{
  // the old code copied the tags; I'm going to assume if you're
  // using strcpy there aren't any pointers in this area, and
  // just clear the tags (if I'm wrong, pull from CVS around 2/01/02
  // to get the tag-copying version)
  verify_nul(src._p, src._e - src._p);
  CHECK_UBOUND(dest._e, (void*)dest._p, strlen(dest._p) + strlen(src._p)+1);
  strcat(dest._p, src._p);
  return NULL;
}

char * strcat_sff(fseqp_char dest, fseqp_char src) // weimer
{
  // the old code copied the tags; I'm going to assume if you're
  // using strcpy there aren't any pointers in this area, and
  // just clear the tags (if I'm wrong, pull from CVS around 2/01/02
  // to get the tag-copying version)
  verify_nul(src._p, src._e - src._p);
  CHECK_UBOUND(dest._e, (void*)dest._p, strlen(dest._p) + strlen(src._p)+1);
  strcat(dest._p, src._p);
  return NULL;
}

// caller will think of this as returning fatp_char, but
// it's more convenient here to call it void b/c malloc_w
// returns fatp_void
fatp_void strdup_ww(fatp_char src)
{
  int len = fatp_verify_nul(src);
  fatp_void res;
  res = malloc_w(len + 1);
  strcpy((char*)res._p, src._p);
  return res;
}

fseqp_char strdup_ff(fseqp_char src)
{
  int len = fseqp_verify_nul(src);

  // allocate space for the result
  fseqp_char ret;
  ret._p = malloc(len + 1);       // include space for nullterm
  ret._e = ret._p + (len + 1);    // end is just beyond last byte alloc'd
  memcpy(ret._p, src._p, len+1);  // copy the bytes!

  return ret;
}
fseqp_char strdup_FF(fseqp_char src) {
  return strdup_ff(src);
}
fseqp_char strdup_Ff(fseqp_char src) {
  return strdup_ff(src);
}

fseqp_char strdup_fq(seqp_char src)
{
  int len = seqp_verify_nul(src);

  // allocate space for the result
  fseqp_char ret;
  ret._p = malloc(len + 1);       // include space for nullterm
  ret._e = ret._p + (len + 1);    // end is just beyond last byte alloc'd
  memcpy(ret._p, src._p, len+1);  // copy the bytes!

  return ret;
}

fatp_char strdup_ws(char *x) {
  return fattenstring(x);
}
*/

fatp_char fattenstring(char *s)
{
  fatp_void res = malloc_w(1 + strlen(s));
  strcpy(res._p, s);
  return *(fatp_char*)&res;
}


// sm: how can this work?  shouldn't our printf be checking that
// the passed base fields are valid?  possibly UNSOUND somewhere
fatp_char strerror_w(int errno)
{
  fatp_char res;
  res._p = strerror(errno);
  res._b = (void*)0;  // !!! No tags or length in there !
  return res;
}


// ----------------------- mem* functions ---------------------
int memcmp_ww(fatp_char dest, fatp_char src, unsigned int size)
{
  fatp_verify_atleast(dest, size);
  fatp_verify_atleast(src, size);
  return memcmp(dest._p, src._p, size);
}

#ifdef _MSVC
  void* memcpy(void *, void const *, unsigned int);
#endif

fatp_char memcpy_www(fatp_char dest, fatp_char src, unsigned int size)
{
  fatp_verify_atleast(dest, size);
  fatp_verify_atleast(src, size);
  fatp_copy_tags(dest, src, size);
  memcpy(dest._p, src._p, size);
  return dest;
}
/*
fseqp_char memcpy_fff(fseqp_char dest, fseqp_char src, unsigned int size)
{
  fseqp_verify_atleast(dest, size);
  fseqp_verify_atleast(src, size);
  memcpy(dest._p, src._p, size);
  return dest;
}

void *memcpy_sff(fseqp_char dest, fseqp_char src, unsigned int size)
{
  fseqp_verify_atleast(dest, size);
  fseqp_verify_atleast(src, size);
  memcpy(dest._p, src._p, size);
  return NULL;    // presumably not used if inferred type was safe
}

void *memmove_sff(fseqp_char dest, fseqp_char src, unsigned int size)
{
  fseqp_verify_atleast(dest, size);
  fseqp_verify_atleast(src, size);
  memmove(dest._p, src._p, size);
  return NULL;
}

seqp_char memcpy_qqq(seqp_char dest, seqp_char src, unsigned int size)
{
  seqp_verify_atleast(dest, size);
  seqp_verify_atleast(src, size);
  memcpy(dest._p, src._p, size);
  return dest;
}

fseqp_char memcpy_fqq(seqp_char dest, seqp_char src, unsigned int size)
{
  fseqp_char res;
  seqp_verify_atleast(dest, size);
  seqp_verify_atleast(src, size);
  memcpy(dest._p, src._p, size);
  res._p = dest._p;     // the checks above ensure that _p >= _b,
  res._e = dest._e;     // so (p,e) is a valid fseq
  return res;
}
*/
fatp_char memmove_www(fatp_char dest, fatp_char src, unsigned int size) 
{
  fatp_verify_atleast(dest, size);
  fatp_verify_atleast(src, size);
  fatp_copy_tags(dest, src, size);
  memmove(dest._p, src._p, size);
  return dest;
}
/*
seqp_char memmove_qqq(seqp_char dest, seqp_char src, unsigned int size) 
{                                 
  // sm: this comment:
  //   "Ought to check that the alignment is correct ! But it should be !"
  // I do not understand.  what depends on alignment here?

  seqp_verify_atleast(dest, size);
  seqp_verify_atleast(src, size);
  memmove(dest._p, src._p, size);
  return dest;
}

fseqp_char memmove_fqq(seqp_char dest, seqp_char src, unsigned int size)
{
  fseqp_char res;
  seqp_verify_atleast(dest, size);
  seqp_verify_atleast(src, size);
  memmove(dest._p, src._p, size);
  res._p = dest._p;
  res._e = dest._e;
  return res;
}

fseqp_char memmove_fff(fseqp_char dest, fseqp_char src, unsigned int size)
{
  fseqp_verify_atleast(dest, size);
  fseqp_verify_atleast(src, size);
  memmove(dest._p, src._p, size);
  return dest;
}
*/

fatp_char memset_ww(fatp_char buffer, int c, size_t size) 
{
  fatp_write_atleast(buffer, size);
  memset(buffer._p, c, size);
  return buffer;
}
/*
seqp_char memset_qq(seqp_char buffer, int c, size_t size)
{
  seqp_verify_atleast(buffer, size);
  memset(buffer._p, c, size);
  return buffer;
}

fseqp_char memset_fq(seqp_char buffer, int c, size_t size)
{
  fseqp_char res;
  seqp_verify_atleast(buffer, size);
  memset(buffer._p, c, size);
  res._p = buffer._p;
  res._e = buffer._e;
  return res;
}

char* memset_sf(fseqp_char buffer, int c, size_t size)
{
  fseqp_verify_atleast(buffer, size);
  memset(buffer._p, c, size);
  return buffer._p;
}

fseqp_char memset_ff(fseqp_char buffer, int c, size_t size)
{
  fseqp_verify_atleast(buffer, size);
  memset(buffer._p, c, size);
  return buffer;
}

void bzero_q(seqp_char buffer, int n)
{
  seqp_verify_atleast(buffer, n);
  memset(buffer._p, 0, n);
}

void bzero_w(fatp_char buffer, int n)
{
  fatp_write_atleast(buffer, n);
  memset(buffer._p, 0, n);
}

#ifndef _MSVC
  void bzero_f(fseqp_char buffer, size_t n)
  {
    fseqp_write_atleast(buffer, n);
    memset(buffer._p, 0, n);
  }
#endif
*/

// ------------------- time ------------------
time_t time_w(fatp_char timer)
{
  if (timer._p) {   // time accepts a NULL argument
    fatp_write_atleast(timer, sizeof(time_t));
  }
  return time((time_t*)timer._p);
}

#ifdef _GNUCC
  clock_t times_w(fatp_char tms)
  {
    fatp_write_atleast(tms, sizeof(struct tms));
    return times((struct tms*)tms._p);
  }


  // int gettimeofday(struct timeval *tv, struct timezone *tz);
  int gettimeofday_ws(fatp_char tv, struct timezone *tz)
  {
    fatp_write_atleast(tv, sizeof(struct timeval));
    return gettimeofday((struct timeval *)tv._p, tz);
  }
#endif


// ----------------- pointer cast log --------------------
// the idea here is, when a program fails with a non-pointer error,
// it's often not obvious where this non-pointer arose.  so,
// optionally, every time a non-pointer is created, we can log the
// source location where that happened.  then when the pointer failure
// happens, we may be able to report the source line that produced the
// bad value.

// does this mean we've *always* been doing this??
// no, it has to be enabled in box.ml
#define LOG_CASTS

#define MIN_POINTER 0x100000U
#define MAX_POINTER (0xFFFFFFFFU - MIN_POINTER)
#if defined(_MSVC) && defined(_DEBUG)
  fatp_void _CrtSetReportFile_ww(int reportMode, fatp_void f) {
    _CrtSetReportFile(reportMode, f._p);
    return f;
  }
#endif /* _MSVC && _DEBUG */

#ifdef WALK_HEAP
  #include <malloc.h>
  _HEAPINFO hpWalk;
#endif /* WALK_HEAP */

#ifdef LOG_CASTS
  typedef struct {
    unsigned long l;
    unsigned short fileid, lineid;
  } LOG_ENTRY;
  #define LOG_SIZE   (1 << 16)
  LOG_ENTRY logBuffer[LOG_SIZE];
  int logBufferP = 0; // Next free entry
  int logFid = -1;
  #include <fcntl.h>

  // Flush the log to a file
  int flushLog()
  {
    // See if we need to open the log file
    int toWrite = logBufferP * sizeof(LOG_ENTRY);
    if(toWrite <= 0)
      return 0;

    if(logFid < 0) {
      logFid = open("__scalar2pointer",
                    O_CREAT  | O_TRUNC | O_RDWR | O_BINARY,
                    0777);
      if(logFid < 0) {
        FP_FAIL(FAIL_SCALARLOG);
      }
    }
    if(toWrite != write(logFid, & logBuffer[0], toWrite)) {
      FP_FAIL(FAIL_SCALARLOG);
    }
    logBufferP = 0;
    return 0;
  }

  /* Scan the log and apply the given function to all entries. Stop when the
   * log is over or when the match function returns true. Scan log returns
   * true if it terminated because match returned true */
  int scanLog(int (*match)(int, LOG_ENTRY *), int closure) {
    flushLog();
    if(logFid < 0) {
      // Log is empty
      return 0;
    }
    {
      LOG_ENTRY *le = & logBuffer[0];
      int toGo = lseek(logFid, 0, SEEK_END);
      int i = LOG_SIZE;  // Pretend we are at the end of the buffer

      lseek(logFid, 0, SEEK_SET); // Go to start
      while(toGo > 0) {
        if(i >= LOG_SIZE) {
          int toRead = sizeof(logBuffer);
          if(toRead > toGo) toRead = toGo;
          if(toRead != read(logFid, & logBuffer[0], toRead)) {
            FP_FAIL(FAIL_SCALARLOG);
          }
          i = 0;
          le = & logBuffer[i];
        }
        if((*match)(closure, le)) {
          return 1;
        }
        i ++;
        le ++;
        toGo -= sizeof(LOG_ENTRY);
      }
    }
    return 0; // Not found
  }

  int foundOne = 0;
  int printMatches(int tomatch, LOG_ENTRY *le) {
    if(le->l == (unsigned)tomatch) {
      foundOne = 1;
      fprintf(stderr, "\tSame as at cast in file 0x%04x on line %d\n",
              le->fileid, le->lineid);
    }
    return 0; // Want all matches
  }

  // Print the contents of the log
  int printAll(int where, LOG_ENTRY *le) {
    fprintf((FILE*)where, "0x%08lx  0x%04x %d\n",
            le->l, (unsigned)(le->fileid), le->lineid);
    return 0;
  }
#endif /* LOG_CASTS */


void* __scalar2pointer(unsigned long l, int fileid, int lineid) 
{
  #ifdef WALK_HEAP
    void *base = NULL;
  #endif

  #ifdef LOG_CASTS
    if(l >= MIN_POINTER && l <= MAX_POINTER) {
      LOG_ENTRY *le = & logBuffer[logBufferP ++];
      le->l = l;
      le->fileid = fileid;
      le->lineid = lineid;
      // See if we need to dump the log out
      if(logBufferP == LOG_SIZE) {
        flushLog();
      }
    }
  #endif /* LOG_CASTS */

  #if defined(WALK_HEAP) && !defined(NO_TAGS)
    hpWalk._pentry = NULL;
    if(l >= MIN_POINTER && l <= MAX_POINTER) {
      while(_heapwalk(&hpWalk)) {
        if(l >= (unsigned long)hpWalk._pentry &&
           l < (unsigned long)hpWalk._pentry + hpWalk._size &&
           hpWalk._useflag == _USEDENTRY) {
          base = (void*)((char*)hpWalk._pentry + 4);
          goto FoundOne;
        }
      }
    }
    return NULL;

  FoundOne:
    {
      unsigned int len = CHECK_FETCHLENGTH(base);
      TAGADDR tag   = CHECK_FETCHTAGADDR(base, len);
      int  nrTagWords = CHECK_NRTAGWORDS(len);
      if(l >= base &&
         l <= (long)(tag.word - 4) &&
         (1 + len + nrTagWords) << 2 == hpWalk._size) {
        return (void*)base;
      }
    }
  #endif /* WALK_HEAP */

  return NULL;
}


// And a function to scan the logs and find the matching
void searchLog(unsigned long l  FP_FAIL_EXTRA_PARAMS) 
{
  #ifdef LOG_CASTS
    if(l >= MIN_POINTER && l <= MAX_POINTER) {
      fprintf(stderr, "Dereferencing a non-pointer (0x%08lx)\n", l);
      foundOne = 0;
      scanLog(& printMatches, (int)l);
      if(! foundOne) {
        // Dump the entire log to a file
        char *theFileName = "__scalar2pointer.txt";
        FILE *txtLog = fopen(theFileName, "w");
        #ifdef _MSVC
          char fullPath[_MAX_PATH];
          theFileName = _fullpath(fullPath, theFileName, _MAX_PATH);
        #endif
        fprintf(stderr,
                "Cannot find the cast in log. Dumping the entire log to %s\n",
                 theFileName);
        scanLog(& printAll, (int)txtLog);
        fclose(txtLog);
      } else {
        fflush(stderr);
      }
    }
    fp_fail(FAIL_NONPOINTER  FP_FAIL_EXTRA_ARGS);

  #else
    fp_fail(FAIL_NONPOINTER  FP_FAIL_EXTRA_ARGS);
  #endif
}


/* --------------- wrappers for main --------------- */
// when the arguments to main become non-safe (as they almost always
// do if they're used), then we need to glue the loader's call to main
// with the program's now-mangled main (which gets renamed to trueMain)

// an initializer for all our CCured stuff
void safecInit()
{
  if (sizeof(failMessages)/sizeof(failMessages[0]) != NUM_FAIL_CODES) {
    FP_FAIL_STR("NUM_FAIL_CODES doesn't agree with failMessages[]");
  }
}

// the program's original main function, renamed; since it gets called
// differently depending on context, below, we don't specify its params
extern int trueMain();     // can't say "..." b/c wants named argument before

// A wrapper main function for the case where the param types don't change
int _mainWrapper(int argc, char **argv)
{
  safecInit();
  return trueMain(argc, argv);
}


// (**) most of these wrappers copy the argv strings; this isn't just
// because the types sometimes impose representation constraints, but
// because some programs like to put the values in the argv array on
// the heap or globals, but the argv strings usually live on the
// stack, triggering a run-time error


// sm: I don't like the obvious duplication in the code that follows,
// but after thinking about it and staring at it for 15 minutes, I
// still have no good ideas about how to fix it (now if this were
// in C++, we'd have some possibilities..)

typedef struct {
  char * * _p;
  void * _e;
} fseqp_p_char;

int _mainWrapper_fs(int argc, char **argv)
{
  int i;
  fseqp_p_char argv1;
  char **nargv = (char**)malloc((argc+1) * sizeof(char *));

  argv1._p = nargv;
  argv1._e = (void*)(nargv + argc + 1);

  for(i=0;i<argc;i++) {
    nargv[i] = strdup(argv[i]);     // see note (**) above
  }
  nargv[argc] = NULL;
  safecInit();
  return trueMain(argc, argv1);
}


typedef struct {
  char * * _p;
  char * * _b;
  void * _e;
} seqp_p_char;

int _mainWrapper_qs(int argc, char **argv)
{
  int i;
  seqp_p_char argv1;
  char **nargv = (char**)malloc((argc+1) * sizeof(char *));

  argv1._p = nargv;
  argv1._b = nargv;
  argv1._e = (void*)(nargv + argc + 1);

  for(i=0;i<argc;i++) {
    nargv[i] = strdup(argv[i]);     // see note (**) above
  }
  nargv[argc] = NULL;
  safecInit();
  return trueMain(argc, argv1);
}


typedef struct {
  seqp_char *_p;
  void *_e;
} fseqp_seqp_char;

int _mainWrapper_fq(int argc, char **argv)
{
  int i;
  fseqp_seqp_char argv1;
  seqp_char *nargv = (seqp_char*)malloc((1 + argc) * sizeof(seqp_char));

  argv1._p = nargv;
  argv1._e = (void*)(nargv + argc + 1);

  for(i=0; i<argc; i++) {
    char *copy = strdup(argv[i]);
    nargv[i]._p = copy;
    nargv[i]._b = copy;
    nargv[i]._e = copy + strlen(copy) + 1;
  }
  nargv[argc]._p = (void*)0;
  nargv[argc]._b = (void*)0;
  nargv[argc]._e = (void*)0;

  safecInit();
  return trueMain(argc, argv1);
}

typedef struct {
  fseqp_char *_p;
  void *_b;
  void *_e;
} seqp_fseqp_char;

int _mainWrapper_qf(int argc, char **argv)
{
  int i;
  seqp_fseqp_char argv1;
  fseqp_char *nargv = (fseqp_char*)malloc((1 + argc) * sizeof(fseqp_char));

  argv1._p = nargv;
  argv1._b = (void*)nargv;
  argv1._e = (void*)(nargv + argc + 1);

  for(i=0; i<argc; i++) {
    char *copy = strdup(argv[i]);
    nargv[i]._p = copy;
    nargv[i]._e = copy + strlen(copy) + 1;
  }
  nargv[argc]._p = (void*)0;
  nargv[argc]._e = (void*)0;

  safecInit();
  return trueMain(argc, argv1);
}


typedef struct {
  fseqp_char *_p;
  void *_e;
} fseqp_fseqp_char;

int _mainWrapper_ff(int argc, char **argv) {
  int i;
  fseqp_fseqp_char argv1;
  fseqp_char *nargv = (fseqp_char*)malloc((argc + 1) * sizeof(fseqp_char));

  argv1._p = nargv;
  argv1._e = (void*)(nargv + argc + 1);

  for(i=0;i<argc;i++) {
    char *copy = strdup(argv[i]);
    nargv[i]._p = copy;
    nargv[i]._e = copy + strlen(copy) + 1;
  }
  nargv[argc]._p = (void*)0;
  nargv[argc]._e = (void*)0;

  safecInit();
  return trueMain(argc, argv1);
}
int _mainWrapper_fF(int argc, char **argv) {
  return _mainWrapper_ff(argc, argv);
}
int _mainWrapper_qF(int argc, char **argv) {
  return _mainWrapper_qf(argc, argv);
}


int _mainWrapper_w(int argc, char **argv)
{
  // Just copy the input arguments. Leave room for the last NULL
  fatp_void tmp = malloc_w(sizeof(fatp_char) * (argc + 1));
  fatp_fatp_char nargv = *(fatp_fatp_char*)&tmp;
  int i;
  unsigned int argvlen = CHECK_FETCHLENGTH(nargv._p, nargv._b);

  for(i=0; i<argc; i++) {
    fatp_char copy = fattenstring(argv[i]);
    // Use CHECK_WILDPOINTERWRITE to write the proper tags also
    CHECK_WILDPOINTERWRITE(nargv._b, argvlen,
                          (void*) & nargv._p[i],
                          copy._b, copy._p);
    nargv._p[i] = copy;
  }
  nargv._p[argc]._p = (void*)0;
  nargv._p[argc]._b = (void*)0;

  safecInit();
  return trueMain(argc, nargv);
}

int _mainWrapper_qw(int argc, char **argv)
{
  int i;
  seq_fatp_char argv1;
  fatp_char *nargv = (fatp_char*)malloc((argc + 1) * sizeof(fatp_char));

  argv1._p = nargv;
  argv1._b = (void*)nargv;
  argv1._e = (void*)(nargv + argc + 1);

  for(i=0; i<argc; i++) {
    fatp_char copy = fattenstring(argv[i]);
    nargv[i] = copy;
  }
  nargv[argc]._p = (void*)0;
  nargv[argc]._b = (void*)0;

  safecInit();
  return trueMain(argc, argv1);
}

int _mainWrapper_fw(int argc, char **argv)
{
  int i;
  fseqp_fatp_char argv1;
  fatp_char *nargv = (fatp_char*)malloc((1 + argc) * sizeof(fatp_char));

  argv1._p = nargv;
  argv1._e = (void*)(nargv + argc + 1);

  for(i=0; i<argc; i++) {
    fatp_char copy = fattenstring(argv[i]);
    nargv[i] = copy;
  }
  nargv[argc]._p = (void*)0;
  nargv[argc]._b = (void*)0;

  safecInit();
  return trueMain(argc, argv1);
}


// --------------------- misc. wrappers ----------------
#ifdef __CYGWIN__
  int _get_ctype(int chr) {
    return (_ctype_+1)[chr];
  }

  int _get_ctype_p1(unsigned int chr) {
    return (_ctype_+1)[chr];
  }
#endif

#if defined(__GLIBC__) && __GLIBC__ == 2
  /// For brooksie
  // sm: glibc 2.0 and 2.1 both have this; so I removed check for minor number
  int _get__ctype_b(int chr) {
    if (!( (unsigned)chr < 256 )) {
      fprintf(stderr, "Bad character index to _get__ctype_b: %d\n", chr);
      abort();
    }
    return (__ctype_b[chr]);
  }
#endif


// sm: I don't have __assert ... we need an #ifdef here
// which identifies the machine which has this ...
// gn: who has this anyway?
#if !defined(__GLIBC__) && !defined(_MSVC)
  extern void __assert(char *, int, char *);
  void __assert_ww(fatp_char file, int line, fatp_char a2)
  {
    fatp_verify_nul(file);
    fatp_verify_nul(a2);
    __assert(file._p, line, a2._b);
  }
#endif


// sm: I don't know if this is a problem on msvc, but I suspect it is
#ifdef _GNUCC
  // sm: for m88k
  typedef void (*sig_fn)(int);
  typedef struct fseq_sig_fn {
    sig_fn fn;
    void *_e;
  } fseq_sig_fn;
  fseq_sig_fn signal_ff(int signum, fseq_sig_fn f)
  {
    // UNSOUND (why? don't remember)
    fseq_sig_fn ret;
    ret.fn = signal(signum, f.fn);
    ret._e = 0;     // people don't call handlers directly (right?)
    return ret;
  }

  // workaround for m88k getopt thing (bug in our tool)
  int m88k_getopt(int argc, seqp_char *argv, char *argdesc)
  {
    char **argv2;
    int i;
    int retval;

    // construct a new array
    argv2 = (char**)malloc((argc+1) * sizeof(char*));

    // copy the string pointers into it; no bounds checking
    for (i=0; i<argc+1; i++) {
      argv2[i] = argv[i]._p;    // last will be NULL (we hope)
    }

    // call the real getopt
    retval = getopt(argc, argv2, argdesc);

    // free the array
    free(argv2);

    return retval;
  }

  // extern char *optarg;
  // int getopt(int argc, char * SAFE const * FSEQ argv, const char * SAFE optstring);
  fseqp_char optarg_f;
  int getopt_fss(int argc, fseqp_p_char argv, char const *optstring)
  {
    int ret;

    if ((char**)argv._e - argv._p < argc) {
      // not enough entries in the argv array; should be at least argc many
      fp_fail_str("getopt called with too-small argv array"  FILE_AND_LINE);
    }

    ret = getopt(argc, argv._p, optstring);
    
    // make optarg available as optarg_f
    optarg_f._p = optarg;
    optarg_f._e = optarg? optarg + strlen(optarg) : NULL;
    
    return ret;
  }
#endif // _GNUCC

/*
// ------------------ strpbrk -----------------
// strpbrk searches a string for any of a set of characters
fatp_char strpbrk_wwf(fatp_char str, fseqp_char accept)
{
  fatp_char res;
  fatp_verify_nul(str);
  fseqp_verify_nul(accept);
  res._p = strpbrk(str._p, accept._p);
  res._b = res._p? str._b : NULL;
  return res;
}

// sm: changed seq_fatp_char to seqp_char .. I don't know how it worked before!
fatp_char strpbrk_wwq(fatp_char str, seqp_char accept)
{
  fatp_char res;
  fatp_verify_nul(str);
  seqp_verify_nul(accept);
  res._p = strpbrk(str._p, accept._p);
  res._b = res._p? str._b : NULL;
  return res;
}

fseqp_char strpbrk_fff(fseqp_char str, fseqp_char accept)
{
  fseqp_char ret;
  fseqp_verify_nul(str);
  fseqp_verify_nul(accept);
  ret._p = strpbrk(str._p, accept._p);
  ret._e = ret._p? str._e : NULL;
  return ret;
}

// sm: couldn't get it to always infer fseq return ..
char *strpbrk_sff(fseqp_char str, fseqp_char accept)
{
  return strpbrk_fff(str, accept)._p;
}

seqp_char strpbrk_qqf(seqp_char str, fseqp_char accept)
{
  seqp_char ret;
  seqp_verify_nul(str);
  fseqp_verify_nul(accept);
  ret._p = strpbrk(str._p, accept._p);
  if (ret._p) {
    ret._b = str._b;
    ret._e = str._e;
  }
  else {
    ret._b = NULL;
    ret._e = NULL;
  }
  return ret;
}
*/

// ------------------- strtok --------------------
fseqp_char strtok_fff(fseqp_char str, fseqp_char delim)
{
  fseqp_char ret;
  static char *prev_end = NULL;     // we must be stateful just like strtok

  if (str._p) {
    fseqp_verify_nul(str);
    prev_end = str._e;              // save for when str._p is NULL
  }
  fseqp_verify_nul(delim);

  ret._p = strtok(str._p, delim._p);
  ret._e = ret._p? prev_end : NULL;

  return ret;
}
fseqp_char strtok_FFf(fseqp_char str, fseqp_char delim) {
  return strtok_fff(str, delim);
}

// sm: copied from above for seq ptrs
seqp_char strtok_qqq(seqp_char str, seqp_char delim)
{
  seqp_char ret;
  static char *prev_base = NULL;    // we must be stateful just like strtok
  static char *prev_end = NULL;

  if (str._p) {
    seqp_verify_nul(str);
    prev_base = str._b;
    prev_end = str._e;              // save for when str._p is NULL
  }
  seqp_verify_nul(delim);

  ret._p = strtok(str._p, delim._p);
  if (ret._p) {
    ret._b = prev_base;
    ret._e = prev_end;
  }
  else {
    ret._b = NULL;
    ret._e = NULL;
  }

  return ret;
}


// A function whose result is either not used or used just for dereference
// This should be automatically constructed
char* strtok_sff(fseqp_char str, fseqp_char delim)
{
  fseqp_char res = strtok_fff(str, delim);
  CHECK_UBOUNDNULL(res._e, res._p, sizeof(char));     // sm: I think this is redundant
  return res._p;
}

char *strtok_sqq(seqp_char str, seqp_char delim) 
{
  return strtok_qqq(str, delim)._p;
}

fatp_char strtok_wws(fatp_char str, char *delim) {
  // Note this is not doing any checking
  fatp_char res;
  fatp_verify_nul(str);
  // CHECK_NULL(delim);
  res._p = strtok(str._p, delim);
  res._b = str._b;
  return res;
}


// --------------------- varargs stuff --------------------
// Define our version of va_start

// sm: I already defined this myself, calling it ALIGNBYTES
//#define OUR_ROUND(size) ((size + sizeof(int) - 1) & ~(sizeof(int) - 1))

#ifdef _GNUCC
  /* On GCC we do not take the address of the last local because if we do that
   * gcc will give us the address of a copy of the local. */
  #define getPNextArg(lastarg) __builtin_next_arg(lastarg)
#else
  #define getPNextArg(lastarg) ((char*)(&(lastarg)) + ALIGNBYTES(sizeof(lastarg)))
#endif
#define our_advance_va_arg(ap,size) ap = (va_list)((char*)ap + ALIGNBYTES(size))

int __ccured_va_count = 0;
int __ccured_va_tags[32];
typedef struct __ccured_va_localinfo {
   int next ;
   int count ;
   unsigned int tags[32] ;
   va_list nextp ;
} CCURED_VAINFO;

void __ccured_va_start_vs(CCURED_VAINFO *vainfo,
                          void *pNextArg) {
  vainfo->next  = 0;
  vainfo->count = __ccured_va_count;
  memcpy(vainfo->tags, __ccured_va_tags, sizeof(__ccured_va_tags));

  vainfo->nextp = (va_list)pNextArg;
  return;
}


void __ccured_va_end_vs(CCURED_VAINFO *vainfo) {
  vainfo->next = -1;
}

#define GET_VA_TAG(tags, idx) ((tags[idx >> 2] >> ((idx & 0x3) << 3)) & 0xFF)

void *__ccured_va_arg_svs(CCURED_VAINFO *vainfo,
                          unsigned int thisTypeSize,
                          int thisTypeIndex) {
  void* lastp = vainfo->nextp;

  if(vainfo->next == -1) {
    FP_FAIL(FAIL_VA_NOSTART);
  }
  if(vainfo->next >= vainfo->count) {
    FP_FAIL(FAIL_VA_MANY);
  }
  // Grab the stored tag
  {
    int tag = GET_VA_TAG(vainfo->tags, vainfo->next);
    if(tag != thisTypeIndex) {
      FP_FAIL(FAIL_VA_BADTYPE);
    }
  }
  our_advance_va_arg(vainfo->nextp, thisTypeSize);

  vainfo->next ++;
  return lastp;
}

// parse a printf-like format string (sm: does this check the
// argument types too?  context suggests yes but I don't see
// where it does)
void CHECK_FORMATARGS(const char *format) {
  int i = 0, check;
  if(! format) {
    FP_FAIL(FAIL_NONPOINTER);
  }
  if(__ccured_va_count == -1) return; // Was checked already
  while(*format) {
    if(*format == '%') {
      format ++;
    KeepLooking:
      switch(*format) {
      case '%': format ++; continue;
      case 's': check = 2; goto DoCheck;
      case 'f': check = 1; goto DoCheck;
      case 'l': case '-': case '.': format ++; goto KeepLooking;
      case 'd': case 'c': case 'x': check = 0; goto DoCheck;
      default:
        if(*format >= '0' && *format <= '9') {
          format ++; goto KeepLooking;
        }
        FP_FAIL(FAIL_UNRECOGNIZEDFORMAT);
      }
    DoCheck:
      if(i >= __ccured_va_count) {
        FP_FAIL(FAIL_VA_MANY);
      }
      if(check != GET_VA_TAG(__ccured_va_tags, i)) {
        FP_FAIL(FAIL_VA_BADTYPE);
      }
      i ++;
    }
    format ++;
  }
  __ccured_va_count = -1; // Was checked already. this will prevent the use
  // of va_arg but that would be unsafe anyway for printf-like functions
}

// ------------------- vsprintf -----------------------

#ifdef _MSVC
  #define VSNPRINTF _vsnprintf
/*  
  //matth:  I'm not sure how to prevent name mangling on _vsprintf, so for now
  // use a wrapper function:
  int _vsnprintf_ssvs(char* buffer, int n, const char *format, va_list args)
  {
    return VSNPRINTF(buffer, n, format, args);
    }*/
#else
  #define VSNPRINTF vsnprintf
/*  
  int vsnprintf_ssvs(char* buffer, int n, const char *format, va_list args)
  {
    return VSNPRINTF(buffer, n, format, args);
    }*/
#endif

int vsprintf_fsvs(fseqp_char buffer, const char *format,
                  CCURED_VAINFO *args)
{
  int res;

  CHECK_FORMATARGS(format);

  if(!buffer._e || buffer._p >= buffer._e) {
    FP_FAIL(FAIL_UBOUND);
  }

  res = VSNPRINTF(buffer._p, (int)buffer._e - (int)buffer._p - 1,
                  format, args->nextp);
  if(res < 0) {
    FP_FAIL(FAIL_UBOUND);
  }
  return res;
}

int vsprintf_qsvs(seqp_char buffer, const char *format,
                  CCURED_VAINFO *args)
{
  fseqp_char buffer_f;
  CHECK_SEQTOFSEQ(buffer._b, buffer._p);
  buffer_f._p = buffer._p;
  buffer_f._e = buffer._e;
  return vsprintf_fsvs(buffer_f, format, args);
}

// ---------------------- sprintf -----------------------

int sprintf_qss(seqp_char buffer, const char *format, ...)
{
  int res;
  CCURED_VAINFO args;
  fseqp_char fbuffer = { buffer._p, buffer._e };

  __ccured_va_start_vs(&args, getPNextArg(format));

  // Make sure there is at least 1 character available 
  CHECK_BOUNDS(buffer._b, buffer._e, (void*)buffer._p, 1);
  res = vsprintf_fsvs(fbuffer, format, &args);
  __ccured_va_end_vs(&args);
  return res;
}

int sprintf_fss(fseqp_char buffer, const char *format, ...)
{
  int res;
  CCURED_VAINFO args;

  __ccured_va_start_vs(&args, getPNextArg(format));

  // Make sure there is at least 1 character available
  CHECK_BOUNDS(buffer._p, buffer._e, (void*)buffer._p, 1);
  res = vsprintf_fsvs(buffer, format, &args);
  __ccured_va_end_vs(&args);
  return res;
}

int sprintf_wss(fatp_char buffer, char *format, ...)
{
  CCURED_VAINFO args;
  int res;
  int wrds = CHECK_FETCHLENGTH(buffer._p, buffer._b);
  fseqp_char fbuffer = { buffer._p, buffer._p + (wrds << 2) };

  __ccured_va_start_vs(&args, getPNextArg(format));
  res = vsprintf_fsvs(fbuffer, format, &args);
  __ccured_va_end_vs(&args);

  return res;
}

int sprintf_wws(fatp_char buffer, fatp_char format, ...)
{
  CCURED_VAINFO args;
  int res;
  int wrds = CHECK_FETCHLENGTH(buffer._p, buffer._b);
  fseqp_char fbuffer = { buffer._p, buffer._p + (wrds << 2) };

  fatp_verify_nul(format);

  __ccured_va_start_vs(&args, getPNextArg(format));
  res = vsprintf_fsvs(fbuffer, format._p, &args);
  __ccured_va_end_vs(&args);
  return res;
}

// --------------------- vfprintf -------------------
int vfprintf_ssvs(FILE * stream, const char *format, CCURED_VAINFO *args)
{
  int res;
  CHECK_NULL(stream);
  CHECK_NULL((char*)format);

  CHECK_FORMATARGS(format);
  res = vfprintf(stream, format, args->nextp);
  return res;
}


int vfprintf_wwvs(fatp_void stream, fatp_char format, CCURED_VAINFO *args)
{
  fatp_verify_nul(format);
  return vfprintf_ssvs(stream._p, format._p, args);
}

int vfprintf_wsvs(fatp_void stream, char *format, CCURED_VAINFO *args)
{
  // Do not check the FILE*
  return vfprintf_ssvs(stream._p, format, args);
}


// ---------------------- fprintf ----------------------
int fprintf_wws(fatp_void stream, fatp_char format, ...)
{
  CCURED_VAINFO args;
  int res;
  fatp_verify_nul(format);

  __ccured_va_start_vs(&args, getPNextArg(format));
  res = vfprintf_ssvs(stream._p, format._p, &args);
  __ccured_va_end_vs(&args);
  return res;
}

int fprintf_sws(FILE *stream, fatp_char format, ...)
{
  CCURED_VAINFO args;
  int res;
  fatp_verify_nul(format);

  __ccured_va_start_vs(&args, getPNextArg(format));
  res = vfprintf_ssvs(stream, format._p, &args);
  __ccured_va_end_vs(&args);
  return res;
}

int fprintf_wss(fatp_void stream, char *format, ...)
{
  CCURED_VAINFO args;
  int res;

  __ccured_va_start_vs(&args, getPNextArg(format));
  res = vfprintf_ssvs(stream._p, format, &args);
  __ccured_va_end_vs(&args);
  return res;
}


// ------------------- [v]printf -------------------
int vprintf_svs(char const *format, CCURED_VAINFO *args)
{
  return vfprintf_ssvs(stdout, format, args);
}

// sm 2/14/02: this used to be called printf_sf .. ??  the mangling
// algorithm gives this function the name printf_fs ..
int printf_fs(fseqp_char format, ...) // weimer
{
  CCURED_VAINFO args;
  int res;
  verify_nul(format._p, format._e - format._p);

  __ccured_va_start_vs(&args, getPNextArg(format));
  res = vfprintf_ssvs(stdout, format._p, &args);
  __ccured_va_end_vs(&args);
  return res;
}

int printf_ws(fatp_char format, ...)
{
  CCURED_VAINFO args;
  int res;
  fatp_verify_nul(format);

  __ccured_va_start_vs(&args, getPNextArg(format));
  res = vfprintf_ssvs(stdout, format._p, &args);
  __ccured_va_end_vs(&args);
  return res;
}


// sm: trying vsyslog too! (conditional on linux to not interfere on cygwin)
#ifdef __linux__
  extern void vsyslog_svs(int pri, char const *format, CCURED_VAINFO *args)
  {
    CHECK_FORMATARGS(format);              // check
    vsyslog(pri, format, args->nextp);     // pass
  }
#endif // __linux__


// ---------------------- scanf and friends -----------------------
int ccured_scanf_count;
FILE *ccured_sscanf_file = NULL;
void resetScanfCount(void) {
  ccured_scanf_count = 0;
}

// Call this function for sscanf
void resetSScanfCount(char *str) {
  if(! str) {
    FP_FAIL(FAIL_NULL_STRING);
  }
  if(ccured_sscanf_file) {
    fprintf(stderr, "Must call getScanfCount after resetSScanfCount");
    exit(2);
  }
  ccured_sscanf_file = fopen("sscanf.out", "w+b");
  if(! ccured_sscanf_file) {
    fprintf(stderr, "Cannot open sscanf.out\n");
    exit(1);
  }
  fputs(str, ccured_sscanf_file);
  fseek(ccured_sscanf_file, 0L, SEEK_SET);
}

void resetSScanfCount_w(fatp_char str)
{
  fatp_verify_nul(str);
  resetSScanfCount(str._p);
}

int getScanfCount() {
  if(ccured_sscanf_file) {
    fclose(ccured_sscanf_file);
  }
  return ccured_scanf_count;
}

double ccured_fscanf_double(FILE *stream, char *format) {
  double d = 0.0;
  if(ccured_scanf_count != EOF) {
    int res = fscanf(stream, format, &d);
    if(res == EOF) {
      ccured_scanf_count = EOF;
    } else
      ccured_scanf_count += res;
  }
  return d;
}                                                                   

double ccured_fscanf_double_ww(fatp_FILE stream, fatp_char format) 
{
  fatp_verify_nul(format);
  return ccured_fscanf_double(stream._p, format._p);
}

int ccured_fscanf_int(FILE *stream, char *format) {
  int i = 0;
  if(ccured_scanf_count != EOF) {
    int res = fscanf(stream, format, &i);
    if(res == EOF) {
      ccured_scanf_count = EOF;
    } else
      ccured_scanf_count += res;
  }
  return i;
}

int ccured_fscanf_int_ww(fatp_FILE stream, fatp_char format)
{
  fatp_verify_nul(format);
  return ccured_fscanf_int(stream._p, format._p);
}


void ccured_fscanf_nothing(FILE *stream, char* what) {
  if(EOF == fscanf(stream, what))  // just advance the pointer and match
    ccured_scanf_count = EOF;
}
void ccured_fscanf_nothing_ws(fatp_FILE stream, char* what) {
  ccured_fscanf_nothing(stream._p, what);
}

void ccured_fscanf_string_ssf(FILE *stream, char* what, fseqp_char buffer) {
  int bufflen = (int)buffer._e - (int)buffer._p;
  char *mybuff;
  int orig_pos, read, scanned;

  if(!buffer._p || !buffer._e) {
    FP_FAIL(FAIL_NONPOINTER);
  }
  if(buffer._p >= buffer._e) {
    FP_FAIL(FAIL_UBOUND);
  }
  // Now create a buffer
  mybuff = (char*)malloc(bufflen);
  if(!mybuff) {
    fprintf(stderr, "Cannot allocate buffer for fscanf_string");
    exit(2);
  }
  // Now remember the positon in the stream
  orig_pos = ftell(stream);
  // Now read that many characters from the stream in the buffer
  read = fread(mybuff, 1, bufflen - 1, stream);
  if(read == bufflen - 1) {
    FILE *myfile = fopen("sscanf_string", "rwb");
    if(! myfile) {
      fprintf(stderr, "Cannot create file for fscanf_string");
      exit(2);
    }
    // Write them out to the new file
    fwrite(mybuff, read, 1, myfile);
    // Rewind
    fseek(myfile, 0L, SEEK_SET);
    // And scan
    scanned = fscanf(myfile, what, buffer._p);
    // Find out how much we have advanced. This is misleading if we
    // are openin text mode on Win and we skipped newlines !
    orig_pos += ftell(myfile);
    fseek(stream, orig_pos, SEEK_SET);
    fclose(myfile);
    if(scanned == EOF) {
      FP_FAIL(FAIL_UBOUND);
    }
  } else {
    // The file is shorter than the buffer. Go ahead and scan
    scanned = fscanf(stream, what, buffer._p);
  }
  if(scanned == EOF) {
    ccured_scanf_count = EOF;
  } else {
    ccured_scanf_count += scanned;
  }
}

void ccured_fscanf_string_ssw(FILE *stream, char* what, fatp_char buffer) {
  fseqp_char buffer_f;
  int buff_wrds = CHECK_FETCHLENGTH(buffer._p, buffer._b);
  buffer_f._p = buffer._p;
  buffer_f._e = buffer._p + (buff_wrds << 2);
  // Now call the above
  ccured_fscanf_string_ssf(stream, what, buffer_f);
  // Now fix the tags !
  // For the entire buffer to be sure
  CHECK_ZEROTAGS(buffer._b, buff_wrds, buffer._p, buff_wrds << 2, 0);
}

// ---------------------- socket functions ----------------
#ifdef _GNUCC
  // the user simply has to pass a buffer in 'optval' that is at least
  // 'optlen' bytes long
  // int setsockopt(int s, int level, int optname, void const *optval, int optlen);
  int setsockopt_w(int s, int level, int optname, fatp_char optval, int optlen)
  {
    fatp_verify_atleast(optval, optlen);
    return setsockopt(s, level, optname, optval._p, optlen);
  }

  // same deal for bind
  // int bind(int sockfd, struct sockaddr *my_addr, int addrlen);
  int bind_w(int sockfd, fatp_char my_addr, int addrlen)
  {
    fatp_verify_atleast(my_addr, addrlen);
    return bind(sockfd, (struct sockaddr*)my_addr._p, addrlen);
  }

  // connect is just like bind
  // int connect(int sockfd, struct sockaddr *my_addr, int addrlen);
  int connect_w(int sockfd, fatp_char my_addr, int addrlen)
  {
    fatp_verify_atleast(my_addr, addrlen);
    return connect(sockfd, (struct sockaddr*)my_addr._p, addrlen);
  }

  // similar for accept, but length is value-result
  // int accept(int s, struct sockaddr *addr, int *addrlen);
  int accept_ws(int s, fatp_char addr, int *addrlen)
  {
    if (addr._p == 0) {
      // addr is ignored when null
    }
    else {
      // verify addr's size
      fatp_verify_atleast(addr, *addrlen);
    }
    return accept(s, (struct sockaddr*)addr._p, addrlen);
  }


  // same as with accept
  //int getpeername(int s, struct sockaddr *name, int *namelen);
  int getpeername_ws(int s, fatp_char name, int *namelen)
  {
    fatp_write_atleast(name, *namelen);
    return getpeername(s, (struct sockaddr*)name._p, namelen);
  }

  // identical to getpeername
  // int getsockname(int s, struct sockaddr *name, int *namelen);
  int getsockname_ws(int s, fatp_char name, int *namelen)
  {
    fatp_write_atleast(name, *namelen);
    return getsockname(s, (struct sockaddr*)name._p, namelen);
  }

  // int sendto(int s, const void *msg, size_t len, int flags,
  //            const struct sockaddr *to, socklen_t tolen);
  int sendto_ws(int s, fatp_char msg, size_t len, int flags,
                const struct sockaddr *to, socklen_t tolen)
  {
    fatp_verify_atleast(msg, len);
    return sendto(s, msg._p, len, flags, to, tolen);
  }

  // int select(int n, fd_set *readfds, fd_set *writefds,
  //            fd_set *exceptfds, struct timeval *timeout);
  int select_sssw(int n, fd_set *readfds, fd_set *writefds,
                  fd_set *exceptfds, fatp_char timeout)
  {
    fatp_write_atleast(timeout, sizeof(struct timeval));    // writes timeout
    return select(n, readfds, writefds, exceptfds,
                  (struct timeval *)timeout._p);
  }

  // int recvfrom(int s, void *buf, size_t len, int flags,
  //              struct sockaddr *from, socklen_t *fromlen);
  int recvfrom_wss(int s, fatp_char buf, size_t len, int flags,
                   struct sockaddr *from, socklen_t *fromlen)
  {
    fatp_write_atleast(buf, len);
    return recvfrom(s, buf._p, len, flags, from, fromlen);
  }

  //gn: I don't have getnameinfo on cygwin
  #if !defined(__CYGWIN__)
    // sm: 'salen' is the length of buffer pointed to by 'sa'; similarly for
    // hostlen/host and servlen/serv.. but these last two aren't checked,
    // and checking them would require adding a model to force them to be
    // at least fseq.. I'll skip that hassle for now (UNSOUND)
    int getnameinfo(/*const struct sockaddr*/ void *sa, int salen,
                    char *host, int hostlen, char *serv, int servlen, int flags);
    int getnameinfo_wss(fatp_char sa, int salen,
                        char *host, int hostlen, char *serv, int servlen,
                        int flags)
    {
      fatp_verify_atleast(sa, salen);
      return getnameinfo(sa._p, salen, host, hostlen, serv, servlen, flags);
    }

    // not a socket function, but specific to unix, so I'll put it here
    //void *mmap(void *start, int length, int prot, int flags, int fd, int offset);
    seqp_void mmap_qs(void *start, int length, int prot, int flags, int fd, int offset)
    {
      seqp_void ret;

      // the only way it makes sense for 'start' to be safe is if it's null
      if (start != NULL) {
        FP_FAIL_STR("expected NULL start to mmap");
      }

      ret._p = mmap(start, length, prot, flags, fd, offset);
      if (ret._p == MAP_FAILED) {
        ret._b = ret._e = NULL;
      }
      else {
        ret._b = ret._p;
        ret._e = ((char*)ret._p) + length;    // ptr to first *invalid* byte
      }

      return ret;
    }

    int munmap_q(seqp_void start, int length)
    {
      CHECK_NULL(start._p);
      if (start._p != start._b) {
        FP_FAIL_STR("must pass pointer to start of the region");
      }
      if ((char*)start._e - (char*)start._b  != length) {
        FP_FAIL_STR("must pass correct length to munmap");
      }

      // fixme?  this could leave dangling pointers.. it's a form
      // of explicit deallocation.. (UNSOUND)
      return munmap(start._p, length);
    }
  #endif // !defined(__CYGWIN__)
#endif // _GNUCC


// -------------------------- glob -------------------------
#if defined(_GNUCC) && !defined(__CYGWIN__)
  // glob_t as inferred by our system with ftpd
  #define GLOB_USES_FSEQ 1
  struct wrapped_glob_t {
     unsigned int gl_pathc ;
     #if GLOB_USES_FSEQ
       // 2/10/02: yet another new inference, and we're back to fseq
       #define GLOB_FSEQP_CHAR fseqp_char     // used below
       fseqp_fseqp_char gl_pathv ;
     #else
       // new inference -- new type ..
       #define GLOB_FSEQP_CHAR seqp_char
       fseqp_seqp_char gl_pathv ;
     #endif
     unsigned int gl_offs ;
     int gl_flags ;
     void (* gl_closedir)(void * ) ;
     struct dirent * (* gl_readdir)(void * ) ;
     void * (* gl_opendir)(char * ) ;
     int (*  gl_lstat)(char * , struct stat * ) ;
     int (*  gl_stat)(char * , struct stat * ) ;
  };

  int wrapped_glob(const char *pattern, int flags,
                   int errfunc(const char * epath, int eerrno),
                   struct wrapped_glob_t *pglob)
  {
    glob_t gl;
    int err, totlen, i;
    char *bigArray;

    // check that we can handle the specified flags
    if (flags & (GLOB_DOOFFS | GLOB_APPEND | GLOB_ALTDIRFUNC)) {
      // none of these are especially difficult, I just don't want to
      // waste time implementing and testing them when I don't need them
      FP_FAIL_STR("some glob flags are specified that I can't handle (yet)");
    }

    // call underlying globber
    err = glob(pattern, flags, errfunc, &gl);
    if (err) {
      return err;
    }

    // copy results into my own area, so I can free the strings that
    // underlying globber allocated

    // first, allocate a big buffer to hold all the strings
    for (totlen=0, i=0; i < gl.gl_pathc; i++) {
      int len = strlen(gl.gl_pathv[i]);
      totlen += len + 1;
    }
    bigArray = (char*)malloc(totlen);

    // allocate an array of [f]seq pointers, to point into this array; the
    // way ftpd uses this, it seems to think the array is terminated by
    // a null pointer, though the man page doesn't mention this ..
    pglob->gl_pathv._p = (GLOB_FSEQP_CHAR*)
      malloc(sizeof(GLOB_FSEQP_CHAR) * (gl.gl_pathc+1));
    pglob->gl_pathv._e = pglob->gl_pathv._p + (gl.gl_pathc+1);    // max+1

    // now copy the strings one at a time
    for (totlen=0, i=0; i < gl.gl_pathc; i++) {
      int len = strlen(gl.gl_pathv[i]);
      strcpy(bigArray+totlen, gl.gl_pathv[i]);
      #if !GLOB_USES_FSEQ
        pglob->gl_pathv._p[i]._b = bigArray+totlen;
      #endif
      pglob->gl_pathv._p[i]._p = bigArray+totlen;
      pglob->gl_pathv._p[i]._e = bigArray+totlen+len+1;   // max+1
      totlen += len + 1;
    }

    // write a final null
    pglob->gl_pathv._p[i]._p = NULL;
    pglob->gl_pathv._p[i]._p = NULL;

    // now that we've copied all the strings (into our garbage-collected
    // heap), we can deallocate whatever libc did
    globfree(&gl);

    return err;
  }

  void wrapped_globfree(struct wrapped_glob_t *pglob)
  {
    // we use a garbage collector
  }

  #undef GLOB_USES_FSEQ
  #undef GLOB_FSEQP_ChAR
#endif // _GNUCC && !__CYGWIN__


// ---------------------- exec ---------------------
int execv_fff(fseqp_char path, fseqp_fseqp_char argv)
{
  int argvLen, i, ret;
  fseqp_char *p = argv._p;
  char **thinArgv;

  fseqp_verify_nul(path);

  // 'argv' is supposed to be a null-terminated array of FSEQ char*'s;
  // how many FSEQs are actually here?
  argvLen = ((int)argv._e - (int)argv._p) / sizeof(fseqp_char);

  // we'll need to copy the pointers to an array of thin pointers
  // for execv to actually use
  thinArgv = (char**)malloc(argvLen * sizeof(char*));

  // walk the argv array, making sure each fseq is ok to read from,
  // and that we find a NULL pointer among them before the end
  for (i=0; i<argvLen; i++, p++) {
    thinArgv[i] = p->_p;

    if (p->_p == NULL) {
      // this is the last one that execv will read from, so we
      // don't need to check the rest
      break;
    }

    // each fseq should be valid to read from
    fseqp_verify_nul(*p);
  }

  if (i >= argvLen) {     // i > argvLen possible if argv._p > argv._e
    // we reached the end of the array without finding a NULL
    // entry, which is bad
    fp_fail_str("execv passed a non-null-terminated argv array"  FILE_AND_LINE);
  }

  ret = execv(path._p, thinArgv);

  // execv doesn't normally return, but if it does, we shouldn't
  // leak memory
  free(thinArgv);

  return ret;
}
int execv_ffF(fseqp_char path, fseqp_fseqp_char argv) {
  return execv_fff(path, argv);
}

// ------------------ startof, endof, ptrof, mkptr  ------------------

#define __CCURED_ASSERT(test, errCode) if(!(test)) {FP_FAIL((errCode));}

// Similar to verify_nul, but returns the length of the string (excluding the NUL)
// rather than the length of the buffer. 
//   __verify_nul_strlen(start, end) <= (end - start - 1)
__inline static int __verify_nul_strlen(const char *start, const char *end)
{
  const char * current = start;
  int max = (int)end - (int)start;

  for(; max > 0; max--) {
    if (*current == 0) {
      // ok
      return current - start;
    }
    current++;
  }

  // didn't find a nul in the accessible area
  FP_FAIL(FAIL_STRINGBOUND);
  return 0;     // silence warning
}


// -- TODO: make all of these functions inline, and put them in ccuredcheck.h:

//
//safe
//
void* __ptrof(void* p){
  return p;
}
void* __ptrof_nocheck(void* p){
  return p;
}
void* __mkptr(void* ptr, void* phome){
  return ptr;    
}
void* __mkptr_ssf(void* ptr, fseqp_void phome){
  return ptr;    
}
void* __mkptr_ssq(void* ptr, seqp_void phome){
  return ptr;    
}
void* __mkptr_ssF(void* ptr, fseqp_void phome){
  return ptr;    
}
void* __mkptr_ssQ(void* ptr, seqp_void phome){
  return ptr;    
}
int __noninteger(void* p){
  return p != 0;
}

//
//seq
//
void* __startof_sq(seqp_void p){
  return p._b;
}
void* __endof_sq(seqp_void p){
  return p._e;
}
void* __ptrof_sq(seqp_void p){
  __CCURED_ASSERT(p._p >= p._b, FAIL_LBOUND);
  __CCURED_ASSERT(!p._b || (p._p < p._e), FAIL_UBOUND);  //allow p to be null
  return p._p;
}
void* __ptrof_nocheck_sq(seqp_void p){
  return p._p;
}
seqp_void __mkptr_qsq(void* ptr, seqp_void phome)
{
  seqp_void res;
  res._p = ptr;
  res._b = ptr == 0 ? 0 : phome._b;
  res._e = ptr == 0 ? 0 : phome._e;
  return res;    
}
seqp_void __mkptr_qsQ(void* ptr, seqp_void phome)
{
  seqp_void res;
  res._p = ptr;
  res._b = ptr == 0 ? 0 : phome._b;
  res._e = ptr == 0 ? 0 : phome._e;
  return res;    
}
int __noninteger_q(seqp_void p){
  return p._b != 0;
}
int __strlen_q(seqp_void p){
  __CCURED_ASSERT(__noninteger_q(p), FAIL_NONPOINTER);
  return __verify_nul_strlen(p._p, p._e);
}
void __write_at_least_q(seqp_void ptr, unsigned int n){
  __CCURED_ASSERT(__noninteger_q(ptr), FAIL_NONPOINTER);
  //n is treated as an unsigned number so that we fail if it's negative
  __CCURED_ASSERT(n <= (U32)ptr._e - (U32)ptr._p, FAIL_UBOUND);
}
void __copytags_qq(seqp_void dest, seqp_void src, unsigned int n){
  __CCURED_ASSERT(__noninteger_q(src) && __noninteger_q(dest), FAIL_NONPOINTER);
  //n is treated as an unsigned number so that we fail if it's negative
  __CCURED_ASSERT(n <= (U32)dest._e - (U32)dest._p, FAIL_UBOUND);
  __CCURED_ASSERT(n <= (U32)src._e - (U32)src._p, FAIL_UBOUND);
}


//
//fseq
//
void* __endof_sf(fseqp_void p){
  return p._e;
}
void* __ptrof_sf(fseqp_void p){
  __CCURED_ASSERT(!p._e || (p._p < p._e), FAIL_UBOUND);  //allow p to be null
  return p._p;
}
void* __ptrof_nocheck_sf(fseqp_void p){
  return p._p;
}
fseqp_void __mkptr_fsf(void* ptr, fseqp_void phome)
{
  fseqp_void res;
  res._p = ptr;
  res._e = ptr == 0 ? 0 : phome._e;
  return res;    
}
fseqp_void __mkptr_fsF(void* ptr, fseqp_void phome)
{
  fseqp_void res;
  res._p = ptr;
  res._e = ptr == 0 ? 0 : phome._e;
  return res;    
}
fseqp_void __mkptr_fsq(void* ptr, seqp_void phome)
{
  fseqp_void res;
  res._p = ptr;
  res._e = ptr == 0 ? 0 : phome._e;
  return res;    
}
int __noninteger_f(fseqp_void p){
  return p._e != 0;
}
int __strlen_f(fseqp_void p){
  __CCURED_ASSERT(__noninteger_f(p), FAIL_NONPOINTER);
  return __verify_nul_strlen(p._p, p._e);
}
void __write_at_least_f(fseqp_void ptr, unsigned int n){
  __CCURED_ASSERT(__noninteger_f(ptr), FAIL_NONPOINTER);
  __CCURED_ASSERT(n <= (U32)ptr._e - (U32)ptr._p, FAIL_UBOUND);
}
void __copytags_ff(fseqp_void dest, fseqp_void src, unsigned int n){
  __CCURED_ASSERT(__noninteger_f(src) && __noninteger_f(dest), FAIL_NONPOINTER);
  __CCURED_ASSERT(n <= (U32)dest._e - (U32)dest._p, FAIL_UBOUND);
  __CCURED_ASSERT(n <= (U32)src._e - (U32)src._p, FAIL_UBOUND);
}
void __copytags_qf(seqp_void dest, fseqp_void src, unsigned int n){
  __CCURED_ASSERT(__noninteger_f(src) && __noninteger_q(dest), FAIL_NONPOINTER);
  //n is treated as an unsigned number so that we fail if it's negative
  __CCURED_ASSERT(n <= (U32)dest._e - (U32)dest._p, FAIL_UBOUND);
  __CCURED_ASSERT(n <= (U32)src._e - (U32)src._p, FAIL_UBOUND);
}


//
//seqN
//
void* __startof_sQ(seqp_void p){
  return p._b;
}
void* __endof_sQ(seqp_void p){
  return p._e;
}
void* __ptrof_sQ(seqp_void p){
  __CCURED_ASSERT(p._p >= p._b, FAIL_LBOUND);
  __CCURED_ASSERT(!p._b || (p._p < p._e), FAIL_UBOUND);  //allow p to be null
  return p._p;
}
void* __ptrof_nocheck_sQ(seqp_void p){
  return p._p;
}
seqp_void __mkptr_QsQ(void* ptr, seqp_void phome){
  return __mkptr_qsq(ptr, phome);
}
int __noninteger_Q(seqp_void p){
  return p._b != 0;
}
int __strlen_Q(seqp_void p){
  __CCURED_ASSERT(__noninteger_Q(p), FAIL_NONPOINTER);
  return __verify_nul_strlen(p._p, p._e);
}
void __write_at_least_Q(seqp_void ptr, unsigned int n){
  __write_at_least_q(ptr, n);
}
void __copytags_QQ(seqp_void dest, seqp_void src, unsigned int n){
  __copytags_qq(dest, src, n);
}
void __copytags_QF(seqp_void dest, fseqp_void src, unsigned int n){
  __copytags_qf(dest, src, n);
}


//
//fseqN
//
void* __endof_sF(fseqp_void p){
  return p._e;
}
void* __ptrof_sF(fseqp_void p){
  __CCURED_ASSERT(!p._e || (p._p < p._e), FAIL_UBOUND);  //allow p to be null
  return p._p;
}
void* __ptrof_nocheck_sF(fseqp_void p){
  return p._p;
}
fseqp_void __mkptr_FsF(void* ptr, fseqp_void phome){
  return __mkptr_fsf(ptr, phome);
}
fseqp_void __mkptr_FsQ(void* ptr, seqp_void phome){
  return __mkptr_fsq(ptr, phome);
}
int __noninteger_F(fseqp_void p){
  return p._e != 0;
}
int __strlen_F(fseqp_void p){
  __CCURED_ASSERT(__noninteger_F(p), FAIL_NONPOINTER);
  return __verify_nul_strlen(p._p, p._e);
}
void __write_at_least_F(fseqp_void ptr, unsigned int n){
  __write_at_least_f(ptr, n);
}
void __copytags_FF(fseqp_void dest, fseqp_void src, unsigned int n){
  __copytags_ff(dest, src, n);
}


//
//wild
//
void* __ptrof_sw(fatp_void p){
  __CCURED_ASSERT(p._p >= p._b, FAIL_LBOUND);
  __CCURED_ASSERT(!p._b || (p._p < CHECK_FETCHEND(p._p, p._b)), FAIL_UBOUND);  //allow p to be null
  return p._p;
}
void* __ptrof_nocheck_sw(fatp_void p){
  return p._p;
}
void* __startof_sw(fatp_void p){
  return p._b;
}
void* __endof_sw(fatp_void p){
  if(p._b == 0)
    return 0;
  else
    return CHECK_FETCHEND(p._p, p._b);
}
fatp_void __mkptr_wsw(void* ptr, fatp_void phome)
{
  fatp_void res;
  res._p = ptr;
  res._b = ptr == 0 ? 0 : phome._b;
  return res;    
}
int __noninteger_w(fatp_void p){
  return p._b != 0;
}
int __strlen_w(fatp_void p){
  __CCURED_ASSERT(__noninteger_w(p), FAIL_NONPOINTER);
  return __verify_nul_strlen(p._p, __endof_sw(p));
}
void __write_at_least_w(fatp_char ptr, unsigned int n){
  fatp_write_atleast(ptr, n);
}
void __copytags_ww(fatp_char dest, fatp_char src, unsigned int n){
  __CCURED_ASSERT(n <= (unsigned int)fatp_length(src), FAIL_UBOUND); // checks that src is not an integer.
  __CCURED_ASSERT(n <= (unsigned int)fatp_length(dest), FAIL_UBOUND); // checks that dest is not an integer.
  fatp_copy_tags(dest, src, n); 
}

/* When we use the WILD solver we might need strange versions of these 
 * wrappers */
fatp_void __ptrof_ww(fatp_void x) {
  return x;
}

fatp_void __mkptr_www(fatp_void ptr, fatp_void phome)
{
  fatp_void res;
  res._p = ptr._p;
  res._b = ptr._p == 0 ? 0 : phome._b;
  return res;    
}

// ------------------------ kind_of ------------------
// sm: what is this for?
// gn: to build test cases in which we can ask about the kind of a pointer

char* ccured_kind_of() {
  return "SAFE";
}
char* ccured_kind_of_sq() {
  return "SEQ";
}
char* ccured_kind_of_sQ() {
  return "SEQN";
}
char* ccured_kind_of_sf() {
  return "FSEQ";
}
char* ccured_kind_of_sF() {
  return "FSEQN";
}
char* ccured_kind_of_sw() {
  return "WILD";
}

struct {
  int len;
  char data[8];
  int tags;
} theWild = { 2, "WILD", 0 };

fatp_char ccured_kind_of_ww() {
  fatp_char res;
  res._p = theWild.data;
  res._b = res._p;
  return res;
}

