// mymark.c
// code for mymark.h

#include "mymark.h"              // this module

// since we're accessing private GC variables, we need the
// declarations to match what were used when gc.a was built
#define ATOMIC_UNCOLLECTABLE 1
#define NO_SIGNALS 1
#define NO_EXECUTE_PERMISSION 1
#define SILENT 1
#define ALL_INTERIOR_POINTERS 1

#include "gc.h"                  // basic interface

#include "private/gc_priv.h"     // necessary first private header
#include "private/gc_mark.h"     // struct 'mse', mark proc stuff

#include <stdio.h>               // printf

#if 1
  #define D(expr) expr
#else
  #define D(expr) ((void)0)
#endif


// ----------- my instrumentation ------------
void noticedReachable(void *block, size_t size)
{
  D(printf("noticedReachable: %p, sz=%d\n", block, size));
}

void noticedUnreachable(void *block, size_t size)
{
  D(printf("noticedUnreachable: %p, sz=%d\n", block, size));
}

void noticedContiguousUnreachable(void *block, size_t size, size_t arraySize)
{
  D(printf("noticedContiguousUnreachable: %p, sz=%d, arraySize=%d\n",
           block, size, arraySize));
}


// --------------- blocks with my marking procedure --------------
// custom mark procedure
struct ms_entry *myMarkProc(
  // pointer to the object from which we want to trace pointers
  word *addr,

  // current top of the mark stack (is mark_stack_ptr[0] or [1] the
  // one to use?)
  struct ms_entry *mark_stack_top,

  // indicator of where the mark stack ends (again, I have the
  // off-by-one question)
  struct ms_entry *mark_stack_limit,

  // 24 bits of extra information associated with the type descriptor
  // maintained within the GC's data structures
  word env)
{
  // runs through the block considering each potential pointer
  register word *current_p = addr;

  // compute the address of the last potential pointer; calling
  // GC_size here is a performance loss since it involves getting
  // the object's header, which we probably had at some point
  // in the recent past
  size_t sz = GC_size(addr);
  register word *limit = (word*)( ((char*)addr + sz) - ALIGNMENT );

  // other variables needed by loop below
  word current;
  register ptr_t greatest_ha = GC_greatest_plausible_heap_addr;
  register ptr_t least_ha = GC_least_plausible_heap_addr;

  D(printf("myMarkProc on addr %p, limit %p, size %d\n",
           current_p, limit, sz));

  // this loop is lifted (almost) without modification from mark.c in the
  // GC itself
  while (current_p <= limit) {
    /* Empirically, unrolling this loop doesn't help a lot.	*/
    /* Since HC_PUSH_CONTENTS expands to a lot of code,	*/
    /* we don't.						*/
    current = *current_p;
    PREFETCH((ptr_t)current_p + PREF_DIST*CACHE_LINE_SIZE);
    if ((ptr_t)current >= least_ha && (ptr_t)current <  greatest_ha) {
      /* Prefetch the contents of the object we just pushed.  It's	*/
      /* likely we will need them soon.				*/
      PREFETCH(current);
      //PUSH_CONTENTS((ptr_t)current, mark_stack_top,
      //              mark_stack_limit, current_p, exit2);

      //# define PUSH_CONTENTS(current, mark_stack_top, mark_stack_limit,
      //                       source, exit_label)
      {
          hdr * my_hhdr;
          ptr_t my_current = (ptr_t)current;

          GET_HDR(my_current, my_hhdr);
          if (IS_FORWARDING_ADDR_OR_NIL(my_hhdr)) {
               my_current = GC_FIND_START(my_current, my_hhdr, (word)current_p);
               if (my_current == 0) goto exit_label;
               my_hhdr = GC_find_header(my_current);
          }
          
          // at this point we've established that 'current' is a valid
          // pointer into a block, where 'my_current' is the beginning
          D(printf("pointer from block %p to block %p\n",
                   addr, my_current));

          PUSH_CONTENTS_HDR(my_current, mark_stack_top, mark_stack_limit,
                        current_p, exit_label, my_hhdr);
      exit_label: ;
      }

    }
    current_p = (word *)((char *)current_p + ALIGNMENT);
  }

  // our contract is we must return this
  return mark_stack_top;
}


// global information for my kind of objects:
// this is the procedure index for my mark procedure
#define MYINDEX (GC_RESERVED_MARK_PROCS + 0)

// this is the object 'kind' (where the normal GC uses this e.g.
// to distinguish pointer-ful and pointer-free areas)
int myKind;

// each kind needs its own free list (?), so this is mine
ptr_t *myFreeList;

// true once this has been initialized
int myMarkInited = 0;


// install knowledge of my kind of object into the GC
// (adapted from GC_init_gcj_malloc in gcl_mlc.c)
void initMyMark()
{
  GC_init();	/* In case it's not already done.	*/

  // install my mark procedure
  GC_mark_procs[MYINDEX] = myMarkProc;

  // create a new kind, so that the GC will find and use our
  // mark procedure
  myFreeList = (ptr_t *)
      GC_INTERNAL_MALLOC((MAXOBJSZ+1)*sizeof(ptr_t), PTRFREE);
  if (myFreeList == 0)
      ABORT("Couldn't allocate myFreeList");
  BZERO(myFreeList, (MAXOBJSZ+1)*sizeof(ptr_t));
  myKind = GC_n_kinds++;
  GC_obj_kinds[myKind].ok_freelist = myFreeList;
  GC_obj_kinds[myKind].ok_reclaim_list = 0;
  GC_obj_kinds[myKind].ok_descriptor =
    MAKE_PROC(MYINDEX, 0 /* env */);
  GC_obj_kinds[myKind].ok_relocate_descr = FALSE;
  GC_obj_kinds[myKind].ok_init = TRUE;
  
  myMarkInited = 1;
}


void *allocMyMark(size_t size)
{
  void *ret;

  if (!myMarkInited) {
    initMyMark();
  }

  ret = GC_malloc_kind(size, myKind);
  //printf("allocMyMark(%d) yielded %p\n", size, ret);
  return ret;
}


// ---------------- test code ---------------------
#ifdef TEST_MYMARK
// global variable to choose which malloc to use
typedef void* (*MallocFunc)(size_t size);
MallocFunc whichMalloc = GC_malloc;

void dropN(int n)
{
  // allocate 'n' things and drop them on the floor
  int i;
  for (i=0; i<n; i++) {
    whichMalloc(16);
  }
}


void *dropHalf(int n)
{
  void *ret = NULL;

  // allocate 'n' blocks; drop half of them; the other
  // half will be strung together and returned
  int i;
  for (i=0; i<n; i++) {
    if (i%2) {
      // drop it
      whichMalloc(16);
    }
    else {
      // string it with previous blocks
      void **p = (void**)whichMalloc(16);
      *p = ret;
      ret = p;
    }
  }

  return ret;
}


void useMyMark()
{
  void **p;

  printf("--------- useMyMark ----------\n");

  // without this, GC_malloc_explicitly_typed segfaults immediately;
  // but this isn't public (compiler will complain), because Boehm
  // expects me to call GC_make_descriptor
  //GC_init_explicit_typing();

  // prepare to use my mark procedure
  initMyMark();

  // allocate something with it
  p = (void**)allocMyMark(16);

  // make it point to another obj with my marker
  *p = allocMyMark(16);

  // gc now, should both be reachable
  printf("first gc..\n");
  GC_gcollect();

  // nullify, making unreachable
  p = NULL;

  // gc again and expect to lose both
  printf("second gc..\n");
  GC_gcollect();
}


void severalTests(char const *banner)
{
  void *local;

  printf("================ %s ==============\n", banner);

  printf("------ dropN(1000) -----\n");
  dropN(1000);
  GC_gcollect();

  printf("------- dropHalf(10) ----\n");
  local = dropHalf(10);
  GC_gcollect();

  printf("nullifying to drop other half..\n");
  local = NULL;
  GC_gcollect();
}


int main()
{
  // run tests with norma GC_malloc
  severalTests("GC_malloc");

  // simple test of my system
  useMyMark();               
  
  // same larger tests with my malloc
  whichMalloc = allocMyMark;
  severalTests("allocMyMark");

  return 0;
}

#endif // TEST_MYMARK
