// safec.c
// runtime library for ccure'd programs; mostly libc wrappers

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>       // __ctype_b

#ifdef _GNUCC
  // sm: stuff I had to add to get this to compile
  #include <unistd.h>    /* struct stat */
  #include <fcntl.h>     /* open; pulled up here to avoid implicit decl */
  #include <signal.h>    // signal
  #include <syslog.h>    // syslog, LOG_ERR(=3)
  #ifndef __CYGWIN__
    #define O_BINARY 0     /* doesn't exist in linux */
    #include <sys/mman.h>  // MAP_FAILED
    #include <glob.h>      // glob, glob_t, GLOB_xxx codes
  #endif
  #define _MAX_PATH 255  /* wouldn't begin to know where to properly find this */
#endif

// sanity check
#if defined(_DEBUG) && defined(RELEASELIB)
  #error _DEBUG and RELEASELIB are inconsistent
#endif

// some random stuff..; declares fp_fail
#include "safec.h"

// pull in declarations for garbage collector
#ifndef RELEASELIB
  #define GC_DEBUG
#endif
#include "gc/include/gc.h"

#if 0
  #define DP(arg) printf arg
#else
  #define DP(arg)
#endif

// 10/28/01: changed from defining 'malloc' etc to defining
// 'ccured_malloc', with a #define in safec.h, because there
// was a problem with the former under cygwin: when libc calls
// malloc (twice) before main runs, it the call to GC_MALLOC
// below would always segfault.  we tried some hackery where
// our malloc would try to avoid handling the first 2 calls,
// but that didn't help.  so we decided to avoid linker tricks
// altogether and rely in the C preprocessor instead to intercept
// malloc/free

// cause references below to malloc/free/realloc to use
// the garbage collector's versions instead
void *ccured_malloc(unsigned int sz)
{
  void *ret;
  
  // sm: I need one more byte allocated, since that's where I stash
  // the final, untouchable null; clearly we don't need this for *every*
  // allocation, only those to which char*s are created.. but I will
  // let someone else be smarter about choosing when to do this (of
  // course, that decision must accompany the decision not to *write*
  // that final null!)
  sz++;

  DP(("malloc(%d)\n", sz));
  ret = GC_MALLOC(sz);
  DP(("malloc(%d) yielded %p\n", sz, ret));
  
  // trying to track down bug in health
  //memset(ret, 0xAA, sz);

  return ret;
}

void *ccured_calloc(size_t x, size_t y)
{
  // GC_MALLOC always zeroes its result
  void *ret = GC_MALLOC(x*y);
  DP(("calloc(%d,%d) yielded %p\n", x, y, ret));
  return ret;
}

void ccured_free(void *p)
{
  DP(("free(%p)\n", p));
  GC_FREE(p);
}

void *ccured_realloc(void *p, unsigned int sz)
{
  void *ret = GC_REALLOC(p, sz);
  DP(("realloc(%p, %d) yielded %p\n", p, sz, ret));
  return ret;
}

// I intend to call this from user code to invoke
// the garbage collector, and verify it has accomplished
// some useful work; returns # of bytes freed
int ccured_explicit_gc()
{
  int before = GC_get_free_bytes();
  GC_gcollect();
  return GC_get_free_bytes() - before;
}

// declarations for malloc/free/realloc that were here before
extern void *malloc(unsigned int );
extern void free(void *);
extern void* realloc(void*, unsigned int);

// permit non-gc mode code to call this anyway
//int explicit_gc() { return 0; }


typedef struct fatp_char {
  char * _p ;
  void * _b ;
} fatp_char;

typedef struct fseqp_char {
  char *_p;
  char *_e;
} fseqp_char;

typedef struct fatp_char1 {
  char const  * _p ;
  void * _b ;
} fatp_char1;

typedef struct fatp_void {
  void  * _p ; 
  void * _b ;
} fatp_void;

typedef struct fseqp_void {  // for realloc_ff
  void *_p;
  void *_e;
} fseqp_void;

typedef struct fatp_void1 {
  void /*const*/ * _p ;      // sm: removed 'const' because triggers warnings -- why was it present?
  void * _b ;
} fatp_void1;

typedef struct seqp_void {
  void /*const*/ * _p ;
  void * _b ;
  void * _e ;
} seqp_void;

typedef struct seqp_char {
  char * _p ; 
  char * _b ;
  char * _e ;
} seqp_char;

typedef struct seqp_int {
  int * _p ;
  int * _b ;
  int * _e ;
} seqp_int;

typedef struct fatp_fatp_char {
  fatp_char * _p ;
  void * _b ;
} fatp_fatp_char;

typedef struct fseqp_fatp_char {
  fatp_char * _p ;
  void * _e ;
} fseqp_fatp_char;

typedef struct seq_fatp_char {
  fatp_char * _p ;
  void * _b ;
  void * _e;
} seq_fatp_char;


#ifdef _MSVC
#define I32 int
#endif
#ifdef _GNUCC
#define I32 int
#endif
#define U32 unsigned I32

#ifndef MIN
#define MIN(x,y) (((x) < (y)) ? (x) : (y))
#endif

/* Copy the tags going forward. Call this in conjuction with any memory 
 * copy. The psrc and pdst and size should be _exactly_ as passed to memcpy. 
 * NOTE THAT THIS MIGHT NOT WORK PROPERLY IN CASE OF OVERLAPPING AREAS 
 * (since the bits are read first and then written as a block. To model 
 * overlapping correctly we need to copy one tag at a time.) Note that this 
 * is used in a somewhat unusual way in realloc.  */
void CHECK_COPYTAGSFORW(void *bsrc, /* base of src */
                        unsigned int lensrc, /* len of src */
                        char* psrc, /* pointer of src */
                        void *bdst, /* base of dest */
                        unsigned int lendst, /* words in dest*/
                        char* pdst, /* pointer of dest */
                        unsigned int size /* Size in bytes 
                                           * of the memory 
                                           * being copied */)
{
#ifndef NO_TAGS
  TAGADDR stag, dtag;
  // Word starting address
  char    *alignSrc      = (char*)((U32)psrc & (~ 3)); 
  char    *alignDest     = (char*)((U32)pdst & (~ 3));
  char    *nextDestByte  = pdst + size; // Byte after the last copied
  int     bitsToCopy     = CHECK_NRTAGBITS(psrc, psrc + size);
  U32     srcTags   = 0;  // Keep some srcTags in here (always at LSB)
  int     nrSrcTags = 0;  // How many live bits we have in srcTags
  int     first = 1;      // A marker that says that we are copying the first

  /* If the alignment of the destination is different from that of the 
   * source then we just zero out all of the destination tags  */
  if(((U32)psrc ^ (U32)pdst) & 3) {
    CHECK_ZEROTAGS(bdst, lendst, pdst, size, 0);
    return;
  }
  // Assume for now the we are copying whole words always
  stag  = CHECK_FETCHTAGADDR(bsrc, lensrc, alignSrc);
  dtag  = CHECK_FETCHTAGADDR(bdst, lendst, alignDest);
  // While there are some more bits in the source
  while(bitsToCopy) {
    // See if we need to refill srcTags
    if(nrSrcTags == 0) {
      srcTags = (*stag.word) >> stag.bit;
      nrSrcTags = MIN(32 - stag.bit, bitsToCopy);
      if(first) { // The first batch of bits
        first = 0;
        // If we are doing partial word read and write
        if(((U32)pdst & 3)) {
          // We know that source and destination have the same alignment
          srcTags &= (~ TAGBITSMASK); // Zero out the tag for the first word
        }
        if(! (srcTags & TAGBITSMASK) || (void*)alignDest > bdst) {
           /* We are starting the write with a tag 0. Must zero the previous 
            * tag as well */
          // Pretend we have started behind with src
          stag.bit -= TAGBITS;
          if(stag.bit == - TAGBITS) {
            stag.word --;
            stag.bit  += 32;
          }
          dtag.bit -= TAGBITS;
          if(dtag.bit == - TAGBITS) {
            dtag.word --;
            dtag.bit += 32;
          }
          srcTags <<= TAGBITS; // Insert a 0 tag
          bitsToCopy += TAGBITS;
        }
      }
      if(bitsToCopy == nrSrcTags) { // This is the last batch
        if(((U32)nextDestByte & 3)  ||
           // We are ending the copy with the first half of a fat pointer
           ((srcTags >> (nrSrcTags - TAGBITS)) & TAGBITSMASK)) {
          // Zero out the tag for the last word
          srcTags &= (~ (TAGBITSMASK << (nrSrcTags - TAGBITS))); 
        }
      }
      stag.bit  += nrSrcTags;
      if(stag.bit >= 32) {
        stag.word ++;
        stag.bit -= 32;
      }
    }
    // Put some tags in the destination 
    {
      int tocopy  = MIN(nrSrcTags, 32 - dtag.bit);
      U32 mask    = tocopy < 32 ? (1 << tocopy) - 1 : (U32)-1;
      U32 towrite = (srcTags & mask) << dtag.bit;
      *dtag.word  = (*dtag.word & (~ (mask << dtag.bit))) | towrite;
      // Advance bit pointers
      nrSrcTags   -= tocopy;
      bitsToCopy  -= tocopy;
      srcTags     >>= tocopy;
      dtag.bit    += tocopy;
      if(dtag.bit == 32) { // Move on to the next destination pointer
        dtag.word ++;
        dtag.bit  = 0;
      }
    }
  }
#else
  FP_FAIL(FAIL_NOTAGS);
#endif
}

/* Copy the tags going backward. The psrc, pdst and size arguments should be 
 * exactly as passed to memcpy. The copying is done going backwards from the 
 * last word. NOTE THAT THIS MIGHT NOT WORK PROPERLY IN CASE OF OVERLAPPING 
 * AREAS (since the bits are read first and then written as a block. To 
 * model overlapping correctly we need to copy one tag at a time.) Note that 
 * this is used in a somewhat unusual way in realloc.  */
void CHECK_COPYTAGSBACK(void *bsrc, /* base of src */
                        unsigned int lensrc, /* len of src */
                        char* psrc, /* pointer of src */
                        void *bdst, /* base of dest */
                        unsigned int lendst, /* words in dest*/
                        char* pdst, /* pointer of dest */
                        unsigned int size)
{
#ifndef NO_TAGS
  TAGADDR stag, dtag;
  // Word starting address
  //  char    *alignSrc      = (char*)((U32)psrc & (~ 3)); 
  char    *nextSrcByte   = psrc + size; // Byte after the last copied
  char    *nextSrcWord   = (char*)((U32)(nextSrcByte + 3) & (~ 3));
  char    *alignDest     = (char*)((U32)pdst & (~ 3));
  char    *nextDestByte  = pdst + size; // Byte after the last copied
  char    *nextDestWord  = (char*)((U32)(nextDestByte + 3) & (~ 3));
  int     bitsToCopy     = CHECK_NRTAGBITS(psrc, psrc + size);
  U32     srcTags   = 0;  // Keep some srcTags in here (always at LSB)
  int     nrSrcTags = 0;  // How many live bits we have in srcTags
  int     first = 1;      // A marker that says that we are copying the first

  /* If the alignment of the destination is different from that of the 
   * source then we just zero out all of the destination tags  */
  if(((U32)psrc ^ (U32)pdst) & 3) {
    CHECK_ZEROTAGS(bdst, lendst, pdst, size, 0);
    return;
  }
  // Assume for now the we are copying whole words always
  stag  = CHECK_FETCHTAGADDR(bsrc, lensrc, nextSrcWord - 4);
  dtag  = CHECK_FETCHTAGADDR(bdst, lendst, nextDestWord - 4);
  // While there are some more bits in the source
  while(bitsToCopy > 0) {
    // See if we need to refill srcTags
    if(nrSrcTags == 0) {
      srcTags = (*stag.word);
      // Maybe all the source tags fit into one word
      nrSrcTags = stag.bit + TAGBITS;
      if(nrSrcTags > bitsToCopy) {
        srcTags >>= (nrSrcTags - bitsToCopy);
        nrSrcTags = bitsToCopy;
      }
      if(first) { // The first batch. Meaning the bits at highest addresses
        first = 0;
        // If partial word write
        if(((U32)nextDestByte & 3) ||
           // We end the copy with the first half of a fat pointer
           ((srcTags >> (nrSrcTags - TAGBITS)) & TAGBITSMASK)) {
          // Zero out the last tag
          srcTags &= (~ (TAGBITSMASK << (nrSrcTags - TAGBITS)));
        }
      }
      if(nrSrcTags == bitsToCopy) {
        // Last batch. Meaning the bits at lowest addr
        // If partial word write, zero out the tag
        if(((U32)pdst & 3)) {
          srcTags &= (~ TAGBITSMASK);
        }
      }
      stag.bit  = 32 - TAGBITS;
      stag.word --;
    }      
    // Put some tags in the destination
    {
      int tocopy;
      int doff;
      U32 mask, towrite;
    WriteToDest:
      if(nrSrcTags < dtag.bit + TAGBITS) {
        tocopy = nrSrcTags;
        doff   = dtag.bit + TAGBITS - nrSrcTags;
      } else {
        tocopy = dtag.bit + TAGBITS;
        doff   = 0;
      }
      mask    = tocopy < 32 ? (1 << tocopy) - 1 : (U32)-1;
      towrite = ((srcTags >> (nrSrcTags - tocopy)) & mask) << doff;
      *dtag.word  = (*dtag.word & (~ (mask << doff))) | towrite;
      // Advance bit pointers
      nrSrcTags   -= tocopy;
      dtag.bit   -= tocopy;
      bitsToCopy  -= tocopy;
      if(dtag.bit < 0) { // Move on to the next destination pointer
        dtag.word --;
        dtag.bit  = 32 - TAGBITS;
      }
      // If we have just copied the tag at the lowest address
      // And the first tag is 0, then we must zero the previous tag as well
      if(bitsToCopy == 0 && (! (srcTags & TAGBITSMASK)) &&
         (void*)alignDest > bdst) {
        tocopy = TAGBITS;
        srcTags = 0;
        nrSrcTags = TAGBITS;
        goto WriteToDest;
      }
    }
  }
#else
  FP_FAIL(FAIL_NOTAGS);
#endif
}

fatp_void malloc_w(unsigned int size) {
  fatp_void res;
  int  nrWords    = (size + 3) >> 2;
#ifndef NO_TAGS
  int  nrTagWords = CHECK_NRTAGWORDS(nrWords);
#else
  int  nrTagWords = 0;
#endif
  int  newSize    = ((1 + nrWords + nrTagWords) << 2);
  U32  *alloc     = (U32*)malloc(newSize);
  if(! alloc) {
    FP_FAIL(FAIL_MALLOC);
  }
  * alloc ++   = nrWords;   // Write the length.
  res._p = (void*)alloc;    // This is the start of the data
  res._b = (void*)alloc;
  alloc += nrWords;         // Skip the data section
#ifndef NO_TAGS
  for(  ; nrTagWords; nrTagWords --) { // Write the tags
    * alloc ++ = 0;
  }
#endif
  return res;
}

fatp_void malloc_fseq_f(unsigned size) {
  fatp_void res;
  res._p = malloc(size);
  res._b = (void*)((char*)res._p + ((size + 3) & (~ 3)));
  return res;
}

void free_safe(void *x) {
  free(x);
}

// sm: this little hack is needed because George calls
// this free on __HEAPIFY'd areas, but the value passed
// is 4 greater than what should be passed to free
void free_minus4(void *x) {
  free((void*)( (int)x - 4 ));
}

void free_w(fatp_void x) {
  int nrWords;
  U32 *p = (U32*)x._p;
  if(! x._b || x._b != p) {
    FP_FAIL(FAIL_FREE);
  }
  // Now find the real start
  // (sm: there's been some debate about whether these -1's
  // should be here.. but for now I think they should..)
  nrWords = *(p-1);
  free((void*)(p-1));
}

void free_f(fatp_void1 x) { 
    // weimer: this is very wrong!
    // free(x._b);
}

void free_q(seqp_void x) {
  free(x._b);
}


extern void* calloc(unsigned int, unsigned int);
fatp_void calloc_w(unsigned int count, unsigned int osize) {
  fatp_void res;
  int  size       = osize * count;
  int  nrWords    = (size + 3) >> 2;
#ifndef NO_TAGS
  int  nrTagWords = ((nrWords + 15) >> 4);
#else
  int  nrTagWords = 0;
#endif
  int  newSize    = ((1 + nrWords + nrTagWords) << 2);
  U32  *alloc     = (U32*)calloc(newSize, 1);
  * alloc ++ = nrWords; // Write the length
  res._p = (void*)alloc;
  res._b = (void*)alloc;
  // No need to write the tags, since the area has been zeroed 
  return res;
}

fatp_void calloc_fseq_f(unsigned int count, 
                        unsigned int osize) {
  fatp_void res;
  res._p = calloc(count, osize);
  res._b = (void*)((char*)res._p + (((osize * count) + 3) & (~ 3)));
  return res;
}


void *calloc_rbnode(unsigned int nrelem, unsigned int osize) {
  char *res = (char*)calloc(nrelem, osize);
  char *where = res + 5 * 4;
  int i;
  //  printf("Allocating %d RBNode's, each of size %d (24 + %d)\n",
  //         nrelem, osize, osize - 24);
  for(i=0;i<(int)nrelem;i++, where += osize) {
    *((int*)where) = osize - 5 * 4;
  }
  return (void*)res;
}

fatp_void realloc_ww(fatp_void orig, unsigned int newsz) {
  int  nrWords    = (newsz + 3) >> 2;
#ifndef NO_TAGS
  int  nrTagWords = ((nrWords + 15) >> 4);
#else
  int  nrTagWords = 0;
#endif
  int  newSize    = ((1 + nrWords + nrTagWords) << 2);
  int  origNrWords;
  U32  *origbase  = (U32*)orig._b;
  U32  *alloc;

  if(! orig._p) { // NULL argument. A straight malloc
    if(origbase) {
      FP_FAIL(FAIL_REALLOC);
    }
    return malloc_w(newsz);
  }
  if(! origbase || orig._p != origbase) {
    FP_FAIL(FAIL_REALLOC);
  }
  if(newsz == 0) { // A free operation. Pass it along.
    free_w(orig);
    orig._p = NULL;
    orig._b = NULL;
    return orig;
  }
  // The old number of words
  origNrWords      = CHECK_FETCHLENGTH(orig._p, origbase); 
  if(origNrWords == nrWords) {
    return orig; // No change
  }
  if(nrWords < origNrWords) {
    // Shrinking. Move the tags to lower addresses before we realloc.
    // Start with the first one!!!
#ifndef NO_TAGS
//    origTagAddr = origbase + origNrWords;
//    newTagAddr  = origbase + nrWords;
//    for(i = nrTagWords ; i; i --) { // Copy the tags
//      * newTagAddr ++ = * origTagAddr ++;
//    }
//    // Zero out some bits in the last tag word !!
//    tag = * -- newTagAddr;
//    tag &= ((((U32)1) << ((nrWords & 0xF) << 1)) - 1);
//    * newTagAddr = tag;
    CHECK_COPYTAGSFORW(origbase, origNrWords, (char*)origbase,
                       origbase, nrWords, (char*)origbase,
                       nrWords << 2);
#endif /* USE_TAGS */
    // Now realloc. The data and the tags will be copied
    alloc = realloc(origbase - 1, newSize);
    if(! alloc) {
      FP_FAIL(FAIL_REALLOC_SHRINK);
    }
  } else {
    // Growing. Realloc first and then move the tags to higher addresses
    alloc = realloc(origbase - 1, newSize);
    if(! alloc) {
      FP_FAIL(FAIL_REALLOC_GROW);
    }
    // Move the tags from last to first !!!
#ifndef NO_TAGS    
//    origNrTagWords = (origNrWords + 15) >> 4;
//    origTagAddr = alloc + origNrWords + origNrTagWords;
//    newTagAddr  = alloc + nrWords + origNrTagWords;
//    for(i = origNrTagWords ; i; i --) { // Copy the tags
//      * -- newTagAddr = * -- origTagAddr;
//    }
//    // Prepare to zero out the rest
//    newTagAddr  = alloc + nrWords + origNrTagWords;
//    // Zero out some bits in the last tag word !!
//    * newTagAddr &= ((((U32)1) << ((origNrWords & 0xF) << 1)) - 1);
//    for(i = nrTagWords - origNrTagWords; i; i --) {
//      * ++ newTagAddr = (U32)0;
//    }
    CHECK_COPYTAGSBACK(alloc + 1, origNrWords, (char*)(alloc + 1),
                       alloc + 1, nrWords, (char*)(alloc + 1),
                       origNrWords << 2);
    
#endif
    // sm: bugfix: need to zero out the cleared space, otherwise the tags
    // left behind lead to bad stuff
    memset(alloc+1+origNrWords, 0, (nrWords-origNrWords) * 4);    
  }
  // Set the new length
  *alloc = nrWords;
  alloc ++;  // Skip the length
  orig._p = (void*)alloc;
  orig._b = (void*)alloc;
  return orig;
}

typedef struct fatp_FILE {
  FILE * _p ;
  void * _b ;
} fatp_FILE;

char* failMessages[] =
{ "Below stack address",
  "Stack address",
  "Lbound",
  "Ubound",
  "Function pointer",  
  "Decrement FSEQ",
  "Null pointer",
  "Null string",
  "Non-terminated string",
  "Unaligned pointer",
  "Malloc",
  "Realloc",
  "Free",
  "Realloc (shrink)",
  "Realloc (grow)",
  "Non-pointer",
  "Scalar log",
  "Null string in strlen",
  "variable argument",
  "call to va_arg without call to va_start",
  "too many calls to va_arg",
  "type mismatch between caller and use of va_arg function",
  "unrecognized format character"
  };

NORETURN  void fp_fail_str(char *str  FP_FAIL_EXTRA_PARAMS) 
{
  #ifdef FP_FAIL_IS_VERBOSE
    #define MSGFMT "Failure at %s:%d: %s"
  #else
    #define MSGFMT "Failure: %s"
  #endif

  fprintf(stderr, MSGFMT "\n"  FP_FAIL_EXTRA_ARGS, str);
  fflush(stderr);

  #if defined(_GNUCC) && !defined(NO_FPFAIL_SYSLOG)
    syslog(LOG_ERR, MSGFMT  FP_FAIL_EXTRA_ARGS, str);
  #endif
  #if defined(_MSVC) && defined(_DEBUG)
    _asm { int 3 }
  #else
    abort();
  #endif

  #undef MSGFMT
}

NORETURN  void fp_fail(int msgId  FP_FAIL_EXTRA_PARAMS)
{
  fp_fail_str(failMessages[msgId]  FP_FAIL_EXTRA_ARGS);
}


 
fatp_FILE fopen_www(fatp_char1 fname, fatp_char1 mode) {
  fatp_FILE res; 
  res._p = fopen(fname._p, mode._p);
  res._b = (void*)0; // We should not be reading from here !!!
  return res;
}


fatp_FILE fopen_wss(const char* fname, const  char * mode) {
  fatp_FILE res; 
  res._p = fopen(fname, mode);
  res._b = (void*)0; // We should not be reading from here !!!
  return res;
}

int fflush_w(fatp_FILE f) {
  // unsigned int l = CHECK_FETCHLENGTH(f._b);
  // don't check the bounds because FILE * do not have lengths
  //  CHECK_CHECKBOUNDS(f._b, l, f._p, sizeof(FILE*));
  return fflush(f._p);
}

int fclose_w(fatp_FILE f) {
  // don't check the bounds because FILE * do not have lengths
  // unsigned int l = CHECK_FETCHLENGTH(f._b);
  // CHECK_CHECKBOUNDS(f._b, l, f._p, sizeof(FILE*));
  return fclose(f._p);
}
               
 //#ifdef _MSVC
 //// A major hack to intercept accesses to _iob on MSVC
 //void * get_iob(int n) {
 //  return (void*)& _iob[n];
 //}
 //fatp_void get_iob_w(int n) {
 //  fatp_void res;
 //  res._p = (void*)& _iob[n];
 //  res._b = (void*)0;
 //  return res;
 //}
 //
fatp_void get_stdin_w() {
  fatp_void res;
  res._p = (void*)stdin;
  res._b = 0;
  return res;
}
fatp_void get_stdout_w() {
  fatp_void res;
  res._p = (void*)stdout;
  res._b = 0;
  return res;
}
fatp_void get_stderr_w() {
  fatp_void res;
  res._p = (void*)stderr;
  res._b = 0;
  return res;
}

#ifdef _GNUCC

/* Appears in putc for GCC */
#ifdef __CYGWIN__
int __swbuf_w(int c, fatp_void fl) {
  return __swbuf(c, (FILE*)fl._p);
}
/* Appears in getc for GCC */
int __srget_w(fatp_void fl) {
  return __srget((FILE*)fl._p);
}
#endif // 0

// sm: internally, my putc uses _IO_putc
// see /usr/include/features.h to find glibc version numbers
#if defined(__GLIBC__) && __GLIBC__ == 2 && __GLIBC_MINOR__ == 1 || \
    defined(__GLIBC__) && __GLIBC__ == 2 && __GLIBC_MINOR__ == 0 || \
    defined(__GLIBC__) && __GLIBC__ == 2 && __GLIBC_MINOR__ == 2
int _IO_putc_w(int c, fatp_void fl) {
  return _IO_putc(c, (FILE*)fl._p);
}

int _IO_getc_w(fatp_void fl) {
  return _IO_getc((FILE*)fl._p);
}

#endif

#endif // !_MSVC

int putc_w(int c, fatp_void fl) {
  return putc(c, (FILE*)fl._p);
}

int getc_w(fatp_void fl) {
  // sm: shouldn't there be some error checking here?
  return getc((FILE*)fl._p);
}

// sm: similar (no error checking..) wrappers for scott/putc
int fputc_w(int c, fatp_void fl) {
  return fputc(c, (FILE*)fl._p);
}

int fgetc_w(fatp_void fl) {
  return fgetc((FILE*)fl._p);
}

int puts_w(fatp_char s) {
  return puts((char*)s._p);
}


// sm: the '-1's below are a major problem because not every call to fread_ww is
// coming from a char[] buffer.  we need a different way to do the ensure-0 thing
// than this.  I am removing the -1 (conditionally on this #define) until we
// can rethink the implementation of strings and null terminators
//#define NULLTERM_ADJ (-1)         // this puts back the -1
#define NULLTERM_ADJ 0              // this is my change

size_t fread_ww(fatp_char buff, size_t size, size_t count, fatp_FILE fl) {
  // Fetch the length in words
  unsigned int blenwrds = CHECK_FETCHLENGTH(buff._p, buff._b);
  int readlen = (int)(blenwrds << 2) - (buff._p - (char*)buff._b) + NULLTERM_ADJ;
  size_t requested = size * count;
  size_t res;

  if(readlen < (int)requested) {
    FP_FAIL(FAIL_UBOUND);
  }
  res = fread(buff._p, size, count, fl._p);
  // Clean the tags
  CHECK_ZEROTAGS(buff._b, blenwrds, buff._p, readlen, 0);
  return res;
}

size_t fread_ws(fatp_char buff, size_t size, size_t count, FILE *fl) {
  // Fetch the length in words
  unsigned int blenwrds = CHECK_FETCHLENGTH(buff._p, buff._b);
  int readlen = (int)(blenwrds << 2) - (buff._p - (char*)buff._b) + NULLTERM_ADJ;
  size_t requested = size * count;
  size_t res;

  if(readlen < (int)requested) {
    FP_FAIL(FAIL_UBOUND);
  }
  res = fread(buff._p, size, count, fl);
  // Clean the tags
  CHECK_ZEROTAGS(buff._b, blenwrds, buff._p, readlen, 0);
  return res;
}

size_t fread_fs(fseqp_char buff, size_t size, size_t count, FILE *fl) {
  size_t requested = size * count;

  CHECK_NULL(fl);
  CHECK_NULL(buff._p);
  CHECK_UBOUND(buff._e, (void*)buff._p, requested);
  return fread(buff._p, size, count, fl);
  
}
size_t fwrite_ww(fatp_char buff, size_t size, size_t count, fatp_FILE fl) {
  // Fetch the length in words
  unsigned int blenwrds = CHECK_FETCHLENGTH(buff._p, buff._b);
  int writelen = (int)(blenwrds << 2) - (buff._p - (char*)buff._b) + NULLTERM_ADJ;
  size_t requested = size * count;
  size_t res;

  if(writelen < (int)requested) {
    FP_FAIL(FAIL_UBOUND);
  }
  res = fwrite(buff._p, size, count, fl._p);
  return res;
}

size_t fwrite_qs(seqp_char buff, size_t size, size_t count, FILE *fl) {
  // Fetch the length in words
  size_t requested = size * count;

  CHECK_NULL(fl);
  CHECK_BOUNDS(buff._b, buff._e, (void*)buff._p, requested);
  return fwrite(buff._p, size, count, fl);
}

size_t fwrite_fs(fseqp_char buff, size_t size, size_t count, FILE *fl) {
  // Fetch the length in words
  size_t requested = size * count;

  CHECK_NULL(fl);
  CHECK_NULL(buff._p);
  CHECK_UBOUND(buff._e, (void*)buff._p, requested);
  return fwrite(buff._p, size, count, fl);
}


fatp_char fgets_wws(fatp_char buffer, int size, FILE *stream) {
  // Fetch the length in words
  unsigned int blenwrds = CHECK_FETCHLENGTH(buffer._p, buffer._b);
  int readlen = (int)(blenwrds << 2) - (buffer._p - (char*)buffer._b) + NULLTERM_ADJ;
  fatp_char res;
  if(size < 0) {
    FP_FAIL(FAIL_LBOUND);
  }
  if(readlen < size) {
    FP_FAIL(FAIL_UBOUND);
  }
  res._p = fgets(buffer._p, readlen, stream);
  res._b = buffer._b;
  // Clean the tags
  CHECK_ZEROTAGS(buffer._b, blenwrds, buffer._p, readlen, 0);
  return res;
}

/* fgets and gets have different behavior concerning \n. Since we want gets
 * to use fgets, we write a separate version of fgets which replaces the \n
 * with \0, as gets would do */

fatp_char fgets_nonewline_wws(fatp_char buffer, int size, FILE *stream) {
  // Fetch the length in words
  unsigned int blenwrds = CHECK_FETCHLENGTH(buffer._p, buffer._b);
  int readlen = (int)(blenwrds << 2) - (buffer._p - (char*)buffer._b) + NULLTERM_ADJ;
  fatp_char res;
  if(size < 0) {
    FP_FAIL(FAIL_LBOUND);
  }
  if(readlen < size) {
    FP_FAIL(FAIL_UBOUND);
  }
  res._p = fgets(buffer._p, readlen, stream);

  {
    char *tmp = res._p;
    while (*tmp != '\n' && *tmp != '\0') tmp++;
    *tmp ='\0';
  }
  res._b = buffer._b;
  // Clean the tags
  CHECK_ZEROTAGS(buffer._b, blenwrds, buffer._p, readlen, 0);
  return res;
}


/* It is extremely dangerous to use gets, so we turn it into a fgets */
fatp_char gets_ww(fatp_char buffer) {
  // Fetch the length in words
  unsigned int blenwrds = CHECK_FETCHLENGTH(buffer._p, buffer._b);
  int readlen = (int)(blenwrds << 2) - (buffer._p - (char*)buffer._b) + NULLTERM_ADJ;

  return fgets_nonewline_wws(buffer, readlen, stdin);
}

fatp_char fgets_www(fatp_char buffer, int size, fatp_FILE stream) {
  return fgets_wws(buffer, size, (FILE*)stream._p);
}

typedef struct fatp_struct__stat {
  struct _stat * _p ;
  void * _b ;
} fatp_struct__stat;

int  __cdecl  fstat_w(int fid ,  fatp_struct__stat buff ) {
  void *bb = buff._b;
  void *bp = buff._p;
  unsigned int blen = CHECK_FETCHLENGTH(bp, bb);
  unsigned int size = sizeof(struct _stat);

  CHECK_BOUNDS_LEN(bb, blen, bp, size);
#ifndef NO_TAGS
  CHECK_ZEROTAGS(bb, blen, bp, size, 0);
#endif
  return fstat(fid, bp);
}

int  __cdecl  stat_ww(fatp_char1 path ,  fatp_struct__stat  buff) {
  void *bb = buff._b;
  void *bp = buff._p;
  unsigned int blen = CHECK_FETCHLENGTH(bp, bb);
  unsigned int size = sizeof(struct _stat);
  unsigned int mp = CHECK_STRINGMAX((void*)path._p, path._b);

  CHECK_BOUNDS_LEN(bb, blen, bp, size);
#ifndef NO_TAGS
  CHECK_ZEROTAGS(bb, blen, bp, size, 0);
#endif

  CHECK_STRINGLEN(path._p, mp);

  return stat(path._p, bp);
}


int  __cdecl  open_w( fatp_char1 path , int oflag , int pmode) {
  unsigned int max = CHECK_STRINGMAX((void*)path._p, path._b);
  CHECK_STRINGLEN(path._p, max);
  return open(path._p, oflag, pmode);
}

int  __cdecl  unlink_w(fatp_char1  path) {
  unsigned int max = CHECK_STRINGMAX((void*)path._p, path._b);
  CHECK_STRINGLEN(path._p, max);
  return unlink(path._p);
}

int fileno_w(fatp_FILE f) {
  // Don't check length because f is not allocated by us
//  unsigned int l = CHECK_FETCHLENGTH(f._p, f._b);
//  CHECK_CHECKBOUNDS(f._b, l, f._p, sizeof(FILE*));
  return fileno(f._p);
}

int  __cdecl  setjmp_w(fatp_void jbuf) {
  return setjmp(jbuf._p);
}

#if defined(_GNUCC) && !defined(__CYGWIN__)
int  __cdecl  __sigsetjmp_w(fatp_void jbuf, int __savemask) {
  return __sigsetjmp(jbuf._p, __savemask);
}
#endif

void  __cdecl longjmp_w(fatp_void jbuf , int res ) {
  longjmp(jbuf._p, res);
}

int  __cdecl  setjmp_q(seqp_void jbuf) {
  CHECK_BOUNDS(jbuf._b, jbuf._e, (void*)jbuf._p, sizeof(int));
  return setjmp((void*)jbuf._p);
}
int  __cdecl  _setjmp_q(seqp_void jbuf) {
  return setjmp_q(jbuf);
}

#if defined(_GNUCC) && !defined(__CYGWIN__)
int  __cdecl  __sigsetjmp_q(seqp_void jbuf, int __savemask) {
  CHECK_BOUNDS(jbuf._b, jbuf._e, (void*)jbuf._p, sizeof(int));
  return __sigsetjmp((void*)jbuf._p, __savemask);
}
#endif


void  __cdecl longjmp_q(seqp_void jbuf , int res ) {
  CHECK_BOUNDS(jbuf._b, jbuf._e, (void*)jbuf._p, sizeof(int));
  longjmp((void*)jbuf._p, res);
}

double  __cdecl  atof_w( fatp_char1 str ) {
  unsigned int max = CHECK_STRINGMAX((void*)str._p, str._b);
  CHECK_STRINGLEN(str._p, max);
  return atof(str._p);
}

int  __cdecl  atoi_w(fatp_char1  str) {
  unsigned int max = CHECK_STRINGMAX((void*)str._p, str._b);
  CHECK_STRINGLEN(str._p, max);
  return atoi(str._p);
}

long  __cdecl  atol_w(fatp_char1 str) {
  unsigned int max = CHECK_STRINGMAX((void*)str._p, str._b);
  CHECK_STRINGLEN(str._p, max);
  return atol(str._p);
}   


#ifdef _GNUCC
//extern int read();
#else
extern int __cdecl read(int fid, void *buff, unsigned int size);
#endif
 
int read_w(int fid, fatp_void buff, unsigned int size) {
  void *bb = buff._b;
  void *bp = buff._p;
  unsigned int blen = CHECK_FETCHLENGTH(bp, bb);

  CHECK_BOUNDS_LEN(bb, blen, bp, size);
#ifndef NO_TAGS
  CHECK_ZEROTAGS(bb, blen, bp, size, 0);
#endif
  return read(fid, bp, size);
}

int read_f(int fid, fseqp_void buff, unsigned int size) {
  CHECK_NULL(buff._p);
  CHECK_UBOUND(buff._e, (void*)buff._p, size);
  return read(fid, buff._p, size);
}

int read_q(int fid, seqp_void buff, unsigned int size) {

  CHECK_BOUNDS(buff._b, buff._e, (void*)buff._p, size);
  return read(fid, buff._p, size);
}


#ifdef _GNUCC
//extern int write();
#else
extern int __cdecl write(int fid, void /* const */ *buff, unsigned int size);
#endif

 
int write_w(int fid, fatp_void buff, unsigned int size) {
  void *bb = buff._b;
  void *bp = (void*)buff._p;
  unsigned int blen = CHECK_FETCHLENGTH(bp, bb);

  CHECK_BOUNDS_LEN(bb, blen, bp, size);
  return write(fid, bp, size);
}

int write_f(int fid, fseqp_void buff, unsigned int size) {
  CHECK_UBOUND(buff._e, buff._p, size);
  return write(fid, buff._p, size);
}

int write_q(int fid, seqp_void buff, unsigned int size) {     // sm
  CHECK_BOUNDS(buff._b, buff._e, (void*)buff._p, size);
  return write(fid, buff._p, size);
}


// sm: this function is bad, right?  it lets the user
// derive a safe pointer from a wild pointer, right?
void qsort_wsss(fatp_void base, size_t nmemb,
                size_t size,
                int (* compar)(const void *, const void *)) {
  unsigned int blenwrds = CHECK_FETCHLENGTH(base._p, base._b);
  CHECK_BOUNDS_LEN(base._b, blenwrds, base._p, size * nmemb);

  // We should change the tags, or at least ZERO them out
  qsort(base._p, nmemb, size, compar);
}

/* weimer : Wed Dec 12 13:12:34 PST 2001 */
void qsort_fsss(fseqp_void __base,
	unsigned int __nmemb,
	unsigned int __size,
	int (* __compar)(void * arg0 , void * arg1 )) {
  CHECK_NULL(__base._p);
  CHECK_UBOUND(__base._e, (void*)__base._p, __size * __nmemb);
  qsort(__base._p, __nmemb, __size, 
        // sm: make gcc happy with a cast
        (int (*)(const void*, const void*))__compar);
}

// this isn't reentrant... threads problem...
// I need these to thunk back to user's sort function
  // if we really need reentrant code, we can use a trick: pass
  // a pointer to a malloc'd structure as the 1st arg to 'qsort',
  // and in our thunk we.. no wait, that won't work.. oh well
void *qsort_wsww_base;
int (*qsort_wsww_compar)(fatp_void, fatp_void);

int qsort_wsww_thunk(void const *left, void const *right)
{
  // map the 'left' and 'right' lean pointers to the
  // fat pointers we need
  fatp_void fatleft, fatright;
  fatleft._p = (void*)left;      // cast away constness since we have no runtime enforcement of it..
  fatleft._b = qsort_wsww_base;
  fatright._p = (void*)right;
  fatright._b = qsort_wsww_base;

  // and call the user-supplied sorting function, which
  // expects fat pointers
  return qsort_wsww_compar(fatleft, fatright);
}

// sm: copied and modified from above
// this function's test case is test/small2/qsort.c
void qsort_wsww(fatp_void base, size_t nmemb,
                size_t size,
                int (* compar)(fatp_void, fatp_void)) {
  unsigned int blenwrds = CHECK_FETCHLENGTH(base._p, base._b);
  CHECK_BOUNDS_LEN(base._b, blenwrds, base._p, size * nmemb);

  // save the pertinent values
  qsort_wsww_base = base._b;
  qsort_wsww_compar = compar;

  // We should change the tags, or at least ZERO them out
  qsort(base._p, nmemb, size, qsort_wsww_thunk);
}


// HERE
int ferror_w(fatp_FILE fp)
{
  return ferror(fp._p);
}





/* ********** STRINGS ************* */
int strncmp_ww(fatp_char s1,
               fatp_char s2,  unsigned int size) {
  unsigned int m1 = CHECK_STRINGMAX(s1._p, s1._b);
  unsigned int m2 = CHECK_STRINGMAX(s2._p, s2._b);
  unsigned int m  = MIN(size, MIN(m1, m2));
  int i;
  char const *x1 = s1._p;
  char const *x2 = s2._p;
  for(i = m ; i > 0; x1 ++, x2 ++, i --) {
    unsigned char c1 = *x1; unsigned char c2 = *x2;
    if(c1 == c2) {
      if(!c1) return 0;
      continue;
    }
    // The shorter string will be "smaller"
    return (c1 < c2) ? -1 : 1;
  } 
  // They are equal so far
  if(m == size) {
    return 0; 
  }
  FP_FAIL(FAIL_STRINGBOUND);
  return 0;  
}

 
int strcmp_ww(fatp_char s1, fatp_char s2) {
  return strncmp_ww(s1, s2, 0x7FFFFFFF); 
}


 
fatp_char strcpy_www(fatp_char dest, fatp_char src) {
  int md = CHECK_STRINGMAX(dest._p, dest._b);
  int ms = CHECK_STRINGMAX(src._p, src._b);
  int m  = MIN(ms, md);
  char *d = dest._p;
  char const *s = src._p;
  for( ; m > 0; m --) {
    if(! *s) {
      break;
    } 
    * d ++ = * s ++;
  }
  if(m <= 0) {
    FP_FAIL(FAIL_STRINGBOUND);
  }
  * d = '\0';
  // Now copy the tags
#ifndef NO_TAGS
  {
    unsigned int slen = CHECK_FETCHLENGTH((void*)src._p, src._b);
    unsigned int dlen = CHECK_FETCHLENGTH((void*)dest._p, dest._b);
    CHECK_COPYTAGSFORW(src._b, slen, (char*)src._p,
                       dest._b, dlen, (char*)dest._p, s - src._p);
  }
#endif
  return dest;
}

#ifdef _MSVC
unsigned int strlen(char const *);
#endif

unsigned int strlen_w(fatp_char s) {
  int m = CHECK_STRINGMAX(s._p, s._b);
  const char *x = (const char*)s._p;
  unsigned int l = 0;
  for( ; m > 0; x ++, m --, l ++) {
    if(! *x) return l;
  }
  FP_FAIL(FAIL_STRINGBOUND);
  return 0;
}



fatp_char  strchr_ww(fatp_char s, int  chr) {
  int m = CHECK_STRINGMAX(s._p, s._b);
  const char *x = (const char*)s._p;
  fatp_char res;
  unsigned int l = 0;
  for( ; m > 0; x ++, m --, l ++) {
    if(! *x) { res._b = 0; res._p = 0; return res; }
    if(*x == chr) {
      res._b = s._b;
      res._p = (char*)x;
      return res;
    }
  }
  FP_FAIL(FAIL_STRINGBOUND);
  return res;
}


fseqp_char strrchr_fs(char * string, int chr) {
  fseqp_char res;
  res._p = strrchr(string, chr);
  res._e = string + strlen(string);
  return res;
}


// sm: for m88k
fseqp_char strchr_ff(fseqp_char str, int chr) {
  fseqp_char res;

  if (str._p >= str._e) {
    FP_FAIL(FAIL_UBOUND);
  }

  res._p = strchr(str._p, chr);
  if (res._p) {
    res._e = str._e;
  }
  else {
    res._e = NULL;
  }
  return res;
}


// sm: for m88k, now it wants to use sequence pointers..
seqp_char strchr_qq(seqp_char str, int chr) {
  seqp_char res;

  if (!( str._b <= str._p && str._p <= str._e )) {
    FP_FAIL(FAIL_UBOUND);
  }

  res._p = strchr(str._p, chr);
  if (res._p) {
    res._b = str._b;
    res._e = str._e;
  }
  else {
    res._b = NULL;
    res._e = NULL;
  }
  return res;
}


fatp_char strcat_www(fatp_char dest,  fatp_char src) {
  int msrc = CHECK_STRINGMAX(src._p, src._b);
  int mdst = CHECK_STRINGMAX(dest._p, dest._b);
  const char *s = (const char*)src._p;
  char *d       = (char*)dest._p;
  char *startd;
  for(; mdst > 0; mdst --, d ++) {
    if(! *d) break;
  }
  startd = d;
  mdst = MIN(mdst, msrc);
  for(; mdst > 0; mdst --) {
    if(! (* d ++ = * s ++)) { break; }
  }
  if(mdst <= 0) {
    FP_FAIL(FAIL_STRINGBOUND);
  }
  // Now copy the tags
#ifndef NO_TAGS
  {
    unsigned int slen = CHECK_FETCHLENGTH((void*)src._p, src._b);
    unsigned int dlen = CHECK_FETCHLENGTH((void*)dest._p, dest._b);
    CHECK_COPYTAGSFORW(src._b, slen, (char*)src._p, 
                       dest._b, dlen, (char*)startd,
                       s - 1 - src._p); // We did s++
  }
#endif
  return dest;
}


#ifdef _MSVC
  char * strcpy(char *, char const *);
#else
  #if !defined(_RELEASE) && !defined(strcpy)
    // sm: sometimes 'strcpy' is a macro, which will cause
    // the following to break
    char * strcpy(char *, const char *);
  #endif
#endif

fatp_char  strdup_ww(fatp_char  src) {
  int msrc = CHECK_STRINGMAX(src._p, src._b);
  int l = 0;
  const char *s = (const char*)src._p;
  fatp_void res;
  for(; msrc > 0; msrc --, s ++, l ++) {
    if(! *s) break;
  }
  if(msrc <= 0) {
    FP_FAIL(FAIL_STRINGBOUND);
  }
  res = calloc_w(1, l + 1);
  if(! res._p) {
    return *(fatp_char*)&res;
  }
  strcpy(res._p, src._p);
  return *(fatp_char*)&res;
}


// verify that the string pointed to by 'src' has a null terminator
// somewhere within the bytes legal to read from it; return the
// length of the string (number of bytes before the nullterm)
static int check_nullterm_fseq(fseqp_char src)
{
  char const *p;

  // verify there's a zero byte between src._p and src._e
  // what is _e?  apparently it's the last byte the user *can* read/write ..
  // except it is supposed to be first byte that can *not* be accessed ..
  for (p = src._p; p <= src._e; p++) {
    if (*p == 0) {
      return p - src._p;
    }
  }
  FP_FAIL(FAIL_STRINGBOUND);
  return 0;    // silence warning
}

fseqp_char strdup_ff(fseqp_char src)
{
  int len = check_nullterm_fseq(src);

  // allocate space for the result
  fseqp_char ret;
  ret._p = malloc(len + 1);       // include space for nullterm
  ret._e = ret._p + (len + 1);    // end is just beyond last byte alloc'd
  memcpy(ret._p, src._p, len+1);  // copy the bytes!

  return ret;
}


// same as check_nullterm_fseq, but for seq
static int check_nullterm_seq(seqp_char src)
{
  char const *p;

  // verify pointer is at least _b
  if (src._b > src._p) {
    FP_FAIL(FAIL_LBOUND);
  }

  // verify there's a zero byte between src._p and src._e
  for (p = src._p; p < src._e; p++) {   // correcting _e bug above?
    if (*p == 0) {
      return p - src._p;
    }
  }
  FP_FAIL(FAIL_STRINGBOUND);
  return 0;    // silence warning
}

fseqp_char strdup_fq(seqp_char src)
{
  int len = check_nullterm_seq(src);

  // allocate space for the result
  fseqp_char ret;
  ret._p = malloc(len + 1);       // include space for nullterm
  ret._e = ret._p + (len + 1);    // end is just beyond last byte alloc'd
  memcpy(ret._p, src._p, len+1);  // copy the bytes!

  return ret;
}


extern char* strerror(int);
 
fatp_char strerror_w(int  errno) {
  fatp_char res;
  res._p = strerror(errno);
  res._b = (void*)0;  // !!! No tags or length in there !
  return res;
}


int  memcmp_ww(fatp_void dest,
               fatp_void  src,
               unsigned int size) {
  void *destb = dest._b;
  void *destp = (void*)dest._p;
  unsigned int destlen ;
  void *srcb = src._b;
  void *srcp = (void*)src._p;
  unsigned int srclen;

  if(size == 0) return 0;
  srclen = CHECK_FETCHLENGTH(srcp, srcb);
  destlen = CHECK_FETCHLENGTH(destp, destb);
  CHECK_BOUNDS_LEN(destb, destlen, destp, size);
  CHECK_BOUNDS_LEN(srcb, srclen, srcp, size);
  return memcmp(destp, srcp, size);
}

#ifdef _MSVC
void* memcpy(void *, void const *, unsigned int);
#endif


fatp_void  __cdecl  memcpy_www(fatp_void   dest,
                               fatp_void  src,
                               unsigned int size ) {
  void *destb = dest._b;
  void *destp = dest._p;
  unsigned int destlen = CHECK_FETCHLENGTH(destp, destb);
  void *srcb = src._b;
  void *srcp = (void*)src._p;
  unsigned int srclen = CHECK_FETCHLENGTH(srcp, srcb);

  CHECK_BOUNDS_LEN(destb, destlen, destp, size);
  CHECK_BOUNDS_LEN(srcb, srclen, srcp, size);
  // Memcpy copies from the start
#ifndef NO_TAGS
  CHECK_COPYTAGSFORW(srcb, srclen, srcp,
                     destb, destlen, destp, size);
#endif
  memcpy(destp, srcp, size);
  return dest;
}

fseqp_void  memcpy_fff(fseqp_void   dest,
                       fseqp_void  src,
                       unsigned int size ) {
  CHECK_UBOUND(dest._e, dest._p, size);
  CHECK_UBOUND(src._e, src._p, size);
  memcpy(dest._p, src._p, size);
  return dest;
}

void *  memcpy_sff(fseqp_void   dest,
                       fseqp_void  src,
                       unsigned int size ) {
  CHECK_UBOUND(dest._e, dest._p, size);
  CHECK_UBOUND(src._e, src._p, size);
  memcpy(dest._p, src._p, size);
  return NULL;
}

void *  memmove_sff(fseqp_void   dest,
                       fseqp_void  src,
                       unsigned int size ) {
  CHECK_UBOUND(dest._e, dest._p, size);
  CHECK_UBOUND(src._e, src._p, size);
  memmove(dest._p, src._p, size);
  return NULL;
}

seqp_void  memcpy_qqq(seqp_void   dest,
                      seqp_void  src,
                      unsigned int size ) {
  CHECK_BOUNDS(dest._b, dest._e, dest._p, size);
  CHECK_BOUNDS(src._b, src._e, src._p, size);
  memcpy(dest._p, src._p, size);
  return dest;
}

fseqp_void  memcpy_fqq(seqp_void   dest,
                      seqp_void  src,
                      unsigned int size ) {
  fseqp_void res;
  CHECK_BOUNDS(dest._b, dest._e, dest._p, size);
  CHECK_BOUNDS(src._b, src._e, src._p, size);
  memcpy(dest._p, src._p, size);
  res._p = dest._p;
  res._e = dest._e;
  return res;
}
 
fatp_void  memmove_www(fatp_void    dest,
                       fatp_void   src,
                       unsigned int size) {
  void *destb = dest._b;
  void *destp = dest._p;
  unsigned int destlen = CHECK_FETCHLENGTH(destp, destb);
  void *srcb = src._b;
  void *srcp = (void*)src._p;
  unsigned int srclen = CHECK_FETCHLENGTH(srcp, srcb);

  CHECK_BOUNDS_LEN(destb, destlen, destp, size);
  CHECK_BOUNDS_LEN(srcb, srclen, srcp, size);
  // Memmove copies from the start or from the end depending on overlap
#ifndef NO_TAGS
  if(destp < srcp || (char*)srcp + size < (char*)destp) {
    CHECK_COPYTAGSFORW(srcb, srclen, srcp,
                       destb, destlen, destp, size);
  } else {
    CHECK_COPYTAGSBACK(srcb, srclen, srcp,
                       destb, destlen, destp, size);
  }
#endif
  memmove(destp, srcp, size);
  return dest;
}


seqp_void  memmove_qqq(seqp_void    dest,
                       seqp_void   src,
                       unsigned int size) {
  CHECK_BOUNDS(dest._b, dest._e, dest._p, size);
  CHECK_NULL(dest._b);
  CHECK_BOUNDS(src._b, src._e, src._p, size);
  CHECK_NULL(src._b);
  // Ought to check that the alignment is correct ! But it should be !
  memmove(dest._p, src._p, size);
  return dest;
}

fseqp_void  memmove_fqq(seqp_void    dest,
                       seqp_void   src,
                       unsigned int size) {
  fseqp_void res;
  CHECK_BOUNDS(dest._b, dest._e, dest._p, size);
  CHECK_NULL(dest._b);
  CHECK_BOUNDS(src._b, src._e, src._p, size);
  CHECK_NULL(src._b);
  // Ought to check that the alignment is correct ! But it should be !
  memmove(dest._p, src._p, size);
  res._p = dest._p;
  res._e = dest._e;
  return res;
}

fseqp_void memmove_fff(fseqp_void dest, fseqp_void src, unsigned int size) {
  CHECK_NULL(dest._p); CHECK_UBOUND(dest._e, dest._p, size);
  CHECK_NULL(src._p); CHECK_UBOUND(src._e, src._p, size);
  memmove(dest._p, src._p, size);
  return dest;
}
            

fatp_void memset_ww(fatp_char buffer, int c, size_t size) {
  // Fetch the length in words
  unsigned int blenwrds = CHECK_FETCHLENGTH(buffer._p, buffer._b);
  fatp_void res;

  // sm: I think the -1 here is wrong (now that we handle strings in
  // a different way), but I won't change this until it becomes necessary
  int readlen = (int)(blenwrds << 2) - (buffer._p - (char*)buffer._b) - 1;
  if (c == 0) {
    // in this case, let them write a little more, since it's safe
    // to write zero in the last byte of any kind of array
    readlen++;
  }

  if(size < 0) {
    FP_FAIL(FAIL_LBOUND);
  }
  if(readlen < size) {
    FP_FAIL(FAIL_UBOUND);
  }
  res._p = memset(buffer._p, c, size);
  res._b = buffer._b;
  // Clean the tags
  CHECK_ZEROTAGS(buffer._b, blenwrds, buffer._p, size, 0);
  return res;
}

seqp_char memset_qq(seqp_char buffer, int c, size_t size) {
  CHECK_BOUNDS(buffer._b, buffer._e, (void*)buffer._p, size);
  memset(buffer._p, c, size);
  return buffer;
}

fseqp_char memset_fq(seqp_char buffer, int c, size_t size) {
  fseqp_char res;
  CHECK_BOUNDS(buffer._b, buffer._e, (void*)buffer._p, size);
  memset(buffer._p, c, size);
  res._p = buffer._p;
  res._e = buffer._e;
  return res;
}

char* memset_sf(fseqp_char buffer, int c, size_t size) {
  CHECK_UBOUND(buffer._e, buffer._p, size);
  memset(buffer._p, c, size);
  return buffer._p;
}

fseqp_char memset_ff(fseqp_char buffer, int c, size_t size) {
  CHECK_UBOUND(buffer._e, buffer._p, size);
  memset(buffer._p, c, size);
  return buffer;
}

// sm: given a fat pointer, return the # of bytes legal to read/write
// starting from the _p field
__inline static int fatp_length(fatp_char ptr)
{
  unsigned nwords = CHECK_FETCHLENGTH(ptr._p, ptr._b);

  // the code that was in memset_ww had a -1, but I think that's
  // wrong now that strings are handled orthogonally
  return (int)(nwords << 2) - (ptr._p - (char*)ptr._b);
}

// sm: fail unless there are >=n bytes read/writable after ptr;
// return # of bytes
__inline static void fatp_verify_atleast(fatp_char ptr, int n)
{
  if (n < 0) {
    FP_FAIL(FAIL_LBOUND);
  }
  if (n > fatp_length(ptr)) {
    FP_FAIL(FAIL_UBOUND);
  }
}

// sm: bzero for anagram
void bzero_w(fatp_char buffer, int n)
{
  unsigned nwords = CHECK_FETCHLENGTH(buffer._p, buffer._b);
  fatp_verify_atleast(buffer, n);

  // zero the data area itself
  memset(buffer._p, 0, n);

  // zero tags (I don't understand 'offset' argument, but it is
  // passed as 0 in memset_ww, so I'll do the same)
  CHECK_ZEROTAGS(buffer._b, nwords, buffer._p, n, 0);
}


#ifndef _MSVC
void bzero_f(fseqp_char buffer, size_t size) {
  CHECK_UBOUND(buffer._e, buffer._p, size);
  bzero(buffer._p, size);
}
#endif


#include <time.h>
typedef struct fatp_time_t1 { 
  time_t * _p ;
  void * _b ;
} fatp_time_t1;

time_t  __cdecl  time_w(fatp_time_t1  timer) {
  if(timer._p) { // time accepts a NULL argument
      unsigned int l = CHECK_FETCHLENGTH(timer._p, timer._b);
      CHECK_BOUNDS_LEN(timer._b, l, timer._p, sizeof(time_t));
  }
  return time(timer._p);
}

#ifdef _GNUCC
#include <sys/times.h>
clock_t times_w(fatp_void tms) {
  return times((struct tms*)tms._p);
}
#endif

#define LOG_CASTS
#define MIN_POINTER 0x100000U
#define MAX_POINTER (0xFFFFFFFFU - MIN_POINTER)
#ifdef _MSVC
#ifdef _DEBUG
fatp_void _CrtSetReportFile_ww(int reportMode, fatp_void f) {
  _CrtSetReportFile(reportMode, f._p);
  return f;
}
#endif /* _DEBUG */
#endif /* _MSVC */

#ifdef WALK_HEAP
#include <malloc.h>
_HEAPINFO hpWalk;
#endif /* WALK_HEAP */

void * malloc_safe(unsigned int size);
/*
  fatp_void calloc_fseq_f(unsigned int count, 
                        unsigned int osize);
*/                        
fatp_void realloc_ww(fatp_void orig, unsigned int newsz);
/* fatp_void malloc_fseq_f(unsigned size); */
void free_safe(void *x);
void free_w(fatp_void x);
// void free_f(fatp_void1 x);
void free_f();

fatp_char fattenstring(char *s) {
  fatp_void tmp = calloc_w(1, 1 + strlen(s));
  fatp_char res = *(fatp_char*)&tmp;
  if(! res._p) {
    return res;
  }
  strcpy(res._p, s);
  return res;
}

fatp_char strdup_ws(char *x) {
  return fattenstring(x);
}

#ifdef LOG_CASTS
typedef struct {
  unsigned long l;
  unsigned short fileid, lineid;
} LOG_ENTRY;
#define LOG_SIZE   (1 << 16)
LOG_ENTRY logBuffer[LOG_SIZE];
int logBufferP = 0; // Next free entry
int logFid = -1;
#include <stdlib.h>
#include <fcntl.h>

// Flush the log to a file
int flushLog(void) {
      // See if we need to open the log file
  int toWrite = logBufferP * sizeof(LOG_ENTRY);
  if(toWrite <= 0)
    return 0;
  
  if(logFid < 0) {
    logFid = open("__scalar2pointer",
                  O_CREAT  | O_TRUNC | O_RDWR | O_BINARY,
                  0777);
    if(logFid < 0) {
      FP_FAIL(FAIL_SCALARLOG);
    }
  }
  if(toWrite != write(logFid, & logBuffer[0], toWrite)) {
    FP_FAIL(FAIL_SCALARLOG);
  }
  logBufferP = 0;
  return 0;
}

/* Scan the log and apply the given function to all entries. Stop when the 
 * log is over or when the match function returns true. Scan log returns 
 * true if it terminated because match returned true */
int scanLog(int (*match)(int, LOG_ENTRY *), int closure) {
  flushLog();
  if(logFid < 0) {
    // Log is empty
    return 0;
  }
  {
    LOG_ENTRY *le = & logBuffer[0];
    int toGo = lseek(logFid, 0, SEEK_END);
    int i = LOG_SIZE;  // Pretend we are at the end of the buffer

    lseek(logFid, 0, SEEK_SET); // Go to start
    while(toGo > 0) {
      if(i >= LOG_SIZE) {
        int toRead = sizeof(logBuffer);
        if(toRead > toGo) toRead = toGo;
        if(toRead != read(logFid, & logBuffer[0], toRead)) {
          FP_FAIL(FAIL_SCALARLOG);
        }
        i = 0;
        le = & logBuffer[i];
      }
      if((*match)(closure, le)) {
        return 1;
      }
      i ++;
      le ++;
      toGo -= sizeof(LOG_ENTRY);
    }
  }
  return 0; // Not found
}

int foundOne = 0;
int printMatches(int tomatch, LOG_ENTRY *le) {
  if(le->l == (unsigned)tomatch) {
    foundOne = 1;
    fprintf(stderr, "\tSame as at cast in file 0x%04x on line %d\n",
            le->fileid, le->lineid);
  }
  return 0; // Want all matches
}

// Print the contents of the log
int printAll(int where, LOG_ENTRY *le) {
  fprintf((FILE*)where, "0x%08lx  0x%04x %d\n",
          le->l, (unsigned)(le->fileid), le->lineid);
  return 0;
}
#endif /* LOG_CASTS */


void* __scalar2pointer(unsigned long l, int fileid, int lineid) {
#ifdef WALK_HEAP  
  void *base = NULL;
#endif  
#ifdef LOG_CASTS
  if(l >= MIN_POINTER && l <= MAX_POINTER) {
    LOG_ENTRY *le = & logBuffer[logBufferP ++];
    le->l = l;
    le->fileid = fileid;
    le->lineid = lineid;
    // See if we need to dump the log out
    if(logBufferP == LOG_SIZE) {
      flushLog();
    }
  }
#endif /* LOG_CASTS */
#ifdef WALK_HEAP
  hpWalk._pentry = NULL;
  if(l >= MIN_POINTER && l <= MAX_POINTER) {
    while(_heapwalk(&hpWalk)) {
      if(l >= (unsigned long)hpWalk._pentry &&
         l < (unsigned long)hpWalk._pentry + hpWalk._size &&
         hpWalk._useflag == _USEDENTRY) {
        base = (void*)((char*)hpWalk._pentry + 4);
        goto FoundOne;
      }
    }
  }
  return NULL;
 FoundOne:
  {
    unsigned int len = CHECK_FETCHLENGTH(base);
    TAGADDR tag   = CHECK_FETCHTAGADDR(base, len);
    int  nrTagWords = CHECK_NRTAGWORDS(len);
    if(l >= base &&
       l <= (long)(tag.word - 4) &&
       (1 + len + nrTagWords) << 2 == hpWalk._size) {
      return (void*)base;
    }
  }
#endif /* WALK_HEAP */
  return NULL;
}

  
// And a function to scan the logs and find the matching
void searchLog(unsigned long l  FP_FAIL_EXTRA_PARAMS) {
#ifdef LOG_CASTS
  if(l >= MIN_POINTER && l <= MAX_POINTER) {
    fprintf(stderr, "Dereferencing a non-pointer (0x%08lx)\n", l);
    foundOne = 0;
    scanLog(& printMatches, (int)l);
    if(! foundOne) {
      // Dump the entire log to a file
      FILE *txtLog = fopen("__scalar2pointer.txt", "w");
#ifdef _MSVC
      char fullPath[_MAX_PATH];
#endif
      char *theFileName = "__scalar2pointer.txt";
#ifdef _MSVC
      theFileName = _fullpath(fullPath, theFileName, _MAX_PATH);
#endif
      fprintf(stderr,
              "Cannot find the cast in log. Dumping the entire log to %s\n",
               theFileName);
      scanLog(& printAll, (int)txtLog);
      fclose(txtLog);
    } else {
      fflush(stderr);
    }
  }
  fp_fail(FAIL_NONPOINTER  FP_FAIL_EXTRA_ARGS);

#else
  fp_fail(FAIL_NONPOINTER  FP_FAIL_EXTRA_ARGS);
#endif
}


// sm: the next 3 functions are not conditionally compiled, because
// they don't hurt in non-gc mode, and it simplifies linking issues
// when switching from gc to non-gc, or back

// the GC has been instrumented to call this function every time
// a block is noticed to be reachable; the idea is I will add code
// to detect if the programmer freed this block, in which case I
// might take some additional action (like warn the user?)
void noticedReachable(void *block, size_t size)
{
  //printf("noticedReachable: %p, sz=%d\n", block, size);
}

void noticedUnreachable(void *block, size_t size)
{
  //printf("noticedUnreachable: %p, sz=%d\n", block, size);
}

void noticedContiguousUnreachable(void *block, size_t size, size_t arraySize)
{
  //printf("noticedContiguousUnreachable: %p, sz=%d, arraySize=%d (ratio: %d)\n",
  //       block, size, arraySize, arraySize/size);
}


/* *************** WRAPPERS for MAIN *************** */
// an initializer for all our c-cured stuff
void safecInit();

// A wrapper main function
extern int trueMain();     // sm: no decl because of diff't usages below?
int _mainWrapper(int argc, char **argv) {
  safecInit();
  return trueMain(argc, argv);
}

int _mainWrapper_w(int argc, char **argv) {
  // Just copy the input arguments. Leave room for the last NULL
  fatp_void tmp = malloc_w(sizeof(fatp_char) * (argc + 1));
  fatp_fatp_char nargv = *(fatp_fatp_char*)&tmp;
  int i;
  unsigned int argvlen = CHECK_FETCHLENGTH(nargv._p, nargv._b);

  for(i=0;i<argc;i++) {
    fatp_char copy = fattenstring(argv[i]);
    // Use CHECK_FATPOINTERWRITE to write the proper tags also
    CHECK_FATPOINTERWRITE(nargv._b, argvlen,
                          (void*) & nargv._p[i],
                          copy._b, copy._p);
    nargv._p[i] = copy;
  }
  nargv._p[argc]._p = (void*)0;
  nargv._p[argc]._b = (void*)0;
  
  safecInit();
  return trueMain(argc, *(fatp_fatp_char*)&nargv);
}

typedef struct {
  char * * _p;
  void * _e;
} fseqp_p_char;
int _mainWrapper_fs(int argc, char **argv) {
  int i;
  fseqp_p_char argv1;
  char **nargv = (char**)malloc((argc+1) * sizeof(char *));

  argv1._p = nargv;
  argv1._e = (void*)(nargv + argc + 1);

  for(i=0;i<argc;i++) {
    nargv[i] = strdup(argv[i]);
  }
  nargv[argc] = NULL;
  safecInit();
  return trueMain(argc, argv1);
}

int _mainWrapper_qw(int argc, char **argv) {
  int i;
  seq_fatp_char argv1;
  fatp_char *nargv = (fatp_char*)malloc((argc + 1) * sizeof(fatp_char));

  argv1._p = nargv;
  argv1._b = (void*)nargv;
  argv1._e = (void*)(nargv + argc + 1);

  for(i=0;i<argc;i++) {
    fatp_char copy = fattenstring(argv[i]);
    nargv[i] = copy;
  }
  nargv[argc]._p = (void*)0;
  nargv[argc]._b = (void*)0;

  safecInit();
  return trueMain(argc, argv1);
}

int _mainWrapper_fw(int argc, char **argv) {
  int i;
  fseqp_fatp_char argv1;
  fatp_char *nargv = (fatp_char*)malloc((1 + argc) * sizeof(fatp_char));

  argv1._p = nargv;
  argv1._e = (void*)(nargv + argc + 1);

  for(i=0;i<argc;i++) {
    fatp_char copy = fattenstring(argv[i]);
    nargv[i] = copy;
  }
  nargv[argc]._p = (void*)0;
  nargv[argc]._b = (void*)0;
  
  safecInit();
  return trueMain(argc, argv1);
}

typedef struct {
  seqp_char *_p;
  void *_e;
} fseqp_seqp_char;

int _mainWrapper_fq(int argc, char **argv) {
  int i;
  fseqp_seqp_char argv1;
  seqp_char *nargv = (seqp_char*)malloc((1 + argc) * sizeof(seqp_char));

  argv1._p = nargv;
  argv1._e = (void*)(nargv + argc + 1);

  for(i=0;i<argc;i++) {
    char *copy = strdup(argv[i]);
    nargv[i]._p = copy;
    nargv[i]._b = copy;
    nargv[i]._e = copy + strlen(copy);
  }
  nargv[argc]._p = (void*)0;
  nargv[argc]._b = (void*)0;
  nargv[argc]._e = (void*)0;
  

  safecInit();
  return trueMain(argc, argv1);
}

typedef struct {
  fseqp_char *_p;
  void *_e;
} fseqp_fseqp_char;

int _mainWrapper_ff(int argc, char **argv) {
  int i;
  fseqp_fseqp_char argv1;
  fseqp_char *nargv = (fseqp_char*)malloc((argc + 1) * sizeof(fseqp_char));

  argv1._p = nargv;
  argv1._e = (void*)(nargv + argc + 1);

  for(i=0;i<argc;i++) {
    char *copy = strdup(argv[i]);
    nargv[i]._p = copy;
    nargv[i]._e = copy + strlen(copy);
  }
  nargv[argc]._p = (void*)0;
  nargv[argc]._e = (void*)0;

  safecInit();
  return trueMain(argc, argv1);
}


#ifdef __CYGWIN__
int _get_ctype(int chr) {
  return (_ctype_+1)[chr];
}
#endif
        

// sm: I don't have __assert ... we need an #ifdef here
// which identifies the machine which has this ...
// gn: who has this anyway?
#if !defined(__GLIBC__) && !defined(_MSVC)
extern void __assert(char *, int, char *);
void __assert_ww(fatp_char file, int line, fatp_char a2) {
  CHECK_STRINGMAX(file._p, file._b);
  CHECK_STRINGMAX(a2._p, a2._b);
  __assert(file._p, line, a2._b);
}
#endif


// c-cured initializer
void safecInit()
{
  if (sizeof(failMessages)/sizeof(failMessages[0]) != NUM_FAIL_CODES) {
    FP_FAIL_STR("NUM_FAIL_CODES doesn't agree with failMessages[]");
  }
}


fatp_char strpbrk_wwf(fatp_char str, fseqp_char accept)
{
  fatp_char res;
  CHECK_STRINGMAX(str._p, str._b);
  check_nullterm_fseq(accept);
  res._b = str._b;
  res._p = strpbrk(str._p, accept._p);
  return res;
}


// sm: changed seq_fatp_char to seqp_char .. I don't know how it worked before!
fatp_char strpbrk_wwq(fatp_char str, seqp_char accept)
{
  fatp_char res;
  CHECK_STRINGMAX(str._p, str._b);
  CHECK_BOUNDS(accept._b, accept._e, accept._p, 1);
  res._b = str._b;
  res._p = strpbrk(str._p, accept._p);
  return res;
}


// sm:
fseqp_char strpbrk_fff(fseqp_char str, fseqp_char accept)
{
  fseqp_char ret;

  check_nullterm_fseq(str);
  check_nullterm_fseq(accept);

  ret._p = strpbrk(str._p, accept._p);
  if (ret._p) {
    ret._e = str._e;     // they point to same block
  }
  else {
    ret._e = 0;          // null return
  }

  return ret;
}

// sm: couldn't get it to always infer fseq return ..
char *strpbrk_sff(fseqp_char str, fseqp_char accept)
{
  return strpbrk_fff(str, accept)._p;
}


fseqp_char strtok_fff(fseqp_char str, fseqp_char delim)
{ 
  fseqp_char ret;
  static char *prev_end = NULL;     // we must be stateful just like strtok

  if (str._p) {
    check_nullterm_fseq(str);
    prev_end = str._e;              // save for when str._p is NULL
  }
  check_nullterm_fseq(delim);

  ret._p = strtok(str._p, delim._p);
  if (ret._p) {
    ret._e = prev_end;          // remember what it was in case str._p is NULL
  }
  else {
    ret._e = NULL;
  }
  
  return ret;
}

// sm: copied from above for seq ptrs
seqp_char strtok_qqq(seqp_char str, seqp_char delim)
{
  seqp_char ret;
  static char *prev_base = NULL;    // we must be stateful just like strtok
  static char *prev_end = NULL;

  if (str._p) {
    check_nullterm_seq(str);
    prev_base = str._b;
    prev_end = str._e;              // save for when str._p is NULL
  }
  check_nullterm_seq(delim);

  ret._p = strtok(str._p, delim._p);
  if (ret._p) {
    ret._b = prev_base;
    ret._e = prev_end;              // remember what it was in case str._p is NULL
  }
  else {
    ret._b = NULL;
    ret._e = NULL;
  }

  return ret;
}


// A function whose result is either not used or used just for dereference
// This should be automatically constructed
char* strtok_sff(fseqp_char str, fseqp_char delim) {
  fseqp_char res = strtok_fff(str, delim);
  CHECK_UBOUNDNULL(res._e, res._p, sizeof(char));
  return res._p;
}

char *strtok_sqq(seqp_char str, seqp_char delim) {
  fseqp_char str_f, delim_f, res_f;
  CHECK_SEQTOFSEQ(str._b, str._p);
  str_f._p = str._p; str_f._e = str._e;
  CHECK_SEQTOFSEQ(delim._b, delim._p);
  delim_f._p = delim._p; delim_f._e = delim._e;
  res_f = strtok_fff(str_f, delim_f);
  return res_f._p;
}

fseqp_char fgets_ffw(fseqp_char buffer, int len, fatp_FILE fp)
{
  // give the caller 1 byte of slack, since we typically set the
  // last byte of character arrays to 0 and then take away that
  // byte.. this isn't sound, but it should work most of the time
  // (change 'len-1' to 'len' to get the sound version)
  if (!buffer._e || ((int)buffer._e - (int)buffer._p < (len-1))) {
    FP_FAIL_STR("fgets called with too-small buffer");
  }

  if (!fgets(buffer._p, len, fp._p)) {
    fseqp_char res;
    res._p = 0;
    res._e = 0;
    return res;
  }
  else {
    return buffer;
  }
}

// another version for seq
seqp_char fgets_qqw(seqp_char buffer, int len, fatp_FILE fp)
{
  // 1 byte of slack, like above
  if (!buffer._b || ((int)buffer._e - (int)buffer._p < (len-1))) {
    FP_FAIL_STR("fgets called with too-small buffer");
  }

  if (!fgets(buffer._p, len, fp._p)) {
    seqp_char res;
    res._b = 0;
    res._p = 0;
    res._e = 0;
    return res;
  }
  else {
    return buffer;
  }
}


int fseek_w(fatp_FILE fp, long offset, int whence)
{
  return fseek(fp._p, offset, whence);
}


// sm: our current implementation actually can't hit this, because we don't
// treat realloc polymorphically; this was from experimenting with James'
// options to do inference unsoundly
fseqp_void realloc_ff(fseqp_void ptr, size_t newsize)
{
  // I can't be sure 'ptr' really is the start of a memory block, so I'll
  // play it safe; GC should clean up after me
  fseqp_void ret;

  size_t oldsize = (int)ptr._e - (int)ptr._p;
  if (oldsize >= newsize) {
    // just give them back the same pointer, and don't release any memory
    ret._p = ptr._p;
    ret._e = (void*)( (int)(ptr._p) + newsize );
    return ret;
  }

  // allocate some new space
  ret._p = malloc(newsize);
  if (!ret._p) {
    FP_FAIL(FAIL_MALLOC);
  }
  ret._e = (void*)( (int)(ret._p) + newsize );

  // copy the original data into this new buffer
  memcpy(ret._p, ptr._p, oldsize);

  // zero out the remainder
  memset((void*)( (int)(ret._p) + oldsize ), 0, newsize-oldsize);
  
  // finally give back this pointer
  return ret;
}


void clearerr_w(fatp_FILE fp)
{
  clearerr(fp._p);
}



// sm: I don't know if this is a problem on msvc, but I suspect it is               
#ifdef _GNUCC
  // sm: for m88k
  typedef void (*sig_fn)(int);
  typedef struct fseq_sig_fn {
    sig_fn fn;
    void *_e;
  } fseq_sig_fn;
  fseq_sig_fn signal_ff(int signum, fseq_sig_fn f)
  {
    // unsound
    fseq_sig_fn ret;
    ret.fn = signal(signum, f.fn);
    ret._e = 0;     // people don't call handlers directly (right?)
    return ret;
  }

  // workaround for m88k getopt thing (bug in our tool)
  int m88k_getopt(int argc, seqp_char *argv, char *argdesc)
  {
    char **argv2;
    int i;
    int retval;

    // construct a new array
    argv2 = (char**)malloc((argc+1) * sizeof(char*));

    // copy the string pointers into it; no bounds checking
    for (i=0; i<argc+1; i++) {
      argv2[i] = argv[i]._p;    // last will be NULL (we hope)
    }

    // call the real getopt
    retval = getopt(argc, argv2, argdesc);

    // free the array
    free(argv2);

    return retval;
  }
#endif // _GNUCC

FILE * get_stdin() {
    return stdin;
}
FILE * get_stdout() {
    return stdout;
}
FILE * get_stderr() {
    return stderr;
} 

#ifdef __CYGWIN__
int _get_ctype_p1(unsigned int chr) {
  return (_ctype_ + 1)[chr]; 
}
#endif

#if defined(__GLIBC__) && __GLIBC__ == 2
/// For brooksie
// sm: glibc 2.0 and 2.1 both have this; so I removed check for minor number
int _get__ctype_b(int chr) { 
  if (!( (unsigned)chr < 256 )) {
    fprintf(stderr, "Bad character index to _get__ctype_b: %d\n", chr);
    abort();
  }
  return (__ctype_b[chr]);
}
#endif


// VARARGS
#include <stdarg.h>
// Define our version of va_start
#define OUR_ROUND(size) ((size + sizeof(int) - 1) & ~(sizeof(int) - 1))
#ifdef _GNUCC
/* On GCC we do not take the address of the last local because if we do that 
 * gcc will give us the address of a copy of the local. */
#define getPNextArg(lastarg) __builtin_next_arg(lastarg)
#else
#define getPNextArg(lastarg) ((char*)(&(lastarg)) + OUR_ROUND(sizeof(lastarg)))
#endif
#define our_advance_va_arg(ap,size) ap = (va_list)((char*)ap + OUR_ROUND(size))
             
int __ccured_va_count = 0;
int __ccured_va_tags[32];
typedef struct __ccured_va_localinfo {
   int next ;
   int count ;
   unsigned int tags[32] ;
   va_list nextp ;
} CCURED_VAINFO;
 
void __ccured_va_start_vs(CCURED_VAINFO *vainfo,
                          void *pNextArg) {
  vainfo->next  = 0;
  vainfo->count = __ccured_va_count;
  memcpy(vainfo->tags, __ccured_va_tags, sizeof(__ccured_va_tags));

  vainfo->nextp = (va_list)pNextArg;
  return;
}


void __ccured_va_end_vs(CCURED_VAINFO *vainfo) {
  vainfo->next = -1;
}

#define GET_VA_TAG(tags, idx) ((tags[idx >> 2] >> ((idx & 0x3) << 3)) & 0xFF)

void *__ccured_va_arg_svs(CCURED_VAINFO *vainfo,
                          unsigned int thisTypeSize,
                          int thisTypeIndex) {
  void* lastp = vainfo->nextp;
  
  if(vainfo->next == -1) {
    FP_FAIL(FAIL_VA_NOSTART);
  }
  if(vainfo->next >= vainfo->count) {
    FP_FAIL(FAIL_VA_MANY);
  }
  // Grab the stored tag
  {
    int tag = GET_VA_TAG(vainfo->tags, vainfo->next);
    if(tag != thisTypeIndex) {
      FP_FAIL(FAIL_VA_BADTYPE);
    }
  }
  our_advance_va_arg(vainfo->nextp, thisTypeSize);

  vainfo->next ++;
  return lastp;
}

void CHECK_FORMATARGS(const char*format) {
  int i = 0, check;
  if(! format) {
    FP_FAIL(FAIL_NONPOINTER);
  }
  if(__ccured_va_count == -1) return; // Was checked already
  while(*format) {
    if(*format == '%') {
      format ++;
    KeepLooking:
      switch(*format) {
      case '%': format ++; continue;
      case 's': check = 2; goto DoCheck;
      case 'f': check = 1; goto DoCheck;
      case 'l': case '-': case '.': format ++; goto KeepLooking;
      case 'd': case 'c': case 'x': check = 0; goto DoCheck;
      default:
        if(*format >= '0' && *format <= '9') {
          format ++; goto KeepLooking;
        }
        FP_FAIL(FAIL_UNRECOGNIZEDFORMAT);
      }
    DoCheck:
      if(i >= __ccured_va_count) {
        FP_FAIL(FAIL_VA_MANY);
      }
      if(check != GET_VA_TAG(__ccured_va_tags, i)) {
        FP_FAIL(FAIL_VA_BADTYPE);
      }
      i ++;
    }
    format ++;
  }
  __ccured_va_count = -1; // Was checked already. this will prevent the use
  // of va_arg but that would be unsafe anyway for printf-like functions
}

// Some wrappers for the format functions
#ifdef _MSVC
#define VSNPRINTF _vsnprintf
#else
#define VSNPRINTF vsnprintf
#endif
int vsprintf_fsvs(fseqp_char buffer, const char *format,
                  CCURED_VAINFO *args) {

  int res;

  CHECK_FORMATARGS(format);
  
  if(!buffer._e || buffer._p >= buffer._e) {
    FP_FAIL(FAIL_UBOUND);
  }
  
  res = VSNPRINTF(buffer._p, (int)buffer._e - (int)buffer._p - 1,
                  format, args->nextp);
  if(res < 0) {
    FP_FAIL(FAIL_UBOUND);
  }
  return res;
}

int vsprintf_qsvs(seqp_char buffer, const char *format,
                  CCURED_VAINFO *args) {
  fseqp_char buffer_f;
  CHECK_SEQTOFSEQ(buffer._b, buffer._p);
  buffer_f._p = buffer._p; buffer_f._e = buffer._e;
  return vsprintf_fsvs(buffer_f, format, args);
}

int sprintf_qss(seqp_char buffer, const char* format, ...) {
  int res;
  CCURED_VAINFO args;
  fseqp_char fbuffer = { buffer._p, buffer._e };
  
  __ccured_va_start_vs(&args, getPNextArg(format));

  /* Make sure there is at least 1 character available */
  CHECK_BOUNDS(buffer._b, buffer._e, (void*)buffer._p, 1);
  res = vsprintf_fsvs(fbuffer, format, &args);
  __ccured_va_end_vs(&args);
  return res;
}

int sprintf_wss(fatp_char buffer, char *format, ...) {
  CCURED_VAINFO args;
  int res;
  int wrds = CHECK_FETCHLENGTH(buffer._p, buffer._b);
  fseqp_char fbuffer = { buffer._p, buffer._p + (wrds << 2) };
  
  __ccured_va_start_vs(&args, getPNextArg(format));
  res = vsprintf_fsvs(fbuffer, format, &args);
  __ccured_va_end_vs(&args);

  return res;
}

int sprintf_wws(fatp_char buffer, fatp_char format, ...) {
  CCURED_VAINFO args;
  int res;
  int wrds = CHECK_FETCHLENGTH(buffer._p, buffer._b);
  fseqp_char fbuffer = { buffer._p, buffer._p + (wrds << 2) };
  
  __ccured_va_start_vs(&args, getPNextArg(format));
  { // Check the format string's length
    unsigned int max = CHECK_STRINGMAX((void*)format._p, format._b);
    CHECK_STRINGLEN(format._p, max);
  }
  res = vsprintf_fsvs(fbuffer, format._p, &args);
  __ccured_va_end_vs(&args);
  return res;
}

// FPRINTF

int vfprintf_ssvs(FILE * stream, const char * format, CCURED_VAINFO *args) {
  int res;
  CHECK_NULL(stream);
  CHECK_NULL((char*)format);

  CHECK_FORMATARGS(format);
  res = vfprintf(stream, format, args->nextp);
  return res;
}


int vfprintf_wwvs(fatp_void stream, fatp_char format, CCURED_VAINFO *args) {
  { // Check the string
    unsigned int max = CHECK_STRINGMAX((void*)format._p, format._b);
    CHECK_STRINGLEN(format._p, max);
  }
  return vfprintf_ssvs(stream._p, format._p, args);
}

int fprintf_wws(fatp_void stream, fatp_char format, ...) {
  CCURED_VAINFO args;
  int res;

  __ccured_va_start_vs(&args, getPNextArg(format));

  { // Check the string
    unsigned int max = CHECK_STRINGMAX((void*)format._p, format._b);
    CHECK_STRINGLEN(format._p, max);
  }
  res = vfprintf_ssvs(stream._p, format._p, &args);
  __ccured_va_end_vs(&args);
  return res;
}

int fprintf_sws(FILE *stream, fatp_char format, ...) {
  CCURED_VAINFO args;
  int res;

  __ccured_va_start_vs(&args, getPNextArg(format));

  { // Check the string
    unsigned int max = CHECK_STRINGMAX((void*)format._p, format._b);
    CHECK_STRINGLEN(format._p, max);
  }
  res = vfprintf_ssvs(stream, format._p, &args);
  __ccured_va_end_vs(&args);
  return res;
}

int fprintf_wss(fatp_void stream, char *format, ...) {
  CCURED_VAINFO args;
  int res;

  __ccured_va_start_vs(&args, getPNextArg(format));
  res = vfprintf_ssvs(stream._p, format, &args);
  __ccured_va_end_vs(&args);
  return res;
}


// PRINTF

int vprintf_svs(char const *format, CCURED_VAINFO *args)
{
  return vfprintf_ssvs(stdout, format, args);
}


int printf_ws(fatp_char format, ...) {
  CCURED_VAINFO args;
  int res;

  __ccured_va_start_vs(&args, getPNextArg(format));

  { // Check the string
    unsigned int max = CHECK_STRINGMAX((void*)format._p, format._b);
    CHECK_STRINGLEN(format._p, max);
  }
  
  res = vfprintf_ssvs(stdout, format._p, &args);
  __ccured_va_end_vs(&args);
  return res;
}


// sm: trying vsyslog too! (conditional on linux to not interfere on cygwin)
#ifdef __linux__
  extern void vsyslog_svs(int pri, char const *format, CCURED_VAINFO *args)
  {
    CHECK_FORMATARGS(format);              // check
    vsyslog(pri, format, args->nextp);     // pass
  }
#endif // __linux__



// Replacements for scanf and friends

int ccured_scanf_count;
FILE *ccured_sscanf_file = NULL;
void resetScanfCount(void) {
  ccured_scanf_count = 0;
}
// Call this function for sscanf
void resetSScanfCount(char *string) {
  if(! string) {
    FP_FAIL(FAIL_NULL_STRING);
  }
  if(ccured_sscanf_file) {
    fprintf(stderr, "Must call getScanfCount after resetSScanfCount");
    exit(2);
  }
  ccured_sscanf_file = fopen("sscanf.out", "w+b");
  if(! ccured_sscanf_file) {
    fprintf(stderr, "Cannot open sscanf.out\n");
    exit(1);
  }
  fputs(string, ccured_sscanf_file);
  fseek(ccured_sscanf_file, 0L, SEEK_SET);
}

void resetSScanfCount_w(fatp_char string) {
  unsigned int max = CHECK_STRINGMAX((void*)string._p, string._b);
  CHECK_STRINGLEN(string._p, max);

  resetSScanfCount(string._p);
}

int getScanfCount() {
  if(ccured_sscanf_file) {
    fclose(ccured_sscanf_file);
  }
  return ccured_scanf_count;
}
  
double ccured_fscanf_double(FILE *stream, char *format) {
  double d = 0.0;
  if(ccured_scanf_count != EOF) {
    int res = fscanf(stream, format, &d);
    if(res == EOF) {
      ccured_scanf_count = EOF;
    } else 
      ccured_scanf_count += res;
  }
  return d;
}
double ccured_fscanf_double_ww(fatp_FILE stream, fatp_char format) {
  unsigned int max = CHECK_STRINGMAX((void*)format._p, format._b);
  CHECK_STRINGLEN(format._p, max);
  return ccured_fscanf_double(stream._p, format._p);
}

int ccured_fscanf_int(FILE *stream, char *format) {
  int i = 0;
  if(ccured_scanf_count != EOF) {
    int res = fscanf(stream, format, &i);
    if(res == EOF) {
      ccured_scanf_count = EOF;
    } else 
      ccured_scanf_count += res;
  }
  return i;
}
int ccured_fscanf_int_ww(fatp_FILE stream, fatp_char format) {
  unsigned int max = CHECK_STRINGMAX((void*)format._p, format._b);
  CHECK_STRINGLEN(format._p, max);
  return ccured_fscanf_int(stream._p, format._p);
}


void ccured_fscanf_nothing(FILE *stream, char* what) {
  if(EOF == fscanf(stream, what))  // just advance the pointer and match
    ccured_scanf_count = EOF;
}
void ccured_fscanf_nothing_ws(fatp_FILE stream, char* what) {
  ccured_fscanf_nothing(stream._p, what);
}

void ccured_fscanf_string_ssf(FILE *stream, char* what, fseqp_char buffer) {
  int bufflen = (int)buffer._e - (int)buffer._p;
  char *mybuff;
  int orig_pos, read, scanned;

  if(!buffer._p || !buffer._e) {
    FP_FAIL(FAIL_NONPOINTER);
  }
  if(buffer._p >= buffer._e) {
    FP_FAIL(FAIL_UBOUND);
  }
  // Now create a buffer
  mybuff = (char*)malloc(bufflen);
  if(!mybuff) {
    fprintf(stderr, "Cannot allocate buffer for fscanf_string");
    exit(2);
  }
  // Now remember the positon in the stream
  orig_pos = ftell(stream);
  // Now read that many characters from the stream in the buffer
  read = fread(mybuff, 1, bufflen - 1, stream);
  if(read == bufflen - 1) {
    FILE *myfile = fopen("sscanf_string", "rwb");
    if(! myfile) {
      fprintf(stderr, "Cannot create file for fscanf_string");
      exit(2);
    }
    // Write them out to the new file
    fwrite(mybuff, read, 1, myfile);
    // Rewind
    fseek(myfile, 0L, SEEK_SET);
    // And scan
    scanned = fscanf(myfile, what, buffer._p);
    // Find out how much we have advanced. This is misleading if we
    // are openin text mode on Win and we skipped newlines !
    orig_pos += ftell(myfile);
    fseek(stream, orig_pos, SEEK_SET);
    fclose(myfile);
    if(scanned == EOF) {
      FP_FAIL(FAIL_UBOUND);
    }
  } else {
    // The file is shorter than the buffer. Go ahead and scan
    scanned = fscanf(stream, what, buffer._p);
  }
  if(scanned == EOF) {
    ccured_scanf_count = EOF;
  } else {
    ccured_scanf_count += scanned;
  }
}

void ccured_fscanf_string_ssw(FILE *stream, char* what, fatp_char buffer) {
  fseqp_char buffer_f;
  int buff_wrds = CHECK_FETCHLENGTH(buffer._p, buffer._b); 
  buffer_f._p = buffer._p;
  buffer_f._e = buffer._p + (buff_wrds << 2);
  // Now call the above
  ccured_fscanf_string_ssf(stream, what, buffer_f);
  // Now fix the tags !
  // For the entire buffer to be sure
  CHECK_ZEROTAGS(buffer._b, buff_wrds, buffer._p, buff_wrds << 2, 0);
}

// ---------------------- socket functions ----------------
#ifdef _GNUCC
  // complain if buffer isn't at least 'length' bytes long
  static void verify_length(fatp_void buffer, int length)
  {
    int bytesTotal = CHECK_FETCHLENGTH(buffer._p, buffer._b) * 4;
    int bytesToEnd = bytesTotal - ((int)buffer._p - (int)buffer._b);
    if (bytesToEnd < length) {
      FP_FAIL(FAIL_UBOUND);
    }
  }

  #if 0
  // be happy if buffer points to at least 1 int; return that int's value
  static int verify_seqp_int(seqp_int buffer)
  {
    if (!buffer._b || buffer._p+1 > buffer._e) {
      FP_FAIL(FAIL_UBOUND);
    }
    return *(buffer._p);
  }
  #endif // 0

  // the user simply has to pass a buffer in 'optval' that is at least
  // 'optlen' bytes long
  int setsockopt(int s, int level, int optname, void const *optval, int optlen);
  int setsockopt_w(int s, int level, int optname, fatp_void optval, int optlen)
  {
    verify_length(optval, optlen);
    return setsockopt(s, level, optname, optval._p, optlen);
  }

  // same deal for bind
  int bind(int sockfd, /*struct sockaddr*/ void *my_addr, int addrlen);
  int bind_w(int sockfd, fatp_void my_addr, int addrlen)
  {
    verify_length(my_addr, addrlen);
    return bind(sockfd, my_addr._p, addrlen);
  }

  // connect is just like bind
  int connect(int sockfd, /*struct sockaddr*/ void *my_addr, int addrlen);
  int connect_w(int sockfd, fatp_void my_addr, int addrlen)
  {
    verify_length(my_addr, addrlen);
    return connect(sockfd, my_addr._p, addrlen);
  }

  // similar for accept, but length is value-result
  int accept(int s, /*struct sockaddr*/ void *addr, int *addrlen);
  int accept_ws(int s, fatp_void addr, int *addrlen)
  {
    if (addr._p == 0) {
      // addr is ignored when null
    }
    else {
      // verify addr's size
      verify_length(addr, *addrlen);
    }
    return accept(s, addr._p, addrlen);
  }


  // same as with accept
  int getpeername(int s, /*struct sockaddr*/ void *name, int *namelen);
  int getpeername_ws(int s, fatp_void name, int *namelen)
  {
    verify_length(name, *namelen);
    return getpeername(s, name._p, namelen);
  }

  // identical to getpeername
  int getsockname(int s, /*struct sockaddr*/ void *name, int *namelen);
  int getsockname_ws(int s, fatp_void name, int *namelen)
  {
    verify_length(name, *namelen);
    return getsockname(s, name._p, namelen);
  }

  //gn: I don't have getnameinfo on cygwin
  #if !defined(__CYGWIN__)
    // sm: 'salen' is the length of buffer pointed to by 'sa'; similarly for
    // hostlen/host and servlen/serv.. but these last two aren't checked,
    // and checking them would require adding a model to force them to be
    // at least fseq.. I'll skip that hassle for now
    int getnameinfo(/*const struct sockaddr*/ void *sa, int salen,
                    char *host, int hostlen, char *serv, int servlen, int flags);
    int getnameinfo_wss(fatp_void sa, int salen,
                        char *host, int hostlen, char *serv, int servlen,
                        int flags)
    {
      verify_length(sa, salen);
      return getnameinfo(sa._p, salen, host, hostlen, serv, servlen, flags);
    }

    // not a socket function, but specific to unix, so I'll put it here
    //void *mmap(void *start, int length, int prot, int flags, int fd, int offset);
    seqp_void mmap_qs(void *start, int length, int prot, int flags, int fd, int offset)
    {
      seqp_void ret;

      // the only way it makes sense for 'start' to be safe is if it's null
      if (start != NULL) {
        FP_FAIL_STR("expected NULL start to mmap");
      }

      ret._p = mmap(start, length, prot, flags, fd, offset);
      if (ret._p == MAP_FAILED) {
        ret._b = ret._e = NULL;
      }
      else {
        ret._b = ret._p;
        ret._e = ((char*)ret._p) + length;    // ptr to first *invalid* byte
      }

      return ret;
    }

    int munmap_q(seqp_void start, int length)
    {
      CHECK_NULL(start._p);
      if (start._p != start._b) {
        FP_FAIL_STR("must pass pointer to start of the region");
      }
      if ((char*)start._e - (char*)start._b  != length) {
        FP_FAIL_STR("must pass correct length to munmap");
      }

      // fixme?  this could leave dangling pointers.. it's a form
      // of explicit deallocation..
      return munmap(start._p, length);
    }

    // glob_t as inferred by our system with ftpd
    struct wrapped_glob_t {
       unsigned int gl_pathc ;
       fseqp_fseqp_char gl_pathv ;
       unsigned int gl_offs ;
       int gl_flags ;
       void (* gl_closedir)(void * ) ;
       struct dirent * (* gl_readdir)(void * ) ;
       void * (* gl_opendir)(char * ) ;
       int (*  gl_lstat)(char * , struct stat * ) ;
       int (*  gl_stat)(char * , struct stat * ) ;
    };
    
    int wrapped_glob(const char *pattern, int flags,
                     int errfunc(const char * epath, int eerrno),
                     struct wrapped_glob_t *pglob)
    {
      glob_t gl;
      int err, totlen, i;
      char *bigArray;

      // check that we can handle the specified flags
      if (flags & (GLOB_DOOFFS | GLOB_APPEND | GLOB_ALTDIRFUNC)) {
        // none of these are especially difficult, I just don't want to
        // waste time implementing and testing them when I don't need them
        FP_FAIL_STR("some glob flags are specified that I can't handle (yet)");
      }

      // call underlying globber
      err = glob(pattern, flags, errfunc, &gl);
      if (err) {
        return err;
      }

      // copy results into my own area, so I can free the strings that
      // underlying globber allocated

      // first, allocate a big buffer to hold all the strings
      for (totlen=0, i=0; i < gl.gl_pathc; i++) {
        int len = strlen(gl.gl_pathv[i]);
        totlen += len + 1;
      }
      bigArray = (char*)malloc(totlen);

      // allocate an array of fseq pointers, to point into this array; the
      // way ftpd uses this, it seems to think the array is terminated by
      // a null pointer, though the man page doesn't mention this ..
      pglob->gl_pathv._p = (fseqp_char*)malloc(sizeof(fseqp_char) * (gl.gl_pathc+1));
      pglob->gl_pathv._e = pglob->gl_pathv._p + (gl.gl_pathc+1);    // max+1

      // now copy the strings one at a time
      for (totlen=0, i=0; i < gl.gl_pathc; i++) {
        int len = strlen(gl.gl_pathv[i]);
        strcpy(bigArray+totlen, gl.gl_pathv[i]);
        pglob->gl_pathv._p[i]._p = bigArray+totlen;
        pglob->gl_pathv._p[i]._e = bigArray+totlen+len+1;   // max+1
        totlen += len + 1;
      }

      // write a final null
      pglob->gl_pathv._p[i]._p = NULL;
      pglob->gl_pathv._p[i]._p = NULL;

      // now that we've copied all the strings (into our garbage-collected
      // heap), we can deallocate whatever libc did
      globfree(&gl);

      return err;
    }

    void wrapped_globfree(struct wrapped_glob_t *pglob)
    {
      // we use a garbage collector
    }
  #endif // !defined(__CYGWIN__)

#endif // _GNUCC
// --------- end of socket functions --------


char* ccured_kind_of() {
  return "SAFE";
}
char* ccured_kind_of_sq() {
  return "SEQ";
}
char* ccured_kind_of_sf() {
  return "FSEQ";
}
char* ccured_kind_of_sw() {
  return "WILD";
}

struct {
  int len;
  char data[8];
  int tags;
} theWild = { 2, "WILD", 0 };


fatp_char ccured_kind_of_ww() {
  fatp_char res;
  res._p = theWild.data;
  res._b = res._p;
  return res;
}
