#!/usr/bin/perl
# A Perl script that substitutes either the GCC or the MSVC compiler. Invokes
# the preprocessor, then another preprocessor and then the real compiler.
#
use strict;
use File::Basename;
use File::Copy;
use FindBin;
if($^O eq 'MSWin32') {
    require Win32;
}

sub printHelp {
    print <<EOL;
Wrapper for GCC and MSVC. 
Usage: safecc [args] [cargs]

args: 
  --help       Prints this help message
  --verbose    Prints a lot of information about what is being done
  --mode=xxx   What tool to emulate: 
                gcc     - GNU CC
                mscl    - MS VC cl compiler
                mslink  - MS VC link linker
               By default it chooses between mscl and gcc based on the form of
               the command line arguments. It chooses gcc if it cannot tell
               by the form of the command line arguments. 

  --patchinclude=xxx Use the directory xxx as a source for patched includes
  --combine    Combine all the sources into one file
  --profile    Compiles and link with profile information

The rest are safec specific arguments:

  --cabs       Uses the CABS version of the file. Stop after parsing.
  --cil        Uses the CIL version of the file (default)
  --box        Uses the BOX version of the file
  --inferbox   Infers the pointer types
  --nobox=xxx  Do not combine and do not box xxx
  --optim      Removes redundant checks
  --release    Runs with the native code version
  --keep=dir   Keep intermediate files in dir
  --safec=xxx  Passes xxx to the safec executable. Multiple occurences of 
               --safec should be used to add more arguments.
  --tr=flag    Passes "-tr flag" (tracing flag) to the safec executable.
  --profilesafec Expects a native code version of the safec executable compiled with
               profiling information. Sums up the profiler statistics into gmon.sum.
               This is needed because the profiler statistics get overwritten in
               gmon.out with each invocation of the executable.

Send bugs to necula\@cs.berkeley.edu.
EOL
}

# Go through the arguments and separate the source names from the other
# arguments. Among the arguments separate those that can be passed to the
# preprocessor (@ppargs) and those that should be passed only to the final
# phase (@otherargs)
my @csources  = ();             # The C or C++ sources
my @osources  = ();             # Other sources passed on the command line

my @ppargs = ();               # Arguments for the preprocessor
my @ccargs = ();               # Arguments for the compiler
my @linkargs = ();             # Arguments for the linker
my $outarg;                    # Argument that names the output file

my @orig_ppargs;
my @orig_ccargs;

my @nobox = ();                #  Files to not box

my $compiler;                  # an instance of either MSVC or GCC packages

my $operation = "TOEXE";
   # One of "PREPROC"  - preprocess only
   #        "TOASM"    - compile to assembly language
   #        "TOOBJ"    - compile to object file
   #        "TOLIB"    - link to library
   #        "TOEXE"    - link to executable (default)

my $debugpatch = 0;
my $verbose = 0;
my $doBox = 0;
my $doInferBox = 0;
my $doOptim=0;
my $profileSafec=0;
my $profile = 0;
my $doCil = 0;
my $doCabs = 0;
my $doCombine = 0;
my $doTV;
my $tvCommand;
my $useRelease = 0;
my $keepDir;                    # Keep outputs here
my $printInferbox = 1;          # whether to print XXX_inferbox.c
my $extraArgs = "";

my $bindir = $FindBin::Bin;

my $safecc;
my $combiner;
my $safeccbin;

&main();


################# Now the subroutine definitions
sub main {
    # Get the command line arguments
    &collectArguments();

    # If we are compiling one file and the output file is not specified     
    # we add it to make sure that no matter how the name of the file
    # to be compiled changes, we produce the required output file
    if($#csources == 0 && $operation == "TOOBJ" && !defined($outarg)) {
        $outarg = 
            $compiler->{'OUTOBJ'} . $compiler->compileOutputFile($csources[0]);
    }
    # Same for executable
    if($operation == "TOEXE" && !defined($outarg)) {
        $outarg = 
            $compiler->{'OUTEXE'} . $compiler->linkOutputFile($csources[0]);
    }
    # See if proofiling is desired
    if($profile) {
        push @ccargs, $compiler->{CPROFILEARG};
        push @linkargs, $compiler->{LPROFILEARG};
    }
    # Now preprocess all C source files 
    my $src;
    my @newcsources = ();
    foreach $src (@csources) {
        push @newcsources, &preprocessFile($src);
    } 
    @csources = @newcsources; # Replace the sources

    
    # Now the files are preprocessed already 
    # If combining 
    if($doCombine) {
        if($operation eq "PREPROC") {
            return;
        }
        if($operation eq "TOOBJ") {
            &combineSourcesCompile();
            return;
        }
        if($operation eq "TOEXE") {
            &combineSourcesLink(); # Changes csources and osources
            goto Finish;
        }
        die "Don't know what to do";
    } 
  Finish:
    # print "CSOURCES = ", join(' ', @csources), "\n";
    # print "OSOURCES = ", join(' ', @osources), "\n";
    @newcsources = ();
    foreach $src (@csources) {
        push @newcsources, &safecPreprocessFile($src);
    }
    @csources = @newcsources;
    
    # Finish the operation
    if($operation eq "PREPROC") {
        return;
    }
    if($operation eq "TOOBJ") {
        # Put the osources into csources. Maybe the compiler knows what to do
        push @csources, @osources;
        # And put the preprocess args into the ccargs
        push @ccargs, @ppargs;
        $compiler->compile(); 
        return;
    }
    if($operation eq "TOEXE") {
        $compiler->link();
        return;
    }
    die "Don't know what to do: $operation\n";
}

###
### A few library routines to help with file name management
###
# Return true if paths should have backward slashes.
my $pathSep = ($^O eq "MSWin32") ? '\\' : '/';
my $dirListSep = ($^O eq "MSWin32") ? ';' : ':'; 

sub openWithMounts {
    my ($fileName, $fdesc) = @_;
    
    my $openFileName = $fileName;
    # Try it outright
    if(open($fdesc, "<$openFileName")) {
        return 1;
    }
    # If we fail then maybe we are using cygwin paths in a Win32 system
    if($compiler->{MODENAME} eq "gcc" && $^O eq 'MSWin32') {
        open(WINNAME, "cygpath -w $fileName|")
            || die "Cannot run cygpath to convert $fileName to a Windows name";
        my $newName = <WINNAME>;
        # print "Converted $fileName to $newName\n";
        close(WINNAME) || die "Cannot run cygpath to convert $fileName";
        if(open($fdesc, "<$newName")) {
            return 1;
        }
    }
    return 0;
}

# Is absolute path name?
sub isAbsolute {
    my($name) = @_;
    if($^O eq "MSWin32") {
        return ($name =~ m%^([a-zA-Z]:)?[/\\]%);
    } else {
        return ($name =~ m%^[/\\]%);
    }
}

# Is it on the same volume?
sub sameVolume {
    my($f1, $f2) = @_;
    if($^O eq "MSWin32") {
        my($v1) = ($f1 =~ m%^([a-zA-Z]:)%);
        my($v2) = ($f2 =~ m%^([a-zA-Z]:)%);
        return (defined $v1 && defined $v2 && lc($v1) eq lc($v2));
    } else {
        return 1;
    }
}


# Are we on Win95?
sub isWin95 {
    if($^O eq 'MSWin32') {
        no strict 'subs'; # To avoid compilation errors on Unix 
	return Win32::IsWin95;
    } else {
	return 0;
    }
}

my $useRelative;           # Whether to use relative paths and to which
                           # directory 
# Take a PATH and normalize it. If $useRelative is the name of a directory
# then relativize with respect to that directory
sub path {
    my ($name) = @_;
    # print "Path on $name is ";
    if ($^O eq "MSWin32") {
        $name =~ s%/%\\%g;
    }
    # Get rid of final /
    if($name =~ m%[/\\]$%) {
        chop $name;
    }
    # Get rid of /./
    $name =~ s|[/\\]\.[/\\]|$pathSep|g;
    $name =~ s|^\.[/\\]||;
    # print "Name = $name\n";
    if(defined($useRelative)) {
        # print " (relative to $PCC::useRelative) ";
        if(&isAbsolute($name) &&
           &isAbsolute($useRelative) &&
           &sameVolume($name, $useRelative)) {
            # print "\n====Here\n";
            my @pbase = split(m%[/\\]%, $useRelative);
            my @path  = split(m%[/\\]%, $name);
            while(@pbase && @path && lc($pbase[0]) eq lc($path[0])) {
                shift @pbase;
                shift @path;
            }
            # print "Here base=@pbase and this=@path\n";
            my $nname = join($pathSep, @path);
            my $nbase = '';
            my $cnt;
            if($#pbase == -1) {
                $nbase = '';
            } elsif ($#pbase == 0) {
                $nbase = '..';
            } else {
                $nbase = '.';
                foreach $cnt (@pbase) {
                    $nbase = "$nbase$pathSep..";
                }
            }
            if($nname eq '' || $nbase eq '') {
                $nname = "$nbase$nname";
                if($nname eq '') {
                    $nname = '.';
                }
            } else {
                $nname = "$nbase$pathSep$nname";
            }
            if(length($nname) < length($name)) {
                $name = $nname;
            }
        }
    }
    # print " $name\n";
    return $name;
}

# Read the user time (in milliseconds)
sub getTime {
    if($^O eq "MSWin32") {
	no strict 'subs';
	return Win32::GetTickCount;
    } else {
	my ($u,$s,$cu,$cs) = times;
	return int(1000 * ($cu + $cs));
    }
}
    

sub quoteIfNecessary {
    my($arg) = @_;
    # If it contains spaces or "" then it must be quoted
    if($arg =~ m|\s| || $arg =~ m|\"|) {
        return "\'$arg\'";
    } else {
        return $arg;
    }
}

sub collectArguments {
    my @patchincludes = ();
    while($#ARGV >= 0) {
        my $arg = shift @ARGV;
        # Handle first our own arguments
        if($arg eq "--help") {
            &printHelp ();  exit 1;
        }
        if($arg =~ m|^--patchinclude=?(.*)$|) {
            push @patchincludes, $1; next;
        }
        if($arg =~ m|--mode|) {
            my $mode;
            if($arg =~ m|--mode=?(.+)$|)  {
                $mode = $1;
            } else {
                $mode = shift @ARGV;
            }
            # Now set the compiler mode
            &setCompilerMode($mode, $arg);
            next;
        }
        if($arg eq "--verbose") {
            $::verbose = 1; next;
        }
        if($arg =~ m|^--keep=(.+)$|) { # Where to keep the outputs
            $keepDir = $1; next;
        }
        if($arg =~ m|^--safec=(.*)$|) { # Extra arguments
            $extraArgs .= " $1"; next;
        }
        if($arg =~ m|^--tr=(.*)$|) {    # Extra tracing arguments
            $extraArgs .= " -tr $1";  next;
        }
        if($arg eq "--box") { # Whether to box or not
            $doBox = 1;   next;
        }
        if($arg eq "--cabs") {
            $doCabs = 1;   next;
        }
        if($arg eq "--combine") { # Whether to combine or not
            $doCombine = 1;   next;
        }
        if($arg =~ m|^--nobox=(.*)$|) { # Whether to leave out some files
            push @nobox, $1; next;
        }

        if($arg eq "--inferbox") { # Whether to infer the boxing or not
            $doInferBox = 1; next;
        }
        if($arg eq "--cil") { # Whether to generate the cil or not
            $doCil = 1; next;
        }
        if($arg =~ m|^--tv=?(.*)$|) { # Whether to use translation validation.
            $doTV = $1;  next;
        }
        if($arg eq "--optim") { # Whether to generate the cabs or not
            $doOptim = 1;  next;
        }
        if($arg eq "--release") {
            $useRelease = 1;  next;
        }
	if($arg eq "--profilesafec") {
	    $profileSafec = 1; next;
	}
	if($arg eq "--profile") {
	    $profile = 1; next;
	}
	if($arg eq "--noPrintInferbox") {
	    $printInferbox = 0; next;
	}
        # now detect whether we are in MSVC or in GCC
        if($arg =~ m|^-|) {
            &setCompilerMode("gcc", $arg);
        }
        # weimer: this regexp does not catch /unix/path/to/file
        # as an MSVC extension
        if($arg =~ m|^/[^/]+$| && ! defined($compiler)) {
            &setCompilerMode("mscl", $arg);
        }
        if(! defined $compiler) {     # Must be a source file, since we have
                                      # not seen an option yet
            my ($base, $dir, $ext) = 
                fileparse($arg, 
                          "(\\.c)|(\\.cc)|(\\.cpp)|(\\.i)");
            if($ext ne "") {
                push @csources, $arg;
            } else {
                push @osources, $arg;
            }
            next;
        }
        &classifyArguments($compiler->{OPTIONS}, $arg);
    }
    # Fixup some arguments
    if(defined $keepDir && $keepDir !~ m|[/\\]$|) {
        $keepDir = $keepDir . "/";
    }
    my $version = ($useRelease | $profileSafec) ? "asm" : "byte";

    $safeccbin = "$bindir/../obj/ccured.$version.exe";
    $combiner = "$bindir/../obj/combiner.$version.exe";
    $safecc = $safeccbin;
    $safecc .= " -cabsindent 2 -verbose $extraArgs ";
    if($compiler->{MODENAME} eq "mscl") {
        $safecc .= " -msvc ";
        $combiner .= " -msvc ";
    }

    # Save the argument lists
    @orig_ppargs = @ppargs;
    @orig_ccargs = @ccargs;

    # See if we must use patchedincludes
    my $pi;
    foreach $pi (@patchincludes) {
        unshift  @ppargs, 
                 $compiler->{INCARG} . $pi . "/" . $compiler->version(), 
    }

    if(! $doCabs && ! $doCil && ! $doBox) {
        $doCil = 1;
    }
    if($doCabs) {
        $doCil = 0; $doBox = 0;
    }
    if($doCil) {
        $doBox = 0;
    }
    if(! $doBox) {
        $doInferBox = 0;
    }
    if($doBox) {
        push @ppargs, $compiler->{DEFARG} . 'CCURED';
        if($operation eq "TOEXE") {
            push @osources, 
            "$bindir/../obj/cillib" . 
                ($useRelease ? "" : "debug") . ".$compiler->{LIBEXT}";
            push @osources, 
            "$bindir/../obj/ccured" . 
                ($useRelease ? "" : "debug") . ".$compiler->{LIBEXT}";
        }
        push @ccargs, $compiler->{INCARG} . $bindir;
        push @ppargs, $compiler->{INCARG} . $bindir;
        push @ccargs, $useRelease ? $compiler->{OPTIM} : $compiler->{DEBUG};
    } else {
        # $doCil = 1;
        
        # sm: Rahul added this on 6/4/01 so it links the runtime even 
        # in 'cil' mode; this substantially changes things, but I'm 
        # going to try to work with it anyway. It seems that we need
        # get_stderr from the library

        # GN. Do not add the safec.a but only the cillib.a
        if($operation eq "TOEXE") {
            push @osources, 
            "$bindir/../obj/cillib" . 
                ($useRelease ? "" : "debug") . ".$compiler->{LIBEXT}";
        }
    }
    if($::verbose) {
        print $compiler->{NAME}, " mode\n";
    }
    
    if(defined $doTV) {
        if($doTV ne "") { $doTV .= " -debug"; }

        if($compiler->{MODENAME} ne "gcc") {
            print "***WARNING: Translation validation only works with GCC\n";
            undef $doTV;
        }

       # Find the translation validator
        my $which = $doTV eq "" ? "asm" : "byte";
        my $tvdir = path("$bindir/../../../Source/TransVal");
        $tvCommand = path("$tvdir/obj/transval.$which.exe");
        if(-f $tvCommand) {
            print "Found $tvCommand\n";
        }
    }
    
}



################ Now the subroutines
    
sub classDebug {
    if(0) { print @_; }
}
sub classifyArguments {
    my($options, $arg) = @_;

    &classDebug("Classifying arg: $arg\n");
    my $idx = 0;
    for($idx=0; $idx < $#{$options}; $idx += 2) {
        my $key = ${$options}[$idx];
        my $action = ${$options}[$idx + 1];
        &classDebug("Try match with $key\n");
        if($arg =~ m|^$key|) {
            &classDebug(" match with $key\n");
            my $onemore;
            if(defined $action->{'ONEMORE'}) {
                &classDebug("  expecting one more\n");
                if($arg eq $key) { # Grab the next argument
                    $onemore = shift @ARGV;
                } else { # The next argument is actually attached
                    ($arg, $onemore) = ($arg =~ m|^($key)(.+)$|);
                }
                $onemore = &quoteIfNecessary($onemore);
            }
            &classDebug(" onemore=$onemore\n");
            my $fullarg = defined($onemore) ? "$arg $onemore" : $arg;
            # Now see what action we must perform
            if(defined $action->{'TYPE'}) {
                &classDebug("  type=$action->{TYPE}\n");
                if($action->{TYPE} eq "PREPROC") {
                    push @ppargs, $fullarg; return;
                }
                if($action->{TYPE} eq "CC") {
                    push @ccargs, $fullarg; return;
                }
                if($action->{TYPE} eq "LINKCC") {
                    push @ccargs, $fullarg; push @linkargs, $fullarg; return;
                }
                if($action->{TYPE} eq "LINK") {
                    push @linkargs, $fullarg; return;
                }
                if($action->{TYPE} eq "CSOURCE") {
                    push @csources, $fullarg; return;
                }
                if($action->{TYPE} eq "OSOURCE") {
                    push @osources, $fullarg; return;
                }
                if($action->{TYPE} eq 'OUT') {
                    if(defined($outarg)) {
                        print "Warning: output file is multiply defined: $outarg and $fullarg\n";
                    }
                    $outarg = $fullarg; return;
                }
                print "  Do not understand TYPE\n"; return;
            }
            if(defined $action->{'RUN'}) {
                &{$action->{'RUN'}}($fullarg, $onemore); return;
            }
            print "Don't know what to do with $arg\n"; return;
        }
    }
    print "Unrecognized option $arg\n";
}


sub runShell {
    my ($cmd) = @_;

    # sm: I want this printed to stderr instead of stdout
    # because the rest of 'make' output goes there and this
    # way I can capture to a coherent file
    print STDERR "$cmd\n";

    # weimer: let's have a sanity check
    if (system($cmd)) {
	die "Possible error with $cmd!\n";
    }
}



sub setCompilerMode {
    my ($mode, $arg) = @_;
    if(defined $compiler && $compiler->{MODENAME} ne $mode) {
#        die "Found a $mode argument ($arg) while in $compiler->{NAME} mode";
    }
    if($mode eq "mscl") {
        $compiler = MSCL->new; return;
    }
    if($mode eq "gcc") {
        $compiler = GCC->new; return;
    }
    if($mode eq "mslink") {
        $compiler = MSLINK->new; return;
    }
    die "Don't know about compiler $mode\n";
}



sub preprocessFile {
    my($src) = @_;

    my ($base, $dir, $ext) = fileparse($src, "\\.[^.]+");

    my $idir = $dir;
    if(defined $::keepDir) {
        $idir = $::keepDir;
    }
    # print "base=$base, dir=$dir, idir=$idir, ext=$ext\n";
    my $interm = $src;
    my $out    = "$idir$base.i";

    # sm: moved from main(); I want this to be done only on
    # the *last* preprocess.
    # To clean some things in the files
    my @savedppargs = @ppargs;
    push @ppargs, $compiler->forceIncludeArg("$bindir/fixup.h");
    if(grep {$_ eq $base} @nobox) {
        # Fist take out the defined CCURED from @ppargs
        my $remove = $compiler->{DEFARG} . 'CCURED';
        @ppargs = grep { $_ ne $remove } @ppargs;
    }
    
    # Now call preprocessor
    $compiler->cpp($interm, $out);
    @ppargs = @savedppargs;

    # Now cleanup
    if($useRelease) {
        if($interm ne $src) {
            unlink $interm;
        }
    }
    return $out;
}


########################################################################
###
### COMBINE SOURCES
###
###
sub combineSourcesCompile {
    my $src;
    my @savecsources = @csources;

    foreach $src (@savecsources) {
        # See if this is one of those that we should not combine
        my ($base, $dir, $ext) = fileparse($src, "\.[^.]+");
        if(grep {$_ eq $base} @nobox) {
            # Do not box this. Just compile it right away
            @csources = ($src);
            if($::verbose) {
                print "Compiling nobox file $src\n";
            }
            $compiler->compile();
        } else {
           # Get the name of the output file
            my $outfile = $compiler->compileOutputFile($src);
            open(OUT, ">$outfile") 
                || die "Cannot create $outfile";
            if($::verbose) {
                print "Creating fake object file $outfile from $src\n";
            }
            my $toprintsrc = $src;
            $toprintsrc =~ s|\\|/|g;
            print OUT "#pragma combiner(\"$toprintsrc\", \"" . 
                join(' ', @ccargs), "\")\n";
            # print OUT $compiler->lineDirective($toprintsrc, 1);
            open(IN, "<$src") 
                || die "Cannot read $src";
            print OUT <IN>;
            close(OUT);
            close(IN);
        }
    }
    @csources = @savecsources;
}

sub combineSourcesLink {
    # Go over all the sources and separate those that are C
    my $src;
    # First add any C sources that we were given
    my @tocombine = @csources;
    my @newosources = ();
    @csources = ();
    foreach $src (@osources) {
        my ($base, $dir, $ext) = fileparse($src, "\\.[^\\.]+");
        if($ext eq ".obj" || $ext eq ".o" || $ext eq ".lib" || $ext eq ".a") {
            # Look inside and see if it is a C or not
            open(IN, "<$src") || die "Cannot read $src";
            my $fstline = <IN>;
            if($fstline =~ m|\#pragma combiner\(|) {
                push @tocombine, $src; next; 
            } else {
                push @newosources, $src; next;
            }
        }
        if($ext eq ".c" || $ext eq ".i") {
            if(grep {$_ eq $base} @nobox) {
                if($::verbose) {
                    print "Skipping nobox file $src\n";
                }
                push @csources, $src; next;
            } else {
                push @tocombine, $src; next;
            }
        }
        die "I don't know how to combine (link) $src\n";
    }
    @osources = @newosources;

    # Now call the combiner 
    my $outfile = $compiler->linkOutputFile();
    my $combinedfile = $outfile . "_all.c";
    my $cmd = "$combiner -o $combinedfile " . join(' ', @tocombine);
    &profilerErrandsPre ();    
    &::runShell($cmd);
    &profilerErrandsPost ();    
    push @csources, $combinedfile;
 }


# AB: Delete the existing gmon.out file ... caused problems sometimes
sub profilerErrandsPre () {
    &::runShell ("rm -f gmon.out");
}

# Keeps a running sum of the gmon.out files (produced by the binary containing profiler
# info) in gmon.sum
# Maintains a copy in the safecc.asm.exe directory to this gmon.sum
# (gprof doesnt like soft links and AB doesn't like too many hard links)
sub profilerErrandsPost {
   if ($profileSafec) {
       # AB: gmon.?u? contains gmon.out and also gmon.sum if it already exists
       #     This way, we keep a running sum in gmon.sum (must be cleaned everytime)
       &::runShell("gprof -s $safeccbin gmon.?u? && cp -f gmon.sum $bindir/../obj/");
    }
}


########################################################################
###
### SAFEC CODE
###
sub safecPreprocessFile {
    my ($src) = @_;

    my ($base, $dir, $ext) = fileparse($src, "\\.[^.]+");

    my $src = "$dir$base$ext";
    my $version = ($useRelease | $profileSafec) ? "asm" : "byte";
    my $safecc = "$bindir/../obj/safec.$version.exe";
    my $cmd = "$safecc -cabsindent 2 -verbose $extraArgs ";
    if($compiler->{MODENAME} eq "mscl") {
        $cmd .= " -msvc ";
    }
    my $result;
    my $resultExt;
    if($doCabs) {
        $resultExt = "cabs.c";
        $cmd .= " -cabsout $dir$base$resultExt";
    }
    if($doCil) {
        $resultExt = "cil.c";
        $cmd .= " -cilout $dir$base$resultExt";
    }
    if($doInferBox) { # Do not change the result, but produce one more file
        if((! $useRelease) && $printInferbox) {
            $cmd .= (" -inferout $dir$base" . "infer.c ");
        }
    }
    if($doBox) {  
        $cmd .= (" -curedout $dir$base" . "box.c ");
        $resultExt = "box.c";
        if($doOptim) {
            $cmd .= (" -optimout $dir$base" . "box.optim.c ");
            $resultExt = "box.optim.c";
        }
    }
    $result = "$dir$base$resultExt";
    $cmd .= " $src";
    &profilerErrandsPre ();
    &runShell($cmd);
    &profilerErrandsPost ();

    # Now run translation validation
    if(defined $doTV) {
        print "Running GCC to extract RTL\n";
        my $makeRTL = "gcc -g -c -dr ";
        my $rtlDir = defined($keepDir) ? $keepDir : "./";
        my $srcRTL = "$rtlDir/$base$ext.rtl";
        my $trgRTL = "$rtlDir/$base$resultExt.rtl";
        &runShell("$makeRTL $src");
        &runShell("$makeRTL $result");
        # The output is in the current directory
        if($rtlDir !~ m|^.[/\\]$|) {
            my $orig = "./$base$ext.trl";
            print "Copying $orig to $srcRTL\n";
            unlink $trgRTL;
            &File::Copy::copy($orig, $srcRTL);
            unlink $orig;

            my $orig = "./$base$resultExt.trl";
            print "Copying $orig to $trgRTL\n";
            unlink $trgRTL;
            &File::Copy::copy($orig, $trgRTL);
            unlink $orig;
        }
        print "Running the translation validator\n";
        &runShell("$tvCommand -L tv.log $doTV $src.rtl $result.rtl");
    }

    return $result;
}

###########################################################################
####
#### MS CL specific code
####
package MSCL;

use strict;
use File::Basename;

sub new {
    my ($proto, %args) = @_;
    my $class = ref($proto) || $proto;
    # Create $self

    my $self = 
    { NAME => 'Microsoft cl compiler',
      MODENAME => 'mscl',
      DEFARG  => "/D",
      INCARG  => "/I",
      DEBUGARG => "/Zi /MLd /DEBUG",
      OPTIMARG => "/Ox /G6 ",
      OBJEXT => "obj",
      LIBEXT => "lib",   # Library extension (without the .)
      EXEEXT => ".exe",  # Executable extension (with the .)
      OUTOBJ => "/Fo",
      OUTEXE => "/Fe",
      LINEPATTERN => "^#line\\s+(\\d+)\\s+\"(.+)\"",

      OPTIONS => 
# Describe the compiler options as a list of patterns and associated actions. 
# The patterns are matched in order against the begining of the argument.
#
# If the action contains ONEMORE => 1 then the argument is expected to be
# parameterized by a following word. The word can be attached immediately to
# the end of the argument or in a separate word. 
#
# If the action contains TYPE => "..." then the argument is put into one of
# several lists, as follows: "PREPROC" in ppargs, "CC" in ccargs, "LINK" in
# linkargs, "LINKCC" both in ccargs and linkargs, "CSOURCE" in csources,
# "OSOURCE" in osources, "OUT" in outarg. 
#
# If the TYPE is not defined but the RUN => sub { ... } is defined then the
# given subroutine is invoked with the argument and the (possibly empty)
# additional word.
#
          ["[^/].*\\.(c|i|cpp|cc)" => { TYPE => 'CSOURCE' },
           "[^/]" => { TYPE => "OSOURCE" },
           "/O" => { TYPE => "CC" },
           "/G" => { TYPE => "CC" },
           "/[DI]" => { TYPE => "PREPROC"},
           "/EH" => { TYPE => "CC" },
           "/G"  => { TYPE => "CC" },
           "/F[aA]"  => { TYPE => 'CC' },
           "/Fo"     => { TYPE => 'OUT' },
           "/F[e]"   => { TYPE => 'OUT' },
           "/F[dprR]" => { TYPE => "CC" },
           "/[CXu]" => { TYPE => "PREPROC" },
           "/U" => { ONEMORE => 1, TYPE => "PREPROC" },
           "/(E|EP|P)" => { RUN => sub { push @ppargs, $_[0]; 
                                         $operation = "PREPROC"; }},
           "/c" => { RUN => sub { $operation = "TOOBJ"; }},
           "/(Q|Z|J|nologo|TC|TP|w|W|Yd|Zm)" => { TYPE => "CC" },
           "/v(d|m)" => { TYPE => "CC" },
           "/F" => { TYPE => "CC" },
           "/M"   => { TYPE => 'LINKCC' },
           "/link" => { RUN => sub { push @linkargs, "/link", @ARGV;
                                     @ARGV = (); } },
           "/"  => { RUN => 
                         sub { print "Unimplemented MSVC argument $_[0]\n"; }},
           ],
      };
    bless $self, $class;
    return $self;
}


sub forceIncludeArg { 
    my($self, $what) = @_;
    return "/FI$what";
}

sub cpp {
    my ($self, $src, $afterpp) = @_;
#    my ($base, $dir, $ext) = fileparse($afterpp, 
#                                       "(\\.c)|(\\.cc)|(\\.cpp)|(\\.i)");
    my ($sbase, $sdir, $sext) = fileparse($src, 
                                       "(\\.c)|(\\.cc)|(\\.cpp)|(\\.i)");
    my $cmd = "cl /nologo /P /D_MSVC " . join(' ', @ppargs);
    &::runShell("$cmd $src");
    # MSVC cannot be told where to put the output. But we know that it
    # puts it in the current directory
    my $msvcout = "./$sbase.i";
    my @st1 = stat $msvcout;
    my @st2 = stat $afterpp;
    while($#st1 >= 0) {
        if(shift @st1 != shift @st2) {
#                print "$msvcout is NOT the same as $afterpp\n";
            if($::verbose) {
                print "Copying $msvcout to $afterpp\n";
            }
            unlink $afterpp;
            &File::Copy::copy($msvcout, $afterpp);
            unlink $msvcout;
            return;
        }
    }
#        print "$msvcout is the same as $afterpp\n";
}

    # MSCL does not understand the extension .i, so we tell it it is a C file
sub fixupCsources {
    my @mod_csources = ();
    my $src;
    foreach $src (@csources) {
        my ($sbase, $sdir, $sext) = fileparse($src, 
                                              "\\.[^.]+");
        if($sext eq ".i") {
            push @mod_csources, "/Tc";
        }
        push @mod_csources, $src;
    }
    return @mod_csources;
}

# Now run the real compilation
sub compile {
    my @mod_csources = &fixupCsources();
    my $cmd = "cl /nologo /D_MSVC /c " .
        join(' ', @ccargs) . " " . 
            join(' ', @mod_csources) . " $outarg";
    &::runShell($cmd);
}

sub link {
    my $cmd = "cl /nologo /D_MSVC " .
        join(' ', @ccargs) . " " . 
            join(' ', @linkargs) . " " .
                join(' ', @csources) . " " .
                    join(' ', @osources) . " $outarg";
    &::runShell($cmd);
}


# Emit a line # directive
sub lineDirective {
    my ($self, $fileName, $lineno) = @_;
    return "#line $lineno \"$fileName\"\n";
}

# The name of the output file
sub compileOutputFile {
    my($self, $src) = @_;
    if($outarg =~ m|/Fo(.+)|) {
        return $1;
    }
    my ($base, $dir, $ext) = fileparse($src, 
                                       "(\\.c)|(\\.cc)|(\\.cpp)|(\\.i)");
    if(! defined($ext) || $ext eq "") { # Not a C source
        die "objectOutputFile: not a C source file\n";
    }
    return "$base.obj"; # In the current directory
}

sub linkOutputFile {
    my($self, $src) = @_;
    if($outarg =~ m|/Fe(.+)|) {
        return $1;
    }
    die "I do not know what is the link output file\n";
}

sub version {
    my $cversion = "";
    open(VER, "cl " . join(' ', @ppargs) . " 2>&1|") 
        || die "Cannot start Microsoft CL\n";
    while(<VER>) {
        if($_ =~ m|Compiler Version (\S+) |) {
            $cversion = "cl_$1";
            close(VER);
            return $cversion;
        }
    }
    die "Cannot find Microsoft CL version\n";
}

#########################################################################
###
###  MS LINK specific code
###
####
package MSLINK;

use strict;
use File::Basename;

sub new {
    my ($proto, %args) = @_;
    my $class = ref($proto) || $proto;
    # Create $self

    my $self = 
    { NAME => 'Microsoft linker',
      MODENAME => 'mslink',
      DEFARG  => " ??DEFARG",
      INCARG  => " ??INCARG",
      DEBUGARG => "/DEBUG",
      OPTIMARG => "",
      OBJEXT => "obj",
      LIBEXT => "lib",   # Library extension (without the .)
      EXEEXT => ".exe",  # Executable extension (with the .)
      OUTOBJ => " ??OUTOBJ",
      OUTEXE => "/OUT:",
      LINEPATTERN => "", 

      OPTIONS => 
          ["[^/]" => { TYPE => 'OSOURCE' },
           "/OUT:" => { TYPE => 'OUT' },
           "/"  => { TYPE => 'LINK' },
           ],
      };
    bless $self, $class;
    return $self;
}


sub forceIncludeArg {  # Same as for CL
    my($self, $what) = @_;
    return "/FI$what";
}

## Now run the real compilation
#sub compile {
#    my $cmd = "cl /D_MSVC /c " .
#        join(' ', @ccargs) . " " . 
#            join(' ', @sources) . " $outarg";
#    &::runShell($cmd);
#}

# Compile and link
sub link {
    my $cmd = "cl /nologo /D_MSVC " .
        join(' ', @ppargs) . " " .
            join(' ', @csources) . " " . 
                join(' ', @osources) . 
                    " /link $outarg " . join(' ', @linkargs) ;
    &::runShell($cmd);
}


sub linkOutputFile {
    my($self, $src) = @_;
    if($outarg =~ m|/OUT:(.+)|) {
        return $1;
    }
    die "I do not know what is the link output file\n";
}

##########################################################################
###
### GCC specific code
###
package GCC;

use strict;
use File::Basename;

sub new {
    my ($proto, %args) = @_;
    my $class = ref($proto) || $proto;
    # Create $self

    my $self = 
    { NAME => 'GNU CC',
      MODENAME => 'gcc',  # do not change this since it is used in code
      DEFARG  => "-D",
      INCARG => "-I",
      DEBUGARG => "-g",
      OPTIMARG => "-O4",
      CPROFILEARG => "-pg ",
      LPROFILEARG => "-pg ",
      OBJEXT => "o",
      LIBEXT => "a",
      EXEEXT => "",
      OUTOBJ => "-o ",
      OUTEXT => "-o ",
      LINEPATTERN => "^#\\s+(\\d+)\\s+\"(.+)\"",
      
      OPTIONS => 
          [ "[^-].*\\.(c|i|cpp|cc)" => { TYPE => 'CSOURCE' },
            "[^-]" => { TYPE => 'OSOURCE' },
            "-[DI]" => { ONEMORE => 1, TYPE => "PREPROC" },
            "-include" => { ONEMORE => 1, TYPE => "PREPROC" },  # sm
            "-c" => { RUN => sub { $operation = "TOOBJ"; }},
            "-x" => { ONEMORE => 1, TYPE => "CC" },
            "-O" => { TYPE => "CC" },
            "-S" => { RUN => sub { $operation = "TOOBJ";
                                   push @ccargs, $_[0]; }},
            "-o" => { ONEMORE => 1, TYPE => 'OUT' },
            "-p" => { TYPE => 'LINKCC' },
            "-W" => { TYPE => 'CC' },
            "-g" => { RUN => sub { push @ccargs, $_[0];
                                   push @linkargs, $_[0]; }},
            "-l" => { TYPE => 'LINK' },
            "-f" => { TYPE => 'LINKCC' },
            "-r" => { TYPE => 'LINK' },
            "-m" => { TYPE => 'LINKCC' },
            "-traditional" => { TYPE => 'PREPROC' },
            "-" => { RUN => sub { print "Unimplemented GCC arg: $_[0]\n";
                                  push @ccargs, $_[0]; }}
            ],
                                  
      };
    bless $self, $class;
    return $self;
}

sub forceIncludeArg { 
    my($self, $what) = @_;
    return "-include $what";
}

sub cpp {
    my ($self, $src, $afterpp) = @_;
    my $cmd = "gcc -E -D_GNUCC " .
        join(' ', @ppargs) . " -o $afterpp";
    &::runShell("$cmd $src");
}

sub compile {
    my $cmd = "gcc -D_GNUCC -c " .
        join(' ', @ccargs) . " " . 
            join(' ', @csources) . " $outarg";
    &::runShell($cmd);
}

sub link {
    my $cmd = "gcc -D_GNUCC " .
        join(' ', @ccargs) . " " . 
            join(' ', @csources) .  " " . join(' ', @osources) .  " " .  
                join(' ', @linkargs) . "  $outarg";
    &::runShell($cmd);
}

# Emit a line # directive
sub lineDirective {
    my ($self, $fileName, $lineno) = @_;
    return "# $lineno \"$fileName\"\n";
}

# The name of the output file
sub compileOutputFile {
    my($self, $src) = @_;
    if($outarg =~ m|-o (.+)|) {
        return $1;
    }
    my ($base, $dir, $ext) = fileparse($src, 
                                       "(\\.c)|(\\.cc)|(\\.cpp)|(\\.i)");
    if(! defined($ext) || $ext eq "") { # Not a C source
        die "objectOutputFile: not a C source file\n";
    }
    return "$base.o"; # In the current directory
}

sub linkOutputFile {
    my($self, $src) = @_;
    if($outarg =~ m|-o (.+)|) {
        return $1;
    }
    return "a.out";
    die "I do not know what is the link output file\n";
}

sub version {
    my $cversion = "";
    open(VER, "gcc -dumpversion " . join(' ', @ppargs) ." |") 
        || die "Cannot start GCC";
    while(<VER>) {
        if($_ =~ m|^(\d+\S+)|) {
            $cversion = "gcc_$1";
            close(VER) || die "Cannot start GCC\n";
            return $cversion;
        }
    }
    die "Cannot find GCC version\n";
}
1;





