// splay.c
// code for splay.h

#include "splay.h"      // this module
#include "malloc.h"     // malloc_usable_size

#include <assert.h>     // assert
#include <stdio.h>      // printf
#include <string.h>     // memset


//#define TRACE_SPLAY_CALLS
#ifdef TRACE_SPLAY_CALLS
  #define DP(expr) expr
#else
  #define DP(expr) ((void)0)
#endif


// my usual bools-in-C definitions
enum { false=0, true=1 };
typedef int bool;

#ifdef _MSVC
#define NORETURN __declspec(noreturn)
#else
#define NORETURN __attribute__((noreturn))
#endif


// ---------------- data structure definition -----------------
struct SplayNode;

// top of the tree; packaged as a struct in case I want to
// add more fields, and because the implementation I'm working
// from does it this way
typedef struct SplayTree {
  struct SplayNode *root;
} SplayTree;

// data structure inserted before the user's
// data in every malloc block
typedef struct SplayNode {
  // threaded binary tree: left, right, parent
  struct SplayNode *left, *right, *parent;

  // why need parent?  it makes moving up the tree *much*
  // easier, and going up is what splaying is all about

  // added later: I now want to represent properly nested intervals,
  // rather than simply disjoint intervals; so I add the concept of
  // subintervals; note that the root of a subinterval tree has NULL
  // in its 'parent' field (just like the root of any splay tree)
  struct SplayTree subintervals;

  /* weimer: let's make these explicit */
  void * base;                  /* the base of the user data */
  unsigned int length;          /* the length of the user data */
  /* This splay node represents the region [base, base+length] */
} SplayNode;





// this tree is used when a 'kind' is not specified; it was for when
// we only had one tree; unfortunately, the one-tree and the multiple-
// tree modes of access have become mixed together in this file..
static SplayTree singleTree;


// we create nine trees: one for each combination of memory
// region and metadata kind
SplayTree __global_tree[NUM_MEMREGIONS][numAreaKinds] = {
  { {NULL}, {NULL}, {NULL} },
  { {NULL}, {NULL}, {NULL} },
  { {NULL}, {NULL}, {NULL} }
};


// prototypes for functions which operate on any splay tree,
// or on a node in any tree
SplayNode *splayInsert(SplayTree *t, SplayNode *n);
void splayDelete(SplayTree *t, SplayNode *n);
SplayNode *splaySplay(SplayTree *t, SplayNode *n);
SplayNode *splayLookup(SplayTree *t, void *ptr);
NORETURN void splayTreeInternalError(SplayTree *t, SplayNode *n, char *s);
void splayInternalTest(SplayTree *t, SplayNode *n, SplayNode *parent,
                       int left_of_parent, void *min, void *max);
void splayCheckTree(SplayTree *t);
void splayPrintTree(SplayTree *t);
void splayPrintNode(SplayNode *n, SplayNode *expectParent, int indent);
SplayNode *splayNondestructiveLookup(SplayNode *n, void *ptr);
SplayNode *innerSplayWalkTree(SplayWalkFn func, SplayNode *n);
void splaySubstitute(SplayNode *n, SplayNode *o);

void splayRemoveBelow(SplayTree *tree, void *cutoffAddr);
SplayNode *splayFindClosestAbove(SplayTree *tree, void *cutoffAddr);
void splayDeallocSubtree(SplayNode *node);

// functions which operate on 'singleTree'
void splayWalkTree(SplayWalkFn func);

// functions which operate on trees of a specified kind, but
// must determine which memory region to use on their own
void splayRemoveKindBelow(enum MemRegion mr, int kind, void *cutoffAddr);



// ================= splay tree code ==================
#ifdef TEST_SPLAY
  #define SPLAY_STRICT_CHECKING 1
#else
  // change this to 1 to have checking even in external uses
  #define SPLAY_STRICT_CHECKING 1       // sigh: FIXME
#endif


#if SPLAY_STRICT_CHECKING
  #define SPLAYCHECKTREE(t) splayCheckTree(t)
#else
  #define SPLAYCHECKTREE(t) ((void)0)
#endif


// given a splay node, return a pointer to the base of the
// user's data
static __inline void *base(SplayNode *n)
{
    return n->base;
}

// export this to the user too
void *mallocBlockToSplayBlock(void *mallocBlock)
{
    abort();    // weimer: fails!
    return base((SplayNode*)mallocBlock);
}

// map from block to node; assumes block is really a block
static __inline SplayNode *baseToNode(void *base)
{
    abort();    // weimer: fails!
    return (SplayNode*)( ((char*)base) - sizeof(SplayNode) );
}


#if 0    // prior version where size was stored in malloc header
  #if 0    // portable but slow due to call overhead
    // return the size of user's data in a node, in bytes
    static __inline size_t size(SplayNode *n)
    {
      return malloc_usable_size(n) - sizeof(SplayNode);
    }

  #else    // faster but uses knowledge of malloc data structures
    #define INTERNAL_SIZE_T size_t
    #define SIZE_SZ                (sizeof(INTERNAL_SIZE_T))

    struct malloc_chunk {
      INTERNAL_SIZE_T      prev_size;  /* Size of previous chunk (if free).  */
      INTERNAL_SIZE_T      size;       /* Size in bytes, including overhead. */

      struct malloc_chunk* fd;         /* double links -- used only if free. */
      struct malloc_chunk* bk;
    };

    typedef struct malloc_chunk* mchunkptr;

    #define PREV_INUSE 0x1
    #define IS_MMAPPED 0x2
    #define SIZE_BITS (PREV_INUSE|IS_MMAPPED)
    #define prev_inuse(p)       ((p)->size & PREV_INUSE)
    #define inuse(p)\
    ((((mchunkptr)(((char*)(p))+((p)->size & ~PREV_INUSE)))->size) & PREV_INUSE)
    #define mem2chunk(mem) ((mchunkptr)((char*)(mem) - 2*SIZE_SZ))
    #define chunk_is_mmapped(p) ((p)->size & IS_MMAPPED)
    #define chunksize(p)         ((p)->size & ~(SIZE_BITS))

    static __inline size_t size(SplayNode *n)
    {
      mchunkptr p = mem2chunk(n);
      if (chunk_is_mmapped(p))
        return chunksize(p) - 2*SIZE_SZ - sizeof(SplayNode);
      else
        return chunksize(p) - SIZE_SZ - sizeof(SplayNode);
    }
  #endif // 1/0

#else  // current version: size is a field
  static __inline size_t size(SplayNode *n)
  {
      return n->length;
  }
#endif // 1/0


// return a pointer to the last+1 byte
static __inline void *extent(SplayNode *n)
{
  return (void*)((char*)base(n) + size(n));
}

// return true if the given node is the left child of its parent
static __inline int left_of_parent(SplayNode *n)
{
  // the original code stored this as a bit in the splay node itself;
  // they may have done that for a reason, but at the moment there
  // isn't a convenient place for that bit, so let's do it this way
  assert(n->parent != NULL);
  return n->parent->left == n;
}

static __inline void set_left_of_parent(SplayNode *n, int val)
{
  // nop since I don't store this
}

static __inline int no_padding(SplayNode *n)
{
  // disable their extent==next thing
  return 1;
}


/* Add a node to the splay tree. We assume the node isn't already in the
 * tree, and that no node of the same value is in the tree. The node will
 * rise to the root of the tree (by splay), and we return it for convenience.
 */
SplayNode *splayInsert(SplayTree *t, SplayNode *n)
{
  SplayNode *tmp = t->root, *prev;

  n->parent = n->left = n->right = NULL;

#if SPLAY_COUNT_OPERATIONS
  add_operations++;
  splay_operations = &splay_add_operations;
  splay_steps = &splay_add_steps;
#endif

  if (tmp == NULL) {
    /* Tree is empty. Add the node at root and return immediately. */
    t->root = n;
    SPLAYCHECKTREE(t);
    return n;
  }

  // find a leaf node we can hang the new node off of
  while (tmp) {
    prev = tmp;
    if (base(n) < base(tmp)) {
      tmp = tmp->left;
    }
    else if (extent(n) < extent(tmp)) {
      // we want to insert a subinterval; this is the one place where
      // I modify the existing splay tree code to change (slightly)
      // the interpretation of 't', the passed-in root
      return splayInsert(&tmp->subintervals, n);
    }
    else {
      tmp = tmp->right;
    }
  }

  /* `prev' points to the last node before we fell off the bottom of the
   * tree.
   */
  if (prev->parent == NULL) {
    /* `prev' is the root of the tree, so we can just add the node here with
     * no problem.
     */
    if (base(n) < base(prev)) {
      prev->left = n;
      n->parent = prev;
      //n->left_of_parent = 1;
    }
    else {
      prev->right = n;
      n->parent = prev;
      //n->left_of_parent = 0;
    }
    return splaySplay(t, n);
  }

  /* If parent to prev and prev to new node are in the same direction, we
   * can just add the new node.
   */
  if (base(n) < base(prev) && left_of_parent(prev)) {
    prev->left = n;
    n->parent = prev;
    //n->left_of_parent = 1;
    return splaySplay(t, n);
  }
  if (base(n) > base(prev) && !left_of_parent(prev)) {
    prev->right = n;
    n->parent = prev;
    //n->left_of_parent = 0;
    return splaySplay(t, n);
  }

  /* If parent to prev and prev to new node are not in a straight line, we
   * insert the new node between parent and prev.
   */
  if (left_of_parent(prev)) {
    SplayNode *parent = prev->parent;

    parent->left = n;
    n->left = prev;
    n->parent = parent; //n->left_of_parent = 1;
    prev->parent = n; //prev->left_of_parent = 1;
  }
  else {
    SplayNode *parent = prev->parent;

    parent->right = n;
    n->right = prev;
    n->parent = parent; //n->left_of_parent = 0;
    prev->parent = n; //prev->left_of_parent = 0;
  }

  return splaySplay(t, n);
}


/* Delete a node from the tree. We assume that the node to be deleted exists
 * in the tree.
 */
void splayDelete(SplayTree *t, SplayNode *n)
{
  SplayNode *left, *right, *largest;
  SplayTree tmp_left;

#if SPLAY_COUNT_OPERATIONS
  del_operations ++;
  splay_operations = &splay_del_operations;
  splay_steps = &splay_del_steps;
#endif

  if (n->left == NULL) {
    right = n->right;
    if (n->parent == NULL) {
      t->root = right;
      if (right) {
        right->parent = NULL;
      }
    }
    else if (left_of_parent(n)) {
      n->parent->left = right;
      if (right) {
        right->parent = n->parent;
        //right->left_of_parent = 1;
      }
    }
    else {
      n->parent->right = right;
      if (right) {
        right->parent = n->parent;
        //right->left_of_parent = 0;
      }
    }
    SPLAYCHECKTREE(t);
    return;
  }

  if (n->right == NULL) {
    left = n->left;
    if (n->parent == NULL) {
      t->root = left;
      if (left) {
        left->parent = NULL;
      }
    }
    else if (left_of_parent(n)) {
      n->parent->left = left;
      if (left) {
        left->parent = n->parent;
        //left->left_of_parent = 1;
      }
    }
    else {
      n->parent->right = left;
      if (left) {
        left->parent = n->parent;
        //left->left_of_parent = 0;
      }
    }
    SPLAYCHECKTREE(t);
    return;
  }

  right = n->right;
  left = n->left;

  /* Search the left orphaned subtree for the largest value it contains. This
   * will necessarily be smaller than the deleted node, and all the nodes in
   * the right subtree.
   * (Notice that the largest value is just found by going right as far as
   * possible in the left subtree).
   */
  largest = left;
  while (largest->right) {
    largest = largest->right;
  }

  /* Splay the left tree again so that the largest node is brought to the top.
   */
  left->parent = NULL;
  tmp_left.root = left;
  splaySplay(&tmp_left, largest);

  /* Now, since the largest valued node in the left tree is at the top, we
   * know that its right pointer is NULL, and we can just attach the right
   * tree there.
   */
  if (n->parent == NULL) {
    t->root = largest;
    largest->parent = NULL;
  }
  else if (left_of_parent(n)) {
    n->parent->left = largest;
    largest->parent = n->parent;
    //largest->left_of_parent = 1;
  }
  else {
    n->parent->right = largest;
    largest->parent = n->parent;
    //largest->left_of_parent = 0;
  }
  largest->right = right;
  right->parent = largest;
  //right->left_of_parent = 0;
  
  SPLAYCHECKTREE(t);
}


/* SPLAY is the fundamental operation on splay trees. It takes node `n'
 * (assumed to exist within the splay tree) and pulls it up to the root
 * of the tree. The node returned (for convenience) is the root node, which
 * will be `n' and `t->root'.
 */
SplayNode *splaySplay(SplayTree *t, SplayNode *n)
{                                          
  // since I don't have this bit, but the code below uses it to record
  // some state which can't be computed by looking at the pointers at
  // that instant, I record the state explicitly here
  int n_left_of_parent = n->parent? left_of_parent(n) : 0;

#if SPLAY_COUNT_OPERATIONS
  ++(*splay_operations);
#endif

  while (n != t->root) {
    SplayNode *parent = n->parent, *grandparent = parent->parent;

#if SPLAY_COUNT_OPERATIONS
    ++(*splay_steps);
#endif

    /* Determine which transformation to perform. Each of these transformations
     * will move the node nearer to the root, and has the side effect of
     * making the tree more balanced.
     */
    /* 1a,b. Parent is the root, so perform `zig' rotation. */
    if (grandparent == NULL) {
      if (left_of_parent(n)) {
        SplayNode *b = n->right;

        n->right = parent;
        n->parent = parent->parent;                   

        // parent is the root!  these don't make sense
        //set_left_of_parent(n, left_of_parent(parent));
        //n_left_of_parent = left_of_parent(parent);

        parent->left = b;
        parent->parent = n;
        set_left_of_parent(parent, 0);

        if (b) {
          b->parent = parent;
          set_left_of_parent(b, 1);
        }
      }
      else {
        SplayNode *b = n->left;

        n->left = parent;
        n->parent = parent->parent;

        // parent is the root!  these don't make sense
        //set_left_of_parent(n, left_of_parent(parent));
        //n_left_of_parent = left_of_parent(parent);

        parent->right = b;
        parent->parent = n; 
        set_left_of_parent(parent, 1);

        if (b) {
          b->parent = parent;
          set_left_of_parent(b, 0);
        }
      }
    }

    /* 2a,b. `n', its parent and grandparent are in a straight line, so perform
     * the `zig-zig' rotation. */
    else if (left_of_parent(n) == left_of_parent(parent)) {
      if (left_of_parent(n)) {
        SplayNode *b = n->right;

        n->right = grandparent;
        n->parent = grandparent->parent;
        if (grandparent->parent) {
          set_left_of_parent(n, left_of_parent(grandparent));
          n_left_of_parent = left_of_parent(grandparent);
        }

        parent->left = b;
        grandparent->parent = n;
        set_left_of_parent(grandparent, 0);

        if (b) {
          b->parent = parent;
          set_left_of_parent(b, 1);
        }
      }
      else {
        SplayNode *b = n->left;

        n->left = grandparent;
        n->parent = grandparent->parent;
        if (grandparent->parent) {
          set_left_of_parent(n, left_of_parent(grandparent));
          n_left_of_parent = left_of_parent(grandparent);
        }

        parent->right = b;
        grandparent->parent = n; 
        set_left_of_parent(grandparent, 1);

        if (b) {
          b->parent = parent;
          set_left_of_parent(b, 0);
        }
      }
    }

    /* 3a,b. `n', its parent and grandparent aren't in a straight line, so
     * perform the `zig-zag' rotation. */
    else {
      if (left_of_parent(n)) {
        SplayNode *b = n->right, *c = n->left;

        n->right = parent;
        n->left = grandparent;
        n->parent = grandparent->parent;
        if (grandparent->parent) {
          set_left_of_parent(n, left_of_parent(grandparent));
          n_left_of_parent = left_of_parent(grandparent);
        }

        parent->left = b;
        parent->parent = n;
        set_left_of_parent(parent, 0);

        grandparent->right = c;
        grandparent->parent = n;
        set_left_of_parent(grandparent, 1);

        if (b) {
          b->parent = parent;
          set_left_of_parent(b, 1);
        }
        if (c) {
          c->parent = grandparent;
          set_left_of_parent(c, 0);
        }
      }
      else {
        SplayNode *b = n->left, *c = n->right;

        n->left = parent;
        n->right = grandparent;
        n->parent = grandparent->parent;
        if (grandparent->parent) {
          set_left_of_parent(n, left_of_parent(grandparent));
          n_left_of_parent = left_of_parent(grandparent);
        }

        parent->right = b;
        parent->parent = n;
        set_left_of_parent(parent, 1);

        grandparent->left = c;
        grandparent->parent = n;
        set_left_of_parent(grandparent, 0);

        if (b) {
          b->parent = parent;
          set_left_of_parent(b, 0);
        }
        if (c) {
          c->parent = grandparent;
          set_left_of_parent(c, 1);
        }
      }
    }

    // the 2nd line is where this n_left_of_parent state
    // comes in; computing it from the pointer structure
    // would be incorrect
    if (n->parent == NULL)      t->root = n;
    else if (n_left_of_parent)  n->parent->left = n;
    else                        n->parent->right = n;
  } /* while */

  SPLAYCHECKTREE(t);

  return n;
} /* tree_splay */



/* Return leftmost object below the current object node.
 */
static __inline SplayNode *leftmost(SplayNode *n)
{
  while (n->left) n = n->left;
  return n;
}


/* Look-up a node in the splay tree. Notice that the act of looking up a
 * node changes the tree. (In particular, the node, if found, moves to
 * the root of the tree). Returns the node, or NULL if not found.
 */
SplayNode *splayLookup(SplayTree *t, void *ptr)
{
  SplayNode *n = t->root;

#if SPLAY_COUNT_STAT
  int hits_steps;
#endif
#if SPLAY_COUNT_OPERATIONS
  int steps = 0;
  look_up_operations ++;
  if (look_up_operations == DUMP_EVERY)
    tree_dump_statistics (t);
  splay_operations = &splay_look_up_operations;
  splay_steps = &splay_look_up_steps;
#endif

/* Handle the toplevel of the splay tree with this inline code. It it
 * faster than using the while loop and the tree_splay function.
 * If the FAST_UPDATE is set to 0 the tree is not balanced anymore. This
 * is faster than keeping the tree balanced.
 */
#ifdef FAST
#undef FAST
#endif
#define FAST 1
#define FAST_UPDATE 0
#if FAST
  if (n) {
    if (ptr < base(n)) {
      n = n->left;
      if (n) {
        if (ptr < base(n))
          n = n->left;
        else if (ptr > extent(n))
          n = n->right;
        else if (ptr != extent(n)) {
#if FAST_UPDATE
          object *parent = n->parent;
          object *b = n->right;

          n->right = parent;
          n->parent = NULL;
          parent->left = b;
          parent->parent = n; set_left_of_parent(parent, 0);
          if (b) b->parent = parent, set_left_of_parent(b, 1);
          t->root = n;
#endif
#if SPLAY_COUNT_STAT
          add_hits(2);
          n->hits ++;
          n->steps += 2;
#endif
#if SPLAY_COUNT_OPERATIONS
          n->hits ++;
          look_up_steps += 2;
#endif
          SPLAYCHECKTREE(t);
          return n;
        }
      }
    }
    else if (ptr > extent(n)) {
      n = n->right;
      if (n) {
        if (ptr < base(n))
          n = n->left;
        else if (ptr > extent(n))
          n = n->right;
        else if (ptr != extent(n)) {
#if FAST_UPDATE
          object *parent = n->parent;
          object *b = n->left;

          n->left = parent;
          n->parent = NULL;
          parent->right = b;
          parent->parent = n; set_left_of_parent(parent, 1);
          if (b) b->parent = parent, set_left_of_parent(b, 0);
          t->root = n;
#endif
#if SPLAY_COUNT_STAT
          add_hits(2);
          n->hits ++;
          n->steps += 2;
#endif
#if SPLAY_COUNT_OPERATIONS
          n->hits ++;
          look_up_steps += 2;
#endif
          SPLAYCHECKTREE(t);
          return n;
        }
      }
    }
    else if (ptr != extent(n)) {
#if SPLAY_COUNT_STAT
      add_hits(1);
      n->hits ++;
      n->steps += 1;
#endif
#if SPLAY_COUNT_OPERATIONS
      n->hits ++;
      look_up_steps += 1;
#endif
      SPLAYCHECKTREE(t);
      return n;
    }
  }
#endif
#if SPLAY_COUNT_STAT
  hits_steps = 2;
#endif

  // now use the general search code
  n = splayNondestructiveLookup(n, ptr);

  if (n)
    {
#if SPLAY_COUNT_STAT
      add_hits(hits_steps);
      n->hits ++;
      n->steps += hits_steps;
#endif
#if SPLAY_COUNT_OPERATIONS
      n->hits ++;
      look_up_steps += steps;
#endif
      return splaySplay(t, n);
    }
  else
    {
#if SPLAY_COUNT_STAT
      add_no_hits(hits_steps);
#endif
#if SPLAY_COUNT_OPERATIONS
      nlook_up_steps += steps;
      nlook_up_operations ++;
#endif
      SPLAYCHECKTREE(t);
      return NULL;
    }
}


// here, 'n' is the root from which the search begins; the node
// found is *not* splayed to the top of the tree
SplayNode *splayNondestructiveLookup(SplayNode *n, void *ptr)
{
  while (n) {
    // moving this into its own function breaks these stats..
#if SPLAY_COUNT_STAT
    hits_steps ++;
#endif
#if SPLAY_COUNT_OPERATIONS
    steps ++;
#endif

    if (ptr < base(n)) n = n->left;
    else
      {
      /* This is tricky. Normally, a pointer refers to an object if the
       * pointer lies between base and extent inclusively. We add extra
       * padding between checked objects to ensure this is true. However,
       * if we aren't able to add this padding, then the next object may
       * lie immediately after the previous object in memory, ie. extent
       * and the base of the next object may be the same.
       */
        if (ptr > extent(n))
          n = n->right;
        else if (ptr == extent(n))
          {
            /* This is the tricky boundary condition. Use the hints given. If
             * `no_padding' is 0, there can be no immediate object on the
             * right, so stop the search. If `no_padding' is true, we must
             * look rightwards for the adjacent object, and go that way if the
             * right base == this extent.
             */
            if (no_padding(n)
                && n->right
                && base(leftmost(n->right)) == extent(n))
              n = n->right;
            else {
              /* if there could be a object next to this one. Do not use the
               * current object
               */
              if (no_padding(n)) n = NULL;
              break;
            }
          }
        else break;
      }
  }

  return n;
}


// check that a claimed splay node really is one
bool splayIsValidNode(SplayNode *n)
{
  // this is a fairly quick check, though it segfaults if the
  // node is not valid
  #if 1
    if (n->parent &&
        n->parent->left != n &&
        n->parent->right != n) {
      return false;
    }

    if (n->left &&
        n->left->parent != n) {
      return false;
    }

    if (n->right &&
        n->right->parent != n) {
      return false;
    }
  #endif

  // full-blown check
  return splayNondestructiveLookup(singleTree.root, base(n)) != NULL;
}


// ------------------- iterator/remover ----------------------
void splayWalkTree(SplayWalkFn func)
{
  if (singleTree.root) {
    innerSplayWalkTree(func, singleTree.root);
  }
}


// walk from node, postorder; return NULL if the subtree is
// to be deleted; if not, return the rightmost node in the
// subtree
SplayNode *innerSplayWalkTree(SplayWalkFn func, SplayNode *n)
{                              
  // rightmost nodes from resp. subtrees
  SplayNode *leftRight = NULL;
  SplayNode *rightRight = NULL;

  SplayNode *retval;
  SplayNode *subRoot = n;        // root of current subtree
  enum SplayWalkOpts action;

  // recursively process children
  if (n->left) {
    leftRight = innerSplayWalkTree(func, n->left);
    if (!leftRight) {
      n->left = NULL;
    }
    else {
      assert(leftRight >= n->left);
    }
  }

  if (n->right) {
    rightRight = innerSplayWalkTree(func, n->right);
    if (!rightRight) {
      n->right = NULL;
    }
    else {
      assert(rightRight >= n->right);
    }
  }

  // perform the action on this node
  action = func(n);

  if (!( action & SW_FREE )) {
    // easy case, not deleting
    retval = rightRight? rightRight : n;
  }
  else {
    // deleting this node; look at subtrees
    if (!leftRight && !rightRight) {
      // easy: no children
      splaySubstitute(NULL, n);
      subRoot = NULL;
      retval = NULL;
    }
    else if (leftRight && !rightRight) {
      // easy: let the left subtree replace this one
      splaySubstitute(n->left, n);
      subRoot = n->left;
      retval = leftRight;
    }
    else if (!leftRight && rightRight) {
      // easy: let the right subtree replace this one
      // (symmetric with above case)
      splaySubstitute(n->right, n);
      subRoot = n->right;
      retval = rightRight;
    }
    else {
      // hard: I am going away, but both my children remain

      // first, pull the left subtree's rightmost node out of
      // that tree
      assert(leftRight->right == NULL);     // must be rightmost
      splaySubstitute(leftRight->left, leftRight);

      // now install leftRight into the position currently
      // occupied by 'n', the current node
      splaySubstitute(leftRight, n);

      // and put n's children under it (splaySubstitute doesn't
      // work if n->left==NULL || n->right==NULL so I just do
      // the ptr manips directly)
      leftRight->left = n->left;
      if (n->left) {
        n->left->parent = leftRight;
      }

      leftRight->right = n->right;
      if (n->right) {
        n->right->parent = leftRight;
      }

      subRoot = leftRight;
      retval = rightRight;
    }

    // if we were the tree's root, use subRoot instead
    if (singleTree.root == n) {
      singleTree.root = subRoot;
    }

    // do the node free
    free(n);
  }

  if (action & SW_STOP) {
    // can't stop because I don't have a return value to indicate
    // that .. I could just setjmp/longjmp, but I'll avoid any
    // implementation until I need it
    assert(!"not implemented");
  }

  // local integrity checks
  if (subRoot) {
    assert(retval >= subRoot);       // should be rightmost
    if (subRoot->left) {
      assert(subRoot->left < subRoot);
      assert(subRoot->left->parent == subRoot);
    }
    if (subRoot->right) {
      assert(subRoot->right > subRoot);
      assert(subRoot->right->parent == subRoot);
      assert(retval >= subRoot->right);
    }
    assert(size(retval) > 0);
    assert(size(subRoot) > 0);
  }

  SPLAYCHECKTREE(&singleTree);

  return retval;
}


// replace the subtree rooted at 'o' with the subtree
// rooted at 'n'
void splaySubstitute(SplayNode *n, SplayNode *o)
{
  // fix o's parent's link
  if (o->parent) {
    if (left_of_parent(o)) {
      o->parent->left = n;
    }
    else {
      o->parent->right = n;
    }
  }

  // fix n's link
  if (n) {
    n->parent = o->parent;
  }
}


/*----------------------------------------------------------------------
 *      Perform a strict check on the contents of the splay tree. This
 *      check is done after each tree operation if 'SPLAY_STRICT_CHECKING'
 *      is defined as 1. However, we include this code anyway so that
 *      we can call it from the debugger or app. if necessary.
 *----------------------------------------------------------------------*/

void splayCheckTree(SplayTree *t)
{
  SplayNode *root = t->root;

  if (root == NULL) return;
  
  if (root->parent != NULL)
    splayTreeInternalError(t, root, "root node has a parent (should be NULL)");

  splayInternalTest(t, root, NULL /*parent*/, 0 /*left_of_par*/,
                    (void*)0 /*min*/, (void*)~0 /*max*/);

  #if 0
  if (root->left)
    splayInternalTest(t, root->left, root, 1, (void *) 0,
                      (void*)((char*)base(root) - 1));
  if (root->right)
    splayInternalTest(t, root->right, root, 0, extent(root), (void *) ~0);
  #endif // 0
}

void splayInternalTest(SplayTree *t, SplayNode *n, SplayNode *parent,
                       int left_of_par, void *min, void *max)
{
  if (n->parent != parent)
    splayTreeInternalError(t, n, "node has parent different from true parent");
  if (parent && (left_of_parent(n) != left_of_par))
    splayTreeInternalError(t, n, "node has left_of_parent different from true direction");
  if (base(n) >= extent(n))
    splayTreeInternalError(t, n, "node has zero or negative size");
  if (base(n) < min || (void*)((char*)extent(n) - 1) > max)
    splayTreeInternalError(t, n, "node is out of order within the tree");
  if (n->left)
    splayInternalTest(t, n->left, n, 1, min, (void*)((char*)base(n) - 1));
  if (n->right)
    splayInternalTest(t, n->right, n, 0, extent(n), max);

  splayCheckTree(&n->subintervals);
}

void splayTreeInternalError(SplayTree *t, SplayNode *n, char *msg)
{
  printf ("tree_check: %s\n"
          "  tree root at %p\n"
          "  object struct at %p\n"
          "  object refers to %p..%p\n",
          msg, t->root, n, base(n), (void*)((char*)extent(n)-1));
  abort();
}


// -------------- sm: stuff I've added -------------------
// print the contents of the splay tree
void splayPrintTree(SplayTree *t)
{
  printf("splay tree {\n");
  if (t->root) {
    splayPrintNode(t->root, NULL, 2);
  }
  printf("}\n");
}

void printIndent(int indent)
{
  int i;
  for (i=0; i<indent; i++) {
    printf(" ");
  }
}

void splayPrintNode(SplayNode *n, SplayNode *expectParent, int indent)
{
  char *alarm;

  printIndent(indent);

  alarm = (expectParent == n->parent? "" : "***");
  printf("node=%p, parent=%s%p%s, base=%p, size=%d",
         n, alarm, n->parent, alarm, base(n), size(n));

  if (n->subintervals.root) {
    printf("; subs {\n");
    splayPrintNode(n->subintervals.root, NULL, indent+2);
    printIndent(indent);
    printf("}\n");
  }
  else {
    printf("\n");
  }

  if (n->left) {
    splayPrintNode(n->left, n, indent+2);
  }
  if (n->right) {
    splayPrintNode(n->right, n, indent+2);
  }
}


// --------------- non-malloc interval stuff ----------------
// given a pointer, classify it as to which region it points
enum MemRegion whichRegion(void *ptr)
{
  unsigned uptr = (unsigned)ptr;

  #ifndef linux
    // we can either hack this so it works under Windows, or else
    // get the garbage collector compiled under Win, in which case
    // we can just call its routines that do similar things
    return MR_STACK;
  #endif

  // linux's stack starts at 0xC0000000 and grows downward; so any
  // valid pointer will have the first hex digit as a B (unless of
  // course the stack grows bigger than 256M..)
  if ((uptr & 0xF0000000) == 0xB0000000) {
    return MR_STACK;
  }

  // linux's libc's first global data symbol is '__data_start', and
  // the linker creates a symbol called _end at the end of the data
  // segment
  {
    #pragma weak __data_start
    extern int __data_start;
    extern int _end;

    if ( (unsigned)(&__data_start) <= uptr &&
         uptr <= (unsigned)(&_end)) {
      return MR_GLOBAL;
    }
  }

  // classify everything else as the heap
  return MR_HEAP;
}


void *splayFindHomeEnd(int kind, void *ptr, void **end)
{
  enum MemRegion mr = whichRegion(ptr);
  SplayNode *n = splayLookup(&__global_tree[mr][kind], ptr);
  void *ret;

  if (!n) {
    // the assumption is this pointer is actually being used to store
    // an integer; were we using the two-word representations, the
    // 'base' field would be NULL; so I return NULL to model that
    ret = NULL;
  }
  else {
    // we're at least within 'n' -- but are we in a subinterval?
    for (;;) {
      SplayNode *sub = splayLookup(&n->subintervals, ptr);
      if (sub) {
        n = sub;    // yes, this subinterval contains 'ptr'
      }
      else {
        break;      // either no subintervals, or none contain 'ptr'
      }
    }

    ret = base(n);
  }

  if (n && end) {
    *end = (void*)((char*)base(n) + size(n));
  }

  if (end) {
    DP(printf("splayFindHomeEnd[%d](%d, %p) returning (%p, %p)\n",
              mr, kind, ptr, ret, *end));
  }
  else {
    DP(printf("splayFindHome[%d](%d, %p) returning %p\n", mr, kind, ptr, ret));
  }

  return ret;
}


void *splayFindHome(int kind, void *ptr)
{
  return splayFindHomeEnd(kind, ptr, NULL /*end*/);
}


void splayRemoveKindBelow(enum MemRegion mr, int kind, void *cutoffAddr)
{
  DP(printf("splayRemoveKindBelow[%d](%d, %p)\n", mr, kind, cutoffAddr));

  splayRemoveBelow(&__global_tree[mr][kind], cutoffAddr);
}


// remove all nodes whose 'base' field is less-or-equal 'cutoffAddr'
void splayRemoveBelow(SplayTree *tree, void *cutoffAddr)
{
  // step 1: find the node whose base field is closest to
  // 'cutoffAddr', but bigger than it
  SplayNode *above = splayFindClosestAbove(tree, cutoffAddr);
  if (!above) {
    // everything is below the cutoff
    splayDeallocSubtree(tree->root);
    tree->root = NULL;
    return;
  }

  // step 2: splay that node
  splaySplay(tree, above);

  // step 3: remove the left subtree, since now all nodes
  // less-or-equal cutoffAddr are in that subtree
  splayDeallocSubtree(above->left);
  above->left = NULL;
}


// find the node whose base field is closest to 'cutoffAddr', but
// bigger than it
SplayNode *splayFindClosestAbove(SplayTree *tree, void *cutoffAddr)
{
  SplayNode *above = NULL, *below = NULL;

  //printf("tree = %p\n", tree);

  if (tree->root == NULL) {
    return NULL;
  }

  if (base(tree->root) > cutoffAddr) {
    above = tree->root;
  }
  else {
    below = tree->root;
    goto secondHalf;
  }

  for (;;) {
    while (base(above) > cutoffAddr) {
      if (!above->left) {
        // 'above' is the node we seek
        return above;
      }
      above = above->left;
    }

    // just walked below my target
    below = above;
    above = above->parent;

  secondHalf:
    while (base(below) <= cutoffAddr) {
      below = below->right;
      if (!below) {
        // can't get back above cutoff; so 'above' is the
        // value we seek
        return above;
      }
    }

    // just walked above my target
    above = below;
    below = below->parent;
  }
}


// recursively deallocate 'node' and all its children
void splayDeallocSubtree(SplayNode *node)
{
  if (node) {
    splayDeallocSubtree(node->left);
    splayDeallocSubtree(node->right);
    splayDeallocSubtree(node->subintervals.root);
    free(node);
  }
}


// count the # of intervals in a subtree
int splaySubtreeCountIntervals(SplayNode *node)
{
  if (node) {
    return 1 + splaySubtreeCountIntervals(node->left)
             + splaySubtreeCountIntervals(node->right)
             + splaySubtreeCountIntervals(node->subintervals.root);
  }
  else {
    return 0;
  }
}

int splayCountIntervals(enum MemRegion mr, int kind)
{
  return splaySubtreeCountIntervals(__global_tree[mr][kind].root);
}


void splayInsertInterval(int kind, void *base, unsigned length)
{
  SplayNode *node;
  enum MemRegion mr = whichRegion(base);

  DP(printf("splayInsertInterval[%d](%d, %p, %d)\n", mr, kind, base, length));

  node = (SplayNode*)malloc(sizeof(SplayNode));
  node->base = base;
  node->length = length;
  node->subintervals.root = NULL;
  splayInsert(&__global_tree[mr][kind], node);

  //DP(splayPrintTreeKind(mr, kind));
}


void splayPrintTreeKind(enum MemRegion mr, int kind)
{
  splayPrintTree(&__global_tree[mr][kind]);
}


// -------------- malloc/free code that uses the splay tree ---------------
void *splayMalloc(size_t size)
{
  void *raw;
  SplayNode *n;

  // currently the code breaks if size is 0 .. I could bump
  // it up, but for now let's just record this requirement
  assert(size > 0);

  // add the amount we're going to take first
  size += sizeof(SplayNode);

  // get a block from malloc
  raw = malloc(size);

  // add it to the tree
  n = (SplayNode*)raw;
  splayInsert(&singleTree, n);

  DP(printf("splayMalloc(%d) returning %p\n", 
            size-sizeof(SplayNode), base(n)));

  // return the block to the user
  return base(n);
}


void splayFree(void *block)
{
  SplayNode *n;

  DP(printf("splayFree(%p)\n", block));

  // map to splay block
  n = baseToNode(block);
  assert(splayIsValidNode(n));

  // remove it from the tree
  splayDelete(&singleTree, n);

  // deallocate the storage
  free(n);
}

void splayRemoveBlock(void *block)
{
  // map to splay block
  SplayNode *n = baseToNode(block);
  assert(splayIsValidNode(n));

  // remove it from the tree
  splayDelete(&singleTree, n);
}


void *splayRealloc(void *oldBlock, size_t newSize)
{ 
  SplayNode *o, *n;     // old,new
  bool origLeft;        // true if we were originally a left child

  // map to splay block
  o = baseToNode(oldBlock);
  assert(splayIsValidNode(o));

  origLeft = (o->parent != NULL) && left_of_parent(o);

  // do the user-prescribed realloc; if the block moves, this
  // will leave the tree temporarily malformed
  n = (SplayNode*)realloc(o, newSize + sizeof(SplayNode));

  if (n != o) {
    // repair the tree structure itself (the index structure is still
    // wrong because it's not in order anymore); we rely on the checks
    // (during debugging) inside 'splayDelete' and 'splayInsert' to
    // verify the pointer manipulations are correct
    if (n->parent) {
      if (origLeft) {
        assert(n->parent->left = o);
        n->parent->left = n;
      }
      else {
        assert(n->parent->right = o);
        n->parent->right = n;
      }
    }

    if (n->left) {
      assert(n->left->parent = o);
      n->left->parent = n;
    }

    if (n->right) {
      assert(n->right->parent = o);
      n->right->parent = n;
    }

    // my hope, and assumption, is that 'splayDelete' won't mind that
    // the node we're deleting is out of order (my reading of the code
    // indicates that's true)

    // remove it from the tree
    splayDelete(&singleTree, n);

    // insert it back into the tree, into the proper place
    splayInsert(&singleTree, n);
  }

  // return the user's data
  return base(n);
}


void *splayMallocZeroed(size_t osize)
{
  void *raw;
  SplayNode *n;
  size_t size;

  // add the amount we're going to take
  size = osize + sizeof(SplayNode);

  // get a block from calloc, which may have some clever
  // internal routines to zero blocks
  raw = calloc(size, 1);

  // add it to the tree
  n = (SplayNode*)raw;
  splayInsert(&singleTree, n);

  DP(printf("splayMallocZeroed(%d) returning %p\n",
            osize, base(n)));

  // return the block to the user
  return base(n);
}


void *findSplayBlock(void *interior)
{
  SplayNode *n = splayLookup(&singleTree, interior);
  if (!n) {
    return NULL;
  }
  else {
    return base(n);
  }
}


void validateBlock(void *block)
{
  assert(splayIsValidNode(baseToNode(block)));
}


size_t getSplayBlockSize(void *block)
{
  // map to splay block
  SplayNode *n = baseToNode(block);
  assert(splayIsValidNode(n));

  return size(n);
}


void printSplayTree()
{
  splayPrintTree(&singleTree);
}


void getSplayExcludedRoots(void **start, void **end)
{
  *start = &singleTree;
  *end = &singleTree + 1;
}


// --------------------- old test code -----------------------
#ifdef OLD_TEST_SPLAY

#define TRACERET(ptr, size)                                    \
  ptr = splayMalloc(size);                                     \
  printf("size %d, block %s: %p\n", size, #ptr, ptr);          \
  printf("  findBlock(+2): %p\n", findSplayBlock(ptr+2));      \
  printf("  findBlock(+100): %p\n", findSplayBlock(ptr+100));  \
  printSplayTree();

#define FREE(ptr)                                              \
  printf("free(%p)...\n", ptr);                                \
  splayFree(ptr);                                              \
  printf("findblock(%p): %p\n", p[5], findSplayBlock(p[5]));   \
  printSplayTree();


void test1()
{
  void *p[6];
  int i;

  memset(p, 0, 6*sizeof(void*));

  printf("--------- test1 ---------\n");

  printSplayTree();
  TRACERET(p[0], 5);
  TRACERET(p[1], 70);
  TRACERET(p[2], 1000);
  TRACERET(p[3], 10000);
  TRACERET(p[4], 100000);
  TRACERET(p[5], 1000000);

  for (i=0; i<12; i++) {
    printf("findblock(p[%d]=%p): ",
           i%6, p[i%6]);
    printf("%p\n", findSplayBlock(p[i%6]));
    printSplayTree();
  }

  FREE(p[0]);
  FREE(p[4]);
  FREE(p[2]);
  FREE(p[5]);
  FREE(p[1]);
  FREE(p[3]);
}


int randSize()
{
  int ret = random() % 100 +
            ((random() % 10)==0) * 400 +
            ((random() % 15)==0) * 1000 +
            ((random() % 20)==0) * 10000;
  if (ret == 0) {
    ret = 1;
  }
  return ret;
}


typedef struct Pair {
  void *mem;
  int size;
} Pair;

void test2()
{
  enum { PAIRS=100, ITERS=1000 };
  Pair p[PAIRS];
  int i;

  printf("---------- test 2 ---------\n");

  memset(p, 0, sizeof(p));

  for (i=0; i<ITERS; i++) {
    int index = random() % PAIRS;

    if (p[index].mem == NULL) {
      // allocate
      p[index].size = randSize();
      p[index].mem = splayMalloc(p[index].size);
    }

    else {
      // allocated block
      switch (random()%4) {
        default: {
          // probe
          void *spot = p[index].mem + random() % (p[index].size);
          void *m = findSplayBlock(spot);
          assert(m == p[index].mem);
          break;
        }

        case 0: {
          // deallocate
          void *spot = p[index].mem + random() % (p[index].size);
          void *m;
          splayFree(p[index].mem);
          m = findSplayBlock(spot);
          assert(m == NULL);

          // nullify my data
          p[index].mem = NULL;
          p[index].size = 0;
          break;
        }

        #if 1
        case 1: {
          // reallocate
          int newSize = randSize();
          void *m = p[index].mem;
          p[index].mem = splayRealloc(m, newSize);
          //printf("realloc(%p, %d) = %p\n",
          //       m, newSize, p[index].mem);
          p[index].size = newSize;
          break;
        }
        #endif // 0/1
      }
    }
  }

  for (i=0; i<PAIRS; i++) {
    if (p[i].mem != NULL) {
      splayFree(p[i].mem);
    }
  }

  // should be empty
  printSplayTree();
}


int test3Blocks = 0;
  
enum SplayWalkOpts test3Walker(void *block)
{
  if (random()%3) {
    test3Blocks--;
    return SW_FREE | SW_GO;
  }
  else {
    return SW_GO;
  }
}

// tests the splay walker interface
void test3()
{
  enum { ITERS=100, ELTS=100 };
  int i,j;

  printf("---------- test3 ----------\n");

  for (i=0; i<ITERS; i++) {
    // put some stuff into the tree
    for (j=0; j<ELTS; j++) {
      splayMalloc(randSize());
      test3Blocks++;
    }

    // deallocate them via the walker interface, somewhat
    // haphazardly
    while (test3Blocks > 0) {
      splayWalkTree(test3Walker);
    }
  }
}


int main()
{
  test1();
  test2();
  test3();
  printf("\nall tests ok!\n");
  return 0;
}

#endif // OLD_TEST_SPLAY


// -------------------- current test code --------------------
#ifdef TEST_SPLAY
  
int initedGlobal = 4;
int uninitedGlobal;

void testWhichRegion()
{
  int local = 9;
  
  printf("testWhichRegion\n");

  assert(whichRegion(&local) == MR_STACK);

  assert(whichRegion(&initedGlobal) == MR_GLOBAL);
  assert(whichRegion(&uninitedGlobal) == MR_GLOBAL);

  {
    void *p = malloc(12);
    assert(whichRegion(p) == MR_HEAP);
    free(p);
  }
}


// elementary tests
int count()
{
  return splayCountIntervals(MR_HEAP, 0);
}

void insert(int base, int len)
{
  splayInsertInterval(0, (void*)base, len);
}

int find(int ptr)
{
  return (int)splayFindHome(0, (void*)ptr);
}

void print()
{
  splayPrintTreeKind(MR_HEAP, 0);
}

void removeBelow(int base)
{
  splayRemoveKindBelow(MR_HEAP, 0, (void*)base);
}


void test1()
{
  printf("--------- test1 --------\n");

  // insert some intervals
  assert(count() == 0);
  insert(100, 20);
  assert(count() == 1);
  insert(200, 20);
  assert(count() == 2);
  insert(150, 20);
  assert(count() == 3);

  print();

  // delete whole tree
  removeBelow(900);
  assert(count() == 0);

  print();
}


// lots of disjoint intervals, arranged as they're likely to
// be under the safec workload
struct Pair {
  char *base;
  unsigned len;
};

int top = 0;
struct Pair stack[100];

void pushInterval(char *base, int len)
{
  assert(top < 99);
  top++;
  stack[top].len = len;
  stack[top].base = base;
  splayInsertInterval(0, stack[top].base, stack[top].len);
}

void test2()
{
  int i;

  printf("--------- test2 --------\n");

  top = -1;     // so first push sets entry 0
  pushInterval((char*)0x100000, 234);
  assert(splayCountIntervals(MR_HEAP,0) == 1);

  for (i=0; i<1000; i++) {
    // select entry to modify or query
    int r = 1 + (rand() % 99);
    if (r <= top) {
      // existing entry
      char *base = stack[r].base;
      int len = stack[r].len;
      switch (rand()%10) {
        default: {
          // query
          char *ptr = base + rand()%len;
          void *end;
          char *b = splayFindHomeEnd(0, ptr, &end);
          assert(b == base);
          assert(end == base+len);
          break;
        }

        case 0: {
          // remove
          splayRemoveKindBelow(MR_HEAP,0, base);
          assert(splayCountIntervals(MR_HEAP,0) == r);
          top = r-1;
        }
      }
    }
    else {
      // new entry
      if (r > top + 5) {
        // add at most 5 at a time
        r = top + 5;
      }
      while (r > top) {
        int len = rand()%20 * 4 + 4;
        char *base = stack[top].base - (rand()%20) * 4 - len;
        pushInterval(base, len);
      }
      assert(splayCountIntervals(MR_HEAP,0) == r+1);
    }
    assert(splayCountIntervals(MR_HEAP,0) == top+1);
  }

  splayRemoveKindBelow(MR_HEAP,0, stack[0].base);
  assert(splayCountIntervals(MR_HEAP,0) == 0);
}


// elementary nested interval tests
void test3()
{
  printf("--------- test3 --------\n");
 
  // first some toplevel intervals
  insert(100, 100);
  insert(200, 100);
  insert(300, 100);
  assert(find(100) == 100);
  assert(find(199) == 100);
  assert(find(200) == 200);
  assert(find(299) == 200);
  assert(find(300) == 300);
  assert(find(399) == 300);
  assert(count() == 3);

  // nest a few in the 200 interval
  insert(210, 10);
  insert(220, 10);
  insert(230, 10);
  assert(find(200) == 200);
  assert(find(299) == 200);
  assert(find(210) == 210);
  assert(find(219) == 210);
  assert(find(220) == 220);
  assert(find(229) == 220);
  assert(find(230) == 230);
  assert(find(239) == 230);
  assert(find(250) == 200);
  assert(count() == 6);

  // and a few more in the 220
  insert(221, 1);
  insert(222, 1);
  insert(223, 1);
  assert(find(220) == 220);
  assert(find(221) == 221);
  assert(find(222) == 222);
  assert(find(223) == 223);
  assert(find(224) == 220);
  assert(count() == 9);

  print();

  // throw them all away
  removeBelow(400);
  assert(count() == 0);
}







#if 0
        case 1: {
          if (top < 99) {
            // add subinterval
            char *begin = base + rand()%(len-1);
            char *end = begin + rand() % (len - (begin-base) - 1) + 1;
            pushInterval(begin, end-begin);
            break;
          }
        }
#endif // 0



int main()
{
  testWhichRegion();
  test1();
  //test2();
  test3();

  printf("\nall tests ok!\n");
  return 0;
}

#endif // TEST_SPLAY
