(*
 *
 * Copyright (c) 2001 by
 *  George C. Necula	necula@cs.berkeley.edu
 *  Scott McPeak        smcpeak@cs.berkeley.edu
 *  Wes Weimer          weimer@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1. XSRedistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products 
 * derived from  this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

open Cil
open Pretty
module E = Errormsg
module H = Hashtbl


(******** STATISTICS **********)
(* Collect statistics about the code *)
let counters : (string, int ref) H.t = H.create 23
class collectCureStats = object (self)
  inherit nopCilVisitor




  method vinst (i: instr) : instr list visitAction = 
    let isCheck name = 
      let prefix p s = 
        let lp = String.length p in
        let ls = String.length s in
        lp <= ls && String.sub s 0 lp = p
      in
      prefix "CHECK_" name
    in
    let incrCount (name: string) = 
      try
        let vr = H.find counters name in
        incr vr
      with Not_found -> begin
        let vr = ref 1 in
        H.add counters name vr
      end
    in
    match i with 
      Call(_, Lval(Var fvi, NoOffset), _, _) when isCheck fvi.vname -> 
        incrCount fvi.vname;
        SkipChildren
    | _ -> SkipChildren



end

(* weimer: visitor routines to check and see if major types have
 * changed *)
let typeSizeHT = Hashtbl.create 555
class typeSizeSetVisitor = object
  inherit nopCilVisitor 
  method vglob g = 
    match g with
      GType (t, _) when t.tname <> "" -> begin
        let bs = try bitsSizeOf t.ttype / 8 with _ -> 0 in
        Hashtbl.add typeSizeHT t.tname bs;
        SkipChildren
      end
    | _ -> SkipChildren
end

class typeSizeCheckVisitor = object
  inherit nopCilVisitor 
  method vglob g =
    match g with 
      GType (t, l) when t.tname <> "" 
                       && Hashtbl.mem typeSizeHT t.tname -> begin
        let new_bs = try bitsSizeOf t.ttype / 8 with _ -> 0 in
        let old_bs = Hashtbl.find typeSizeHT t.tname in
        if (new_bs <> old_bs) then begin
          let used_in_interface = ref None in
          (* Wes, there are better ways to do this !!! *)
          Hashtbl.iter (fun k n -> 
            if n.Ptrnode.btype = t.ttype && 
               Ptrnode.hasFlag n Ptrnode.pkInterface then
              used_in_interface := Some(n.Ptrnode.where)
                   ) Ptrnode.idNode ; 
          match !used_in_interface with
            None -> 
              if !E.verboseFlag then 
                ignore (printf "**** Sizeof %s changed from %d to %d\n" 
                          t.tname old_bs new_bs)
          | Some(p,i) -> 
              ignore (printf "****!!! Sizeof %s changed from %d to %d\n" 
                        t.tname old_bs new_bs);
              ignore (printf "  and is used in interface function %a\n" 
                        Ptrnode.d_place p)
        end;
        SkipChildren
      end
    | _ -> SkipChildren
end

let print (f: file) = 
  H.clear counters;
  ignore (visitCilFile (new collectCureStats) f);
  let allChecks = ref 0 in
  let l = 
    H.fold (fun name counter acc -> 
      allChecks := !allChecks + !counter;
      (name, !counter) :: acc) counters [] in
  (* Now sort for most frequent first *)
  let l = List.sort (fun (_, c1) (_, c2) -> (compare c2 c1)) l in
  H.clear counters;
  if !allChecks = 0 then 
    ignore (E.log "No checks were inserted!\n")
  else begin
    ignore (E.log "Static count of CHECKs inserted:\n");
    List.iter 
      (fun (n, c) -> ignore (E.log " %-30s %6d (%6.2f%%)\n"
                               n c (100.0 *. (float c) /. (float !allChecks))))
      l
  end
