(*
 *
 * Copyright (c) 2001 by
 *  George C. Necula	necula@cs.berkeley.edu
 *  Scott McPeak        smcpeak@cs.berkeley.edu
 *  Wes Weimer          weimer@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products 
 * derived from  this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

open Cil
open Pretty

module H = Hashtbl
module E = Errormsg

module N = Ptrnode

module MU = Markutil


let lu = locUnknown

let currentFile = ref dummyFile
let currentFunctionName = ref ""
let currentResultType = ref voidType

let boxing = ref true

let useGenerics = ref false
let genericId   = ref 0

let callId = ref (-1)  (* Each call site gets a new ID *)


let rec stripCasts (e: exp) = 
  match e with CastE(_, e') -> stripCasts e' | _ -> e

(* weimer: utility function to ease the transition between our flag formats *)
let setPosArith n = begin N.setFlag n N.pkPosArith end
let setReachSeq n = begin N.setFlag n N.pkReachSeq end
let setReachString n = begin N.setFlag n N.pkReachString end
let setArith n = begin N.setFlag n N.pkArith end
let setNull n = begin N.setFlag n N.pkNull  end
let setUpdated n = begin N.setFlag n N.pkUpdated end
let setIntCast n = begin N.setFlag n N.pkIntCast end
let setInterface n = begin N.setFlag n N.pkInterface end

let setOnStack n = begin N.setFlag n N.pkOnStack end
let setEscape n = begin N.setFlag n N.pkEscape end

(* We keep track of a number of type that we should not unroll *)
let dontUnrollTypes : (string, bool) H.t = H.create 19

let rec mustUnroll (ti: typeinfo) : bool = 
  (not (H.mem dontUnrollTypes ti.tname)) &&
  (match ti.ttype with 
  | TPtr _ -> true
  | TArray _ -> true
  | TNamed (ti', _) -> mustUnroll ti'
  | _ -> false)


(* if some enclosing context [like the attributes for a field] says that
 * this array should be sized ... we do not want to forget it! *)
let addArraySizedAttribute arrayType enclosingAttr =
  if filterAttributes "sized" enclosingAttr <> [] then
    typeAddAttributes [Attr("sized",[])] arrayType
  else
    if hasAttribute "safeunion" enclosingAttr then
      typeAddAttributes [Attr("safeunion",[])] arrayType
    else
      arrayType

(* Grab the node from the attributs of a type. Returns dummyNode if no such 
 * node *)
let nodeOfType t = 
  match unrollType t with 
    TPtr(_, a) -> begin
      match N.nodeOfAttrlist a with
        Some n -> n
      | None -> N.dummyNode
    end
  | _ -> N.dummyNode




(* Pass also the place and the next index within the place. Returns the 
 * modified type and the next ununsed index *)
let rec doType (t: typ) (p: N.place) 
               (nextidx: int) : typ * int = 
  match t with 
    (TVoid _ | TInt _ | TFloat _ | TEnum _ ) -> t, nextidx
  | TBuiltin_va_list _ -> E.s (E.bug "doType: va_list")
  | TPtr (bt, a) -> begin
      match N.nodeOfAttrlist a with
        Some n -> TPtr (bt, a), nextidx (* Already done *)
      | None -> 
          let bt', i' = doType bt p (nextidx + 1) in
          let n = N.getNode p nextidx bt' a in
          TPtr (bt', n.N.attr), i'
  end
  | TArray(bt, len, a) -> begin
      (* wes: we want a node for the array, just like we have a node for
       * each pointer *)
      let sized = filterAttributes "sized" a <> [] in 
      match N.nodeOfAttrlist a with
        Some n -> TArray (bt, len, a), nextidx (* Already done *)
      | None -> 
          let bt', i' = doType bt p (nextidx + 1) in
          let n = N.getNode p nextidx bt' a in
          TArray (bt', len, n.N.attr), i'
  end
          
  | TComp(c, at)-> 
      if Poly.isPolyComp c then begin
        (* We must make a copy of it. If we need to create new ones we must 
         * mark them and add them to the file *)
        let c' = Poly.instantiatePolyComp c 
            (fun c'' -> 
              markCompInfo c'';
              (* Add a global declaration for it *)
              MU.theFile := GCompTag (c'', !currentLoc) :: !MU.theFile) in
        TComp(c', at), nextidx
      end else
        t, nextidx (* A reference to a regular composite type, leave alone *)

    (* Strip the type names so that we have less sharing of nodes. However, 
     * we do not need to do it if the named type is a structure, and we get 
     * nicer looking programs. We also don't do it for base types *)
  | TNamed (bt, a) -> 
      if mustUnroll bt then begin
        let t', nextidx' = doType bt.ttype p nextidx in
        t', nextidx'
      end else
        (* A type reference. Leave alone. We'll handle the type inside when 
         * we do the GType *)
        t, nextidx
        
  | TFun (restyp, args, isva, a) -> 
      let noproto = hasAttribute "missingproto" a in
      let restyp = 
        if noproto && not (isPointerType restyp) then 
          voidPtrType
        else restyp 
      in
      let restyp', i0 = doType restyp p nextidx in
      let i' = 
        List.fold_left 
          (fun nidx arg -> 
            let t', i' = doType arg.vtype p nidx in
            arg.vtype <- t'; (* Can change in place because we shall copy the 
                              * varinfo for polymorphic functions *)
            i') i0 (argsToList args) 
      in
      let newtp = TFun(restyp', args, isva, a) in
      newtp, i'

and markCompInfo (comp: compinfo) = 
  (* Maybe we must turn this composite type into a struct *)
  if not comp.cstruct &&
    hasAttribute "safeunion" comp.cattr then
    comp.cstruct <- true;
  List.iter 
    (fun f -> 
      let fftype = addArraySizedAttribute f.ftype f.fattr in 
      let t', i' = doType fftype (N.PField f) 1 in
      f.ftype <- t') comp.cfields
  
(* For each node corresponding to a struct or union or array type we will 
 * create successor node corresponding to various offsets. In the 
 * particular case of an array, we use the field name "@field" to refer to 
 * the first element *)
(* Create a new offset node *)
let newOffsetNode (n: N.node)  (fname: string) 
                  (btype: typ) (battr: attribute list) 
                  (next_should_be : N.node option) = 
  let mkNext () = 
    match next_should_be with
      Some(node) -> node
    | None -> N.newNode (N.POffset(n.N.id, fname)) 0 btype battr 
  in
  (* ignore (E.log "newOffsetNode: n=%d, fname=%s\n" n.N.id fname); *)
  (* Add edges between n and next *)
  let next = 
    match unrollType n.N.btype with
      TComp (c, _) when not c.cstruct -> (* A union *)
        let next = mkNext () in
        let _ = N.addEdge n next N.ECast (-1) None in 
        let _ = N.addEdge n next N.ESafe (-1) None in 
        next
          
    | TArray (bt, l, a) -> (* An index *)
      (* Maybe the array already has a node *)
        let next = 
          match N.nodeOfAttrlist a with
            Some oldn -> 
              (* ignore (E.log "Reusing node %d\n" oldn.N.id); *)
              oldn
          | _ -> 
              (* ignore (E.log "Creating a node for array\n"); *)
              mkNext () (* Shouldn't there always be a node here ? *)
        in
        let _ = N.addEdge n next N.EIndex (-1) None in 
        next
          
    | TComp (c, _) when c.cstruct -> (* A struct *)
        let next = mkNext () in
        let _ = N.addEdge n next N.ESafe (-1) None in 
        next
          
    | _ -> E.s (bug "Unexpected offset")
  in
  next

(* Create a field successor *)
(* weimer : Tue Dec 11 23:47:00  2001
 * I need the node (which is associated with the address of the field)
 * to be stored in the fi.fattr attribute list so that I can get it
 * out later when I want to typecheck "&(struct.field)". Previously
 * the node was just created and returned. In addition, this function
 * could get called multiple times for the same structure field, creating
 * multiple distinct node to stand for the address of that field. As
 * I understand the CCured type system, it would be non-sensical for
 * them to have different kinds, so here we do not create a new node
 * if there is already one for this field. 
 *
 * weimer : Fri Feb  1 15:19:40 PST 2002
 * However, we should still call "newOffsetNode" anyway because it
 * adds some important edges. For example, if you have:
 *   typedef struct { Foo * NODE(3) f; } S;
 *   ...
 *   S * NODE(1) s1, * NODE(2) s2; 
 * We want ESAFE edges between 1-3 and 2-3. 
 * *)
let fieldOfNode (n: N.node) (fi: fieldinfo) : N.node =
  match N.nodeOfAttrlist fi.fattr with
    Some(old_n) -> 
      let _ = newOffsetNode n fi.fname fi.ftype [] (Some(old_n)) in
      old_n 
  | None -> let answer = newOffsetNode n fi.fname fi.ftype [] None in
    fi.fattr <- addAttributes fi.fattr answer.N.attr ; 
    answer

let startOfNode (n: N.node) : N.node =
  match unrollType n.N.btype with
    TArray (bt, _, a) -> newOffsetNode n "@first" bt a None
  | _ -> n (* It is a function *)
  



(* Compute the sign of an expression. Extend this to a real constant folding 
 * + the sign rule  *)
type sign = SPos | SNeg | SAny | SLiteral of int64

let rec signOf = function
    Const(CInt64(n, _, _)) -> SLiteral n
  | Const(CChr c) -> SLiteral (Int64.of_int (Char.code c))
  | SizeOf _ -> SPos (* We do not compute it now *)
  | UnOp (Neg, e, _) -> begin
      match signOf e with
        SPos -> SNeg
      | SLiteral n -> SLiteral (Int64.neg n)
      | SNeg -> SNeg
      | _ -> SAny
  end
  | UnOp (LNot, e, _) -> SPos
  | BinOp (PlusA, e1, e2, _) -> begin
      match signOf e1, signOf e2 with
        SPos, SPos -> SPos
      | SLiteral n, SPos when n >= Int64.zero -> SPos
      | SPos, SLiteral n when n >= Int64.zero -> SPos
      | SLiteral n1, SLiteral n2 -> SLiteral (Int64.add n1 n2)
      | SNeg, SNeg -> SNeg
      | SLiteral n, SNeg when n <= Int64.zero -> SNeg
      | SNeg, SLiteral n when n <= Int64.zero -> SNeg
      | _ -> SAny
  end
  | BinOp (MinusA, e1, e2, _) -> begin
      match signOf e1, signOf e2 with
        SPos, SNeg -> SPos
      | SLiteral n, SNeg when n >= Int64.zero -> SPos
      | SPos, SLiteral n when n <= Int64.zero -> SPos
      | SLiteral n1, SLiteral n2 -> SLiteral (Int64.sub n1 n2)
      | SNeg, SPos -> SNeg
      | SLiteral n, SPos when n <= Int64.zero -> SNeg
      | SNeg, SLiteral n when n >= Int64.zero -> SNeg
      | _ -> SAny
  end
  | _ -> SAny


(* Do varinfo. We do the type and for all variables we also generate a node 
 * that will be used when we take the address of the variable (or if the 
 * variable contains an array) *)
let doVarinfo (vi : varinfo) = 
  (* Compute a place for it *)
  let place = 
    if vi.vglob then
      if vi.vstorage = Static then 
        N.PStatic (!currentFile.fileName, vi.vname)
      else
        N.PGlob vi.vname
    else
      N.PLocal (!currentFile.fileName, !currentFunctionName, vi.vname)
  in
  let vi_vtype = addArraySizedAttribute vi.vtype vi.vattr in 
  (* Do the type of the variable. Start the index at 1 *)
  let t', _ = doType vi_vtype place 1 in
  vi.vtype <- t';
(*  ignore (E.log "Did varinfo: %s. T=%a\n" vi.vname
            d_plaintype vi.vtype); *)
  (* Associate a node with the variable itself. Use index = 0 *)
  let n = N.getNode place 0 vi.vtype vi.vattr in
  (* Add this to the variable attributes. Note that this node might have been 
   * created earlier. Merge the attributes and make sure we get the _ptrnode 
   * attribute  *)
  vi.vattr <- addAttributes vi.vattr n.N.attr;
(*  ignore (E.log "varinfo: T=%a. A=%a\n" 
            d_plaintype vi.vtype (d_attrlist true) vi.vattr) *)

  (* sg: not global and not static means a variable goes on the stack *)
  if (not vi.vglob) && (vi.vstorage <> Static) then
	setOnStack n;

  ()

(* Do an expression. Return an expression, a type and a node. The node is 
 * only meaningful if the type is a TPtr _. In that case the node is also 
 * refered to from the attributes of TPtr. Otherwise the node is N.dummyNode
  *)
let rec doExp (e: exp) : exp * typ * N.node= 
  match e with 
    Lval lv -> 
      let lv', lvn = doLvalue lv false in
      Lval lv', lvn.N.btype, nodeOfType lvn.N.btype

  | AddrOf (lv) -> 
      let lv', lvn = doLvalue lv false in
      AddrOf (lv'), TPtr(lvn.N.btype, lvn.N.attr), lvn

  | StartOf lv -> 
      let lv', lvn = doLvalue lv false in
      let next = startOfNode lvn in
      StartOf lv', TPtr(next.N.btype, next.N.attr), next

  | UnOp (uo, e, tres) -> (* tres is an arithmetic type *)
      UnOp(uo, doExpAndCast e tres, tres), tres, N.dummyNode

  | SizeOf (t) ->
      let t', _ = doType t (N.anonPlace()) 1 in
      SizeOf (t'), uintType, N.dummyNode

  | SizeOfE (e) -> 
      let e', et', en' = doExp e in
      SizeOfE(e'), uintType, N.dummyNode

        (* arithmetic binop *)
  | BinOp (((PlusA|MinusA|Mult|Div|Mod|Shiftlt|Shiftrt|
    Lt|Gt|Le|Ge|Eq|Ne|BAnd|BXor|BOr|LtP|GtP|LeP|GeP|EqP|NeP|MinusPP) as bop), 
           e1, e2, tres) -> 
             BinOp(bop, doExpAndCast e1 tres,
                   doExpAndCast e2 tres, tres), tres, N.dummyNode
       (* pointer arithmetic *)
  | BinOp (((PlusPI|MinusPI|IndexPI) as bop), e1, e2, tres) -> 
      let e1', e1t, e1n = doExp e1 in
      let sign = 
        signOf 
          (match bop with PlusPI|IndexPI -> e2 | _ -> UnOp(Neg, e2, intType)) 
      in
      (match sign with
        SLiteral z -> 
          if z < Int64.zero then setArith e1n else
          if z > Int64.zero then setPosArith e1n else
          ()

      | SPos -> setPosArith e1n

      | _ -> 
          if bop = IndexPI then (*  Was created from p[e] *)
             setPosArith e1n
          else 
             setArith e1n);
      if sign = SLiteral Int64.zero then
        e1', e1t, e1n
      else
        BinOp (bop, e1', doExpAndCast e2 intType, e1t), e1t, e1n
      
      
  | CastE (newt, e) -> 
      let newt', _ = doType newt (N.anonPlace ()) 1 in
      CastE (newt', doExpAndCast e newt'), newt', nodeOfType newt'

  | Const (CStr s) as e -> 
      (* Add a cast in front of all strings. This way we have a place where 
       * to attach a node *)
      let newt', _ = doType charPtrType (N.anonPlace ()) 1 in
      CastE (newt', e), newt', nodeOfType newt'

  | _ -> (e, typeOf e, N.dummyNode)


(* Do initializers. All the initliazers at this point better be non-pointers *)
and doInit (i: init) : init * typ = 
  match i with 
  | SingleInit e -> 
      let e', t, n = doExp e in
      if n != N.dummyNode then 
        E.s (bug "Found pointer initializer: %a\n" d_init i);
      SingleInit e', t
          
  | CompoundInit (t, initl) -> 
      let t', _ = doType t (N.anonPlace ()) 1 in
      if nodeOfType t' != N.dummyNode then
        E.s (bug "found pointer initializer: %a\n" d_init i);
        (* Construct a new initializer list *)
      let doOneInit (off: offset) (ei: init) (tei: typ) acc = 
        let ei', _ = doInit ei in
        (off, ei') :: acc
      in
      let newinitl = 
        List.rev (foldLeftCompound ~doinit:doOneInit ~ct:t' 
                                   ~initl:initl ~acc:[]) in
      CompoundInit (t', newinitl), t'

(* Do an lvalue. We assume conservatively that this is for the purpose of 
 * taking its address. Return a modifed lvalue and a node that stands for & 
 * lval. Just ignore the node and get its base type if you do not want to 
 * take the address of. *)
and doLvalue ((base, off) : lval) (iswrite: bool) : lval * N.node = 
  let base', startNode = 
    match base with 
      Var vi -> begin 
        (* ignore (E.log "doLval (before): %s: T=%a\n"
                  vi.vname d_plaintype vi.vtype); *)
        doVarinfo vi; (* It was done when the variable was declared !!! *)
        let vn = 
          match N.nodeOfAttrlist vi.vattr with Some n -> n | _ -> N.dummyNode
        in
        (* Now grab the node for it *)
(*        ignore (E.log "doLval: %s: T=%a (ND=%d)\n"
                  vi.vname d_plaintype vi.vtype vn.N.id); *)
        base, vn
      end
    | Mem e -> 
        let e', et, ne = doExp e in
        if iswrite then
          setUpdated ne ;
        Mem e', ne
  in
  let newoff, newn = doOffset off startNode in
  (base', newoff), newn
        
(* Now do the offset. Base types are included in nodes. *)
and doOffset (off: offset) (n: N.node) : offset * N.node = 
  match off with 
    NoOffset -> off, n
  | Field(fi, resto) -> 
      (* We might need to change the fi in case we are in a polymorphic 
       * structure *)
      let fi' = 
        match unrollType n.N.btype with 
          TComp (nci, _) when nci != fi.fcomp -> 
            let fi' = 
              try 
                List.find (fun fi' -> fi'.fname = fi.fname) nci.cfields
              with Not_found -> 
                E.s (bug "Cannot find field %s in instance %s\n"
                       fi.fname (compFullName nci))
            in
            fi'
         
        | _ -> fi
      in
      let nextn = fieldOfNode n fi' in
      let newo, newn = doOffset resto nextn in
      Field(fi', newo), newn

  | Index(e, resto) -> begin
      let nextn = startOfNode n in
      setPosArith nextn ;
      let newo, newn = doOffset resto nextn in
      let e', et, _ = doExp e in
      Index(e', newo), newn
  end


  
(* Now model an assignment of a processed expression into a type *)
and expToType (e,et,en) t (callid: int) : exp = 
  let etn = nodeOfType et in
  let tn  = nodeOfType t in
(*  ignore (E.log "expToType e=%a (NS=%d) -> TD=%a (ND=%d)\n"
            d_plainexp e etn.N.id d_plaintype t tn.N.id); *)
  match etn == N.dummyNode, tn == N.dummyNode with
    true, true -> e (* scalar -> scalar *)
  | false, true -> e (* Ignore casts of pointer to non-pointer *)
  | false, false -> (* pointer to pointer *)
      if isZero e then begin
        setNull tn ;
        let _ = N.addEdge etn tn N.ENull callid None in 
        e
      end else 
        let addCastEdge (e: exp) (etn: N.node) (desttn : N.node) 
                        (callid: int) : exp = 
           (* See if it is a cast to or from void * *)
          let isvoidstar, tovoid = 
            match etn.N.btype, desttn.N.btype with
              TVoid _, TVoid _ -> false, false
            | TVoid _, _ -> true, false
            | _, TVoid _ -> true, true
            | _ -> false, false
          in
          if isvoidstar && !useGenerics then begin
(*            ignore (E.log "Found a generic: %a\n" d_exp e); *)
            (* Now create a new function *)
            let genericFun =   
              let fdec = emptyFunction 
                  ("generic" ^ (if tovoid then "put" else "get") ^ 
                   (string_of_int !genericId)) in
              incr genericId;
              let argx  = makeLocalVar fdec "x" (TPtr(etn.N.btype, 
                                                      etn.N.attr)) in
              fdec.svar.vtype <- TFun(TPtr(desttn.N.btype, desttn.N.attr), 
                                      Some [ argx ], false, []);
              fdec.svar.vstorage <- Extern;
              MU.theFile := GVarDecl (fdec.svar,lu) :: !MU.theFile;
              fdec
            in
            (* Now use the function instead of adding an edge *)
            e
          end else begin
(*            ignore (E.log "Setting %a : %a -> %d\n"
                      d_plainexp e d_plaintype et desttn.N.id); *)
            let _ = N.addEdge etn desttn N.ECast callid None in 
            e
          end
        in
        addCastEdge e etn tn callid

  | true, false -> (* scalar -> pointer *)
      (* Check for zero *)
      (if isZero e then
        setNull tn 
      else begin
        setIntCast tn
(*        match e with
          Const(CStr(_)) -> () 
        | _ -> 
          setIntCast tn 
*)
      end);
      e
    
and doExpAndCast e t = 
  (* Get rid of cascades of casts of 0 *)
  let e' = if isZero e then zero else e in
  expToType (doExp e') t (-1)

and doExpAndCastCall e t callid = 
  (* Get rid of cascades of casts of 0 *)
  let e' = if isZero e then zero else e in
  expToType (doExp e') t callid





(*****************************************************************)


(* return the first few items of a list *)
let rec list_first l n = begin
  if n >= 0 then (List.hd l) :: (list_first (List.tl l) (n-1))
  else []
end


(* Some utility functions for decomposing a function call so that we can 
 * process them in a special way *)
type callArgDescr = 
    { caOrig: exp; (* The original actual, without the cast that was added to 
                    * match it with the formal type *)
      caOrigType: typ; (* Type of caOrig *)
      caOrigNode: N.node; (* The node corresponding to caOrigType, if TPtr *)
      caFormal: varinfo;  (* The formal varinfo *)
      caFormalType: typ; (* The type of the corresponding formal *)
      caFormalNode: N.node; (* The node corrsponding to caFormalNode,if TPtr *)
    } 

let decomposeCall 
    (reso: lval option)
    (f: exp)
    (args: exp list) : (* result *) callArgDescr * 
                       (* args *) callArgDescr list = 
  (* Fetch the function type *)
  let rt, formals = 
    match unrollType (typeOf f) with
      TFun (rt, formals, _, _) -> rt, argsToList formals
    | _ -> E.s (E.bug "decomposingCall to a non-function")
  in
  if List.length formals <> List.length args then 
    E.s (E.bug "decomposeCall: mismatch of argument list length");
  (* A function to split a pointer type into base type, attribute and 
  * node of attributes *)
  let splitPtrType (t: typ) = 
    match unrollType t with
      TPtr(bt, a) -> begin
        bt, a, 
        (match N.nodeOfAttrlist a with
          Some n -> n
        | None -> N.dummyNode)
      end
    | _ -> voidType, [], N.dummyNode
  in
  let argDescr = 
    List.map2
      (fun a f ->
        (* Get the types of the arguments. But be prepared to strip the top 
         * level cast since it could have been added by cabs2cil *)
        let orig = stripCasts a in
        let origt = typeOf orig in
        let orig_bt, orig_a, orig_n = splitPtrType origt in
        (* So the same for the formal *)
        let formt = f.vtype in
        let form_bt, form_a, form_n = splitPtrType formt in
        { caOrig = orig; caOrigType = origt; caOrigNode = orig_n;
          caFormal = f; caFormalType = formt; caFormalNode = form_n })
      args formals
  in
  (* Now the return type *)
  let rt_bt, rt_a, rt_n = splitPtrType rt in
  let resDescr = 
    match reso, rt with 
      None, TVoid _ -> 
        { caOrig = zero; caOrigType = voidType; caOrigNode = N.dummyNode;
          caFormal = dummyFunDec.svar; 
          caFormalType = voidType; caFormalNode = N.dummyNode }
    | None, _ -> { caOrig = zero; caOrigType = voidType; 
                   caOrigNode = N.dummyNode;
                   caFormal = dummyFunDec.svar; 
                   caFormalType = rt; caFormalNode = rt_n }
    | Some _, TVoid _ -> E.s (E.bug "decomposeCall: assigned subroutine")
    | Some lv, _ -> 
        let origt = typeOfLval lv in
        let orig_bt, orig_a, orig_n = splitPtrType origt in
        { caOrig = Lval (lv); caOrigType = origt; 
          caOrigNode = orig_n; caFormal = dummyFunDec.svar; 
          caFormalType = rt; caFormalNode = rt_n }
  in
  resDescr, argDescr


let rec doBlock blk = 
  if hasAttribute "nobox" blk.battrs then 
    blk
  else 
    { bstmts = List.map doStmt blk.bstmts; battrs = blk.battrs }

and doStmt (s: stmt) : stmt = 
  (match s.skind with 
    Goto _ | Break _ | Continue _ -> ()
  | Return (None, _) -> ()
  | Return (Some e, l) -> 
      currentLoc := l;
      let e' = doExpAndCast e !currentResultType in
      (*sg:need to mark exp as escaped*)
      expMarkEscape e';
      s.skind <- Return (Some e', l)
  | Instr il -> 
      s.skind <- Instr (mapNoCopyList doInstr il)
  | Loop (b, l) -> 
      currentLoc := l;
      s.skind <- Loop (doBlock b, l)
  | Block b -> s.skind <- Block (doBlock b)
  | If(e, b1, b2, l) -> 
      currentLoc := l;
      s.skind <- If (doExpAndCast e intType, doBlock b1, doBlock b2, l)
  | Switch (e, b, cases, l) -> 
      currentLoc := l;
      s.skind <- Switch(doExpAndCast e intType, doBlock b, cases, l));
  s

and doInstr (i:instr) : instr list = 
  match i with
  | Asm _ -> [i]
  | Set (lv, e,l) -> 
      currentLoc := l;
      let lv', lvn = doLvalue lv true in
      (* Now process the copy *)
(*      ignore (E.log "Setting lv=%a\n lvt=%a (ND=%d)\n" 
        d_plainlval lv' d_plaintype lvn.N.btype lvn.N.id);  *)
      let e' = doExpAndCast e lvn.N.btype in
      (* sg: If assigning thru a pointer or to a global, mark adresses in e' 
      * as pkEscape. The former is a conservative approximation since 
      * depending on where lv can point, the value may probably not escape *)
      (match lv' with
	Mem _, _ -> expMarkEscape e' (* thru a pointer *)
      | Var vi, _ -> 
          if vi.vglob then expMarkEscape e' else () ); (* to a global *)
      
      [Set (lv', e', l)]

  | Call (reso, orig_func, args, l) as i -> begin
      currentLoc := l;
      (match orig_func with
        Lval(Var vf, NoOffset) -> 
          if not (H.mem MU.calledFunctions vf.vname) then
            H.add MU.calledFunctions vf.vname vf
      | _ -> ());
      match interceptFunctionCalls i with
        Some il -> mapNoCopyList doInstr il
      | None -> doFunctionCall reso orig_func args l
  end

and expMarkEscape (e : exp) : unit = 
	(*ignore (printf "--%a--\n" d_plainexp e); *)
  match e with 
    
    Lval lv -> 	
      let lvnode = nodeOfType (typeOfLval lv) (*get node for lv*)
      in setEscape lvnode 
        
  | StartOf lv -> 
      let lvnode = (* like typeOf, but keeps attrs of arrays *)
	(match unrollType (typeOfLval lv) with
	  TArray (t,_,al) -> nodeOfType (TPtr(t, al))
	| _ -> E.s (E.bug "expMarkEscape: StartOf on a non-array") )
      in
      setEscape lvnode
        
  | AddrOf lv  -> 
      let _, alvnode = doLvalue lv false (* gets node for &lv *)
      in setEscape alvnode
        
  | CastE(_, e1)   -> expMarkEscape e1
  | UnOp((Neg|BNot), e1, _) -> expMarkEscape e1 
        
  | BinOp( (Lt|Gt|Le|Ge|Eq|Ne|LtP|GtP|LeP|GeP|EqP|NeP), _, _, _) -> ()
  | BinOp(_, e1, e2, _) -> expMarkEscape e1; expMarkEscape e2 
        (*
  | Question(_, e1, e2) -> expMarkEscape e1; (*expMarkEscape e2*)
        *)
  | _ -> ()
        

and interceptFunctionCalls = function
  | i -> None
      
and doFunctionCall 
    (reso: lval option)
    (orig_func: exp)
    (args: exp list) 
    (l: location) = 
  incr callId; (* A new call id *)
  (*      ignore (E.log "Call %a args: %a\n" 
                d_plainexp orig_func
                (docList (chr ',') (fun a -> d_plaintype () (typeOf a)))
  args); *)
  let func, ispoly = (* check and see if it is polymorphic *)
    match orig_func with
      (Lval(Var(v),NoOffset)) -> 
        let newvi, ispoly = Poly.instantiatePolyFunc v in
        doVarinfo newvi;
        if ispoly then begin
          Lval (Var newvi, NoOffset), true
        end else (* Not polymorphic *) orig_func, false
    | _ -> orig_func, false
  in
  (* Do the function as if we were to take its address *)
  let (pfunc, _, pfuncn), funct = 
    match func with 
      Lval lv -> doExp (mkAddrOf lv), typeOfLval lv
    | _ -> E.s (unimp "Called function is not an lvalue")
  in
  (* Now fetch out the real function and its type *)
  let func' = Lval (mkMem pfunc NoOffset) in
  let (rt, formals, isva) = 
    match unrollType funct with
      TFun(rt, formals, isva, _) -> rt, argsToList formals, isva
    | _ -> E.s (bug "Call to a non-function")
  in
  let preinstr, args' = 
    if isva then 
      Vararg.prepareVarargArguments func' (List.length formals) args
    else
      [], args
  in
  (* If the function has more actual arguments than formals then mark 
  * the function node as used without prototype  *)
  let make_it_wild = 
    if List.length args' <> List.length formals && not isva then begin
      (* Bark if it is polymorphic. No prototype + polymorphism (or 
      * allocation) do not work together *)
      if ispoly then 
        E.s (error "Calling polymorphic (or allocation) function %a without proper prototype" d_exp func);
      ignore (warn "Calling function %a with too many arguments: will be WILD"
                d_exp func) ;
      N.setFlag pfuncn N.pkNoPrototype ;
      true
    end 
    else false
  in
  (* Now check the arguments *)
  let rec loopArgs formals args = 
    match formals, args with
      [], [] -> []
    | [], a :: args -> (* We ran out of formals. This is bad, so we 
        * make sure to mark that the argument is used 
        * in a function without prototypes  *)
        (* Do the arguments because they might contain pointer types *)
        let a', _, an = doExp a in
        if an != N.dummyNode && not isva  then
	  N.setFlag an N.pkNoPrototype ; 
        a' :: loopArgs [] args
                
    | fo :: formals, a :: args -> 
        (* See if this is a polymorphic argument. Then strip the cast to 
         * void* *)
        let a' = 
          match ispoly, unrollType fo.vtype with
            true, TPtr(TVoid _, _) -> true
          | _, _ -> false
        in
        (*            ignore (E.log "Call arg %a: %a -> %s\n" 
                      d_exp a' d_plaintype (typeOf a') fo.vname); *)
        let a' = doExpAndCastCall a fo.vtype !callId in
        a' :: loopArgs formals args
                
    | _, _ -> E.s (E.unimp "Markptr: not enough arguments in call to %a" 
                     d_exp orig_func)
  in  
  let reso' = 
    (* Now check the return value*)
    match reso, unrollType rt with
      None, TVoid _ -> None
    | Some _, TVoid _ -> 
        ignore (warn "Call of subroutine is assigned");
        None

    | None, _ -> None (* "Call of function is not assigned" *)
    | Some dest, _ -> begin
        (* Do the lvalue, just so that the type is done *)
        let dest', _ = doLvalue dest true in
        (* Add the cast from the return type to the destination of the call. 
         * Make up a phony expression and a node so that we can call 
         * expToType.  *)
(*        ignore (E.log "Call to %a. \n\trt=%a\n\ttypeOFLval(dest') = %a\n"
                  d_exp func' d_type rt d_plaintype (typeOfLval dest')); *)
        ignore (expToType (Const(CStr("a call return")),
                           rt, N.dummyNode) (typeOfLval dest') 
                  !callId);
        Some dest'
    end 
  in
  (* Now do the arguments *)
  let args'' = loopArgs formals args' in
  (* Take a look at a few special functions *)
  Wrappers.addConstraints func' args'' rt !callId;
  (* We need to mark all instructions that we have generated *)
  let preinstr' = mapNoCopyList doInstr preinstr in
  preinstr' @ [Call(reso', func', args'', l)]


let doFunctionBody (fdec: fundec) = 
  currentFunctionName := fdec.svar.vname;
  (* See if this is a vararg function. Must do this first because we might 
   * change a lot of temporaries, which we want to process *)
  Vararg.processVarargBody fdec;
  (* Go through the formals and copy their type and attributes from 
  * the type. Then restore the sharing  *)
  let rt, targs, isva, fa = splitFunctionType fdec.svar in
  let rec scanFormals targs sformals = 
    match targs, sformals with
      [], [] -> ()
    | ta :: targs, sf :: sformals -> 
        sf.vtype <- ta.vtype;
        sf.vattr <- ta.vattr;
        scanFormals targs sformals
    | _ -> E.s (bug "scanFormals(%s) non-matching formal lists"
                  fdec.svar.vname)
  in
  scanFormals (argsToList targs) fdec.sformals;
  (* Restore the sharing by writing the type *)
  setFormals fdec fdec.sformals;
  currentResultType := rt;
  (* Do the other locals *)
  List.iter doVarinfo fdec.slocals;
  (* Do the body *)
  fdec.sbody <- doBlock fdec.sbody

(* let disableModelCheck = ref false *)

(** A visitor that finds the function calls *)
class findCallSites : cilVisitor = object
  inherit nopCilVisitor

    (* Change all call instructions to use the marker *)
  method vinst (i: instr) : instr list visitAction = 
    match i with
      Call(opt, Lval(Var vf, NoOffset), params, loc) -> begin
        let record_call s = 
          if not (H.mem MU.calledFunctions s) then
            H.add MU.calledFunctions s vf in
        try 
          if Wrappers.isAWrapper !currentFunctionName then begin
            if Wrappers.debugWrappers then
              ignore (E.log "ignoring call to %s in wrapped function %s.\n" vf.vname !currentFunctionName);
            record_call vf.vname;
	    SkipChildren
	  end else
	    let wrapper = Wrappers.findWrapperOf vf.vname in
            if Wrappers.debugWrappers then
              ignore (E.log "changing %s to %s in %s.\n" vf.vname wrapper.svar.vname !currentFunctionName);
            record_call wrapper.svar.vname;
	    ChangeTo [Call(opt, Lval(Var wrapper.svar, NoOffset), params, loc)]
        with Not_found ->
          (*This function is not wrapped *)
          if Wrappers.debugWrappers then
            ignore (E.log "ignoring call to %s in function %s.\n" vf.vname !currentFunctionName);
          record_call vf.vname;
	  SkipChildren
      end

      | _ -> SkipChildren
      
  method vexpr _ = SkipChildren

  method vtype _ = SkipChildren
  end
let findCallSitesVisitor = new findCallSites
  
(* Now do the globals *)
let doGlobal (g: global) : global = 
  try
    match g with
    | GPragma (a, l) as g -> begin
        (* Most of the pragmas have been processed. Only a few have to be 
        * processed at the same time with the other globals *)
        currentLoc := l;
        (match a with
        | Attr("box", [ACons("on", _)]) -> boxing := true
        | Attr("box", [ACons("off", _)]) -> boxing := false
        | _ -> ());
        g
    end
    | _ -> begin
        if not !boxing then g
        else match g with
        | GText _ | GAsm _ | GEnumTag _ | GCompTagDecl _ | GEnumTagDecl _ -> g
              
              (* We process here only those types that we must not unroll. 
               * The others we'll process as we see them used.*)
        | GType (t, l) -> 
            currentLoc := l;
            if H.mem dontUnrollTypes t.tname then begin
              let t', _ = doType t.ttype (N.PType t.tname) 1 in
              t.ttype <- t'
            end;
            g
              
        | GCompTag (comp, l) -> 
            currentLoc := l;
            if not (Poly.isPolyComp comp) then begin
              markCompInfo comp;
              g
            end else begin
              (* It is polymorphic, so nobody should be using this copy 
               * anymore *)
              GText("// definition of polymorphic struct " ^ comp.cname ^ 
                    "used to be here")
            end
              
        | GVarDecl (vi, l) -> 
            currentLoc := l;
            (* ignore (E.log "Found GVarDecl of %s. T=%a\n" vi.vname
                    d_plaintype vi.vtype); *)
            if not (Poly.isPolyFunc vi) then doVarinfo vi; 
            (match vi.vtype with 
              TFun _ -> Vararg.processFormatAttribute vi
            | _ -> ());
            g
              
        | GVar (vi, init, l) -> 
            currentLoc := l;
            let init' = 
              match init with
                None -> None
              | Some i -> 
                  let i', _ = doInit i in
                  Some i'
            in
            GVar (vi, init', l)
              
        | GFun (fdec, l) ->
            currentLoc := l;
            Vararg.processFormatAttribute fdec.svar;
            let dobox = not (hasAttribute "nobox" fdec.svar.vattr) in
            if dobox then begin
              (* If it is polymorphic then remember it for later. *)
              if Poly.isPolyFunc fdec.svar then begin
                Poly.rememberFunctionDefinition fdec
              end else begin
                doVarinfo fdec.svar;
                doFunctionBody fdec;
                g
              end
            end else
              g
                
                
        | GPragma _ -> g (* Should never be reached *)
    end
  with e -> begin
    ignore (E.log "Markptr: global error (%s) at %t\n" 
              (Printexc.to_string e) d_thisloc);
    E.hadErrors := true;
    dGlobal (dprintf "markptr_error_in_global") !currentLoc
  end
  

      
(* Now do the file *)      
let markFile fl = 
  currentFile := fl;
  boxing := true;
(*  disableModelCheck := false; *)
  E.hadErrors := false;
  MU.init();
  MU.theFile := [];
  (* Add some prototypes *)
  Vararg.initFile ();
  Poly.initFile ();
  Models.initFile ();
  Wrappers.initFile ();
  H.clear dontUnrollTypes;
  (* Some globals that are exported and must thus be considered part of the 
   * interface *)
  let exported : (string, bool) H.t = H.create 111 in
  (* Find the globals that are declared but not defined. They are part of 
   * the interface. Process the pragmas as well *)
  let interfglobs : (int, varinfo) H.t = H.create 111 in
  
  let processDecls = function
      GVarDecl (vi, l) ->
        currentLoc := l;
        if not (H.mem interfglobs vi.vid) then 
          (* Do not add multiple times since then deleting does not remove 
           * all copies *)
          H.add interfglobs vi.vid vi;
        (match vi.vtype with
          TFun _ -> MU.registerFunction (MU.Declared vi)
        | _ -> ())

    | GFun(fdec, _) -> 
        MU.registerFunction (MU.Defined fdec);
        currentFunctionName := fdec.svar.vname;
        (* And scan its body to find the called functions *)
        ignore (visitCilBlock findCallSitesVisitor fdec.sbody)

    | GPragma (a, l) -> begin
        currentLoc := l;
        match a with
        | Attr("boxalloc", AStr(s) :: _) -> 
            (* Allocators are not cured *)
            MU.applyToFunction s
              (fun vi -> vi.vattr <- addAttribute (Attr("nobox",[])) vi.vattr);
            (* Allocators are polymorphic *)
            Poly.processPragma a

        | Attr("boxexported", [AStr s]) ->
            H.add exported s true
        | Attr("boxnounroll", [AStr s]) ->
            H.add dontUnrollTypes s true

        | Attr("nobox", funcs) ->
            List.iter 
              (function (AStr s) -> 
                (* Make nobox functions polymorphic *)
                Poly.processPragma a;
                if MU.alreadyDefinedFunction s then
                  ignore 
                    (warn "#pragma nobox appears after definition of %s\n" s)
                else begin
                  MU.applyToFunction s
                    (fun vi -> 
                      vi.vattr <- addAttribute (Attr("nobox",[])) vi.vattr);
                end
                | _ -> ignore (warn "Invalid #pragma nobox"))
              funcs
            

        | _ -> Vararg.processPragma a;
               Models.processPragma a;
               Poly.processPragma a
    end
    | _ -> ()
  in
  let processWrapperPragmas = function
      GPragma (a, l) -> 
        Wrappers.processPragma a;
    | _ -> ()
  in

  (* Add all the ccuredwrapperof pragmas, then the rest of the global declarations. *)
  List.iter processWrapperPragmas fl.globals;
  List.iter processDecls fl.globals;
  (* Take out the defined functions *)
  let rec processDefs = function
      GVar (vi, _, _) -> processVarinfo vi
    | GFun (fdec, _) -> processVarinfo fdec.svar
    | _ -> ()
  and processVarinfo vi = 
    if H.mem exported vi.vname then begin
      if not (H.mem interfglobs vi.vid) then
        H.add interfglobs vi.vid vi
    end else
      H.remove interfglobs vi.vid
  in
  List.iter processDefs fl.globals;
(*  ignore (E.log "inferfglob:\n");
  H.iter (fun sid v -> ignore (E.log "%s: %d@!" v.vname sid)) interfglobs; *)

  (* Now we have all the information to add the models *)
  (* Add models for the caled but not defined functions *)
  let models : global list = Models.addModels () in 
  MU.theFile := [];
  (* This is where we process all the functions *)
  List.iter (fun g -> let g' = doGlobal g in 
                      MU.theFile := g' :: !MU.theFile) fl.globals;
  (* Process and add the models at the end *)
  List.iter (fun g -> let g' = doGlobal g in 
                      MU.theFile := g' :: !MU.theFile) models;

  (* Now do the globinit *)
  let newglobinit = 
    match fl.globinit with
      None -> None
    | Some g -> begin
        match doGlobal (GFun(g, locUnknown)) with
          GFun (g', _) -> Some g'
        | _ -> E.s (bug "markptr: globinit")
    end
  in

  (* Now we must create the copies for the polymorphic functions *)
  Poly.finishInstantiations 
    (fun f -> 
(*    ignore (E.log "Marking body of %s after instantiation\n" f.svar.vname);*)
      let g = doGlobal (GFun(f, locUnknown)) in
      MU.theFile :=  g :: !MU.theFile);

  (* There might be some polymorphic functions that were not used. Make sure 
   * we have an instance of those as well, except if the body is a model. 
GN: I forgot why I added this. We'll see what breaks.
  H.iter 
    (fun oldname -> fun body ->
      if  H.mem boxModels (Poly.stripPoly body.svar.vname) 
       && List.filter (fun (on, _) -> on = oldname) !allinstantiations != [] 
      then
        let g' = doGlobal (GFun(body, locUnknown)) in
        MU.theFile :=  g' :: !MU.theFile)
    polyBodies;
  *)
  let newglobals = List.rev !MU.theFile in
        
     
  (* Now go through the types of interfglobs and mark all nodes that are part 
   * of the interface *)
  H.iter
    (fun id vi -> 
      ignore 
        (existsType 
           (fun t -> 
             match N.nodeOfAttrlist (typeAttrs t) with
               Some n -> setInterface n; ExistsMaybe
             | _ -> ExistsMaybe)
           vi.vtype))
    interfglobs;
    
  if !E.verboseFlag || !E.hadErrors then
    ignore (E.log "Markptr: %s\n"
              (if !E.hadErrors then "Error" else "Success"));
  let newfile = {fl with globals = newglobals; globinit = newglobinit} in
  if !Util.doCheck then
    Check.checkFile [] newfile;
  H.clear dontUnrollTypes;
  MU.init ();
  Models.initFile ();
  Poly.initFile ();
  currentFile := dummyFile;
  newfile

        
let solver = ref "fourth"

(* A special file printer *)
let printFile (c: out_channel) fl = 
  Stats.time "printMarkedfile" (Cil.printFile N.ccuredInferPrinter c) fl;
  output_string c "#if 0\n/* Now the graph */\n";
  output_string c "/* Now the solved graph (simplesolve) */\n";
  Stats.time "printgraph" N.printGraph c;
  output_string c "/* End of solved graph*/\n#endif\n";
  Stats.time "graph stats" N.printGraphStats ();
  if !E.hadErrors then 
    E.s (E.error "There were errors during markptr\n");
  ()

