(*
 *
 * Copyright (c) 2001 by
 *  George C. Necula	necula@cs.berkeley.edu
 *  Scott McPeak        smcpeak@cs.berkeley.edu
 *  Wes Weimer          weimer@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products 
 * derived from  this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

open Cil
open Pretty

module H = Hashtbl
module E = Errormsg

module N = Ptrnode

module MU = Markutil

let debugModels = false

(* We keep track of the models to use *)
let boxModeledBy: (string, string) H.t = H.create 15 (* Map the name of the 
                                                    * modeled function to 
                                                    * the model *)
(* And of all models *)
let boxModels: (string, bool) H.t = H.create 13

let processPragma = function
  | Attr("boxmodelof", AStr modelname :: lst) -> 
      if lst = [] then 
        ignore 
          (warn "#pragma boxmodelof with empty list of modeled functions")
      else begin
        H.add boxModels modelname true;
        (* Make sure we mark the body as only a model so that it can be 
         * removed later *)
        MU.applyToFunction modelname 
          (fun fvar -> 
            fvar.vattr <- addAttribute (Attr("boxmodel",[])) fvar.vattr);
        (* Once you see the definition of the model apply it to the modeled 
        * functions *)
        List.iter 
          (function 
              AStr modeled -> 
                if debugModels then 
                  ignore (E.log "Will use model %s for %s\n"
                            modelname modeled);
                H.add boxModeledBy modeled modelname
            | _ -> ignore (warn "invalid #pragma boxmodelof"))
          lst
      end

  | _ -> ()

(* Find the model of a function. Might raise Not_found *)
let findModelOf (s: string) = 
  let modelname = H.find boxModeledBy s in
  try
    match H.find MU.allFunctions modelname with
      MU.Defined fdec -> fdec
    | _ -> 
        E.s (error "Cannot find the definition of model %s of %s\n"
               modelname s)
  with Not_found -> 
    E.s (error "Cannot find the definition of model %s of %s\n"
           modelname s)
    


(** Scan all of the declared functions and see if they have a model. is give 
 * a function that can be used to process the newly created models *)
let addModels () : global list = 
  let res : global list ref = ref [] in
  (* Iterate over all functions that were actually called *)
  H.iter 
    (fun fname -> fun fvi -> 
      (* Skip it if defined also in the file *)
      let isNotDefined =
        try match (H.find MU.allFunctions fname) with 
          MU.Defined _ -> false
        | _ -> true
              (* Maybe it is not even declared *)
        with Not_found -> true
      in
      if isNotDefined then begin
        (* Do we have a model? *)
        try 
          let model : fundec = findModelOf fname in
          (* Here if we have a model *)          
          (* For now we dislike modeled functions without prototype *)
          if [] != (filterAttributes "missingproto" 
                      (MU.getFunctionTypeAttributes (Lval (var fvi)))) then
            E.s (unimp 
                   "Function %s is modeled but does not have a prototype\n" 
                   fname);
          if debugModels then
            ignore (E.log "Creating body for %s based on model %s\n"
                      fname model.svar.vname);
          
          (* Make the sformals *)
          let rt, sformals, va, l = 
            match fvi.vtype with
            TFun(rt, args, va, l) -> rt, argsToList args, va, l 
            | _ -> 
                E.s (E.bug "Modeled function %s does not have a function type"
                       fname)
          in
          (* Go over the formals and make sure that we assign the right ids. 
           * Being in a type they might not have the right IDS. This is safe 
           * since we know that there is no definition already for this func. 
          *)
          let vid = ref 0 in (* for local ids *)
          List.iter (fun s -> s.vid <- !vid; incr vid) sformals;
          let modeledFun = { svar    = fvi;
                             sformals = sformals;
                             slocals  = [];
                             smaxid   = !vid - 1;
                             sinline  = false;
                             sbody    = mkBlock [] ;
			     smaxstmtid = None ;} in
          (* Now make the body *)
          let reslvo, reso = 
            match rt with 
              TVoid _ -> None, None
            | t -> 
                let tmp = makeTempVar modeledFun t in
                Some (var tmp), Some (Lval (var tmp))
          in
          let call = 
            mkStmtOneInstr
              (Call (reslvo, Lval (var model.svar),
                     List.map (fun a -> Lval (var a)) sformals,
                     locUnknown)) in
          let return = mkStmt (Return (reso, locUnknown)) in
          modeledFun.sbody <- mkBlock [call; return];
          (* Mark it as modeled *)
          fvi.vattr <- addAttribute (Attr("modeledbody",[])) fvi.vattr;
          (* Just mark the body *)
          res := GText ("// Modeling body of " 
                               ^ fvi.vname ^ " based on model " ^
                               model.svar.vname) :: !res;
(*          let g' = processFunction modeledFun in *)
          res :=  GFun(modeledFun, locUnknown) :: !res
        with Not_found (* model *) -> ()
      end)
    MU.calledFunctions;
  (* Now return all the models *)
  !res



let initFile () = 
  H.clear boxModels;
  H.clear boxModeledBy
