(*
 *
 * Copyright (c) 2001 by
 *  George C. Necula	necula@cs.berkeley.edu
 *  Scott McPeak        smcpeak@cs.berkeley.edu
 *  Wes Weimer          weimer@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products 
 * derived from  this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

(* Weimer
 *
 * CCUred pointer kind inference using an extension of the algorithm
 * described in the paper submitted to PLDI'02. 
 *
 * INPUT ASSUMPTIONS: 
 *
 * The node hashtable includes the following edges:
 *      n1 ECAST n2     for the top-level nodes of casts
 *      n1 ENULL n2     cast or assignment where value of exp of n1 is 0 
 *      n1 ESAFE n2     from the top of a struct to its non-array fields
 *      n1 EINDEX n2    from the top of a struct to array fields
 *
 * Nodes have the following bits set iff the corresponding fact actually
 * originate at that node: 
 *      pkPosArith
 *      pkArith
 *      pkNull
 *      pkUpdated
 *      pkIntCast
 *      pkInterface
 *      pkSized
 *      pkNoPrototype (makes a function node wild)
 *
 * The "is_p x y" function returns true if x--cast-->y is from a polymorphic
 * function and should not be considered. 
 *
 * Returns "true" if we should typecheck the result because we let some
 * user-annotated qualifier stand. 
 *)

(*
 * How does the solver work? 
 *
 * We treat every "void *" as a separate type variable and we keep
 * equivalence classes of void* nodes that are connected by ECast or
 * ECompat edges. ECast edges come from Markptr as part of our
 * precondition. We create ECompat edges. 
 *
 * (1) We examine every CAST in the program. While scanning the
 *      types to see if the cast is valid we add ECompat edges between
 *      inner types that must be equal and mark as Wild types that
 *      do not match up. We also mark nodes as Not Safe in casts where
 *      the sizes are wrong. 
 * (2) Check equivalence classes. All nodes joined by ECompat edges must
 *      have equal types. Make sure all nodes in all equiv classes have
 *      the same flags. 
 * (3) We assign base-case flags. For example, we assign
 *      "pkReachString" to nodes that are of type string.
 * (4) We push those flags around using standard data-flow tricks. Nodes
 *      connected by ECompat edges have identical flags. 
 * (5) Once we have all the flags in place we distinguish between all
 *      the kinds of sequences. For example, a node that cannot be safe
 *      and has an integer cast into it but does not reach a string becomes 
 *      FSEQ. 
 * (6) We turn all string-like nodes that lead only into ROString nodes 
 *      into ROString nodes themselves. Note all nodes connected by ECompat
 *      edges must have the same kind. 
 * (7) We push WILD as far as it can go. Generally WILD contaminates all
 *      outgoing edges and all points-to targets, however an edge from
 *      WILD to ROString is allowed. 
 * (8) All otherwise unconstrainted nodes become SAFE. 
 *
 * Compared to previous solvers, this solver is much stricter about
 * strings. No "safe -> string" or "safe -> rostring" casts should remain. 
 *)

open Cil
open Ptrnode
open Pretty
module E = Errormsg

let interface_char_ptr_is_ROSTRING = false
  (* should all interface 'char *'s be made into ROSTRINGS? *)

(* Set this to true and initialize the watch_which_nodes with a list of nodes
 * to be watched *)
let logUpdates = false
let logNodeUpdate = 
  let which_nodes = [ 1; ] in
  let which_nodes = 
    let h : (int, bool) Hashtbl.t = Hashtbl.create 13 in
    List.iter (fun nid -> Hashtbl.add h nid true) which_nodes;
    h
  in
  fun  (n: node) (k: opointerkind) (new_w : whykind) (old_w: whykind) ->
    if Hashtbl.mem which_nodes n.id then
      ignore (E.warn "updating %d to %a (%a). Old kind: %a (%a)"
                n.id d_opointerkind k d_whykind new_w
                d_opointerkind n.kind
                d_whykind old_w)

(* helper function: does a node represent a pointer to a character? *)
let is_char_pointer n =
  match n.btype with
    TInt(IChar,_) -> true
  | TInt(ISChar,_) -> true
  | TInt(IUChar,_) -> true
  | _ -> false

(* helper function: is this node's kind some form of string? *)
let is_string k = (k = String || k = ROString) 

(* was the node's kind set by some outside force? *)
let set_outside n = n.why_kind = UserSpec || n.why_kind = PrintfArg

(* go through the source code and find where tau was declared *)
class findTypeVisitor tau = object
  inherit nopCilVisitor
  method vtype some_type = 
    if tau = some_type then 
      ignore (error "type %a was not given a node by markptr" d_type tau) ;
    DoChildren
end

(* equiv classes of nodes connected by ECompat edges. *)
module OrderedNode = struct type t = node let compare = compare end
module NodeSet = Set.Make(OrderedNode)
module NodeUF = Unionfind.Make(NodeSet)

(*
 **
 *** The Solver!
 ** 
 *)
let solve (the_file : Cil.file) (node_ht : (int, node) Hashtbl.t) : bool = begin
  let node_eq = ref (NodeUF.empty) in (* ECompat equivalence classes *)

  let existsECompatEdge ~(start : node) ~(dest : node) = 
    List.fold_left (fun acc e -> acc ||
     (e.eto.id = dest.id && match e.ekind with ECompat _ -> true | _ ->
     false)) false start.succ
  in 

  (* Say that n1 and n2 (which are usually matching inner pointers) must
   * really be equal. This adds an ECompat edge between them and places
   * them in the same equivalence class. We avoid making duplicate edges.
   *)
  let rec addECompatEdge n1 n2 ecast_edges location =
    if existsECompatEdge n1 n2 || 
       existsECompatEdge n2 n1 || 
       n1.id = n2.id then
      ()
    else begin
      assert(Clist.fold_left (fun acc e -> acc && e.ekind = ECast)
        true ecast_edges) ;
      let new_edge = addEdge n1 n2 (ECompat(ecast_edges)) 
        (-1) (Some(location)) in
      if (isVoidType n1.btype || isVoidType n2.btype) then 
        join n1 n2 ecast_edges
      else 
        node_eq := NodeUF.make_equal (!node_eq) n1 n2  ;

    end
  in 
  
  (* whenever we have to "loop until things settle", we used
   * finished to keep track of our status *)
  let finished = ref false in 

  (* should we force a typecheck later? *)
  let we_should_typecheck = ref false in 

  (* update changes node "n"'s kind to "k" because of "w" and sets
   * finished to false. It also does logging and sanity checking. *)
  let update_warnings = ref 5 in
  let update n k w = begin
    if (k <> n.kind) then begin
      if n.why_kind = UserSpec then begin
        if !update_warnings > 0 then 
          ignore (E.warn 
            "Solver: not changing User Specified %a node %d to %a" 
            d_opointerkind n.kind n.id d_opointerkind k);
        decr update_warnings ; 
        if !update_warnings = 0 then
          ignore (E.warn 
            "Solver: skipping further 'not changing User Specified' warnings"
            );
        if (n.kind = Wild) then 
          E.s (E.warn 
            "Solver: not changing User Specified %a node %d to %a" 
            d_opointerkind n.kind n.id d_opointerkind k);
        we_should_typecheck := true 
      end else begin
        if logUpdates then logNodeUpdate n k w n.why_kind;
        n.kind <- k ; 
        n.why_kind <- w ;
        finished := false 
      end
    end
  end in

  (* Help Function: find the attributes of a type *)
  let node_of_type tau = 
    nodeOfAttrlist (typeAttrs tau)
  in

  (* Step 1
   * ~~~~~~
   * Consider every cast.
   *
   * Generate ECOMPAT edges between aligned sub-pointers.
   * Generate BADCAST constaints on failed pointers (make 'em WILD)
   * Generate ARITH constaints on upcasts (make 'em not SAFE). 
   *)
  if !E.verboseFlag then ignore (E.log "Solver: Step 1  (Casts)\n") ;

  (* try to find a node for this type *)
  let pointer_with_no_node tau1 =
    match node_of_type tau1 with
      Some(n) -> ()
    | None -> begin

    end
  in 


  (* Whenever we compare two types for equality we should mark all
   * matching inner pointers with ECOMPAT edges. This function is called
   * by the type-scanning phase on all pairs of pointers that should
   * match up. *)
  let handle_inner_pointers loc explanation tau1 tau2 = 
    try begin
      match (node_of_type tau1),(node_of_type tau2) with
      | Some(n1),Some(n2) -> 
          addECompatEdge n1 n2 explanation loc;

      | Some(n),None 
      | None,Some(n) when (isVoidType n.btype) -> begin
          (* special case: link a "void*" equivalence class with the magic
           * scalar node *)
          addECompatEdge n scalarNode explanation loc;
          end

      | _,_ -> (* in this unfortunate case, we don't know how to get
              * to the nodes of these types *)
        begin 
        if node_of_type tau1 = None then 
          ignore (visitCilFile (new findTypeVisitor tau1) the_file) ;
        if node_of_type tau2 = None then 
          ignore (visitCilFile (new findTypeVisitor tau2) the_file) ;
        E.s (E.bug "Solver: cannot link inner pointers:@!%a@!%a\n"
          d_type tau1 d_type tau2)  ; E.hadErrors := true 
        end 
    end with _ -> ()
  in 

  (* In the process of scanning types involved in a cast, we may come
   * up with two pointer types that fail to match up. This function is
   * called on those types. We make them WILD. *)
  let handle_failure_node why n1 =
  (* E.warn "failure: node %d (Ptr(%a)) is now WILD" n1.id d_type n1.btype ; *)
    update n1 Wild why 
  in 
  let rec handle_failure e why tau =
    try
      match (node_of_type tau) with
        Some(n1) -> handle_failure_node why n1
      | None -> 
          if (tau = unrollType e.eto.btype || 
              tau = unrollType e.efrom.btype) then begin
            handle_failure_node why e.eto ;
            handle_failure_node why e.efrom ;
          end else begin match tau with
            TFun(r,None,b,a) -> if isPointerType r then handle_failure e why r  
          | TFun(r,Some(valist),b,a) -> 
              if isPointerType r then handle_failure e why r ; 
              List.iter (fun va -> if isPointerType va.vtype then 
                  handle_failure e why va.vtype) valist
          | _ -> ignore (E.s (E.bug "Solver: unknown failure case (1)@!%a"
            d_type tau)) ; E.hadErrors := true 
          end
    with _ -> () 
  in 

  (* replace a "void *" type with its representative *)
  let polymorphic_replace tau = (* tau is already unrolled *)
    match tau with
    | (TPtr((TVoid _),attr)) -> begin
        match nodeOfAttrlist attr with
          Some(n) -> 
            let rep,why = get_rep_why n in 
            (TPtr((rep).btype,attr)),why
        | None -> tau,(Clist.empty)
      end
    | _ -> tau,(Clist.empty)
  in 
    
  (* 
   * Step 1 Loop : examine every cast
   *)
  let step1_oneEdge e = (* look at every forward edge *)
    if e.ekind = ECast then begin
      (* E.warn "Considering Edge %d->%d" e.efrom.id e.eto.id ;  *)
      let from_rep,ex1 = get_rep_why e.efrom in
      let to_rep,ex2 = get_rep_why e.eto in 
      let from_t = from_rep.btype in
      let to_t = to_rep.btype in 
      let explanation = Clist.append (Clist.append ex1 ex2) (Clist.single e)in
      if (isVoidType from_t || isVoidType to_t) then 
        addECompatEdge e.eto e.efrom (Clist.single e) e.eloc
      else begin
        let from_size = try bitsSizeOf(from_t) with _ -> 1 in
        let to_size = try bitsSizeOf(to_t) with _ -> 1 in
        if from_size < to_size then begin
          setFlag e.efrom pkNotSafe ; (* ARITH constraint *)
        end ;
        
        Stats.time "subtype" (fun () -> 
          if not (Type.subtype 
              ~compat:(handle_inner_pointers e.eloc)
              ~failure:(handle_failure e (BadCast(e)))
              ~replace:(polymorphic_replace)
              ~explanation:(explanation)
              to_t from_t) 
            then begin (* they are not subtypes *)
              if (Type.all_scalars ~replace:(polymorphic_replace) to_t) && 
                 (Type.all_scalars ~replace:(polymorphic_replace) from_t) 
              then begin
                setFlag e.eto pkNotSafe ; (* ARITH constraint *)
                setFlag e.efrom pkNotSafe ; (* ARITH constraint *)
                (* In this one special case, we'll infer that the two
                 * types should be SEQ *)
              end else begin
                handle_failure_node (BadCast(e)) e.efrom ;
                handle_failure_node (BadCast(e)) e.eto 
              end
            end 
            ) ()
      end
    end else if e.ekind = EIndex then begin
      (* while we're here, these arrays cannot be safe *)
      setFlag e.eto pkNotSafe ;
      setFlag e.eto pkReachSeq 
    end
  in
  let rec handle_union_failure why tau =
    try
      match (node_of_type tau) with
        Some(n1) -> handle_failure_node why n1
      | None -> begin
          match tau with
            TFun(r,None,b,a) -> if isPointerType r then 
                handle_union_failure why r  
          | TFun(r,Some(valist),b,a) -> handle_union_failure why r ; 
              List.iter (fun va -> if isPointerType va.vtype then 
                  handle_union_failure why va.vtype) valist
          | _ -> ignore (E.s (E.bug "Solver: unknown failure case (2)@!%a"
            d_type tau)) ; E.hadErrors := true 
          end
    with _ -> () 
  in 

  Hashtbl.iter (fun id cur ->
    List.iter step1_oneEdge cur.succ ; (* look at every forward edge *)
    let rep,ex1 = get_rep_why cur in
    match rep.btype with
      TComp(ci,al) when not ci.cstruct ->
        let okay = Type.check_union rep.btype 
          (handle_inner_pointers locUnknown)
          (handle_union_failure (UnsafeUnion(rep.btype)))
          (polymorphic_replace)
          (ex1)
        in if not okay then
          update rep Wild (UnsafeUnion(rep.btype))
    | _ -> () 
  ) node_ht ; 

  (* Step 2
   * ~~~~~~
   * Check our equivalence classes for consistency.
   *
   * All "void *" equivalence class nodes should have an ECompat edge to
   * their rep. The rep has the flags for the whole class. 
   *
   * All nodes in "void*" equivalence classes should link up everything
   * they point to in "void*" equivalence classes as well. 
   *
   * All nodes in equiv classes should have the same flags. 
   *)
  (* First we add transitive closure of the ECast edges on "void *" *)
  if !E.verboseFlag then ignore (E.log "Solver: Step 2  (equiv)\n") ;

  let check_compat_fun e =
    if isECompat e then begin
      let to_t = e.eto.btype in
      let from_t = e.efrom.btype in 
      if not (isVoidType to_t) &&
         not (isVoidType from_t) && 
         not (Stats.time "subtype" (fun () -> Type.equal 
          ~failure:(handle_failure e (Incompat(e))) 
          ~replace:(polymorphic_replace)
          ~explanation:(Clist.single e)
          to_t from_t) ()) 
        then begin
              handle_failure_node (Incompat(e)) e.eto ;
              handle_failure_node (Incompat(e)) e.efrom
          end
    end
  in 

  Hashtbl.iter (fun id cur ->
    let rep = get_rep cur in
    setFlag rep cur.flags ;
    List.iter check_compat_fun cur.succ ;
    List.iter check_compat_fun cur.pred ;
  ) node_ht ;

  (* Check the Special Scalar Node 
   * All nodes connected to the special scalar node must be in the same EQ
   * class and their base type must all be "void *". *)
  let scalar_ok = ref true in

  List.iter (fun e -> 
    if not (isVoidType e.eto.btype) || 
       not (isVoidType e.efrom.btype) then
      scalar_ok := false
  ) (scalarNode.succ @ scalarNode.pred) ;

  if not !scalar_ok then begin
    List.iter (fun e -> update e.eto Wild (Incompat(e))) scalarNode.succ ;
    List.iter (fun e -> update e.efrom Wild (Incompat(e))) scalarNode.pred ;
  end ;

  (* Now we have equivalence classes of ECompat-joined nodes *)
  let compat_eq_classes = NodeUF.eq_classes !node_eq in 

  (* 
   * Step 2 Loop #2 : examine each "void *" equiv class 
   *)

  (* share all flags within equivalence classes *)
  List.iter (fun eq_class ->
    List.iter (fun from_elt ->
      List.iter (fun to_elt ->
        setFlag to_elt from_elt.flags
      ) eq_class
    ) eq_class
  ) compat_eq_classes ;

  (* Step 3
   * ~~~~~~
   * Our second pass over the set of nodes. 
   * Set all of the flag starting conditions that we know about.  
   *)
  if !E.verboseFlag then ignore (E.log "Solver: Step 3  (Base Case)\n") ;

  (* loop over all the nodes ... *)
  Hashtbl.iter (fun id n -> 
    (* Mark interface character pointers as strings. *)
    if interface_char_ptr_is_ROSTRING then begin
      if hasFlag n pkInterface && is_char_pointer n then begin
        (* the user had something to say here *)
        if (set_outside n) then begin
          match n.kind with
            String | ROString -> ()
          | _ -> ignore (E.warn 
                           "Solver: %a annotation on interface (char *)@!%a" 
                           d_opointerkind n.kind d_node n)
        end else 
          update n String BoolFlag 
      end
    end ;

    (* calling a function without a prototype makes it wild *)
    if hasFlag n pkNoPrototype then begin
        update n Wild BoolFlag
    end ; 

    (* Set initial flags based on the pointer kind *)
    begin
    match n.kind with
    | String
    | ROString -> setFlag n pkReachString

    | Seq
    | Seq      -> setFlag n pkReachSeq

    | FSeqN
    | SeqN     -> setFlag n pkReachString ;
                  setFlag n pkReachSeq

    | Index    -> setFlag n pkReachIndex

    | _        -> () 
    end ;

  ) node_ht ; 

  (* Step 4
   * ~~~~~~
   * Push all of the boolean flags around. 
   *)
  if !E.verboseFlag then ignore (E.log "Solver: Step 4  (Data-Flow)\n") ;
  (* loop over all the nodes ... *)
  finished := false ; 
  Stats.time "data-flow" (fun () ->
  while (not !finished) do 
    finished := true ; 
    Hashtbl.iter (fun id cur -> 
      (* First consider all ECompat edges:
       * flags should be equal across them. This is motivated by
       * test/small1/apachebuf.c. Merely making ECompat-linked nodes have
       * the same kind does not suffice: a pred of an ecompat-linked node
       * may need to be made FSEQ because of a posarith on the other side
       * of the ecompat link. *)
      let inner_fun e = 
        if isECompat e then begin
          let old_flags = e.efrom.flags in setFlag e.efrom e.eto.flags ;
          finished := !finished && (old_flags = e.efrom.flags) ;
          let old_flags = e.eto.flags in setFlag e.eto e.efrom.flags ;
          finished := !finished && (old_flags = e.eto.flags)
        end
      in 
      List.iter inner_fun cur.pred ;
      List.iter inner_fun cur.succ ;

      (* Consider all Successor Edges, do data-flow *)
      List.iter (fun e ->
        let old_flags = e.eto.flags in 
        (match e.ekind with
           ECast -> setFlag e.eto (cur.flags land pkCastSuccFlags) ;
         | ESafe -> setFlag e.eto (cur.flags land pkSafeSuccFlags) ;
         | _ -> ()) ;
        finished := !finished && (old_flags = e.eto.flags)
      ) cur.succ ;

      (* Consider all Predecessor Edges, do data-flow *)
      List.iter (fun e ->
        let old_flags = e.efrom.flags in 
        (match e.ekind with
           ECast ->  setFlag e.efrom (cur.flags land pkCastPredFlags) ;
                     setFlag e.efrom (cur.flags land pkCNIPredFlags) ;
         | ENull 
         | EIndex -> setFlag e.efrom (cur.flags land pkCNIPredFlags) ;
         | _ -> ()) ;
        finished := !finished && (old_flags = e.efrom.flags)
      ) cur.pred ;
        
    ) node_ht ; 
  done ) () ; 

  (* Step 5
   * ~~~~~~
   * Distinguish between sequences. We must do this after boolean flags
   * (otherwise we cannot tell what reaches what, etc.) but before we do
   * WILDs (because they interact with read-only strings). 
   *
   * Also generate ARITH constraints: q != SAFE. 
   *)
  if !E.verboseFlag then ignore (E.log "Solver: Step 5  (sequences)\n") ;

  (* n is a polymorphic "void *" node if it points to void* and its
   * representative points to void* as well. *)
  let is_polymorphic_voidstar n =
    match n.btype, (get_rep n).btype with
      TVoid _ , TVoid _->  true
    | _ -> false
  in 

  Hashtbl.iter (fun id cur -> 
    (* Generate "ARITH" style constraints: q != SAFE *)
    if     hasFlag cur pkArith 
        || hasFlag cur pkPosArith 
        || hasFlag cur pkIntCast then begin 
        setFlag cur pkNotSafe ;
    end ;

    if interface_char_ptr_is_ROSTRING && hasFlag cur pkInterface && 
       is_char_pointer cur && cur.kind = Unknown then begin
        update cur String BoolFlag 
    end ;

    if cur.kind <> Wild && 
        (hasFlag cur pkNotSafe || 
         hasFlag cur pkReachString ||
         hasFlag cur pkReachSeq || 
         hasFlag cur pkReachIndex || 
         hasFlag cur pkIntCast) then 
    begin
      let new_kind = 
        if     (hasFlag cur pkReachIndex)   then Index
        else if 
               (hasFlag cur pkReachString)  &&
           not (hasFlag cur pkReachSeq)     &&
           not (hasFlag cur pkIntCast)      &&
           not (hasFlag cur pkArith)        &&
           not (hasFlag cur pkPosArith)     &&
           not (cur.kind = ROString)        then String
        else if 
               (hasFlag cur pkReachString)  &&
           not (hasFlag cur pkReachSeq)     &&
           not (hasFlag cur pkIntCast)      &&
           not (hasFlag cur pkArith)        &&
           not (hasFlag cur pkPosArith)     &&
               (cur.kind = ROString)        then ROString
        else if 
               (hasFlag cur pkReachString)  &&
               (hasFlag cur pkArith)        then SeqN
        else if 
           not (hasFlag cur pkReachString)  &&
               (hasFlag cur pkArith)        then Seq
        else if 
               (hasFlag cur pkReachString)  then FSeqN

        else if
               (is_polymorphic_voidstar cur)&&
               (hasFlag cur pkIntCast)      &&
           not (hasFlag cur pkArith)        &&
           not (hasFlag cur pkPosArith)     then Safe

        else                                     FSeq
      in 
      update cur new_kind BoolFlag 
    end
  ) node_ht ; 

  (* Step 6
   * ~~~~~~
   * Determine which Strings/FSeqNs should be ROStrings. 
   *
   * Note that nodes connected by ECompat edges must all agree. 
   *)
  if !E.verboseFlag then ignore (E.log "Solver: Step 6  (ROString)\n") ;

  let any_wilds_in_graph = ref false in 
  (* if not, we can skip step 7 *)

  (* helper function to determine if a node is ready to become ROString *)
  let ready_for_rostring cur = (* modulo edges *)
    (cur.kind = FSeqN || cur.kind = Safe || cur.kind = String) &&
    not (hasFlag cur pkUpdated) &&
    not (hasFlag cur pkIntCast) &&
    not (hasFlag cur pkArith) &&
    List.length cur.succ > 0 &&
    List.fold_left (fun acc e ->     
      acc && ( (isECompat e) ||
               (e.ekind = ECast && e.eto.kind = ROString))) true cur.succ 
  in 
  finished := false ; 
  while not !finished do 
    finished := true ; 
    (* 
     * Consider every node in the graph in two phases.
     *  (1) Consider all nodes w/o ECompat edges
     *  (2) Consider all Eq-Classes of ECompat nodes
     *)
    Hashtbl.iter (fun id cur -> 
      if cur.kind = Wild then any_wilds_in_graph := true ; 
      let any_ecompat = 
        List.fold_left (fun acc e -> acc || isECompat e)
          false (cur.succ @ cur.pred) in
      if not any_ecompat && 
         ready_for_rostring cur 
         then begin
          update cur ROString (SpreadFromEdge ((List.hd cur.succ).eto)) ;
      end
    ) node_ht ;

    (* now look at an equivalence class of ECompat nodes *)
    List.iter (fun ecompat_node_list ->
      (* if every node in this list is ready, move every node in this
       * list to ROString *)
      let all_are_ready = 
        List.fold_left (fun acc cur -> acc &&
           ready_for_rostring cur && 
           (cur.why_kind <> UserSpec || cur.kind = ROString)) 
           true ecompat_node_list 
      in
      if all_are_ready then 
        List.iter (fun cur -> 
          update cur ROString (SpreadFromEdge ((List.hd cur.succ).eto)) ;
        ) ecompat_node_list 
    ) compat_eq_classes 
  done ;

  (* Step 7
   * ~~~~~~
   * Spread WILD as far as it will go. 
   *)
  if !E.verboseFlag then ignore (E.log "Solver: Step 7  (Wild)\n") ;
  if (!any_wilds_in_graph) then begin 
    finished := false ; 
    while not !finished do
      finished := true ; 
      Hashtbl.iter (fun id cur -> 
        if cur.kind = Wild then begin
          (* consider all of the successor edges *)
          List.iter (fun e ->
            if e.eto.kind <> ROString then 
              update e.eto Wild (SpreadFromEdge cur)
          ) cur.succ ;

          (* consider all of the predecessor edges *)
          List.iter (fun e ->
            if e.efrom.kind <> ROString (* && e.ekind <> ESafe *) then 
              update e.efrom Wild (SpreadFromEdge cur)
          ) cur.pred ;

          (* handle points-to information *)
          List.iter (fun n ->
            update n Wild (SpreadPointsTo cur) ;
          ) cur.pointsto ;

        end
      ) node_ht ;
    done ;
  end ;

  (* Step 8
   * ~~~~~~~
   * All other nodes are safe. 
   *)
  if !E.verboseFlag then ignore (E.log "Solver: Step 8  (Safe)\n") ;
  let ecompat_warnings = false in 
  Hashtbl.iter (fun id n -> 
    if n.kind = Unknown then begin
      update n Safe Unconstrained
    end ;

    (* Sanity Check! Typecheck.ml does much more than this. *)
    if n.kind = Safe &&
      (hasFlag n pkNotSafe || 
       hasFlag n pkReachString ||
       hasFlag n pkReachSeq || 
       hasFlag n pkReachIndex || 
       hasFlag n pkIntCast) && 
      not (n.why_kind = UserSpec) &&
      not (is_polymorphic_voidstar n) then begin
      E.s (E.bug "Solver: botched node (left/made it safe)@!%a"
        d_node n) ; E.hadErrors := true 
    end ;
  ) node_ht ;

  !we_should_typecheck
end 

(* 
 * TODO (The Wishlist of Wes): 
 * (1) change markptr so that ENULL is never used, use ECAST instead
 * (2) shrink the graph instead of using ECompat edges
 *)
