(*
 *
 * Copyright (c) 2001 by
 *  George C. Necula	necula@cs.berkeley.edu
 *  Scott McPeak        smcpeak@cs.berkeley.edu
 *  Wes Weimer          weimer@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products 
 * derived from  this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

(*
 * C Structural Type Equality
 *)
open Cil
open Clist
open Pretty
module E = Errormsg

(* Idealized representation of C types to make structural equality
 * comparisons easier. *)
type layout =     
    Scalar of int   (* length in bits *)
  | Array of (layout list) * int
  | Union of ((layout list) list) * int (* total union size in bits *)
  | Pointer of typ (* original CIL pointer type
                    * either TPtr or TFun *)
  | Anything of int (* length in bits, matches anything *)

(* Pretty printing *)
let rec d_layout () l = match l with
    Scalar(i) -> dprintf "Scalar(%d)" i 
  | Anything(i) -> dprintf "Anything(%d)" i 
  | Array(ll,i) -> dprintf "Array %d %a" i d_layout_list ll
  | Union(lll,i) -> dprintf "Union @[(%a)@]"
      (docList (chr ',' ++ break) (d_layout_list ())) lll
  | Pointer(tau) -> dprintf "Ptr (%a)" d_type tau
and d_layout_list () ll = 
  dprintf "@[[%a]@]" (docList (chr ';' ++ break) (d_layout ())) ll

let global_layout_list_cache = Hashtbl.create 2047

(* peephole optimization for layout lists *)
let rec peephole ll = 
  match ll with
    [] -> []
  | Scalar(i) :: Scalar(j) :: tl -> peephole (Scalar(i+j) :: tl)
  | hd :: tl -> hd :: (peephole tl)

(* is a layout list a complete list of scalars? *)
let rec list_of_scalars ll =
  match ll with
    [] -> true
  | Scalar(i) :: tl -> list_of_scalars tl
  | _ -> false
  (* We don't want a match case like this: 
   * | Anything _ :: tl -> list_of_scalars tl
   * because it throws away information. "Anything" is "more powerful"
   * than a list of scalars in the matching world . *)
    
(* Convert a C/CIL type into a list of layouts. This usually amounts
 * to flattening structures and making padding and alignment explicit. *)
let rec convert_type_to_layout_list (orig_t : typ) =
  if Hashtbl.mem global_layout_list_cache orig_t then
    Hashtbl.find global_layout_list_cache orig_t 
  else 
  let t = unrollType orig_t in 
  let result = 
  match t with
    TVoid _ -> (* failwith "convert_type_to_layout_list: void" *) [] 
  | TInt _ | TFloat _ | TEnum _ -> [Scalar((bitsSizeOf t))]
  | TPtr(_) -> [Pointer(t)]
  | TFun(_) -> [Pointer(t)]
  | TArray(tau,Some(e),_) when e = one -> convert_type_to_layout_list tau 
  | TArray(tau,Some(e),_) -> 
      let ll = convert_type_to_layout_list tau in
      let e = constFold true e in 
      let len = begin match isInteger e with
        Some(i64) -> Int64.to_int i64
      | None -> 
          E.s (E.unimp 
            "type: convert_type_to_layout_list: non-const length %a" 
            d_type t)
      end in
      ( match ll with
          [Scalar(length)] -> [Scalar(length * len)]
        | _ -> [Array(ll,len)])
  | TArray(tau,None,_) -> 
      E.s (E.unimp "type: convert_type_to_layout_list: empty array %a"
            d_type t)
  | TComp(ci,al) when ci.cstruct ->
      let total_size = (bitsSizeOf t) in
      let seen_size = ref 0 in 
      let ll = ref [] in 
      List.iter (fun fi -> 
        let field_size_bits = 
          match fi.fbitfield with 
            Some(size) -> size
          | None -> bitsSizeOf fi.ftype 
        in 
        let field_offset_bits, field_width_bits = 
          bitsOffset t (Field(fi,NoOffset)) in
        let field_offset = field_offset_bits in 
        if (!seen_size < field_offset) then begin
          ll := !ll @ [Scalar(field_offset - !seen_size)];
          seen_size := field_offset 
        end ;
        let field_ll = match fi.fbitfield with
          Some(size) -> [Scalar(size)]
        | None -> convert_type_to_layout_list fi.ftype 
        in
        ll := !ll @ field_ll ;
        seen_size := !seen_size + field_size_bits ; 
      ) ci.cfields ;
      if (!seen_size < total_size) then begin
        ll := !ll @ [Scalar(total_size - !seen_size)];
      end ;
      !ll
  | TComp(ci,al) when not ci.cstruct -> 
      let total_size = (bitsSizeOf t) in
      let lll = List.map (fun fi ->
        let this_size = (bitsSizeOf fi.ftype) in
        let ll = convert_type_to_layout_list fi.ftype in
        if this_size < total_size then
          ll @ [Anything(total_size - this_size)]
        else
          ll 
      ) ci.cfields in
      (* determine if all disjuncts are all scalars *)
      begin
      match List.fold_left (fun acc elt -> acc && 
        list_of_scalars elt) true lll with
        true ->  [Scalar(total_size)]
      | false -> [Union(lll,total_size)]
      end
  | TComp _ -> E.s (E.bug "type: convert_type_to_layout_list: mystery comp")
  | TNamed _ -> E.s (E.bug "type: convert_type_to_layout_list: named")
  | TBuiltin_va_list _ -> 
      E.s (E.bug "type: convert_type_to_layout_list: va_list")
  in
  let result = peephole result in 
  Hashtbl.add global_layout_list_cache t result ;
  result
  
(* Strips 'i' bits of scalars from the beginning of the layout list 'll'.
 * Returns a boolean success code and new layout lest representing the
 * suffix or raises an exception on failure. *)
let rec strip_scalar_prefix (ll : layout list) i =
  if i = 0 then 
    true, ll
  else if i < 0 then 
    true, (Scalar(-i) :: ll)
  else match ll with
    [] -> false, [] 
  | Scalar(j) :: tl -> (strip_scalar_prefix tl (i - j))
  | Anything(j) :: tl -> 
      if (i >= j) then (* eat up the entire "Anything" *)
        (strip_scalar_prefix tl (i - j))
      else
        true, (Anything(j - i) :: tl)
  | Array(inner_ll,0) :: tl -> (strip_scalar_prefix tl i)
  | Array(inner_ll,1) :: tl -> (strip_scalar_prefix (inner_ll @ tl) i)
  | Array(inner_ll,j) :: tl -> 
      let new_ll = inner_ll @ (Array(inner_ll,j-1) :: tl) in
      (strip_scalar_prefix new_ll i )
  | Union(lll,ts) :: (tl : layout list) ->
      if ts > i then begin (* strip just part of the union *)
        let final_answer = ref true in 
        let lll = List.map (fun ll -> 
          let answer, new_ll = strip_scalar_prefix ll i in
          if answer = false then final_answer := false ;
          new_ll) lll in
        !final_answer, (Union(lll,(ts - i)) :: tl )
      end else begin (* strip the whole union away! *)
        let final_answer = ref true in 
        List.iter (fun ll -> let answer, _ = (strip_scalar_prefix ll ts) in
          if answer = false then final_answer := false
          ) lll ;
        let answer, result = strip_scalar_prefix tl (i - ts) in
        (!final_answer && answer),result
      end
  | Pointer(_) :: tl -> 
      let _, res = strip_scalar_prefix (Scalar(4*8) :: tl) i in
      false, res

(* Physical/Structural Equivalence Classes of type *)
module OrderedType = struct type t = typ let compare = compare end
module TypeSet = Set.Make(OrderedType)
module TypeUF = Unionfind.Make(TypeSet)
(* INVARIANT: always put unrolled types in here *)

(* a global set of type equivalence classes *)
let global_eq = ref (TypeUF.empty)
let global_subtype = Hashtbl.create 511

(* When we are computing "t1 < t2" or "t1 > t2", which one can be
 * smaller? *)
type subtype_direction = LeftCanEndEarly | RightCanEndEarly | MustBeEqual

let flip_subtyping st = match st with
    LeftCanEndEarly -> RightCanEndEarly 
  | RightCanEndEarly -> LeftCanEndEarly 
  | MustBeEqual -> MustBeEqual 

(*
 * Check to see if a type is made up entirely of scalars. 
 *)
let all_scalars
  ?(replace=(fun a -> a,empty)) 
  t1
= 
  let t1,ex1 = replace (unrollType t1) in
  try
    let ll = convert_type_to_layout_list t1 in
    let answer = match ll with
      [Scalar(_)] -> true
    | _ -> false
    in answer
  with _ -> false


(* Check to see if two types are physically/structurally equal. This
 * is the function that other modules should call to check equality. 
 *
 * NOTE that this returns TRUE in the case where the top-level types are
 * equal assuming that all types that "failure" was called on (and all
 * types below those) are WILD (or otherwise not considered). 
 *
 * failure : type -> unit
 *   is called on INNER POINTER types enclosing equality failiures
 * compat : edge Clist -> type -> type -> unit
 *   is called on INNNER POINTER types that must be equal for these types
 *   to be equal
 * replace : type -> type * (edge clist)
 *   is called on every type that is considered, giving you a chance to
 *   replace it with some other type and explain why that was done
 *)
let rec equal 
  ?(compat=(fun _ _ _ -> ())) 
  ?(failure=(fun _ -> ())) 
  ?(replace=(fun a -> a,empty)) 
  ?(explanation=empty)
  t1 t2 
=
  let t1,ex1 = replace (unrollType t1) in
  let t2,ex2 = replace (unrollType t2) in
  let explanation = append explanation ( append ex1 ex2 ) in 
  (* Special Polymorphism Check: "void *" is really a type variable,
   * so whenever we are tempted to check its equality with something
   * we instead unify those things *)
  if (isVoidPtrType t1 || isVoidPtrType t2) then begin
    compat explanation t1 t2 ;
    true 
  end else begin 
    let s1 = try bitsSizeOf t1 with _ -> -1 in
    let s2 = try bitsSizeOf t2 with _ -> -1 in
    if s1 <> s2 then begin
      false
    end else if TypeUF.check_equal !global_eq t1 t2 then
      true
    else begin
      try 
        let answer, new_eq_classes = 
          equal_types t1 t2 !global_eq compat failure replace explanation
        in
        let answer = answer || (t1 = t2) in 
        if answer then begin 
          global_eq := new_eq_classes ; true 
        end else begin
          false
        end
      with e -> (t1 = t2) (* fall back on ML equality *) 
    end
  end
(* 
 * Is small a physical/structural subtype of big? Only the top-level layout
 * can actually be a subtype. Must be invariant under pointers. 
 *
 * NOTE that this returns TRUE in the case where the top-level types are
 * subtypes assuming that all types that "failure" was called on (and all
 * types below those) are WILD (or otherwise not considered). 
 *)
and subtype 
  ?(compat=(fun _ _ _ -> ())) 
  ?(failure=(fun _ -> ())) 
  ?(replace=(fun a -> a,(empty))) 
  ?(explanation=(empty))
  small' big' 
=
  (* ignore (E.warn "subtype: %a %a@!" d_type small' d_type big') ;
  flush stderr ;  *)
  let small,ex1 = replace (unrollType small') in
  let big,ex2 = replace (unrollType big') in
  let explanation = append explanation (append ex1 ex2) in 
  if (isVoidPtrType small || isVoidPtrType big) then begin
    compat explanation small big ;
    true
  end else if small == big then 
    true
  else begin
  let s1 = try bitsSizeOf small with _ -> -1 in
  let s2 = try bitsSizeOf big with _ -> -1 in
  if s1 > s2 then begin
    (* Pretty.printf "subtype: %a %a failed because %d > %d\n"
      d_type small d_type big s1 s2 ;   *)
    false
  end else if Hashtbl.mem global_subtype (small,big) || 
              Hashtbl.mem global_subtype (big,small) then 
    true
  else begin
    try 
      let l1 = convert_type_to_layout_list small in
      let l2 = convert_type_to_layout_list big in
      if l1 = l2 then begin
        true
      end else begin
        let answer, new_eq_classes = 
          equal_ll l1 l2 !global_eq LeftCanEndEarly compat failure replace 
          explanation in 
        if answer then begin
          global_eq := new_eq_classes ;
          Hashtbl.add global_subtype (small,big) true ;
          true
        end else begin
          false
        (* Pretty.printf "subtype: %a %a failed because@!--> %a@!--> %a@!"
            d_type small d_type big 
            d_layout_list l1 d_layout_list l2 ;  *)
        end ; 
      end
    with e -> (big = small) 
  end
  end
(* is the given union internally consistent? *)
and check_union
    tau1
    compat 
    failure 
    replace 
    explanation
=
  let tau1,ex1 = replace (unrollType tau1) in
  let explanation = append explanation ex1 in 
  let ll = convert_type_to_layout_list tau1 in 
  try 
    match ll with
      [Union(lll,sz)] -> 
        let lla = Array.of_list lll in
        let lla_len = Array.length lla in 
        (* I think this really has to be O(n^2) *) 
        for i = 0 to lla_len - 1 do
          for j = (i+1) to lla_len - 1 do
            let answer,nec = equal_ll lla.(i) lla.(j) !global_eq 
              (MustBeEqual) compat failure replace explanation in
            global_eq := nec ;
            if not answer then failwith "unsafe union"
          done
        done ; 
        true
    | _ -> true (* not a union, so OK! *)
  with Failure(_) -> false
and equal_types 
    tau1 tau2 
    (assumed_eq_classes : TypeUF.t) 
    compat 
    failure 
    replace 
    explanation
=
  (* ignore (E.warn "equal_types: %a %a@!" d_type tau1 d_type tau2) ;
  flush stderr ;  *)
  let tau1,ex1 = replace (unrollType tau1) in
  let tau2,ex2 = replace (unrollType tau2) in
  let explanation = append explanation (append ex1 ex2 ) in 
  if (isVoidPtrType tau1 || isVoidPtrType tau2) then begin
    compat explanation tau1 tau2 ;
    true, assumed_eq_classes
  end else if tau1 == tau2 || 
              TypeUF.check_equal assumed_eq_classes tau1 tau2 then 
    true, assumed_eq_classes
  else 
    let l1 = convert_type_to_layout_list tau1 in
    let l2 = convert_type_to_layout_list tau2 in
    if l1 = l2 then 
      true, assumed_eq_classes 
    else 
    let new_eq_classes = TypeUF.make_equal assumed_eq_classes tau1 tau2  in
    equal_ll l1 l2 (new_eq_classes : TypeUF.t) 
      MustBeEqual compat failure replace explanation 
(* 
 * arguments: 
 *   l1 l2     // check and see if these two layout lists are equal
 *   aec       // assumed equivalence classes 
 *   subtyping // if this is true, l1 can be shorter than l2
 *   compat    // call this function on all lined-up pointers in l1, l2
 *   failure   // call on pointer types that fail to match up 
 *)
and ok_to_call_compat t1 t2 = isVoidPtrType t1 || isVoidPtrType t2 ||
     (isPointerType t1 && isPointerType t2)
and equal_ll 
  l1 l2 
  (aec : TypeUF.t) 
  subtyping 
  compat 
  failure 
  replace 
  explanation
=
(*
  ignore (E.warn "equal_ll:@!%a@!%a@!" d_layout_list l1 d_layout_list l2) ;
  flush stderr ; 
  *)

  let final_answer, aec = 
  match l1, l2 with
    [], [] -> (true,aec)
  | [], _ when subtyping = LeftCanEndEarly -> (true, aec)
  | _, [] when subtyping = RightCanEndEarly -> (true, aec)

  | [], _ 
  | _, [] -> (false, aec)

  | (Scalar(i) :: tl), _ -> 
      let new_lhs = tl in
      let worked, new_rhs = strip_scalar_prefix l2 i in
      let answer, nec = 
        equal_ll new_lhs new_rhs aec subtyping compat failure replace
        explanation in
      (answer && worked), nec

  | (Anything(i) :: tl), _ -> 
      let new_lhs = tl in
      let _, new_rhs = strip_scalar_prefix l2 i in
      equal_ll new_lhs new_rhs aec subtyping compat failure replace
      explanation

  (* array on left *)
  | (Array(inner_ll1,i1) :: tl1) , _ ->
      let new_lhs = 
        if i1 = 0 then
          tl1 
        else if i1 = 1 then 
          inner_ll1 @ tl1 
        else 
          inner_ll1 @ (Array(inner_ll1,i1-1) :: tl1)
      in 
      equal_ll new_lhs l2 aec subtyping compat failure replace explanation

  (* union on left *)
  | (Union(lll,ts) :: tl1) , _ -> 
      let assumes = ref aec in
      let union_answer = List.fold_left (fun acc elt -> 
        if not acc then false else begin
          let answer, nec = 
            equal_ll (elt @ tl1) l2 !assumes subtyping 
              compat failure replace explanation in
          assumes := nec ;
          answer
        end) true lll in
      union_answer, !assumes

  (* array/union/anything on right *)
  | _, (Array(_) :: _)
  | _, (Union(_) :: _) 
  | _, (Anything(_) :: _) -> 
      (* flip things around, use the code above *)
      equal_ll l2 l1 aec (flip_subtyping subtyping) compat failure replace 
        explanation

  | (Pointer(tau1) :: tl1) , (Pointer(tau2) :: tl2) -> 
      (* already unrolled *)
      let tau1,ex1 = replace (tau1) in
      let tau2,ex2 = replace (tau2) in
      let explanation = append explanation (append ex1 ex2 ) in 
      if ok_to_call_compat tau1 tau2 then compat explanation tau1 tau2 ; 
      let result_eq_class = 
        if isVoidPtrType tau1 || isVoidPtrType tau2 then begin
          aec
        end else if tau1 = tau2 || TypeUF.check_equal aec tau1 tau2 then begin
          aec
        end else begin
          (* first, check under the pointers *)
          match tau1, tau2 with
        | TPtr(TFun(t1,vlo1,_,_),_), TPtr(TFun(t2,vlo2,_,_),_)  
        |      TFun(t1,vlo1,_,_),         TFun(t2,vlo2,_,_) -> 
          (* two fun ptrs *)
            let aec = ref aec in 
            if ok_to_call_compat t1 t2 then compat explanation t1 t2 ; 
            let ret_ok, nec1 = equal_types t1 t2 !aec 
              compat failure replace explanation in
            aec := nec1 ; 
            let al1 = argsToList vlo1 in
            let al2 = argsToList vlo2 in
            let args_ok = 
              if List.length al1 <> List.length al2 then begin 
                false 
              end else begin
                (* check every argument *)
                let answer = ref true in
                List.iter2 (fun arg1 arg2 ->
                  if ok_to_call_compat arg1.vtype arg2.vtype then 
                    compat explanation arg1.vtype arg2.vtype ; 
                  let ans, nec2 = equal_types arg1.vtype arg2.vtype 
                    !aec compat failure replace explanation in
                  aec := nec2 ;
                  answer := ans && !answer ;
                ) al1 al2 ; 
                !answer
              end 
            in 
            if (not args_ok) || (not ret_ok) then begin
              failure tau1 ;
              failure tau2
            end ;
            !aec

        | TFun _ , _  (* function pointer and non-function ptr *)
        | _, TFun _ -> begin
                          failure tau1 ;
                          failure tau2 ;
                          aec 
                        end
              
        | TPtr(inner1,_), TPtr(inner2,_) -> begin (* two non-fun pointers *)
            let answer, nec = 
                equal_types inner1 inner2 aec compat failure replace
                explanation 
            in
            if not answer then begin
              failure tau1 ;
              failure tau2 ;
            end ;
            nec
        end

        | _, _ -> E.s (E.bug "type: unexpected mystery pointers %a and %a"
          d_type tau1 d_type tau2) 
      end 
      in (* now check the rest of the layout list! *)
      equal_ll tl1 tl2 result_eq_class subtyping compat failure replace 
        explanation 

  | (Pointer(p) :: tl1) , something 
  | something , (Pointer(p) :: tl1) when isVoidType p -> 
    (* special scalar node handling *)
    E.s (E.bug "type: unify Ptr(%a) with the scalar node" d_type p)

  | (_ :: tl1), (_ :: tl2) -> (* NOT EQUAL *)
      (* but check the rest anyway! *)
      let answer, new_aec = 
        equal_ll tl1 tl2 aec subtyping compat failure replace explanation in
      (false, new_aec) 
  in
  (* ignore (E.warn "subtype: %b@!%a@!%a"
    final_answer d_layout_list l1 d_layout_list l2) ;*)
  (final_answer, aec)

(* Debugging Information *)
let global_eq_classes () = 
  TypeUF.eq_classes !global_eq

