(*
 *
 * Copyright (c) 2001 by
 *  George C. Necula	necula@cs.berkeley.edu
 *  Scott McPeak        smcpeak@cs.berkeley.edu
 *  Wes Weimer          weimer@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products 
 * derived from  this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

(*
 * Union-Find Functor (no side effects, purely applicative)
 *
 * (used by physical type equality to keep equivalence classes of types)
 *) 

module type UF = 
  sig
    type t
    type elt 
    val empty : t
    val check_equal : t -> elt -> elt -> bool 
    val make_equal : t -> elt -> elt -> t
    val eq_classes : t -> elt list list
    val class_of : t -> elt -> elt list
  end

module Make(S : Set.S) =
  struct
    type t = S.t list
    type elt = S.elt 
    let empty = []
    let check_equal uf e1 e2 =
      List.fold_left (fun acc elt -> acc ||
        (S.mem e1 elt && S.mem e2 elt)) false uf
    let make_equal (uf : S.t list) e1 e2 =
      let uninvolved = ref [] in
      let s1 = ref (S.singleton e1) in
      let s2 = ref (S.singleton e2) in 
      List.iter (fun (s : S.t) ->
        if S.mem e1 s then 
          s1 := s
        else if S.mem e2 s then
          s2 := s
        else 
          uninvolved := s :: !uninvolved
      ) uf ;
      let merged_set = S.union !s1 !s2 in
      let final_list_of_sets = merged_set :: !uninvolved in
      (final_list_of_sets : S.t list)
    let eq_classes uf = 
      List.map (fun eqclass -> S.elements eqclass) uf
    let class_of uf elt =
      let rec search ll = match ll with 
        [] -> []
      | set :: tl -> if (S.mem elt set) 
                      then (S.elements set)
                      else search tl 
      in search uf
  end


    
  
