(*
 *
 * Copyright (c) 2001 by
 *  George C. Necula	necula@cs.berkeley.edu
 *  Scott McPeak        smcpeak@cs.berkeley.edu
 *  Wes Weimer          weimer@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products 
 * derived from  this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

open Cil
open Pretty

module H = Hashtbl
module E = Errormsg

module MU = Markutil

module N = Ptrnode

let debugVararg = false

(* Check that s starts with prefix p *)
let prefix p s = 
  let lp = String.length p in
  let ls = String.length s in
  lp <= ls && String.sub s 0 lp = p


(*********************** VARARG ********************************)

(* Once we see the definition of the printf format union we will remember it 
 * here *)
let printfFormatUnion : typ =
  TComp(mkCompInfo false
          "printf_format_union"
          (fun _ -> 
            [ ("f_int", intType, None, []);
              ("f_double", doubleType, None, []);
              ("f_string", TPtr(charType, [N.k2attr N.ROString]), None, []);
              ("f_longlong", TInt(ILongLong, []), None, []) ])
          [],
        [])
  

let markPrintfFunction (s: string) (format_idx: int) = 
  if !E.verboseFlag then 
    ignore (E.log "Will treat %s as a printf-like function\n" s);
  MU.applyToFunction s 
    (fun vi -> 
      MU.addFunctionTypeAttribute 
        (Attr("boxvararg", [ASizeOf printfFormatUnion])) vi;
      MU.addFunctionTypeAttribute 
        (Attr("boxformat", [AInt format_idx])) vi)

(* First process the pragmas *)
let processPragma (a: attribute) : unit = 
  match a with 
  | Attr("boxvararg", [AStr s; ASizeOf t]) -> 
(*
          if H.mem calledFunctions s then 
            ignore (warn "#pragma boxvararg appears after call to %s\n"
                      s);
*)
      if debugVararg then 
        if !E.verboseFlag then 
          ignore (E.log "Will treat %s as a vararg function\n" s);
      MU.applyToFunction s 
        (fun fvi -> 
          if not (hasAttribute "boxvararg" 
                    (MU.getFunctionTypeAttributes (Lval(Var fvi, NoOffset)))) 
          then 
            MU.addFunctionTypeAttribute 
              (Attr("boxvararg", [ASizeOf t]))
              fvi)
        

  | Attr("boxvararg_printf", [AStr s; AInt format_idx]) -> 
(*
          if H.mem calledFunctions s then 
            ignore (warn "#pragma boxvararg_printf appears after call to %s\n"
                      s);
*)
      markPrintfFunction s format_idx


  | _ -> ()


(* Turn the gcc format attribute into our own notation *)
let processFormatAttribute (fvi: varinfo) = 
  match filterAttributes "format" 
                          (MU.getFunctionTypeAttributes (Lval (var fvi))) with
    [Attr(_, [printf; AInt format_idx; AInt _]) as a] 
      when (match printf with 
                ACons("printf",[]) -> true 
              | _ -> false) -> 
      markPrintfFunction fvi.vname format_idx

  | [] -> ()
  | al -> ignore (warn "Do not understand %a" d_attrlist al)


(* Compute the signature of a type for comparing arguments *)
let argumentTypeSig (t: typ) = 
  let argumentPromotion (t : typ) : typ =
    match unrollType t with
      (* We assume that an IInt can hold even an IUShort *)
      TInt ((IShort|IUShort|IChar|ISChar|IUChar|ILong|IULong|IUInt), a) -> 
        TInt(IInt, a)
    | TInt _ -> t
          (* floats are passed as double *)
    | TFloat (FFloat, a) -> TFloat (FDouble, a)
    | TEnum (_, a) -> TInt(IInt, a)
    | t -> t
  in
  (* Ignore all attributes *)
  typeSigWithAttrs (fun x -> []) (argumentPromotion t)

type vaKind  = int (* An index starting from 0 *) 
             * string (* A name of the kind *)
             * typ (* The type of the kind *)
             * typsig (* The pre-computed type signature *)

(* Get the descriptor type for a va_list variable. May raise Not_found *)
let getValistDescriptorType (dv:varinfo) = 
  match unrollType dv.vtype with 
    TPtr(TComp(ci, _), pa) when prefix "__ccured_va_list" ci.cname -> begin
      match filterAttributes "boxvararg" pa with 
        Attr(_, [ASizeOf t]) :: _ -> t
      | _ -> raise Not_found
    end
  | _ -> raise Not_found
         

let getVarargFunctionDescriptorType (fexp: exp) = 
  match unrollType (typeOf fexp) with
    TFun(_, _, _, a) -> begin
      match filterAttributes "boxvararg" a with
        Attr(_, [ASizeOf t]) :: _ -> t
      | _ -> raise Not_found
    end
  | _ -> raise Not_found


(* Keep track for each function whether it is a vararg, and the 
 * alternatives, with an index, a field name, a type and a type signature. *)
let getVarargKinds (descrt: typ) : vaKind list = 
  let kinds = 
    match unrollType descrt with
      TComp (ci, _) when not ci.cstruct -> 
        let rec loop (idx: int) = function
            [] -> []
          | fi :: restfi -> 
              (idx, fi.fname, fi.ftype, argumentTypeSig fi.ftype) 
              :: loop (idx + 1) restfi
        in
        loop 0 ci.cfields
    | t' -> [(0, "anon", t', argumentTypeSig t')]
  in
  (* Make sure that no two types are compatible *)
  let _ = 
    List.fold_left
      (fun prev ((_, thisn, _, thiss) as this) -> 
        List.iter
          (fun (_, pn, pt, ps) -> 
            if thiss = ps then 
              E.s (error "Vararg type %a has compatible fields %s and %s\n"
                     d_type descrt pn thisn)) prev;
        this :: prev)
      []
      kinds
  in
  if kinds == [] then 
    ignore (warn "Empty descriptor for vararg function");
  kinds

(* May raise Not_found *)
let findTypeIndex (t: typ) 
                  (argkinds: vaKind list)
    : int * string * typ * typsig = 
  let ts = argumentTypeSig t in
  List.find (fun (_, _, _, ks) -> ks = ts) argkinds

(* Fetch references to the Vararg globals *)
let ccured_va_count_o : lval option ref = ref None
let ccured_va_tags_o : lval option ref = ref None
let ccured_va_count_max = 32
let fetchVarargGlobals () =
  let g1 = 
    match !ccured_va_count_o with 
      Some lv -> lv
    | None -> 
        let g1 = makeGlobalVar "__ccured_va_count" intType in
        g1.vstorage <- Extern;
        g1.vdecl <- !currentLoc;
        MU.theFile := GVarDecl(g1, g1.vdecl) :: !MU.theFile;
        let g1lval = var g1 in
        ccured_va_count_o := Some g1lval;
        g1lval
  in
  let g2 = 
    match !ccured_va_tags_o with 
      Some lv -> lv
    | None -> 
        let g2 = makeGlobalVar "__ccured_va_tags" 
            (TArray(intType, Some (integer ccured_va_count_max), [])) in
        g2.vstorage <- Extern;
        g2.vdecl <- !currentLoc;
        MU.theFile := GVarDecl(g2, g2.vdecl) :: !MU.theFile;
        let g2lval = var g2  in
        ccured_va_tags_o := Some g2lval;
        g2lval
  in
  g1, g2


(* take apart a format string and return a list of int/pointer type choices *)
let rec parseFormatString f start = begin
  (* interpret a conversion specifier: the stuff that comes after the % in a 
   * printf format string  *)
  let rec parseConversionSpec f start = begin
    try 
      match Char.lowercase f.[start] with
        'e' | 'f' | 'g' |   (* double arg *)
        'a' ->              (* double arg *)
          [TFloat(FDouble, [])], (start+1)
      | 'd' | 'i' |         (* signed int arg *)
        'o' | 'u' | 'x' |   (* unsigned int arg *)
        'c' |               (* unsigned char *)
        'p' ->              (* void pointer treated as int *)
          [intType], (start+1)
      | '*' ->            (* integer *)
          let a,b = parseConversionSpec f (start+1) in
          intType :: a, b 
      | 's' -> [charPtrType], (start+1) (* char pointer *)
      | 'n' -> E.s (bug "Cannot handle 'n' character in format string [%s]" f)
      | _ -> parseConversionSpec f (start+1)
    with _ ->
      ignore (warn "Malformed format string [%s]" f);
      raise Not_found
  end
  in
  try 
    let i = String.index_from f start '%' in
    if f.[i+1] = '%' then (* an escaped %, not a format char *) 
      parseFormatString f (i+2)
    else begin
      (* look after the % to see if they want an Int or a Pointer *)
      let t, next_start = parseConversionSpec f (i+1) in
      t @ (parseFormatString f next_start) (* look for another % *)
    end
  with _ -> []  (* no more % left in format string *)
end

let checkFormatArgsFun = 
  let fdec = emptyFunction "CHECK_FORMATARGS" in
  let argp  = makeVarinfo "format" (TPtr(charType, [Attr("rostring",[])])) in
  fdec.svar.vtype <- TFun(intType, Some [ argp ], false, []);
  fdec

(* See if a function has a boxformat argument and return it *)
let getFormatArg (func: exp) 
                 (args: exp list) : exp option = 
  try
    match filterAttributes "boxformat" (MU.getFunctionTypeAttributes func) with
      [Attr(_, [AInt format_idx])] -> begin
        let format_arg = 
          try List.nth args (format_idx - 1)
          with _ -> begin
            ignore (warn "Call to format function with too few arguments");
            raise Not_found
          end
        in
        Some format_arg
      end
    | _ -> None
  with Not_found -> None
 
(* Special handler for format strings *)
let checkFormatString 
    (format_str: string)
    (argkinds: vaKind list) 
    (indices: int list) : bool =
  try
    let format_types = parseFormatString format_str 0 in (* May raise 
    * Not_found  *)
    let format_indices = 
      List.map (fun t -> 
        let n, _, _, _ = findTypeIndex t argkinds in n) format_types in
    (* Now commpare the indices. Allow any kind of argument for an 
    * index 0 in the format_indices (corresponds to int) *)
    let intIndex, _, _, _ = 
      try findTypeIndex intType argkinds 
      with _ -> E.s (bug "handleFormatString")
    in
    let rec compare indices formats = 
      match indices, formats with
        [], [] -> ()
      | i :: indices, f :: formats when (i = f || f = intIndex) -> 
          compare indices formats
      | _ -> begin
          ignore (warn "Format string \"%s\" not compatible with arguments: %a\n" 
                    format_str (docList (chr ',') num) format_indices);
          raise Not_found
      end
    in
    compare indices format_indices;
    true
  with Not_found -> false

let rec stripCasts (e: exp) = 
  match e with CastE(_, e') -> stripCasts e' | _ -> e
  
(* Prepare the argument in a call to a vararg function *)
let prepareVarargArguments
    ~(func: exp) 
    ~(nrformals: int)
    ~(args: exp list) : instr list * exp list = 
  try
    let descrt = getVarargFunctionDescriptorType func in
    let argkinds = getVarargKinds descrt in
    if debugVararg then 
      ignore (E.log "Preparing args in call to %a\n" d_exp func);
    let format_argo = getFormatArg func args in
    let nrargs = List.length args in
    let (_, indices, args') = 
      List.fold_right
        (fun a (arg_idx, indices, args) -> 
          if arg_idx > nrformals then begin
            let t = typeOf a in
            (* Search for a compatible type in the kinds *)
            let k_idx, _, kt, _ = 
              try findTypeIndex t argkinds
              with Not_found ->  begin
                (* Maybe it is a format function. In that case cast to int *)
                if format_argo != None then begin
                  ignore (warn "Unexpected argument %d of format function %a. Turned into int.@!" arg_idx d_exp func);
                  try findTypeIndex intType argkinds 
                  with _ -> E.s (bug "prepareVarArgs")
                end else
                  E.s (error "Argument %d (%a : %a) does not match any expected type for vararg function %a.@! Expected one of: %a" 
                         arg_idx d_exp a d_type t d_exp func 
                         (docList (chr ',' ++ break) 
                            (fun (_,_,tau,_) -> d_type () tau)) argkinds )
              end
            in
            let a' = mkCastT a t kt in
            (arg_idx - 1, k_idx :: indices, a' :: args)
          end else
            (arg_idx - 1, indices, a :: args))
        args
        (nrargs, [], [])
    in
    if debugVararg then 
      ignore (E.log "Indices are: [%a]\n" (docList (chr ',') num) indices);
    (* fetch the global variables *)
    let ccured_va_count, ccured_va_tags = fetchVarargGlobals () in
    (* Now construct the tag *)
    let rec loopIndices (count: int) (indices: int list) 
        : exp (* a partial word*) * instr list (* set the following words 
        * *) = 
      match indices with 
      | [] -> zero, [Set(ccured_va_count, integer count, !currentLoc)]
      | i :: rest -> 
          let p_rest, rest = loopIndices (count + 1) rest in
          let amount = 8 * (count mod 4) in
          let shifted = 
            if amount = 0 then integer i else 
            BinOp(Shiftlt, integer i, integer amount, intType) in
          let bor_ed = 
            if isZero p_rest then shifted else 
            BinOp(BOr, shifted, p_rest, intType) in
          if amount = 0 then (* time to put into rest *)
            (zero,
             Set(addOffsetLval (Index(integer (count / 4), 
                                      NoOffset)) ccured_va_tags,
                 bor_ed, !currentLoc) :: rest) 
          else
            (bor_ed, rest)
    in
    (* Now check the special case of the format-functions with constant 
    * format strings *)
    let instr = 
      match format_argo with
      | Some format_arg -> begin
          match stripCasts format_arg with
            (Const(CStr format_str))
              when (checkFormatString format_str argkinds indices) -> 
                [Set(ccured_va_count, integer (-1), !currentLoc)]
          | _ -> 
              (* format string not a constant or it cannot be checked 
               * statically  *)
              (* Generate code to fill in the indices *)
              let _, instr = loopIndices 0 indices in 
              (* Generate a call to a run-time function that checks the 
               * indices  *)
              let call = Call(None, Lval(var checkFormatArgsFun.svar), 
                              [ format_arg ],
                              !currentLoc)
              in
              instr @ [call]
      end

      | None -> (* Not a printf function *)
          let _, instr = loopIndices 0 indices in 
          instr
    in
    instr, args'
      
  with Not_found -> E.s (error "Call to vararg function %a without descriptor"
                           d_exp func)
        

(* Create the type for storing the vararg information *)
let varargCompInfo: compinfo option ref = ref None
let getVarArgCompInfo () : compinfo  =  
  let ci = 
    match !varargCompInfo with 
      Some x -> x
    | None -> 
        let ci = 
          mkCompInfo true "__ccured_va_localinfo"
            (fun _ ->
              [ ("next", intType, None, []);
                ("count", intType, None, []);
                ("tags", TArray(intType, 
                                Some (integer ccured_va_count_max), []),
                 None, []);
                ("nextp", voidPtrType, None, []);
              ])
            [] in
        MU.theFile := GCompTag(ci, !currentLoc) :: !MU.theFile;
        varargCompInfo := Some ci;
        ci
  in
  ci
    
    (* The last formal in this body *)
let lastformal : varinfo ref = ref dummyFunDec.svar

(** Now the handling of the body of vararg functions **)
class processVarargClass (fdec: fundec) = object
  inherit nopCilVisitor

    (* Change all call instructions to use the marker *)
  method vinst (i: instr) : instr list visitAction = 
    match i with 
      Call(reso, (Lval(Var fvi, NoOffset) as f), args, l) -> 
        let getNthArg (n: int) = 
          try
            stripCasts (List.nth args n) 
          with _ -> E.s (error "Not enough arguments in call to %s\n" 
                           fvi.vname)
        in
        if fvi.vname = "__ccured_va_start" then begin
          match getNthArg 0 with 
            Lval (Var markervi, NoOffset) -> begin
              let markerdescrt = 
                try getValistDescriptorType markervi 
                with Not_found -> 
                  E.s (error "Call to va_start on a non va_list variable\n")
              in
              let fundescrt = 
                try getVarargFunctionDescriptorType (Lval(var fdec.svar))
                with Not_found -> 
                  E.s (error "Call to va_start in a non vararg function\n")
              in
              if markerdescrt <> fundescrt then
                E.s (error "Cannot match the descriptor types in va_start");
              SkipChildren
            end
          | _ -> E.s (error "Invalid call to %s\n" fvi.vname)
        end else if "__ccured_va_arg" = fvi.vname then begin
          (* Find the type of the desired argument *)
          let markervi, desiredt = 
            match args with 
              Lval(Var markervi, NoOffset) :: SizeOf t :: _ :: [] -> 
                markervi, t
            | _ -> E.s (error "Invalid call to %s" fvi.vname)
          in
          let desiredts = argumentTypeSig desiredt in
          (* Find the index in the union type of a compatible type *)
          let kinds = 
            try getVarargKinds (getValistDescriptorType markervi) 
            with Not_found -> 
              E.s (error "Cannot find descriptor type in call to %s\n" fvi.vname)
          in
          let ki, kn, kt, ks = 
            match List.filter (fun (_, _, _, ks) -> ks = desiredts) kinds with 
              [x] -> x
            | _ -> 
                E.s (error "Cannot match any of the descriptor type for call to va_arg")
          in
          (* Now change the type of the temporary variable that takes this 
           * result to a pointer to kt *)
          (match reso with 
            Some (Var tmpvi, NoOffset) -> 
              tmpvi.vtype <- TPtr(kt,[])
          | None -> ignore (warn "Call to va_arg is not assigned")
          | _ -> E.s (unimp "Result of va_arg does not go into a variable"));
          (* it is important that the argument of sizeof shares the node with 
           * the actualy union field *)
          ChangeTo ([Call(reso, f, [ Lval (var markervi); 
                                     SizeOf kt; 
                                     integer ki ], l)])
        end else
          SkipChildren
    | _ -> SkipChildren
end


(** Make sure you call this before marking anything in the function body. If 
 * you do not do that the newly created temporaries might not get marked *)
let processVarargBody (fdec: fundec) : unit = 
  let formals, isva = 
    match fdec.svar.vtype with
      TFun(_, formals, isva, _) -> formals, isva
    | _ -> E.s (E.bug "Function does not have function type")
  in
  let foundVarargs = ref false in
  (* Get and memoize the descriptor for a va_list local. Return None if this 
   * is not a va_list local. *)
  let getDescriptorLocal (l: varinfo) : typ option = 
    match unrollType l.vtype with 
      TPtr(TComp(ci, a), pa) when prefix "__ccured_va_list" ci.cname -> begin
        foundVarargs := true;
        let descrt = 
          try getValistDescriptorType l 
          with Not_found -> begin
            (* It does not have a descriptor already. Grab it from the 
            * function type *)
            try
              let descrt = 
                getVarargFunctionDescriptorType (Lval(var fdec.svar)) in
              l.vtype <- 
                 TPtr(TComp(ci, a), 
                      addAttribute (Attr("boxvararg", [ASizeOf descrt])) pa);
              descrt
            with Not_found -> 
              E.s (error "Cannot find the descriptor type for va_list local %s in %s"
                     l.vname fdec.svar.vname)
          end
        in
        if debugVararg then 
          ignore (E.log "va: found va_list variable %s with descriptor %a\n"
                    l.vname d_type descrt);
        Some descrt
      end
    | _ -> None
  in
  let doLocal (isformal: bool) 
              (acc: instr list) 
              (l: varinfo) = 
    (* Get and memoize the descriptor type. We'll need it later *)
    match getDescriptorLocal l with 
    | Some descrt when not isformal -> 
        (* Make a local to store the va info *)
        let ci = getVarArgCompInfo () in
        let vainfo = 
          makeTempVar fdec ~name:(l.vname ^ "__vainfo") (TComp(ci, [])) in
        vainfo.vaddrof <- true;
        if debugVararg then 
          ignore (E.log "va: creating local va_local %s for %s\n"
                    vainfo.vname l.vname);
        (* And an instruction to initialize the local *)
        let init = Set(var l, AddrOf (var vainfo), !currentLoc) in
        init :: acc
    | _ -> acc
  in
  (* Go through all the locals (but not formals) and when we see a variable 
   * of type __ccured_va_list then we allocate a local structure to which the 
   * variable should point *)
  let initlocals = List.fold_left (doLocal false) [] fdec.slocals in
  (* Go through the formals also,mainly to verify that they have descriptors *)
  List.iter (fun f -> ignore (doLocal true [] f)) fdec.sformals;
  if !foundVarargs then begin
    lastformal := 
       if isva then begin
         (* Get the last formal *)
         let rec getlast = function 
             [] -> E.s (E.bug "Vararg function %s does not have formals" 
                          fdec.svar.vname)
           | [l] -> l
           | _ :: rest -> getlast rest
         in getlast (argsToList formals)
       end else dummyFunDec.svar;
    let va_visit = new processVarargClass(fdec) in
    fdec.sbody <- visitCilBlock va_visit fdec.sbody;
    (* Now insert code to initialize the next fields *)
    fdec.sbody.bstmts <- 
       mkStmt (Instr initlocals) ::
       fdec.sbody.bstmts
  end


let initFile () = 
  MU.theFile := GVarDecl(checkFormatArgsFun.svar, locUnknown) :: !MU.theFile;
