(*
 *
 * Copyright (c) 2001 by
 *  George C. Necula	necula@cs.berkeley.edu
 *  Scott McPeak        smcpeak@cs.berkeley.edu
 *  Wes Weimer          weimer@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products 
 * derived from  this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

open Cil
open Pretty

module H = Hashtbl
module E = Errormsg

module N = Ptrnode

module MU = Markutil

let debugWrappers = false

(* maps functions to wrappers *)
let ccuredWrappedBy: (string, string) H.t = H.create 15

(* a list of wrappers *)
let ccuredWrappers: (string, unit) H.t = H.create 15

let processPragma = function
  | Attr("ccuredwrapperof", lst) -> begin
      match lst with 
	[AStr wrappername; AStr wrapped] -> begin
          if debugWrappers then 
            ignore (E.log "Will use wrapper %s for %s\n"
                      wrappername wrapped);

          H.add ccuredWrappers wrappername ();
	  try
            (* Check whether the function already has a wrapper. *)
	    let previousWrapper = H.find ccuredWrappedBy wrapped in
	    if previousWrapper <> wrappername then
              ignore (E.unimp "Function %s has more than one wrapper" wrapped )
            (*if the names are the same, it's probably just a repeated
              .h file. *)
	  with Not_found ->
            (* Good, there were no other wrappers for this function. *)
	    H.add ccuredWrappedBy wrapped wrappername;
	  
	end 
      | AStr wrappername :: AStr funcName :: rest ->   ignore 
            (warn "#pragma ccuredwrapperof can take only one function name.")
      | _ ->   ignore 
            (warn "Not enough parameters to #pragma ccuredwrapperof.")
    end
  | _ -> ()


(* Find the wrapper of a function. Might raise Not_found *)
let findWrapperOf (s: string) = 
  let wrappername = H.find ccuredWrappedBy s in
  try
    match H.find MU.allFunctions wrappername with
      MU.Defined fdec -> fdec
    | _ -> 
        E.s (error "Cannot find the definition of wrapper %s of %s\n"
               wrappername s)
  with Not_found -> 
    E.s (error "Cannot find the definition of wrapper %s of %s\n"
           wrappername s)

let isAWrapper (s:string) = 
  H.mem ccuredWrappers s


let initFile () = 
  H.clear ccuredWrappedBy;
  H.clear ccuredWrappers


let addConstraints42 () = 
  H.clear ccuredWrappedBy;
  H.clear ccuredWrappers


(*Helpers for addConstraints, from markptr *)

let setPosArith n = begin N.setFlag n N.pkPosArith end
let setReachSeq n = begin N.setFlag n N.pkReachSeq end
let setReachString n = begin N.setFlag n N.pkReachString end
let setArith n = begin N.setFlag n N.pkArith end

(* Grab the node from the attributs of a type. Returns dummyNode if no such 
 * node *)
let nodeOfType t = 
  match unrollType t with 
    TPtr(_, a) -> begin
      match N.nodeOfAttrlist a with
        Some n -> n
      | None -> N.dummyNode
    end
  | _ -> N.dummyNode


(* Adds constraints to the arguments of e.g. __endof, __mkptr, etc. *)
let addConstraints (func:exp) (args:exp list) (rt: typ) (callId: int) = 
  match func with 
    Lval(Var v, NoOffset) -> begin
      let strippedName = Poly.stripPoly v.vname in
      match args with
        [a] -> 
          if "__endof" = strippedName then begin
            let n = nodeOfType (typeOf a) in
            if n == N.dummyNode then
              E.s (error "Call to __endof on a non pointer: %a" d_plainexp a);
            setPosArith n (* To make sure we have an end *) ;
            setReachSeq n 
          end else if "__strlen" = strippedName then begin
            let n = nodeOfType (typeOf a) in
            if n == N.dummyNode then
              E.s (error "Call to __strlen on a non pointer: %a" d_plainexp a);
            setPosArith n (* To make sure we have an end *) ;
            setReachString n;
            setReachSeq n 
          end else if "__startof" = strippedName then begin
            let n = nodeOfType (typeOf a) in
            if n == N.dummyNode then
              E.s (error "Call to __startof on a non pointer");
            setArith n (* To make sure we have a start and an end *) ;
            setReachSeq n 
          end 
      | [a; b] -> 
          if "__mkptr" = strippedName then begin
	    (* We must be able to cast b to the return value *)
            let restn = nodeOfType rt in
            let btn = nodeOfType (typeOf b) in
            if btn == N.dummyNode then
              E.s (error "Call to __mkptr on a non pointer: %a" d_plainexp b);
	    ignore (N.addEdge btn restn N.ECast callId None)
          end else if "__write_at_least" = strippedName then begin
	    (* a must have an end *)
            let n = nodeOfType (typeOf a) in
            if n == N.dummyNode then
              E.s (error "Call to __write_at_least on a non pointer: %a" d_plainexp a);
            setPosArith n;
            setReachSeq n 
          end
      | [a; b; c] -> (* __copytags(dest, src, len) *)
          if "__copytags" = strippedName then begin
	    (* a and b should be compatible *)
            let atn = nodeOfType (typeOf a) in
            let btn = nodeOfType (typeOf b) in
            if atn == N.dummyNode then
              E.s (error "Call to __copytags on a non pointer: %a" d_plainexp a);
            if btn == N.dummyNode then
              E.s (error "Call to __copytags on a non pointer: %a" d_plainexp b);
            (* matth: is this correct?*)
	   ignore (N.addEdge atn btn N.EIndex callId None);
          end
      | _ -> ()
    end
  | _ -> ()

