(*
 *
 * Copyright (c) 2001 by
 *  George C. Necula	necula@cs.berkeley.edu
 *  Scott McPeak        smcpeak@cs.berkeley.edu
 *  Wes Weimer          weimer@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1. XSRedistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products 
 * derived from  this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

 
(* main *)
(* this module is the program entry point for the 'ccured' *)
(* program, which instruments C programs with runtime checks to *)
(* prevent memory safety violations *)


module type FRONTEND =
  sig
   (* Signal that we are in MS VC mode *)
    val setMSVCMode: unit -> unit

   (* Parse a file in *)
    exception ParseError of string

    (* additional command line arguments *)
    val args  : (string * Arg.spec * string) list
    (* the main command to parse a file. Return a thunk that can be use to 
     * convert the AST to CIL *)
    val parse : string -> (unit -> Cil.file)

  end

                                        (* Use the frontc frontend *)
module F : FRONTEND = Frontc
module C = Cil
module CK = Check
module E = Errormsg
open Pretty
open Trace

let newoptim  = ref false (* Use the new optimizer *)

let boxChannel : out_channel option ref = ref None
let annChannel : out_channel option ref = ref None
let optimChannel : out_channel option ref = ref None
    (* channel for printing out annotated file, needed by PCC *)
let optimFileName : string ref = ref ""
let mergedName : string ref = ref "stdout"
let mergedChannel : out_channel option ref = ref None
let cilChannel : out_channel option ref = ref None
let tableAll = ref false
let tableInterface = ref false
let inferChannel : out_channel option ref = ref None
let optimElimAll : string list ref = ref [] (* A list of CHECK_xxx to remove 
                                              * *)
let preproc = ref ""                    (* The preprocess command *)
let typecheck = ref false
exception Done_Processing

let parseOneFile fname : Cil.file = 
  if !Util.printStages then ignore (E.log "Parsing %s\n" fname);
  let cil = F.parse fname () in
  if !Util.doCheck then begin
    ignore (E.log "Checking CIL after CABS2CIL\n");
    CK.checkFile [] cil;
  end;
  cil

let rec processOneFile cil =
  try
    if !cilChannel = None && !boxChannel = None then raise Done_Processing;

     (* sm: remove unused temps to cut down on gcc warnings  *)
     (* (Stats.time "usedVar" Rmtmps.removeUnusedTemps cil);  *)
    (trace "sm" (dprintf "removing unused temporaries\n"));
    ignore (E.log "removing unused temporaries\n");
    Rmtmps.removeUnusedTemps cil;

    if !cilChannel <> None && !boxChannel = None && !Util.logCalls then
      Logcalls.logCalls cil;

    if !cilChannel <> None && !boxChannel = None && !Util.logWrites then
      Logwrites.logWrites cil;

    if !Util.doCheck then begin
      ignore (E.log "Checking CIL before printing CIL\n");
      CK.checkFile [] cil;
    end;

    (match !cilChannel with
      None -> ()
    | Some c -> Stats.time "printCIL" (C.printFile C.defaultCilPrinter c) cil);

    if !E.hadErrors then
      E.s (E.error "Cabs2cil has some errors");
    (* Pull out the global initializer if we either infer or box *)
    let withglob =
      match !inferChannel, !boxChannel with
        None, None -> cil
      | _, _ ->
          let globinit = Globinit.doFile cil in
          globinit
    in
    let addGraphToOutput c = 
      begin
        (trace "sm" (dprintf "printing pointer graph\n"));
        (* AB: These flags are no longer used by Pretty *)
        (*
        let o_noal = !noAligns in
        let o_nobr = !noBreaks in
        noAligns := true; noBreaks := true;
        *)
        output_string c "#if 0\n/* Now the graph */\n";
        output_string c "/* Now the solved graph (simplesolve) */\n";
        Stats.time "printgraph" Ptrnode.printGraph c;
        (*
        noAligns := o_noal; noBreaks := o_nobr;
        *)
        output_string c "/* End of solved graph*/\n#endif\n";
      end
    in
    let infer =
      let solver =
        if !E.verboseFlag then 
          ignore (E.log "Using %s solver\n" !Markptr.solver);
        match !Markptr.solver with
          "none" -> None
        | "wild" -> Some Solveutil.wild_solve
        | _ -> Some (Solver.solve withglob)
      in
      match solver with
        None -> withglob
      | Some s ->
        (* Mark the file *)
          (trace "sm" (dprintf "calling markptr\n"));
          if !Util.printStages then ignore (E.log "Collecting constraints\n");
          let infer = Markptr.markFile withglob in
        (* Solve the graph *)
          if !Util.printStages then ignore (E.log "Solving constraints\n");
          (trace "sm" (dprintf "invoking the solver\n"));
          (Util.tryFinally
             (fun _ ->
               Stats.time "solver"
                 (fun graph ->
                   let should_typecheck = s graph in
                   if should_typecheck && not !typecheck then begin 
                     ignore (E.warn "The inference results should be type-checked.")
                   end ; 
                   if (!tableInterface) then begin
                     Ptrnode.useLeanFats := true;
                     Solveutil.table_interface graph;
                   end ;
                   if (!tableAll) then begin
                     Ptrnode.useLeanFats := true;
                     Solveutil.table_it_all graph
                   end
                 ) Ptrnode.idNode
             )
             (fun _ ->
               (match !inferChannel with
                 None -> ()
               | Some c ->
                   C.printFile Ptrnode.ccuredInferPrinter c infer;
                   addGraphToOutput c))
             ());
          infer
    in
    if (!E.verboseFlag || !Util.printStats) && !Markptr.solver <> "none" then 
      Stats.time "graphstats" Ptrnode.printGraphStats ();
    if !typecheck then begin
      Stats.time "typecheck" Typecheck.typecheck_file infer
    end ; 

    (* weimer: remember the size of types *)
    Stats.time "type-size-check"
      (fun x -> ignore (C.visitCilFile (new Boxstats.typeSizeSetVisitor) x))
      infer;

    if !Util.printStages then ignore (E.log "Adding run-time checks\n");
    let box : Cil.file =
      match !boxChannel with
        None -> infer
      | Some c ->
          (trace "sm" (dprintf "invoking the boxer\n"));
          let box = Stats.time "box" Box.boxFile infer in
          if !Util.doCheck then
            CK.checkFile [CK.NoCheckGlobalIds] box;
          if !Util.logCalls then
            Logcalls.logCalls box;
          if !Util.logWrites then
            Logwrites.logWrites box;

          Stats.time "printCured"
            (C.printFile Ptrnode.ccuredPrinter c) box;
          if !E.hadErrors then begin
            E.s (E.error "There were errors during boxing\n");
          end;
          (trace "sm" (dprintf "boxer has finished\n"));
          box
    in

    (* spr: emit an annotated file for use by PCC *)
    let annotated : Cil.file =
      match !annChannel with
        None -> box (* don't annotate *)
      | Some c ->  (* annotate *)
          if !Util.printStages then ignore (E.log "Adding annotations\n");
          let annot = 
            Switch.make_cfg box;
            Stats.time "annot" Annot.annotate box
          in
          Stats.time "printAnnot" (C.printFile Annot.ccuredPrinter c) annot;
          annot
    in

    let final =
      match !optimChannel with
        None -> annotated
      | Some c ->
          (trace "sm" (dprintf "invoking the optimizer\n"));
          if !Util.printStages then ignore (E.log "Optimizing checks\n");
          Optim.checkToRemove := None;
          let optimized1 = 
            if !newoptim then Stats.time "Optim" Seoptim.optim box
            else box
          in
          let optimized2 = 
            Stats.time "Optim" Choptim.optim optimized1
          in 
          let optimized3 = Stats.time "Optim" Optim.optimFile optimized2 in
	  ignore (E.log "Optimizer statistics : %t\n" Optim.getStatistics);
          Stats.time "printOptim" 
            (C.printFile Ptrnode.ccuredPrinter c) optimized3;
          (* Now see if we must also produce a number of optimizer variants *)
          List.iter 
            (fun name -> 
              let toremove = 
                if name = "CHECKS" then "" (* remove all *)
                else "CHECK_" ^ name
              in
              (* Find the new file name *)
              let newname = 
                if Filename.check_suffix !optimFileName ".optim.c" then 
                  (Filename.chop_suffix !optimFileName ".optim.c") 
                  ^ ".no" ^ name ^ ".optim.c"
                else
                  E.s (E.bug "doOneVariant: wrong suffix in ouput file name")
              in
              (* Make a private copy of the file *)
              let copyGlobal = function
                  C.GFun (fd, l) -> 
                    C.GFun (C.copyFunction fd fd.C.svar.C.vname, l)
                | g -> g
              in
              let copy = 
                { optimized3 with
                       C.globals = List.map copyGlobal optimized3.C.globals;
                       C.globinit = 
                        match optimized3.C.globinit with
                           None -> None
                         | Some g -> Some (C.copyFunction g
                                             g.C.svar.C.vname)} in
              if !newoptim then 
                if toremove = "" then Seoptim.forcefullyRemoveAll := true 
                else Seoptim.forcefullyRemove := [toremove];
              
              Optim.checkToRemove := Some toremove;
              if !E.verboseFlag then 
                ignore (E.log " removing checks: %s\n" toremove);
              let removed = 
                (if !newoptim then Seoptim.optim else Optim.optimFile) copy in
              let variant_out = open_out newname in
              C.printFile Ptrnode.ccuredPrinter variant_out removed;
              close_out variant_out)
            !optimElimAll;
          (* Return the optimized version *)
          optimized3
    in
    if !E.verboseFlag || !Util.printStats then begin
      Boxstats.print final;
    end;
    if !Util.printStages then ignore (E.log "Cure complete\n");
    (* weimer: check those types *)
    ignore (Stats.time "type-size-check"
              (C.visitCilFile (new Boxstats.typeSizeCheckVisitor)) box);
    ()
  with Done_Processing -> ()
        
(***** MAIN *****)  
let rec theMain () =
  let doCombine = ref "" in 
  let usageMsg = "Usage: ccured [options] source-files" in
  let files : string list ref = ref [] in
  let recordFile fname = files := fname :: (!files) in
  (* Parsing of files with additional names *)
  let parseExtraFile (s: string) = 
    try
      let sfile = open_in s in
      while true do
        let line = try input_line sfile with e -> (close_in sfile; raise e) in
        let linelen = String.length line in
        let rec scan (pos: int) (* next char to look at *)
                     (start: int) : unit (* start of the word, 
                                            or -1 if none *) =
          if pos >= linelen then 
            if start >= 0 then 
              recordFile (String.sub line start (pos - start))
            else 
              () (* Just move on to the next line *)
          else
            let c = String.get line pos in
            match c with 
              ' ' | '\n' | '\r' | '\t' -> 
                (* whitespace *)
                if start >= 0 then begin
                  recordFile (String.sub line start (pos - start));
                end;
                scan (pos + 1) (-1)

            | _ -> (* non-whitespace *)
                if start >= 0 then 
                  scan (pos + 1) start 
                else
                  scan (pos + 1) pos
        in
        scan 0 (-1)
      done
    with Sys_error _ -> E.s (E.error "Cannot find extra file: %s\n" s)
   |  End_of_file -> () 
  in
  (* Processign of output file arguments *)
  let openFile (what: string) (takeit: out_channel -> unit) (fl: string) = 
    if !E.verboseFlag then
      ignore (Printf.printf "Setting %s to %s\n" what fl);
    (try takeit (open_out fl)
    with _ ->
      raise (Arg.Bad ("Cannot open " ^ what ^ " file")))
  in
  let whichSolver sname = 
    Markptr.solver := sname in
  let setDebugFlag v name = 
    E.debugFlag := v;
    if v then Pretty.flushOften := true
  in
  whichSolver "none"; (* No solver for now *)
  (* sm: for print depth *)
  let setTraceDepth n =
    Pretty.printDepth := n
  in
  (* sm: to test stack depth expansion *)
  let rec recursiveAddThing n =
    if (n = 0) then 0 else 1 + (recursiveAddThing (n-1))
  in
  (*********** COMMAND LINE ARGUMENTS *****************)
  let argDescr = [
    "--verbose", Arg.Unit (fun _ -> E.verboseFlag := true),
                "turn on verbose mode";
    "--debug", Arg.String (setDebugFlag true),
                     "<xxx> turns on debugging flag xxx";
    "--entryPoint", Arg.String (fun s -> Globinit.mainname := s),
                     "<xxx> call globinit from program entry point xxx";
    "--flush", Arg.Unit (fun _ -> Pretty.flushOften := true),
                     "Flush the output streams often (aids debugging)" ;
    "--check", Arg.Unit (fun _ -> Util.doCheck := true),
                     "turns on consistency checking of CIL";
    "--typecheck", Arg.Unit (fun _ -> typecheck := true),
                     "turns on consistency checking of CCured types";
    "--nocheck", Arg.Unit (fun _ -> Util.doCheck := false),
                     "turns off consistency checking of CIL";
    "--logcalls", Arg.Unit (fun _ -> Util.logCalls := true),
                     "turns on generation of code to log function calls in CIL";
    "--logstyle", Arg.Int (fun i -> (Logcalls.setStyle i)),
                     Logcalls.styleHelp;
    "--logwrites", Arg.Unit (fun _ -> Util.logWrites := true),
                     "turns on generation of code to log memory writes calls in CIL";
    "--nodebug", Arg.String (setDebugFlag false), 
                     "<xxx> turns off debugging flag xxx";
    "--noStackChecks", Arg.Unit (fun _ -> Box.stackChecks := false),
                     "turns off storing-local-variable checks";
    "--log", Arg.String (openFile "log" (fun x -> E.logChannel := x)), 
             "the name of the log file";
    "--cilout", Arg.String (openFile "CIL" (fun x -> cilChannel := Some x)), 
               "the name of the output CIL file";
    "--curedout", Arg.String (openFile "CCURED" 
                                (fun x -> boxChannel := Some x)),
                 "the name of the boxed output file";
    "--mergedout", Arg.String (fun mn -> 
                                 mergedName := mn;
                                 openFile "merged output" 
                                   (fun oc -> mergedChannel := Some oc) mn),
                "specify the name of the merged file";

    "--tableAll", Arg.Unit (fun _ -> tableAll := true), "Make all pointers use an external table.";
    "--tableInterface", Arg.Unit (fun _ -> tableInterface := true), "Make interface pointers use an external table.";
    "--inferout", Arg.String (openFile "INFER" 
                                (fun x -> inferChannel := Some x)), 
                 "Infer the pointer qualifiers";
    "--curetype", Arg.String whichSolver, "Which solver to use [infer or wild]";
    "--warnall", Arg.Unit (fun _ -> E.warnFlag := true),
                 "Show all warnings";
(*    "-warnsolver", Arg.Unit (fun _ -> Secondsolve.warn := true),
      "Warn if solver disagrees with manual annotations"; *)
    "--MSVC", Arg.Unit (fun _ -> if Machdep.hasMSVC then begin
                                   C.msvcMode := true;
                                   F.setMSVCMode ()
                                  end else 
                                     E.s (E.error "MSVC mode is not supported on this build\n")),   "Produce MSVC output. Default is GNU";
    (* sm: the next two lines appeared twice?! *)
    (*"-boxout", Arg.String boxFile, "the name of the boxed output file";*)
    (*"-inferbox", Arg.String inferFile, "Infer the pointer qualifiers";*)
    "--boxdefaultwild", Arg.Unit (fun _ -> Ptrnode.defaultIsWild := true),
                       "the default pointer qualifier is wild";
    "--boxallpoly", Arg.Unit (fun _ -> Ptrnode.allPoly := true),
                       "make all external functions polymorphic";
    "--boxexternpoly", Arg.Unit (fun _ -> Ptrnode.externPoly := true),
                       "make all external functions polymorphic";

    "--annout", Arg.String (openFile "annotated" 
                              (fun x -> annChannel := Some x)),
                                  "the name of the output file that contains\
                                    annotations required for PCC";

    "--optimout", Arg.String (openFile "cured-optimized"
                                (fun x -> optimChannel := Some x)),
                                         "the name of the boxing-optimized\
                                           output file";
    "--optimelimall", 
                   Arg.String (fun s -> optimElimAll := s :: !optimElimAll), 
                   "Produce several variants of optimized code";
    "--newoptim", Arg.Unit (fun _ -> newoptim := true), 
                   "Uses the new optimizer";
    "--keepunused", Arg.Unit (fun _ -> Rmtmps.keepUnused := true),
                "do not remove the unused variables and types"; 
    "--noPrintLn", Arg.Unit (fun _ -> Cil.printLn := false;
                                     Cprint.printLn := false),
               "don't output #line directives";
    "--commPrintLn", Arg.Unit (fun _ -> Cil.printLnComment := true;
                                       Cprint.printLnComment := true),
               "output #line directives in comments";
    "--stats", Arg.Unit (fun _ -> Util.printStats := true),
               "print some statistics";
    "--stages", Arg.Unit (fun _ -> Util.printStages := true),
               "print the stages of the algorithm as they happen";
    (* sm: some more debugging options *)
    "--tr",         Arg.String Trace.traceAddMulti,
                     "<sys>: subsystem to show debug printfs for";
    "--pdepth",     Arg.Int setTraceDepth,
                      "<n>: set max print depth (default: 5)";
    "--recurse",    Arg.Int (fun n -> (ignore (recursiveAddThing n)); ()),
                      "<n>: deliberately make stack grow to O(n) bytes";
    "--extrafiles", Arg.String parseExtraFile,
    "<filename>: the name of a file that contains a list of additional files to process, separated by whitespace of newlines";

  ] @ F.args in
  begin
    (* this point in the code is the program entry point *)

    Stats.reset ();

    (* parse the command-line arguments *)
    Arg.parse argDescr recordFile usageMsg;
    let files = List.rev !files in

    (* parse each of the files named on the command line, to CIL *)
    let cils = List.map parseOneFile files in

    (* if there's more than one source file, merge them together; *)
    (* now we have just one CIL "file" to deal with *)
    let one =
      match cils with
        [one] -> one
      | [] -> E.s (E.error "No arguments\n")
      | _ ->
          let merged =
            Stats.time "merge" (Mergecil.merge cils) !mergedName in
          if !E.hadErrors then
            E.s (E.error "There were errors during merging\n");
          (* See if we must save the merged file *)
          (match !mergedChannel with
            None -> ()
          | Some mc -> begin
              let oldpci = !C.print_CIL_Input in
              C.print_CIL_Input := true;
              Stats.time "printMerged"
                (C.printFile C.defaultCilPrinter mc) merged;
              C.print_CIL_Input := oldpci
          end);
          merged
    in

    (* process the CIL file (merged if necessary) *)
    processOneFile one
  end
;;
                                        (* Define a wrapper for main to 
                                         * intercept the exit *)
let failed = ref false 
let main () = 
  let term = 
    try 
      theMain (); 
      fun () -> exit (if !failed then 1 else 0)
    with e ->  
      (fun () -> 
        print_string ("Uncaught exception: " ^ (Printexc.to_string e)
                      ^ "\n");
        if !E.verboseFlag then Stats.print stderr "Timings:\n";
        exit 2)
  in
  begin
    if !E.verboseFlag || !Util.printStats then
      Stats.print stderr "Timings:\n";
    close_out (! E.logChannel);  
    (match ! cilChannel with Some c -> close_out c | _ -> ());
    (match ! boxChannel with Some c -> close_out c | _ -> ());
    (match ! annChannel with Some c -> close_out c | _ -> ());
    (match ! optimChannel with Some c -> close_out c | _ -> ());
    (match ! inferChannel with Some c -> close_out c | _ -> ());
    term ()
  end
;;

Printexc.catch main ()
;;
