(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: fociAst.ml,v 1.2 2003/06/17 17:09:27 rupak Exp $
 *
 * This file is modified from the BLAST Project by KLM.
 *)


(**
 * This module allows the parsing of formulas
 *)

module Misc = FociMisc

module Symbol = 
  struct 
    type symbol = string (* change later *)

    let toString s = s       (* name of symbol *)

    let print fmt s =
      Format.pp_print_string fmt s
  end

module Constant =
  struct
    type constant = 
	Int     of int

    let toString c =
      match c with
	Int i -> string_of_int i
 
  end

module Expression = 
  struct
    type expression =
	  Constant of Constant.constant
        | Variable of Symbol.symbol
	| Application of Symbol.symbol * expression list
	| Sum of expression list
	| Coeff of Constant.constant * expression

    let rec print fmt e =
      begin
    	match e with

    	Constant ac -> 
	    Format.fprintf fmt "%s" (Constant.toString ac)

	| Application (s, ev) ->
            Format.fprintf fmt "%s" (Symbol.toString s);
	    Format.fprintf fmt "(";
	    List.iter (fun a-> print fmt a) ev;
	    Format.fprintf fmt ")"

    	| Variable s -> 
            Format.fprintf fmt "%s" (Symbol.toString s)

	| Sum(l) -> 
	    Format.fprintf fmt "+[";
	    print_vec fmt l;
	    Format.fprintf fmt "]"

	| Coeff(c,t) ->
	    Format.fprintf fmt "( * ";
	    Format.fprintf fmt "%s" (Constant.toString c);
	    Format.fprintf fmt "  ";
	    print fmt t;
	    Format.fprintf fmt ")"
	    

      end

    and print_vec fmt ev =
          List.iter (fun a -> print fmt a; Printf.printf " ") ev


   let rec toString e =
     begin
       match e with

    	Constant ac -> 
	    (Constant.toString ac)
	    
	| Application (exp, ev) ->
	    (Symbol.toString exp)^
	    "("^
	    (toStringList  ev)^")"

    	| Variable s -> 
	    begin
	      (Symbol.toString s)
	    end

	| Sum(l) -> 
	    "+[" ^ (toStringList l) ^ "]"

	| Coeff(c,t) ->
	    "(* " ^ (Constant.toString c) ^ " " ^ (toString t) ^ ")"
      end

    and toStringList  ev =
          List.fold_left (fun a -> fun b -> a ^ " " ^ (toString b)) "" ev

    let rec support exp =
      match exp with
      Constant(_) -> []
      | Application (func, args) -> 
	    (List.fold_left Misc.union [] (List.map support args))
      | Variable(sym) -> [sym]
      | Sum(args) -> (List.fold_left Misc.union [] (List.map support args))
      | Coeff(c,t) -> (support t)


    let rec subst (e,v,vv) =
      let rec s e =
	if e = v then vv
	else
	  match e with
	      Constant(_) -> e
	    | Application (func, args) -> 
		Application (func, List.map s args)
	    | Variable(_) -> e
	    | Sum(args) -> Sum (List.map s args)
	    | Coeff(c,t) -> Coeff(c,(s t))
      in s e
  end

(**
   A predicate is a boolean combination of atomic expressions.
   *)

module Predicate =
    struct
      type predicate =
	  True
	| False
	| And of predicate list
	| Or of predicate list
	| Not of predicate
        | Implies of predicate * predicate
	| Equality of Expression.expression * Expression.expression
	| Leq of Expression.expression * Expression.expression
        | Atom of Symbol.symbol


      let rec toString pred =
	match pred with
	  True -> "true"
	| False -> "false"
	| And plist -> "& ["^(toStringList plist)^"]"
	| Or plist -> "| ["^(toStringList plist)^"]"
	| Not p -> "~ ("^(toString p)^")"
	| Implies (p1, p2) -> "("^(toString p1)^"->"^(toString p2)^")"
        | Equality (x,y) -> "(= "^(Expression.toString x)^" "^(Expression.toString y)^")"
        | Leq (x,y) -> "(<= "^(Expression.toString x)^" "^(Expression.toString y)^")"
	| Atom (s) -> (Symbol.toString s)
      
      and toStringList plist =
	List.fold_left (fun a -> fun b -> a ^ " " ^ (toString b)) "" plist

      let print fmt pr = Format.pp_print_string fmt (toString pr)


      let rec support pred =
        match pred with
	  True -> []
        | False -> []
        | And plist -> List.fold_left (fun l1 l2 -> Misc.union l1 (support l2)) [] plist
        | Or plist -> List.fold_left (fun l1 l2 -> Misc.union l1 (support l2)) [] plist
        | Not p -> support p
        | Implies (p1, p2) -> Misc.union (support p1) (support p2)
        | Equality (x,y) -> Misc.union (Expression.support x) (Expression.support y)
        | Leq (x,y) -> Misc.union (Expression.support x) (Expression.support y)
	| Atom (s) -> []

      let rec subst (e,v,vv) =
	let rec s e =
	  match e with
	      True -> e
            | False -> e
            | And plist -> And(List.map s plist)
            | Or plist -> Or(List.map s plist)
            | Not p -> Not(s p)
            | Implies (p1, p2) -> Implies (s p1, s p2)
            | Equality (x,y) -> Equality(Expression.subst(x,v,vv),Expression.subst (y,v,vv))
	    | Atom (_) -> e
	    | Leq(x,y) -> Leq(Expression.subst(x,v,vv),Expression.subst (y,v,vv))
	in s e
	     
    end

let zero = Expression.Constant(Constant.Int(0))
