(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: fociMisc.ml,v 1.1 2003/06/16 18:52:26 jhala Exp $
 *
 * This file is part of the BLAST Project.
 *)

(**
 * This module provides some miscellaneous useful helper functions.
 *)
open Str

let swap (x,y) = (y,x)

let fst3 (x,y,z) = x
let snd3 (x,y,z) = y
let thd3 (x,y,z) = z

(* Poor man's set operations (in terms of lists) *)

(* add an element to a list if it's not already in there *)
let add lst elem = if List.mem elem lst then lst else elem::lst

(* union two lists representing sets *)
let union l1 l2 = List.fold_left add l2 l1

let compact l = union l []

(* set difference using lists to represent sets *)
let difference l1 l2 = List.filter (function elem -> not(List.mem elem l2)) l1

let intersection l1 l2 = difference l1 (difference l1 l2)


(* set containment using lists *)
let rec subsetOf l1 l2 =
  match l1 with
    [] -> true
  | a::rest -> (List.mem a l2) && (subsetOf rest l2)

(* given a list of strings, produce a single string that is the
   comma-separated list of the original strings *)
let strList l = String.concat ", " l

(* returns the to_string function corresponding to the given
   pretty-printer *)
let to_string_from_printer printer =
  function data ->
    printer Format.str_formatter data ;
    Format.flush_str_formatter ()

(* given a printer for a type t, returns a printer for lists of elements of
   type t *)
let list_printer_from_printer printer fmt list =
  Format.fprintf fmt "@[[@[" ;
  begin
    match list with
        [] ->
          ()
      | head::tail ->
          printer fmt head ;
          List.iter (function e -> Format.fprintf fmt ";@ " ;
                                   printer fmt e)
                    tail
  end ;
  Format.fprintf fmt "@]]@]"

(* returns true if the character is a digit [0-9] *)
let is_digit c =
	match c with
		'0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9' -> true
	| _ -> false

let map_exn ex f l = 
  let rec _map _l = 
    match _l with
      [] -> []
    | h::t -> try (f h)::(_map t) with ex -> (_map t)
  in
  _map l

let filter_cut f l = 
  let l1 = List.filter f l in
  let l2 = difference l l1 in
  (l1,l2)

(* returns the truncated list containing the first n elements Raise Failure "nth" if list is too short *)
let truncateList lst n =
  let rec _truncateWorker l rest len =
    if len = 0 then rest else 
    match l with
      [] -> failwith "truncateList: list too short"
    | head::tail -> _truncateWorker tail (rest @[head]) (len - 1)
  in
  _truncateWorker lst [] n

let chop_list f l = 
  let rec search st fin = 
    match fin with
	[] -> (st,fin)
      | h::tl -> if f h then (st,fin) else search (st@[h]) tl
  in
    search [] l

(* chop s chopper returns Some([x;y;z...]) if s = x.chopper.y.chopper ..., None o.w. *)
let chop s chopper = 
  let reg = Str.regexp chopper in
    Str.split reg s  

let bounded_chop s chopper i = 
  let reg = Str.regexp chopper in
    Str.bounded_split reg s i 
      
let is_prefix p s = 
  let reg = Str.regexp p in
    Str.string_match reg s 0

(*  chop_after_prefix p s = if s = p.s' then s' *)
let chop_after_prefix prefix s =
  let reg = Str.regexp prefix in
    (*if (Str.string_match reg s 0) then *)
    (List.hd(Str.bounded_split reg s 1))
    
(* Hashtable union: hash_union t1 t2 adds all the data of t2 to t1 *)

let hashtbl_addtable t1 t2 = 
  let f_add key data = Hashtbl.add t1 key data in
    Hashtbl.iter f_add t2

(* search_list f l walks through l and finds the first guy st f(x) is some..., 
   it returns that some *)
let search_list f list = 
  let rec search l = 
    match l with 
	[] -> None
      | h::t -> 
	  begin
	    match f h with 
		Some(s) -> Some(s)
	      | None -> search t
	  end
  in
    search list

let rec list_gather_option l = 
  match l with
      [] -> []
    | Some(a)::tl -> a::(list_gather_option tl)
    | None::tl -> (list_gather_option tl)

let rec list_add l =
  let add x y = x + y in
    List.fold_right add l 0
