(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(**
   aliasAnalyzer.ml (the engine that produces the alias database
   and answers alias queries).

   Depends on Bddptsto
   *)
  
open Ast


let listOfAliases = ref []
let pointsToInfo : Bddptsto.pointsToInfo ref = ref (Bddptsto.FlowInsensitivePointsTo Bddptsto.bddZero)

let constructAliasDatabase () =
  let aliasFile = Options.getValueOfString "alias" in
  if (aliasFile = "") then
    (* Nothing to do *)
    ()
  else
    if (aliasFile = "bdd") then
    begin
      (* do the bdd based alias analysis *)
      pointsToInfo := Bddptsto.do_bdd_alias (if (Options.getValueOfString "pta" = "flow") then true else false)
    end
    else 
    (* add each pair to the alias database *)
      begin
       (* we simulate the effect of an alias analysis by letting the user input a file 
	  with alias pairs.
        *)
	let readAliasPairs filename =
	  let doParse inchan =
	    let _ = VampyreErrormsg.startFile filename in
	    let lexbuf = Lexing.from_channel inchan in
	    let _ = VampyreErrormsg.theLexbuf := lexbuf in
	    Inputparse.main Inputflex.token lexbuf 
	  in
	  ignore (Message.msg_string Message.Normal "Reading in alias pairs...");
	  try
	    let inchan = open_in filename
	    in
	    let a = doParse inchan
	    in
	    ignore (Message.msg_string Message.Normal "Alias pairs read.\n") ;
	    List.iter (fun a -> Message.msg_string Message.Normal (Predicate.toString a)) a ;
	    a
	  with
	    Sys_error _ -> (Message.msg_string Message.Error ("Cannot find alias information in "^filename^".\n"); [])
	  | VampyreErrormsg.Error -> (Message.msg_string Message.Error ("Error raised in reading alias pairs.\n"); [])
	in
	
	let rec loopOverAliasPairs lst =
	  match lst with
	    [] -> ()
	  | p :: rest ->
	      begin
		match p with
		  Predicate.Atom a ->
		  begin
		    match a with
		      Expression.Binary (Expression.Eq, Expression.Lval l1, Expression.Lval l2) ->
			listOfAliases := (l1, l2) :: !listOfAliases ; loopOverAliasPairs rest
		    | _ -> failwith ("Strange expression read by readAliasPairs "^(Expression.toString a))
		  end
		|	_ -> failwith ("Strange predicate read by readAliasPairs "^(Predicate.toString p))
	      end
	in
	loopOverAliasPairs (readAliasPairs aliasFile)(* ;
							transitivelyClose !listOfAliases *)
      end
	
let askAliasDatabase l1 l2 =
  Message.msg_string Message.Debug "In askAliasDatabase" ;
  Message.msg_string Message.Debug ("Lvalues are "^(Expression.lvalToString l1)^" and "^(Expression.lvalToString l2)) ;
  if Options.getValueOfString "alias" = "bdd" then
    begin
      let t1 = BlastCSystemDescr.C_System_Descr.get_type_of_lval l1 
      and t2 = BlastCSystemDescr.C_System_Descr.get_type_of_lval l2 in
      if t1 = t2 then (* only check aliasing between values of the same type *) 
	match !pointsToInfo with
	  Bddptsto.FlowSensitivePointsTo locbddhashtbl ->
	  (* Not implemented yet *)
	    false
	|	Bddptsto.FlowInsensitivePointsTo ptstobdd ->
          (* Look up aliasing relationship from pointsToBdd *)
	    Bddptsto.checkAliasFI (ptstobdd) l1 l2
      else
	false
    end
  else
    (* Assume the user supplied list of aliases *)
    let checker (l,l') = ((l = l1) && (l' = l2)) || ((l=l2) && (l'=l1)) in
    List.exists checker !listOfAliases

  (*
    let rec loop l1 l2 lst =
    match lst with 
    [] -> false
    | (l, l') :: rest ->
    begin
    if ((l = l1) && (l' = l2)) || ((l=l2) && (l'=l1)) then true
    else
    loop l1 l2 rest
    end
    in
    loop l1 l2 !listOfAliases
  *)

let getAllAliases lv =
  if Options.getValueOfString "alias" = "bdd" then
    begin
      failwith "getAllAliases : bdds : Not implemented yet" ;
      match !pointsToInfo with
	Bddptsto.FlowSensitivePointsTo locbddhashtbl ->
	  []
      |	Bddptsto.FlowInsensitivePointsTo ptstobdd ->
          (* Look up aliasing relationship from pointsToBdd *)
	  []
    end
  else
    let lst =
      Misc.union 
	(List.map snd (List.filter (fun (l,l') -> l = lv) !listOfAliases))
	(List.map fst (List.filter (fun (l,l') -> l' = lv) !listOfAliases))
    in
    Message.msg_string Message.Debug ("In getAllAliases with "^(Expression.lvalToString lv)) ;
    Message.msg_string Message.Debug ("Aliases are ") ;
    List.iter (fun l -> Message.msg_string Message.Debug (Expression.lvalToString l)) lst ;
    lst
      
let can_escape lv globals =
  Message.msg_string Message.Debug ("In can_escape with "^Expression.lvalToString lv) ;
  let rec can_escape_worker lv globals seen =
    let rec exp_can_escape e globals =
      match e with
      |	Expression.Lval l -> can_escape_worker l globals seen
      |	Expression.Chlval (l,_) -> 
	  begin
	    can_escape_worker l globals seen
	  end
      | Expression.Binary (_, e1, e2) -> (exp_can_escape e1 globals) || (exp_can_escape e2 globals)
      |	Expression.Cast (t, e1) -> exp_can_escape e1 globals
      | Expression.Constant(_) -> false
      |	Expression.Constructor (_, _) -> failwith "Constructor not handled in can_escape"
      | Expression.FunctionCall (func, args) ->
	  begin
	    let rec loop lst =
	      match lst with 
		[] -> false
	      | a:: rest -> if exp_can_escape a globals then true else loop rest
	    in
	    exp_can_escape func globals || (loop args) 
	  end
      | Expression.Unary(_, e) -> 
	  exp_can_escape e globals
      |	_ -> failwith "can_escape : Strange expression"
    in
    if List.mem lv seen then
      false else
      begin
	let escapes_directly = 
	  match lv with
	    Expression.Symbol s -> Hashtbl.mem globals lv 
	  | Expression.Dereference e -> exp_can_escape e globals 
	  | Expression.Access (_, e, _) -> exp_can_escape e globals
	  | Expression.Indexed (e1, e2) -> failwith "Indexed not handled in can_escape"
	in
	escapes_directly || 
	(List.exists (fun l -> 
	  can_escape_worker l globals (lv::seen)) (getAllAliases lv))
      end
  in
  let r = can_escape_worker lv globals []
  in
  Message.msg_string Message.Debug ("can escape returns "^(if r then "true" else "false"));
  r
    
let rec generateAliasConstraints (var1 : Expression.lval) (var2 : Expression.lval) : Predicate.predicate =
  match (var1, var2) with
  | Expression.Symbol v, Expression.Symbol v' -> if askAliasDatabase var1 var2 then Predicate.True else Predicate.False
  | Expression.Access (Expression.Dot, Expression.Lval l, fld1), Expression.Access (Expression.Dot, Expression.Lval l', fld2)
  | Expression.Access (Expression.Arrow, Expression.Lval l, fld1), Expression.Access (Expression.Arrow, Expression.Lval l', fld2)->
      if fld1 = fld2 then
	generateAliasConstraints l l'
      else
	Predicate.False
  | Expression.Indexed (e1, e2), Expression.Indexed (e1', e2') ->
      Predicate.conjoinL 
	[ Predicate.Atom (Expression.Binary (Expression.Eq, e1, e1')); 
	  Predicate.Atom (Expression.Binary (Expression.Eq, e2, e2'))]
  | _ -> 
      if askAliasDatabase var1 var2 then Predicate.True
      else Predicate.False (* not aliased *)


(* The main interface.
*)
let rec queryAlias (v1 : Expression.expression) (v2 : Expression.expression) = 
    Message.msg_string Message.Debug ("Alias query for "^(Expression.toString v1)^" and "^(Expression.toString v2)) ;
  let lvalOfExp e =
    match e with 
      Expression.Lval l -> l
    | _ -> failwith ("queryAlias: bug in replaceChlvals?" ^ (Expression.toString e))

  in
  try
  let var1 = lvalOfExp (Expression.replaceChlvals v1) 
  and var2 = lvalOfExp (Expression.replaceChlvals v2) 
  in
  let a' = 
    if var1 = var2 then false else (* RJ: why is this false ?
				      Answer: If var1 is syntactically equal to var2, then
				      we substitute var2 anyway; and moreover var1 not = var2
				      is false. So in this case we can do without aliasing.
				      *)
    begin
	askAliasDatabase var1 var2
     end
  in
  if a' = true then 
    Message.msg_string Message.Debug ("Got back alias answer true for "^(Expression.toString v1)^" and "^(Expression.toString v2)) ;
  let a'' =
    match (var1,var2) with
	(Expression.Dereference (Expression.Binary(Expression.Plus,a1,o1))
	   ,Expression.Dereference (Expression.Binary(Expression.Plus,a2,o2))) ->
	  (a1 = a2)
      | (Expression.Access(_ ,Expression.Lval(Expression.Dereference(Expression.Binary(Expression.Plus,a1,o1))), f1),
	 Expression.Access(_ ,Expression.Lval(Expression.Dereference(Expression.Binary(Expression.Plus,a2,o2))), f2))
	-> (a1=a2 && f1 = f2)
      | (Expression.Indexed(e1,e2),Expression.Indexed(e1',e2')) -> (e1 = e2)
      | _ -> false
  in 
  let a = a' || (a'' && (not (var1 = var2))) (* this last conjunct just to be consistent *) 
  in
    Message.msg_string Message.Debug (Printf.sprintf "QueryAlias %s %s (%s)\n\n" (Expression.lvalToString var1) (Expression.lvalToString var2) (if a then "true" else "false")); 
    a
    with _ -> 
      begin
	Message.msg_string Message.Debug "queryAlias raises exception for the query" ;
	Message.msg_string Message.Debug ((Expression.toString v1) ^" and " ^ (Expression.toString v2)) ;
	Message.msg_string Message.Debug "returning false" ;
	false  
      end

