(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: ast.ml,v 1.23 2003/08/20 19:19:32 rupak Exp $
 *
 * This file is part of the BLAST Project.
 *)


(**
 * This module allows the parsing of C programs.
 *)


type ast_type = string (* change later *)


module Symbol = 
  struct 
    type symbol = string (* change later *)
    type t = symbol

    let compare = Pervasives.compare

    let toString s = s       (* name of symbol *)

    let print fmt s =
      Format.pp_print_string fmt s
  end

(* The variable undefinedVariable is used to encode nondeterministic
   choice. If it appears in an If, then the predicate is discarded
   and the If can nondeterministically branch in either direction.

   See also build_cfa, where If statements are checked for this variable, and
   replaced by nondeterministic choice.
*)
let undefinedVariable = "__BLAST_NONDET" (* formerly, arbit *)

module Constant =
  struct
    type constant = 
	Int     of int		(* SLAM uses int32?? so that we get all 32 bits *)
      | Float   of float
      | String  of string
      |	EnumC   of int * string (* SLAM uses int32 *)

    let toString c =
      match c with
	Int i -> string_of_int i
      |	Float f -> string_of_float f
      |	String s -> Printf.sprintf "\"%s\"" (String.escaped s)
      |	EnumC (i,name) -> Printf.sprintf "%s(%s)" name (string_of_int i)

  end

module Expression = 
  struct
    type access_op = Dot | Arrow
      
    type assign_op = Assign | AssignBitAnd | AssignBitOr | AssignDiv | AssignLShift | AssignMinus
                   | AssignMul | AssignPlus | AssignRem | AssignRShift | AssignXor
      
      
      type unary_op  = UnaryMinus | Not | BitNot | Address | UnaryPlus | DestructorCall | Assume | Assert
                     | PreInc | PreDec | PostInc | PostDec

    type binary_op = BitAnd | BitOr | Div | LShift | Minus | Mul | Plus | Rem | RShift | Xor 
                   | Eq | Ge | Gt | Le | Lt | Ne | Index | And | Or | Comma 
                   | FieldOffset
      
      type lval =
	  Symbol of Symbol.symbol
	| Access of access_op * expression * Symbol.symbol
	| Dereference of expression
	| Indexed of expression * expression

      and expression =
	  Lval of lval
        | Chlval of lval * string (* changed lvals to help implement skolemization in wp, sp computation *)
   	| Assignment of assign_op * lval * expression (* op, target, value *)
	| Binary of binary_op * expression * expression
        | Cast of ast_type * expression
	| Constant of Constant.constant
	| Constructor of Symbol.symbol * ast_type
	| FunctionCall of expression * expression list
	| Sizeof of int
	| Unary of unary_op * expression

    type t = expression

    let compare = Pervasives.compare

    let unaryOpToString = function
	UnaryMinus     -> "-"
      | Not            -> "!"
      | BitNot         -> "~"
      | Address        -> "&"
      | UnaryPlus      -> "+"
      | DestructorCall -> "Destruct"
      | Assume         -> "__assume"
      | Assert         -> "__assert"
      | PreInc       -> "++"
      | PreDec       -> "--"
      | PostInc      -> "++"
      | PostDec      -> "--"

    let unaryOpToVanillaString = function
	UnaryMinus     -> "UnaryMinus"
      | Not            -> "Not"
      | BitNot         -> "BitNot"
      | Address        -> "Address"
      | UnaryPlus      -> "UnaryPlus"
      | DestructorCall -> "Destruct"
      | Assume         -> "Assume"
      | Assert         -> "Assert"
      | PreInc       -> "PreInc"
      | PreDec       -> "PreDec"
      | PostInc      -> "PostInc"
      | PostDec      -> "PostDec"

    let assignOpToString = function
	Assign       -> "="
      | AssignBitAnd -> "&="
      | AssignBitOr  -> "|="
      | AssignDiv    -> "/="
      | AssignLShift -> "<<="
      | AssignMinus  -> "-="
      | AssignMul    -> "*="
      | AssignPlus   -> "+="
      | AssignRem    -> "%="
      | AssignRShift -> ">>="
      | AssignXor    -> "^="

    let accessOpToString = function
	Dot       -> "."
      | Arrow     -> "->"

    let binaryOpToString = function
    	BitAnd    -> "&"
      | BitOr     -> "|"
      | Div       -> "/"
      | LShift    -> "<<"
      | Minus     -> "-"
      | Mul       -> "*"
      | Plus      -> "+"
      | Rem       -> "%"
      | RShift    -> ">>"
      | Xor       -> "^"
      | Eq        -> "=="
      | Ge        -> ">="
      | Gt        -> ">"
      | Le        -> "<="
      | Lt        -> "<"
      | Ne        -> "!="
      | Index     -> "[]"
      | And       -> "&&"
      | Or        -> "||"
      | Comma     -> ","
      |	FieldOffset -> "foffset"

    let binaryOpToVanillaString = function
    	BitAnd    -> "BitAnd"
      | BitOr     -> "BitOr"
      | Div       -> "Div"
      | LShift    -> "LShift"
      | Minus     -> "Minus"
      | Mul       -> "Mul"
      | Plus      -> "Plus"
      | Rem       -> "Rem"
      | RShift    -> "RShift"
      | Xor       -> "Xor"
      | Eq        -> "Eq"
      | Ge        -> "Ge"
      | Gt        -> "Gt"
      | Le        -> "Le"
      | Lt        -> "Lt"
      | Ne        -> "Ne"
      | Index     -> "Index"
      | And       -> "And"
      | Or        -> "Or"
      | Comma     -> "Comma"
      |	FieldOffset -> "Foffset"


    let rec print fmt exp =
      begin
    	match exp with
	  Lval lv -> print_lval fmt lv
	| Chlval (lv,i) -> 
	    begin
	      Format.fprintf fmt "Primed(" ;
	      print_lval fmt lv ;
	      Format.fprintf fmt ", %s)" i 
	    end
    	| Assignment (op, target, value) ->
	    let op_string = assignOpToString op in
	    begin
	      print_lval fmt target;
	      Format.fprintf fmt "@ %s@ " op_string;
	      print fmt value
	    end
	| Binary (bop, left, right) ->
	    let op_string = " "^(binaryOpToString bop)^" " in
	    print fmt left;
	    Format.fprintf fmt "@ %s@ " op_string;
	    print fmt right;

	| Cast (typ, exp) ->
	    begin
		  Format.fprintf fmt "(<cast>)(@[%a@])" print exp
	    end

    	| Constant ac -> 
	    Format.fprintf fmt "%s" (Constant.toString ac)

	| Constructor (sym, typ) ->
	    begin
	      try Format.fprintf fmt "%s" (Symbol.toString sym)
	      with
	      	_ -> Format.fprintf fmt "exn in Constructor@."
	    end

	| FunctionCall (exp, es) ->
	    print fmt exp;
	    Format.fprintf fmt "(";
            begin match es with
              [] -> ()
            | e0 :: et ->
                print fmt e0;
	        List.iter (fun a -> Format.fprintf fmt ", "; print fmt a) et
            end;
	    Format.fprintf fmt ")"

	| Sizeof i ->
	    Format.fprintf fmt "sizeof(<%d>)" i
    	| Unary (op, arg) ->
	    let op_string = unaryOpToString op in
	    begin
	      match op with
	      	(PostInc | PostDec) ->
		  Format.fprintf fmt "(" ;
		  print fmt arg;
		  Format.fprintf fmt ")" ;
		  Format.fprintf fmt "%s" op_string
	      | _ ->
	      	  Format.fprintf fmt "%s" op_string;
		  Format.fprintf fmt "(" ;
	      	  print fmt arg ;
		  Format.fprintf fmt ")"
	    end

      end

    and print_vec fmt ev =
          List.iter (fun a -> print fmt a; Printf.printf " ") ev

    and print_lval fmt lv =
      match lv with
    	| Symbol s -> 
	    begin
	      try Format.fprintf fmt "%s" (Symbol.toString s)
	      with
		_ -> Format.fprintf fmt "exn in Symbol (expr)@."
	    end
	| Access (op, exp, sym) ->
	    begin
	      let op_string = accessOpToString op in
	      print fmt exp;
	      Format.fprintf fmt "%s" op_string;
              Format.fprintf fmt "%s" (sym)
	    end
	| Dereference exp ->
	    begin
	      Format.fprintf fmt "* (" ;
	      print fmt exp ;
	      Format.fprintf fmt " )" ;
	    end
	| Indexed (exp1, exp2) ->
	    begin
	      print fmt exp1 ;
	      Format.fprintf fmt "[ " ;
	      print fmt exp2 ;
	      Format.fprintf fmt " ]" 
	    end

   let rec toString e =
     begin
       match e with
       | Lval lv -> (lvalToString lv)
       | Chlval (lv, suffix) ->
	   "Primed ("^(lvalToString lv)^","^suffix^" )"
       | Assignment (op, target, value) ->
	   let op_string = assignOpToString op in
	   (lvalToString target)^op_string^(toString value)

	| Binary (bop, left, right) ->
	    let op_string = binaryOpToString bop in
	    (toString left)^op_string^(toString right)

	| Cast (typ, exp) ->
	    "(<cast>)("^(toString exp)^")"

    	| Constant ac -> 
	    (Constant.toString ac)
	    
	| Constructor (sym, typ) ->
	    (Symbol.toString sym)
	    
	| FunctionCall (exp, ev) ->
	    (toString exp)^
	    "("^
	    (toStringList  ev)^")"

	| Sizeof i ->
	    "sizeof(<"^(string_of_int i)^">)"

    	| Unary (op, arg) ->
	    let op_string = unaryOpToString op in
	    begin
	      match op with
	      	(PostInc | PostDec) ->
		  "("^(toString arg)^")"^op_string
	      | _ ->
	      	  op_string^"("^(toString arg)^")"
	    end

      end

    and toStringList  ev =
          List.fold_left (fun a -> fun b -> a ^ "," ^ (toString b)) "" ev

   and lvalToString lv =
     match lv with
       Access (op, exp, sym) ->
	 begin
	   let op_string = accessOpToString op in
	   "("^(toString exp)^")"^op_string^(sym)
	 end
     | Symbol s -> 
	 begin
	   (Symbol.toString s)
	 end
     | Dereference e ->
	 begin
	   "* ("^(toString e)^")"
	 end
     | Indexed (e1, e2) ->
	 (toString e1)^"[ ("^(toString e2)^")  ]"

	     

    (* miscellaneous utilities *)
       let isRelOp op =
	 match op with
	   Eq | Ge | Gt | Le | Lt | Ne -> true
	 | _ -> false

  (* compute the address of an Expression *)
    let rec addressOf (e : expression) = 
      begin
	match e with 
	  Lval lv ->
	    begin
	      match lv with
		Dereference e1 -> e1
	      |	Access (Dot, e1, fld) ->
		  Binary (FieldOffset, addressOf e1, Lval (Symbol fld))
	      |	Indexed (e1, e2) -> Binary(Plus, e1,e2)
	      | Symbol _ -> Unary(Address, e)
	      |	_ -> failwith ("Pattern matching failed in addressOf with "^(lvalToString lv))
	    end
	| Chlval (l, _) -> (* We push address-of operators through Chlvals *)
	    addressOf (Lval l)
(*
	    Unary (Address, e)
*)
	| _ -> failwith ("addressOf : strange expression "^(toString e))
      end

    (* Push the top level dereference below address-ofs and fieldoffsets.
       Use:
          * (& x) = x
          * (fieldoffset e fld) = access (Dot, *e, fld)
    *)
    let rec push_deref (l : lval) =
      begin
        match l with
          Dereference e ->
           begin
            match e with
              Unary(Address, Lval l') -> l'
            | Unary(Address, _) -> failwith ("push_deref: Address applied to non-lval "^(lvalToString l))
            | Binary(FieldOffset, e, Lval (Symbol fld)) ->
              begin
                match e with
                  Lval l' -> Access(Dot, Lval (push_deref l'), fld)
                | _ -> failwith ("push_deref : Strange lvalue "^(lvalToString l))
              end
            | Binary(FieldOffset, _, _) -> failwith ("push_deref: strange field offset "^(lvalToString l))
            | _ -> l  
           end
        | _ -> l
      end


(* RJ: this could be easily rewritten with deep_alpha_convert *)
       (* replace all occurrences of Chlval by the corresponding Lval *)
    let rec replaceChlvals exp =
      let replaceChlvals_lval l =
	match l with
    	  Access(op, e, f) -> 
	    begin
	      Access (op, replaceChlvals e, f) 
	    end
	| Dereference e -> 
	    begin
	      push_deref (Dereference (replaceChlvals e))
	    end
	| Symbol(_) -> l
	| Indexed (e1, e2) -> Indexed (replaceChlvals e1, replaceChlvals e2) 
      
      in
      match exp with
      |	Lval l -> Lval (replaceChlvals_lval l)
      |	Chlval (l, suff) -> Lval (replaceChlvals_lval l)
	  
      | Assignment(op, target, value) -> Assignment (op, replaceChlvals_lval target, replaceChlvals value)
      | Binary (op, e1, e2) -> Binary (op, replaceChlvals e1, replaceChlvals e2)
      |	Cast (t, e1) -> Cast(t, replaceChlvals e1)
      | Constant(_) -> exp
      |	Constructor (_, _) -> exp
      | FunctionCall (func, args) -> 
	  FunctionCall (replaceChlvals func, List.map replaceChlvals args)
      |	Unary (Address, e) ->
	  let e' = replaceChlvals e in
	  addressOf e'
      | Unary(op, e) -> 
	  Unary (op, replaceChlvals e)
      |	_ -> failwith ("Pattern matching failed in replaceChlvals with "^(toString exp))

      

      (* check if an expression is a memory op *)
    let rec isDeref e =
      match e with
	Lval lv ->
	  begin
	    match lv with
	      Symbol _ -> false
	    |	_ -> true
	  end
      |	Chlval (lv, _) -> isDeref (Lval lv)
      |	_ -> failwith ("Strange expression in isDeref "^(toString e))
	    
      (* given a deref lval *p or ( *p ).f or p->f, return the derefed symbol p *)
    let derefedVar lv =
      match lv with
    	Dereference e -> e
      | Access(Arrow, e,_) -> e
      | Access(Dot, e,_) -> 
	  begin
	    match e with
	      Lval l -> 
		if isDeref e then e 
		else 
		  failwith ("derefedVar:  parameter is not a deref expression "^(lvalToString l))
	    | Chlval (l, _) ->
		failwith ("derefedVar: Chlval not handled")
(* 
		if isDeref e then e 
		else 
		  failwith ("derefedVar:  parameter is not a deref expression "^(lvalToString l))
*)
	    | _ -> failwith ("derefedVar: strange operator to Dot : "^(toString e))
	  end
      | _ -> failwith ("derefedVar:  parameter is not a deref expression "^(lvalToString lv))

      (* get the variables p s.t. *p or ( *p ).f or p->f appear in varExps *)
    let getParents varExps =
      let derefExps = (List.filter (fun l -> isDeref (Lval l)) varExps) in
      let rec getDerefVars vars exps =
    	match exps with
	  [] -> vars
    	| (hd::tl) -> getDerefVars (Misc.add vars (derefedVar hd)) tl
      in
      getDerefVars [] derefExps
	
      (** return the expressions of the form x, x.f, *x, &x, x->f, and ( *x ).f in the 
	  given expression.  we assume those are the only forms. 
	  also, because of our assumption that whole-struct assignments do not occur, we
	  do not go "inside" an expression of the form x->f or ( *x ).f .  that is, we return it, but not
	  *x as a variable in the expression.  
	  *)
    let rec allVarExps exp =
      let allVarExps_lval l =
	match l with
    	  Access(Dot, e, _) -> 
	    begin
	      match e with
		Lval (Symbol (_)) -> 
		  [(Lval l)]
	      | Lval (Dereference e) -> [Lval l]
              |   _ -> (Lval l) :: (allVarExps e)
	    end
	| Access(Arrow, e, _) ->
	    begin
	      match e with
		Lval (Symbol _) -> [Lval l]
	      | _ -> failwith ("Strange expression in allVarExps :"^(toString exp))
	    end
	| Dereference e -> 
	    begin
	      match e with
		Unary (Address, Lval l' ) -> [Lval l']  
	      | Unary (Address, _ ) -> failwith "allVarExps: Address taken of something which is not an lvalue"
	      | _ -> [Lval l] 
	    end
	| Symbol(_) -> [Lval l]
	| Indexed (e1, e2) -> (Lval l) :: (Misc.union (allVarExps e1) (allVarExps e2))
      
      in
      match exp with
      |	Lval l -> allVarExps_lval l
      |	Chlval (l, suff) -> 
	  begin
	    let _t = 
	      match l with
		Symbol _ -> []
	      | Access (_, e, label) -> allVarExps e
	      | Dereference e -> allVarExps e
	      | Indexed (e1, e2) -> Misc.union (allVarExps e1) (allVarExps e2)
	    in
	    Misc.union [exp] _t
	  end
      | Assignment(op, target, value) -> Misc.union [(Lval target)] (allVarExps value)
      | Binary (FieldOffset, e1, e2) -> 
	  allVarExps e1
      | Binary (op, e1, e2) -> 
	  Misc.union (allVarExps e1) (allVarExps e2)
      |	Cast (t, e1) -> allVarExps e1
      | Constant(_) -> []
      |	Constructor (_, _) -> []
      | FunctionCall (func, args) -> 
	  Misc.union (allVarExps func) 
	    (List.fold_left Misc.union [] (List.map allVarExps args))
      | Unary(op, e) -> 
	  (match op with
	  |	Address -> [exp]
	  |	_ -> allVarExps e) 
      |	_ -> []

    let negateRel exp =
      match exp with
	Binary (op, e1, e2) -> 
	  begin
	    if (isRelOp op)
	    then
	      begin
		match op with
		  Eq -> Binary(Ne, e1, e2)
		| Ge -> Binary(Lt, e1, e2)
		| Gt -> Binary(Le, e1, e2)
		| Le -> Binary(Gt, e1, e2)
		| Lt -> Binary(Ge, e1, e2)
		| Ne -> Binary(Eq, e1, e2)
		| _ -> failwith ("negateRel: expected binary relational predicate, got "^(toString exp))
	      end
	    else
	      failwith ("negateRel: expected binary relational predicate, got "^(toString exp))
	  end
      |	_ -> failwith ("negateRel: expected binary relational predicate, got "^(toString exp))

  (* A fresh lval given some lval -- ie given an expression which is an lval it returns an appropriate new chlval *)
    let fresh_lval_name = ref 0
    
    let new_tag () = 
      fresh_lval_name := !fresh_lval_name + 1;
      string_of_int !fresh_lval_name
	
    let fresh_lval lv = Chlval (lv,new_tag () )

  (* Renaming of variables *)
  let rec alpha_convert f = 
    begin
      function
	| Lval l -> Lval (f l) (* Lval (alpha_convert_lval f l) *)
	| Chlval (l,s) -> 
	    begin
	      match l with
	      Symbol _ -> Chlval (l, s)
		| Access (op, e, label) -> Chlval (Access (op, alpha_convert f e, label), s)
		| Dereference e -> Chlval (Dereference (alpha_convert f e), s)
		| Indexed (e1, e2) -> Chlval (Indexed (alpha_convert f e1, alpha_convert f e2), s)
	    end
	| Assignment (op, l, r) ->
	    (*WAS:        Assignment (op, alpha_convert_lval f l, alpha_convert f r) *)
            Assignment (op, f l, alpha_convert f r)
	| Binary (FieldOffset, l, fld) -> Binary (FieldOffset, alpha_convert f l, fld)
	| Binary (op, l, r) -> Binary (op, alpha_convert f l, alpha_convert f r)
	| Cast (ty, e) -> Cast (ty, alpha_convert f e)
	| Constant _ as e -> e
	| Constructor _ as e -> e
	| FunctionCall (fct, args) ->
            FunctionCall (alpha_convert f fct, List.map (alpha_convert f) args)
	| Sizeof _ as e -> e
	| Unary (op, e) -> Unary (op, alpha_convert f e)
    end 
  and alpha_convert_lval f =
    function
      | Access (op, e, label) -> Access (op, alpha_convert f e, label)
      | Symbol sym -> f (Symbol sym)
      | Dereference exp -> Dereference (alpha_convert f exp)
      | Indexed (exp1, exp2) -> Indexed (alpha_convert f exp1, alpha_convert f exp2)


  (** deep_alpha_convert f: lval -> expr 
    where f: lval -> expr essentially f has the chlval information ... this substitutes the lval at the
    top level and then goes inside and recursively goes in and substitutes
    *)
  let rec deep_alpha_convert f = 
    begin
      function
	| Lval l -> 
	    begin
	      match f l with
		  Lval l' -> Lval (deep_alpha_convert_lval f l')
		| Chlval (l',s) -> Chlval (deep_alpha_convert_lval f l',s)
		| _ -> failwith "deep_alpha_convert fails to pattern match"
	    end
	| Chlval (l,s) -> Chlval (deep_alpha_convert_lval f l,s)
	    (* was 
	       begin
	       match l with
	       Symbol _ -> Chlval (l, s)
	       | Access (op, e, label) -> Chlval (Access (op, alpha_convert f e, label), s)
	       | Dereference e -> Chlval (Dereference (alpha_convert f e), s)
	       | Indexed (e1, e2) -> Chlval (Indexed (alpha_convert f e1, alpha_convert f e2), s)
	       end
	    *)
	| Assignment (op, l, r) ->
	    begin
	      match deep_alpha_convert f (Lval l) with
		  Lval l' -> Assignment ( op, l', deep_alpha_convert f r)
	      |	_ -> failwith "deep_alpha_convert : strange assignment"
	    end

	| Binary (FieldOffset, l, r) -> Binary (FieldOffset, deep_alpha_convert f l, r)
	| Binary (op, l, r) -> Binary (op, deep_alpha_convert f l, deep_alpha_convert f r)
	| Cast (ty, e) -> Cast (ty, deep_alpha_convert f e)
	| Constant _ as e -> e
	| Constructor _ as e -> e
	| FunctionCall (fct, args) ->
            FunctionCall (deep_alpha_convert f fct, List.map (deep_alpha_convert f) args)
	| Sizeof _ as e -> e
	| Unary (op, e) -> 
	    begin
	      match op with
		Address -> Unary (op, deep_alpha_convert f e)
	      | _ -> Unary (op, deep_alpha_convert f e)
	    end
    end 
  and deep_alpha_convert_lval f =
    function
      | Access (op, e, label) -> Access (op, deep_alpha_convert f e, label)
      | Symbol sym -> (Symbol sym)
      | Dereference exp -> Dereference (deep_alpha_convert f exp)
      | Indexed (exp1, exp2) -> Indexed (deep_alpha_convert f exp1, deep_alpha_convert f exp2)
	  


  (* substitute -- this is required by wp *)
                   (* substitute in an expression. *)
                   (* Arguments :
                      exp : Expression
                      pairs:  [(old1,new1),...,(old_n, new_n)], where new_i and old_i are
                      Expression.expression's *)
                   (* return exp[new_i/old_i] *)  
  let rec substituteExpression pairs exp =
    let subsExpInLval l =
      match l with 
	Access(op, e, sym) -> Access(op, substituteExpression pairs e, sym)
      |	Dereference e -> push_deref (Dereference (substituteExpression pairs e))
      |	Symbol _ -> l
      |	Indexed (e1, e2) -> Indexed (substituteExpression pairs e1, substituteExpression pairs e2)
    in
    (try List.assoc exp pairs
     with Not_found ->
       match exp with
	 Lval l -> Lval (subsExpInLval l)
       | Chlval (l, s) -> 
	   begin
	     match l with
	       Symbol _ -> Chlval (l, s)
	     | Access (op, e, label) -> Chlval (Access (op, substituteExpression pairs e, label), s)
	     | Dereference e -> Chlval (push_deref (Dereference (substituteExpression pairs e)), s)
	     | Indexed (e1, e2) -> Chlval (Indexed (substituteExpression pairs e1, substituteExpression pairs e2), s)
	   end
       | Binary(FieldOffset, e1, e2) -> Binary(FieldOffset,
                                      substituteExpression pairs e1,
                                      e2)
       | Binary(op, e1, e2) -> Binary(op,
                                      substituteExpression pairs e1,
                                      substituteExpression pairs e2)
       | Constant(_) -> exp
       | Unary(op, e) ->
           (match op with
                  (* don't substitute underneath an & operator *)
             Address -> exp
           | _ -> Unary(op, substituteExpression pairs e))
       | _ -> failwith ("substitute: not implemented. Offending expression "^(toString exp)))


  let rec assignments_of_expression exp =
    match exp with
        Assignment (op, l, r) -> (Lval l, r)::(assignments_of_expression r)
      | Binary (op, l, r) -> (if (op = Eq) then [(l,r)] else [])@((assignments_of_expression l)@(assignments_of_expression r))
      | Cast (ty, e) -> assignments_of_expression e
      | Unary (op, e) -> assignments_of_expression e
      | _ -> []

  end

(**
   A predicate is a boolean combination of atomic expressions.
   *)
module Predicate =
    struct
      type predicate =
	  True
	| False
	| And of predicate list
	| Or of predicate list
	| Not of predicate
	| Implies of predicate * predicate
	| Iff of predicate * predicate
	| All of Symbol.symbol * predicate (* Forall quantifier -- symbol is the variable 
					      being quantified
					      *)
	| Exist of Symbol.symbol * predicate (* Exists quantifier -- symbol is the variable 
					      being quantified
					      *)
	| Atom of Expression.expression


      let rec toString pred =
	match pred with
	  True -> "true"
	| False -> "false"
	| And plist -> "And ["^(toStringList plist)^"]"
	| Or plist -> "Or ["^(toStringList plist)^"]"
	| Not p -> "Not ("^(toString p)^")"
	| Implies (p1, p2) -> "("^(toString p1)^"=>"^(toString p2)^")"
	| Iff (p1, p2) -> "("^(toString p1)^"<=>"^(toString p2)^")"
	| All (sym, p) -> "All "^(Symbol.toString sym)^". "^(toString p)
	| Exist (sym, p) -> "Exists "^(Symbol.toString sym)^". "^(toString p)
	| Atom expr -> (Expression.toString expr)
      and toStringList plist =
	List.fold_left (fun a -> fun b -> a ^ "," ^ (toString b)) "" plist

      let print fmt pr = Format.pp_print_string fmt (toString pr)



      let rec normalize pred =
	let drop_equal plist =
	  List.fold_left (fun a -> fun b -> if (List.mem b a) then a else b::a) [] plist
	in
 	let flatten predlist i =
	  if (i==0) (* And *)
	  then
	    begin
	      List.fold_left 
		(fun a -> fun b ->
		  begin
		    match b with
		      And pl -> List.append pl a 
		    | _ -> b::a
		  end) 
		[] predlist 
	    end
	  else (* Or *)
	    begin
	      List.fold_left 
		(fun a -> fun b ->
		  begin
		    match b with
		      Or pl -> List.append pl a 
		    | _ -> b::a
		  end) 
		[] predlist 
	    end
	in
	match pred with
	  And plist ->
	    begin
	      match plist with
	      	[]  -> True
	      | [a] -> normalize a
	      | _   -> 
		  begin
		    let normlist = 
		      List.filter (fun a-> (a!= True)) 
			(drop_equal (flatten (List.map normalize plist) 0))
		    in
		    if (List.mem False normlist)
		    then False
		    else
		      begin
			match normlist with 
			  [] -> True
			| [a] -> a
			| _ -> And normlist
		      end
		  end
	    end
	| Or plist ->
	    begin
	      match plist with
	      	[]  -> False
	      | [a] -> normalize a
	      | _   -> 
		  begin
		    let normlist = 
		      List.filter (fun a-> (a!=False)) 
			(drop_equal (flatten (List.map normalize plist) 1))
		    in
		    if (List.mem True normlist)
		    then True
		    else
		      begin
			match normlist with
			  [] -> False
			| [a] -> a
			| _ -> Or normlist
		      end
		  end
	    end
	| Implies (p, q) -> Implies (normalize p, normalize q)
	| Iff (p, q) -> Iff (normalize p, normalize q)
	| Not p -> Not (normalize p)
	| All (a,b) -> All (a, normalize b)
	| Exist (a,b) -> Exist (a, normalize b)
	| _ -> pred

      let conjoinL plist = normalize (And (plist))
	  
      let disjoinL plist = normalize (Or (plist))

  let extract_terms_and p = 
	match p with 
	  And l -> l
	| _ -> [p]
       
      let extract_terms_or p =
	  match p with 
	  Or l -> l
	| _ -> [p] 
	      
      let rec unzip ll = 
	match ll with
	  [t] ->  List.map (function x -> [x] ) t
	| h::tl ->  List.flatten (List.map (function x -> (List.map (function y -> x::y) (unzip tl) )) h)
	| [] -> []

      let rec to_dnf p =
	match p with 
	| And l -> (List.map conjoinL (unzip (List.map to_dnf l)))
	| Or l -> (List.flatten (List.map to_dnf l))
	|  _ -> [p]


      let implies p1 p2 = normalize (Implies (p1, p2))

      let iff p1 p2 = normalize (Iff (p1, p2))
	 
      let rec negate p =
	match p with
	  True -> False
	| False -> True
	| And pl -> Or (List.map negate pl)
	| Or pl -> And (List.map negate pl)
	| Not q -> q
	| Implies (p,q) -> And [p; negate q]
	| Iff (p,q) -> Or [ And [p; negate q] ; And [negate p; q] ]
	| All (a,b) -> Exist (a, negate b)
	| Exist (a,b) -> All (a, negate b)
	| Atom _ -> Not p

      let negateAtom p =
	match p with
	  Atom e -> Atom (Expression.negateRel e)
	| _ -> failwith "Predicate.negateAtom expects an atomic predicate!"

      let rec convertDNF p = (* EXPENSIVE !! *)
	match p with
	  True -> True
	| False -> False
	| Or plist -> normalize (Or (List.map convertDNF plist))
	| And plist -> (* difficult case *)
	    begin
	      let listOfOrs = List.map (fun a -> normalize (convertDNF a)) plist in
	      let (pureOrs, rest) = List.partition (function a -> match a with Or _ -> true | _ -> false) listOfOrs
	      in
	      let cartesianProduct ll =
		let  cartprodIt xs ys = List.fold_right 
		    (function x -> function pairs -> 
		      List.fold_right (function y -> function l -> 
			(x::y):: l) ys pairs) 
		    xs []
		in
		List.fold_right cartprodIt ll [[]]
	      in
	      let andLists = 
		cartesianProduct (List.map (function p -> match p with Or a -> a | _-> failwith "convertDNF:only Ors") pureOrs)
	      in
	      normalize (Or (List.map (function a -> normalize (And (a @ rest))) andLists)) 
	    end
	| Not pr -> 
	    begin
	      match pr with
		Atom _ -> p
	      |	_ -> convertDNF (negate p)
	    end
	| Atom _ -> p
	| Implies (p1, p2) -> Implies (convertDNF p1, p2)
	| _ -> failwith ("convertDNF: Unexpected input "^(toString p))
		
  (* return the Expressions of the form x, x.f, *x, &x, x->f and ( *x ).f in the 
     given Predicate.  we assume those are the only forms. 
     also, because of our assumption that whole-struct assignments do not occur, we
     do not go "inside" an expression of the form ( *x ).f or x->f.  that is, we return it, but not
     *x as a variable in the expression.  *)
	      
      let rec allVarExps pred =
	match pred with
          Atom e -> Expression.allVarExps e
        |        True -> []
        |        False -> []
        |        And plist -> List.fold_left (fun a -> fun b -> Misc.union a (allVarExps b)) [] plist
        |        Or plist -> List.fold_left (fun a -> fun b -> Misc.union a (allVarExps b)) [] plist
        |        Implies (p1,p2) -> Misc.union (allVarExps p1) (allVarExps p2)
        |        Iff (p1,p2) -> Misc.union (allVarExps p1) (allVarExps p2)
        |        Not p -> allVarExps p
        |        _ -> failwith "allvarExps : Not implemented"
	      

(* Deep allVarExp -- go inside the x.foo, x -> foo etc. to return "x" *)
      let rec allVarExps_deep pred =
	let rec _allvars exp =
	  match exp with
	  | Expression.Lval l -> l::(_allvarsLval l)
	  | Expression.Chlval(l,_) -> [l] (* failwith "chlval not handled"*)
      	  | Expression.Assignment(op, target, value) -> 
	      Misc.union (_allvarsLval target) (_allvars value)
	  | Expression.Binary (Expression.FieldOffset, e1, e2) -> 
	      _allvars e1
      	  | Expression.Binary (op, e1, e2) -> 
	      Misc.union 
		(_allvars e1) 
		(_allvars e2)
      	  | Expression.Constant(_) -> []
      	  | Expression.FunctionCall (func, args) -> 
	      Misc.union (_allvars func) 
		(List.fold_left Misc.union [] (List.map _allvars args))
      	  | Expression.Unary(op, e) -> 
	      (match op with
	      |	Expression.Address -> _allvars e
	      |	_ -> _allvars e) 
      	  | _ -> [] (* Jhala: Say some sort of message ?? *)
	and _allvarsLval l =
      	  match l with
      	  | Expression.Symbol(sym) -> [] (* [sym] *)
	  | Expression.Dereference e ->
	      begin
		(*match e with 
		    Expression.Lval (Expression.Symbol (x)) -> [x]
		  |	_ ->*) _allvars e
	      end
    	  | Expression.Access(Expression.Dot, e, sym) -> 
	      ((*match e with
		Expression.Lval (Expression.Symbol(x)) -> [x] 
	      |	Expression.Lval (Expression.Dereference (Expression.Lval (Expression.Symbol(x)))) -> 
		  [x]
	      |	_ ->*) _allvars e)
      	  |	Expression.Access(Expression.Arrow, e, sym) -> _allvars e
	  | Expression.Indexed (e1, e2) -> Misc.union (_allvars e1) (_allvars e2)
	in
	match pred with
	  Atom e -> _allvars e
	|	True -> []
	|	False -> []
	|	And plist -> List.fold_left (fun a -> fun b -> 
	    Misc.union a (allVarExps_deep b)) [] plist
	|	Or plist -> List.fold_left (fun a -> fun b -> Misc.union 
	      a (allVarExps_deep b)) [] plist
	|	Implies (p1,p2) -> Misc.union (allVarExps_deep p1) 
	      (allVarExps_deep p2)
	|	Iff (p1,p2) -> Misc.union (allVarExps_deep p1) 
	      (allVarExps_deep p2)
	|	Not p -> allVarExps_deep p
	|	_ -> failwith "allvarExps_deep : Not implemented"
	      
	      


      let rec getAtoms pred =
        match pred with
	  True -> []
        | False -> []
        | And plist -> List.fold_left (fun l1 l2 -> Misc.union l1 (getAtoms l2)) [] plist
        | Or plist -> List.fold_left (fun l1 l2 -> Misc.union l1 (getAtoms l2)) [] plist
        | Not p -> getAtoms p
        | Implies (p1, p2) -> Misc.union (getAtoms p1) (getAtoms p2)
        | Iff (p1, p2) -> Misc.union (getAtoms p1) (getAtoms p2)
        | Atom expr -> [pred]
        | _ ->
            failwith "getAtoms: input predicate should quantifier-free!"

  module SymbolSet = Set.Make(
    struct type t = Expression.lval
      let compare = Pervasives.compare
    end
      )
  let map sigma eta p0 =
    let rec traverse bound = function
      | True | False as p -> p
      | And ps -> And (List.map (traverse bound) ps)
      | Or ps -> Or (List.map (traverse bound) ps)
      | Not p -> Not (traverse bound p)
      | Implies (p1, p2) -> Implies (traverse bound p1, traverse bound p2)
      | Iff (p1, p2) -> Iff (traverse bound p1, traverse bound p2)
      | All (sym, p) -> All (sigma sym, traverse (SymbolSet.add (Expression.Symbol sym) bound) p)
      | Exist (sym, p) -> Exist (sigma sym, traverse (SymbolSet.add (Expression.Symbol sym) bound) p)
      | Atom e -> Atom (eta (fun sym -> SymbolSet.mem sym bound) e)
    in
    traverse SymbolSet.empty p0

  let alpha_convert f =
    map (fun x -> x) 
      (function is_bound ->
         Expression.alpha_convert
         (fun lv -> if is_bound lv then lv else f lv))
      

  let deep_alpha_convert f = 
    map (fun x -> x) 
      (function is_bound ->
         Expression.deep_alpha_convert
         (fun lv -> if is_bound lv then Expression.Lval lv else f lv))
     
    (* Substitute In A Predicate *)
    (* Arguments:
       f: Predicate
       pairs:  [(old1,new1),...,(old_n, new_n)], where new_i and old_i are
       Expression.expression's *)
    (* return f[new_i/old_i] *)
    (* NOTE:  We make use of the fact that the only things we are substituting
       for are expressions of the form x and *x. *)

  let rec substitute pairs f =
    match f with
        True      -> True
      |        False     -> False
      |        Atom e    -> Atom (Expression.substituteExpression pairs e)
      |        And plist -> And (List.map (substitute pairs) plist)
      |        Or plist  -> Or (List.map (substitute pairs) plist)
      |        Not p     -> Not (substitute pairs p)
      |        Implies (p1, p2) -> Implies (substitute pairs p1, substitute pairs p2)
      |        Iff (p1, p2) -> Iff (substitute pairs p1, substitute pairs p2)
      |        _         -> failwith "Substitute: Quantified predicate not implemented"

    (* We will represent predicates in (pseudo-)canonical form:
       - Equalities rather than disequalities
       - <= rather than >=
       - < rather than >
       - lexicographically smaller symbol to the left in case of sym1 <relop> sym2
       *)
       
  let rec canonicize pred = 
    match pred with
	Atom e ->
	  begin
	    let atm = 
	      match e with
		  Expression.Binary (Expression.Ne, 
				     Expression.Lval (Expression.Symbol sym1), 
				     Expression.Lval (Expression.Symbol sym2)) ->
				       if (sym1 < sym2) then 
					 Expression.Binary (Expression.Eq, 
							    Expression.Lval (Expression.Symbol sym1), 
							    Expression.Lval (Expression.Symbol sym2))
				       else
					 Expression.Binary (Expression.Eq, 
							    Expression.Lval (Expression.Symbol sym2), 
							    Expression.Lval (Expression.Symbol sym1))
		| Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol sym1), Expression.Lval (Expression.Symbol sym2)) ->
		    if (sym1 < sym2) then 
		      Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol sym1), 
					 Expression.Lval (Expression.Symbol sym2))
		    else
		      Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol sym2), 
					 Expression.Lval (Expression.Symbol sym1))

		| Expression.Binary (Expression.Ne, Expression.Constant c, e1) 
		| Expression.Binary (Expression.Eq, Expression.Constant c, e1) ->
		    Expression.Binary (Expression.Eq, e1, Expression.Constant c)
		      
		|	Expression.Binary (Expression.Ne, ee, Expression.Lval (Expression.Symbol sym2)) ->
			  Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol sym2), ee) 
		|	Expression.Binary (Expression.Eq, ee, Expression.Lval (Expression.Symbol sym2)) ->
			  Expression.Binary (Expression.Eq, Expression.Lval (Expression.Symbol sym2), ee)
		|	Expression.Binary (Expression.Ne, e1, e2) -> Expression.Binary(Expression.Eq, e1, e2)
		|	Expression.Binary (Expression.Ge, e1, e2) -> Expression.Binary(Expression.Le, e2, e1)
		|	Expression.Binary (Expression.Gt, e1, e2) -> Expression.Binary(Expression.Lt, e2, e1)

                (* Further syntactic matches to take care of predicates produced by Foci *)
		| Expression.Binary (Expression.Le, Expression.Constant (Constant.Int c1), e2) ->
		    begin
		      match e2 with
			Expression.Binary(Expression.Plus, Expression.Constant (Constant.Int c), Expression.Lval l) ->
			  Expression.Binary (Expression.Le, Expression.Constant (Constant.Int (c1-c)), Expression.Lval l)
		      |	Expression.Binary(Expression.Plus, 
					  Expression.Constant (Constant.Int c), 
					  Expression.Unary (Expression.UnaryMinus, Expression.Lval l)) ->
			  Expression.Binary (Expression.Le, Expression.Lval l, Expression.Constant (Constant.Int (c-c1)))
		      |	_ -> e
		    end

		      (* anything else: leave as is *)
 		      
		|	_ -> e
	    in
	      Atom atm
	  end
      |	_ -> failwith ("canonicize: Predicate "^(toString pred)^" is not atomic")
end
 
