(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: ast.mli,v 1.12 2003/06/19 18:07:50 rupak Exp $
 *
 * This file is part of the BLAST Project.
 *)


(**
 * This module defines the internal representation used
 * by BLAST to represent expressions and predicates.
 *)


type ast_type = string (* change later *)


module Symbol : 
sig 
  type symbol = string (* change later *)
  type t = symbol
  val compare : t -> t -> int (* So that Symbol : Set.OrderedType *)
  val toString : symbol -> string 
  val print : Format.formatter -> symbol -> unit 
end

module Constant :
sig
  type constant = 
      Int     of int		(* SLAM uses int32?? so that we get all 32 bits *)
    | Float   of float
    | String  of string
    |	EnumC   of int * string (* SLAM uses int32 *)

  val toString : constant -> string
end

module Expression : 
sig
  type access_op = Dot | Arrow
      
  type assign_op = 
      Assign | AssignBitAnd | AssignBitOr | AssignDiv | AssignLShift | AssignMinus
    | AssignMul | AssignPlus | AssignRem | AssignRShift | AssignXor
	  
  type unary_op  = 
      UnaryMinus | Not | BitNot | Address | UnaryPlus | DestructorCall | Assume | Assert
    | PreInc | PreDec | PostInc | PostDec
	  
  type binary_op = 
      BitAnd | BitOr | Div | LShift | Minus | Mul | Plus | Rem | RShift | Xor 
    | Eq | Ge | Gt | Le | Lt | Ne | Index | And | Or | Comma 
    | FieldOffset

  type lval =	
      Symbol of Symbol.symbol
    | Access of access_op * expression * Symbol.symbol
    | Dereference of expression
    | Indexed of expression * expression
  and
	expression =
      Lval of lval 
    | Chlval of lval * string
    | Assignment of assign_op * lval * expression (* op, target, value *)
    | Binary of binary_op * expression * expression
    | Cast of ast_type * expression	(* only deal with cast (T)e *)
    | Constant of Constant.constant
    | Constructor of Symbol.symbol * ast_type
    | FunctionCall of expression * expression list
    | Sizeof of int
    | Unary of unary_op * expression

  (** t and compare are added to make Expression a Set.OrderedType.
      With these, we can put expressions in sets.
      The type t is of course the same as the type expression,
      and compare is the system compare
      *)
  type t = expression
  val compare : t -> t -> int

  val unaryOpToString : unary_op -> string 
  
  (** a vanilla string is a string over the alphabet 
     \[a-zA-Z0-9\] ; it is useful because the theorem provers/other backend tools dislike
     expressions like "|". Vanilla string for this is BitOr.
  *)
  val unaryOpToVanillaString : unary_op -> string 

  val assignOpToString : assign_op -> string

  val accessOpToString : access_op -> string 

  val binaryOpToString : binary_op -> string
  val binaryOpToVanillaString : binary_op -> string

  (** {b Printers.} *)

  val print : Format.formatter -> expression -> unit
  val print_lval : Format.formatter -> lval -> unit
  val toString : expression -> string
  val lvalToString : lval -> string

  (** {b Miscellaneous utilities.} *)

  (** Push all address operators to the variables.
      See the memory model discussions.
      *)
  val addressOf : expression -> expression

  (** replace all occurrences of a Chlval by corresponding Lval *)
  val replaceChlvals : expression -> expression

  val isRelOp : binary_op -> bool  
  (* check if an Expression is of the form *p or ( *p ).f or p->f *)
  val isDeref : expression -> bool
  
  (* push dereferences below address of operations *)
  val push_deref : lval -> lval
 
  (* given a deref expression *p or ( *p ).f or p->f, return the derefed symbol p *)
  val derefedVar : lval -> expression
  (* get the variables p s.t. *p or ( *p ).f or p->f appear in varExps *)
  val getParents : lval list -> expression list
  (* return the expressions of the form x, x.f, *x, &x, x->f, and ( *x ).f in the 
     given expression.  we assume those are the only forms. 
     also, because of our assumption that whole-struct assignments do not occur, we
     do not go "inside" an expression of the form x->f or ( *x ).f .  that is, we return it, but not
     *x as a variable in the expression. *)
  val allVarExps : expression -> expression list
  (* Negate a relational expression. *)
  val negateRel : expression -> expression

  (** returns a chlval that is essentially a "Fresh" version of that lval *)
  val fresh_lval : lval -> expression

  (* substitute *)
  val alpha_convert : (lval -> lval) -> expression -> expression
  val alpha_convert_lval : (lval -> lval) -> lval -> lval
 
  (** deep_alpha_convert f: lval -> expr 
    where f: lval -> expr essentially f has the chlval information ... this substitutes the lval at the
    top level and then goes inside and recursively goes in and substitutes
    *)
  val deep_alpha_convert : (lval -> expression) -> expression -> expression

  val substituteExpression : (expression * expression) list -> expression -> expression
  val assignments_of_expression : expression -> (expression * expression) list

end

(**
 * A predicate is a boolean combination of atomic expressions.
 * The signature allows All and Exist quantifiers, but we never
 * use them. 
 *)
module Predicate :
sig
  type predicate =
      True
    | False
    | And of predicate list
    | Or of predicate list
    | Not of predicate
    | Implies of predicate * predicate
    | Iff of predicate * predicate
    | All of Symbol.symbol * predicate (* Forall quantifier -- symbol is the variable 
					  being quantified *)
    | Exist of Symbol.symbol * predicate (* Exists quantifier -- symbol is the variable 
					    being quantified *)
    | Atom of Expression.expression

  (** Various useful normalizations : e.g., And \[...; False; ...\] -> False etc *)
  val normalize : predicate -> predicate 

  val conjoinL : predicate list -> predicate (* And of predicate list *)
  val disjoinL : predicate list -> predicate (* Or of predicate list *)
  val implies : predicate -> predicate -> predicate
  val iff : predicate -> predicate -> predicate
  (* Negate a predicate: does not go inside Atoms *)
  val negate : predicate -> predicate
  (* Negate an atomic predicate deeply *)
  val negateAtom : predicate -> predicate

  (** Convert a boolean expression to DNF form. *)
  val convertDNF : predicate -> predicate
  val to_dnf : predicate -> predicate list
  
  val toString : predicate -> string
  val print : Format.formatter -> predicate -> unit


  val allVarExps : predicate -> Expression.expression list
  val allVarExps_deep : predicate -> Expression.lval list
  
  val getAtoms : predicate -> predicate list

  (** [map sym_f expr_f p] ==> [p']
      Return a predicate with the same structure, and [sym_f] applied to each
      *bound* symbol at the place where it is bound, and [expr_f] applied to
      each expression. The first argument passed to [expr_f] is a function
      that tests whether its argument is *bound*. *)
  val map : (Symbol.symbol -> Symbol.symbol) ->
            ((Expression.lval -> bool) ->
              Expression.expression -> Expression.expression) ->
            predicate -> predicate
  val alpha_convert : (Expression.lval -> Expression.lval) -> predicate -> predicate
  val deep_alpha_convert : (Expression.lval -> Expression.expression) -> predicate -> predicate
  (** substitute *)
  val substitute: (Expression.expression * Expression.expression) list -> predicate -> predicate

  val canonicize : predicate -> predicate
end

