(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: blastArch.ml,v 1.32 2003/08/14 04:09:56 jhala Exp $
 *
 * This file is part of the BLAST Project.
 *)


(**
 * This module defines the main signatures for Blast.
 *)


(**
 * The PRINTABLE signature specifies a type equipped with a print function
 * that allows to pretty print values of this type.
 *)
module type PRINTABLE =
sig
  (* the type *)
  type t

  (** the pretty-printer for this type *)
  val print : Format.formatter -> t -> unit

  (** the string converter for this type *)
  val to_string : t -> string
end


(**
 * The COMMAND signature specifies a set of commands.  These commands
 * typically label the edges of a System description.
 *)
module type COMMAND =
sig
  include PRINTABLE
end

(**
 * The SYSTEM_DESCRIPTION signature specifies how an input system shall be
 * described.  This description is an automaton-like representation suitable
 * for model checking.
 * 
 *)
module type SYSTEM_DESCRIPTION =
sig
  module Command : COMMAND

  (** the type of (control) locations *)
  type location

  (** the type of edges (i.e. control transitions) *)
  type edge

  (** a command that does nothing *)
  val skip : Command.t

  (** a location, used to build regions where only the data part is relevant *)
  val fake_location : location

  (** get the source position for a location *)
  val get_source_position : location -> (string * int * int) option

  (** the pretty-printer for locations *)
  val print_location : Format.formatter -> location -> unit

  (** the pretty-printer for edges *)
  val print_edge     : Format.formatter -> edge     -> unit

  (** the string converter for locations *)
  val location_to_string : location -> string


  val location_coords : location -> int * int

  (** the string converter for edges *)
  val edge_to_string : edge -> string

  (** Extracting an edge's associated command *)
  val edge_to_command : edge -> Command.t
      
  (** gives the list of edges originating from a given location *)
  val get_outgoing_edges : location -> edge list

  (** gives the list of edges pointing to a given location *)
  val get_ingoing_edges : location -> edge list

  (** gives the source of an edge *)
  val get_source  : edge -> location

  (** gives the target of an edge *)
  val get_target  : edge -> location

  (** gives the command labeling an edge *)
  val get_command : edge -> Command.t

  val reads_and_writes_at_edge : edge -> Ast.Expression.lval list * Ast.Expression.lval list

  (** for an expression x returns *x, **x etc etc ... for now only goes down a depth of 2 *)
  val expression_closure : Ast.Expression.expression -> Ast.Expression.expression list


  (** Call a function on every global variable (noit including tid and
      notTid) *)
  val iter_global_variables : (Ast.Expression.lval -> unit) -> unit

  val iterate_all_lvals : (Ast.Expression.lval -> unit) -> unit


  (** Functions that return the Cil type of an lvalue.
	These are quite fragile. Use at your own risk.
  *)
  
  (** return the type of an lvalue *)
  val get_type_of_lval : Ast.Expression.lval -> Cil.typ
  (** return the type of an expression.
      Right now this only deals with Lvals and Address-es
      *)
  val get_type_of_exp : Ast.Expression.expression -> Cil.typ

  val is_array : Ast.Expression.lval -> bool
  val is_ptr : Ast.Expression.lval -> bool


  

  (** returns true if the symbol is in the table of global variables *)
  val is_global : Ast.Expression.lval -> bool

  (** returns true if the lvalue can escape. Uses the alias analyzer *)
  val can_escape : Ast.Expression.lval -> bool

  (** returns true if the string is the name of a lock variable *)
  val is_a_lock : Ast.Expression.lval -> bool

  (** A variable that contains the id of the running thread. *)
  val tid : string

  (** A variable that is always different from tid *)
  val notTid : string

  (** A variable that refers to a generic memory location *)
  val self : string

  (** gives the entry location of a function *)
  (* Greg: strings for function calls: to change? *)
  exception NoSuchFunctionException of string
  
  (** returns true if the function is defined, false otherwise *)
  val is_defined : string -> bool
  
  (** find the entry location for a function *)
  val lookup_entry_location : string -> location

 (** Given some location, return what function it belongs to *)
  val get_location_fname : location -> string

  (** return list of functions *)
  val list_of_functions : unit -> string list

   (** is a given location a loopback location ? *)
  val is_loopback : location -> bool


  (* val instrument_CFA : unit -> unit *)

  (** Type of formal arguments to a function. *)
  type formals_kind =
      Fixed of string list
    | Variable of string list (** variable arguments *)

  (** gives the locals of a function *)
  val lookup_locals : string -> string list

  (** gives the formals of a function *)
  val lookup_formals : string -> formals_kind

  (** gives the list of global variables *)
  val globals : unit -> string list

  (** makes a sym_var out of a var : make_symvar fname var *) 
  val make_symvar : string -> string -> string

  (** the locations corresponding to a label name in the C source *)
  val get_locations_at_label : string -> location list

  (** map_edges f applies f to all the edges in the program *)
  val map_edges : (edge -> 'a) -> 'a list

  (** map_edges f fname applies f to all the edges in the function fname *)
  val map_edges_fname : (edge -> 'a) -> string -> 'a list

  (** call graph related functions *)
(* a call node describes the local calling structure for a
 * single function: which functions it calls, and which
 * functions call it *)
type callnode = {
  (* the function this node describes *)
  cnInfo: string;

  (* set of functions this one calls *)
  cnCallees: (string, callnode) Hashtbl.t;

  (* set of functions that call this one *)
  cnCallers: (string, callnode) Hashtbl.t;
}

(* a call graph is a hashtable, mapping a function name to
 * the node which describes that function's call structure *)
type callgraph =
  (string, callnode) Hashtbl.t


  val compute_callgraph : unit -> unit
  val get_callgraph : unit -> callgraph
  val get_callers : string -> (string, callnode) Hashtbl.t
  val get_callees : string -> (string, callnode) Hashtbl.t

  (* update_callgraph caller callee *)
  val update_callgraph : string -> string -> unit
  (* output call graph to file in dot format *)
  val output_callgraph : string (* filename *) -> unit
  val print_callgraph : out_channel -> unit

end


(**
 * The OPERATION signature specifies a set of operation.  Operations are used
 * in Abstractions whereas Commands are used in System Descriptions.  In
 * pratice, Operations can be e.g. edges of a System Description.
 *
 * @see 'blastCSystemDescr.ml' for operations for C programs
 *)
module type OPERATION =
sig
  include PRINTABLE

  type exp
  type lval

  type info =
    | Normal
    | Call of lval option * string * exp list
    | Ret of exp

  val get_info : t -> info
end


(**
 * The REGION signature specifies a printable type used to denote sets,
 * equipped with equality, inclusion, (approximate) union, (approximate)
 * intersection, a bottom element for the empty set, and a function to check
 * for emptiness.
 *
 * Regions are an abstract description of sets of states. 
 * We do not make a distinction between "abstract regions" and "concrete
 * regions" at this level, but see {!BlastArch.ABSTRACTION}
 *
 * @see <http://www.eecs.berkeley.edu/~rupak/Papers/a_classification_of_sts.ps>
 *       to see a discussion on regions in model checking
 *)
module type REGION =
sig
  include PRINTABLE

  (** the equivalence relation corresponding to set equality *)
  val eq : t -> t -> bool

  (** the quasi-ordering corresponding to set inclusion *)
  val leq : t -> t -> bool

  (** the (approximate) union operation *)
  val cup : t -> t -> t

  (** the (approximate) intersection operation *)
  val cap : t -> t -> t

  (** the region representing the emptyset *)
  val bot : t

  (** tests whether a region denotes an empty set *)
  val is_empty : t -> bool

  (** make a region concrete *)
  val concretize : t -> t

  val rename_vars : t -> (Ast.Expression.lval -> Ast.Expression.lval) -> t
  val change_location_and_stack : t -> t -> t


  (* coverers store data structures for quick lookup of a region which covers another *)
  type 'a coverers
  val emptyCoverers : unit -> 'a coverers
  val clearCoverers : 'a coverers -> unit
  val addCoverer : 'a coverers -> t -> 'a -> unit
  val findCoverer : 'a coverers -> t -> 'a option

  (* Restricting regions to contain only certain kinds of predicates *)
  type restriction
  val negate_restriction : restriction -> restriction
  val locals_and_globals : string -> restriction
  val only_globals : restriction
  val only_locals : string -> restriction
  val restrict : restriction -> t -> t
end

module type ACCESS_REGION = sig
  module Region : REGION

  type 'a lock_predicate = {
      local_lp_region : Region.t;
      lp_region : Region.t;
      lp_node : 'a;
    }

  (* The type of access regions (\phi's for race condition detection).
     Each component of an access region has a ``node'' information of type
     ['a]. *)
  type 'a t

  (* The empty access region *)
  val empty : 'a t

  (* [add_region n r phi interfering]
     Add a region [r] to the access region [phi]. The node for [r] is [n].
     The region [r] must have been stripped of its location component for
     this function to make sense when analysing a multi-threaded program.
     Return [New_region phi'] if [r] does not interfere with [phi], i.e., 
     if [interfering r ri] is false for every [ri] in that has previously
     been added to [phi]. Otherwise return [Error_at lp] where [lp] contains
     the region and node for the component of [phi] that interferes with
     [r]. *)
  type 'a add_region_outcome =
    | New_region of 'a t
    | Error_at of 'a lock_predicate * Region.t

  val add_region : 'a -> Region.t -> Region.t -> 'a t ->
                   (Region.t -> Region.t -> Region.t option) ->
                   'a add_region_outcome

  val concrete : 'a t -> Region.t

  val to_string : 'a t -> string
end

(**
 * The ABSTRACTION signature groups an operation module and a region module
 * and specifies abstract post and pre functions.  These two functions are
 * supposed to be conservative with respect to the precise semantics of the
 * operations.  It also specifies the precise and focus functions used for
 * refinement.
 *)
module type ABSTRACTION =
sig
  (** the operation module *)
  module Operation : OPERATION

  (** the region module *)
  module Region : REGION

  (** create alias database *) (* Code for David *)
  (* val create_alias_database : unit -> unit  *)

  (** creates a string representation of an (atomic) region -- used by gui *)
  val reg_to_strings : Region.t -> (string list * string )

  (** gets the name of the function for an atomic region *)
  val reg_function_name : Region.t -> string

  (** gets a string deswcription of a region *)
  val reg_to_string : Region.t -> string

  (** creates a string from an operation *)
  val op_to_string : Operation.t -> (int * int * string)

(*  module AR : ACCESS_REGION with module Region = Region*)

  (** the table mapping variables names to phi regions *) 
  type phi_table = (Ast.Expression.lval, Region.t ref) Hashtbl.t 

  (** the symbolic post function: (approximate) strongest post-condition *)
  val post : Region.t -> Operation.t -> phi_table -> Region.t

  (** the symbolic pre function: (approximate) weakest pre-condition *)
  val pre  : Region.t -> Operation.t -> phi_table -> Region.t

  (** [summary_post caller exit] creates a region to be used after the function call whose
    * post-condition caller represents, where exit is the region of the used summary
    * node's exit
    *)
  val summary_post : Operation.t -> Region.t -> Region.t -> phi_table -> Region.t

  (** this does pre only it makes syntactic substitutions explicit. This is required
    * in counterexample analysis
    *
    * @see "abstraction.ml" and the operator {!BlastArch.ABSTRACTION.focus}
    *)
  val spre : Region.t -> Operation.t -> phi_table -> Region.t    

  (** forget all data information in a region *)
  val erase_data : Region.t -> Region.t

  (** make a region out of a predicate *)
  val predicate_to_region : Ast.Predicate.predicate -> Region.t

  (** create a concrete "true" region with location as the destination of the operation **)
  val create_concrete_true_region : Operation.t -> Region.t

  (** create an abstract "true" region with location as the destination of the operation **)
  val create_abstract_true_region : Operation.t -> Region.t

  (** keep only the data part of a region, and concretize it *)
  val trim_region_to_globals : Region.t -> Region.t
  
  (** **)
  val new_name : string -> Ast.Expression.lval -> Ast.Expression.lval
  val brand_new_suffix : unit -> string
  val assumed_invariants_region : Region.t ref
  (** the concretization function: take an abstract region and concretize it *)
  val precise  : Region.t -> Region.t

  (** Raised when no new predicates are found after counterexample analysis. (Blast fails) *)
  exception NoNewPredicatesException

  (** the refinement operator *)
  val focus  : Region.t -> Region.t -> Region.t

    (** a dummy refinement operator that really checks if the supplied region is as precise as it can be *)
  val focus_stamp : Region.t -> Region.t 

  (** analyze a trace using uclid *)
  val analyze_trace : Region.t list -> Operation.t list -> unit

  (** analyze a trace as a bunch of blocks foci *)
  val block_analyze_trace : Region.t list -> Operation.t list -> int list -> (int * Region.t)

    (** analyze all subtraces of a trace to find all possible predicates *)
  val find_all_predicates : Region.t list -> Operation.t list -> unit

    (** have we not discovered new predicates since the last time this function was called ? *)
  val predicate_fixpoint_reached : unit -> bool


  (** check if it is even worthwhile to check if the region is covered *)
  val coverable : Region.t -> bool

  (** the quasi-ordering corresponding to coverability, that is, basically,
     Region.leq modulo "safety" equivalence *)
  val covered : Region.t -> Region.t -> bool

  (** abstracts reg2 wrt preds of reg1 and then takes the intersection *)
  val intersect_with_abstraction : Region.t -> Region.t -> Region.t

  (** allows the printing of various statistics *)
  val print_abs_stats : unit -> unit

  (** resets statistics ... *)
  val reset_abs_stats : unit -> unit

  (** gives the possibly enabled operations in a given region *)
  val enabled_ops  : Region.t -> Operation.t list

  val print_debug_info : Operation.t list -> unit

  (** returns true if the operation is an assume predicate *)
  val isPredicate : Operation.t -> bool

  (** returns true if the operation is an assume block *)
  val isBlock : Operation.t -> bool


  (** tells you whether or not to bother to check for emptiness now *)
  val check_emptiness : Operation.t -> bool

  val reset : unit -> unit
  (* prepare restart the analysis on the same program with possibly
     different havocs. *)
 
  (** checks if the lvalue can ever be aliased to a global variable. 
      Uses alias information from AliasAnalyzer
      *)
  val can_escape : Ast.Expression.lval -> bool

  val is_lock : Ast.Expression.lval -> bool

      (** {b Data Structures to represent Proofs} *)

  type proof_region
  type lemma_index
  val post_proof: Region.t -> Operation.t -> phi_table -> (proof_region * lemma_index)
  val check_reg_eq: proof_region -> Region.t -> bool
  val extract_proof_region: Region.t -> proof_region
  val print_proof_region : proof_region -> unit
  val print_proofTable : unit -> unit
      
  (** {b Fields related to race condition detection} *)

  (** Invoke a function on each global variable *)
  val iter_global_variables : (Ast.Expression.lval -> unit) -> unit

  (** Check whether an operation reads or writes to a certain variable *)
  type access_type = Anone | Aread | Awrite | Aboth
  
  (** checks how variable string is accessed in Operation.t *)
  val accessing : Operation.t -> Ast.Expression.lval -> access_type
  
  (** returns all strings that are written by this operation *)
  val get_all_writes : Operation.t -> Ast.Expression.lval list
  (** returns all strings that are read by this operation *)
  val get_all_reads : Operation.t -> Ast.Expression.lval list
  
  val reads_and_writes : Operation.t -> (Ast.Expression.lval list * Ast.Expression.lval list)
  val aliasesOf : Ast.Expression.lval -> Ast.Expression.lval list

  (** [multiply_instantiable r r'] is true iff $r[\th\gets\th_1] \wedge
      r'[\th\gets\th_2] \implies \th_1 = \th_2$. *)
  val multiply_instantiable : Region.t -> Region.t -> Ast.Expression.lval -> Ast.Expression.lval -> Region.t option
  val make_multiply_instantiable_region : Region.t -> Region.t -> Ast.Expression.lval -> Ast.Expression.lval -> Region.t


   (* Interface checking *)
   type spec
   type spec_result =
     | Yes
     | No
     | Maybe of Ast.Predicate.predicate

   (** Prepares to check specifications. Call once after all normal checking is complete. *)
   val init_spec : unit -> unit
   (** Builds a predicate that describes a function's observed input-output behavior,
    *  as shown by a list of function entry points paired with their observed exit points.
    *)
   val build_spec : string -> (Region.t * Region.t list) list -> spec
   (** Generates a predicate equivalent to the implementation of the second parameter spec
    *  by the first parameter spec. (i.e., first parameter => second parameter) *)
   val check_spec : spec -> spec -> spec_result

  val predIndex : int ref
end

module ModelCheckOutcome =
struct
  (* Outcome type polymorphic in error_path type *)
  type 'a outcome =
    | No_error
    | Error_reached of 'a
    | Race_condition of 'a * 'a * Ast.Expression.lval * Ast.Expression.lval
    | Bad_lock_spec of 'a * Ast.Expression.lval
end

(**
 * The LAZY_MODEL_CHECKER specifies the functionality expected from the
 * Lazy Model Checker.
 *)
module type LAZY_MODEL_CHECKER =
sig
  (** the abstraction module used by the model checker *)
  module Abstraction : ABSTRACTION

  (** the type of error paths *)
  type error_path = (Abstraction.Region.t *
                     (Abstraction.Operation.t list) *
                     Abstraction.Region.t)

  type model_check_outcome = error_path ModelCheckOutcome.outcome

  (** the TAR algorithm implementation 
   * {[model_check init error]}
   * searches for a path from {i init} to {i error} in the
   * model and returns an error path or that the system is correct   
  *)
  val model_check_tar : Abstraction.Region.t  ->
                     Abstraction.Region.t option   ->
                     model_check_outcome

  (** the lazy model checking algorithm implementation 
   * {[model_check init error]}
   * searches for a path from {i init} to {i error} in the
   * model and returns an error path or that the system is correct   
  *)
  val model_check : Abstraction.Region.t  ->
                     Abstraction.Region.t option   ->
                     model_check_outcome

end
