(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(**
   Depends on:
     Ast

**)


module Symbol = Ast.Symbol
module Constant = Ast.Constant
module Expression = Ast.Expression
module Predicate = Ast.Predicate

module FakeAliasAnalyzer = (* To hack around the problem of alias analysis
			      calls made by the TAR-havoc code, we introduce
			      this analyzer
			      *)
struct
  let getAllAliases lval = []
end


module C_Command =
struct
  type statement =
      Expr of Expression.expression
    | Return of Expression.expression (* option *)

  type command =
      Skip
    | Block of statement list
      (* Remark: at the C level, a predicate is just another expression
         but it may prove useful to differentiate them when we can (e.g.
         for conditionals.) *)
    | Pred of Predicate.predicate
      (* Remark: the expression of a function call is assumed to be
         either:
           - an Expression.FunctionCall(_), or
           - an Expression.Assignment (_, _, Expression.FunctionCall(_)) *)
    | FunctionCall of Expression.expression
    | Havoc of Expression.lval (*Symbol.symbol*)
        (* [Havoc x]
           Forget all that is known about x, unless \phi(x) is known to
           hold. This action models the behavior of other threads. *)
    | HavocAll
        (* Havoc all global variables *)
    | Phi of Ast.Expression.lval * bool
        (* [Phi (x, sign)]
           This is semantically equivalent to [Pred <<sign iff \phi(x)>>]. *)

  type t = {
      code : command;
      read : Expression.lval list;
      written : Expression.lval list;
    }

  let local_cmd code = {code = code; read = []; written = []}
  let skip = local_cmd Skip
  let cmd_true = local_cmd (Pred Predicate.True)


  let gather_accesses code =
    let module S = Set.Make(
      struct type t = Expression.lval
	    let compare = Pervasives.compare
      end
	    ) 
    in
    let is_global sym =
      not (String.contains sym '@')
    in
      (* JHALA: note that expr_w is not used anywhere ! 
	 expr_w w  = function
	 | Expression.Lval lv -> lvalue_w w lv
	 | Expression.Chlval _ -> failwith "gather_accesses : Chlval unexpected!"
	 | Expression.Binary (Expression.FieldOffset, e1, e2) -> w (* CHECK THIS *) 
	 | Expression.Binary (op, e1, e2) -> let w' = expr_w w e1 in expr_w w' e2
	 | Expression.Cast (ty, e) -> expr_w w e
	 | Expression.Constant _ -> w
	 | Expression.Constructor _ -> w
	| Expression.FunctionCall (e, es) -> List.fold_left expr_w w (e :: es)
	 | Expression.Sizeof _ -> w
         | Expression.Unary (Expression.AddressOf, e) -> w (* nothing under an address-of is written *)
	 | Expression.Unary (op, e) -> expr_w w e
	 | e -> failwith ("expr_w : Strange expression "^(Expression.toString e))
	 and *)

(* JHALA: so in the next 2 functions,
   the l is an lval, but we have to 
   recursively go down inside it ... *)

  let rec
      lvalue_w (r, w as rw) l = 
      let (r', w') = 
	match l with
	  | Expression.Symbol s -> rw
	  | Expression.Access (op,e,s) -> expr rw e
	  | Expression.Dereference e -> expr rw e 
	  | Expression.Indexed (e1, e2) -> 
	      begin 
		let rw' = expr rw e1 in expr rw' e2
	      end
      in
	(r', S.add l w')
	
    and
      lvalue (r, w as rw) l = 
      
      let (r',w') = 
	match l with
	  | Expression.Symbol s -> rw
	  | Expression.Access (op,e,s) -> expr rw e
	  | Expression.Dereference e -> expr rw e 
	  | Expression.Indexed (e1, e2) -> 
	      begin
		let rw' = expr rw e1 in
		expr rw' e2
	      end
      in
	(S.add l r', w')
    and
      expr (r, w as rw) = function
	| Expression.Lval lv -> lvalue rw lv
	| Expression.Chlval _ -> failwith "gather_accesses : Chlval unexpected!"
	| Expression.Assignment (op, lv, e2) ->
            let rw' = expr rw e2 in
  	      lvalue_w rw' lv 
	| Expression.Binary (Expression.FieldOffset, e1, e2) -> expr rw e1 
	| Expression.Binary (op, e1, e2) -> let rw' = expr rw e1 in expr rw' e2
	| Expression.Cast (ty, e) -> expr rw e
	| Expression.Constant _ -> rw
	| Expression.Constructor _ -> rw
	| Expression.FunctionCall (e, es) -> List.fold_left expr rw ((* e ::*) es)
	    (* when the day dawns where we deal with function pointers we shall throw the "e" back in  
	       but for now it puts the function names in which is not so nice... *)
	| Expression.Sizeof _ -> rw
	| Expression.Unary (Expression.Address, e) -> rw (* don't go below address-ofs *)
	| Expression.Unary (op, e) -> expr rw e
    in
    let rec predicate rw = function
      | Predicate.True | Predicate.False -> rw
      | Predicate.And ps -> List.fold_left predicate rw ps
      | Predicate.Or ps -> List.fold_left predicate rw ps
      | Predicate.Not p -> predicate rw p
      | Predicate.Implies (p1, p2) -> predicate (predicate rw p1) p2
      | Predicate.Iff (p1, p2) -> predicate (predicate rw p1) p2
      | Predicate.All (sym, p) -> predicate rw p
      | Predicate.Exist (sym, p) -> predicate rw p
      | Predicate.Atom e -> expr rw e
    in
    let collect (r, w) =
      { code = code;
	read = S.fold (fun v vs -> v :: vs) r [];
	written = S.fold (fun v vs -> v :: vs) w [] }
    in
      match code with
	| Skip -> skip
	| Block statements ->
	    collect (List.fold_left
		       (fun rw block ->
				  expr rw (match block with Expr e -> e | Return e -> e))
		       (S.empty, S.empty) statements)
	| Pred p -> collect (predicate (S.empty, S.empty) p)
	| FunctionCall e -> collect (expr (S.empty, S.empty) e)
	| Havoc _ | HavocAll -> assert false
	| Phi _ -> assert false
	    
  (**RUPAK -- old version
  let gather_accesses code =
    let module S = Set.Make(Symbol) in
    let is_global sym =
      not (String.contains sym '@')
    in

    let rec
	expr_w w  = function
	  | Expression.Lval lv -> lvalue_w w lv
	  | Expression.Chlval _ -> failwith "gather_accesses : Chlval unexpected!"
	  | Expression.Binary (op, e1, e2) -> let w' = expr_w w e1 in expr_w w' e2
	  | Expression.Cast (ty, e) -> expr_w w e
	  | Expression.Constant _ -> w
	  | Expression.Constructor _ -> w
	  | Expression.FunctionCall (e, es) -> List.fold_left expr_w w (e :: es)
	  | Expression.Sizeof _ -> w
	  | Expression.Unary (op, e) -> expr_w w e
	  | e -> failwith ("expr_w: Strange expression = " ^ (Expression.toString e))
    and
	lvalue_w w = function
      | Expression.Symbol sym ->
          if is_global sym then S.add sym w else w
      | (Expression.Dereference e1) -> expr_w w e1
      | Expression.Access (_, e1, field) ->
          expr_w w e1
      | _ -> failwith ("BlastCSystemDescr.C_Command.gather_accesses.lvalue ")
    and
	lvalue (r, w as rw)  = function
	  | Expression.Access (op, e, label) -> expr rw e
	  | Expression.Symbol sym -> (if is_global sym then S.add sym r else r), w
	  | Expression.Dereference e -> expr rw e
	  | Expression.Indexed (e1, e2) -> let rw' = expr rw e1 in expr rw' e2 
    and
	expr (r, w as rw) = function
	  | Expression.Lval lv -> lvalue rw lv
	  | Expression.Chlval _ -> failwith "gather_accesses : Chlval unexpected!"
	  | Expression.Assignment (op, e1, e2) ->
              let r', w' = expr rw e2 in
              r', lvalue_w w e1
	  | Expression.Binary (op, e1, e2) -> let rw' = expr rw e1 in expr rw' e2
	  | Expression.Cast (ty, e) -> expr rw e
	  | Expression.Constant _ -> rw
	  | Expression.Constructor _ -> rw
	  | Expression.FunctionCall (e, es) -> List.fold_left expr rw (e :: es)
	  | Expression.Sizeof _ -> rw
	  | Expression.Unary (op, e) -> expr rw e
    in
    let rec predicate rw = function
      | Predicate.True | Predicate.False -> rw
      | Predicate.And ps -> List.fold_left predicate rw ps
      | Predicate.Or ps -> List.fold_left predicate rw ps
      | Predicate.Not p -> predicate rw p
      | Predicate.Implies (p1, p2) -> predicate (predicate rw p1) p2
      | Predicate.All (sym, p) -> predicate rw p
      | Predicate.Exist (sym, p) -> predicate rw p
      | Predicate.Atom e -> expr rw e
    in
    let collect (r, w) =
      { code = code;
        read = S.fold (fun v vs -> v :: vs) r [];
        written = S.fold (fun v vs -> v :: vs) w [] }
    in
    match code with
    | Skip -> skip
    | Block statements ->
        collect (List.fold_left
                   (fun rw block ->
                     expr rw (match block with Expr e -> e | Return e -> e))
                   (S.empty, S.empty) statements)
    | Pred p -> collect (predicate (S.empty, S.empty) p)
    | FunctionCall e -> collect (expr (S.empty, S.empty) e)
    | Havoc _ | HavocAll -> assert false
    | Phi _ -> assert false
**)

  (* the pretty-printer for this type *)
  let print fmt cmd =
    Format.fprintf fmt "@[" ;
    begin
      match cmd.code with
          Skip ->
            Format.fprintf fmt "Skip"

        | Block(l) ->
            let print_statement = function
                Expr(e) ->
                  Expression.print fmt e ;
                  Format.fprintf fmt ";"

              | Return(e) ->
                  Format.fprintf fmt "Return(" ;
                  Expression.print fmt e ;
                  Format.fprintf fmt ");"
            in
              Format.fprintf fmt "Block(" ;
              List.iter print_statement l ;
              Format.fprintf fmt ")"

        | Pred(p) ->
            Format.fprintf fmt "Pred(" ;
            Predicate.print fmt p ;
            Format.fprintf fmt ")"

        | FunctionCall(e)  ->
            Format.fprintf fmt "FunctionCall(" ;
            Expression.print fmt e ;
            Format.fprintf fmt ")"
        | Havoc v ->
            Format.fprintf fmt "Havoc (%s)" (Expression.lvalToString v)
        | HavocAll ->
            Format.fprintf fmt "HavocAll"
        | Phi (v, sign) ->
            Format.fprintf fmt "Pred(%s\\phi(%s))"
                               (if sign then "" else "\\not") (Expression.lvalToString v)
    end ;
    Format.fprintf fmt "@]"

  (* the string converter for this type *)
  let to_string = Misc.to_string_from_printer print
end


module C_System_Descr =
struct

  module CFA = BlastControlFlowAutomaton



  (* The list of CIL files that were processed.
     This is initially "None".
     It is set in initilize 
     *)
  let global_cil_files : Cil.file list option ref = ref None
  
  (* get the list of Cil files.
     This call makes sense only after the files have been parsed.
     *)
  let get_cil_files () =
    match !global_cil_files with
      Some f -> f
    | None -> failwith "No CIL file found!"
	  

  (*******************************************************)
  (** Parameter variables                                *)
  (** These are variables with special meaning for blast *)
  (*******************************************************)

  let tid = "tid"       (* thread id of this thread    *)
  let notTid = "notTid" (* thread id of another thread *)
  let self = "self"     (* generic memory location     *)
  let self_lval = Expression.Symbol self
      
  (************************************************************************)
  (* Some generic utilities to get information out of the Cil structures. *)
  (************************************************************************)
      
  (* Given a string, find the varinfo structure corresponding to that
     variable name (ie with the same vname)
     
   Since, in the files we consider, all variables have unique names,
     the list vinfoL will have exactly one member.
     
     Returns an empty list if there is no variable with that name.
     *)
  let vinfoL = ref []
      
  class getVarinfoClass str = object
    inherit Cil.nopCilVisitor
	
	
    method vvdec vardec =
      if (vardec.Cil.vname = str) then
        vinfoL := vardec :: !vinfoL
      else ();
      Cil.DoChildren
	
  end
      
	(* given a string, return the Cil varinfo associated with it,
	   and raise Not_found if there is no such variable.
	   *)
  let get_varinfo str =
    let getVars str file =
      let ob = new getVarinfoClass str in
      let _ = Cil.visitCilFile ob file in
      let vL = !vinfoL
      in
      vinfoL := []; vL
    in
    let cil_files = get_cil_files () in
    let vars =  List.fold_left (fun vars f -> List.append (getVars str f) vars) [] cil_files
    in
    match vars with
      [] -> raise Not_found
    | [vinfo] -> vinfo
    | vinfo:: _ -> (*Message.msg_string Message.Normal ("Warning: More than one varinfo found for the string "^str) ;*) vinfo
	  
	  (* get the Cil.typ type of an lvalue *)
  let rec get_type_of_lval lv =
(*    Message.msg_string Message.Normal ("get_type_of_lval : "^(Expression.lvalToString lv)) ; *)
    match lv with
      Expression.Symbol s -> 
	begin
	  if Misc.is_prefix "BLAST_UNKNOWN" s then Cil.intType
	  else
	  try
	    Cil.unrollType (get_varinfo s).Cil.vtype
	  with
	    Not_found -> failwith ("get_type_of_lval: No such variable "^s)
	end
    |	Expression.Dereference e ->
	begin
	  let etyp = Cil.unrollType (get_type_of_exp e) in
	  match etyp with
	    Cil.TPtr (t, att) -> Cil.unrollType t
	  | _ -> failwith ((Expression.toString e)^ "is Not a pointer type "^(Expression.lvalToString lv))
	end
    |	Expression.Access (Expression.Dot, e, fld) ->
	begin
	  let etyp = Cil.unrollType (get_type_of_exp e) in
	  match etyp with
	    Cil.TComp (c, att) -> 
	      begin
		let fldinfo = Cil.getCompField c fld in
		Cil.unrollType fldinfo.Cil.ftype
	      end
	  | _ -> failwith "Not a composite type"	
	end
    |	Expression.Access (Expression.Arrow, e, fld) ->
	get_type_of_lval (Expression.Access (Expression.Dot, Expression.Lval (Expression.Dereference e), fld))
    |	Expression.Indexed (e1, e2) ->
	  begin
	  let etyp = Cil.unrollType (get_type_of_exp e1) in
	  match etyp with
	    Cil.TArray (at, _, _) -> 
	      begin
		Cil.unrollType at
	      end
	  | _ -> failwith "Not an array type"	
	  end
	  (* get the type of an expression *)
  and get_type_of_exp exp =
    begin
      match exp with
	Expression.Lval l -> get_type_of_lval l
      | Expression.Chlval _ -> failwith ("get_type : Chlval not handled "^(Expression.toString exp))
      | Expression.Binary (_, _, _) -> failwith ("get_type :  not handled "^(Expression.toString exp))
      |	Expression.Unary (Expression.Address, e) -> 
	  Cil.TPtr (get_type_of_exp e, [])
      | Expression.Unary (_, _) ->failwith ("get_type :  not handled "^(Expression.toString exp))
      |	Expression.Constant c ->
	  begin
	    match c with
	      Constant.Int _ -> Cil.intType
	    | Constant.Float _ -> Cil.doubleType
	    | _ -> failwith ("get_type : not handled "^(Expression.toString exp))
	   
	  end
      | _ -> failwith ("get_type : not handled "^(Expression.toString exp))
    end
      
	  (* Given an "access" lvalue, find the Cil fieldinfo associated with it *)
  let get_fieldinfo lval =
    match lval with
      Expression.Access (Expression.Dot, e, fld) ->
	begin
	  try
	    let typ = get_type_of_exp e
	    in
	    match typ with
	      Cil.TComp (c, _) -> Cil.getCompField c fld
	    | _ -> failwith "get_fieldinfo : Strange type "
	  with
	    Not_found -> failwith ("get_fieldinfo : No such field " ^ (Expression.lvalToString lval))
	end
    |  Expression.Access (Expression.Arrow, e, fld) ->
	begin
	  try
	    let typ = get_type_of_exp e
	    in
	    match typ with
	      Cil.TPtr (Cil.TComp (c, _), _) -> Cil.getCompField c fld
	    | _ -> failwith "get_fieldinfo : Strange type "
	  with
	    Not_found -> failwith ("get_fieldinfo : No such field " ^ (Expression.lvalToString lval))
	end
    | _ -> failwith "get_fieldinfo: Not a field access"
	  
	  

   let is_array lval =
     let tp = get_type_of_lval lval in
     match tp with
       Cil.TArray _ -> true
     | _ -> false

   let is_ptr lval =
     let tp = get_type_of_lval lval in
     match tp with
       Cil.TPtr _ -> true
     | _ -> false

  (*********************************************************************************************)
  (* End general Cil utilities                                                                 *)
  (*********************************************************************************************)

  (*********************************************************************************************)
  (* Blast internal data structures                                                            *)
  (*********************************************************************************************)

  (* Source position information in Blast *)
  type source_position = {
      file_name : string ;
      line      : int ;
      column    : int ;
    }

  let print_source_position fmt sp =
    Format.fprintf fmt "@[src=\"%s\";@ line=%d@]" sp.file_name sp.line
   
  let cil_location_to_source_position cil_loc =
    { line      = cil_loc.Cil.line ;
      column    = -1 ;
      file_name = cil_loc.Cil.file ;
    }
      
  let source_position_to_string = Misc.to_string_from_printer print_source_position
  (* End Source position information in Blast *)

  type location_attributes =
      Original of source_position
    | Artificial

  let print_location_attributes fmt = function
      Original (sp) ->
        Format.fprintf fmt "@[Original(" ;
        print_source_position fmt sp ;
        Format.fprintf fmt ")@]"
    | Artificial ->
        Format.fprintf fmt "@[Artificial@]"

  type automaton_attributes = {
    label_table : (string, (automaton_attributes,
                            location_attributes,
                            C_Command.t) CFA.location) Hashtbl.t ;
  }

  type formals_kind =
      Fixed of Symbol.symbol list
    | Variable of Symbol.symbol list

  let print_formals fmt = function
      Fixed l -> (Misc.list_printer_from_printer Symbol.print) fmt l
    | Variable l -> (Misc.list_printer_from_printer Symbol.print) fmt (l@["..."])

  type function_definition = {
    function_name    : string ;
    formals          : formals_kind ;
    locals           : string list ;
    cfa              : (automaton_attributes,
                        location_attributes,
                        C_Command.t) CFA.automaton ;
    fun_src_pos      : source_position ;
    fun_attributes   : Cil.attributes ; }

  type variable_definition = {
    var_name         : string ;
    var_src_pos      : source_position ;
    var_is_lock      : bool ;
    var_vaddrof : bool ; (* the Cil vaddrof flag: true if the address of this variable has been taken *)
    }

  let global_function_table = Hashtbl.create 31
  let global_label_table = Hashtbl.create 31


  (************************************************************************)
  (* Global lvalues and associated information maintained as a hash table *)
  (* Functions for handling global lvalues                                *)
  (************************************************************************)
  (*                                                                                          *)
  (* Iterate over all lvalues in a Cil file and add them to the all_lvals_hashtbl *)
	  
  let all_lvals_hashtbl = Hashtbl.create 31 
      
  class addLvalClass = object  
    inherit Cil.nopCilVisitor
	
    method vlval lv =
      let blast_lval = BlastCilInterface.convertCilLval lv in
      if Hashtbl.mem all_lvals_hashtbl blast_lval then () 
      else
	begin
	  
                                                       (* keep the original Cil lval as well,
                                                          to get type info etc.
                                                          *)
          match blast_lval with
            Expression.Symbol s -> 
	      if (Misc.is_prefix s "__BLAST_NONDET") 
	      then () 
	      else
		Hashtbl.add all_lvals_hashtbl blast_lval (Some lv)  
		  
          | Expression.Dereference e -> 
	      Hashtbl.add all_lvals_hashtbl blast_lval (Some lv)  
          | Expression.Access (Expression.Arrow, Expression.Lval l, _) -> 
              Hashtbl.add all_lvals_hashtbl blast_lval (Some lv) ;
              (* add *l to lvals *)
              if not (Hashtbl.mem all_lvals_hashtbl l) then Hashtbl.add all_lvals_hashtbl l None 
          | Expression.Access (Expression.Dot, Expression.Lval l, _) -> 
              Hashtbl.add all_lvals_hashtbl blast_lval (Some lv) ;
              (* add l to lvals *)
              if not (Hashtbl.mem all_lvals_hashtbl l) then Hashtbl.add all_lvals_hashtbl l None 
	  | Expression.Indexed (e1, e2) -> 
	      Hashtbl.add all_lvals_hashtbl blast_lval (Some lv) 
	  | _ -> failwith ("add_lvals : Strange lval "^(Expression.lvalToString blast_lval))
	end ;
      Cil.DoChildren
	
  end
      
  let add_all_lvals cil_file =
    let ob = new addLvalClass in
    let _ = Cil.visitCilFile ob cil_file
    in
    ()
      
  let iterate_all_lvals f = Hashtbl.iter (fun lv _ -> f lv) all_lvals_hashtbl

  let remove_lval l = Hashtbl.remove all_lvals_hashtbl l
      
  let print_all_lvals () = 
    iterate_all_lvals 
      (fun l -> 
	Message.msg_string 
	  Message.Normal 
	  ("Lval " ^ (Expression.lvalToString l)))
      
(**********************************************************************************)

  (* Information associated with an lvalue: 
     Cil.varinfo if the lvalue is a variable.
     Cil.fieldinfo if the lvalue is a field access.
     *)
  type global_lval_information = 
      Varinfo of Cil.varinfo
    | Fieldinfo of Cil.fieldinfo
	  
  let global_variable_table = Hashtbl.create 31
      
  (* iterate over the global variables table and apply function f to each element *)
  let iter_global_variables f = Hashtbl.iter (fun name _ -> f name) global_variable_table 
      
  (* TODO : disambiguate two field references of the same name:
     e.g., struct foo { int data ; }
     struct bar { int data ; }
     will now have just one entry : this.data.
     *)
  let rec add_global lval info =
    let rec_add styp =
      let rec walk_ptrs t =
	match t with
	  Cil.TPtr (t', _) -> walk_ptrs t'
	| t' -> t
      in
      match styp with
	Cil.TComp (c, _) -> 
	  begin (* composite type : recursively add all field information to global table. *)
		(* for each fieldinfo, add this.fieldinfo to global table *)
	    List.iter 
	      (fun f -> 
		add_global (Expression.Access (Expression.Dot, Expression.Lval self_lval, f.Cil.fname)) (Fieldinfo f))
	      c.Cil.cfields
	  end
      | Cil.TPtr (pt, _) ->
	  begin
	    match walk_ptrs pt with 
	      Cil.TComp (c, _) ->
				(* for each fieldinfo, add this.fieldinfo to global table *)
		List.iter 
		  (fun f -> 
		    add_global (Expression.Access (Expression.Dot, Expression.Lval self_lval, f.Cil.fname)) (Fieldinfo f))
		  c.Cil.cfields
	    | Cil.TVoid _ -> () (* ignore void * !!  failwith "Do not handle void *"*)
	    | _ -> ()
	  end
      |	_ -> () (* do not add anything *)
    in
    if (Hashtbl.mem global_variable_table lval) then (* this has already been seen *)
      () 
    else
      begin
	Hashtbl.add global_variable_table lval info ;
	match lval with
	  Expression.Symbol s -> 
	    begin
	      if s = self then () 
	      else
		let styp = get_type_of_lval lval in
		rec_add styp
	    end
	| Expression.Access (Expression.Dot, e, fld) ->
	(* if the fieldinfo is composite, recursively add these to the global table *)
	    begin
	      match info with
		Fieldinfo finfo ->
		  begin
		    let ftyp = Cil.unrollType finfo.Cil.ftype in
		    rec_add ftyp
		  end
	      | _ -> failwith "add_global: an access is accompanied by the wrong information!"
	    end
(*
	|  Expression.Access (Expression.Arrow, e, _) 
	|  Expression.Dereference e 
*)
	| _ -> failwith "add_global : Not handled"
      end

(*****************************************************************************************)
	
  let rec is_global s =
    let rec is_global_e e =
      match e with
	Expression.Lval l -> is_global l
      |	Expression.Chlval _ -> false
      |	_ -> failwith ("is_global: Strange expression "^(Expression.toString e))
    in
    match s with 
      Expression.Symbol ss ->
(*	assert (Hashtbl.mem global_variable_table s == not (String.contains ss '@')) ; *)
	not (String.contains ss '@')
    | Expression.Dereference e -> is_global_e e
    | Expression.Access (_, e, _) -> is_global_e e
    | Expression.Indexed (e1, _) -> is_global_e e1

  let can_escape lv =
    Message.msg_string Message.Debug ("IN can_escape, with lvalue "^(Expression.lvalToString lv)) ;
    let rec _can_escape seen lv =
      let rec exp_can_escape e =
	match e with
	|	Expression.Lval l -> _can_escape seen l
	|	Expression.Chlval (l,_) -> 
	    begin
	      _can_escape seen l
	    end
	| Expression.Binary (_, e1, e2) -> exp_can_escape e1 || exp_can_escape e2
	|	Expression.Cast (t, e1) -> exp_can_escape e1
	| Expression.Constant(_) -> false
	|	Expression.Constructor (_, _) -> failwith "Constructor not handled in can_escape"
	| Expression.FunctionCall (func, args) ->
	    begin
	      let rec loop lst =
		match lst with 
		  [] -> false
		| a:: rest -> if exp_can_escape a then true else loop rest
	      in
	      exp_can_escape func || (loop args) 
	    end
	| Expression.Unary(_, e) -> 
	    exp_can_escape e
	|	_ -> failwith "abstractor.can_escape : Strange expression"
      in
      if List.mem lv seen then false else
      begin
	let escapes_directly = 
	  match lv with
	  Expression.Symbol s -> is_global lv 
	  | Expression.Dereference e -> exp_can_escape e 
	  | Expression.Access (_, e, _) -> exp_can_escape e 
	  | Expression.Indexed (e1, e2) -> exp_can_escape e1
	in
	escapes_directly || (List.exists (_can_escape (lv::seen)) (FakeAliasAnalyzer.getAllAliases lv))
      end
    in
    let r = _can_escape [] lv
    in
    Message.msg_string Message.Debug ("can_escape returns " ^ (if r then "true" else "false")) ;
    r

  (*
    Make is_a_lock more comprehensive as follows.
    
    An lvalue is a lock if
    (1) it is a symbol, and the symbol has attribute lock.
    (2) it is a field access, and the field has attribute lock.
    (3) it is a dereference, and *all* basic aliases (i.e., symbols
        and field references) of this dereference are locks.
    (4) it is an array, and *all* elements of the array are locks.
    We need all basic aliases to be locks to catch the (unlikely) case when the dereference
    is used sometimes as a lock, and sometimes as data.
    *)
  let rec is_a_lock v =
    Message.msg_string Message.Normal ("In is_a_lock with lvalue "^(Expression.lvalToString v)) ;
    try
      match v with 
	Expression.Symbol s ->    
	  begin
	    (* only a moron would have a local lock variable. :-0 *)
            (* Prove me wrong? *)
	    match (Hashtbl.find global_variable_table v) with
	      Varinfo cilv -> 
		let ans = Cil.hasAttribute "lock" cilv.Cil.vattr
		in
		Message.msg_string Message.Debug ("Answer to is_a_lock is "^(if ans then "true" else "false")) ;
		ans
	    | _ -> failwith ("Incorrect information in global table associated with variable "^s)
	  end
      | Expression.Access (_, _, fld) -> 
	  begin
	    let l = Expression.Access (Expression.Dot, Expression.Lval self_lval, fld)
	    in
	    match (Hashtbl.find global_variable_table l) with
	      Fieldinfo f -> 
		let ans = Cil.hasAttribute "lock" f.Cil.fattr
		in
		Message.msg_string Message.Debug ("Answer to is_a_lock is "^(if ans then "true" else "false")) ;
		ans
	    | _ -> failwith ("Incorrect information in global table associated with lvalue "^(Expression.lvalToString l))
	  end
      | Expression.Dereference e ->
	  begin
	    let rec all_locks lst at_least_one_lock_seen =
	      Message.msg_string Message.Debug ("In all_locks, flag is "^(if at_least_one_lock_seen then "true" else "false")) ;
	      match lst with
		[] -> at_least_one_lock_seen
	      |	a :: rest ->
		  begin
		    Message.msg_string Message.Debug ("In all_locks, current alias is "^(Expression.lvalToString a)) ;
		    match a with
		      Expression.Symbol s -> 
			begin
			  try
			    match (Hashtbl.find global_variable_table a) with
			      Varinfo vinfo -> 
				if Cil.hasAttribute "lock" vinfo.Cil.vattr
				then
				  all_locks rest true
				else
				  false
			    | _ -> failwith ("Incorrect information in global table associated with variable "^s)
			  with Not_found -> Message.msg_string Message.Debug "Not found! Retur false" ; false
			end
		    | Expression.Access (_, _, fld) ->
			begin		
			  let l = Expression.Access (Expression.Dot, Expression.Lval self_lval, fld)
			  in
			  try
			    match (Hashtbl.find global_variable_table l) with
			      Fieldinfo f -> 
				if Cil.hasAttribute "lock" f.Cil.fattr then
				  all_locks rest true
				else
				  false
			    | _ -> failwith 
				  ("Incorrect information in global table associated with lvalue "^(Expression.lvalToString l))
			  with Not_found -> false (* failwith ("Field "^fld^" not found in global fields table.") *)
			end
		    | _ -> all_locks rest at_least_one_lock_seen
		  end
	    in
	    let list_of_aliases = FakeAliasAnalyzer.getAllAliases v in
	    List.iter (fun l -> print_string (Expression.lvalToString l)) list_of_aliases ;
	    let ans = all_locks list_of_aliases false
	    in
	    Message.msg_string Message.Debug ("Answer is "^(if ans then "true" else "false")) ;
	    ans
	  end
      | Expression.Indexed (e1, e2) -> 
	  begin
	    match e1 with
	      Expression.Lval (l) -> is_a_lock l
	    | _ -> failwith ("is_a_lock: Do not handle indexed lvalues of this type! (To be done)"^(Expression.lvalToString v))
	  end
      | _ -> failwith ("is_a_lock : Do not handle lvalues of this type "^(Expression.lvalToString v))
    with Not_found -> Message.msg_string Message.Debug "is_a_lock: Not found!! Returning false." ; false
	
  let rec is_in_scope x fname =
    let rec is_in_scope_e e fname =
      match e with
	Expression.Lval l -> is_in_scope l fname
      | _ -> 
	  begin
	    Message.msg_string Message.Normal ("Warning: is_in_scope_e : Strange expression "^(Expression.lvalToString x)) ;
	    false
	  end
    in
    match x with
      Expression.Symbol s ->
	begin 	  (* s should have @fname or no \@ *)
	  not (String.contains s '@') ||
	  ((Str.string_after s ((String.index s '@')+1)) = fname)
	end
    | Expression.Dereference e 
    | Expression.Access (_, e, _) ->
	is_in_scope_e e fname
    | Expression.Indexed (e1, e2) ->
  	is_in_scope_e e1 fname


  (********************************************************************************************)


  exception CilAsmUnimplemented

  let skip = C_Command.skip

  let rec cil_instr_to_stmt_and_src_pos cil_instr =
    let convert_assignment target value =
      Expression.Assignment (Expression.Assign,
                             BlastCilInterface.convertCilLval target,
                             BlastCilInterface.convertCilExp value)
    and convert_call fn_exp args =
      Expression.FunctionCall (BlastCilInterface.convertCilExp fn_exp,
                               List.map BlastCilInterface.convertCilExp
                                 args)
    in
      match cil_instr with
          Cil.Set (target, value, loc) ->
            (convert_assignment target value,
             cil_location_to_source_position loc)
        | Cil.Call (None, fn_exp, args, loc) ->
            (convert_call fn_exp args,
             cil_location_to_source_position loc)
        | Cil.Call (Some target, fn_exp, args, loc) ->
            (Expression.Assignment (Expression.Assign,
                                    BlastCilInterface.convertCilLval target,
                                    (convert_call fn_exp args)),
             cil_location_to_source_position loc)
        | Cil.Asm _ -> raise CilAsmUnimplemented

  (* this function splits a CIL instruction list into a list of couples
     (cmd, src_pos) where cmd is a C command (block or function call) and
     src_pos is the source_position of the command's starting point *)
  let rec split_cil_instr_list instr_l =
    match instr_l with
        []          -> []
      | instr::tail ->
          match instr with
              Cil.Set _ ->
                let (stmt, src_pos) = cil_instr_to_stmt_and_src_pos instr
                in
                  split_cil_instr_list_cont tail [stmt] src_pos
            | Cil.Call _ ->
                let (stmt, src_pos) = cil_instr_to_stmt_and_src_pos instr in
		let code = match stmt with
		| Expression.FunctionCall (Expression.Lval (Expression.Symbol "__blast_assume"), [e]) ->
		    C_Command.Pred (Predicate.Atom e)
		| _ -> C_Command.FunctionCall stmt
		in
                (C_Command.gather_accesses code, src_pos) ::
                (split_cil_instr_list tail)
            | Cil.Asm (_, _, _, _, _, loc) ->
                let msg = loc.Cil.file ^ ":" ^ (string_of_int loc.Cil.line) ^
                          ": " ^ "warning: inline assembly not supported --- skipping it."
                in
                  Message.msg_string Message.Minor msg ;
                  split_cil_instr_list tail

  and split_cil_instr_list_cont instr_l curr_block_stmt_l curr_block_src_pos =
    let make_curr_block () =
      let statements =
        List.rev_map (fun x -> C_Command.Expr x) curr_block_stmt_l
      in
      C_Command.gather_accesses (C_Command.Block statements)
    in
    match instr_l with
        [] ->
          (* remark: current_block_stmt_l cannot be empty *)
          [(make_curr_block ()), curr_block_src_pos]
      | instr::tail ->
            match instr with
                Cil.Set _ ->
                  let (stmt, src_pos) = cil_instr_to_stmt_and_src_pos instr
                  in
                    split_cil_instr_list_cont tail (stmt::curr_block_stmt_l)
                                              curr_block_src_pos
              | Cil.Call _ ->
                  let (stmt, src_pos) = cil_instr_to_stmt_and_src_pos instr in
                  let code = C_Command.FunctionCall stmt in
                  let cmd = C_Command.gather_accesses code in
                    ((make_curr_block ()),
                     curr_block_src_pos) ::
                    (cmd, src_pos) ::
                    (split_cil_instr_list tail)
              | Cil.Asm (_, _, _, _, _, loc) ->
                  let msg = loc.Cil.file ^ ":" ^ (string_of_int loc.Cil.line)
                            ^ ": " ^ "warning: inline assembly not supported --- skipping it."
                  in
                    Message.msg_string Message.Minor msg ;
                    split_cil_instr_list_cont tail curr_block_stmt_l
                                              curr_block_src_pos


  exception CilInstrWithIncorrectNumberOfSuccessors
  exception CilGotoWithIncorrectNumberOfSuccessors
  exception CilIfWithIncorrectNumberOfSuccessors
  exception CilAssumeWithIncorrectNumberOfSuccessors
  exception CilBlockWithIncorrectNumberOfSuccessors
  exception CilBreakUnimplemented
  exception CilContinueUnimplemented
  exception CilSwitchUnimplemented
  exception CilBlockWithNoStatement


  module Havoc = struct
    type havoc_implementation =
      | Undifferentiated
      | Locks_and_optimized_data

    let implementation = Undifferentiated


    let insert_havoc automaton l vl =
      let fname = CFA.location_function_name l in
      if not (can_escape vl)
      then begin
        Message.msg_string Message.Debug ("No havoc of " ^ (Expression.lvalToString vl));
        l (* [vl] is actually a function, hence immutable, so don't havoc it *)
      end else
        let l' = CFA.create_location fname automaton Artificial in
        Message.msg_string Message.Debug ("Inserting havoc of " ^ (Expression.lvalToString vl));
        let havoc_cmd =
          { C_Command.code = C_Command.Havoc vl;
            C_Command.read = []; (* TODO: this changes if \phi changes *)
	    C_Command.written = [ ]; }
        in
        CFA.add_edge l l' havoc_cmd;
        l'

    let insert_havocs automaton vars l0 =
      List.fold_left (insert_havoc automaton) l0 vars

    let insert_all_havocs automaton l =
      let fname = CFA.location_function_name l in
      let l' = CFA.create_location fname automaton Artificial in
      Message.msg_string Message.Debug ("Inserting HavocAll");
      let havoc_cmd =
        { C_Command.code = C_Command.HavocAll;
          C_Command.read = []; (* should be all locks *)
          C_Command.written = []; (* should be all globals *) }
      in
      CFA.add_edge l l' havoc_cmd;
      l'

    let global_variables = ref ([] : Ast.Expression.lval list)
        
    let havoc_cmd lv =
      { C_Command.code = C_Command.Havoc lv;
        C_Command.read = [];
	C_Command.written = [] }

    let phi_cmd sym sign =
      { C_Command.code = C_Command.Phi (sym, sign);
        C_Command.read = [ ];
        C_Command.written = [] }

    let und_one automaton choice_loc entry_loc (x : Expression.lval) =
      
      let fname = CFA.location_function_name entry_loc in
      
      let _add_cfa_edges x =
	if ((is_in_scope x fname) && (can_escape x )) then 
	  begin
(*
	    Message.msg_string Message.Normal ("Adding havoc for variable "^(Expression.lvalToString x)) ;
*)
	    let havoc_loc = CFA.create_location fname automaton Artificial in
	    CFA.add_edge choice_loc havoc_loc (phi_cmd x true);
	    CFA.add_edge havoc_loc entry_loc (havoc_cmd x);
	    ()
	  end
	else
	  begin
	    Message.msg_string Message.Debug 
	      ("Not adding havoc for lvalue "^(Expression.lvalToString x)^" (Not in scope, or cannot escape).") ;
	    ()
	  end

      in
      match x with
	Expression.Symbol sym ->
	  begin
	    match sym with
	        "tid" 
	    |	"notTid" 
(*
	    |	"self"
*)
	      -> ()
	    |	_	->
		begin
		  _add_cfa_edges x 
		end
	  end
(* Don't need self anymore...
      |	Expression.Access (Expression.Dot, Expression.Lval (Expression.Symbol "self"), fld) ->
	  begin
	    _add_cfa_edges x
	  end
*)
      |	_ -> _add_cfa_edges x

    let notTid_assumption_cmd =
      let zero = Expression.Constant (Constant.Int 0)
      and not_eq z =
        Predicate.Atom (Expression.Binary (Expression.Ne,
                                           (Expression.Lval (Expression.Symbol notTid)),
                                           z))
      in
      { C_Command.code =
        C_Command.Pred
          (Predicate.And [not_eq zero; not_eq (Expression.Lval (Expression.Symbol tid))]);
        C_Command.read = [];
        C_Command.written = [] }

    let undifferentiated automaton vars l0 =
      let fname = CFA.location_function_name l0 in
      let l0' = CFA.create_location fname automaton Artificial in
      let l0'' = CFA.create_location fname automaton Artificial in
      let l1 = CFA.create_location fname automaton Artificial in
      let l1' = CFA.create_location fname automaton Artificial in
      (* let l2 = CFA.create_location fname automaton Artificial in *)
      CFA.add_edge l0 l0' C_Command.cmd_true;
      CFA.add_edge l0 l0'' C_Command.cmd_true;
      CFA.add_edge l0' l1 (havoc_cmd (Expression.Symbol notTid) );
      CFA.add_edge l1 l1' notTid_assumption_cmd;
      (* CFA.add_edge l1' l2 (havoc_cmd self_lval) ; *)
      List.iter (und_one automaton l1' l0) (List.filter (fun lv -> List.mem lv !global_variables) vars);
      CFA.add_edge l1' l0 C_Command.skip ;
      l0''

    let undifferentiated_all automaton l0 =
      undifferentiated automaton (!global_variables) l0
  end

  let insert_havocs, insert_all_havocs =
    match Havoc.implementation with
    | Havoc.Undifferentiated ->
        Havoc.undifferentiated,
        Havoc.undifferentiated_all
    | Havoc.Locks_and_optimized_data ->
        Havoc.insert_havocs, Havoc.insert_all_havocs

  let cil_fun_block_to_cfa atomic cil_fun_block fun_name =
    let label_table = Hashtbl.create 31 in
    let automaton = CFA.create_automaton { label_table = label_table ; } in
    let end_location = CFA.create_location fun_name automaton Artificial
    and entry_location = CFA.create_location fun_name automaton Artificial in
    let start_location =
      if (Options.getValueOfBool "checkRace") && atomic then begin
        Message.msg_string Message.Debug "Adding all havocs at the start of an atomic function";
        insert_all_havocs automaton entry_location
      end else entry_location
    and void_return_c_cmd =
      { C_Command.code = C_Command.Block [C_Command.Return
                                           (Expression.Constant (Constant.Int 0))];
        C_Command.read = [];
        C_Command.written = [] }
    in
    let get_real_labels cil_stmt =
      List.fold_left (function acc -> function
                          Cil.Label (label, _, b) ->
                            if b then label::acc else acc
                        | _                       ->
                            acc)
                     [] cil_stmt.Cil.labels
    (* This table maps CIL statements to their entry locations.  Several CIL
       statements might be mapped to the same location, see e.g. the
       processing of blocks and rerouting with last_skipped_cil_stmts *)
    and cil_stmt_to_loc_table = Hashtbl.create 31 in
    (* This recursive function is the heart of the automaton creation
       process: the translation from the CIL CFG to our CFA structure
       takes place here.  The meaning of the first 2 arguments is quite
       straightforward.  Let us detail the last one, last_skipped_cil_stmts:
       This list keeps track of the latest consecutive CIL statements we
       skipped during our translation, in order to avoid introducing skips,
       such as empty blocks or empty instruction lists.  This is necessary
       to avoid screwing up the translation of e.g. { label: } along with
       goto label: C constructs -- see rerouting *)
    let rec process_cil_stmt entry_loc cil_stmt last_skipped_cil_stmts =
      (* This entry_loc's attribute should be Artificial, the default
         attribute upon creation *)
      assert ((CFA.get_location_attributes entry_loc) = Artificial) ;
      (* let's first check whether we already processed this statement *)
      if (Hashtbl.mem cil_stmt_to_loc_table cil_stmt.Cil.sid)
      then
        begin
          (* entry_loc should not have any successor *)
          assert ((CFA.get_succs entry_loc) = []) ;
          (* entry_loc should have exactly one predecessor *)
          assert ((List.length (CFA.get_preds entry_loc)) = 1) ;
          let (inc_edge_cmd, inc_edge_loc) = List.hd (CFA.get_preds entry_loc)
          and reroute_loc = Hashtbl.find cil_stmt_to_loc_table
                                         cil_stmt.Cil.sid
          in
            (* We redirect entry_loc's incoming edge to reroute_loc *)
            CFA.remove_edge inc_edge_loc entry_loc inc_edge_cmd ;
            CFA.add_edge inc_edge_loc reroute_loc inc_edge_cmd ;
            (* entry_loc should not have any predecessor now *)
            assert ((CFA.get_preds entry_loc) = []) ;
            (* We update cil_stmt_to_loc_table so that reroute_loc becomes
               the new entry location of all last skipped statements *)
            List.iter (fun s ->
                         Hashtbl.replace cil_stmt_to_loc_table
                                         s.Cil.sid reroute_loc)
                      last_skipped_cil_stmts ;
            (* We update label_table so that reroute_loc becomes the new
               location associated with all labels of all last skipped
               statements *)
            List.iter (fun s ->
                         List.iter (fun l ->
                                      Hashtbl.replace label_table l
                                                      reroute_loc)
                                   (get_real_labels s))
                      last_skipped_cil_stmts
            (* No recursive call, meaning that the CIL successors of this
               CIL statement should not be explored.
               Here the reason is: this statement, and its successors, have
               already been processed *)
        end
      else
        begin
          (* entry_loc becomes the entry location for this statement *)
          Hashtbl.add cil_stmt_to_loc_table cil_stmt.Cil.sid entry_loc ;
          (* update the label table *)
          List.iter (function label ->
                       assert (not (Hashtbl.mem label_table label)) ;
                       Hashtbl.add label_table label entry_loc)
                    (get_real_labels cil_stmt) ;
          match cil_stmt.Cil.skind with
              Cil.Instr (instr_l) ->
                begin
                  let process_cmd entry_loc (cmd, src_pos) =
                    (* A new location has by default an Artificial attribute.
                       If when processing the next statement this location
                       turns out to be the entry location for that statement,
                       then its attributes shall be changed --- see below *)
(*
		    Message.msg_string Message.Normal "Command ";
		    Message.msg_string Message.Normal (C_Command.to_string cmd) ;
		    Message.msg_string Message.Normal "Reads:" ;
		    List.iter (fun a -> Message.msg_string Message.Normal (Expression.lvalToString a)) cmd.C_Command.read ;
		    Message.msg_string Message.Normal "Writes:" ;		    
		    List.iter (fun a -> Message.msg_string Message.Normal (Expression.lvalToString a)) cmd.C_Command.written ;
		    Message.msg_string Message.Normal "Done" ;		    
*)
                    let new_loc = CFA.create_location fun_name automaton Artificial in
                    let start_loc =
                      if (Options.getValueOfBool "checkRace" = false) || atomic then entry_loc else
                      match cmd.C_Command.read, cmd.C_Command.written with
                      | _, [v] when is_a_lock v -> (* mutex action *)
                          Message.msg_string Message.Debug (Printf.sprintf "Adding havocs before action on lock %s at %s" (Expression.lvalToString v) (source_position_to_string src_pos));
                          insert_all_havocs automaton entry_loc
                      | [], [] -> (* local action *)
                          Message.msg_string Message.Debug (Printf.sprintf "Not adding havocs before local action at %s" (source_position_to_string src_pos));
                          entry_loc
                      | [], _ -> (* write action *)
			  Message.msg_string Message.Debug (Printf.sprintf "Not adding havocs before write action at %s" (source_position_to_string src_pos));
                          entry_loc
                      | vars, _ -> (* read action *)
                          Message.msg_string Message.Debug (Printf.sprintf "Adding havocs before read action at %s" (source_position_to_string src_pos));
                          insert_havocs automaton vars entry_loc
                      | _ -> failwith "C_System_Descr.cil_fun_block_to_cfa.process_cil_stmt"
                    in
                      CFA.set_location_attributes start_loc
                                                  (Original src_pos) ;
                      CFA.add_edge start_loc new_loc cmd ;
                    
                      (* Greg: We inssert a skip after each function call,
                         since the abstraction module assumes that the
                         target of each function call has only one incoming
                         edge *)
                      match cmd.C_Command.code with
                          C_Command.FunctionCall _ ->
                            let extra_loc = CFA.create_location fun_name automaton
                                                                Artificial
                            in
                              CFA.add_edge new_loc extra_loc C_Command.skip;
                              extra_loc
                        | _ ->
                            new_loc
                  in
                  let new_loc = List.fold_left process_cmd entry_loc
                                               (split_cil_instr_list instr_l)
                  in
                    (* This CIL statement should have a unique successor *)
                    match cil_stmt.Cil.succs with

                        [next_cil_stmt] ->
                          (* Recursive call, meaning that the CIL successor
                             of this CIL statement should be explored, with
                             an entry location of new_loc *)
                          (* But we first have to check whether new_loc is
                             equal to entry_loc *)
                          if (new_loc == entry_loc)
                          then
                            begin
                              (* this means that instr_l was actually empty *)
                              (* [Greg] well, actually, since we just drop
                                 Cil.Asm instructions (for now), the previous
                                 claim is not true, so I comment the following
                                 assert *)
                              (* assert (instr_l = []) ; *)
                              skip_cil_stmt cil_stmt entry_loc next_cil_stmt
                                            last_skipped_cil_stmts None
                            end
                          else
                            process_cil_stmt new_loc next_cil_stmt []
                      | _ -> raise CilInstrWithIncorrectNumberOfSuccessors
                end

            | Cil.Return (cil_exp_opt, cil_loc) ->
                begin
                  let src_pos = cil_location_to_source_position cil_loc
                  and cmd =
                    match cil_exp_opt with
                        None ->
                          (* [Greg:] Hack to be cleaned...  This is because
                             our Return datatype in C_Command enforces to
                             return something *)
                          void_return_c_cmd
                      | Some (cil_exp) ->
                          let e = BlastCilInterface.convertCilExp cil_exp in
                          let code = C_Command.Block [C_Command.Return e] in
                          C_Command.gather_accesses code
                  in
                    (* all return edges point to the end location *)
                    CFA.set_location_attributes entry_loc (Original src_pos) ;
                    let start_loc =
(*RUPAK: This was:
                      if atomic then entry_loc else 
that is, do not add havocs at the end of atomic routines.
We change it as follows.
*)
		      if (Options.getValueOfBool "checkRace" = false) then entry_loc 
		      else
		      begin
		        (* Do add havocs at the end of atomic functions.
			   This takes care of local variables in atomic functions
			   that go out of scope, yet can be pointing to globals.
		         *)
                        (* A return instruction may not contain assignments or
                           mutex actions *)
                        assert (cmd.C_Command.written = []);
                        Message.msg_string Message.Debug (Printf.sprintf "Adding havocs before read action (return ...) at %s" (source_position_to_string src_pos));
                        insert_havocs automaton cmd.C_Command.read entry_loc
                      end
                    in
                      CFA.add_edge start_loc end_location cmd ;
                    (* This CIL statement should not have any successor *)
                    assert (cil_stmt.Cil.succs = []) ;
                    (* No recursive call, since this CIL statement does not
                       have any CIL successors *)
                end

            | Cil.Goto (next_cil_stmt_ref, cil_loc) ->
                begin
                  let next_cil_stmt = !next_cil_stmt_ref
                  in
                    (* This CIL statement should have a unique successor:
                       !next_cil_stmt_ref *)
                    (match cil_stmt.Cil.succs with
                         [s] -> assert (s == next_cil_stmt)
                       | _   -> raise CilGotoWithIncorrectNumberOfSuccessors) ;
                    (* Recursive call on the successor of this CIL statement *)
                    skip_cil_stmt cil_stmt entry_loc next_cil_stmt
                                  last_skipped_cil_stmts (Some cil_loc)
                end

            | Cil.Break _ -> raise CilBreakUnimplemented
            | Cil.Continue _ -> raise CilContinueUnimplemented

            | Cil.If (cil_exp, block_t, block_f, cil_loc) ->
                begin
                  let src_pos = cil_location_to_source_position cil_loc
                  and (cmd_t, cmd_f) =
                    let pred = BlastCilInterface.convertCilExpToPred cil_exp
                    in
                      (* [Greg] DIRTY... DIRTY... DIRTY...
                         But it's the way we model non-determinism... :-/ *)
                    match pred with
		      (Predicate.Atom (Expression.Binary (_, Expression.Lval (Expression.Symbol nondet), _)))
                      when (Misc.is_prefix "__BLAST_NONDET" nondet) ->
			(C_Command.cmd_true, C_Command.cmd_true)
		    | (Predicate.Atom (Expression.Binary (_, _, Expression.Lval (Expression.Symbol nondet))))
                      when (Misc.is_prefix "__BLAST_NONDET" nondet) ->
			(C_Command.cmd_true, C_Command.cmd_true)
                    | _ ->
                        let cmd_t = C_Command.gather_accesses (C_Command.Pred pred) in
                        (cmd_t,
                         { cmd_t with
                           C_Command.code =
                             C_Command.Pred (Predicate.negate pred)})
                  in
                    CFA.set_location_attributes entry_loc (Original src_pos) ;
                    match cil_stmt.Cil.succs with
                        [next_cil_stmt] ->
                          (* only one successor: this should happen when both
                             blocks are empty.  Let's check it *)
                          assert ((cil_block_is_empty block_t) &&
                                  (cil_block_is_empty block_f)) ;
                          let new_loc = CFA.create_location fun_name automaton Artificial
                          in
                            let start_loc =
                              if (Options.getValueOfBool "checkRace" = false) || atomic then entry_loc else begin
                                (* A predicate may not contain assignments or
                                   mutex actions (?? for the latter) *)
                                assert (cmd_t.C_Command.written = []);
                                Message.msg_string Message.Debug (Printf.sprintf "Adding havocs before read action (if ...) at %s" (source_position_to_string src_pos));
                                insert_havocs automaton cmd_t.C_Command.read entry_loc
                              end
                            in
                            CFA.add_edge start_loc new_loc cmd_t ;
                            CFA.add_edge start_loc new_loc cmd_f ;
                            (* Recursive call, meaning that the CIL successor
                               of this CIL statement should be explored, with
                               an entry location of new_loc *)
                            process_cil_stmt new_loc next_cil_stmt []
                      | [next_cil_stmt_f ; next_cil_stmt_t] ->
                          let new_loc_t = CFA.create_location fun_name automaton Artificial
                          and new_loc_f = CFA.create_location fun_name automaton Artificial
                          in
                            CFA.add_edge entry_loc new_loc_t cmd_t ;
                            CFA.add_edge entry_loc new_loc_f cmd_f ;
                            (* two successors:  Let's check that the CIL
                               successors are correctly set.  We only check
                               the sucessors corresponding to the non-empty
                               blocks *)
                            let check_successors () =
                              if (cil_block_is_empty block_f)
                              then
                                  let first_stmt_t = List.hd block_t.Cil.bstmts
                                  in
                                    (not (cil_block_is_empty block_t)) &&
                                    (first_stmt_t == next_cil_stmt_t)
                              else
                                if (cil_block_is_empty block_t)
                                then
                                    let first_stmt_f = List.hd block_f.Cil.bstmts
                                    in
                                      (first_stmt_f == next_cil_stmt_f)
                                else
                                    let first_stmt_t = List.hd block_t.Cil.bstmts
                                    and first_stmt_f = List.hd block_f.Cil.bstmts
                                    in
                                      (first_stmt_t == next_cil_stmt_t) &&
                                      (first_stmt_f == next_cil_stmt_f)
                            in
                              assert (check_successors ()) ;
                              (* Recursive calls, meaning that the 2 CIL successors
                                 of this CIL statement should be explored, with
                                 entry locations of new_loc_t and new_loc_f *)
                              process_cil_stmt new_loc_t next_cil_stmt_t [] ;
                              process_cil_stmt new_loc_f next_cil_stmt_f []
                      | _ -> raise CilIfWithIncorrectNumberOfSuccessors
                end

            | Cil.Switch _ -> raise CilSwitchUnimplemented

	    | Cil.Loop ({Cil.bstmts =
			 [{Cil.skind = Cil.If (cond,
					       {Cil.bstmts = [{Cil.skind = Cil.Break _}]},
					       {Cil.bstmts = []}, _)}]},
			cil_loc) ->
	      (* C: while (not_cond); *)
              (* Cil: while (1) {if (cond) {break;}} *)
	      (* We prefer to see this as: assume(cond); *)
              let src_pos = cil_location_to_source_position cil_loc in
              let pred = BlastCilInterface.convertCilExpToPred cond in
	      let cmd = C_Command.gather_accesses (C_Command.Pred pred) in
	      let new_loc = CFA.create_location fun_name automaton Artificial in
      	      let start_loc =
		if (Options.getValueOfBool "checkRace" = false) || atomic then entry_loc else begin
                  (* A predicate may not contain assignments or mutex actions
		     (?? for the latter) *)
                  assert (cmd.C_Command.written = []);
                  Message.msg_string Message.Debug (Printf.sprintf "Adding havocs before read action (if ...) at %s" (source_position_to_string src_pos));
                  insert_havocs automaton cmd.C_Command.read entry_loc
		end
	      in
	      CFA.add_edge start_loc new_loc cmd;
	      let next_cil_stmt =
		match cil_stmt.Cil.succs with
		| [s] -> s
		| _ -> raise CilAssumeWithIncorrectNumberOfSuccessors
	      in
	      process_cil_stmt new_loc next_cil_stmt []

              (* We process a CIL loop just as a block, since control flow
                 information from the CIL CFG takes care of the looping,
                 unless the loop's block is empty -- see below *)
            | Cil.Loop (block, _)
            | Cil.Block block ->
		Message.msg_string Message.Debug ("In cil_fun_block_to_cfa: looping on\n" ^
						  Pretty.sprint 100 (Cil.d_stmt () cil_stmt));
                begin
                let next_cil_stmt =
                    (match cil_stmt.Cil.succs with
                         [s] -> s
                       | _   -> raise CilBlockWithIncorrectNumberOfSuccessors)
                  in
                    if (cil_block_is_empty block)
                    then
                      begin
                        (* here we have to distinguish between loops and
                           simple blocks *)
                        match cil_stmt.Cil.skind with
                            Cil.Loop (block, cil_loc) ->
                              let src_pos = cil_location_to_source_position cil_loc
                              and cmd = C_Command.cmd_true
                              in
                                CFA.set_location_attributes entry_loc (Original src_pos) ;
                                CFA.add_edge entry_loc entry_loc cmd
                                (* No recursive call, since this CIL statement
                                   does not have any CIL successors *)
                          | Cil.Block b ->
                              match b.Cil.battrs with
				[attr] -> 
				  let (aname,attrexp_list) = BlastCilInterface.convertCilAttrToExpList attr in
				    if (aname = "assume" && (List.length attrexp_list = 1)) then 
				      begin
					
					let  pred = Predicate.Atom (List.hd attrexp_list) in
                                        Message.msg_string Message.Debug (Predicate.toString pred); 
					let cmd = C_Command.gather_accesses (C_Command.Pred pred) in
					let new_loc = CFA.create_location fun_name automaton Artificial in
					
					  CFA.add_edge entry_loc new_loc cmd;
					  let next_cil_stmt =
					    match cil_stmt.Cil.succs with
					      | [s] -> s
					      | _ -> raise CilAssumeWithIncorrectNumberOfSuccessors
					  in
					     (* Recursive call on the successor of this CIL
						statement *)
					   (*  skip_cil_stmt cil_stmt entry_loc next_cil_stmt
					      last_skipped_cil_stmts None *)
					      process_cil_stmt new_loc next_cil_stmt [] 
				      end
				    else
				      (* Recursive call on the successor of this CIL
                                       statement *)
				    skip_cil_stmt cil_stmt entry_loc next_cil_stmt
                                    last_skipped_cil_stmts None

				| _  -> 
				    (* Recursive call on the successor of this CIL
                                       statement *)
				    skip_cil_stmt cil_stmt entry_loc next_cil_stmt
                                    last_skipped_cil_stmts None

                          | _ -> assert false (* can't happen for empty blocks *)
                      end
                    else
                      begin
                        (* The successor of this CIL statement should be the
                           first statement of the block *)
                        assert (next_cil_stmt == (List.hd block.Cil.bstmts)) ;
                        (* Recursive call on the successor of this CIL statement *)
                        skip_cil_stmt cil_stmt entry_loc next_cil_stmt
                                      last_skipped_cil_stmts None
                      end
                end
        end
    (* This function tells whether we can safely skip a cil_stmt.  It is
       used during the translation on statements for which skipping
       preserves the semantics of the original CIL file.  This function
       just checks for other attributes (e.g. labels) *)
    and skipping_ok cil_stmt =
      (* It would be nicer to check only whether this statement has some
         original label atached to it (not some CIL generated label), but
         CIL does not allow that *)
      (get_real_labels cil_stmt) = []
    (* This function takes care of extra checks when processing a CIL
       statement that could be safely (w.r.t to the semantics of the original
       CIL file) skipped *)
    and skip_cil_stmt cil_stmt entry_loc next_cil_stmt last_skipped_cil_stmts
                      cil_loc_option =
      if (skipping_ok cil_stmt)
      then
        process_cil_stmt entry_loc next_cil_stmt
                         (cil_stmt::last_skipped_cil_stmts)
      else
        (* We can not skip it.  So we create a new location and introduce a
           skip.  Let's try to get the source location for entry_loc. *)
        let get_label_loc = function
            Cil.Label (_, cil_loc, _)
          | Cil.Case (_, cil_loc)
          | Cil.Default (cil_loc) ->
              cil_loc in
        let get_best_cil_loc cil_loc label =
          let label_loc = get_label_loc label
          in
            { Cil.file =
                if (cil_loc.Cil.file = "")
                then label_loc.Cil.file
                else
                  begin
                    assert (label_loc.Cil.file = (max "" cil_loc.Cil.file)) ;
                    cil_loc.Cil.file
                  end ;
              Cil.line =
                if (label_loc.Cil.file = "")
                then cil_loc.Cil.line
                else max (cil_loc.Cil.line) (label_loc.Cil.line) } in
        let cil_loc =
          match cil_loc_option with
              None ->
                List.fold_left get_best_cil_loc
                               { Cil.file = "" ;
                                 Cil.line = -1 }
                               cil_stmt.Cil.labels
            | Some c -> c in
        let src_pos = cil_location_to_source_position cil_loc
        and new_loc = CFA.create_location fun_name automaton Artificial
        in
          (* We insert a skip *)
          CFA.set_location_attributes entry_loc (Original src_pos) ;
          CFA.add_edge entry_loc new_loc C_Command.skip ;
          process_cil_stmt new_loc next_cil_stmt []
    (* this function returns true if cil_block is either an empty list of
       statements or a list of empty blocks statements *)
    and cil_block_is_empty cil_block =
      List.fold_left (fun curr stmt -> if (curr)
                                       then (cil_stmt_is_empty_block stmt)
                                       else false)
                     true cil_block.Cil.bstmts
    and cil_stmt_is_empty_block stmt =
      match stmt.Cil.skind with
          Cil.Block block -> cil_block_is_empty block
        | _               -> false
    in
      CFA.set_start_location automaton entry_location ;
      CFA.set_end_location automaton end_location ;
      process_cil_stmt start_location (List.hd cil_fun_block.Cil.bstmts) [] ;
      automaton


  exception CilFunDecTypeMismatch

  let cil_fundec_to_fundef (cil_fundec, cil_location) =
    let src_pos = cil_location_to_source_position cil_location
    and formals_as_symbols = List.map BlastCilInterface.cil_varinfo_to_symbol
                                      cil_fundec.Cil.sformals in
    let locals = Misc.union (List.map BlastCilInterface.cil_varinfo_to_symbol cil_fundec.Cil.slocals)
                            formals_as_symbols
    in
    let formals =
      if (match cil_fundec.Cil.svar.Cil.vtype with
            | Cil.TFun (_, _, b, _) -> b
            | _                      -> raise CilFunDecTypeMismatch)
      then Variable formals_as_symbols
      else Fixed formals_as_symbols
    in
    let atomic = Cil.hasAttribute "atomic" cil_fundec.Cil.svar.Cil.vattr in
    let res =
      { function_name    = cil_fundec.Cil.svar.Cil.vname ;
        locals           = locals ;
        formals          = formals ;
        cfa              = cil_fun_block_to_cfa atomic cil_fundec.Cil.sbody cil_fundec.Cil.svar.Cil.vname;
        fun_src_pos      = src_pos ; 
	fun_attributes   = cil_fundec.Cil.svar.Cil.vattr ; }
    in
    res



  exception DuplicateFunctionName of string

  let process_cil_file cil_file =
    let process_cil_global_declarations cil_global =
      match cil_global with
      | Cil.GVar (cil_varinfo, _, cil_location) -> 
	  let name = cil_varinfo.Cil.vname in
          Message.msg_string Message.Normal ("Encountered global variable " ^ name);
	  (*RUPAK: was :
	    let cil_vardec_to_vardef cil_varinfo cil_location = 	  
	    let src_pos = cil_location_to_source_position cil_location in
	    let flag = Cil.hasAttribute "lock" cil_varinfo.Cil.vattr in
	    { var_name      = name ;
	      var_src_pos   = src_pos ;
	      var_is_lock   = flag ;
              var_vaddrof = cil_varinfo.Cil.vaddrof ;
            }
	    in
	    let var_def = cil_vardec_to_vardef cil_varinfo cil_location in
	    Hashtbl.add global_variable_table name var_def
	     *)
	  add_global (Expression.Symbol name) (Varinfo cil_varinfo)
      | _ -> ()
    in
    let process_cil_global_functions cil_global =
      match cil_global with
          Cil.GFun (cil_fundec, cil_location) ->
            let fn_name = cil_fundec.Cil.svar.Cil.vname in
            (* First check that a function with the same name has not
               already been processed *)
              begin
                if (Hashtbl.mem global_function_table fn_name)
                then
                  raise (DuplicateFunctionName fn_name)
              end ;
              Message.msg_string Message.Debug ("Encountered function " ^ fn_name);
              let fundef = cil_fundec_to_fundef (cil_fundec, cil_location)
              in
                Hashtbl.add global_function_table fn_name fundef ;
                Hashtbl.iter (fun label loc ->
                                Hashtbl.add global_label_table label loc)
                             (CFA.get_automaton_attributes fundef.cfa).label_table
      |	_ -> ()

    in
    let is_fun_name lv =
      match lv with
	Expression.Symbol s ->
	  let rec _find_fname lst =
	    match lst with 
	      [] -> false
	    | a :: rest ->
		begin
		  match a with
		    Cil.GFun (cil_fundec, _) -> 
		      if s = cil_fundec.Cil.svar.Cil.vname then true else
		      _find_fname rest
		  | Cil.GVarDecl (vinfo, _) ->
		      begin
			match vinfo.Cil.vtype with
			  Cil.TFun _ ->
			    if s = vinfo.Cil.vname then true else _find_fname rest
			| _ -> _find_fname rest
		      end	
		  | _ -> _find_fname rest
		end
	  in
	  _find_fname cil_file.Cil.globals
      |	_ -> false
    in
      (* RJ : adding global info for cf reachability as well *)
    (*if (Options.getValueOfBool "checkRace") then *)
      add_all_lvals cil_file ;

    (* start comment: This is no longer needed: since we keep track of all lvals. 
       I will expunge this after I add the field/varinfo to the lval table. *)
    (*if (Options.getValueOfBool "checkRace") then*)
      List.iter process_cil_global_declarations cil_file.Cil.globals ;
    (* end comment *)

(** At this point, the global_variable_table has been
    populated. So I can iterate over its entries and get all the lvalues
    that need to be havoced.
    *)

    if (Options.getValueOfBool "checkRace") then
	(* add variables to havoc only if we are running in TAR mode for races *)
      iterate_all_lvals 
	(fun lv -> 
	  if is_fun_name lv then () else 
	  Havoc.global_variables := lv :: !Havoc.global_variables) 
    else () ;
    
    List.iter process_cil_global_functions cil_file.Cil.globals
      


  module OrderedInt = struct type t = int let compare a b = b - a end
  module IntSet = Set.Make(OrderedInt)

  let output_cfa_dot ch fundef =
    let known_locs = ref IntSet.empty in
    let rec output_loc loc =
      Printf.fprintf ch "  %d;\n" loc.CFA.loc_id;
      known_locs := IntSet.add loc.CFA.loc_id !known_locs;
      List.iter
        (fun (cmd, l) ->
            Printf.fprintf ch "  %d -> %d [label=\"%s\"]\n"
              loc.CFA.loc_id l.CFA.loc_id
              (String.escaped (C_Command.to_string cmd));
          if not (IntSet.mem l.CFA.loc_id !known_locs) then output_loc l)
        loc.CFA.succs
    in
    match fundef.cfa.CFA.start_location with
    | None -> ()
    | Some loc ->
        Printf.fprintf ch "digraph %s {\n" fundef.function_name;
        output_loc loc;
        Printf.fprintf ch "}\n\n";
        ()





(* Stuff required by the SYSTEM_DESCRIPTION signature *)

  module Command = C_Command


  type location = (automaton_attributes,
                   location_attributes,
                   C_Command.t) CFA.location

  let fake_location =
    { CFA.loc_id = -1;
      CFA.succs = [];
      CFA.preds = [];
      CFA.automaton = { CFA.aut_id = -1;
                        CFA.start_location = None;
                        CFA.end_location = None;
                        CFA.aut_attributes = {label_table = Hashtbl.create 1}};
      CFA.loc_attributes = Artificial}

  let get_source_position loc =
    match (CFA.get_location_attributes loc) with
      Original sp ->  Some (sp.file_name, sp.line, sp.column)
    | Artificial -> None

  let print_location fmt loc =
    Format.fprintf fmt "@[Location(@[id=%d#%d;@ attrs="
                       loc.CFA.automaton.CFA.aut_id
                       loc.CFA.loc_id ;
    print_location_attributes fmt (CFA.get_location_attributes loc) ;
    Format.fprintf fmt "@])@]"

  type edge = location * Command.t * location

  let edge_to_command (_, c, _) = c

  let print_edge fmt (s, c, t) =
    Format.fprintf fmt "@[" ;
    print_location fmt s ;
    Format.fprintf fmt "---" ;
    Command.print fmt c ;
    Format.fprintf fmt "--->" ;
    print_location fmt t ;
    Format.fprintf fmt "@]"

  let location_to_string = Misc.to_string_from_printer print_location

  let location_coords loc = (loc.CFA.automaton.CFA.aut_id , loc.CFA.loc_id)

  let edge_to_string = Misc.to_string_from_printer print_edge

  let get_source (s, c, t) = s
  let get_target (s, c, t) = t
  let get_command (s, c, t) = c

  let get_outgoing_edges loc =
    let succs = CFA.get_succs loc in
      List.map (fun (cmd, succ) -> (loc, cmd, succ)) succs

  let get_ingoing_edges loc = 
    let preds = CFA.get_preds loc in
      List.map (fun (cmd, pred) -> (pred, cmd, loc)) preds


  let reads_and_writes_at_edge e =
    let cmd = get_command e in
    cmd.C_Command.read, cmd.C_Command.written

(* lval -> lval list *)
  let exp_closure_table = Hashtbl.create 1009

  let expression_closure e = 
    let rec _rfl depth e  = 
      if depth = 0 then []
      else 
	begin
	  match get_type_of_exp e with
	      Cil.TPtr (pt, _) -> 
		let e' = Expression.Lval (Expression.Dereference e) in
		  e'::(_rfl (depth - 1) e') 
	    | Cil.TComp (c,_) -> 
		let e's = 
		  List.map 
		    (fun f -> 
		       Expression.Lval (Expression.Access (Expression.Dot, e , f.Cil.fname)))
		    c.Cil.cfields
		in
		  e's @ (List.flatten (List.map (_rfl (depth - 1)) e's))
	    | _ -> []
	end
    in
      if (Options.getValueOfString "alias" = "" || Options.getValueOfBool "incref")
      then [e]
      else
	try
	  Hashtbl.find exp_closure_table e
	with Not_found ->
	  begin
	    let clos = e::(_rfl 2 e) in
	      Hashtbl.replace exp_closure_table e clos;
	      clos
	  end


  exception NoSuchFunctionException of string

  let is_defined fn_name =
    Hashtbl.mem global_function_table fn_name

  let lookup_entry_location fn_name =
    try
      CFA.get_start_location (Hashtbl.find global_function_table fn_name).cfa
    with Not_found ->
      raise (NoSuchFunctionException fn_name)

   let get_location_fname loc = 
     CFA.location_function_name loc

  let list_of_functions () = Misc.run_on_table (fun fname _ -> fname) global_function_table


  let list_of_functions () = Misc.run_on_table (fun fname _ -> fname) global_function_table

  let is_loopback loc = ((List.length (CFA.get_preds loc)) > 1) (* Hashtbl.mem loopback_location_table loc *)

  let lookup_formals fn_name =
    try
      (Hashtbl.find global_function_table fn_name).formals
    with Not_found ->
      raise (NoSuchFunctionException fn_name)

  let lookup_locals fn_name =
    try
      (Hashtbl.find global_function_table fn_name).locals
    with Not_found ->
      raise (NoSuchFunctionException fn_name)

  let get_locations_at_label label =
    Hashtbl.find_all global_label_table label


  let map_nodes_fname f fname  = 
    let module LocationSet =
      Set.Make (struct
                  type t = int 
                  let compare x y =  compare x  y
		end) 
    in
    let already_visited_source_locations = ref LocationSet.empty in
    let rec map_loc_and_edges loc_list =
      match loc_list with
	  [] -> []
	| loc::t -> 
	    begin
	      let (_,loc_id) = location_coords loc in
		if (LocationSet.mem loc_id !already_visited_source_locations) 
		then 
		  (map_loc_and_edges t)
                else
		  begin
		    let _ = 
		      (already_visited_source_locations := LocationSet.add loc_id !already_visited_source_locations) 
		    in
		    let children_h = List.map get_target (get_outgoing_edges loc) in
		      (f loc)::(map_loc_and_edges (children_h @ t))
		  end
	    end
    in
      map_loc_and_edges [(lookup_entry_location fname)]
    
  let map_edges_fname f fname  = 
    let node_fn n = List.map f (get_outgoing_edges n) in
      List.flatten (map_nodes_fname node_fn fname)
	
  let map_edges f =
    List.flatten (List.map (map_edges_fname f) (list_of_functions ()))

  let map_nodes f = 
    List.flatten (List.map (map_nodes_fname f) (list_of_functions ()))

  let print_system fmt () =
    let module LocationSet =
      Set.Make(struct
                 type t = location
                 let compare loc_1 loc_2 = compare loc_1.CFA.loc_id loc_2.CFA.loc_id 
               end) in
    let print_function function_def =
      let already_printed_locations = ref LocationSet.empty in
      let rec print_location_and_edges loc =
        if (LocationSet.mem loc !already_printed_locations) then
          ()
        else
          begin
            let ingoing_edges = List.rev (get_ingoing_edges loc)
            and outgoing_edges = List.rev (get_outgoing_edges loc)
            in
              Format.fprintf fmt "@ @    @[<v>Location: " ;
              print_location fmt loc ;
              Format.fprintf fmt "@ " ;
              Format.fprintf fmt "   @[<v>Outgoing edges:" ;
              List.iter (fun e -> Format.fprintf fmt "@    " ; print_edge fmt e)
                        outgoing_edges ;
              Format.fprintf fmt "@ Ingoing edges:" ;
              List.iter (fun e -> Format.fprintf fmt "@    "; print_edge fmt e)
                        ingoing_edges ;
              Format.fprintf fmt "@]@]" ;
              already_printed_locations := (LocationSet.add loc !already_printed_locations) ;
              List.iter (fun e -> print_location_and_edges (get_target e)) outgoing_edges
          end in
      let label_table = (CFA.get_automaton_attributes function_def.cfa).label_table
      in
        Format.fprintf fmt "@ @ @ Function: %s" function_def.function_name ;
        if (Cil.hasAttribute "atomic" function_def.fun_attributes)
        then Format.fprintf fmt "@  atomic";
        if (Cil.hasAttribute "noreturn" function_def.fun_attributes) 
        then Format.fprintf fmt "@  __noreturn";
        Format.fprintf fmt "@  Formals: " ;
        print_formals fmt (function_def.formals) ;
        Format.fprintf fmt "@  Labels: @[<v>" ;
        Hashtbl.iter (fun lbl loc ->
                        Format.fprintf fmt "%s: " lbl ;
                        print_location fmt loc ;
                        Format.fprintf fmt "@ ")
                     label_table ;
        Format.fprintf fmt "@]" ;
        print_location_and_edges (CFA.get_start_location function_def.cfa)
    and print_variable var_def =
      match var_def with
	Varinfo v -> 
	  Format.fprintf fmt "@ %s%s@\n"
            v.Cil.vname
            (if (Cil.hasAttribute "lock" v.Cil.vattr) then " (lock)" else "");
	  ()
      |	Fieldinfo f -> 
	  Format.fprintf fmt "@ %s.%s%s@\n"
	    self
            f.Cil.fname
            (if (Cil.hasAttribute "lock" f.Cil.fattr) then " (lock)" else "");
	  ()
    in
    Format.fprintf fmt "@[<v>" ;
    Format.fprintf fmt "Functions are: @[" ;
    Hashtbl.iter (fun fn _ -> Format.fprintf fmt "%s@ " fn) global_function_table ;
    Format.fprintf fmt "@]" ;
    Hashtbl.iter (fun _ fn_def -> print_function fn_def) global_function_table ;
    Format.fprintf fmt "@]@\n" ;
    Format.fprintf fmt "@[Global variables are:@\n";
    Hashtbl.iter (fun _ var_def -> print_variable var_def) global_variable_table;
    Format.fprintf fmt "@]@\n";
    Format.fprintf fmt "@[Lvalues are:@\n";
    print_all_lvals () ;
    Format.fprintf fmt "@]@\n";
    ()

  let get_spec_functions spec_function_prefix =
    Hashtbl.fold (fun fn_name _ curr ->
                    if (Misc.is_prefix spec_function_prefix fn_name)
                    then fn_name::curr
                    else curr)
                 global_function_table []

  let get_args_for_spec_call spec_formal_prefix formals spec_formals fun_name =
    let process_formal s =
      (* [Greg] I believe the following comment is no longer relevant, but
         I leave it there just in case *)
      (* Jhala: change the thing below to change the prefix *)
      if (not (Misc.is_prefix spec_formal_prefix s)) then
        failwith("Bad formal '"^ s ^"' in spec function!")
      else
	begin
	  let  _s = Misc.chop_after_prefix spec_formal_prefix s
          in
	    if _s = "ret" then Expression.Lval (Expression.Symbol("__retres_" ^ fun_name))
	    else (* the string is an int *)
	      let num = (int_of_string _s) - 1 in
		Expression.Lval (Expression.Symbol(List.nth formals num))
	end
    in
      List.map (process_formal) spec_formals

  let hook_cmd_before_start cfa hook_cmd =
    let start_loc = CFA.get_start_location cfa in
    let fname = CFA.location_function_name start_loc in
    let new_start_loc = CFA.create_location fname cfa Artificial
    in
      CFA.set_start_location cfa new_start_loc ;
      CFA.add_edge new_start_loc start_loc hook_cmd

  let hook_cmd_before_return cfa hook_cmd =
    let end_loc = CFA.get_end_location cfa
    in
      (* Since we call oneret on each function, there can be only one edge
         leading to the end location (and it is labeled by a return) *)
      assert ((List.length (CFA.get_preds end_loc)) = 1) ;
      let (return_cmd, end_loc_pred) = List.hd (CFA.get_preds end_loc) in
      let end_loc_pred_attr = CFA.get_location_attributes end_loc_pred in
      let fname = CFA.location_function_name end_loc in
      let new_end_loc_pred = CFA.create_location fname cfa end_loc_pred_attr
      in
        (* end_loc_pred should have only one successor *)
        assert ((List.length (CFA.get_succs end_loc_pred)) = 1) ;
        (* Let's replace end_loc_pred by new_end_loc_pred *)
        CFA.set_location_attributes end_loc_pred Artificial ;
        (* We redirect end_loc's incoming edge to new_end_loc_pred *)
        CFA.remove_edge end_loc_pred end_loc return_cmd ;
        CFA.add_edge new_end_loc_pred end_loc return_cmd ;
        (* end_loc_pred should not have any successor now *)
        assert ((CFA.get_succs end_loc_pred) = []) ;
        (* we now can hook the command hook_cmd *)
        CFA.add_edge end_loc_pred new_end_loc_pred hook_cmd

  let instrument_with_spec spec_function_prefix spec_formal_prefix =
    let instrument_spec_function spec_fun_name =
      let chopped_sfn = Misc.chop_after_prefix spec_function_prefix
                                               spec_fun_name
      in
      let (fun_name, is_call, is_return) =
        if (Misc.is_prefix "call_" chopped_sfn)
        then
          (Misc.chop_after_prefix "call_" chopped_sfn, true, false)
        else
          if (Misc.is_prefix "return_" chopped_sfn)
          then
            (Misc.chop_after_prefix "return_" chopped_sfn, false, true)
          else
            ("", false, false)
      in
        if (not (is_call || is_return))
        then
          (* We do nothing, this might be an internal spec function *)
          let src_pos = (Hashtbl.find global_function_table
                                      spec_fun_name).fun_src_pos in
          let msg = src_pos.file_name ^ ":" ^ (string_of_int src_pos.line)
                    ^ ": " ^ "warning: un-hooked spec-like function '" ^
                    spec_fun_name ^ "'."
          in
            Message.msg_string Message.Normal msg ;
        else
          let get_formals_as_string_list = function
              Variable l | Fixed l -> l in
          let formals =
            get_formals_as_string_list (lookup_formals fun_name)
          and spec_formals =
            get_formals_as_string_list (lookup_formals spec_fun_name) in
          let spec_args = get_args_for_spec_call spec_formal_prefix formals
                                                 spec_formals fun_name in
          let hook_code =
            C_Command.FunctionCall
              (Expression.FunctionCall(Expression.Lval (Expression.Symbol(spec_fun_name)),
                                       spec_args))
          and cfa = (Hashtbl.find global_function_table fun_name).cfa
          in
          let hook_cmd = C_Command.gather_accesses hook_code in
            if (is_call)
            then hook_cmd_before_start cfa hook_cmd
            else hook_cmd_before_return cfa hook_cmd
    in
      List.iter instrument_spec_function (get_spec_functions spec_function_prefix)

(* dealing with symbolic variables and parameter loading etc etc 
   for each function f, with formals x, globals g, add an assume p, where p is:
   x = x#f  && g = g#f 
*)
	  (* this function should be called BEFORE the model checking begins,
	     if we are doing CF reachability *)


  
let globals () = 
  
  Misc.map_partial 
    (function Expression.Symbol s -> Some s | _ -> None) 
    (Misc.hashtbl_keys global_variable_table) 
    
let make_symvar f x = 
  if (is_global (Expression.Symbol x)) then (x^"#@"^f) 
  else 
    let [pre;suff] = Misc.bounded_chop x "@" 2 in
      pre^"#@"^suff

  
let add_symbolic_hooks () =
  Message.msg_string Message.Debug "In add_symbolic_hooks";
  let globs = globals () (* actually we should be doing a mod analysis blah blah blah *) in
  let sym_assume f =
    let formals = 
      match lookup_formals f with
	  Variable l -> l
	| Fixed l -> l
    in
    let make_eq x = 
      Command.Expr( Expression.Assignment (Expression.Assign,Expression.Symbol x, Expression.Lval( Expression.Symbol(make_symvar f x))))
    in
      { Command.code = C_Command.Block (List.map make_eq (formals@(globs)));
	Command.read = [];    (* and its not so hard to change *)
	Command.written = []; (* um er. Should be changed ...*)
      }
  in
  let process_function f = 
    Message.msg_string Message.Debug ("Processing function: "^f);
    let cfa = (Hashtbl.find global_function_table f).cfa in
      hook_cmd_before_start cfa (sym_assume f)
  in
    List.iter process_function (Misc.hashtbl_keys global_function_table)
	  
	
let initialize c_files =
  let cil_files = List.map BlastCilInterface.make_cil_file c_files
  in
    global_cil_files := (Some cil_files);
    (* first build the CFAs -- the global_cfa_table *)
    let _ = 
      try
	List.iter process_cil_file cil_files;
	let cfa_file = Options.getValueOfString "cfa-dot" in
          if cfa_file <> "" then
            let ch = open_out cfa_file in
              List.iter (output_cfa_dot ch)
		(List.sort (fun fundef1 fundef2 ->
                              compare fundef1.function_name fundef2.function_name)
		   (Hashtbl.fold (fun _ fundef l -> fundef :: l)
                      global_function_table []));
              close_out ch
      with
          DuplicateFunctionName fn_name ->
            failwith ("C_System_Descr.initialize: duplicate function name ("
                      ^ fn_name ^ "): unimplemented")
    in
      if (Options.getValueOfBool "cf") then add_symbolic_hooks ();

   	  


(****************************************************************)
(* ------------------- call graph interface ------------------- *)
(****************************************************************)

(* a call node describes the local calling structure for a
 * single function: which functions it calls, and which
 * functions call it *)
type callnode = {
  (* the function this node describes *)
  cnInfo: string;

  (* set of functions this one calls *)
  cnCallees: (string, callnode) Hashtbl.t;

  (* set of functions that call this one *)
  cnCallers: (string, callnode) Hashtbl.t;
}

(* a call graph is a hashtable, mapping a function name to
 * the node which describes that function's call structure *)
type callgraph =
  (string, callnode) Hashtbl.t


(* ------------------- implementation --------------------------- *)
(* add something to a hashtable if it's not already there, instead of
 * the default pushback behavior of Hashtbl.add *)
let add_if (table : ('a,'b) Hashtbl.t) (key:'a) (data:'b) : unit =
begin
  if (not (Hashtbl.mem table key)) then
    (Hashtbl.add table key data)
end

  (* this is the graph we are computing, 
   * it is created empty when the object is constructed, and will
   * be added-to during the computation *)
let graph : callgraph = Hashtbl.create 101


  (* given the name of a function, retrieve its callnode; this
   * will create a node if one doesn't already exist *)
  let getNode (fname: string) : callnode =
  begin
    try
      (Hashtbl.find graph fname)

    with Not_found -> (
      (* make a new node *)
      let ret:callnode = {
        cnInfo = fname;
        cnCallees = (Hashtbl.create 5);
        cnCallers = (Hashtbl.create 5);
      }  in

      (* add it to the table, then return it *)
      (Hashtbl.add graph fname ret);
      ret
    )
  end

let findNode (fname: string) : callnode =
  begin
    try
      Hashtbl.find graph fname
    with
      Not_found -> Message.msg_string Message.Error (Printf.sprintf 
                       "findNode: Cannot find function %s in call graph!" fname);
		   raise Not_found
  end

let update_callgraph caller callee =
   let callernode = getNode caller 
   in
   let calleenode = getNode callee 
   in
            (* add one entry to each node's appropriate list *)
   (add_if callernode.cnCallees callee calleenode);
   (add_if calleenode.cnCallers caller callernode)

let get_callers (fname : string) =
  let fnode = findNode fname   
  in
  fnode.cnCallers

let get_callees (fname : string) =
  let fnode = findNode fname   
  in
  fnode.cnCallees
  

let compute_callgraph () : unit =
begin
  let the_curr_func = ref None in
  let get_caller () = 
    match !the_curr_func with
      Some f -> f 
    | None -> failwith "computeGraph: Function called in non-context" 
  in
  let f_edge e =
    let get_fname fname =
      match fname with
        Expression.Lval (Expression.Symbol name) -> name
      | _                                        -> "__BLAST_fnptr"
    in
    match (get_command e).Command.code with
      Command.Skip -> ()
    | Command.FunctionCall (Expression.FunctionCall (fname, _)) 
    | Command.FunctionCall (Expression.Assignment (_, _, Expression.FunctionCall (fname, _))) ->  
     let callee = get_fname fname in
       let caller = get_caller () in
       (* update call graph edge *)
       update_callgraph 
                caller 
                callee
    | _ -> ()
  in
  List.map (fun fname -> the_curr_func := Some fname ; ignore (getNode fname) ; map_edges_fname f_edge fname) 
           (list_of_functions ());
  () (*   graph  *) 
end

let get_callgraph () = graph

let check_cycles (node : callnode) =
        let rec print_cycle node stack =
	  match stack with 
            [] -> failwith "print_cycle: node not found in stack"
          | h:: t -> 
		if h = node then Message.msg_string Message.Normal ("Calls "^ (node.cnInfo))
		else 
		  begin
			Message.msg_string Message.Normal ("Calls "^ (h.cnInfo)) ;
			print_cycle node t
		  end
        in
	let rec _do_dfs (node : callnode) (stack : callnode list) (marked : callnode list) =
          if (List.mem node marked) then 
	   begin
              if (List.mem node stack) then 
		(* cycle found *)
                print_cycle node stack ;
	     marked
           end 
          else 
            begin
              if (List.mem node stack) then 
		(* cycle found *)
                print_cycle node stack ;
              Hashtbl.fold (fun child -> fun childnode -> fun markednodes -> _do_dfs childnode (node::stack) markednodes) 
			        (node.cnCallees) 
				(node :: marked)
            end
        in
        ignore (_do_dfs node [] [])

  module OrderedString = struct type t = string let compare = compare end
  module StringSet = Set.Make (OrderedString)

let output_callgraph filename = 
  let output_graph ch graph =
    let known_locs = ref StringSet.empty in
    let rec output_loc (loc : callnode) =
      Printf.fprintf ch "  %s;\n" loc.cnInfo;
      known_locs := StringSet.add loc.cnInfo !known_locs;
      Hashtbl.iter
        (fun called calledNode ->
            Printf.fprintf ch "  %s -> %s \n"
              loc.cnInfo called
              ;
          if not (StringSet.mem called !known_locs) then output_loc calledNode)
        loc.cnCallees ;
      ()
    in

    let main = List.hd (Options.getValueOfStringList "main") in
    let startcallnode = 
      try
        Hashtbl.find graph main
      with Not_found -> (Message.msg_string Message.Error ("Main function "^main^" not found in call graph!"); raise Not_found ) 
    in
    Message.msg_string Message.Normal "Checking for cycles in the call graph." ;
    check_cycles startcallnode ;
    Message.msg_string Message.Normal "Done checking for cycles in the call graph." ;
    Printf.fprintf ch "digraph callGraph {\n" ;
    output_loc startcallnode;
    Printf.fprintf ch "}\n\n" ;
    ()   

  in
  let outch = open_out filename in
  output_graph outch graph ;
  close_out outch

let print_callgraph (out:out_channel)  : unit =
begin
  let printEntry (s:string) (n:callnode) : unit =
    (Printf.fprintf out " %s" s) in

  let printNode (name:string) (node:callnode) : unit =
  begin
    (Printf.fprintf out "%s:\n" name);
    (Printf.fprintf out "  calls:");
    (Hashtbl.iter printEntry node.cnCallees);
    (Printf.fprintf out "\n  is called by:");
    (Hashtbl.iter printEntry node.cnCallers);
    (Printf.fprintf out "\n")
  end in
  
  (Hashtbl.iter printNode graph)
end

      

end
