(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

module Constant = Ast.Constant
module Expression = Ast.Expression
module Predicate = Ast.Predicate

(*
module CilExpressionTranslator =
struct
(* Functions to convert from CIL representation to Expressions and Predicates *)
*)

let cilUOptoUnaryOp op =
  match op with
    Cil.Neg -> Expression.UnaryMinus
  | Cil.BNot -> Expression.BitNot
  | Cil.LNot -> Expression.Not

let cilBOptoBinaryOp op =
  match op with
    Cil.PlusA  -> Expression.Plus                              (** arithmetic + *)
  | Cil.PlusPI -> Expression.Plus                             (** pointer + integer *)
  | Cil.IndexPI -> Expression.Plus                            (** pointer[integer]. The difference 
                                           form PlusPI is that in this case 
                                           the integer is very likely 
                                           positive *)
  | Cil.MinusA -> Expression.Minus                             (** arithemtic - *)
  | Cil.MinusPI -> Expression.Minus                            (** pointer - integer *)
  | Cil.MinusPP -> Expression.Minus                            (** pointer - pointer *)
  | Cil.Mult    -> Expression.Mul                            (** * *)
  | Cil.Div     -> Expression.Div                            (** / *)
  | Cil.Mod     -> Expression.Rem                            (** % *)
  | Cil.Shiftlt -> Expression.LShift                            (** shift left *)
  | Cil.Shiftrt -> Expression.RShift                            (** shift right *)

  | Cil.Lt      -> Expression.Lt                            (** <  (arithmetic comparison) *)
  | Cil.Gt      -> Expression.Gt                            (** >  (arithmetic comparison) *)  
  | Cil.Le      -> Expression.Le                            (** <= (arithmetic comparison) *)
  | Cil.Ge      -> Expression.Ge                            (** >  (arithmetic comparison) *)
  | Cil.Eq      -> Expression.Eq                            (** == (arithmetic comparison) *)
  | Cil.Ne      -> Expression.Ne                            (** != (arithmetic comparison) *)            

  | Cil.LtP     -> Expression.Lt                            (** <  (pointer comparison) *)
  | Cil.GtP     -> Expression.Gt                          (** >  (pointer comparison) *)
  | Cil.LeP     -> Expression.Le                            (** <= (pointer comparison) *)
  | Cil.GeP     -> Expression.Ge                            (** >= (pointer comparison) *)
  | Cil.EqP     -> Expression.Eq                            (** == (pointer comparison) *)
  | Cil.NeP     -> Expression.Ne                            (** != (pointer comparison) *)

  | Cil.BAnd    -> Expression.BitAnd                            (** bitwise and *)
  | Cil.BXor    -> Expression.Xor                            (** exclusive-or *)
  | Cil.BOr     -> Expression.BitOr                            (** inclusive-or *)

let rec convertCilLval (lb, off) =
  let rec genOffset lvalue o =
    match o with
      Cil.NoOffset -> lvalue
    | Cil.Field (fid, o1) ->
	begin
	  (* try *)
	  if Cil.hasAttribute "lock" fid.Cil.fattr then
	    Message.msg_string Message.Normal ("Lock found! "^fid.Cil.fname) ;
	  let ee = Expression.Access (Expression.Dot, Expression.Lval lvalue, fid.Cil.fname) in
	  genOffset ee o1
	end
    | Cil.Index (e, o1) ->
	begin
	  let iexp = convertCilExp e in
(*
	  let ee = (Expression.Dereference (Expression.Binary (Expression.Plus, Expression.Lval lvalue, iexp))) in
*)
	  let ee = Expression.Indexed (Expression.Lval lvalue, iexp) in
	  genOffset ee o1
	end
  in
  
  
  match lb with
    Cil.Var vinfo -> 
      begin
	let m = Expression.Symbol vinfo.Cil.vname in
	genOffset m off
      end
  | Cil.Mem m ->
      begin
	let _lval_mem_worker m off =
	  match off with
	    Cil.NoOffset -> Expression.Dereference m
	  | Cil.Field (finfo, offset) -> 
	      begin
                (* Have only Dot's, no Arrow's *)
		(* OLD:
                let e = Expression.Access (Expression.Arrow, m, finfo.Cil.fname) in
                *)
	  (* try *)
	  if Cil.hasAttribute "lock" finfo.Cil.fattr then
	    Message.msg_string Message.Normal ("Lock found! "^finfo.Cil.fname) ;
                (* NEW: *)
                let e = Expression.Access (Expression.Dot, Expression.Lval (Expression.Dereference m), finfo.Cil.fname) in 
		genOffset e offset
	      end
	  | Cil.Index (exp, offset) -> 
	      begin
		let e = (Expression.Indexed (m, convertCilExp exp)) in
		genOffset e offset
	      end
	in
	let m = convertCilExp m
	in
	_lval_mem_worker m off
      end

and convertCilConst c =
  match c with
      Cil.CInt64 (i64, _, _) -> Constant.Int (Int64.to_int i64)
    | Cil.CStr str -> Constant.String str
    | Cil.CChr c   -> Constant.String ("char") (* Check : why dont we have char constants? *)
    | Cil.CReal (f,_,_) -> Constant.Float f

and convertCilExp e =
  match e with
    Cil.Const const ->
      Expression.Constant (convertCilConst const)
  | Cil.Lval (lb,loff) ->
      begin
	Expression.Lval (convertCilLval (lb,loff))
      end
  | Cil.SizeOf t -> 
      Expression.Constant 
	(Constant.Int 
	   (Int32.to_int (Int32.shift_left (Int32.of_int (Cil.bitsSizeOf t))  3))) (* size in bytes *)
  | Cil.SizeOfE exp ->
      Expression.Constant
      (Constant.Int
         (Int32.to_int (Int32.shift_left (Int32.of_int (Cil.bitsSizeOf (Cil.typeOf exp)))  3))) (* size in bytes *)
  | Cil.AlignOf t -> failwith "convertCilExp: align not handled"
  | Cil.AlignOfE e -> failwith "convertCilExp: AlignOfE not handled"
  | Cil.UnOp (op, exp, _) ->
      let uop = cilUOptoUnaryOp op and e = convertCilExp exp in
      Expression.Unary(uop, e)
  | Cil.BinOp (op, e1,e2,_) ->
      let bop = cilBOptoBinaryOp op in
      let (ex1, ex2) = (convertCilExp e1, convertCilExp e2) in
      Expression.Binary(bop, ex1, ex2)
  | Cil.CastE (t, e) -> convertCilExp e (* drop casts *)
  | Cil.AddrOf lval -> 
      let l = convertCilLval lval in
      Expression.addressOf (Expression.Lval l)
      (* Expression.Unary(Expression.Address, Expression.Lval (convertCilLval lval)) *)
  | Cil.StartOf lv -> Expression.Lval (convertCilLval lv)
  | _ -> failwith "Not handled" 

let convertCilAttrToExpList a : (string * (Expression.expression list)) =
  let rec convertCilAttrparam ap =
    match ap with
      Cil.AInt i -> Expression.Constant (Constant.Int i)
    | Cil.AStr s -> Message.msg_string Message.Debug ("In AStr, string is "^s); (Expression.Lval (Expression.Symbol s))
    | Cil.ACons (s,aplist) -> Message.msg_string Message.Debug ("In ACons, string is "^s); 
	if (aplist = []) then (Expression.Lval (Expression.Symbol s)) else
	begin
	match s with
          "sel" -> Expression.Lval (Expression.Dereference (convertCilAttrparam (List.hd aplist)))
	| "at" -> 
	    begin
	      let s1 = convertCilAttrparam (List.nth aplist 0) and 
		  s2 = convertCilAttrparam (List.nth aplist 1) in
	      match (s1, s2) with
		(Expression.Lval (Expression.Symbol s), Expression.Lval (Expression.Symbol s'))  ->
		  Expression.Lval (Expression.Symbol (s^"@"^s'))
	      |	_ -> failwith "convertCilAttrparam : problem in at"
	    end
	| "arrow" ->
	    begin 
	      let s1 = convertCilAttrparam (List.nth aplist 0) and
		  s2 = convertCilAttrparam (List.nth aplist 1) in
	      match s2 with
		Expression.Lval (Expression.Symbol s) ->
		  Expression.Lval (Expression.Access (Expression.Arrow, s1, s))
	      | _ -> failwith "convertCilAttrparam : problem in arrow"	
	    end
	| "dot" ->
	    begin 
	      let s1 = convertCilAttrparam (List.nth aplist 0) and
		  s2 = convertCilAttrparam (List.nth aplist 1) in
	      match s2 with
		Expression.Lval (Expression.Symbol s) ->
		  Expression.Lval (Expression.Access (Expression.Dot, s1, s))
	      | _ -> failwith "convertCilAttrparam : problem in arrow"	
	    end
          | _ -> 
	      begin
		Message.msg_string Message.Error ("Error in parsing attributes: Strange function call "^s) ;
		Expression.FunctionCall (Expression.Lval (Expression.Symbol s), List.map convertCilAttrparam aplist)
	      end 
	end
    | Cil.ASizeOf _ -> failwith "convertCilAttrparam: SizeOf not handled"
    | Cil.ASizeOfE _ -> failwith "convertCilAttrparam: SizeOfE not handled"
    | Cil.AUnOp (uop, at) -> Expression.Unary (cilUOptoUnaryOp uop, convertCilAttrparam at)
    | Cil.ABinOp (bop, at1, at2) -> Expression.Binary (cilBOptoBinaryOp bop, convertCilAttrparam at1, convertCilAttrparam at2)
  in
  match a with
    Cil.Attr (name, alist) -> Message.msg_string Message.Debug ("Name of attribute is "^name) ; (name, List.map convertCilAttrparam alist)



let rec convertCilExpToPred e =
  (* Greg: I modify this to make use of convertCilExp.
           This way, we just have to make sure convertCilExp works and then
           this one will follow *)     
  (*
  match e with
      Cil.BinOp (op, e1, e2, _) ->
        let bop = cilBOptoBinaryOp op in
          if (Expression.isRelOp bop) then
            let (ex1, ex2) = (convertCilExp e1, convertCilExp e2) in
              Predicate.Atom (Expression.Binary(bop, ex1, ex2))
          else
            Predicate.Atom (Expression.Binary(Expression.Ne,
                                              convertCilExp e,
                                              Expression.Constant (Constant.Int 0)))
    | Cil.UnOp  (op, e, _) ->
        begin
          match op with
              Cil.LNot ->
                begin
                  Predicate.negate (convertCilExpToPred e)
                end
            | _        -> failwith "convertCilExpToPred: Strange unary operator"
        end
    | Cil.Const _
    | Cil.Lval _ -> Predicate.Atom (Expression.Binary(Expression.Ne,
                                                      convertCilExp e,
                                                      Expression.Constant (Constant.Int 0)))
    | _ -> failwith ("convertCilExpToPred: Non relational expression! Cannot convert to predicate.")
  *)
  let rec convertExpToPred exp =
    match exp with
        Expression.Binary (bop, e1, e2) ->
          if (Expression.isRelOp bop) then
            Predicate.Atom exp
          else 
            Predicate.Atom (Expression.Binary(Expression.Ne,
                                              exp,
                                              Expression.Constant (Constant.Int 0)))
      | Expression.Unary (uop, e1) ->
          begin
            match uop with
                Expression.Not ->
                  Predicate.negate (convertExpToPred e1)
              | _ ->
                  Predicate.Atom (Expression.Binary(Expression.Ne,
                                                    exp,
                                                    Expression.Constant (Constant.Int 0)))
          end
      | _ ->
          Predicate.Atom (Expression.Binary(Expression.Ne,
                                            exp,
                                            Expression.Constant (Constant.Int 0)))
  in
    convertExpToPred (convertCilExp e)

(*******************************************************************************)

let partiallyEvaluate file =
(*
  let constructSubgoals () =
    let readAndParseSubgoals filename =
      let doParse inchan =
	let _ = VampyreErrormsg.startFile filename in
	let lexbuf = Lexing.from_channel inchan in
	let _ = VampyreErrormsg.theLexbuf := lexbuf in
	Subgoalparse.main Subgoalflex.token lexbuf 
      in
      ignore (Message.msg_string Message.Normal "Reading in subgoals ...");
      try
	let inchan = open_in filename
	in
	let sglist = doParse inchan
	in
	ignore (Message.msg_string Message.Normal "Subgoals read.\n") ;
	sglist
      with
	Sys_error _ -> (Message.msg_string Message.Error ("Cannot find subgoal information in "^filename^"\n"); [])
      | VampyreErrormsg.Error -> (Message.msg_string Message.Error ("Error raised in reading subgoals\n"); [])
    in
    let sugfile = Options.getValueOfString "subgoal" in
    if (sugfile = "") then [] 
    else
      readAndParseSubgoals sugfile
  in    
  let subgoals = constructSubgoals () in
*)
  let subgoals = [] in
  match subgoals with
    [] ->  Partial.partial  file [ ]
  | (p, subs) :: _ -> (* can only handle the first *)
      begin
	let rec construct_assumption l accum =
	  match l with [] -> accum
	  | (varnam, v):: rest  -> 
	      let var = (* get cil svar structure from string varnam *)
		let v = ref (Cil.makeVarinfo "__none" Cil.voidType) in
		Cil.iterGlobals file (fun g -> match g with
		  Cil.GVar (vinf, _, _) -> Printf.printf "Variable %s \n" vinf.Cil.vname;
		    if vinf.Cil.vname = varnam then v:= vinf else ()
		| Cil.GFun (fund, _) -> 
                    (* TODO: look for an "atomic" attribute in [fund.svar.vtype] *)
		    let rec find_in_list varnam ls =
		      match ls with [] -> () |
		      a::rest -> if a.Cil.vname = varnam then v:=a else find_in_list varnam rest
		    in
		    find_in_list varnam fund.Cil.sformals;
		    find_in_list varnam fund.Cil.slocals
		      
		| _ -> ()
		    ); if ((!v).Cil.vname = "__none") then failwith "Variable not found!" else !v
	      and
		  valu = Cil.integer v in
       
	      construct_assumption rest ((Cil.var var, valu)::accum)
	in
	Partial.partial file (construct_assumption subs [])
      end


class oneret_visitor = object
  inherit Cil.nopCilVisitor 
  method vfunc f =
    Oneret.oneret f ; Cil.SkipChildren
end


(* [Greg] This is very dirty, but it's the only way to know whether Frontc
   encountered an error *)

exception FrontcError of Cil.location

class frontc_error_detection_visitor = object
  inherit Cil.nopCilVisitor 
  method vinst i =
    match i with
        (* Check the Asm to see wether frontc encountered an error *)
        Cil.Asm (_, templates, _, _, _, loc) ->
          begin
            if ((List.length templates) == 1) then
              if (Misc.is_prefix "booo_exp(" (List.hd templates)) then
                raise (FrontcError loc)
          end ;
          Cil.SkipChildren
      | _ ->
          Cil.SkipChildren
end

let rename_locals f =
  Cil.iterGlobals f (fun g -> match g with
    Cil.GFun(fd,_) -> 
  let fnName = fd.Cil.svar.Cil.vname in
  let locals = List.map (fun var -> (var.Cil.vname <- (var.Cil.vname^"@"^fnName));var) fd.Cil.slocals
  and formals = List.map (fun var -> (var.Cil.vname <- (var.Cil.vname^"@"^fnName));var) fd.Cil.sformals
  in
  fd.Cil.slocals <- locals ;
  fd.Cil.sformals <- formals
  | _ -> ())

let junk_location = {Cil.line = 0; Cil.file = ""}
let diverge = Cil.Loop ({Cil.bstmts = []; Cil.battrs = []}, junk_location)

module VarSet = Set.Make(struct
  type t = Cil.varinfo
  let compare v1 v2 = compare v1.Cil.vid v2.Cil.vid
end)

class add_symbolic_constants file =
  let folder globals global =
    match global with
      Cil.GVar (v, _, _) -> VarSet.add v globals
    | Cil.GVarDecl (v, _) -> VarSet.add v globals
    | _ -> globals in
  let globals = Cil.foldGlobals file folder VarSet.empty in
  let globals = VarSet.elements globals in
    
  object
    inherit Cil.nopCilVisitor
	
    method vfunc fundec =
      let mk_symbolic formal =
	let name = formal.Cil.vname ^ "#" in
	Cil.makeLocalVar fundec name formal.Cil.vtype in

      let formal_symbol_pairs = List.map (fun formal -> (formal, mk_symbolic formal))
	  fundec.Cil.sformals in
      let global_symbol_pairs = List.map (fun global -> (global, mk_symbolic global))
	  globals in
      let symbol_pairs = global_symbol_pairs @ formal_symbol_pairs in
      
      let wrap_assertion (formal, symbolic) block =
	let assertion = Cil.BinOp (Cil.Eq,
				   Cil.Lval (Cil.Var formal, Cil.NoOffset),
				   Cil.Lval (Cil.Var symbolic, Cil.NoOffset),
				   Cil.intType) in
	let wrapper = Cil.mkStmt (Cil.If (assertion, block,
					  {Cil.bstmts = [Cil.mkStmt diverge];
					   Cil.battrs = []}, junk_location)) in
	{Cil.bstmts = [wrapper]; Cil.battrs = []} in
      (*fundec.Cil.sbody <- List.fold_right wrap_assertion symbol_pairs fundec.Cil.sbody;*)
      Cil.SkipChildren
  end

let make_cil_file file_name =
  let cil_file = (Frontc.parse file_name) ()
  in
    begin
      try
        let _ = Cil.visitCilFile (new frontc_error_detection_visitor) cil_file in
        ()
      with FrontcError loc ->
        failwith (loc.Cil.file ^ ":" ^ (string_of_int loc.Cil.line) ^
                  ": " ^ "error: frontc reported an error --- see messages above.")
    end ;
    let simplified_cil_file = Simplemem.simplemem cil_file in
    Rmtmps.removeUnusedTemps simplified_cil_file ;

    let cil_file =
      if Options.getValueOfBool "cf" then
	Cil.visitCilFile (new add_symbolic_constants cil_file) cil_file
      else
	cil_file in
    (*
      Message.msg_string Message.Debug "File with symbolic constants added:";
      List.iter (fun x -> Pretty.fprint stdout 80 
      (Cil.printGlobal Cil.defaultCilPrinter () x);
      print_char '\n')
      cil_file.Cil.globals;
    *)

    Cfg.make_cfg simplified_cil_file ;
    if (Options.getValueOfInt "O" >= 1 or Options.getValueOfBool "pe") then begin
      let _ = partiallyEvaluate simplified_cil_file in ()
    end;
    let _ = Cil.visitCilFile (new oneret_visitor) cil_file in
    Cfg.make_cfg simplified_cil_file ;
    rename_locals simplified_cil_file; 
    if (Options.getValueOfInt "O" >= 2) then ignore (Panalysis.t_simplify simplified_cil_file []);
    if (Options.getValueOfInt "O" >= 1) then ignore (Panalysis.slice simplified_cil_file);
(*
    print_string "Starting alias analysis...";
    print_newline();
    Ptranal.analyze_file simplified_cil_file;
   
    let lst = Ptranal.compute_results true in
    List.iter
      (fun (a, b) ->
        print_string ("name ="^ a^" points to ") ;
        List.iter (fun a -> print_string (a^"  ")) b; print_string "\n")
      lst       ;

        Ptranal.compute_aliases true;
*)
    
    ignore (if (Options.getValueOfInt "O">= 1) then
	      let opfile = open_out "driver.i.simp" in
		Cil.printFile Cil.defaultCilPrinter opfile simplified_cil_file ; close_out opfile) ; 
    simplified_cil_file
      
let cil_varinfo_to_symbol v = v.Cil.vname
