(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: lazyModelChecker.ml,v 1.36 2003/08/16 00:17:53 jhala Exp $
 *
 * This file is part of the BLAST Project.
 *)


(**
 * This module implements the lazy model checking algorithm.
 *)

open Ast
open BlastArch


(* [Greg] Very dirty hack because we renamed Stats to VampyreStats in
   vampyre...  We should clean this stats business anyway. *)
module Stats = VampyreStats

module Make_Lazy_Model_Checker =
  functor(Abstraction : ABSTRACTION) ->
struct
  module Abstraction = Abstraction

  module Region = Abstraction.Region
(*  module AR = Abstraction.AR *)
  module Operation = Abstraction.Operation

  module Tree = BiDirectionalLabeledTree

  type error_path = (Region.t * (Operation.t list) * Region.t)

  (**
   * The time stamp ts of a processed covered node indicates that this node
   * is covered by processed uncovered nodes of time stamp lesser than ts.
   * Conversely, the time stamp ts of processed uncovered node indicates
   * that this node may partially cover any processed covered of time stamp
   * greater than ts.
   *)
  type marking_data = {
    mutable time_stamp : int ;
    mutable region     : Region.t ;
  }

  let print_marking_data fmt md =
    Format.fprintf fmt "@[time_stamp=%d;@ region=" md.time_stamp ;
    Region.print fmt md.region ;
    Format.fprintf fmt "@]"

  type marking =
      (* for leaves that have never been processed *)
      Unprocessed
      (* for (already processed) error-free leaves that are covered by
         previously processed (cf. time stamps) nodes *)
    | Processed_Covered of marking_data
      (* for (already processed) error-free nodes that were not covered
         by previously processed nodes at the time they were last
         processed *)
    | Processed_Uncovered of marking_data
      (* for (already processed) error-free leaves that were covered but
         by previously processed nodes that have been refined, and hence
      have to be re-processed *)
    | Processed_Was_Covered_To_Reprocess of marking_data

  let print_marking fmt = function
      Unprocessed ->
        Format.fprintf fmt "@[Unprocessed@]"
    | Processed_Covered md ->
        Format.fprintf fmt "@[Processed_Covered(" ;
        print_marking_data fmt md ;
        Format.fprintf fmt ")@]" ;
    | Processed_Uncovered md ->
        Format.fprintf fmt "@[Processed_Uncovered(" ;
        print_marking_data fmt md ;
        Format.fprintf fmt ")@]" ;
    | Processed_Was_Covered_To_Reprocess md ->
        Format.fprintf fmt "@[Processed_Was_Covered_To_Reprocess(" ;
        print_marking_data fmt md ;
        Format.fprintf fmt ")@]" ;

  type node_data = {
    id           : int ;
    mutable mark : marking ;
  }

  
  let print_node_data fmt nd =
    Format.fprintf fmt "@[Data(@[id=%d;@ mark=" nd.id ;
    print_marking fmt nd.mark ;
    Format.fprintf fmt "@])@]"

  type tree_node = (node_data, Operation.t) Tree.node

  let tree_node_table = Hashtbl.create 1009

  (* helper functions to access a tree node's data *)
  let get_node_data n = Tree.get_node_label n

  let get_op n =
    match (Tree.get_parent_edge n) with
        Some(e) ->
          (Tree.get_edge_label e)
      | None ->
          invalid_arg "get_op: input node has no parent edge"

  let get_parent n =
    match (Tree.get_parent_edge n) with
        Some(e) ->
          (Tree.get_source e)
      | None ->
          invalid_arg "get_parent: input node has no parent edge"

  let get_children n = Tree.get_children n

  let get_id n = (get_node_data n).id

  let get_marking n = (get_node_data n).mark

  let get_marking_data n =
    match (get_marking n) with
        Unprocessed ->
          invalid_arg "get_marking: input node is unprocessed"
      | Processed_Covered md
      | Processed_Uncovered md
      | Processed_Was_Covered_To_Reprocess md ->
          md

  let get_time_stamp n =
    try (get_marking_data n).time_stamp
    with Invalid_argument "get_marking: input node is unprocessed" ->
      invalid_arg "get_time_stamp: input node is unprocessed"

  let print_tree_node fmt n =
    Format.fprintf fmt "@[TreeNode(@[label=" ;
    print_node_data fmt (get_node_data n) ;
    begin
      match (Tree.get_parent_edge n) with
          None ->
            Format.fprintf fmt ";@ no parent"
        | Some(e) ->
            Format.fprintf fmt ";@ incoming_op=" ;
            Operation.print fmt (Tree.get_edge_label e) ;
            Format.fprintf fmt ";@ parent_id=%d"
              (get_id (Tree.get_source e))
    end ;
    let children_ids = List.map get_id (Tree.get_children n)
    in
      Format.fprintf fmt ";@ children_ids=" ;
      Misc.list_printer_from_printer Format.pp_print_int fmt children_ids ;
      Format.fprintf fmt "@])@]"


  let get_region n =
    try (get_marking_data n).region
    with Invalid_argument "get_marking: input node is unprocessed" ->
      invalid_arg "get_region: input node is unprocessed"

  (* helper class to create nodes with appropriate ids *)
  class tree_node_creator =
  object
    val mutable next_id = 0
    method create_root m =
      let this_id = next_id in
        next_id <- next_id + 1 ;
        let rv = ((Tree.create_root { id  = this_id ;
                             mark = m }) : tree_node) 
	in
	  Hashtbl.add tree_node_table this_id (ref rv);
	  rv

    method create_child m op par =
      let this_id = next_id in
        next_id <- next_id + 1 ;
	let rv = 
          ((Tree.create_child { id  = this_id ;
                              mark = m } op par) : tree_node)
	in
	  Hashtbl.add tree_node_table this_id (ref rv);
	  rv

	(* Jhala: It would be nice to have a destroy as well -- or are we relying on some sort of Garbage collection to handle that ? *)
    method delete_children par = 
      Tree.delete_children (par : tree_node)

  end

  exception RestartException

  (* statistic variables *)
  let stats_nb_iterations = ref 0
  let stats_nb_iterations_of_outer_loop = ref 0 (* added for races *)
  let stats_nb_created_nodes = ref 0
  let stats_nb_refinment_processes = ref 0
  let stats_nb_refined_nodes = ref 0
  let stats_nb_proof_tree_nodes = ref 0
  let stats_nb_proof_tree_covered_nodes = ref 0
  let stats_nb_deleted_nodes = ref 0
  let local_restart_counter = ref 1

  let reset_stats =
    stats_nb_iterations := 0 ;
    stats_nb_refinment_processes := 0 ;
    stats_nb_created_nodes := 0 ;
    stats_nb_refined_nodes := 0 ;
    stats_nb_proof_tree_nodes := 0;
    stats_nb_proof_tree_covered_nodes := 0
      
  let print_stats fmt () =
    Format.fprintf fmt "@[<v>@[Nb iterations of outer while loop:@ %d@]" !stats_nb_iterations_of_outer_loop ;
    Format.fprintf fmt "@ " ;
    Format.fprintf fmt "@[<v>@[Nb iterations of reachability:@ %d@]" !stats_nb_iterations ;
    Format.fprintf fmt "@ " ;
    Format.fprintf fmt "@[Nb created nodes:@ %d@]" !stats_nb_created_nodes ;
    Format.fprintf fmt "@ " ;
    Format.fprintf fmt "@[Nb refinment processes:@ %d@]" !stats_nb_refinment_processes ;
    Format.fprintf fmt "@ " ;
    Format.fprintf fmt "@[Nb refined nodes:@ %d@]" !stats_nb_refined_nodes ;
    Format.fprintf fmt "@ " ;
    Format.fprintf fmt "@[Nb proof tree nodes:@ %d@]" !stats_nb_proof_tree_nodes;
    Format.fprintf fmt "@ " ;
    Format.fprintf fmt "@[Nb proof tree covered nodes:@ %d@]@]" !stats_nb_proof_tree_covered_nodes;
    Format.fprintf fmt "@[Nb deleted nodes:@ %d@]@]" !stats_nb_deleted_nodes;
    ()

  (************************************************************************************************)
  (** Model checker data types **)
  (*type model_check_outcome =
    | No_error
    | Error_reached of error_path
    | Race_condition of error_path * error_path * Ast.Expression.lval * Ast.Expression.lval
    | Bad_lock_spec of error_path * Ast.Expression.lval*)

  include BlastArch.ModelCheckOutcome
  type model_check_outcome = error_path outcome

  exception Exit_from_check_concurrent_access of model_check_outcome
	  
  type 'a currentApproximation = 
      { 
      read_whole : Region.t ;
      write_whole : Region.t ;
      read_accesses : (('a * Region.t) list) ; (* the 'a is a tree node *)
      write_accesses : (('a * Region.t) list) (* the 'a is a tree node *)
      } 


  type raceKind =      Read | Write | Both | NoDistinction

  exception RaceConditionException of model_check_outcome 

  (**************************************************************************************************)
  (** Data structures to hold the phi's for lvalues                                                **)

  let global_variable_table = Hashtbl.create 31
  let iter_global_variables f = Hashtbl.iter f global_variable_table 
  let get_global_var_phi sym = 
    try
      Hashtbl.find global_variable_table sym
    with Not_found ->
      (* this lvalue has not been seen so far. So return an empty phi. *)
      ref Region.bot

  let global_lock_asm_table = Hashtbl.create 31
  (* the last is because the global_variable_table gets messed with too often *)
				 

  (**************************************************************************************************)

  (* the following takes a list of (lval,predicate) and adds it to the above table *)

  let process_env_assumptions ll =
    (* initialize all environment assumptions to false *)
    let _ = Abstraction.iter_global_variables 
	      (fun name -> 
		 Hashtbl.add global_variable_table name (ref Region.bot)
	      ) 
    in  
      (* now add the stuff about the interesting locks ... *)                     
    let add (l,p) = 
      Hashtbl.replace global_variable_table l (ref (Abstraction.predicate_to_region p));
      Hashtbl.replace global_lock_asm_table l (Abstraction.predicate_to_region p)
    in
      List.iter add ll

  let print_phis verbose_flag = 
    let phi_print flag name ref_phi = 
      if (flag) then
	begin
	  Message.msg_string Message.Debug 
	    ("Variable " ^ (Ast.Expression.lvalToString name) ^ 
	     "\nPhi =\n" ^ (Region.to_string !ref_phi) ^ "\n\n"); 
	  ()
	end
      else
	begin
	  if !ref_phi = Region.bot 
	  then () 
	  else
	    (Message.msg_string Message.Debug 
	      ("Variable " ^ (Ast.Expression.lvalToString name) ^ 
	       "\nPhi =\n" ^ (Region.to_string !ref_phi) ^ "\n\n"); 
	     ())
	end
    in
    iter_global_variables (phi_print verbose_flag);
    "Done"



  type error_check =
      EmptyErrorAtNode of Region.t * tree_node * (tree_node list)
    | NonEmptyErrorAtRoot of Region.t * tree_node * (tree_node list)

  (**
   * Checks whether a reached error region corresponds to a real error
   * path.  This function takes a node and an error region contained in
   * the node's region as arguments, and does a backward pre computation
   * from that node, along the branch from the root to that node, until
   * it gets an empty region or reaches the root.  It returns the obtained
   * region at the reached ancestor along with the path from this ancestor
   * to the initial node.
   * 
   * @param n     a node of the reachability tree
   * @param err_n a non-empty error region contained in n.region
   * @return a value of type error_check
   *)

  (* do the block-trace analysis, this will also add predicates wherever (locations) required 
     of course we want this function to return the same sort of thing that the original (i.e.
     below function returns :: EmptyErrorAtNode(reg, *)

  let check_error node error_at_node =
    let r_path = Tree.path_to_root node in
let  _ = Message.msg_string Message.Error ("counterex. size:"^(string_of_int (List.length r_path))) in
  
    match Options.getValueOfBool "block" with
	true -> 
	  begin
	
            let reg_l = 
	      ((List.map (get_region)(Misc.chop_last r_path))@[error_at_node] )
	    in
	    let id_l = List.map get_id r_path in 
	    
	    let op_l = List.map (get_op) ( List.tl r_path) in
	      match (Stats.time "block_analyze_trace" (Abstraction.block_analyze_trace reg_l op_l) id_l) with
		  (-1,a) -> 
		    (* the assumption is that -1 indicates it goes all the way to the top *)
		    let (root::path) = r_path in (* should never fail ! *)
		      NonEmptyErrorAtRoot(a,root,path)
		| (i,a) ->
		    (* "i" is the number of hops from the root the falsehood comes *)		     
		    let (n',path) = Misc.list_cut i (r_path) in
		      EmptyErrorAtNode(a,n',path)
          end
      | false ->
	  begin
	    let rec backward_pre n (err_n,subs_err_n) path =
	      match (Tree.get_parent_edge n) with
		  Some(e) ->
		    (* f is n's father and op labels the edge f --> n *)
		    let f = Tree.get_source e
		    and op = Tree.get_edge_label e in
		    let f_reg = get_region f in
		      (* the error subregion at f *)
		      (* JHALA: this subregion business is rubbish. It only blows up the
			  formulas. Not required *)
		      (* JHALA: MAJOR CHANGE! dropping the "capping" in the recursive call *)
	
		    (* JHALA: temporarily comment out!::  let _ = Abstraction.setWPCapRegion f_reg in *)
		    let pre_err_f = (Abstraction.pre err_n op global_variable_table) in
		    let err_f = Region.cap f_reg pre_err_f in
		    let pre_subs_err_f = (Abstraction.spre subs_err_n op global_variable_table) 
		    in

		    let subs_err_f = Region.cap f_reg pre_subs_err_f in
		    (* JHALA: temporarily comment out!:: let _ = Abstraction.resetWPCapRegion () in *)
		      (*let check_emptiness_flag = Abstraction.check_emptiness op in*)
		      if ((*check_emptiness_flag &&*) Stats.time "Emptiness Check:" Region.is_empty subs_err_f)   
		      then
			begin
			  (*Message.msg_printer Message.Error print_tree_node f;*)
			EmptyErrorAtNode(err_f, f, n::path)
			end
		      else
			backward_pre f (pre_err_f,pre_subs_err_f) (n::path)
		| None ->
		    (* n is the root *)
		    NonEmptyErrorAtRoot(err_n, n, path)
	    in
	    let prec_err_n = Abstraction.precise error_at_node in
	    let rv = backward_pre node (prec_err_n,prec_err_n) [] in
	      (* double checking *)
	      (*
	    let _ = 
	      begin
		let reg_l = 
		  ((List.map (get_region)(Misc.chop_last r_path))@[error_at_node] )
		in
		let op_l = List.map (get_op) ( List.tl r_path) in
		  match (Stats.time "block_analyze_trace" (Abstraction.block_analyze_trace reg_l) op_l) with
		      (-1,a) -> 
			(* the assumption is that -1 indicates it goes all the way to the top *)
			let (root::path) = r_path in (* should never fail ! *)
			  NonEmptyErrorAtRoot(a,root,path)
		    | (i,a) ->
			(* "i" is the number of hops from the root the falsehood comes *)		     
			let (n',path) = Misc.list_cut i (r_path) in
			let _ = Message.msg_printer Message.Error print_tree_node n' in
			  EmptyErrorAtNode(a,n',path)
              end
	    in *)
	      rv
	  end

  let check_error_fwd node error_at_node  =
    let rec get_path node path proj_path =
      match (Tree.get_parent_edge node) with
        Some (e) ->
            (* f is node's father and op labels the edge f --> node *)
          let f = Tree.get_source e and op = Tree.get_edge_label e
          in 
          get_path f ((f, Some op):: path) (f::proj_path)
      | None -> (path, proj_path)
    in
    let (error_path, path_projected_on_tree_nodes) = get_path node [ (node, None) ] [ node ]
    in
    Message.msg_string Message.Debug "Check error forward called with" ;
    Message.msg_string Message.Debug "Path = ";
    List.iter (function node -> Message.msg_printer Message.Debug print_tree_node node) path_projected_on_tree_nodes ;
    Message.msg_string Message.Debug "end of path\n\n\n" ;
    let root = match path_projected_on_tree_nodes with
      h :: _ -> h
    | _ -> failwith "Empty error path in check_error_fwd"
    in
    let reg_true = Abstraction.precise (get_region root)
    in
    (* now do a concrete post on the error path *)
    let rec forward_post current_reg path =
      (*Message.msg_string Message.Debug ("\n\nforward post called with region "^(Region.to_string current_reg));
	*)
      match path with
        [] -> (* Valid error *) 
	  let r = (Region.cap (Abstraction.trim_region_to_globals current_reg) error_at_node) in
	  if (Region.is_empty r) then
	    failwith "Rupak: this case not handled yet"
	  else
	    NonEmptyErrorAtRoot (r , root, List.tl path_projected_on_tree_nodes)
      | (f, op) :: rest ->
            begin
              match op with
		None -> (* reached the node where the error is without trouble *)
		  (* check if satisfiable *)
		  let err_f = Region.cap (Abstraction.trim_region_to_globals error_at_node) 
					 (Abstraction.trim_region_to_globals current_reg) in
		  if (Stats.time "Emptiness Check (Forward):" Region.is_empty err_f) then
		    begin
		      Message.msg_string Message.Debug "Emptiness check succeeds at end of path: this is a false error";
		      check_error f error_at_node
		    end
		  else
		    begin
		      Message.msg_string Message.Debug "Emptiness check fails at end of path: this is a true error";
		      NonEmptyErrorAtRoot ( err_f , root, List.tl path_projected_on_tree_nodes)
		    end
	      |	Some oper ->
		  begin
		    (* do a strongest postcondition computation.
		       if oper is an assume, do a satisfiability check
		       (find predicates if unsatisfiable).
		       otherwise recursively call forward_post on the rest of the path
		       *)
		    Message.msg_string Message.Debug ("Now doing post of "^(Operation.to_string oper)) ;
		    let next_reg = Abstraction.post current_reg oper global_variable_table 
		    in
(*
		    if (Abstraction.check_emptiness oper) then 
*)
		    if (true) then
		      if (Stats.time "Emptiness Check (Forward):" Region.is_empty next_reg) then
			begin
			  Message.msg_string Message.Debug "Emptiness check succeeds: this is a false error";
			  let nd = match rest with
			    (h, _) :: _ -> h
			  | _ -> failwith "strange error path"
			  in
			  try
			    check_error nd (get_region nd)
			  with Invalid_argument _ -> (* This node was marked Unprocessed *)
			    begin (* get region in the tedious way *)
			      let reg = Abstraction.post (get_region (get_parent nd)) (get_op nd) global_variable_table in
			      (get_node_data nd).mark <- 
				 Processed_Was_Covered_To_Reprocess { time_stamp = 0; region = reg } ;
			      check_error nd (reg)
			    end
			end
		      else
			forward_post next_reg rest
		    else
		      forward_post next_reg rest
		  end
            end
    in
    (* forward simulate starting from root and True *)
    forward_post reg_true error_path


  (* Greg: I don't like this get_op... try to add the operations in the path? *)



  let rec refine_path anc path =
    match path with
        []   -> ()
      | n::p ->
          stats_nb_refined_nodes := !stats_nb_refined_nodes + 1 ;

          Message.msg_string Message.Debug "Refining node:" ;
          Message.msg_printer Message.Debug print_tree_node n ;
          Message.msg_string Message.Debug "whose ancestor is:" ;
          Message.msg_printer Message.Debug print_tree_node anc ;

          let anc_reg = get_region anc
          and n_marking_data = get_marking_data n in

            Message.msg_string Message.Debug "calling Abstraction.post with region:" ;
            Message.msg_string Message.Debug ("    region    : " ^ (Region.to_string anc_reg)) ;
            Message.msg_string Message.Debug ("    operation : " ^ (Operation.to_string (get_op n))) ;

            n_marking_data.region <- Abstraction.post anc_reg (get_op n) global_variable_table;

            Message.msg_string Message.Debug "The refined node looks like:" ;
            Message.msg_printer Message.Debug print_tree_node n ;

            refine_path n p


  let update_tree_after_refinment root unprocessed_nodes ts =
    let update_and_compute_bfs n = 
      let nl_int = ref [] in
      let nl_leaf = ref [] in
      let reached_reg = ref Region.bot in      
      let f_int = (fun n -> reached_reg := Region.cup !reached_reg (get_region n)) in
      let f_leaf = (fun n -> nl_leaf := n::(!nl_leaf)) in
      let _ = Tree.traverse_bfs f_leaf f_int n in
      let nl_leaf = List.rev !nl_leaf in
      let _ = Message.msg_string Message.Debug "Leaf Nodes" in
      let _ = List.iter 
		(fun n -> Message.msg_printer Message.Debug  print_tree_node n) 
		nl_leaf
      in
      let process n = 
	let n_data = get_node_data n in 
	  match n_data.mark with
	      Unprocessed ->
		assert (not (Tree.has_child n)) ;
		begin
		  match (Options.getValueOfString "comp") with
		      "path" -> ();
		    | _ -> 	
			unprocessed_nodes#add_element n ;
			Message.msg_string Message.Debug "Adding to unprocessed#nodes";
			Message.msg_printer Message.Debug print_tree_node n ;
		end
	    |  Processed_Was_Covered_To_Reprocess md ->
		 (* --- for debugging --- *)
		 assert (not (Tree.has_child n)) ;
		 begin
		   match (Options.getValueOfString "comp") with
		       "path" -> ()
		     | _ -> 	
			 Message.msg_string Message.Debug "Adding to unprocessed#nodes";
			 Message.msg_printer Message.Debug print_tree_node n ;
			 unprocessed_nodes#add_element n ;
		    	 (* Region.cup reached_region md.region  JHALA CHECK CHECK CHECK *)
		 end
	    | Processed_Covered md ->
		(* --- for debugging --- *)
		assert (not (Tree.has_child n));
		if (not (Abstraction.covered md.region !reached_reg)) then
                  begin
                    n_data.mark <- Processed_Was_Covered_To_Reprocess md ;
                    Message.msg_string Message.Debug "Adding to unprocessed#nodes";
		    Message.msg_printer Message.Debug print_tree_node n ;
		    unprocessed_nodes#add_element n 
                  end
		else
                  ()
            | Processed_Uncovered md ->
		begin
		  (* it should be the case that this node, oddly enough, has NO children, i.e. this pc has no children,
		     so it doesnt matter if I add it to cover or not. Indeed I shouldnt, and I shouldnt try to add it to 
		     the list of things to process either -- RJ. *)
		  if (Abstraction.enabled_ops (get_region n) <> []) then
		    begin
		      Message.msg_string Message.Normal "About to die";
		      Message.msg_printer Message.Normal print_tree_node n ;
		      failwith "ugh! process_n_leaf gets an internal node!"
		    end
		end
      in
      let _ = List.iter process nl_leaf in
	!reached_reg
    in
    let rec update_and_compute reached_region n =
      let n_data = get_node_data n in
        match n_data.mark with
            Unprocessed -> 
	      assert (not (Tree.has_child n)) ;
	      begin
		match (Options.getValueOfString "comp") with
		    "path" -> ();
		  | _ -> 	
		      unprocessed_nodes#add_element n ;
		      Message.msg_string Message.Debug "Adding to unprocessed#nodes";
		      Message.msg_printer Message.Debug print_tree_node n ;
	      end;
	      reached_region
		
          | Processed_Was_Covered_To_Reprocess md ->
              (* --- for debugging --- *)
              assert (not (Tree.has_child n)) ;
	      begin
		match (Options.getValueOfString "comp") with
		    "path" -> reached_region
		  | _ -> 	
		      Message.msg_string Message.Debug "Adding to unprocessed#nodes";
		      Message.msg_printer Message.Debug print_tree_node n ;
		      unprocessed_nodes#add_element n ;
		      
		      Region.cup reached_region md.region
	      end
          | Processed_Covered md ->
              (* --- for debugging --- *)
              assert (not (Tree.has_child n)) ;
              if (md.time_stamp >= ts) then
                begin
                  n_data.mark <- Processed_Was_Covered_To_Reprocess md ;
                  Message.msg_string Message.Debug "Adding to unprocessed#nodes";
		  Message.msg_printer Message.Debug print_tree_node n ;
		  unprocessed_nodes#add_element n ;
                  reached_region
                end
              else
                Region.cup reached_region md.region
          | Processed_Uncovered md ->
              let new_reached_region = Region.cup reached_region md.region in
                List.fold_left update_and_compute new_reached_region (get_children n)
    in
      match (Options.getValueOfBool "bfs") with
	  true -> update_and_compute_bfs root
	| false -> update_and_compute Region.bot root


(* ********************************************************************** *) 
  (* Jhala: Adding proof tree data structures *)
    (* module BlastProof = Abstraction.BlastProof *)
			  
    type proof_tree_node_data = {
      pt_id       : int; 
      pt_region   : Abstraction.proof_region;
      covering_nodes :  (Abstraction.lemma_index * (proof_tree_node ref list)) option; 
     (* Jhala: Right now, just dumping the tree, 
	the last field will say which nodes cover the present node *)
    }
    and proof_tree_edge = (Operation.t * Abstraction.lemma_index) 
    and proof_tree_node = (proof_tree_node_data, proof_tree_edge) Tree.node(* Jhala: not storing the operation on the edge *)
  (* Jhala: end proof tree data structures -- 
     rest are in the module "BlastProof" in abstraction.ml 
     -- Abstraction.lemma_index is just "int" *)


let print_proof_tree (root: proof_tree_node) =
  let print_pt_id n1 = print_int (Tree.get_node_label n1).pt_id;print_string " ; "
  in
  let print_pt_children_ids n1 = 
    print_string "Children of ";
    print_pt_id n1;print_string ": \n"; 
    print_string "["; 
    List.iter print_pt_id (Tree.get_children n1);
    print_string "]\n"
  in
  let print_pt_edge pt_e = 
    match pt_e with
	(op,i) ->  
	  begin 
	    print_string "op: ";
	    Message.msg_printer Message.Debug Operation.print op ; 
	    print_string "\n proof: " (* URGENT ! ;print_int i --TYPE ISSUES *)
	  end
  in
  let print_pt_node_data n_data = 
    print_int n_data.pt_id;print_string "\n";
    Abstraction.print_proof_region n_data.pt_region;
    match n_data.covering_nodes with
	None -> print_string "Not covered \n"
      | Some(i,l) -> print_string "Covered \n"
  in

  let rec print_pt_node n =
    let _ = 
      match (Tree.get_parent_edge n) with
	  None -> print_string "ROOT NODE Id: ";
	| Some(e) ->
	    let par = Tree.get_source e in
	      begin
		print_string "----par: ";print_pt_id par;print_string "\n";
		print_pt_edge (Tree.get_edge_label e);
		print_string "---->\n";
		print_string "INT NODE Id: "
	      end 
    in
      print_pt_node_data (Tree.get_node_label n);
      print_pt_children_ids n;
      List.iter print_pt_node (Tree.get_children n)
  in
    print_pt_node root
      
(* count the size of the proof tree *)
let count_tree_nodes root = 
  Message.msg_string Message.Debug "Counting tree nodes";
  let update_counter n = 
    (* let us just count how many covered nodes there are *)
    if ((List.length (get_children n ) = 0)) then stats_nb_proof_tree_covered_nodes := !stats_nb_proof_tree_covered_nodes + 1;
    stats_nb_proof_tree_nodes := !stats_nb_proof_tree_nodes + 1; ()
  in
  let leaf_check_fn n = (get_children n) = [] in
    stats_nb_proof_tree_nodes := Tree.count_nodes_descendants  root;
    stats_nb_proof_tree_covered_nodes := Tree.count_f_descendants leaf_check_fn root

  (* Jhala: Functions to construct the proof tree *)
   let build_proof_tree () root =
      Message.msg_string Message.Debug "In build_proof_tree"; 
      Message.msg_string Message.Error "In build_proof_tree"; (* Say that we are building proof 
                                                                 on both stdout and stderr *)
      (* open file for outputting vampyre proofs *)
      let lfPfFile = Options.getValueOfString "pffile" in
      let _ =    (VampyreErrormsg.proofChannel := open_out lfPfFile;
		  ignore (VampyrePretty.fprintf (!VampyreErrormsg.proofChannel) "ANN_STARTLF\n");
		  ignore (VampyrePretty.fprintf (!VampyreErrormsg.proofChannel) "IMPORT blast.ans\n"))
      in
      let rec process_node (n (* : tree_node *) ,np (* : proof_tree_node*) ):proof_tree_node =
	let process_child n_1  = 
       	  let op = get_op n_1 in (* n --op--> n_1 *)
      (* We make a corresponding np --op--> np_1 *)
	  let (child_proof_region,lemma_token) = 
	    Abstraction.post_proof (get_region n) op global_variable_table
	  in
          let np_1_data = 
	    {  pt_id  = (get_id n_1);
	       pt_region = child_proof_region;
	       covering_nodes = None; (* Jhala: TBD, adding the covering proof *)
	    }
	  in
	    stats_nb_proof_tree_nodes := (!stats_nb_proof_tree_nodes) + 1;
	    let np_1 = ((Tree.create_child np_1_data (op,lemma_token) np) : proof_tree_node)
	    in
	      if  (Abstraction.check_reg_eq child_proof_region (get_region n_1)) 
	      then
		process_node (n_1,np_1)
	      else  (Message.msg_string Message.Error "mismatched regions!";
		     process_node (n_1, np_1))
		
	in
	let children = get_children n in
	  (* let us just count how many covered nodes there are *)
	  if ((List.length children) = 0) then stats_nb_proof_tree_covered_nodes := !stats_nb_proof_tree_covered_nodes + 1;
	  let _ = List.map process_child children in
	    np
      in		
	(* first make the root of the proof tree *)
      let pt_root_data = 
	{ pt_id  = (get_id root);
	  pt_region = Abstraction.extract_proof_region (get_region root) ;
	  covering_nodes = None; (* Jhala: TBD, adding the covering proof *)
	}
      in
      let pt_root = process_node (root, ((Tree.create_root pt_root_data) : proof_tree_node))	
      in
        (*let dump_node_to_chan chan node = Marshal.to_channel chan  (Tree.get_node_label node) [] 
	  in *)
	
      let tree_chan = open_out_bin "foo.tree" in
      Marshal.to_channel tree_chan pt_root [];
      close_out tree_chan;
      Abstraction.print_proofTable ();
	(* print_proof_tree pt_root;*)
      ignore (VampyrePretty.fprintf !VampyreErrormsg.proofChannel "ANN_ENDLF\n");
      close_out (!VampyreErrormsg.proofChannel);
      let _ = Unix.system ("gzip -9 -f foo.tree") in
      let lfsz = (Unix.stat lfPfFile).Unix.st_size in 
      let treesz = (Unix.stat "foo.tree.gz").Unix.st_size in
      Message.msg_string Message.Normal ("Proof size: (tree)"^(string_of_int treesz));
      Message.msg_string Message.Normal ("         (Vampyre)"^(string_of_int lfsz));
      
      ()
	(* Jhala: to begin with just print this damn tree *)
(****************************************************************************)
(****************************************************************************)

(* GUI hooks *)
(* some functions that dumps the counterexample data structure to a file *)
(* the type of a trace is a (bool * Operation.t list * Region.t list) *)

   let dumptrace (b,op_l,reg_l) =
     let creg_l = List.map Region.concretize reg_l in
     let trace_chan = open_out_bin (Options.getValueOfString "tracefile") in
     let op_string = ref "" in
     let check_o2s x = 
       let (l1,l2,rv) = Abstraction.op_to_string x in
         let s = Printf.sprintf "%d :: %s :: %d \n\n" l1 rv l2 in
	   op_string := s^(!op_string);
           Printf.printf "%d :: %s :: %d \n" l1 rv l2 ; 
  	   (l1,l2,rv) 
     in
       Marshal.to_channel trace_chan (b,List.map check_o2s op_l,List.map Abstraction.reg_to_strings creg_l) [];
       close_out trace_chan;
       Misc.append_to_file "traces" ("Start trace \n "^(!op_string)^"End trace\n");
       ()

  let dumppath node region_at_node =
    let _ = Message.msg_string Message.Error "in dumppath" in
    let p_to_root = Tree.path_to_root node in
    let list_of_regs = (List.map (get_region)  (Misc.chop_last p_to_root))@[region_at_node] in
    let list_of_ops = List.map (get_op) ( List.tl p_to_root) in
    dumptrace (false, list_of_ops, list_of_regs)


let dump_trace_to_node n n_region genuine =
   let n_data   = (Tree.get_node_label n) in
   let path_to_root = Tree.path_to_root n in
   let list_of_regs = (List.map (get_region)  
			 (Misc.chop_last path_to_root))@[n_region] in
   let list_of_ops = List.map (get_op) ( List.tl path_to_root) in
     dumptrace (genuine, list_of_ops,list_of_regs)


let print_whole_tree root = 
  Message.msg_string Message.Debug "Printing Reachability Tree";
  let f n = Message.msg_printer Message.Debug print_tree_node n
  in
  let rec _ac_print n = 
    f n; 
    List.iter _ac_print (Tree.get_children n)
  in
    _ac_print root
      

let demo_process n n_region genuine =

 match (Options.getValueOfBool "demo" (*|| genuine *)) with
      true ->
	begin
	  let s = if genuine then "Genuine" else "Spurious" in
	  let outs = Printf.sprintf "*** %s Counterexample Found!***" s in
	    Message.msg_string Message.Normal outs;
	    dump_trace_to_node n n_region genuine;
	Message.msg_string Message.Error "waiting to hear from gui";
	let _ = input_line stdin in
	  Message.msg_string Message.Error "heard from gui!";
	  ()
 end 
| false -> ()


(****************************************************************************)
(* TAR algorithm *)
(****************************************************************************)
	
  let model_check_tar init o_err = 
    Message.msg_string Message.Debug "Entering model_check" ;
    Message.msg_string Message.Debug ("   init region: " ^ (Region.to_string init)) ;
    Message.msg_string Message.Debug 
      (match o_err with None -> "No error region" | Some e -> "   error region: " ^ (Region.to_string e)) ;
    try   
     let flag = ref true 
      in
      while (!flag) do
      begin
	Message.msg_string Message.Debug "\n\n%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n" ;
        Message.msg_string Message.Debug "Next iteration of model-check's outer while-loop\n\n\n" ;
	Message.msg_string Message.Debug "Printing Assumptions\n" ;
	(* debug *)
	print_phis true;
	
	stats_nb_iterations_of_outer_loop := !stats_nb_iterations_of_outer_loop + 1;
	flag := false;
	
	
	(* Routines to check for concurrent access and to check for valid errors / do refinement *)
	let do_focus_and_refine err_anc anc path =
	  Message.msg_string Message.Debug "\n\nLet's refine the nodes along that path" ;
	  let anc_region = (get_region anc) in
	  let _ = Message.msg_string Message.Debug (* Normal*) "About to call focus" in
	  try
            let focused_anc_region = 
              Stats.time "focus" (Abstraction.focus anc_region) err_anc
            in
            Message.msg_string Message.Debug "Done calling Focus";
            Message.msg_string Message.Debug ("The focused region for the ancestor is: " ^ (Region.to_string focused_anc_region)) ;
            stats_nb_refinment_processes := !stats_nb_refinment_processes + 1
	  with Abstraction.NoNewPredicatesException ->
            Message.msg_string Message.Debug (* Normal *)"do_focus_and_refine: This is a false error" ;
            Message.msg_string Message.Debug "The check_error function ended up with" ;
            Message.msg_string Message.Debug ("     an empty error: " ^ (Region.to_string err_anc)) ;
            Message.msg_string Message.Debug "     at ancestor: " ;
            Message.msg_printer Message.Debug print_tree_node anc ;
            Message.msg_string Message.Debug "     along the path (from the ancestor to the current node):" ;
            Message.msg_printer Message.Debug  print_tree_node anc ;
(*	    let path_to_root = Tree.path_to_root n in 
	    let list_of_regs = (List.map (get_region)  (Misc.chop_last path))@[n_region] in
*)
	    let list_of_regs = (List.map (get_region)  path) in
	    let list_of_ops = List.map (get_op) ( List.tl path) in
	    dumptrace (false, list_of_ops, list_of_regs) ;
	    failwith "No new predicates" 
(*
            List.iter (function x -> Message.msg_printer Message.Normal print_tree_node x) path
*)
	in
	let focus_and_refine args =
	  let n_refinements = !stats_nb_refinment_processes in
	  List.iter (function (err_anc, anc, path) ->
            do_focus_and_refine err_anc anc path)
            args;
	  if n_refinements = !stats_nb_refinment_processes
	  then failwith "No new predicates found by focus!"
	in
	(* raise exception if there is a possible race
	   return true if the command on edge accesses global_var and false otherwise *)
	(* 
	   rkind -> kind of race to check for. rkind is Read, Write, Both, or NoDistinction
	   operation -> the edge label of the cfa that needs to be checked for races
	   node -> the current node in the reachability tree
	   region_at_node -> R /\ wp (op, true) , where R is the region at the node computed in the reachability tree
	   global_var -> the variable being checked
	   phi_list -> the current assumptions about the global variable
	   *)

	let check_concurrent_access rkind operation node region_at_node global_var gv_alias phi_list  =
	  (* the phi_list given above is gv_alias's phi_list *)
	  (* this function should only be called with global_var/gv_alias such that the pair of lvals is
	     such that they might be aliased ... if it is known that they are NOT ever aliased then they should not 
	     be sent to this function ... 
	  *)
	  Message.msg_string Message.Debug "Checking concurrent access." ;
	  Message.msg_string Message.Debug ("Region at node: "^(Region.to_string region_at_node)) ;
	  Message.msg_string Message.Debug ("Main lval = "^(Expression.lvalToString global_var)) ;
	  Message.msg_string Message.Debug ("Alias lval = "^(Expression.lvalToString gv_alias));
	  if (Abstraction.is_lock global_var || Abstraction.is_lock gv_alias) then
	    begin
	      (* JHALA: TO BE DONE TBD *)
	      (* we have to insert the checking for locks -- i.e. that region is contained inside the predicate
		 business 
	      *)
	      (* 1. is global_var \in the global_lock_asm_table -- if so then check: leq region_at_node spec
		 if not -- then do counterexample guided refinement !! see if present path is feasible if so is sp contained inside the spec -- if so learn new preds and restart -- if not -- ERROR !
		 else gv_alias is in the damn table --> do the check with (cap region_at_node eq_addr_const) spec
	      *)
	      Message.msg_string Message.Debug "This is a lock variable. Need not check this." ;
	      ()
	    end
	  else
	    (* check for race condition *)
	  begin
	    let (* the trim_region_to_globals simply projects away all control info -- ie pc, stack *)
		r = Abstraction.trim_region_to_globals region_at_node 
	    in
	      begin
		let check_for_race (othernode, otherregion) =
		  (* check for race does the sat. check for (node, region_at_node) (othernode, otherregion) *)
		  (* raise an exception if a possible error is found *)
		  (* construct the formula and check satisfiability *)
		  (* this function returns the following: 
		     either it raises no exceptions -- 
		     meaning all the intersections in the present phi_tables are empty ...
		     or it raises  No_error -- which means that there WAS some non-empty intersection but it was
		     bogus because either a. at least one of the 2 offending paths is infeasible or b. the conjunction
		     of the 2 sp is infeasible ... in either case No_error is caught later and the model checker 
		     "restarted" appropriately 
		     Question: what if the address constraint predicate is returned by focus ? 
		     ... where will we add it ? will we need to add it at all ? 
		  *)
		  begin
		    match 
		      Abstraction.multiply_instantiable 
			r (Abstraction.trim_region_to_globals otherregion) global_var gv_alias
		    with
		      None -> (* Good, not satisfiable, do nothing, 
				 means the present abs is enough to say that the regions dont intersect *) ()
		    | Some reg -> (* the region is satisfiable: potential race! *)
			begin
			  Message.msg_string Message.Normal "Found a possible race! Checking feasibility." ;
			  Message.msg_string Message.Error "Found a possible race! Checking feasibility." ;
			  Message.msg_printer Message.Debug print_tree_node node ;
			  Message.msg_string Message.Debug ("This region is "^(Region.to_string region_at_node)) ;
			  Message.msg_string Message.Debug "\n\nOther node :" ;
			  Message.msg_printer Message.Debug print_tree_node othernode ;
			  Message.msg_string Message.Debug ("Other region is "^(Region.to_string otherregion)) ;
		                            (* satisfiable, check feasibility of each error *)
				 (* DEBUG *)
			  
			  Message.msg_string Message.Normal "\n\nTrace 1:" ;
			  dumppath node region_at_node ;
			  
			  Message.msg_string Message.Normal "\n\nTrace 2:" ;
			  dumppath othernode otherregion ;


			  (* first I need to put back the control info (pc, stack) in otherregion *)
			  (* RUPAK: this does not work now, I simply replace otherregion with the region in othernode to test
			   *)
			  begin
			    match (check_error_fwd node region_at_node, check_error_fwd othernode otherregion) with
			      (NonEmptyErrorAtRoot (err_root1, root1, path1), NonEmptyErrorAtRoot (err_root2, root2, path2)) ->
				begin
				Message.msg_string Message.Normal "\nBoth paths are feasible! But is their intersection feasible?" ;
				Message.msg_string Message.Error "\nBoth paths are feasible! But is their intersection feasible?" ;
				let intersection =
				  Abstraction.make_multiply_instantiable_region
				    (Abstraction.trim_region_to_globals err_root1)
				    (Abstraction.trim_region_to_globals err_root2)
				    global_var 
				    gv_alias
				in
				  match Region.is_empty intersection with
				      true ->
					begin
					  (* the paths are not feasible together *)
					  (* in this case we take the empty intersection and focus an arbitrary
					     node with respect to this empty region.
					     We ignore the result returned by focus, because we are only interested
					     in the side effect: some new predicates get added.
					     Not functional. But functional. (Bad pun here).
					     
					  *)
					  Message.msg_string Message.Normal "The intersection is empty. Refining." ;
					  try
					    ignore (Stats.time "focus" 
						      (Abstraction.focus 
						       (Abstraction.create_abstract_true_region operation) )
						      intersection)
					  with Abstraction.NoNewPredicatesException ->
					    begin
					      Message.msg_string Message.Normal "Intersection: This is a false error" ;
					      Message.msg_string Message.Normal "The check_error function ended up with" ;
					      Message.msg_string Message.Normal ("     an empty intersection: " ^ 
										 (Region.to_string
										    intersection)) ;
					      print_phis true ;
                                              Message.msg_string Message.Normal "\n\nTrace 1:" ;
					      dumppath node region_at_node ;
					      Message.msg_string Message.Normal "\n\nTrace 2:" ;
					      dumppath othernode otherregion ;

					      
					      failwith "No new predicates."
					    end
					end
				    | false ->
					begin
					  Message.msg_string Message.Debug "The intersection is feasible. Real race." ;
					  raise (Exit_from_check_concurrent_access 
						   (Race_condition ((err_root1,
								     List.map get_op path1,
								     region_at_node), 
								    (err_root2,
								     List.map get_op path2,
								     otherregion),global_var , gv_alias ) ))
					end
				end
			    | (NonEmptyErrorAtRoot (_, _, _),
			       EmptyErrorAtNode (err_anc2, anc2, path2)) ->
		              (* find more predicates and restart *)
				Message.msg_string Message.Normal "\nSecond path is infeasible! " ;
				Message.msg_string Message.Error "\nSecond path is infeasible! " ;

				 focus_and_refine [(err_anc2, anc2, path2)]
			    | (EmptyErrorAtNode (err_anc1, anc1, path1),
			       NonEmptyErrorAtRoot (_, _, _)) ->
				Message.msg_string Message.Normal "\n1st path is infeasible! " ;
				Message.msg_string Message.Error "\n1st path is infeasible! " ;
		             (* find more predicates and restart *)
				 focus_and_refine [(err_anc1, anc1, path1)]
			    | (EmptyErrorAtNode (err_anc1, anc1, path1), EmptyErrorAtNode (err_anc2, anc2, path2))
			      ->
				Message.msg_string Message.Normal "\nBoth paths is infeasible! " ;
				Message.msg_string Message.Error "\nBoth paths is infeasible! " ;
				focus_and_refine [(err_anc1, anc1, path1);
						    (err_anc2, anc2, path2)]
			  end;
			  raise (Exit_from_check_concurrent_access No_error)		
			end;
		
		  end
		in
		match rkind with
		  NoDistinction ->
		    (* first check if node_region is included in phi_list -- this is the easy case *)
		    if (Region.leq region_at_node (phi_list.write_whole)) 
		    then ()
		    else
		      begin
			List.iter check_for_race ((node, region_at_node) :: phi_list.write_accesses) ; 
		        (* if I am here, then there was no error *)
			()
		      end
		| Read -> 
		    (* first check if node_region is included in phi_list -- this is the easy case *)
		    if (Region.leq region_at_node (phi_list.read_whole)) 
		    then ()
		    else
		      begin
		       (* check for conflicts with write accesses. Since this node was a read, we
			  need not check for errors against this node. *)
			List.iter check_for_race (phi_list.write_accesses) ;
                        (* if I am here, then there was no error *)
			()
		      end
		| Write ->
		    		    (* first check if node_region is included in phi_list -- this is the easy case *)
		    if (Region.leq region_at_node (phi_list.write_whole)) 
		    then ()
		    else
		      begin
		       (* check for conflicts with write accesses. *)
			List.iter check_for_race ((node, region_at_node) :: phi_list.write_accesses) ;
		       (* check for conflicts with read accesses. *)
			List.iter check_for_race (phi_list.read_accesses) ;
                        (* if I am here, then there was no error *)
			()
		      end
		| Both ->
		    		    (* first check if node_region is included in the write phi_list  *)
		    if (Region.leq region_at_node (phi_list.write_whole) ) 
		    then (* no new bugs can occur due to this write, so check for read bugs *)
		      begin
			(* this could be done by a recursive call to check_concurrent_access Read .. *)
			if (Region.leq region_at_node (phi_list.read_whole)) 
			then ()
			else
			  begin
		            (* check for conflicts with write accesses. Since the write region at this node
			       was included in write_whole, we
			       need not check for errors against this node. *)
			    List.iter check_for_race (phi_list.write_accesses) ;
                            (* if I am here, then there was no error *)
			    ()
			  end
		      end
		    else
		      begin
		       (* check for conflicts with write accesses. *)
			(* step 1*) List.iter check_for_race ((node, region_at_node) :: phi_list.write_accesses) ;
		       (* check for conflicts with read accesses. *)
			(* step 2*) List.iter check_for_race ((node, region_at_node) :: phi_list.read_accesses) ;
                        (* if I am here, then there was no write error *)
			(* also, the possible read errors due to this region have already been
			   checked in step 1. So I need not check for read errors again.
			   RUPAK: check this!
			   *)
			()
		      end
	      end
	  end
	(* end check_concurrent_access *) 
	in
	(* initialize current approximations to environment assumptions to empty list *)
	(* the current approximations are a map Global Variables -> List of tree nodes *)
	(* the environment assumption for x is \/ { n_reg | n\in accNodes(x) } *)
	let accNodes = Hashtbl.create 31 
	in
	(* initialize the hash table to [] for each global *)
	Abstraction.iter_global_variables 
	  (fun name -> Hashtbl.add accNodes name { read_whole = Region.bot  ; read_accesses = [] ;
						   write_whole = Region.bot ; write_accesses = []
						 } 
	      ) ;

	(* set up the reachability search *)
        (* the reached region is initially empty *)
	let reached_region = ref Region.bot
	    
        (* helper object to create nodes *)
	and tnc = new tree_node_creator
	    
	    
        (* the set of pending nodes to be processed *)
	and unprocessed_nodes = 
	  if (Options.getValueOfBool "bfs") 
	  then new Search.bfsiterator
	  else new Search.dfsiterator

        (* the global time used for time stamps *)
	and time = ref 0 in
	
        (* The function for constructing children .. and adding them to the tree *)
	let construct_children node accu op =
	  (tnc#create_child Unprocessed op node)::accu
						     
	in				 
	
        (* the root of the reachability tree *)
	let tree_root =
	  tnc#create_root (Processed_Was_Covered_To_Reprocess { time_stamp = !time ;
								region = init ; })
	in
	let do_subtree_cut_refinement anc new_anc_region =
          let _ = 
	    stats_nb_deleted_nodes := !stats_nb_deleted_nodes + 
	    (Tree.count_nodes_descendants anc) 
	  in
	  let anc_data = Tree.get_node_label anc in
	  let anc_marking_data = get_marking_data anc in
        (* update the ancestor region *)
	  Message.msg_string Message.Debug "Updating the ancestor's region..." ;
	  anc_marking_data.region <- new_anc_region ;
          
	  Message.msg_string Message.Debug "This ancestor now looks like:" ;
	  Message.msg_printer Message.Debug print_tree_node anc ;
        (* Note that here there is no path to update as the entire subtree is going to be chopped off... *)
	  Message.msg_string Message.Debug "Now deleting subtree:" ;
	  tnc#delete_children anc;
	  unprocessed_nodes#empty_out ();
      (* Jhala: I'm assuming that its ok to just empty this out and fill it back up when we are traversing the tree later ... this whole traversal business is not good by the way, there should be a much more incremental way to do it *)
	  Message.msg_string Message.Debug "Now re-adding children:" ;
	  let children = List.fold_left (construct_children anc) [] 
	      (Abstraction.enabled_ops new_anc_region) in
	  Message.msg_string Message.Debug "The children are:" ;
	  List.iter (function child -> Message.msg_printer Message.Debug print_tree_node child) children ;
	  
       (* add the children to the set of pending nodes *)
      (* Jhala: This should all be done by the the subtree cut 
	 Message.msg_string Message.Debug "Adding the children to the set of pending nodes" ;
	 unprocessed_nodes#add_list children ;
	 *)
      (* mark n as processed and uncovered *)
	  Message.msg_string Message.Debug "Updating the node's marking" ;
	  
	  anc_data.mark <- Processed_Uncovered anc_marking_data ;
	  Message.msg_string Message.Debug "This node now looks like:" ;
	  Message.msg_printer Message.Debug print_tree_node anc;
      (* add n's region to the reached region *)
	  Message.msg_string Message.Debug "Updating the currently reached region" ;
	  reached_region := 
	     Stats.time "update_tree_after_refinement[CUT]" 
	       (update_tree_after_refinment tree_root unprocessed_nodes) anc_marking_data.time_stamp
	in
	
	let no_error n n_data n_marking_data =
          (* no error found at this node *)
	  Message.msg_string Message.Debug "No error found at this node" ;
	  
          (* let's test whether this node is covered *)
	  Message.msg_string Message.Debug "Let's test whether this node is covered" ;
	  let coverable = 
	    if (Options.getValueOfBool "cov") then Abstraction.coverable n_marking_data.region else true
	  in
	  let covered = 
	  if coverable then 
	    (Stats.time "Covered check:"
	       (Abstraction.covered n_marking_data.region) !reached_region)
	  else false in
	  if covered
	  then
            begin
              (* this node is covered *)
              Message.msg_string Message.Debug "This node is covered" ;
	      
              (* mark n as processed and covered *)
              n_data.mark <- Processed_Covered n_marking_data
            end
	  else
            begin
              (* this node is not covered *)
              Message.msg_string Message.Debug "This node is not covered" ;
              Message.msg_string Message.Debug "Constructing its successor children..." ;
	      
              (* we construct its successor children *)
	      
              let children = List.fold_left (construct_children n) []
                  (Abstraction.enabled_ops n_marking_data.region) in
	      
              Message.msg_string Message.Debug "The children are:" ;
              List.iter (function child -> Message.msg_printer Message.Debug print_tree_node child) children ;
	      
              (* add the children to the set of pending nodes *)
              Message.msg_string Message.Debug "Adding the children to the set of pending unprocessed#nodes" ;
              unprocessed_nodes#add_list children ;
	      
              (* mark n as processed and uncovered *)
              Message.msg_string Message.Debug "Updating the node's marking" ;
              n_data.mark <- Processed_Uncovered n_marking_data ;
	      
              Message.msg_string Message.Debug "This node now looks like:" ;
              Message.msg_printer Message.Debug print_tree_node n ;
	      
              (* add n's region to the reached region *)
              Message.msg_string Message.Debug "Updating the currently reached region" ;
              reached_region := Region.cup !reached_region n_marking_data.region
            end
	in	      

	(* This function checks if an error region reached in the thread is a real
	   error trace (given the current assumption about the other thread).
	   *)
	let error_reached n err_n n_marking_data =
	  begin
	    let n_data = (Tree.get_node_label n) in
            let n_region = n_marking_data.region in
                    (* this node will have to be reprocessed after refinment *)
		    (* JHALA: why will this node have to be reprocessed after refinement ? *) 
            Message.msg_string Message.Debug "Re-adding this node to the set of pending unprocessed nodes" ;
            n_data.mark <- Processed_Was_Covered_To_Reprocess n_marking_data ;
	    Message.msg_string Message.Debug "Adding to unprocessed#nodes";
	    Message.msg_printer Message.Debug print_tree_node n ;
            unprocessed_nodes#add_element n ;
	    
                (* let's now test whether this error is a real one *)
            Message.msg_string Message.Normal "Testing whether this (single threaded) error is a real one" ; 
	    Message.msg_string Message.Debug "Testing whether this (single threaded) error is a real one" ; 
	    let check_refinement = 
(*	      if (Options.getValueOfString "refine" = "fwd") then
		check_error_fwd 
	      else *) check_error
	    in
	    match (Stats.time "check_error" (check_refinement n) err_n) with
	      EmptyErrorAtNode(err_anc, anc, path) ->
				(* this is a false error *)
		begin
		  Message.msg_string Message.Normal "This is a false (single threaded) error" ;
		  Message.msg_string Message.Debug "The check_error function ended up with" ;
		  Message.msg_string Message.Debug "     at ancestor: " ;
		  Message.msg_printer Message.Debug print_tree_node anc ;
		  Message.msg_string Message.Debug "     along the path (from the ancestor to the current node):" ;
		  Message.msg_printer Message.Debug print_tree_node anc ;
		  List.iter (function x -> Message.msg_printer Message.Debug print_tree_node x) path ;
		  demo_process n n_region false;
		  
		  if (Options.getValueOfBool "stop") then
		    begin
		      dumppath n n_region;
		      failwith ("FALSE ERROR CHOKE");
		    end
				(* let's refine *)
		
		  Message.msg_string Message.Debug "\n\nLet's refine the nodes along that path" ;
		  let anc_region = (get_region anc) in
		  let focused_anc_region = 
		    try
		      let rv = Stats.time "focus" (Abstraction.focus anc_region) err_anc in
		      rv
		    with Abstraction.NoNewPredicatesException ->
		      begin
			Message.msg_string Message.Normal "No new predicates raised for single-threaded error." ;
			let path_to_root = Tree.path_to_root n in
			let list_of_regs = (List.map (get_region)  (Misc.chop_last path_to_root))@[n_region] in
			let list_of_ops = List.map (get_op) ( List.tl path_to_root) in
			dumptrace (false, list_of_ops, list_of_regs);
			failwith ("No new preds found !-- and not running find_all_preds ...")
		      end
		  in (* for the focused_anc_region *)
		  Message.msg_string Message.Debug "Done calling Focus";
		  Message.msg_string Message.Debug ("The focused region for the ancestor is: " ^ (Region.to_string focused_anc_region)) ;
		  
		  stats_nb_refinment_processes := !stats_nb_refinment_processes + 1 ;
		  Stats.time "subtree_cut_refinement" (do_subtree_cut_refinement anc) focused_anc_region;
		  end
	    | NonEmptyErrorAtRoot(err_root, root, path) ->
			    (* --- for debugging --- *)
		begin
		  assert (root == tree_root) ;
			    (* this is a real error *)
			    (* we put the error path in the appropriate variable *)

                              (* print the unknown functions on the path *)
		  Abstraction.print_debug_info (List.map get_op path);
		  
                  raise (RaceConditionException (Error_reached 
						  (err_root,
						   List.map get_op path,
						   err_n) ) )

		end
	  end
	in
	(* continue_reachability says if there was an error *)
	let continue_reachability = ref true in
	stats_nb_iterations := 0; 
        (* add the root to the pending unprocessed nodes *)
	unprocessed_nodes#add_element (tree_root) ;
	while (!continue_reachability && unprocessed_nodes#has_next ()) do
	  stats_nb_iterations := !stats_nb_iterations + 1;
	  Message.msg_string Message.Debug "\n\n************************************************************************\n" ;
          Message.msg_string Message.Debug "Next iteration of model-check's inner reachability loop" ;
	  Message.msg_printer Message.Debug Format.pp_print_int (!stats_nb_iterations);
          Message.msg_string Message.Debug "Currently reached region:" ;
          Message.msg_printer Message.Debug Region.print !reached_region ;
	  
        (* let's process the next pending node *)
          let n        = unprocessed_nodes#next () in
          Message.msg_string Message.Debug (*Normal*)  "Now processing tree node:" ;
          Message.msg_printer Message.Debug (*Normal*) print_tree_node n ;
	  
          let n_data   = (Tree.get_node_label n) in
	  
          let n_marking_data = 
            match n_data.mark with
              Unprocessed ->
                let op = (get_op n) in
                { time_stamp = (time := !time + 1 ; !time) ;
                  region = Abstraction.post (get_region (get_parent n)) op global_variable_table; }
            | Processed_Was_Covered_To_Reprocess md ->
                md
            | _ -> failwith "LazyModelChecker.model_check: unexpected marking data"
		  
          in
            (* --- for debugging --- *)
            (* assert (n_data.mark == Unprocessed) ; *)
	  
          let n_region = n_marking_data.region in
	  (** look for bugs **)
	  (* continue is set to false if there was a bug **)
	  begin
	    Message.msg_string Message.Debug "Checking for race conditions." ;
	    (** Helper functions *)

	    (* construct_enabled_region returns region /\ wp(action on edge, true) *)
	    let construct_enabled_region n_region edge =
	      if (Abstraction.isPredicate edge) then
		begin
		  let true_reg = Abstraction.create_concrete_true_region edge in (* region true *)
		  let wp = Abstraction.spre true_reg edge global_variable_table in
		  Message.msg_string Message.Debug ("wp created, it is "^(Region.to_string wp)) ;
		  Abstraction.intersect_with_abstraction (n_region) wp
		end
	      else
		n_region
	    in
	    try 
	    (* for each enabled successor check if there is a potential race condition *)
	      (* RUPAK: this is not the most efficient. need to optimize this. e.g., only
		 a few variables are touched by the edge, so need not iterate over all global vars *)
	      
	      (* JHALA: add other parameters to check_concurrent_access lval_now aliased_lval *)

	      (* JHALA: for each edge: 
		 1. find the lvals written to/ read from for that edge and update
		 their assumption tables appropriately... or do we do the updating AFTER the check ?
		 2. then do the checking i.e. for each lval that may be aliased
		 to something in the lvals_written \cup lvals_read *)
	      
	      let update_read_assumptions trimmed_reg lval = 
		Message.msg_string Message.Debug ("Updating the current read assumption on "^(Expression.lvalToString lval));
		try
		  let phi_list = Hashtbl.find accNodes lval in
		  Hashtbl.replace 
		    accNodes 
		    lval 
		    { read_whole = Region.cup phi_list.read_whole trimmed_reg ; 
		      read_accesses = (n, trimmed_reg) :: phi_list.read_accesses ;
		      write_whole = phi_list.write_whole ; 
		      write_accesses = phi_list.write_accesses 
		    }
		with Not_found -> (* if accNodes does not find lval, this means that there is no
				     write in the program through this lvalue. 
				     *)
		  begin
		    Message.msg_string Message.Debug ("This lvalue was not found in accNodes. Not updating read assumption.");
		    ()
		  end
	      in
	      let update_write_assumptions trimmed_reg lval =
		Message.msg_string Message.Debug ("Updating the current write assumption on "^(Expression.lvalToString lval));
		try
		  let phi_list = Hashtbl.find accNodes lval in
		  Hashtbl.replace 
		    accNodes 
		    lval
		    { read_whole = phi_list.read_whole ; 
		      read_accesses = phi_list.read_accesses ;
		      write_whole =  Region.cup phi_list.write_whole trimmed_reg; 
		      write_accesses = (n, trimmed_reg) ::phi_list.write_accesses 
		    }
		with Not_found ->
		  (* Something wrong! *)
		  failwith ("Lvalue being written to not found in accNodes! "^(Expression.lvalToString lval))
	      in
	      let strong_src_reg = n_region (* construct_enabled_region n_region edge *) in
	      let trimmed_reg = Abstraction.trim_region_to_globals strong_src_reg in
              let checking_access_and_updating edge race_kind lval =
		(* first find the list of all aliases of this lval *)
		let list_of_aliases = Abstraction.aliasesOf lval in
		let race_string = match race_kind with Read -> "read" | Write -> "write" in
		  Message.msg_string Message.Normal 
		    ("Checking for races on a " ^ race_string ^" to variable "^(Expression.lvalToString lval)) ;
		  Message.msg_string Message.Normal ("for the operation "^(Operation.to_string edge)) ;
		let checkfn l = 
		  let l_phi_list = 
		    try Hashtbl.find accNodes l 
		    with Not_found ->
		      (* RUPAK: check that this is correct.
			 If phi is not found for an lvalue l in the accNodes table, this
			 means that the lvalue l never occurs syntactically in the program.
			 Hence, this means that no address is ever updated in the program
			 through the lvalue l, nor is the lvalue l ever read.
			 So we can return Region.Bot for any such l
			 *)
		      begin
			Message.msg_string Message.Debug ("phi not found for "^(Expression.lvalToString l));
			{
			read_whole = Region.bot ; read_accesses = [] ;
			write_whole = Region.bot ; write_accesses = [] ;
		        }	
		      end
		  in 
		  Message.msg_string Message.Debug ("\n\nwith the possibly aliased "^(Expression.lvalToString l));
		  Message.msg_string Message.Debug "Now calling check_concurrent_access" ;
		  Message.msg_string Message.Debug "Phi list is : " ;
		  Message.msg_string Message.Debug ("Read = "^(Region.to_string l_phi_list.read_whole)) ;
		  Message.msg_string Message.Debug ("Write = "^(Region.to_string l_phi_list.write_whole)) ;
		  
		  check_concurrent_access race_kind edge n strong_src_reg lval l l_phi_list; 
		  Message.msg_string Message.Debug "Check concurrent access returned without error, start updating phi" 
		in
		let _  = List.iter checkfn list_of_aliases in
		  (* if we get this far then there are no races between lval and all his list_of_aliases *)
		  match race_kind with
		      Read -> update_read_assumptions trimmed_reg lval
		    | Write -> update_write_assumptions trimmed_reg lval
	      in
	      	List.iter 
		  (fun edge -> 
		     let (lvals_read, lvals_written) = Abstraction.reads_and_writes edge in
		       List.iter (checking_access_and_updating edge Read) lvals_read;
		       List.iter (checking_access_and_updating edge Write) lvals_written)
		  (Abstraction.enabled_ops n_region);

	    with Exit_from_check_concurrent_access a ->
	      begin
		match a with
		  No_error ->
	           (* found an infeasible race *)
		    Message.msg_string Message.Debug "Found an infeasible race. Setting continue to false." ;
		    continue_reachability := false 
		  | Race_condition _ -> raise (RaceConditionException a) 
		      (* send back error condition to caller *)
		  | _ -> failwith "model_check: Strange return value!"
	      end;
          end;
		    (** finished looking for bugs **)
	  (** do the following only if continue remains true **)
	  if (!continue_reachability = true) then
	    (* continue the reachability *)
	    begin
	      (* check for location errors *)
	      match o_err with
		None ->	 no_error n n_data n_marking_data (* no error to check: pure race detection *)
	      |	Some err ->
		  (* check for error location reachability *)
		  begin
		    let err_n = Stats.time "cap" (Region.cap n_region) err in
		    if (Region.is_empty err_n) then
		      no_error n n_data n_marking_data
		    else
		      begin
  		        (* error found. Check if real error *)
			Message.msg_string Message.Debug "Error found : checking validity." ;
			error_reached n err_n n_marking_data
		      end   
		  end
		    
	    end ;
	done; (* inner reachability loop *)

      (* outer loop: if continue is false, then there was an error, so reset environment assumptions loop back *)
      if (!continue_reachability = false) then
	begin
	  (* reset all environment assumptions to false *)
	  let _ = Hashtbl.iter (fun a b -> Hashtbl.replace global_variable_table a (ref Region.bot)) 
	                       global_variable_table in
	  Message.msg_string Message.Debug "Resetting all assumptions in the global_variable_table and continuing reachability" ;
	  (* note that refine already adds predicates *)
	  flag := true (*;
	  Abstraction.reset ()  do we need this? *)
	end
      else
	begin
	  let _checkEntailmentAndUpdate name currAppx =
	    let ax = !(get_global_var_phi name) in
	    let curr = (* current approximation to the environment assumption *)
	      currAppx.write_whole
	    in
	    if (not (Region.leq curr ax)) then
	      begin
	        (* update global environment assumptions *)
		let newreg = (Region.cup ax curr) in
		Message.msg_string Message.Debug ("Updating global invariant about "^(Expression.lvalToString name)) ;
		Message.msg_string Message.Debug ("The new environment assumption is "^(Region.to_string newreg)) ;
		Hashtbl.replace global_variable_table name (ref newreg) ;
		flag := true
	      end
	  in
	  (* check for each global if the current environment assumptions are subsumed *)
	  Hashtbl.iter (_checkEntailmentAndUpdate) accNodes
	end
      end
    done;
    (* send back "correct program" to caller *)
    No_error
    with RaceConditionException a -> 
      begin
(*
	match a with
	  Race_condition (p1, p2, _, _) ->
	    dumptrace p1 ;
	    dumptrace p2
	| Error_reached p1 -> dumptrace p1
	| _ -> failwith "Race condition exception raised with strange model check outcome" ;
*)
	a
      end
	
(*******************************************************************************)
(*******************************************************************************)
(*******************************************************************************)
(* vanilla *)

(*******************************************************************************)
(*******************************************************************************)
  (* Parallel model checker.
     Instead of assume guarantee checking, just check all interleavings.

     

     Need to synchronize this with the TAR algorithm.
     *)
  let model_check init o_err =
    Message.msg_string Message.Debug "Entering parallel_model_check" ;
    Message.msg_string Message.Debug ("   init region: " ^ (Region.to_string init)) ;
    Message.msg_string Message.Debug 
      (match o_err with None -> "No error region" | Some e -> "   error region: " ^ (Region.to_string e)) ;
    let err = 
      match o_err with
	None -> failwith "parallel_model_check: error region not specified."
      | Some e -> e
    in 
    (* the final result of the search will be stored here *)
    let resulting_error_path = ref None
    and offending_error_node = ref None
      
    (* this boolean variable allows to stop the while-loop *)
    and continue = ref true

    (* the reached region is initially empty *)
    and reached_region = ref Region.bot

    (* helper object to create nodes *)
    and tnc = new tree_node_creator

   
    (* the set of pending nodes to be processed *)
    and unprocessed_nodes = 
      if (Options.getValueOfBool "bfs") then new Search.bfsiterator
      else new Search.dfsiterator

    (* the global time used for time stamps *)
    and time = ref 0 in

    (* The function for constructing children .. and adding them to the tree *)
    let construct_children node accu op =
      (tnc#create_child Unprocessed op node)::accu
    in				 
						 
 (* the root of the reachability tree *)
    let tree_root =
      tnc#create_root (Processed_Was_Covered_To_Reprocess { time_stamp = !time ;
                                                            region = init ; })
    in
    let do_subtree_cut_refinement anc new_anc_region =
      let anc_data = Tree.get_node_label anc in
      let anc_marking_data = get_marking_data anc in
        (* update the ancestor region *)
      Message.msg_string Message.Debug "Updating the ancestor's region..." ;
      anc_marking_data.region <- new_anc_region ;
      
      Message.msg_string Message.Debug "This ancestor now looks like:" ;
      Message.msg_printer Message.Debug print_tree_node anc ;
        (* Note that here there is no path to update as the entire subtree is going to be chopped off... *)
      Message.msg_string Message.Debug "Now deleting subtree:" ;
      tnc#delete_children anc;
      unprocessed_nodes#empty_out ();
      (* Jhala: I'm assuming that its ok to just empty this out and fill it back up when we are traversing the tree later ... this whole traversal business is not good by the way, there should be a much more incremental way to do it *)
      Message.msg_string Message.Debug "Now re-adding children:" ;
      let children = List.fold_left (construct_children anc) [] 
	  (Abstraction.enabled_ops new_anc_region) in
      Message.msg_string Message.Debug "The children are:" ;
      List.iter (function child -> Message.msg_printer Message.Debug print_tree_node child) children ;

       (* add the children to the set of pending nodes *)
      (* Jhala: This should all be done by the the subtree cut 
	 Message.msg_string Message.Debug "Adding the children to the set of pending nodes" ;
	 unprocessed_nodes#add_list children ;
	 *)
      (* mark n as processed and uncovered *)
      Message.msg_string Message.Debug "Updating the node's marking" ;
      
      anc_data.mark <- Processed_Uncovered anc_marking_data ;
      Message.msg_string Message.Debug "This node now looks like:" ;
      Message.msg_printer Message.Debug print_tree_node anc;
      (* add n's region to the reached region *)
      Message.msg_string Message.Debug "Updating the currently reached region" ;
      reached_region := Stats.time "update_tree_after_refinement[CUT]" 
	   (update_tree_after_refinment tree_root unprocessed_nodes) anc_marking_data.time_stamp
    in


    (* the refinment process *)
    let do_refinment anc path new_anc_region =
      let anc_data = Tree.get_node_label anc in
      let anc_marking_data = get_marking_data anc in
        (* update the ancestor region *)
      Message.msg_string Message.Debug "Updating the ancestor's region..." ;
      anc_marking_data.region <- new_anc_region ;
      
      Message.msg_string Message.Debug "This ancestor now looks like:" ;
      Message.msg_printer Message.Debug print_tree_node anc ;
      
        (* update the path *)
      Message.msg_string Message.Debug "Refining the nodes along the path..." ;
      refine_path anc path ;
      
        (* update the tree and the reached region *)
      Message.msg_string Message.Debug "Updating the markings and the reached region..." ;
      reached_region := Stats.time "update after refine [PATH]" (update_tree_after_refinment tree_root
								   unprocessed_nodes) anc_marking_data.time_stamp

    in

    (* this function deletes covered nodes from the tree *)
    let rec reclaim n = 
      if (Tree.has_child n || (not (Tree.has_parent n))) then ()
      else 
	begin
	  let par = get_parent n in
	  Tree.delete_child n; 
	  reclaim par
	end
    in
      (* add the root to the pending unprocessed nodes *)
      unprocessed_nodes#add_element (tree_root) ;

      while (!continue && unprocessed_nodes#has_next ()) do
        stats_nb_iterations := !stats_nb_iterations + 1;

        (* the search is not finished yet *)
	if (!stats_nb_iterations mod 100 == 0) then
	  begin
            Message.msg_string Message.Normal ("While loop tick:"^(string_of_int !stats_nb_iterations));
	    flush stdout; flush stderr;
	    Abstraction.print_abs_stats ()
	  end ;
	Message.msg_string Message.Debug "\n\n****************************************************************************\n" ;
        Message.msg_string Message.Debug "Next iteration of model-check's big while-loop" ;
	Message.msg_printer Message.Debug Format.pp_print_int (!stats_nb_iterations);
        (* Message.msg_string Message.Debug "Currently reached region:" ;
           Message.msg_printer Message.Debug Region.print !reached_region ;*)

        (* let's process the next pending node *)
        let n        = unprocessed_nodes#next () in
          Message.msg_string Message.Debug (*Debug Normal*) "Now processing tree node:" ;
          Message.msg_printer Message.Debug (*Debug Normal*) print_tree_node n ;

          let n_data   = (Tree.get_node_label n) in

          let n_marking_data = 
            match n_data.mark with
                Unprocessed ->
                  let op = (get_op n) in
                    { time_stamp = (time := !time + 1 ; !time) ;
                      region = Stats.time "main post" (Abstraction.post (get_region (get_parent n)) op)  global_variable_table; }
              | Processed_Was_Covered_To_Reprocess md ->
                  md
              | _ -> failwith "LazyModelChecker.model_check: unexpected marking data"

          in
            (* --- for debugging --- *)
            (* assert (n_data.mark == Unprocessed) ; *)

          let n_region = n_marking_data.region in

          (* the error region at this node *)
          let err_n = Stats.time "cap" (Region.cap n_region) err in
            Message.msg_string Message.Debug ("Error region at this node: " ^ (Region.to_string err_n)) ;

            if (Region.is_empty err_n)
            then
              begin
                (* no error found at this node *)
                Message.msg_string Message.Debug "No error found at this node" ;

                (* let's test whether this node is covered *)
                Message.msg_string Message.Debug "Let's test whether this node is covered" ;
		let coverable =  
		  if (Options.getValueOfBool "cov") 
		  then Abstraction.coverable n_region
		  else true
		in
		let covered = 
		  if coverable
		  then (Stats.time "Covered check:" (Abstraction.covered n_region) !reached_region)
		  else false
		in
		if (covered)
                then
                  begin
                    (* this node is covered *)
		    Message.msg_string Message.Debug "This node is covered" ;
			
                    (* mark n as processed and covered *)
		    n_data.mark <- Processed_Covered n_marking_data;
		    if (Options.getValueOfBool "reclaim") 
		    then
		      reclaim n;
		    ()
                  end
		else
		  begin
                    (* this node is not covered *)
                    Message.msg_string Message.Debug "This node is not covered" ;
                    Message.msg_string Message.Debug "Constructing its successor children..." ;
		    
                    (* we construct its successor children *)
		    
                    let children = List.fold_left (construct_children n) []
                        (Abstraction.enabled_ops n_region) 
		    in
		    
                    Message.msg_string Message.Debug (*Normal*) "The children are:" ;
                    List.iter (function child -> Message.msg_printer Message.Debug (*Normal*) print_tree_node child) children ;
		    
                      (* add the children to the set of pending nodes *)
                    Message.msg_string Message.Debug "Adding the children to the set of pending unprocessed#nodes" ;
                    unprocessed_nodes#add_list children ;
		    Message.msg_string Message.Debug ("Remaining nodes:"^(string_of_int (unprocessed_nodes#sizeof ()))); 
		    unprocessed_nodes#iter (fun x -> Message.msg_printer Message.Debug print_tree_node x);
                      (* mark n as processed and uncovered *)
                    Message.msg_string Message.Debug "Updating the node's marking" ;
                    n_data.mark <- Processed_Uncovered n_marking_data ;
		    
                    Message.msg_string Message.Debug "This node now looks like:" ;
                    Message.msg_printer Message.Debug print_tree_node n ;
	            
                      (* add n's region to the reached region *)
		    if (coverable) then 
		      begin
			Message.msg_string Message.Debug "Updating the currently reached region" ;
			reached_region := Region.cup !reached_region n_region
		      end
		  end
              end
            else
              begin
                (* error found *)
                Message.msg_string Message.Debug "Error found : checking validity." ;
		Message.msg_string Message.Debug ("Error at depth"^(string_of_int (Tree.node_depth n))); 
	(*	List.iter (function x -> Message.msg_printer Message.Debug print_tree_node x) (Tree.path_to_root n) ; *)
		if (Options.getValueOfBool "cref") 
                then 
		  begin
		    let path_to_root = Tree.path_to_root n in
		    let list_of_regs = (List.map (get_region)  (Misc.chop_last path_to_root))@[n_region] in
		    let list_of_ops = List.map (get_op) ( List.tl path_to_root) in
		    (* note that list_of_regs should be 1 longer than list of ops *)
		    let _ = 
		      try 
			Message.msg_string Message.Debug ("Found error length: "^(string_of_int (List.length list_of_regs)));
			Abstraction.analyze_trace list_of_regs list_of_ops
		      with fail_exception ->  
			begin
			  List.iter (function x -> Message.msg_printer Message.Debug print_tree_node x)
			    path_to_root;
			  print_string "Uh Oh! Error in analyze_trace!";
			  raise fail_exception
                        end 
		    in
		    stats_nb_refinment_processes := !stats_nb_refinment_processes + 1;
		    do_subtree_cut_refinement tree_root (get_region tree_root)
		      
                  end
                else 
                  begin
                    Message.msg_string Message.Debug "Error found at this node" ;
		    
                    (* this node will have to be reprocessed after refinment *)
		    (* JHALA: why will this node have to be reprocessed after refinement ? *) 
                    Message.msg_string Message.Debug "Re-adding this node to the set of pending unprocessed nodes" ;
                    n_data.mark <- Processed_Was_Covered_To_Reprocess n_marking_data ;
		    Message.msg_string Message.Debug "Adding to unprocessed#nodes";
		    Message.msg_printer Message.Debug print_tree_node n ;
                    unprocessed_nodes#add_element n ;
		    
                (* let's now test whether this error is a real one *)
                    Message.msg_string Message.Debug "Testing whether this error is a real one" ; 
		    let check_refinement = 
(*
		      if (Options.getValueOfString "refine" = "fwd") then
			check_error_fwd 
		      else *)
			check_error
		    in
		      try 
			match (Stats.time "check_error" (check_refinement n) err_n) with
			    EmptyErrorAtNode(err_anc, anc, path) ->
				(* this is a false error *)
			      begin
				Message.msg_string Message.Debug "This is a false error (seq)" ;
				Message.msg_string Message.Debug "The check_error function ended up with" ;
				Message.msg_string Message.Debug "     at ancestor: " ;
				Message.msg_printer Message.Debug print_tree_node anc ;
				Message.msg_string Message.Debug "     along the path (from the ancestor to the current node):" ;
				Message.msg_printer Message.Debug print_tree_node anc ;
				List.iter (function x -> Message.msg_printer Message.Debug print_tree_node x) path ;
				demo_process n n_region false;
				
				if (Options.getValueOfBool "stop") then
				  failwith ("FALSE ERROR CHOKE");
				(* let's refine *)
				Message.msg_string Message.Debug "\n\nLet's refine the nodes along that path" ;
				if (Options.getValueOfBool "traces") then dumppath n n_region;
				
			      	let anc_region = (get_region anc) in
				let focused_anc_region = 
				  try
				    let rv = 
				      match Options.getValueOfBool "block" with
					  false ->
					    Stats.time "focus" (Abstraction.focus anc_region) err_anc 
					| true -> Abstraction.focus_stamp anc_region 
					    (* analysis is already done! *) 
				    in
				      if (Options.getValueOfBool "localrestart") then
					local_restart_counter := !local_restart_counter + 1;
				      if (Options.getValueOfBool "restart" || ( Options.getValueOfBool "localrestart" && !local_restart_counter mod 15 == 0)) then 
					begin
					  Message.msg_string Message.Error "About to Restart";
					  					  
					  raise RestartException
					    
					end
					else 
					  begin
					    Options.setValueOfBool "localrestart" false;
					    rv
					  end
				  with Abstraction.NoNewPredicatesException ->
				    begin
 				      if (Options.getValueOfInt "predH" = 3) then
					begin
					  let path_to_root = Tree.path_to_root n in
					  let list_of_regs = (List.map (get_region)  (Misc.chop_last path_to_root))@[n_region] in
					  let list_of_ops = List.map (get_op) ( List.tl path_to_root) in
					  Message.msg_string Message.Debug 
					    ("\n No of regs: "^(string_of_int (List.length list_of_regs)));
					  Message.msg_string Message.Debug 
					    ("\n No of ops : "^(string_of_int (List.length list_of_ops)));
					  try 
					    (Abstraction.find_all_predicates list_of_regs list_of_ops; 
					     raise RestartException
					       )
					  with e -> 
					    
					    begin
					      if (e = RestartException) then raise e 
					      else 
						begin
						  Message.msg_string Message.Minor "This is a false error" ;
						  Message.msg_string Message.Minor "The check_error function ended up with" ;
						  Message.msg_string Message.Minor ("     an empty error: " ^ (Region.to_string err_anc)) ;
						  Message.msg_string Message.Minor "     at ancestor: " ;
						  Message.msg_printer Message.Minor print_tree_node anc ;
						  Message.msg_string Message.Minor "     along the path (from the ancestor to the current node):" ;
						  Message.msg_printer Message.Minor  print_tree_node anc ;
						  List.iter (function x -> Message.msg_printer Message.Minor print_tree_node x) path ;
						  Message.msg_string Message.Debug ("find_all_predicates raises : "^(Printexc.to_string e)) ; 
						  dumptrace (false, list_of_ops, list_of_regs);
						  failwith ("problem in find_all_predicates !") 
						end
					    end
					end
				      else 
					if (Options.getValueOfInt "predH" = 1 && 
					    (not (Abstraction.predicate_fixpoint_reached ()))) 
					then (Message.msg_string Message.Debug "Raising Restart"; raise RestartException)
					else
					  begin
					    let path_to_root = Tree.path_to_root n in
					    let list_of_regs = (List.map (get_region)  (Misc.chop_last path_to_root))@[n_region] in
					    let list_of_ops = List.map (get_op) ( List.tl path_to_root) in
					    dumptrace (false, list_of_ops, list_of_regs);
					      failwith ("No new preds found !-- and not running allPreds ...")
					  end
				    end
				in (* for the focused_anc_region *)
				Message.msg_string Message.Debug "Done calling Focus";
				Message.msg_string Message.Debug ("The focused region for the ancestor is: " ^ (Region.to_string focused_anc_region)) ;
				
				stats_nb_refinment_processes := !stats_nb_refinment_processes + 1 ;
				begin
				match (Options.getValueOfString "comp") with
				    "path" -> 
				      begin
					Stats.time "do_ref [PATH]" (do_refinment anc path) focused_anc_region ;
					Message.msg_string Message.Error "PATH"; 
					Message.msg_string Message.Debug "The refined path looks like:" ;
					Message.msg_printer Message.Debug print_tree_node anc ;
					List.iter (function x -> Message.msg_printer Message.Debug print_tree_node x) path ;
				      end
				  | _ -> Stats.time "subtree_cut_refinement" (do_subtree_cut_refinement anc) focused_anc_region;
			      end
 end
		| NonEmptyErrorAtRoot(err_root, root, path) ->
			    (* --- for debugging --- *)
			    begin
			    assert (root == tree_root) ;
			    (* this is a real error *)
			    (* we put the error path in the appropriate variable *)
			
                              resulting_error_path := Some(err_root,
							   List.map get_op path,
							   err_n) ;
			      offending_error_node := Some (Misc.list_last path) ;
                              (* print the unknown functions on the path *)
			      Abstraction.print_debug_info (List.map get_op path);
			      demo_process n n_region true;
			      
                              (* and break the while-loop *)
                              continue := false
			    end
		      with RestartException -> 
			begin
			  stats_nb_refinment_processes := !stats_nb_refinment_processes + 1;
			  do_subtree_cut_refinement tree_root (get_region tree_root)
			end
		  end
	      end
      done ;
    Message.msg_string Message.Debug ("Depth of tree: "^(string_of_int (Tree.subtree_depth tree_root)));
    print_whole_tree tree_root;
    match (!resulting_error_path) with 
      Some ((_,_,err_n) as e_path) -> 
	begin
	  let err_node = 
	    match !offending_error_node with
		    Some n -> n
	    | None -> failwith "Quit while w/ error loop w/o setting offending_error_node!"
	  in
	  let path_to_root = Tree.path_to_root err_node in
	  let list_of_regs = (List.map (get_region)  (Misc.chop_last path_to_root))@[get_region err_node] in
	  let list_of_ops = List.map (get_op) ( List.tl path_to_root) in
	  dumptrace (true,list_of_ops,list_of_regs);
	  Error_reached e_path
	    
	end
    | None -> (* print the tree *) 
	begin
	      if (Options.getValueOfBool "pf") 
	      then 
		(Stats.time "Generate Proof:" (build_proof_tree ()) tree_root) 
	      else (count_tree_nodes tree_root);
	  (* !resulting_error_path *)
	      No_error
	end
	 

end 
