(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

module Predicate = Ast.Predicate
module Expression = Ast.Expression
module C_System_Descr = BlastCSystemDescr.C_System_Descr
module C_Command = BlastCSystemDescr.C_Command

module IntSet = Set.Make (struct type t = int let compare a b = b - a end )

(* [Greg] Very dirty hack because we renamed Stats to VampyreStats in
   vampyre...  We should clean this stats business anyway. *)
module Stats = VampyreStats

open BlastArch.ModelCheckOutcome

(*****************************************************************************************)

let inputCustomAbs () =
  let read_abs_info filename =
    let doParse inchan =
      let _ = VampyreErrormsg.startFile filename in
      let lexbuf = Lexing.from_channel inchan in
      let _ = VampyreErrormsg.theLexbuf := lexbuf in
      Absparse.main Absflex.token lexbuf 
    in
    ignore (Message.msg_string Message.Normal "Reading in abstraction ...");
    try
      let inchan = open_in filename
      in
      let a = doParse inchan
      in
      ignore (Message.msg_string Message.Normal " Abstraction read.\n") ;
      a
    with
      Sys_error _ -> (Message.msg_string Message.Error ("Cannot find abstraction info in "^filename^".\n"); ([],None,[]))
    | VampyreErrormsg.Error -> (Message.msg_string Message.Error ("Error raised in reading abstraction \n"); ([],None,[]))
  in
  let abs_file = Options.getValueOfString "custom" in
  if (abs_file = "") then
    ([],None,[]) (* no special types nothing at all ... *)
  else
    read_abs_info abs_file

(*****************************************************************************************)


let constructInitialPredicates predFile =
  let readAndAddPredicates filename =
    let doParse inchan =
      let _ = VampyreErrormsg.startFile filename in
      let lexbuf = Lexing.from_channel inchan in
      let _ = VampyreErrormsg.theLexbuf := lexbuf in
      Inputparse.main Inputflex.token lexbuf 
    in
    ignore (Message.msg_string Message.Normal "Reading in seed predicates...");
    try
      let inchan = open_in filename
      in
      let a = doParse inchan
      in
	ignore (Message.msg_string Message.Normal "Seed predicates read.\n") ;	
	List.iter (fun a -> Message.msg_string Message.Normal (Predicate.toString a)) a ;
	Message.msg_string Message.Normal ("Read "^(string_of_int (List.length a))^" predicates");
	a
    with
      Sys_error _ -> (Message.msg_string Message.Error ("Cannot find predicate information in "^filename^".\n"); [])
    | VampyreErrormsg.Error -> (Message.msg_string Message.Error ("Error raised in reading seed predicates.\n"); [])
  in
  if (predFile = "") then
    (* Nothing to do *)
    []
  else
    readAndAddPredicates predFile

let constructInvariants () =
  let readAndAddInvariants filename =
    let doParse inchan =
      let _ = VampyreErrormsg.startFile filename in
      let lexbuf = Lexing.from_channel inchan in
      let _ = VampyreErrormsg.theLexbuf := lexbuf in
      Inputparse.main Inputflex.token lexbuf 
    in
    ignore (Message.msg_string Message.Normal "Reading in invariants...");
    try
      let inchan = open_in filename
      in
      let a = doParse inchan
      in
      ignore (Message.msg_string Message.Normal "Invariants read.\n") ;
      List.iter (fun a -> Message.msg_string Message.Normal (Predicate.toString a)) a ;
       a
    with
      Sys_error _ -> (Message.msg_string Message.Error ("Cannot find invariant information in "^filename^".\n"); [])
    | VampyreErrormsg.Error -> (Message.msg_string Message.Error ("Error raised in reading invariants.\n"); [])
  in
  let invFile = Options.getValueOfString "inv" in
  if (invFile = "") then
    (* Nothing to do *)
    []
  else
    readAndAddInvariants invFile

let constructEnvironmentAssumptions () = 
   let readAndAddAssumptions filename =
    let doParse inchan =
      let _ = VampyreErrormsg.startFile filename in
      let lexbuf = Lexing.from_channel inchan in
      let _ = VampyreErrormsg.theLexbuf := lexbuf in
      Inputparse.env_assumptions Inputflex.token lexbuf 
    in
    ignore (Message.msg_string Message.Normal "Reading in environment assumptions ...");
    try
      let inchan = open_in filename
      in
      let a = doParse inchan 
      in
      ignore (Message.msg_string Message.Normal "Assumptions read.\n") ;
      List.iter (fun (l,p) -> Message.msg_string Message.Normal ((Expression.lvalToString l )^" : "^(Predicate.toString p))) a ;
       a
    with
      Sys_error _ -> (Message.msg_string Message.Error ("Cannot find assumption information in "^filename^".\n"); [])
    | VampyreErrormsg.Error -> (Message.msg_string Message.Error ("Error raised in reading assumptions.\n"); [])
  in
  let asmFile = Options.getValueOfString "easm" in
  if (asmFile = "") then
    (* Nothing to do *)
    []
  else
    readAndAddAssumptions asmFile


(* JHALA: reading in kill predicates *)
let constructKillPredicates () =
  let readAndAddPredicates_kill filename =
    let doParse inchan =
      let _ = VampyreErrormsg.startFile filename in
      let lexbuf = Lexing.from_channel inchan in
      let _ = VampyreErrormsg.theLexbuf := lexbuf in
      Inputparse.main Inputflex.token lexbuf 
    in
    ignore (Message.msg_string Message.Normal "Reading in kill predicates...");
    try
      let inchan = open_in filename
      in
      let a = doParse inchan
      in
      ignore (Message.msg_string Message.Normal "kill predicates read.\n") ;
      List.iter (fun a -> Message.msg_string Message.Normal (Predicate.toString a)) a ;
       a
    with
      Sys_error _ -> (Message.msg_string Message.Error ("Cannot find kill information in "^filename^"\n"); [])
    | VampyreErrormsg.Error -> (Message.msg_string Message.Error ("Error raised in reading kill predicates\n"); [])
  in
  let killFile = Options.getValueOfString "killFile" in
  if (killFile = "") then
    (* Nothing to do *)
    []
  else
    readAndAddPredicates_kill killFile

let modelCheck o_err_loc =
  (* Get the abstraction of this C system description *)
  let module This_C_Abstraction = 
      Abstraction.Make_C_Abstraction(C_System_Descr) 
  in
  let module This_Cf_C_Abstraction = 
    CfAbstraction.Make_C_Abstraction(C_System_Descr) 
  in
  (* Create the lazy model checker for this abstraction *)
  let module This_LMC = LazyModelChecker.Make_Lazy_Model_Checker(This_C_Abstraction) in
  let module This_LMC_CF = CfLazyModelChecker.Make_Lazy_Model_Checker(This_Cf_C_Abstraction) in 
  let module Region = This_C_Abstraction.Region in
  This_C_Abstraction.check_races := Options.getValueOfBool "checkRace";

  This_C_Abstraction.assumed_invariants := Stats.time "read invariants" constructInvariants();  
  This_C_Abstraction.assumed_invariants_region := This_C_Abstraction.assumed_invariants_list_to_region();
  
(*
     ( let data = Region.Concrete { Region.pred = Predicate.conjoinL !assumed_invariants ; } in
       Region.Atomic { Region.location = C_SD.fake_location ;
		       Region.data = data ;
		       Region.stack = Region.EveryStack ; } );
*)
  (* Parsing in and adding the environment assumptions *)
  
  let _ = This_LMC.process_env_assumptions (constructEnvironmentAssumptions ()) in

  (* Add seed predicates, if any to system *)
  let seedPredsFile = Options.getValueOfString "pred" in
  let seedPreds = Predicate.conjoinL (Stats.time "read seeds" constructInitialPredicates seedPredsFile)
  in
  let initStateFile = Options.getValueOfString "init" in
  let initialState = Predicate.conjoinL (Stats.time "read initial state" constructInitialPredicates initStateFile) in
  let mainFunctions = Options.getValueOfStringList "main" in
    (* JHALA: Kill hacks *)
  let killLabel = Options.getValueOfString "killLoc" in
  let kill_location =
    match (C_System_Descr.get_locations_at_label killLabel) with
        [l] -> Some(l)
      | []  ->
          if (killLabel = "") then ()
          else
            Message.msg_string Message.Normal ("Kill label '" ^ killLabel ^
                                            "' was not found in the source file(s)") ;
          None
      | _   ->
          Message.msg_string Message.Error ("Kill label '" ^ killLabel ^
                                            "' appears multiple times in the source file(s)") ;
          exit 1
  in  
  let killPreds = constructKillPredicates () in
  let _ = This_C_Abstraction.initKill kill_location killPreds in
  let _ = This_C_Abstraction.initialize_abstraction () in
  (* JHALA: end kill hacks *)
  (* JHALA: custom abstraction hacks 
  let _ = if not (Options.getValueOfString "custom" = "") then
    This_C_Abstraction.build_custom_abstraction (inputCustomAbs ()) in *)
  (* JHALA: end custom abstraction hacks *)

  let _ = Message.msg_string Message.Error "Done adding seed predicates" in
  if (Options.getValueOfInt "predH" > 0) then 
    This_C_Abstraction.make_const_table () ;

    Message.msg_string Message.Normal "\n\n\n********** Now running the model-checker **********\n\n" ;
    flush stdout;
    begin
            (* [Greg] This is really a first draft...t
               Put this in a printer function, with formatting boxes, etc. *)
      let indent_loc = "                                                " in
      let print_op indent_op op =
        let target_loc = C_System_Descr.get_target op
        and c_command = C_System_Descr.get_command op
        in
        Message.msg_string Message.Normal
          (indent_op ^ (C_Command.to_string c_command)) ;
        Message.msg_string Message.Normal
          (indent_loc ^ (C_System_Descr.location_to_string target_loc)) ;
        match c_command.C_Command.code with
          C_Command.FunctionCall _ ->
            indent_op ^ "   "
        | C_Command.Block l ->
            begin
              match (List.hd (List.rev l)) with
              | C_Command.Return _ ->
                  String.sub indent_op 0 ((String.length indent_op) - 3)
              | _ -> indent_op
            end
        | _ -> indent_op
      in
      let mc_return_val = 
	try
	  if (Options.getValueOfBool "cf")
	  then
	    begin
	      let init_region =
		let rec buildloclist lst loclist =
		  match lst with [] -> loclist
		  | a ::rest -> 
		      try
			buildloclist rest ((C_System_Descr.lookup_entry_location a) :: loclist) 
		      with 
			Not_found -> 
			  (Format.printf "Creating init_region: No function %s defined in program. Exiting\n" (a);
			   failwith ("No function "^(a)^" defined"))
	in
		let loclist = buildloclist mainFunctions [] in
		This_Cf_C_Abstraction.create_region 
		  loclist false seedPreds initialState
		  
	      and error_region =
		match o_err_loc with
		| Some err_loc ->
		    if (Options.getValueOfBool "par") then
		      begin
	    (* first get the number of threads *)
			let num_of_threads = List.length (Options.getValueOfStringList "main") in
			Some (This_Cf_C_Abstraction.create_error_region err_loc num_of_threads Predicate.True) 
		      end
		    else
		      Some (This_Cf_C_Abstraction.create_region [err_loc] true Predicate.True Predicate.True)
		| None -> None
		
	      in let mc_return_val = This_LMC_CF.model_check init_region error_region
	      in match (mc_return_val) with
		No_error ->
		  Message.msg_string Message.Normal "\n\nNo error found.  The system is safe :-)\n"
	      | Error_reached (reg_i, op_list, reg_e) ->
		  Message.msg_string Message.Normal "\n\nError found! The system is unsafe :-(\n\n Error trace:\n" ;
		  Message.msg_printer Message.Normal This_Cf_C_Abstraction.Region.print reg_i ;
		  Message.msg_string Message.Normal "..\n\n" ;
	  
		  Message.msg_string Message.Normal
		    (indent_loc ^ (C_System_Descr.location_to_string
				     (C_System_Descr.get_source (List.hd op_list)))) ;
		  let _ = List.fold_left print_op "" op_list in
		  Message.msg_string Message.Normal "\n\n" ;
		  Message.msg_printer Message.Normal This_Cf_C_Abstraction.Region.print reg_e;
		  if (Options.getValueOfBool "xml") then This_Cf_C_Abstraction.dumpXml op_list
	      | Race_condition ((reg1,oplist1, reg1'), (reg2, oplist2, reg2'), lv1, lv2) ->
		  Message.msg_string Message.Normal "\n\nRace condition found! The system is unsafe :-(\n\n\n";
		  Message.msg_string Message.Normal (Printf.sprintf "\n On the lvalues %s, %s \n" (Ast.Expression.lvalToString lv1) (Ast.Expression.lvalToString lv2));
	          (* print out the first trace *)
		  Message.msg_string Message.Normal "Printing out the first trace.\n" ;
		  Message.msg_printer Message.Normal This_Cf_C_Abstraction.Region.print reg1 ;
		  Message.msg_string Message.Normal "\n\n" ;
	  
		  Message.msg_string Message.Normal
		    (indent_loc ^ (C_System_Descr.location_to_string
				     (C_System_Descr.get_source (List.hd oplist1)))) ;
		  let _ = List.fold_left print_op "" oplist1 in
		  Message.msg_string Message.Normal "\n\n" ;
		  Message.msg_printer Message.Normal This_Cf_C_Abstraction.Region.print reg1';
		  if (Options.getValueOfBool "xml") then This_Cf_C_Abstraction.dumpXml oplist1 ;
	          (* print out the second trace *)
		  Message.msg_string Message.Normal "\n\nPrinting out the second trace.\n" ;
		  Message.msg_printer Message.Normal This_Cf_C_Abstraction.Region.print reg2 ;
		  Message.msg_string Message.Normal "\n\n" ;
	  
		  Message.msg_string Message.Normal
		    (indent_loc ^ (C_System_Descr.location_to_string
				     (C_System_Descr.get_source (List.hd oplist2)))) ;
		  let _ = List.fold_left print_op "" oplist2 in
		  Message.msg_string Message.Normal "\n\n" ;
		  Message.msg_printer Message.Normal This_Cf_C_Abstraction.Region.print reg2';
		  if (Options.getValueOfBool "xml") then This_Cf_C_Abstraction.dumpXml oplist2 
	    end
	  else
	    begin
	      let init_region =
		let rec buildloclist lst loclist =
		  match lst with [] -> loclist
		  | a ::rest -> 
		      try
			buildloclist rest ((C_System_Descr.lookup_entry_location a) :: loclist) 
		      with 
			Not_found -> 
			  (Format.printf "Creating init_region: No function %s defined in program. Exiting\n" (a);
			   failwith ("No function "^(a)^" defined"))
		in
		let loclist = buildloclist mainFunctions [] in
		This_C_Abstraction.create_region 
		  loclist false seedPreds initialState
		  
	      and error_region =
		match o_err_loc with
		| Some err_loc ->
		    if (Options.getValueOfBool "par") then
		      begin
	    (* first get the number of threads *)
			let num_of_threads = List.length (Options.getValueOfStringList "main") in
			Some (This_C_Abstraction.create_error_region err_loc num_of_threads Predicate.True) 
		      end
		    else
		      Some (This_C_Abstraction.create_region [err_loc] true Predicate.True Predicate.True)
		| None -> None

	      in
	      let mc_return_val =
		if (Options.getValueOfBool "checkRace") 
		then (This_LMC.model_check_tar init_region error_region) 
		else (This_LMC.model_check init_region error_region)
	      in match (mc_return_val) with
		No_error ->
		  Message.msg_string Message.Normal "\n\nNo error found.  The system is safe :-)\n"
	      | Error_reached (reg_i, op_list, reg_e) ->
		  Message.msg_string Message.Normal "\n\nError found! The system is unsafe :-(\n\n Error trace:\n" ;
		  (* this next line raises a seg fault for some reason...
		     Message.msg_printer Message.Normal This_C_Abstraction.Region.print reg_i ;
		     Message.msg_string Message.Normal "\n\n" ;
		  *)
		  Message.msg_string Message.Normal
		    (indent_loc ^ (C_System_Descr.location_to_string
				     (C_System_Descr.get_source (List.hd op_list)))) ;
		  let _ = List.fold_left print_op "" op_list in
		  Message.msg_string Message.Normal "\n\n" ;
		  Message.msg_printer Message.Normal This_C_Abstraction.Region.print reg_e;
		  if (Options.getValueOfBool "xml") then This_C_Abstraction.dumpXml op_list 
	      | Race_condition ((reg1,oplist1, reg1'), (reg2, oplist2, reg2'), lv1, lv2) ->
		  Message.msg_string Message.Normal "\n\nRace condition found! The system is unsafe :-(\n\n\n";
		  Message.msg_string Message.Normal (Printf.sprintf "\n On the lvalues %s, %s \n" (Ast.Expression.lvalToString lv1) (Ast.Expression.lvalToString lv2));
	          (* print out the first trace *)
		  Message.msg_string Message.Normal "Printing out the first trace.\n" ;
		  Message.msg_printer Message.Normal This_C_Abstraction.Region.print reg1 ;
		  Message.msg_string Message.Normal "\n\n" ;
	  
		  Message.msg_string Message.Normal
		    (indent_loc ^ (C_System_Descr.location_to_string
				     (C_System_Descr.get_source (List.hd oplist1)))) ;
		  let _ = List.fold_left print_op "" oplist1 in
		  Message.msg_string Message.Normal "\n\n" ;
		  Message.msg_printer Message.Normal This_C_Abstraction.Region.print reg1';
		  if (Options.getValueOfBool "xml") then This_C_Abstraction.dumpXml oplist1 ;
	          (* print out the second trace *)
		  Message.msg_string Message.Normal "\n\nPrinting out the second trace.\n" ;
		  Message.msg_printer Message.Normal This_C_Abstraction.Region.print reg2 ;
		  Message.msg_string Message.Normal "\n\n" ;
	  
		  Message.msg_string Message.Normal
		    (indent_loc ^ (C_System_Descr.location_to_string
				     (C_System_Descr.get_source (List.hd oplist2)))) ;
		  let _ = List.fold_left print_op "" oplist2 in
		  Message.msg_string Message.Normal "\n\n" ;
		  Message.msg_printer Message.Normal This_C_Abstraction.Region.print reg2';
		  if (Options.getValueOfBool "xml") then This_C_Abstraction.dumpXml oplist2 
	    end
	with e -> 
	  begin 
	    This_C_Abstraction.dump_abs (); 
	    Message.msg_string Message.Error ("Exception raised :("^(Printexc.to_string e));
	    raise e
	  end
      in 
      ()
    end ;
  Message.msg_string Message.Normal "\n\nPhi Regions:\n" ;
  Message.msg_string Message.Normal (This_LMC.print_phis false) ;
  Message.msg_string Message.Normal "\n\nModel checker stats:\n" ;
  Message.msg_printer Message.Normal This_LMC.print_stats () ;
  Message.msg_string Message.Normal "\n\nAbstractor stats:\n" ;
  Message.msg_printer Message.Normal This_C_Abstraction.print_stats () 

let testModelCheck () =
  (* Get the abstraction of this C system description *)
  let module This_C_Abstraction = 
      Abstraction.Make_C_Abstraction(C_System_Descr) 
  in
  (* Create the lazy model checker for this abstraction *)
  let module This_LMC = LazyModelChecker.Make_Lazy_Model_Checker(This_C_Abstraction) in
  let module Region = This_C_Abstraction.Region in

(*
  This_C_Abstraction.assumed_invariants := Stats.time "read invariants" constructInvariants();  
  This_C_Abstraction.assumed_invariants_region := This_C_Abstraction.assumed_invariants_list_to_region();
 *) 

  (* Add seed predicates, if any to system *)
  let seedPredsFile = Options.getValueOfString "pred" in
  let seedPreds = Predicate.conjoinL (Stats.time "read seeds" constructInitialPredicates seedPredsFile)
  in
  let initStateFile = Options.getValueOfString "init" in
  let initialState = Predicate.conjoinL (Stats.time "read initial state" constructInitialPredicates initStateFile) in
  let mainFunctions = Options.getValueOfStringList "main" in
  let _ = This_C_Abstraction.initialize_abstraction () in
  let _ = Message.msg_string Message.Error "Done adding seed predicates" in
  if (Options.getValueOfInt "predH" > 0) then 
    This_C_Abstraction.make_const_table () ;

  Message.msg_string Message.Normal "\n\n\n********** Now running the tester **********\n\n" ;
  flush stdout;
  begin
    (* compute callgraph *)
    C_System_Descr.compute_callgraph ();
    
    (* create an array of priority -> location option, and a map location -> priority *)
    let sinfo_map = Hashtbl.create 511 in
    let get_sinfo loc =
      try
	Hashtbl.find sinfo_map (C_System_Descr.location_coords loc)
      with
	_ -> Hashtbl.add sinfo_map (C_System_Descr.location_coords loc) (-1) ; -1
    in
    (* depth-first number the locations of the call graph *)
    let toposort_counter = ref 0 in
    let temp_list = ref [] in
    let rec toposort loc =
      let prio = get_sinfo loc in
      if prio >= 0 then
	() (* already visited *)
      else
	begin
	  Hashtbl.replace sinfo_map (C_System_Descr.location_coords loc) (!toposort_counter) ; (* currently visiting *)
	  temp_list := (!toposort_counter, loc) :: !temp_list ;
	  incr toposort_counter ;
	  (* visit children *) 
	  let succs = C_System_Descr.get_outgoing_edges loc in
	  (* handle function calls in edges *)
	  List.iter 
	    (fun e -> 
	      let c = C_System_Descr.get_command e in 
	      match c.C_Command.code with 
		C_Command.FunctionCall exp ->
		  begin
		    match exp with
		      Expression.Assignment(Expression.Assign, _, Expression.FunctionCall(e,_) ) 
		    | Expression.FunctionCall (e, _) ->
			begin
			  match e with
			    Expression.Lval (Expression.Symbol fname) ->
			      try
				let fname_start = (C_System_Descr.lookup_entry_location fname)
				in
				toposort fname_start
			      with _ -> ()
			  | _ ->
			      if (Options.getValueOfBool "nofp") 
			      then 
				(Message.msg_string Message.Error "Warning: Ignoring function pointer call") 
			      else
				failwith ("Function pointer call: " ^ (Expression.toString exp)) ;
			      ()
				
			end
		  end
	      |	_ -> ()
		    ) 
	    succs ;
	  List.iter (fun e -> toposort (C_System_Descr.get_target e)) succs ;
	  ()
	end
    in
    let startloc =  (C_System_Descr.lookup_entry_location (List.hd mainFunctions)) in
    toposort startloc ;
    

    (* At this point toposort is the number of locations marked, since we started at zero *)
    let num_locs = !toposort_counter in
    
    (* priority_to_loc_array is an array mapping priorities to locations *)
    let priority_to_loc_array = Array.make (!toposort_counter) (None) in
    List.iter (fun (p, l) -> priority_to_loc_array.(p) <- Some l ) !temp_list ;
    temp_list := [] ; (* clear space *)

    for i = 0 to !toposort_counter - 1 do
      let (i1,i2) = 
	match (priority_to_loc_array.(i)) with
	  Some l ->	  C_System_Descr.location_coords  l
	| None -> failwith "Umm" 
      in
      Message.msg_string Message.Normal (Format.sprintf "Priority %d, Location (%d, %d)" i i1 i2)
    done ;
    
    (* Initialize two sets: dead, live *)
    let dead = Hashtbl.create 1009 in
    let live = Hashtbl.create 1009 in
    let live_locs = ref 0 
    and dead_locs = ref 0
    and num_tests = ref 0 in
    
    Hashtbl.add live startloc true ;
    priority_to_loc_array.(0) <- None ;
    incr live_locs ;


    let init_region =
      let rec buildloclist lst loclist =
	match lst with [] -> loclist
	| a ::rest -> 
	    try
	      buildloclist rest ((C_System_Descr.lookup_entry_location a) :: loclist) 
	    with 
	      Not_found -> 
		(Format.printf "Creating init_region: No function %s defined in program. Exiting\n" (a);
		 failwith ("No function "^(a)^" defined"))
      in
      let loclist = buildloclist mainFunctions [] in
      This_C_Abstraction.create_region 
	loclist false seedPreds initialState
    in	

    (* toposort_counter is now used as the current node to be tested *)
    let iteration_count = ref 0 in 
    
    let anymore = ref true in
      Message.msg_string Message.Normal "\n\n\n********** Running the tester's while loop **********\n\n" ;
    (* big test while loop until get_next raises Not_found *)
    while (!anymore) do 
      Message.msg_string Message.Normal "\n\n\n***** Running the next iteration of the tester's while loop *******\n" ;
      Message.msg_string Message.Normal (Format.sprintf "Iteration Number %d \n\n" !iteration_count );

      incr iteration_count ;
      decr toposort_counter; 
      let get_next () =
	let continue = ref true in 
	while (!toposort_counter > 0 && !continue) do
	  match priority_to_loc_array.(!toposort_counter) with
	    Some loc -> (* found the next one *)
	      continue := false 
	  | None -> decr toposort_counter
	done ;
	if (!continue ) then
	    (* found something *)
	  raise Not_found
      in
      try 
	get_next () ;
	try
	  begin
	    let is_after_return loc =
	      (* check if this location is immediately after a return statement *)
	      let incoming = C_System_Descr.get_ingoing_edges loc in
	      List.for_all
		(fun e -> 
		  let cmd = C_System_Descr.get_command e in
		  match cmd.C_Command.code with
		    C_Command.Block l -> List.exists (fun x -> match x with C_Command.Return _ -> true | _ -> false) l
		  | _ -> false
		  )
		incoming 
	    in
	    let error_loc =
	      match priority_to_loc_array.(!toposort_counter) with
	      | Some err_loc ->
		  err_loc
	      | None -> failwith "testModelCheck: Something wrong with get_next"
	    in
	    Message.msg_string Message.Normal (Format.sprintf "Checking reachability of %s \n" (C_System_Descr.location_to_string error_loc) );
	    let error_region =
		  Some (This_C_Abstraction.create_region [error_loc] true Predicate.True Predicate.True)
	    in
	    let mc_return_val =
	      if (is_after_return error_loc) then
		(* get the easy case: the location immediately following a return statement can
		   never be reached 
		   *)
		No_error 
	      else
		try
		  (This_LMC.model_check init_region error_region)
		with e ->
		  begin
		    
		    Message.msg_string Message.Error ("Exception raised by model_check!: "^(Printexc.to_string e));
		    raise e
		  end
	    in 
	    match (mc_return_val) with
	      No_error ->
		let rec expand_dead_locs error_loc =
		(*
		  let ic = List.filter 
		      (fun e -> try not (Hashtbl.find dead (C_System_Descr.get_source e)) with _ -> true) 
		      (C_System_Descr.get_ingoing_edges error_loc) in
		  if List.length ic = 1 then
		    let edg = List.hd ic in
		    match (C_System_Descr.get_command edg).C_Command.code with
		      C_Command.Skip 
		    | C_Command.Block _ ->
			let src = (C_System_Descr.get_source edg) in
			(* can mark the source dead *)
			Hashtbl.add dead src ; 
			incr dead_locs ;
			let pr = Hashtbl.find sinfo_map (C_System_Descr.location_coords src) in
			priority_to_loc_array.(pr) <- None ;
			expand_dead_locs src
		    | _ -> ()
		  else ()
*)
		  ()
		in
		Message.msg_string Message.Normal "\n\nNo error found. Location unreachable.\n" ;
		Hashtbl.add dead error_loc true ;
		incr dead_locs ;
		(* expand as far as possible *)
		expand_dead_locs error_loc 
	    | Error_reached (reg_i, op_list, reg_e) ->
		let expand_error_path (reg_i, op_list, reg_e) =
		  (* should get the state at the end of the error path from the constraints *)
		  
		  ()
		in
		Message.msg_string Message.Normal "\n\nError found! The system is unsafe :-(\n" ;
		(* Add all locations in the error trace to the live hash table *)
		(* set all these locations to null in the priority_to_loc_array *)
		incr num_tests ;

		if (Hashtbl.mem live error_loc) then () else
		begin
		  Hashtbl.add live error_loc true ;
		  incr live_locs 
		end;
		List.iter 
		  (fun e -> 
		    let l = C_System_Descr.get_source e in
		    if (Hashtbl.mem live l) then () else
		    begin
		      Hashtbl.add live l true ;
		      incr live_locs 
		    end;
		    let pr = Hashtbl.find sinfo_map (C_System_Descr.location_coords l) in
		    priority_to_loc_array.(pr) <- None
		  ) 
		  op_list ;
		  (* expand the error path as far as possible *)
		expand_error_path (reg_i, op_list, reg_e) 
		  (* generate a test *)
		  (* TO BE DONE *)

	    | _ -> failwith "testModelCheck: Unexpected return value from model_check"
	  end ;
	  priority_to_loc_array.(!toposort_counter) <- None ;
	with e -> 
	  begin 
	    This_C_Abstraction.dump_abs (); 
	    Message.msg_string Message.Error ("Exception raised :("^(Printexc.to_string e));
	    raise e
	  end ;
      with Not_found -> anymore := false 
    done ;
    Message.msg_string Message.Normal "\n\nTest stats:\n" ;
    Message.msg_string Message.Normal ("\nNumber of iterations = " ^ (string_of_int !iteration_count) ) ; 
    Message.msg_string Message.Normal "\n\nModel checker stats:\n" ;
    Message.msg_printer Message.Normal This_LMC.print_stats () ;
    Message.msg_string Message.Normal "\n\nAbstractor stats:\n" ;
    Message.msg_printer Message.Normal This_C_Abstraction.print_stats () ;
    Message.msg_string Message.Normal ("\nNumber of locations = " ^ (string_of_int num_locs) ) ; 
    Message.msg_string Message.Normal ("\nNumber of live locations = " ^ (string_of_int !live_locs) ) ; 
    Message.msg_string Message.Normal ("\nNumber of dead locations = " ^ (string_of_int !dead_locs) ) ; 
    Message.msg_string Message.Normal ("\nNumber of tests = " ^ (string_of_int !num_tests) )  ;
    (* color code the cfa graph *)
    (* red: dead *)
    (* green: reachable *)
    (* white: syntactically unreachable *)
    let output_cfa_dot ch fundef =
      let known_locs = ref IntSet.empty in
      let rec output_loc loc =
	let color = 
	  if (Hashtbl.mem live loc) then "green" 
	  else
	    if (Hashtbl.mem dead loc) then "red"
	    else "blue" 
	in
	let locid = (snd (C_System_Descr.location_coords loc)) in
	Printf.fprintf ch "  %d[color=%s,style=filled];\n" locid color;
	known_locs := IntSet.add locid !known_locs;
	List.iter
          (fun e ->
	    let l = C_System_Descr.get_target e in
	    let lid = snd (C_System_Descr.location_coords l) in
	    let cmd = C_System_Descr.get_command e in
            Printf.fprintf ch "  %d -> %d [label=\"%s\"]\n"
              locid lid
              (String.escaped (C_Command.to_string cmd));
            if not (IntSet.mem lid !known_locs) then output_loc l)
          (C_System_Descr.get_outgoing_edges loc)
      in
      try
	let loc = (C_System_Descr.lookup_entry_location fundef) in
        Printf.fprintf ch "digraph %s {\n" fundef;
        output_loc loc;
        Printf.fprintf ch "}\n\n";
        ()
      with C_System_Descr.NoSuchFunctionException _ -> ()
    in
    let cfa_file = "test.cfa" in
    let ch = open_out cfa_file in
    List.iter (output_cfa_dot ch)
      (C_System_Descr.list_of_functions ()) ;
    close_out ch
  end 


let interactive () =
    (* Shell mode. Read inputs from stdin and execute them. *)
  let prompt () =
   Format.printf "> @?" in
  let parse_error () = Format.eprintf "Syntax Error\n@?" in
  let lb = Lexing.from_channel stdin in
  try
    while true do
      prompt ();
      begin try
(* Need to write the command interpreter. *)
raise End_of_file (*
       Command.command Lexer.token lb
*)
      with
        | Parsing.Parse_error -> parse_error () 
      end;
      print_newline(); flush stdout
    done
  with
    | End_of_file    -> Format.printf "\n@?"; exit 0


(* subgoal manager *)

let print_copyright () =
  Message.msg_string Message.Normal "Blast 1.0. Copyright  2003
Regents of the 
University of California at Berkeley";
  Message.msg_string Message.Normal "This program contains Foci,
Copyright  2003
 Cadence Berkeley Labs," ;
  Message.msg_string Message.Normal "Cadence Design Systems. All rights
reserved."



(* main function: the glue *)
let main () =
(* try *)
  Message.set_verbosity Message.Default;
  Stats.reset ();
  Options.buildOptionsDatabase ();

 print_copyright () ;
  
  if (Options.getValueOfBool "interactive" = true) then
    (* Shell mode. Read inputs from stdin and execute them. *)
    interactive ()
  else begin
    let reachLabel = (Options.getValueOfString "L")
    and source_files = (Options.getSourceFiles ()) in
    (* No reachLabel has been specified on the command line *)
    if(reachLabel = "") then
      begin
        Printf.printf "No error label specified!\n" ;
        exit 1
      end
    else
      () ;
    if (source_files == [])
    then
      begin
        Message.msg_string Message.Error ("Error: no input source file specified") ;
        exit 1
      end ;
    Message.msg_string Message.Normal ("Parsing files: " ^ (Misc.strList source_files)) ;
    
    Stats.time "Parse:" C_System_Descr.initialize source_files ;
    AliasAnalyzer.constructAliasDatabase () ; 
    
    if (Options.getValueOfBool "test") then
      begin
	testModelCheck () ; exit 1 
      end 
    else () ; 
  
    (* Add the sym_var_hooks for CF reachability *)
    let error_location =
      match (C_System_Descr.get_locations_at_label reachLabel) with
	[l] -> Some l
      | [] when Options.getValueOfBool "checkRace" -> None
      | []  ->
          Message.msg_string Message.Error ("Error: label '" ^ reachLabel ^
                                            "' was not found in the source file(s)") ;
          exit 1
      | _   ->
          Message.msg_string Message.Error ("Error: label '" ^ reachLabel ^
                                            "' appears multiple times in the source file(s)") ;
          exit 1
    in
      (* Hooking up the spec *)
      (* RJ: old "spec" stuff. eschewed. *)
      (*
	C_System_Descr.instrument_with_spec (Options.getValueOfString "spec_fun_pref")
        (Options.getValueOfString "spec_arg_pref") ;
	*)
    Message.msg_string Message.Debug "\n\n\n\n\nPrinting System Description\n\n" ;
    Message.msg_printer Message.Debug C_System_Descr.print_system () ;
(*
  if (Options.getValueOfBool "debug") then begin
  C_System_Descr.compute_callgraph ();
  C_System_Descr.output_callgraph "callgraph.dot" ;
  end ;
  *)
    
    Stats.time "modelCheck" modelCheck error_location ;
  end;
  let memstats = Gc.stat () in
  Message.msg_string Message.Normal ("Minor Words : "^(string_of_float memstats.Gc.minor_words)) ;
      Message.msg_string Message.Normal ("Major Words : "^(string_of_float memstats.Gc.major_words)) ;
      Message.msg_string Message.Normal ("Total size of heap in words : "^(string_of_int memstats.Gc.heap_words)) ;
      Stats.print stderr "Timings:\n"
(* with a -> flush stdout; flush stderr; raise a*)
;;


(* [Printexc.catch main ()] is deprecated as of OCaml 3.01: run the bytecode
   with OCAMLRUNPARAM=b to get a diagnostic with backtrace *)
let () = try main () with e -> 
  begin
    Printf.printf "Ack! The gremlins again";
    Message.msg_string Message.Normal "Ack! The gremlins again!";
    raise e;
  end

