(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: misc.ml,v 1.18 2003/08/13 19:32:20 jhala Exp $
 *
 * This file is part of the BLAST Project.
 *)

(**
 * This module provides some miscellaneous useful helper functions.
 *)
open Str

let swap (x,y) = (y,x)

let fst3 (x,y,z) = x
let snd3 (x,y,z) = y
let thd3 (x,y,z) = z
 
(* Poor man's set operations (in terms of lists) *)

(* add an element to a list if it's not already in there *)
let add lst elem = if List.mem elem lst then lst else elem::lst

(* union two lists representing sets *)
let union l1 l2 = List.fold_left add l2 l1

(* let compact l = union l [] *)

(* set difference using lists to represent sets *)
let difference l1 l2 = List.filter (function elem -> not(List.mem elem l2)) l1

let list_remove l e = List.filter (fun x -> not (x = e)) l

let intersection l1 l2 = difference l1 (difference l1 l2)


(* set containment using lists *)
let rec subsetOf l1 l2 =
  match l1 with
    [] -> true
  | a::rest -> (List.mem a l2) && (subsetOf rest l2)

(* given a list of strings, produce a single string that is the
   comma-separated list of the original strings *)
let strList l = String.concat ", " l


(* returns the to_string function corresponding to the given
   pretty-printer *)
let to_string_from_printer printer =
  function data ->
    printer Format.str_formatter data ;
    Format.flush_str_formatter ()

(* given a printer for a type t, returns a printer for lists of elements of
   type t *)
let list_printer_from_printer printer fmt list =
  Format.fprintf fmt "@[[@[" ;
  begin
    match list with
        [] ->
          ()
      | head::tail ->
          printer fmt head ;
          List.iter (function e -> Format.fprintf fmt ";@ " ;
                                   printer fmt e)
                    tail
  end ;
  Format.fprintf fmt "@]]@]"

(* returns true if the character is a digit [0-9] *)
let is_digit c =
	match c with
		'0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9' -> true
	| _ -> false

let map_exn ex f l = 
  let rec _map _l = 
    match _l with
      [] -> []
    | h::t -> try (f h)::(_map t) with ex -> (_map t)
  in
  _map l

let rec map_partial f l = 
  match l with
      [] -> []
    | h::t -> 
	begin
	  match f h with
	      Some x -> x::(map_partial f t)
	    | None -> (map_partial f t)
	end

  

let filter_cut f l = 
(*  let rec _ac (l1,l2) _l = 
    match _l with
	h::tl -> (if f h then _ac (h::l1,l2) else _ac (l1,h::l2)) tl
      | [] -> (l1,l2)
  in
    _ac ([],[]) l *)
  let l1 = List.filter f l in
  let l2 = difference l l1 in
  (l1,l2)

(* returns the truncated list containing the first n elements Raise Failure "nth" if list is too short *)
let truncateList lst n =
  let rec _truncateWorker l rest len =
    if len = 0 then rest else 
    match l with
      [] -> failwith "truncateList: list too short"
    | head::tail -> _truncateWorker tail (rest @[head]) (len - 1)
  in
  _truncateWorker lst [] n

let chop_list f l = 
  let rec search st fin = 
    match fin with
	[] -> (st,fin)
      | h::tl -> if f h then (st,fin) else search (st@[h]) tl
  in
    search [] l

(* chop s chopper returns Some([x;y;z...]) if s = x.chopper.y.chopper ..., None o.w. *)
let chop s chopper = 
  let reg = Str.regexp chopper in
    Str.split reg s  


(* like chop only the chop is by chop+ *)
let chop_star  chopper s = 
  let reg = Str.regexp (Printf.sprintf "[%s+]" chopper) in
    Str.split reg s

let bounded_chop s chopper i = 
  let reg = Str.regexp chopper in
    Str.bounded_split reg s i 
      
let is_prefix p s = 
  let reg = Str.regexp p in
    Str.string_match reg s 0

let is_substring s subs = 
  let reg = Str.regexp subs in
    try 
      Str.search_forward reg s 0; true
    with
	Not_found -> false

let k_suffix k s = 
  let s_len = String.length s in
  let rem_len = s_len - k in 
    if rem_len < 0 then failwith "bad call to k_suffix"
    else String.sub s k rem_len
      


(*  chop_after_prefix p s = if s = p.s' then s' *)
let chop_after_prefix prefix s =
  let reg = Str.regexp prefix in
    (*if (Str.string_match reg s 0) then *)
    (List.hd(Str.bounded_split reg s 1))
   
(* insert_before c s ins = s1 ins c s2 where s = s1 c s2 else s *)

let insert_before c s ins =
  let l = bounded_chop s c 2 in
    match l with
	[h] -> s
      | [s1;s2] -> Printf.sprintf "%s%s%s%s" s1 ins c s2
      | _ -> failwith "This cannot happen!"


(** unzip  a,b-list returns a list, b list *)
  
let unzip l =
  let rec ac l l1 l2 =
    match l with
        (a,b)::tl -> ac tl (a::l1) (b::l2)
      | [] -> (List.rev l1, List.rev l2)
  in
    ac l [] []


(** hashtbl_keys tbl returns the list of keys that have bindings in the table -- assumes that every key has
  a unique binding *)

let hashtbl_keys t =
  Hashtbl.fold (fun x y l -> x::l) t []
 
(* Hashtable union: hash_union t1 t2 adds all the data of t2 to t1 *)

let hashtbl_addtable t1 t2 = 
  let f_add key data = Hashtbl.add t1 key data in
    Hashtbl.iter f_add t2

(* search_list f l walks through l and finds the first guy st f(x) is some..., 
   it returns that some *)
let search_list f list = 
  let rec search l = 
    match l with 
	[] -> None
      | h::t -> 
	  begin
	    match f h with 
		Some(s) -> Some(s)
	      | None -> search t
	  end
  in
    search list

let rec list_gather_option l = 
  match l with
      [] -> []
    | Some(a)::tl -> a::(list_gather_option tl)
    | None::tl -> (list_gather_option tl)

let rec list_add l =
  let add x y = x + y in
    List.fold_right add l 0

let string_option_to_string so = 
  match so with
    Some(s) -> s
  | None -> "NONE"

let simplify_boolean_unroll l op base_case =
  (List.fold_left 
	 (function a -> function b -> "( "^op^" "^b^" "^a^")") base_case l)

let rec list_max l = 
  match l with
    [] -> failwith("empty list given to list_max !!")
  | [i] -> i
  | h::tl -> max h (list_max tl);;
	
let write_to_file fname string = 
	     let oc = open_out fname in
	     output_string oc string; 
	     close_out oc

let write_list_to_file fname string_list =
  let oc = open_out fname in
    List.iter (output_string oc) string_list;
    close_out oc

let append_to_file fname _string = 
   (*let _ = Printf.printf "Writing to file %s : %s" fname _string in*)
   let string = _string in
   let oc = Unix.openfile fname  [Unix.O_WRONLY; Unix.O_APPEND; Unix.O_CREAT] 755  in
     ignore (Unix.write oc string 0 ((String.length string)-1) ); 
     Unix.close oc

let chop_last l = 
  if l = [] then [] else (List.rev (List.tl (List.rev l)))

let rec make_all_pairs l = 
  match l with
      [] -> []
    | h::tl -> (List.map (fun x -> (h,x)) tl)@(make_all_pairs tl)

let rec cross_product l1 l2 = 
  let _cross x = List.map (fun y -> (x,y)) l2 in
  List.flatten (List.map _cross l1)

let array_reset arr resetval = 
  Array.fill arr 0 (Array.length arr) resetval

(* ambitious -- take a text file and return a list of lines that is the file *)
let string_list_of_file fname = 
  let ic  = open_in fname in
  let doneflag = ref false in
  let listoflines = ref [] in
    while not (!doneflag) do
      try
	listoflines := (input_line ic)::(!listoflines)
      with
	  End_of_file -> doneflag := true
    done;
    close_in ic;
  List.rev !listoflines

let words_of_string string = 
  let reg = Str.regexp " +" in
  let l = Str.split reg string in
    List.filter (fun x -> not ((x = "") || (x = " "))) l

(* source file is some random text file and grepfile is a list of interesting 
words *)

let rec string_list_cat sep l = 
  match l with 
      [] -> "" 
    | [h] -> h
    | h::tl -> h^sep^(string_list_cat sep tl)

let grep_file sourcefile grepfile =
  let wordtable = Hashtbl.create 101 in
  let interesting_words = List.flatten (List.map words_of_string (string_list_of_file grepfile)) in
  let _ = List.iter (fun x -> Hashtbl.add wordtable x true) interesting_words in
  let lines_of_file = List.map words_of_string (string_list_of_file sourcefile) in
    List.filter (fun x -> List.exists (Hashtbl.mem wordtable) x) lines_of_file
  

let run_on_table f tab = 
  let reslist_ref = ref [] in
    Hashtbl.iter (fun x y -> reslist_ref := (f x y)::(!reslist_ref)) tab;
    !reslist_ref
      
let list_last l = 
  try List.hd (List.rev l) with Not_found -> failwith "list_last on Empty list!"
 
let bool_to_string b = 
  match b with
      true -> "true"
    | false -> "false"

(** unzip  a,b-list returns a list, b list *)

let unzip l =
  let rec ac l l1 l2 = 
    match l with
	(a,b)::tl -> ac tl (a::l1) (b::l2)
      | [] -> (List.rev l1, List.rev l2)
  in
    ac l [] []

(** l is a list of even length -- this returns a pair of lists -- the even elts and the odd elts *)
let unzip_joined l = 
  let rec _uj l l1 l2 = 
    match l with
	[] -> (l1,l2)
      | h1::h2::t -> _uj t (h1::l1) (h2::l2)
      | [h1] -> (h1::l1,l2)
  in
  let (r1,r2) = _uj l [] [] in
    (List.rev r1,List.rev r2)

(** hashtbl_keys tbl returns the list of keys that have bindings in the table -- assumes that every key has 
  a unique binding *)

let hashtbl_keys t = 
  Hashtbl.fold (fun x y l -> x::l) t []
    
(** hashtbl_map tbl f returns the list f a b where the bindings in the table are a -> b *)
let hashtbl_map tbl f = 
  let keys = hashtbl_keys tbl in
    List.map (fun x -> f x (Hashtbl.find x)) keys

(** compact l -> l' where l' is the list l without redundancies *)
let compact l = 
  let _t = Hashtbl.create (List.length l) in
  let _ = List.iter (fun x -> Hashtbl.replace _t x true) l in
  hashtbl_keys _t

  (* compute the power set of a given set of elements represented as a 
     list.  the presence of an element e in a subset is represented as 
     (true, e), and the absence is represented as (false, e). 
  *)
  let powerSet l = 
(*
    let rec powAux currSet elems =
      match elems with
          [] -> [currSet]
        | head::tail ->
            (powAux ((true,head)::currSet) tail)@
            (powAux ((false,head)::currSet) tail) in
      powAux [] l
*)
  (* Generate all the (exponentially many) minterms to check. 
   * We are given l, a list of absPredExpressions. For example, l might
   * be [a;b;c]. We are supposed to generate:
   * 	a, ~a, b, ~b, c, ~c
   * 	ab, a~b, ~a~b, ~ab, ac, a~c, ~ac, ~a~c, bc, b~c ...
   * 	abc, ab~c, a~bc, ...
   *
   * Through amazingly convoluted coding, this procedure uses an amount of
   * stack space that is LINEAR in MaxCubeLen (which is <= l).
   * Heap space is exponential, but so is the answer. *)
  match l with [] -> [[]] 
  | _ ->
  (* what is the maximum cube length for one of our answers? *)
    let maxCubeLen = List.length l
    in 
  (* we'll store our answers here *)
    let answer_list = ref [] in 
    
  (* we need random access to the input list of predicates, so we convert
     * it to an array *)
    let l_array = Array.of_list l in 
    let l_len = maxCubeLen - 1 in 
    
(*[rupak]    for cubeLen = maxCubeLen downto 1 do *) let cubeLen = maxCubeLen in (*[rupak]*)
    (* now we want to spit out all minterm-lists of length cubeLen *)
    (* so we will consider all subsets of l of length cubeLen *)
    (* inThisSubset is an array of indices into l_array *)
      let inThisSubset = Array.make cubeLen (-1) in
    (* exampe: if inThisSubset.(0) = 3 and inThisSubset.(1) = 5 then
       * this subset consists of elements 3 and 5 *)
      
    (* we also want to know if we are including the input term or its
       * negation *)
      let polarity = Array.make cubeLen 0 in 
      
    (* Here comes the black magic. Normally to enumerate all subsets of
       * length three you would use three nested for-loops. To enumerate
       * all subsets of length cubeLen we need cubeLen nested for loops.
       * We will build our loop nest up dynamically using function pointers.
       *
       * To avoid considering both 3,5 and 5,3 as subsets of length 2 we will
       * only consider subsets that have their l-indices in ascending order.
       *)

    (* "iter" is one such for loop. 
       * "index" tells us which element of inThisSubset and polarity we are
       * setting. "last_value" helps us stay in ascending order.
       * "continuation" is what to call when we are done. We call it with
       * our current value as its only argument. 
       *)
      let iter (index : int) (last_value : int) (continuation : int -> unit) =
	for i = last_value+1 to l_len do (* consider every element in l *)
	  inThisSubset.(index) <- i; (* it just became element 'index' 
			 		* of our subset *)
	  for pol = 0 to 1 do 	   (* include it once positively *)
	    polarity.(index) <- pol ; (* and once negatively *)
	  (continuation) i
	  done
	done
      in 
      (* "construct" is the inner-most loop. Given that previous iterations
       * have set up the inThisSubset and polarity arrays for us, construct
       * the associated cube. *)
      let construct _ = 
	let this_cube = ref [] in (* we'll build the cube here *)
	for gather = 0 to cubeLen-1 do (* pick out all elts in this subset *)
	  let elt = l_array.(inThisSubset.(gather)) in 
	  let elt' = if polarity.(gather) = 0 then 
	    (false,elt) else (true,elt) in 
	  this_cube := elt' :: !this_cube ; 
	done ; (* now that we have it, append it to the answer list *)
	answer_list := !this_cube :: !answer_list
      in 
      (* Now it's time to built up our loop nest. Recall that we'll have
       * cubeLen for loops. We'll store them in this array. The default
       * array element is "construct", which handles the innermost loop. *)
      let funPtrArray = Array.make cubeLen construct in
      (* NOTE: ocaml will not do what you want if you try to define these
       * guys like: "myfun := iter x y (!myfun)". So we need the array. *)
     
      (* Actually constuct the loop nest. Each iteration calls another
       * iteration. *)
    for i = 1 to cubeLen-1 do
      funPtrArray.(i) <- (fun last_val -> iter i last_val funPtrArray.(i-1)) 
    done ; 
    (* Here's the outermost loop: This one is an actual function call. *)
      iter 0 (-1) (funPtrArray.(cubeLen-1)) ;
    (*[rupak]done;[rupak]*)
    !answer_list


let list_ordered_map f l = 
  let rec _acm _l =
    match _l with
	[] -> []
      | h::t -> let r = f h in r::(_acm t)
  in
    _acm l

let list_cut i l1 =
  let rec _lc i l1 l2 =
    (* assert (i >= List.length l1); *)
    if i = 0 then (List.hd l1, l2)
    else 
      let h::t = l1 in
	_lc (i-1) t (h::l2)
  in
    _lc i l1 []

let list_cut' i l1 =
  let rec _lc i l1 l2 =
    (* assert (i >= List.length l1); *)
    if i = 0 then (List.tl l1, List.hd l1, l2)
    else 
      let h::t = l1 in
	_lc (i-1) t (h::l2)
  in
    _lc i l1 []

(* assumes that all the supplied integers are  >= 0 *)
let min (l: int list) = 
  let rec _min _l curr =
    match _l with 
	[] -> curr 
      | h::t -> 
	  _min t (if (curr = -1 or h < curr) then h else curr)
  in
    _min l (-1)

(** repeats f: unit - > unit i times *)
let rec repeat_fn f i = 
  if i = 0 then ()
  else (f (); repeat_fn f (i-1))

(** takes a dom_rel should be a p.o. I guess, and returns the maximal elts of list_l *)
(* as you can see, a most frivolous algorithm, but hey! *)
let maximal dom_rel list_l = 
  let stored_ll = list_l in
  let is_not_dominated l = not (List.exists (fun x -> dom_rel x l) stored_ll) in
    List.filter is_not_dominated list_l

  
(* tmp *)




(* input: 'a array, f: 'a -> paren. Output: l,r : int -> int *)

let paren_match f arr =
  let n = Array.length arr in
  let left_array = Array.create n (-1) in
  let right_array = Array.create n (-1) in
  let pot = ref [] in
  let _ = 
    for i = 0 to n-1 do
      match (f (arr.(i))) with
	  "(" -> 
	    begin
	      pot := i::!pot;
	      Array.set left_array i i
	    end
	| ")" -> 
	    begin
		let (tos::tl) = try !pot with _ -> failwith "Bad Nesting" in
		  Array.set right_array tos i;
		  Array.set left_array i tos;
		  pot := tl
	    end
	| _ -> 
	    let tos = try List.hd !pot with _ -> -1 in
	      Array.set left_array i tos
    done
  in  
  let l i = if i < 0 || i >= Array.length left_array then -1 else left_array.(i) in
  let r i = if (l i = -1) then -1 else right_array.(l i) in
    (l,r)

let rec list_tabulate f start stop = 
  if start >= stop then []
  else 
    (f start)::(list_tabulate f (start+1) stop)
