(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(**
  Process command line options.
  
   Build database with 
     buildOptionsDatabase : unit -> unit

   Values of flags accessible with 
	isSet : string -> bool
  and
	getValueOf : valueType -> valueType
  and its incarnations.

  Note: values of all types must be added to the database with
   their default values.
   To get the value of an option, query the database with 
   the corresponding command line option, e.g., query "L"
   to get the value of the reachLabel

*)

type valueType = 
  Bool of bool
| String of string
| Int of int
| Float of float	
| StringList of string list

exception OptionNotFoundException of string

let optionsDatabase = Hashtbl.create 13

(*** Command line arguments and global variables to be set ***)
let _ = Hashtbl.add optionsDatabase "L" (String "ERROR")
let _ = Hashtbl.add optionsDatabase "interactive" (Bool false)

(* O stands for optimization level. Default is 0 (no optimizations).
   Optimization level 1 : do cone of influence slicing.
   Optimization level 2 : do partial evaluation (see also -pe) 
                          and cfg compaction. Note that cfg compaction
                          can add infeasible paths in the program, so
                          you can end up with infeasible error traces.
*)
let _ = Hashtbl.add optionsDatabase "O" (Int 0)
    
let _ = Hashtbl.add optionsDatabase "main" (StringList [ "main" ])
let _ = Hashtbl.add optionsDatabase "pta" (String "noflow")
let _ = Hashtbl.add optionsDatabase "pred" (String "")
let _ = Hashtbl.add optionsDatabase "init" (String "")
let _ = Hashtbl.add optionsDatabase "ax" (String "")
let _ = Hashtbl.add optionsDatabase "custom" (String "")
let _ = Hashtbl.add optionsDatabase "refine" (String "fwd")
let _ = Hashtbl.add optionsDatabase "subgoal" (String "")
let _ = Hashtbl.add optionsDatabase "spec_fun_pref" (String "__BLAST_")
let _ = Hashtbl.add optionsDatabase "spec_arg_pref" (String "__BLAST_PARAM_")
let _ = Hashtbl.add optionsDatabase "nofp" (Bool false)
let _ = Hashtbl.add optionsDatabase "pe" (Bool false)
let _ = Hashtbl.add optionsDatabase "pf" (Bool false)
let _ = Hashtbl.add optionsDatabase "pfgen" (Bool false)
let _ = Hashtbl.add optionsDatabase "pffile" (String "/tmp/lf.pfs")
let _ = Hashtbl.add optionsDatabase "forget" (Bool false)
(*let _ = Hashtbl.add optionsDatabase "killLoc" (String "_BLAST_KILL_LOCATION")*)
let _ = Hashtbl.add optionsDatabase "killLoc" (String "")
let _ = Hashtbl.add optionsDatabase "killFile" (String "")
let _ = Hashtbl.add optionsDatabase "forget" (Bool false)
let _ = Hashtbl.add optionsDatabase "xml" (Bool false)
let _ = Hashtbl.add optionsDatabase "cfa-dot" (String "")
let _ = Hashtbl.add optionsDatabase "nocache" (Bool false)
let _ = Hashtbl.add optionsDatabase "cov" (Bool false)
let _ = Hashtbl.add optionsDatabase "reclaim" (Bool false)
let _ = Hashtbl.add optionsDatabase "stop" (Bool false)
let _ = Hashtbl.add optionsDatabase "bfs" (Bool true)
let _ = Hashtbl.add optionsDatabase "bddcov" (Bool true)
let _ = Hashtbl.add optionsDatabase "scope" (Bool false)
let _ = Hashtbl.add optionsDatabase "dc" (Bool false)
let _ = Hashtbl.add optionsDatabase "restart" (Bool false)
let _ = Hashtbl.add optionsDatabase "cref" (Bool false)
let _= Hashtbl.add optionsDatabase "wpsat" (Bool false)
(*
let _ = Hashtbl.add optionsDatabase "allPreds" (Bool false)
*)

(* the first thing to do is in main is to set this to true if ... *)
(* Possible value of predH
   0 : just the vanilla thing where it extracts predicates from proofs
   1 : every time it adds a predicate p, it adds all the constant predicates associated with it
   2 : does the above AND calls find_all_predicates if a trace fails
*)

let _ = Hashtbl.add optionsDatabase "traces" (Bool false)
let _ = Hashtbl.add optionsDatabase "predH" (Int 0)
let _ = Hashtbl.add optionsDatabase "block" (Bool true)
let _ = Hashtbl.add optionsDatabase "craig" (Int 0)
let _ = Hashtbl.add optionsDatabase "bset" (Int 8)
let _ = Hashtbl.add optionsDatabase "localrestart" (Bool false)
let _ = Hashtbl.add optionsDatabase "tsd" (String "")

(* the craig option is only useful if you have chosen block *)

(* Options relating to dumping traces *)
let _ = Hashtbl.add optionsDatabase "tracefile" (String "error.btr")
let _ = Hashtbl.add optionsDatabase "file" (String "")
let _ = Hashtbl.add optionsDatabase "demo" (Bool false)
let _ = Hashtbl.add optionsDatabase "loadabs" (Bool false)
let _ = Hashtbl.add optionsDatabase "scheck" (Bool false)
let _ = Hashtbl.add optionsDatabase "incref" (Bool false)

 (* Possible algorithms for post.
   Options are
   1. H -- for the DasDillPark algorithm
   2. slam -- for the SLAM-like algorithm with no optimizations
   3. slam1 -- slam + basic block optimizations
   4. slam2 -- slam1 + branch optimizations
   5. slam-dumb -- slam ignoring branches
*) 
let _ = Hashtbl.add optionsDatabase "post" (String "slam2")
let _ = Hashtbl.add optionsDatabase "comp" (String "cut")

(* Possible theorem provers that can be used.
   Options are 
   1. ICS
   2. Vampyre
   3. Simplify
   The default is simplify
*)

let _ = Hashtbl.add optionsDatabase "s" (String "simplify")

(* What kind of analysis to do: single-threaded (look for an ERROR label), or
   multi-threaded race analysis (look for unprotected variable accesses) *)

let _ = Hashtbl.add optionsDatabase "checkRace" (Bool false)
let _ = Hashtbl.add optionsDatabase "test" (Bool false)
let _ = Hashtbl.add optionsDatabase "cf" (Bool false)
let _ = Hashtbl.add optionsDatabase "par" (Bool false)
let _ = Hashtbl.add optionsDatabase "inv" (String "")
let _ = Hashtbl.add optionsDatabase "alias" (String "") 
	  (* the alias file containing a list of alias pairs *)
let _ = Hashtbl.add optionsDatabase "easm" (String "") (* the env. assumptions file *)

let _ = Hashtbl.add optionsDatabase "depth" (Int 1)

let _ = Hashtbl.add optionsDatabase "interface" (String "")
let _ = Hashtbl.add optionsDatabase "component" (String "")

let usageMsg = "usage: blast [-traces] [-debug] [-nodebug] [-quiet] [-verbose] [-cfa-dot cfa.dot] [-nofp] [-pe] [-O lev] [-pred pfile] [-init initstatefile] [-custom absfile] [-easm asmfile] [-ax axiomfile] [-L label] [-main startfn] [-post postalgo] [-refine fwd|bwd] [-comp] [-pf] [-pffile lffile] [-pfgen] [-forget] [-s thmprvr] [-xml] [-cf] [-checkRace] [-rw] [-inv invFile] [-interface funcname1 -component funcname2] <source files>\n"


let sourceFiles = ref []

let arg_spec = [
  ("", Arg.Unit (fun () -> ()), "\n\t\t Model Checking Options\n");  
  ("-main", Arg.String (fun a -> 
	Hashtbl.replace optionsDatabase "main" (StringList [a]))
	(*
	let oldlist = let ol = Hashtbl.find optionsDatabase "main" in
	  match ol with StringList l -> l 
	  | _ -> failwith "Strange type of function list" in
	 Hashtbl.replace optionsDatabase "main" (StringList (a::
	oldlist)))*)
	, "start reachability from this function (default main)");
  ("-L", Arg.String (fun a -> Hashtbl.replace optionsDatabase "L" (String a)), 
   "target label for reachability (default ERROR)");  
  ("-cf", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "cf" (Bool true)),
   "use context free reachability");
  ("-bfs", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bfs" (Bool true)), "breadth first search (Default)");
  ("-dfs", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bfs" (Bool false)), "depth first search");
  ("-pred", Arg.String (fun a -> Hashtbl.replace optionsDatabase "pred" (String a)), "seed predicates file");
  ("-init", Arg.String (fun a -> Hashtbl.replace optionsDatabase "init" (String a)), "initial state file");
  ("-inv",  Arg.String (fun a -> Hashtbl.replace optionsDatabase "inv" (String a)), "invariants file");
  ("-nocache", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "nocache" (Bool true)), 
   "dont cache theorem prover calls");
  ("-cov", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "cov" (Bool true)), 
   "keep only coverable nodes");
  ("-reclaim", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "rec" (Bool true)), 
   "dont keep whole tree around reclaim on backtrack");
  
  ("-s", Arg.String (fun a -> Hashtbl.replace optionsDatabase "s" (String a)), 
   "satisfiability solver to be used:\n\tOptions are vampyre/simplify/cvc (default simplify; craig uses vampyre)");
  ("-ax", Arg.String (fun a -> Hashtbl.replace optionsDatabase "ax" (String a)), 
   "additional axioms for simplify to add reasoning about other theories");
  ("-alias",  Arg.String (fun a -> Hashtbl.replace optionsDatabase "alias" (String a)), 
   "alias options\n\tOptions: If file name is bdd, invokes a bdd based alias analysis,\n\totherwise reads the alias relationships from a file. See also -pta");
  ("-pta", Arg.String (fun a -> Hashtbl.replace optionsDatabase "pta" (String a)), 
   "Mode for points to analysis. noflow runs a flow insensitive version, flow a flow sensitive version.\n\tThis version of Blast only supports noflow.");  
  ("-nofp", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "nofp" (Bool true)), "ignore function pointer calls (Blast does not handle function pointers)");

  ("-tsd", Arg.String (fun a -> Hashtbl.replace optionsDatabase "tsd" (String a)), 
   "trace search direction. [zig] goes zigzagging [] restarts from the end");  
  ("-bddcov", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bddcov" (Bool true)), 
   "use only bdds in cover check (default)");
  ("-nobddcov", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bddcov" (Bool false)), 
   "do full cover check");
  ("-scope", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "scope" (Bool true)), "scoping via inferred predicates ");

  ("-wpsat", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "wpsat" (Bool true)), " keep only satisfiable disjuncts in wp ");
  ("-restart", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "restart" (Bool true)), 
   "restart computation (SLAM mode)");
(*  
  ("-allPreds", Arg.Unit (fun ()-> Hashtbl.replace optionsDatabase "allPreds" (Bool true)), "do focus on ALL subtraces ");
*)

  ("-scheck", Arg.Unit (fun a-> Hashtbl.replace optionsDatabase "scheck" (Bool true)), "Additional safety checks for aliasing"); 
("-incref", Arg.Unit (fun a-> Hashtbl.replace optionsDatabase "incref" (Bool true)), "incomplete counterex. analysis"); 
  ("-predH", Arg.Int (fun a -> Hashtbl.replace optionsDatabase "predH" (Int a)), 
   "new predicate heuristic level\n\tOptions: 2: add x = c for all constant c, 6, 7: Additional Heuristics. Recommended Level is 7.");
  ("-block", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "block" (Bool true)), "Analyze trace as blocks ");
  ("-noblock", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "block" (Bool false)), "Use the old trace analysis ");
  ("-craig", Arg.Int (fun a 
    -> Hashtbl.replace optionsDatabase "craig" (Int a);
      Hashtbl.replace optionsDatabase "block" (Bool true)), 
   "Use craig interpolants (FOCI) to get predicates ([1] use with scope, [2] goes the whole hog)\n\tSee the paper Abstractions from Proofs");



  ("", Arg.Unit (fun () -> ()), "\n\t\t Old Heuristics that are no longer used/supported\n");

  ("-comp", Arg.String (fun a -> Hashtbl.replace optionsDatabase "comp" (String a)), 
   "Completeness level:\n\tOptions are cut and path (default cut)");  
  ("-forget", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "forget" (Bool true)), 
   "forget predicates outside subtree");
  ("-post", Arg.String (fun a -> Hashtbl.replace optionsDatabase "post" (String a)), 
   "post algorithm to be used:\n\tOptions are slam and H (default slam)");
  ("-dc", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "dc" (Bool true)), " dont care heuristic for predicates not in scope ");
  ("-refine", Arg.String (fun a -> Hashtbl.replace optionsDatabase "refine" (String a)), 
   "direction of refinement (fwd or bwd), default fwd. No longer used, superceded by -block and -craig.");
  ("-custom", Arg.String (fun a -> Hashtbl.replace optionsDatabase "custom" (String a)), "custom abstraction file");




  ("", Arg.Unit (fun () -> ()), "\n\t\t Program Optimization Options\n");
  ("-pe", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "pe" (Bool true)), "partial evaluation (Constant propagation)");
(*  ("-subgoal", Arg.String (fun a -> Hashtbl.replace optionsDatabase "subgoal" (String a)), "subgoals file"); *)
  ("-O", Arg.Int (fun a -> Hashtbl.replace optionsDatabase "O" (Int a)), "Program optimizations"); 
  ("-depth", Arg.Int (fun a -> Hashtbl.replace optionsDatabase "depth" (Int a)), "Unroll Depth"); 

(*
  ("-killFile", Arg.String (fun a -> Hashtbl.replace optionsDatabase "killFile" (String a)), 
   "kill predicates file");
  ("-killLoc", Arg.String (fun a -> Hashtbl.replace optionsDatabase "killLoc" (String a)), 
   "kill label (default _BLAST_KILL_LOCATION)"); 
*)

  ("", Arg.Unit (fun () -> ()), "\n\t\t Parallel Model Checking and Races\n");  
  ("-checkRace", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "checkRace" (Bool true)),
   "check for races on variable accesses");
  ("-par", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "par" (Bool true)),
   "use data structures for parallel blast");
  ("-easm",  Arg.String (fun a -> Hashtbl.replace optionsDatabase "easm" (String a)), "environment assumptions file");

  ("", Arg.Unit (fun () -> ()), "\n\t\t Saved Abstractions and Summarization\n");  
  ("-loadabs", Arg.Unit (fun a-> Hashtbl.replace optionsDatabase "loadabs" (Bool true)), " Load Abstraction from fname.abs "); 
  ("-interface", Arg.String (fun a -> Hashtbl.replace optionsDatabase "interface" (String a)), "name of a function to take as describing a component interface");
  ("-component", Arg.String (fun a -> Hashtbl.replace optionsDatabase "component" (String a)), "name of a function to check for satisfying the interface");

  ("", Arg.Unit (fun () -> ()), "\n\t\t Proof Generation Options\n");
  ("-pf", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "pf" (Bool true)), "generate proof tree");
  ("-pfgen",Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "pfgen" (Bool true)), "spit out vampyre proofs");
  ("-pffile", Arg.String (fun a -> Hashtbl.replace optionsDatabase "pffile" (String a)), 
   "file to write vampyre proofs (default /tmp/lf.pfs)");

  ("", Arg.Unit (fun () -> ()), "\n\t\t General Options\n");
(*
  ("-quiet", Arg.Unit (fun () -> Message.set_verbosity Message.Quiet), "quiet mode");
  ("-nodebug", Arg.Unit (fun () -> Message.set_verbosity Message.Default), "normal verbosity mode");
  ("-verbose", Arg.Unit (fun () -> Message.set_verbosity Message.Verbose), "verbose mode");
*)
  ("-debug", Arg.Unit (fun () -> Message.set_verbosity Message.Debugging), "debug mode");
  ("-test", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "test" (Bool true)), "test mode");
  ("-cfa-dot", Arg.String (fun a -> Hashtbl.replace optionsDatabase "cfa-dot" (String a)), "output the CFA of every function in ATT dot format");
(*  ("-I", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "interactive" (Bool true)), "interactive shell mode"); *)
  ("-msvc", Arg.Unit (Frontc.setMSVCMode), "parse MS VC");
  ("-stop", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "stop" (Bool true)), "stop when it hits first counterex");  
  ("-traces", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "traces" (Bool true)), "dump bogus traces");
  ("-tracefile", Arg.String (fun a-> Hashtbl.replace optionsDatabase "tracefile" (String a)), " file containing trace information -- for trace viewer");
  ("-file", Arg.String (fun a-> Hashtbl.replace optionsDatabase "file" (String a)), " file containing trace information -- for trace viewer"); 
  ("-demo", Arg.Unit (fun a-> Hashtbl.replace optionsDatabase "demo" (Bool true)), "Demo Mode for the GUI"); 
  ("-xml", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "xml" (Bool true)), "generate error trace as xml files readable by SLAM's gui (Deprecated. Use the blast gui.)");

] 

    

let buildOptionsDatabase () = 
  Arg.parse arg_spec (fun s -> (sourceFiles := s :: (!sourceFiles))) usageMsg ;
  sourceFiles := List.rev !sourceFiles;
  let filename_prefix =
    if !sourceFiles = [] then "__foo"
    else
    begin
    let fname = List.hd !sourceFiles in
      match Misc.chop fname "\." with 
	  [] -> fname
	|  e -> List.hd e
     end
  in
    Hashtbl.replace optionsDatabase "mainsourcename" (String filename_prefix)


let getSourceFiles () = !sourceFiles

let isSet arg =
  try
    let a = Hashtbl.find optionsDatabase arg in
    match a with 
      Bool b -> b
    | _ -> failwith "Options: isSet -- non boolean type"
  with
    Not_found ->
      (Message.msg_string Message.Major "\nOptions: isSet: arg not found in database!\n";
      (* continue *) 
       false)

let getValueOf arg =
  try
	Hashtbl.find optionsDatabase arg
  with Not_found -> raise (OptionNotFoundException arg)

let getValueOfString arg =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    String s -> s
  | _ -> failwith "Options: getValueOfString -- non string type"
  with Not_found -> raise (OptionNotFoundException arg)

let setValueOfString arg str_v =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    String s -> Hashtbl.replace optionsDatabase arg (String str_v)
  | _ -> failwith "Options: setValueOfString -- non string type"
  with Not_found -> raise (OptionNotFoundException arg)

let getValueOfInt arg =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    Int s -> s
  | _ -> failwith "Options: getValueOfInt -- non int type"
  with Not_found -> raise (OptionNotFoundException arg)

let setValueOfInt arg int_v =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    Int s -> Hashtbl.replace optionsDatabase arg (Int int_v)
  | _ -> failwith "Options: setValueOfInt -- non int type"
  with Not_found -> raise (OptionNotFoundException arg)


let getValueOfBool arg =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    Bool s -> s
  | _ -> failwith "Options: getValueOfBool -- non bool type"
  with Not_found -> raise (OptionNotFoundException arg)

let setValueOfBool arg bool_v =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    Bool s -> Hashtbl.replace optionsDatabase arg (Bool bool_v)
  | _ -> failwith "Options: setValueOfBool -- non bool type"
  with Not_found -> raise (OptionNotFoundException arg)


let getValueOfStringList arg =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    StringList s -> s
  | _ -> failwith "Options: getValueOfStringList -- non StringList type"
  with Not_found -> raise (OptionNotFoundException arg)
