(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: theoremProver.ml,v 1.23 2003/08/20 18:44:43 rupak Exp $
 *
 * This file is part of the BLAST Project.
 *)


(**
 * Interface to Vampyre and ICS.  This module contains the interface to the theorem provers.
 *)


open Ast


(* This code maintains a mapping from field names of structs to unique indices.  This allows us
   to translate an expression of the form x.f into (select x i), where i is the unique index of f. 
   This is a hack (that works). A more proper translation should look at
   the types, etc.
*)

(* convert an Expression.expression to Vampyre Input *)
let listOfAddrVars = ref []    
(*  let prfnos = string_of_int (!prfno) in*) (* get rid of prfnos once reinit is fixed *)

let rec convertExp e =
  match e with
    Expression.Lval lv -> convertLval lv
  | Expression.Chlval (l, s) -> Input.Fun ("chlval", [ convertLval l; Input.Var ("_"^s) ])
  | Expression.Binary (op, e1, e2) -> 
      begin
	match op with 
	  Expression.Minus -> Input.Minus(convertExp e1, convertExp e2)
	|	Expression.Plus  -> Input.Plus(convertExp e1, convertExp e2)
	|	Expression.Mul   -> Input.Times(convertExp e1, convertExp e2)
	|	Expression.BitAnd 
	|	Expression.BitOr 
	|	Expression.Div 
	|	Expression.LShift 
	|	Expression.Rem 
	|	Expression.RShift 
	|	Expression.Xor 
	|	Expression.Comma ->
	    Input.Fun(Expression.binaryOpToVanillaString op, [convertExp e1; convertExp e2])
	| Expression.FieldOffset ->
	    begin
	      let ve1 = convertExp e1 and ve2 = convertExp e2 in
              Input.Fun ("foffset", [ve1; ve2])
	    end
	|	_ -> failwith ("convertExp (Vampyre): This operator should not be handled here "^(Expression.toString e))
      end
  | Expression.Constant (c) -> 
      begin
	match c with
	  Constant.Int i -> Input.Int i
	| Constant.String s -> Input.Var ("_STRINGCONSTANT"^s)
	| Constant.Float f -> 
	    if (f -. (ceil f) = 0.0) then Input.Int (int_of_float f)
	    else Input.Var "_FLOATINGPTCONSTANT"
	| _ -> failwith ("Cannot handle constant "^(Constant.toString c))
      end

  | Expression.Unary (uop, e) -> 
      begin
	match uop with 
	  Expression.UnaryMinus -> Input.Minus(Input.Int 0, convertExp e)
	| Expression.UnaryPlus -> convertExp e
	| Expression.Address -> 
	    let v = convertExp e in
	    if (List.mem v (!listOfAddrVars)) then () 
	    else
	      listOfAddrVars := v::(!listOfAddrVars);
	    Input.Fun("addrOf", [v])
	(*$| Expression.Deref -> failwith "Deref"*)
	|	_ -> Input.Fun(Expression.unaryOpToVanillaString uop, [convertExp e])
      end
  | _ -> failwith ("convertExp: " ^ (Expression.toString e) ^ "not handled")
	
  and convertLval l =
    match l with
    | Expression.Symbol(str) -> Input.Var ("_"^str) (* prepend all symbols with "_" for later print out in LF *)
    | Expression.Access(Expression.Dot, e, s) -> (* NEEDS DEBUGGING *) 
	let v = convertExp e in
	Input.Fun("select", [v; Input.Var ("_"^(s))])
	  
    | Expression.Access(Expression.Arrow, e, s) ->
(*
	Input.Fun ("sel4", [convertExp e; Input.Var ("_"^(s))])
*)
	failwith "Arrrow"
    | Expression.Dereference e -> Input.Fun("select", [convertExp e; Input.Int 0])
    | Expression.Indexed (e1, e2) -> Input.Fun("sel4", [convertExp e1; convertExp e2])


  and convertAtom e =
    match e with
      Expression.Binary (op, e1, e2) -> 
	begin
	  match op with 
	    Expression.Eq  -> Input.Eq(convertExp e1, convertExp e2)
	  |	Expression.Ne  -> Input.Neq(convertExp e1, convertExp e2)
	  |	Expression.Gt  -> Input.Gt(convertExp e1, convertExp e2)
	  |	Expression.Lt  -> Input.Lt(convertExp e1, convertExp e2)
	  |	Expression.Ge  -> Input.Geq(convertExp e1, convertExp e2)
	  |	Expression.Le  -> Input.Leq(convertExp e1, convertExp e2)
	  |	_   -> failwith ("convertAtom: Cannot handle this expression: "^(Expression.toString e))
	end
    | _ -> failwith ("convertAtom: "^(Expression.toString e)^ " Not a binary relation operator")
  
  and convertPred e =
    match e with
      Predicate.True -> Input.True
    | Predicate.False -> Input.False
    | Predicate.And plist -> 
	List.fold_left (function a -> function b -> Input.And(convertPred b, a)) 
	  Input.True (plist)
    | Predicate.Or plist -> 
	List.fold_left (function a -> function b -> Input.Or(convertPred b, a)) 
	  Input.False (plist)
    | Predicate.Implies (e1, e2) -> Input.Imp(convertPred e1, convertPred e2)
    | Predicate.Not e1 -> Input.Not(convertPred e1)
    | Predicate.Atom e ->  Input.Atom (convertAtom e)
    | _ -> failwith ("Cannot handle this predicate "^(Predicate.toString e))
 
and convertToVampyreSyntax e =
(*  let prfnos = string_of_int (!prfno) in*) (* get rid of prfnos once reinit is fixed *)(*  prfno := !prfno + 1; *)
  listOfAddrVars := [];
  let v_exp = convertPred e
  in
  let genAddrOfEq b =
    Input.Atom(Input.Eq(b, Input.Fun("sel4", [Input.Fun("addrOf", [b]); Input.Int 0])))
  in
  begin
    let vv = List.fold_left (function a -> function b -> Input.Imp(genAddrOfEq b, a)) 
	v_exp !listOfAddrVars
    in
    vv
  end
      
      
      
(* Call to prover *)
(* An Engine.Failure is assumed to indicate falsehood *)
      
(* Initialize Vampyre *)
module S = Satproc
module CC : S.SATPROC = Equality
module SX : S.SATPROC = Simplex
module SI : S.SATPROC = Sup_inf
module GT : S.SATPROC = Tactical
module TYP: S.SATPROC = Typing
module SP: S.SATPROC = Special

let (vampyreProver,vampyre_assume_list, vampyre_prove_context, vampyre_prove_context_proof) =
  let allSatProcs = 
    [CC.initialize "CC";
      SX.initialize "SX";
      GT.initialize "GT";
     (* SI.initialize "SI"; 
      TYP.initialize "TYP";
      SP.initialize "SP" *)
    ]
  in
  (* circular dependency hack *)
  ignore(Proof.simp_and_subtract_is_zero := Sup_inf.simp_and_subtract_is_zero;
	 Proof.simp_and_subtract_is_at_least_zero := Sup_inf.simp_and_subtract_is_at_least_zero);
  (* end circular dependency hack *)
  Engine.prove allSatProcs

(* Interface functions*)
let proof = ref None 

let queryExpUsingVampyre exp =
  try
    let vampyreInput = convertToVampyreSyntax exp
    in
    let vampyreInput_pushneg = (Input.pushNeg  vampyreInput) in
    try
(*
	let _ = VampyrePretty.printf "Vampyre Saw queryExp: %a \n" Input.dp vampyreInput_pushneg
    	in
*)
      let _ = proof := Some (vampyreProver "foo" (vampyreInput_pushneg)) in
      ignore !proof ; true
    with Engine.Failure -> (Message.msg_string Message.Debug "Proof failed.\n") ; false
  with
    VampyreErrormsg.Error -> (failwith ("Bug in Vampyre ..."))

(* for now only vampyre supports contexts -- assume etc *)
let assume assumptions = 
  let atomize_asm asm_pred = 
    match asm_pred with
      Predicate.Atom e -> (convertAtom e)
    | Predicate.Not Predicate.Atom e -> Input.negate  (convertAtom e)
    | _ -> 
	Message.msg_string Message.Error "Bad assumption in atomize_asm";
	Message.msg_printer Message.Error Predicate.print asm_pred;
	failwith ("bad asm !")
  in

  (* try *) 
    let vam_assumptions = List.map atomize_asm assumptions 
    in
    vampyre_assume_list vam_assumptions
  (* with
    _ -> failwith "Vampyre Bug due to Jhala !" *)

let printproof pred = 
  let _eToCheck = Predicate.implies pred (Predicate.False) in
  let vampyreInput = convertToVampyreSyntax _eToCheck in
  let vampyreInput_pushneg = (Input.pushNeg  vampyreInput) in
    try
      let proof = vampyreProver "foo" (vampyreInput_pushneg)
      in
	VampyrePretty.printf "Vampyre proof: %a \n" Proof.dp proof; ()
    with Engine.Failure -> (Message.msg_string Message.Debug "Proof failed.\n") ; ()
     
	
(* This converts a vampyre predicate to one that blast can absorb *)
let convertToPredicate w =
  let rec convertExp e =
    match e with
      Input.Var var -> 
	      (* drop initial "_" *)
	if (String.get var 0 = '_') then
	  Expression.Lval (Expression.Symbol (String.sub var 1 ((String.length var) - 1)))
	else (* Huh? ConvertExp did not prepend a "_"! *)
	  Expression.Lval (Expression.Symbol var)
	    
    | Input.Int i   -> Expression.Constant (Constant.Int i)
    | Input.Plus (e1, e2) -> Expression.Binary (Expression.Plus, convertExp e1, convertExp e2)
    | Input.Minus (e1, e2) -> Expression.Binary (Expression.Minus, convertExp e1, convertExp e2)
    | Input.Times (e1, e2) -> Expression.Binary (Expression.Mul, convertExp e1, convertExp e2)
    | Input.Div (e1, e2) -> Expression.Binary (Expression.Div, convertExp e2, Expression.Constant (Constant.Int e1))
    | Input.Mod (e1, e2) -> Expression.Binary (Expression.Rem, convertExp e2, Expression.Constant (Constant.Int e1))
    | Input.Fun (s, el) -> Expression.FunctionCall (Expression.Lval (Expression.Symbol s), List.map convertExp el)
    | _ -> failwith "getUsefulPredicates: Do not handle Floor and Ceiling"
  in
  let rec convertAtom a =
    match a with
      Input.Beq (a1, a2) -> Expression.Binary(Expression.Eq, convertAtom a1, convertAtom a2)
    | Input.Eq (e1, e2) -> Expression.Binary (Expression.Eq, convertExp e1, convertExp e2)
    | Input.Neq (e1, e2) -> Expression.Binary (Expression.Ne, convertExp e1, convertExp e2)
    | Input.Geq (e1, e2) -> Expression.Binary (Expression.Ge, convertExp e1, convertExp e2)
    | Input.Leq (e1, e2) -> Expression.Binary (Expression.Le, convertExp e1, convertExp e2)
    | Input.Gt (e1, e2) -> Expression.Binary (Expression.Gt, convertExp e1, convertExp e2)
    | Input.Lt (e1, e2) -> Expression.Binary (Expression.Lt, convertExp e1, convertExp e2)
    | _ -> failwith "getUsefulPredicates: BFun and NotBFun not handled."
	in
  let _w = convertAtom w in
  Predicate.Atom _w

  

(* Interface functions*)

(* check validity of exp using Ics *)
                                                                                                  
let queryExpUsingIcs exp =
  failwith ("We dont want ICS, thank you very much")
(*
  let icsInput = convertToIcsSyntax exp
  and icsState = Ics.init () 
  in
  Printf.printf "ICS OUTPUT = ";
  let tms = Ics.terms_empty () in
  let tm2 = Ics.terms_add icsInput tms in
  Ics.terms_pp tm2;
  flush stderr;
  Ics.is_valid icsState icsInput
*)    

(******************************************************************)
(* Interface to CVC                                               *)
(******************************************************************)
	     
(* keep around info: do we have a cvc server (= other process) 
	     * running? *)

(*
let cvcServerWithProof = ref None   
    
let getCvcServerWithProof () = 
  match !cvcServerWithProof with
    None -> begin
      Message.msg_string Message.Normal "Forking CVC process (with proofs)..." ;
      let ic,oc = Unix.open_process "cvc +proof" 
      in
      cvcServerWithProof := Some(ic,oc) ;
      Message.msg_string Message.Normal "done!\n";
      flush stdout ;
      at_exit (fun () -> 
	ignore (Unix.close_process (ic,oc) ) ) ;
      (ic,oc)
	end
  | Some(a) -> a
*)
	
let cvcServer = ref None   
    
let getCvcServer () = 
  match !cvcServer with
    None -> 
      begin
	Message.msg_string Message.Debug "Forking CVC process (no proofs)..." ;
	let ic,oc = Unix.open_process "cvc" 
	in
	cvcServer := Some(ic,oc) ;
	Message.msg_string Message.Debug "done!\n";
	flush stdout ;
	at_exit (fun () ->
	  ignore (Unix.close_process (ic,oc) ) ) ;
	(ic,oc)
      end
  | Some(a) -> a
	
	
let rec askCvcIfValid ic =
  let line = input_line ic in
  let result = if Misc.is_prefix "Valid" line then 
    true
  else if Misc.is_prefix "Invalid" line then
    false
  else if Misc.is_prefix "Error" line then failwith "Error in CVC output!"
  else (askCvcIfValid ic)
  in
  result

    
let cvc_ground_term_list = ref [] 
    
let add_cvc_ground str = 
  if List.mem str (!cvc_ground_term_list) 
  then () 
  else 
    begin
      if (Misc.is_prefix "ASSERT" str) then
        cvc_ground_term_list := (!cvc_ground_term_list) @ [ str ]
      else 
        cvc_ground_term_list := str::(!cvc_ground_term_list)
    end 
      
let reset_cvc_ground () = 
  cvc_ground_term_list := 
     [
     "select4 : [ [ REAL, REAL ] -> REAL ] ;"; 
     "select_dot : [ [ REAL, REAL ] -> REAL ] ;"; 
     "addrof : [ REAL -> REAL ] ;" ;
     "BitNot : [ REAL -> REAL ] ;" ;
     "BitOr : [ [ REAL, REAL ] -> REAL ] ;"; 
     "BitAnd : [ [ REAL, REAL ] -> REAL ] ;"; 
     "guf : [ [ REAL, REAL ] -> REAL ] ;"; 
   ]
       
let convertToCvcSyntax exp =
  let rec convertExp e =
    match e with
    | Expression.Lval lv -> convertLval lv
    | Expression.Chlval (lv, s) -> 
	begin
	  add_cvc_ground ("FLD"^s ^": REAL") ;
	  Printf.sprintf "(chlval %s %s)" (convertLval lv) ("FLD"^ s)
	end
    | Expression.Binary (op, e1, e2) -> 
	  begin
	    match op with 
	      Expression.Minus -> 
		begin
		    match e2 with
	              Expression.Constant (Constant.Int 0) -> (convertExp e1)
	            | _ -> Printf.sprintf "(%s - %s)" (convertExp e1) (convertExp e2)
		end
	      |	Expression.Plus  -> 
		  begin
		    match e2 with
                      Expression.Constant (Constant.Int 0) -> (convertExp e1)
	            | _ -> Printf.sprintf "(%s + %s)" (convertExp e1) (convertExp e2)
		  end
	      |	Expression.Mul ->   
	          Printf.sprintf "(%s * %s)" (convertExp e1) (convertExp e2)
	      |	Expression.Div ->
	          Printf.sprintf "(%s / %s)" (convertExp e1) (convertExp e2)
	      |	Expression.BitAnd ->
	          Printf.sprintf "BitAnd(%s,%s)" (convertExp e1) (convertExp e2)
	      |	Expression.BitOr -> 
	          Printf.sprintf "BitOr(%s,%s)" (convertExp e1) (convertExp e2)
	      |	Expression.LShift 
	      |	Expression.Rem 
	      |	Expression.RShift 
	      |	Expression.Xor 
	      |	Expression.Comma ->
	          Printf.sprintf "guf(%s,%s)" (convertExp e1) (convertExp e2)
	 
	      |	_ -> failwith ("convertExp: (CVC) This operator should not be handled here "^(Expression.toString e))
	  end
    | Expression.Constant (c) -> 
	begin
	  match c with
		Constant.Int i ->  
	          string_of_int i
	  | Constant.String s -> add_cvc_ground ("STRING"^" : REAL ;");("STRING") 
	  | _ -> failwith ("unknown constant type")
	end
    | Expression.Unary (uop, e) -> 
	begin
	  match uop with 
	    Expression.UnaryMinus -> Printf.sprintf ("( - %s)") (convertExp e)
	  | Expression.UnaryPlus -> Printf.sprintf ("( + %s)") (convertExp e)
	(*  | Expression.Deref -> 
	      begin
		match e with
		  Expression.Binary(Expression.Plus,e1,e2) -> Printf.sprintf "(select4(%s,%s))" (convertExp e1) (convertExp e2)
		| _ -> Printf.sprintf "(select4(%s,0))" (convertExp e)
	      end
*)
	  | Expression.Address -> 
	      begin
	        let e_cvc = convertExp e in
	        let expr = Printf.sprintf "(addrof(%s))" e_cvc in
	           (* Also add an assertion relating deref and address for this expression. *)
                add_cvc_ground (Printf.sprintf "ASSERT (select4 (%s, 0) = %s);" expr e_cvc) ;
	        expr
              end
	  |	_ -> failwith "unknown unop" 
	end
    | _ -> failwith ("convertExp (Cvc): " ^ (Expression.toString e) ^ "not handled")
  and convertLval l =
    match l with
    | Expression.Symbol(str) -> 
	begin
	  let cvc_name = "TMP"^(Str.global_replace (Str.regexp "@") "LOCAL" str) in
	  add_cvc_ground (cvc_name ^ ": REAL ;") ;
	  cvc_name
	end
    | Expression.Dereference drfe -> "select4( "^(convertExp drfe)^ ", 0)"
    | Expression.Access(Expression.Dot, e, s) -> 
	  begin
	    add_cvc_ground ("FLD"^s^" : REAL;");
	    (Printf.sprintf "(select4(%s,%s))" (convertExp e) ("FLD"^s))
	  end
    | Expression.Access(Expression.Arrow, e, s) ->
	begin
	  add_cvc_ground ("FLD"^s^" : REAL;"); 
	  (Printf.sprintf "(select_dot(%s,%s))" (convertExp e) ("FLD"^s))
	end
    | _ -> failwith ("convertLval (Cvc): " ^ (Expression.lvalToString l) ^ "not handled")
	 
  and convertAtom e =
    match e with
      Expression.Binary (op, e1, e2) -> 
	begin
	  let opstring = 
	    match op with 
	      Expression.Eq  -> "="
	    | Expression.Ne  -> "/="
	    | Expression.Gt  -> ">"
	    | Expression.Lt  -> "<"
	    | Expression.Ge  -> ">="
	    | Expression.Le  -> "<="
	    | _   -> "" 
	  in
          if not (opstring = "") then
		Printf.sprintf "(%s %s %s)" (convertExp e1) opstring (convertExp e2)
	  else (convertExp e)
	end
    | _ -> convertExp e 
	  
  and convertPred e =
    match e with
      Predicate.True -> "TRUE"
    | Predicate.False -> "FALSE"
    | Predicate.And plist -> 
	List.fold_left 
	  (fun a -> fun b -> Printf.sprintf " (%s AND %s) " (convertPred b) a) 
	  "1" (plist)
    | Predicate.Or plist -> 
	List.fold_left 
	  (function a -> function b -> Printf.sprintf "(%s OR %s)" (convertPred b) a )
	    "0" (plist)
    | Predicate.Implies (e1, e2) -> Printf.sprintf "(%s => %s)" (convertPred e1) (convertPred e2)
	  
    | Predicate.Not e1 -> Printf.sprintf "(NOT %s)" (convertPred e1)
    | Predicate.Atom e ->  (convertAtom e)
    | _ -> failwith ("Cannot handle this predicate "^(Predicate.toString e))
	  
  in
  (convertPred exp)
    
    
    
let cvcQueryNum = ref 1
let cvcQueryNumPf = ref 1
    
    
(*    
let fdgreedy = open_out "blast-trace-greedy.cvc" 

let _ = at_exit (fun () -> close_out fdgreedy)
*)
    
let queryExpUsingCvcGreedy exp =
  let _ = Message.msg_string Message.Debug ("querying exp in queryExpUsingCvcGreedy :"^(Predicate.toString exp)^"\n") in
   (* first we have to convert the formula and collect all the ground terms that appear in it *)

  let fdgreedy = open_out "blast-trace-greedy.cvc" in

  reset_cvc_ground ();
  let formulas = match exp with Predicate.And l -> l | _ -> failwith ("bad call to queryExpUsingCvcGreedy") in
  let all_asserts = List.map (convertToCvcSyntax) formulas
  in
	     (* at this point all_asserts is a list of all assertions, and !cvc_ground_term_list is
	        the list of all type declarations.
	     *)
  let rec output_asserts oc predlist =
    match predlist with 
      [] -> () 
    | a::rest -> 
	begin
          output_string oc (Printf.sprintf "ASSERT %s;\n" a) ; 
          output_asserts oc rest 
	end
  in 
  let rec output_list oc lst =
    match lst with 
      [] -> () 
    | a::rest -> 
        begin
          output_string oc (Printf.sprintf "%s\n" a) ; 
          output_list oc rest 
        end
  in
  let ic,oc = getCvcServer () in
  let rec getUsefulPredicates usefulPreds  =
    let rec _worker assertlist =
      match assertlist with 
	[] -> failwith "Exception : Formula is not valid!"
      | a :: rest ->
	  output_string oc (Printf.sprintf "ASSERT %s;\nECHO \"Query %s\";\nQUERY FALSE;\n" a (string_of_int !cvcQueryNum)) ;

	  output_string fdgreedy (Printf.sprintf "ASSERT %s;\nECHO \"Query %s\";\n\n" a (string_of_int !cvcQueryNum)) ;

(*	  output_string fdgreedy (Printf.sprintf "ASSERT %s;\nECHO \"Query %s\";\nQUERY FALSE;\n" a (string_of_int !cvcQueryNum)) ; *)
          incr cvcQueryNum ;
          flush oc ;
          flush fdgreedy ;
	  let result = 
	     try
   	       askCvcIfValid ic
	     with e ->
	       failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
          in
          if result then a
          else
             _worker rest
    in
    output_string oc "PUSH 1;\n" ;
    output_string fdgreedy "PUSH 1;\n" ;
    output_list oc !cvc_ground_term_list ;
    output_asserts oc usefulPreds ;
    output_list fdgreedy !cvc_ground_term_list ;
    output_asserts fdgreedy usefulPreds ;
    output_string oc (Printf.sprintf "ECHO \"Query %s\" ; \n" (string_of_int !cvcQueryNum)) ;
(*    output_string fdgreedy (Printf.sprintf "ECHO \"Query %s\" ; \n" (string_of_int !cvcQueryNum)) ; *)
    incr cvcQueryNum ;
    output_string oc "QUERY FALSE; \n" ;
(*    output_string fdgreedy "QUERY FALSE; \n" ; *)
    flush oc ;
    flush fdgreedy ;
    let result = 
      try
	askCvcIfValid ic
      with e ->
	failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
    in
    if result then usefulPreds 
    else 
      begin
        (* go on adding predicates from assertlist and check for validity.
	   When valid, add the last formula to usefulPreds and repeat
           *)
        let nextpred = _worker all_asserts in
        output_string oc "POPTO 1; \n" ;
        output_string fdgreedy "POPTO 1; \n" ;
        getUsefulPredicates (nextpred :: usefulPreds) 
      end
    in
  let listOfPreds = getUsefulPredicates [] 
  in
  output_string oc "POPTO 1;\n" ;
  output_string fdgreedy "POPTO 1;\n" ;
  flush oc ;
  output_string fdgreedy "ECHO\"Predicates are \" ;\n" ;
  output_list fdgreedy (List.map (fun s -> "ECHO \""^s^"\";\n") listOfPreds) ;
  flush fdgreedy ;
  close_out fdgreedy ;
  listOfPreds

(*    
let filed = open_out "blast-trace.cvc" 

let _ = at_exit (fun () -> close_out filed)
*)
(*    
let queryExpUsingCvcProof exp =
  let _ = Message.msg_string Message.Debug ("querying exp:"^(Predicate.toString (Predicate.convertDNF exp))^"\n") in
  reset_cvc_ground ();
  let formulas = match exp with Predicate.And l -> l | _ -> failwith ("bad call to cvcSat") in
  let all_asserts = List.map convertToCvcSyntax formulas
  in
  let rec output_asserts oc predlist =
    match predlist with 
      [] -> () 
    | a::rest -> 
        begin
          output_string oc (Printf.sprintf "ASSERT %s;\n" a) ; 
          output_asserts oc rest
        end
  in 
  let rec output_list oc lst =
    match lst with 
      [] -> () 
    | a::rest -> 
        begin
          output_string oc (Printf.sprintf "%s\n" a) ; 
          output_list oc rest
        end
  in
  let ic,oc = getCvcServerWithProof () in
  Message.msg_string Message.Normal "Asking CVC (Proof)";
  output_string oc (Printf.sprintf "ECHO \"Query %s\";\nPUSH 1;\n" (string_of_int !cvcQueryNumPf)) ;
  output_string filed (Printf.sprintf "ECHO \"Query %s\";\nPUSH 1;\n" (string_of_int !cvcQueryNumPf)) ;
  output_list oc !cvc_ground_term_list ;
  output_list filed !cvc_ground_term_list ;
  let rec assert_loop asrtlist =
    match asrtlist with
      [] ->
        begin
          output_string oc "\nDUMP_ASSUMPTIONS; QUERY FALSE;\nPOPTO 1;\n" ;
          output_string filed "\nDUMP_ASSUMPTIONS; QUERY FALSE;\nPOPTO 1;\n" ;
          incr cvcQueryNumPf ;
          flush oc ; 
          flush filed ;
          let result = 
	     try
	       askCvcIfValid ic
	     with e ->
	      failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
          in
          reset_cvc_ground () ;
          flush oc;
          result       
        end
    | a::rest ->
       begin 
         output_string oc (Printf.sprintf "ASSERT %s;\n QUERY FALSE;\n" a);
         output_string filed (Printf.sprintf "ASSERT %s;\n QUERY FALSE;\n" a);
         flush oc ;
         let result = 
	     try
	       askCvcIfValid ic
	     with e ->
	       failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
         in
         if result then
           begin
             output_string oc "\nDUMP_ASSUMPTIONS; \nPOPTO 1;\n" ;
             output_string filed "\nDUMP_ASSUMPTIONS; \nPOPTO 1;\n" ;
             incr cvcQueryNumPf ;
             flush oc ; 
             flush filed ;
             reset_cvc_ground () ;
             flush oc;
             result      
           end
         else  
           assert_loop rest
       end
  in
  assert_loop all_asserts
*)
(* 
  output_asserts oc all_asserts ;
  output_list filed !cvc_ground_term_list ;
  output_asserts filed all_asserts ;
  output_string oc "\nDUMP_ASSUMPTIONS; QUERY FALSE;\nPOPTO 1;\n" ;
  output_string filed "\nDUMP_ASSUMPTIONS; QUERY FALSE;\nPOPTO 1;\n" ;
  incr cvcQueryNumPf ;
  flush oc ; 
  flush filed ;
  let result = 
	     try
	       askCvcIfValid ic
	     with e ->
	      failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
  in
  reset_cvc_ground () ;
  flush oc;
  result       
*)

(*	     
let queryExpUsingCvc exp =
  let result =	    
    queryExpUsingCvcProof exp
  in
  if result then 
      begin
        Message.msg_string Message.Normal "Doing Harald's method." ;
        ignore (queryExpUsingCvcGreedy exp) ;
	result
      end
    else 
    failwith "CVC Failed"
*)


(* This implements the greedy algorithm (SLAM, Ruess, ...) to get back a list
   of useful predicates.
*)
exception CVCException 

let getProofsFromCvc exp =
  let _ = Message.msg_string Message.Debug ("querying exp in getProofsFromCvc :"^(Predicate.toString exp)^"\n") in
   (* first we have to convert the formula and collect all the ground terms that appear in it *)
  try  
  reset_cvc_ground ();
  let formulas = 
    match exp with 
      Predicate.And l -> l 
    | _ -> failwith ("bad call to getProofsFromCvc -- we only handle listd of conjunections") in
  let all_blast_asserts_array = Array.of_list formulas in
  let all_asserts = List.map (convertToCvcSyntax) formulas
  in
	     (* at this point all_asserts is a list of all assertions, and !cvc_ground_term_list is
	        the list of all type declarations.
	     *)
  let rec output_asserts oc predlist =
    match predlist with 
      [] -> () 
    | a::rest -> 
	begin
          output_string oc (Printf.sprintf "ASSERT %s;\n" a) ; 
          output_asserts oc rest 
	end
  in 
  let rec output_list oc lst =
    match lst with 
      [] -> () 
    | a::rest -> 
        begin
          output_string oc (Printf.sprintf "%s\n" a) ; 
          output_list oc rest 
        end
  in
  let ic,oc = getCvcServer () in
  let rec usefulPredicates usefulPreds  usefulBlastPreds =
    let rec _worker assertlist num =
      match assertlist with 
	[] -> 
	  begin
	    Message.msg_string Message.Error "Exception : Formula is not valid! Reverting to Vampyre" ;
	    raise CVCException 
	  end
      | a :: rest ->
	  output_string oc (Printf.sprintf "ASSERT %s;\nECHO \"Query %s\";\nQUERY FALSE;\n" a (string_of_int !cvcQueryNum)) ;
	  incr cvcQueryNum ;
          flush oc ;
	  let result = 
	     try
   	       askCvcIfValid ic
	     with e ->
	       failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
          in
          if result then (a, num)
          else
             _worker rest (num+1)
    in
    output_string oc "PUSH 1;\n" ;
    output_list oc !cvc_ground_term_list ;
    output_asserts oc usefulPreds ;
    output_string oc (Printf.sprintf "ECHO \"Query %s\" ; \n" (string_of_int !cvcQueryNum)) ;
    incr cvcQueryNum ;
    output_string oc "QUERY FALSE; \n" ;
    flush oc ;
    let result = 
      try
	askCvcIfValid ic
      with e ->
	failwith ("Cvc raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
    in
    if result then usefulBlastPreds 
    else 
      begin
        (* go on adding predicates from assertlist and check for validity.
	   When valid, add the last formula to usefulPreds and repeat
           *)
        let (nextpred, i) = _worker all_asserts 0 in
        output_string oc "POPTO 1; \n" ;
        usefulPredicates (nextpred :: usefulPreds) ((all_blast_asserts_array.(i)) :: usefulBlastPreds)
      end
    in
  let listOfPreds = usefulPredicates [] []
  in
  output_string oc "POPTO 1;\n" ;
  flush oc ;
  List.map 
    (fun a -> match a with Predicate.Not b -> b | _ -> a)
    listOfPreds
with _ ->
	    begin
	      let _eToCheck = Predicate.implies exp (Predicate.False) in
              let ans = queryExpUsingVampyre _eToCheck
              in
              let _ = Message.msg_string Message.Debug "Got UsefulPred" in
              if ans = false
              then
		begin
		  failwith "Vampyre could not prove formula valid!" ; []
		end
              else
                begin
		  let _lst = Engine.getUsefulPredicates () in
            (*
	      let inferred_atoms = 
	      match (!proof) with 
	      Some(prf) -> 
	      Engine.getInferredAtoms (prf)
	      | None -> []
	      in
	      *)
		  List.map convertToPredicate _lst
		end
	    end

(*************************************************************************************)
(* Interface to Simplify.                                                            *)
(* Simplify is run as a server and queries are piped to it.                          *)
(* Thanks to Wes Weimer for the code.                                                *)
(*************************************************************************************)

(* Return a "\n" terminated string corresponding to simplify input *)
let clear_varHashtbl_flag = ref true (* the table will get cleared only if this flag is true *)
		    (* moreover -- when simplify_assume is called the flag gets set to false *)
		    (* when pop is called, the flag is again set to true *)

let varHashtbl = Hashtbl.create 53 
let next_var = ref 0 
let chlvalHashtbl = Hashtbl.create 53 

let convertToSimplifySyntax () exp =
(*Simplify has issues with variable names starting with "_", so we map all vars to v<num> *)
 
  let get_next_var () = next_var := !next_var + 1; !next_var in
  let get_current_var () = !next_var in
  if (!clear_varHashtbl_flag) then 
    begin
      Hashtbl.clear varHashtbl;next_var :=0; 
      Hashtbl.clear chlvalHashtbl
    end ;
  let fieldId str =
    if Hashtbl.mem varHashtbl str then
      Hashtbl.find varHashtbl str
    else
      begin
	let new_id = "v"^(string_of_int (get_next_var ())) in
	Hashtbl.add varHashtbl str  new_id;
	new_id
      end
  in
  let id_of_var lv =
    match lv with
      Expression.Symbol s -> 
	begin
	  try
	    let vid = Hashtbl.find varHashtbl s in
	    int_of_string (Misc.chop_after_prefix "v" vid)
	  with 
	    Not_found -> failwith ("id_of_var: (Simplify) variable not found "^(s))
	end
	  
    | _ -> failwith ("id_of_var: (Simplify) lvalue not a symbol. "^(Expression.lvalToString lv))
  in
  let id_of_fld s =
    
    let vid = try Hashtbl.find varHashtbl s 
    with Not_found -> failwith ("id_of_fld: (Simplify) variable not found "^(s))	
    in
    int_of_string (Misc.chop_after_prefix "v" vid)
	
  in
  let rec convertExp e =
    match e with
    | Expression.Lval lv -> convertLval lv
    | Expression.Chlval (lv,s)  -> (* Printf.sprintf "(chlval %s %s)" (convertLval lv) (fieldId s) *)
	  begin
	    try (Hashtbl.find chlvalHashtbl e) 
	    with 
	      Not_found ->
		begin
		  let s_e = get_next_var () (* Expression.toString e*) in
        
		  let new_id = "chlval"^(string_of_int s_e) in
		  Hashtbl.add chlvalHashtbl e new_id;
		  new_id
		end
	  end
    | Expression.Binary (op, e1, e2) -> 
	begin
	  match op with 
	    Expression.Minus -> Printf.sprintf "(- %s %s)" (convertExp e1) (convertExp e2)
	  |	Expression.Plus  -> Printf.sprintf "(+ %s %s)" (convertExp e1) (convertExp e2)
		  (* |     Expression.Ordset_Succ ->  Printf.sprintf "(SUCC %s)" (convertExp e1)*)
	  |	Expression.Mul   -> Printf.sprintf "(* %s %s)" (convertExp e1) (convertExp e2)
	  |	Expression.BitAnd 
	  |	Expression.BitOr 
	  |	Expression.Div 
	  |	Expression.LShift 
	  |	Expression.Rem 
	  |	Expression.RShift 
	  |	Expression.Xor 
	  |	Expression.Comma ->
              Printf.sprintf "(%s %s %s)" (Expression.binaryOpToVanillaString op) (convertExp e1) (convertExp e2)
	  |     Expression.FieldOffset ->
	      (* the second argument to field offset is a field id *)
	      let fld =
		match e2 with
		  Expression.Lval (Expression.Symbol s) -> s
		| _ -> failwith ("convertExp : (Simplify) Second operator of field offset not a field id "^
				 (Expression.toString e))
	      in
	      Printf.sprintf "(foffset %s %s)" (convertExp  e1) (fieldId fld)
	  |	_ -> failwith ("convertExp: (Simplify) This operator "^(Expression.binaryOpToVanillaString op)^" should not be handled here "^(Expression.toString e))
	  | _ -> failwith "foo"
	end
    | Expression.Constant (c) -> 
	begin
	  match c with
	    Constant.Int i ->  if (i < 0) then "(- 0 "^(string_of_int (-i))^")" else string_of_int i 
(*
	  | Constant.String s -> ("CONSTANTSTRING"^s)
*)
	  | Constant.String s -> ("CONSTANTSTRING"^(string_of_int (get_next_var ())))
	  | Constant.Float f -> 
	      if (f -. (ceil f) = 0.0) then string_of_int (int_of_float f)
	      else ("FLOATINGPT"^(string_of_int (get_next_var ())))
	  | _ -> failwith ("Cannot handle constant "^(Constant.toString c))
	end

    | Expression.Unary (uop, e1) -> 
	begin
	  match uop with 
	    Expression.UnaryMinus -> "(- 0 "^(convertExp e1)^")"
	  | Expression.UnaryPlus -> convertExp e1
	  | Expression.Address -> 
	      begin
		match e1 with
		  Expression.Lval lv -> 
		    let l = convertLval lv
		    in
		    let id = (id_of_var lv) in
		    Printf.sprintf "(addrOf %s %d)" l id
		| Expression.Chlval (lv, f) -> 
		    (* let lstr = (convertLval lv) in
		    let fstr = (fieldId f) in
		    let id = (id_of_fld f) in
		       *)
		    let chlv = convertExp e1 in
		    let id = int_of_string (Misc.chop_after_prefix "chlval" chlv) in
		    Printf.sprintf "(addrOf %s %d)" chlv id
		      
		| _ -> failwith ("convertExp: (Simplify) Address: strange expression "^(Expression.toString e))
	      end
(*
	      let v = convertExp e in
	      "(addrOf "^v^")"
*)
	  |	_ -> "("^Expression.unaryOpToVanillaString uop^" "^(convertExp e)^")"
	end
    | _ -> failwith ("convertExp (Simplify): " ^ (Expression.toString e) ^ "not handled")
	 
  and convertLval l =
    match l with
    | Expression.Symbol(str) -> 
	if Hashtbl.mem varHashtbl str then
	  (* (Message.msg_string Message.Debug ("In hasthbl:"^str^":"^(Hashtbl.find varHashtbl str)); Delete this line asap *)
	  (Hashtbl.find varHashtbl str)
	else
	  begin
	    let new_id = "v"^(string_of_int (get_next_var ())) in
	    Hashtbl.add varHashtbl str new_id;
	    (*Message.msg_string Message.Debug ("NOT in hashtbl:"^str^":"^new_id);*)
	    new_id
	  end
    | Expression.Dereference drfe  -> "(select "^(convertExp drfe)^" 0)"

    | Expression.Access(Expression.Dot, e, s) ->
(* 
	(Printf.sprintf "(select_dot %s %s)" (convertExp e) (fieldId s))
*)
	     (Printf.sprintf "(select %s %s)" (convertExp e) (fieldId s))
    | Expression.Access(Expression.Arrow, e, s) ->
	let fieldId str =
	  if Hashtbl.mem varHashtbl str then
	    Hashtbl.find varHashtbl str
	  else
	    begin
	      let new_id = "v"^(string_of_int (get_next_var ())) in
	      Hashtbl.add varHashtbl str new_id;
	      new_id
	    end
	in
	(Printf.sprintf "(select %s %s)" (convertExp e) (fieldId s))
	  (* JHALA: MAJOR BUG!! this should not be (+ e1 e2) surely *)
    | Expression.Indexed (e1, e2) -> Printf.sprintf "(select %s %s)" (convertExp e1) (convertExp e2)

  and convertAtom e =   
    match e with
      Expression.Binary (op, e1, e2) -> 
	begin
	  let opstring = 
	    match op with 
		Expression.Eq  -> "EQ"
	      |	Expression.Ne  -> "NEQ"
	      |	Expression.Gt  -> ">"
	      |	Expression.Lt  -> "<"
	      |	Expression.Ge  -> ">="
	      |	Expression.Le  -> "<="
	      |	_   -> "" (**** see below ** failwith ("convertAtom: Cannot handle this expression: "^(Expression.toString e)) *)
	  in
             if opstring <> "" then
	      Printf.sprintf "(%s %s %s)" opstring (convertExp e1) (convertExp e2)
	    else (Message.msg_string Message.Error "failed opthing";convertExp e)
	end
    | _ -> (Message.msg_string Message.Error "failed opthing2";convertExp e)
	(*** instead i shall simply call convert expression on it *** failwith ("convertAtom: "^(Expression.toString e)^ " Not a binary relation operator") *)
  
  and convertPred e =
    match e with
      Predicate.True -> "TRUE"
    | Predicate.False -> "FALSE"
    | Predicate.And plist -> 
	List.fold_left (function a -> function b -> "(AND "^(convertPred b)^" "^a^")") 
	  "TRUE" (plist)
    | Predicate.Or plist -> 
	List.fold_left (function a -> function b -> "(OR "^(convertPred b)^" "^a^")") 
	  "FALSE" (plist)
    | Predicate.Implies (e1, e2) -> "(IMPLIES "^(convertPred e1)^" "^(convertPred e2)^")"
    | Predicate.Not e1 -> "(NOT "^(convertPred e1)^")"
    | Predicate.Atom e ->  (convertAtom e)
    | _ -> failwith ("Cannot handle this predicate "^(Predicate.toString e))
  in
  (convertPred exp)^"\n"

(* keep around info: do we have a simplify server (= other process) 
 * running? *)
let simplifyServer = ref None
let simplifyServerCount = ref 0 
let simplifyCacheCount = ref 0 

let simplify_query_flag = ref false
let set_simplify_query_flag () = 
  simplify_query_flag := true

let fixed_simplify_axioms = ref 
	     ["(DEFPRED (BitAnd x y))\n" ;
	     "(DEFPRED (BitOr x y))\n" ;
	     "(DEFPRED (BitNot x))\n" ;
	     "(DEFPRED (addrOf x y))\n" ;
	     "(DEFPRED (select y x))\n" ;
	     "(DEFPRED (foffset x f))\n" ;
	     "(BG_PUSH (FORALL (x y z) (IMPLIES (EQ z (BitOr x y)) (NEQ (BitAnd z y) 0))))\n" ;
	     "(BG_PUSH (FORALL (x y) (EQ (select (addrOf x y) 0) x)))" ;
	     "(BG_PUSH (FORALL (x y d1) (IMPLIES (EQ (foffset x d1) (foffset y d1)) (EQ  x y))))\n"
(*
	     "(BG_PUSH (FORALL (x y z) (IMPLIES (EQ z (addrOf x)) (EQ (select z 0) x))))\n" ;
	     "(BG_PUSH (FORALL (x) (EQ (addrOf (select x 0)) x)))" ;
	     "(BG_PUSH (FORALL (x y d1 d2) (IMPLIES (EQ (addrOf (select x d1)) (addrOf (select y d1))) (EQ  (addrOf x) (addrOf y)))))\n"
*)
	     ]

let rec addAxioms axiom_list =
  Message.msg_string Message.Minor "(Adding axioms)" ;
  let oc  = match (!simplifyServer) with
      None -> snd (getServer ()) 
    | Some(_,oc1) -> oc1
  in
    List.iter 
      (function x -> (Message.msg_string Message.Normal (x^"\n"); 
		      output_string oc (x^"\n")))
      axiom_list
  and  getServer () = 
    match !simplifyServer with
       None -> begin
	  Message.msg_string Message.Normal "Forking Simplify process..." ;
	  let ic,oc,ec = Unix.open_process_full "Simplify -nosc" 
			   (Unix.environment ()) in
	    simplifyServer := Some(ic,oc) ;
    (* add axioms to simplify  -- assume axioms are in $BLASTHOME/bin/axioms.in *)
	    addAxioms (!fixed_simplify_axioms);
	    Message.msg_string Message.Normal "done!\n";
	    flush stdout ;
	    at_exit (fun () -> 
		       ignore (Unix.close_process_full (ic,oc,ec) ) ) ;
	    (ic,oc)
	end
      | Some(a) -> a

	
let simplify_assume () pred = 
  let simplifyInput = VampyreStats.time "convertToSimplifySyntax" (convertToSimplifySyntax ()) pred
  in 
    clear_varHashtbl_flag := false;
    Message.msg_string Message.Debug ("Simplify Assuming: "^simplifyInput);
    flush stdout;
    flush stderr;
    let (_,oc) = getServer () in
    let s = ("(BG_PUSH "^simplifyInput^" )\n") in
      output_string oc s ;
      if (!simplify_query_flag) then Misc.append_to_file "simp_query" s;
      flush oc ;
      ()

let simplify_pop () = 
  Message.msg_string Message.Debug ("Simplify Popping");
  let (_,oc) = getServer () in
  let s = "(BG_POP)\n" in
  output_string oc s;
    if (!simplify_query_flag) then Misc.append_to_file "simp_query" s;
  flush oc;
  clear_varHashtbl_flag := true;
  ()

(* doesnt clear the varHashtbl *)
let simplify_pop_noflush () = 
  Message.msg_string Message.Debug ("Simplify Popping");
  let (_,oc) = getServer () in
  let s = "(BG_POP)\n" in
    output_string oc s;
    if (!simplify_query_flag) then Misc.append_to_file "simp_query" s;
 flush oc;
     ()
 
 (*
  let lengthOfAxiomsList = 7 in
  let axioms = Array.make lengthOfAxiomsList "" in
  Array.set axioms 0 "(DEFPRED (BitAnd x y))\n" ;
  Array.set axioms 1 "(DEFPRED (BitOr x y))\n" ;
  Array.set axioms 2 "(DEFPRED (BitNot x))\n" ;
  Array.set axioms 3 "(DEFPRED (addrOf x))\n" ;
  Array.set axioms 4 "(DEFPRED (select y x))\n" ;
  Array.set axioms 5 "(BG_PUSH (FORALL (x y z) (IMPLIES (EQ z (BitOr x y)) (NEQ (BitAnd z y) 0))))\n" ;
  Array.set axioms 6 "(BG_PUSH (FORALL (x y z) (IMPLIES (EQ z (addrOf x)) (EQ (select z 0) x))))\n"; 
  begin
    for i = 0 to lengthOfAxiomsList -1 do
      output_string oc axioms.(i) 
    done ;
    output_string oc "\n" ;  
  end
    *)


let rec isValid ic =
  let line = input_line ic in
  let result = if String.contains line 'V' then 
    true
  else if String.contains line 'I' then
    false
  else isValid ic
  in
    result

let simplify_is_contra () =
  Message.msg_string Message.Debug "simplify check contra";
  let ic,oc = getServer () in
  let s = "FALSE\n"  in
    output_string oc s ;
    flush oc ;
    if (!simplify_query_flag) then Misc.append_to_file "simp_query" s;
    try
      let rv = isValid ic in
      Message.msg_string Message.Debug ("Contra ? "^(string_of_bool rv));
	rv
    with e ->
      failwith ("Simplify raised exception "^(Printexc.to_string e)^" in simplify_is_contra")
   
      
let queryExpUsingSimplify exp =
  let _ = Message.msg_string Message.Debug ("querying exp:"^(Predicate.toString (Predicate.convertDNF exp))^"\n") in
  let simplifyInput = VampyreStats.time "convertToSimplifySyntax" (convertToSimplifySyntax ()) exp
  in 
  Message.msg_string Message.Debug ("Simplify: "^simplifyInput);
  flush stdout;
  flush stderr;
  let ic,oc = getServer () in
  output_string oc (simplifyInput) ;
    if (!simplify_query_flag) then Misc.append_to_file "simp_query" simplifyInput;
  flush oc ;
  let result = try
    isValid ic
  with e ->
    failwith ("Simplify raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
  in
  Message.msg_string Message.Debug (if result then "true" else "false");
  result




	       
let getProofsFromSimplify exp =
  let _ = Message.msg_string Message.Debug("querying exp in getProofsFromSimplify :"^(Predicate.toString exp)^"\n") in
   (* first we have to convert the formula and collect all the ground terms that appear in it *)
  let formulas = 
    match exp with 
      Predicate.And l -> l 
    | _ -> failwith ("bad call to getProofsFromSimplify -- we only handle listd of conjunections") in
  let ic,oc = getServer () in
  let counter = ref 0 in
  let output_asserts oc lst =
    List.iter (fun pred -> simplify_assume () pred; incr counter) lst
  in
  let rec usefulPredicates usefulPreds  =
try
    let rec _worker assertlist =
      match assertlist with 
	[] -> 
	  begin
	    Message.msg_string Message.Error "Exception : Formula is not valid! Simplify cannot prove validity." ;
	    Message.msg_string Message.Debug "Exception : Formula is not valid! Simplify cannot prove validity." ;
	    Message.msg_string Message.Debug "\nThe formula was :" ;
	    Message.msg_string Message.Debug (Predicate.toString exp) ;
	     
	    raise Not_found
	  end
      | a :: rest ->
	  simplify_assume () a ;
          incr counter ;
	  output_string oc "FALSE\n" ;
          flush oc ;
	  if (!simplify_query_flag) then Misc.append_to_file "simp_query" "FALSE\n";
	  let result = 
	     try
   	       isValid ic
	     with e ->
	       failwith ("Simplify raised exception "^(Printexc.to_string e)^" for "^(Predicate.toString exp))
          in
          if result then a
          else
             _worker rest 
    in
      begin
        Message.msg_string Message.Debug "In useFulPredicates, List is " ;
	List.iter (fun p -> Message.msg_string Message.Debug (Predicate.toString p)) usefulPreds ;
        Message.msg_string Message.Debug "end List. " ;

        output_asserts oc usefulPreds ;
        output_string oc "FALSE\n" ;
        flush oc ; 
	  if (!simplify_query_flag) then Misc.append_to_file "simp_query" "FALSE\n";

        Message.msg_string Message.Debug "Asserted useful predicates. " ;

        let result = try isValid ic with e -> failwith "Simplify raised exception" in
        if result then 
	  begin
            while !counter > 0 do simplify_pop () ; decr counter done ;
            usefulPreds 
          end
        else 
          begin 
           (* go on adding predicates from assertlist and check for validity.
	      When valid, add the last formula to usefulPreds and repeat
           *)
           let nextpred = _worker formulas  in
           while !counter > 0 do
             simplify_pop () ; decr counter 
           done ;
           usefulPredicates (nextpred :: usefulPreds) 
          end
      end
with
 Not_found -> begin (* failwith "simplify" ;*) while !counter >0 do simplify_pop (); decr counter done; formulas end
  in
  let listOfPreds = usefulPredicates [] 
  in
  (* iterate simplify pop to get rid of assumptions *)
  List.map 
    (fun a -> match a with Predicate.Not b -> b | _ -> a)
    listOfPreds

      

(*************************************************************************************)
let tp = ref ""
let simplify_fun () e = queryExpUsingSimplify e 

let rec queryExp exp =

  Message.msg_string Message.Debug ("Query : "^(Predicate.toString exp)); 
  if (!tp = "ics") then
    begin
      Message.msg_string Message.Debug ("Query is "^(Predicate.toString exp)^"\n");
      let icsans =  queryExpUsingIcs exp and
	  vamans =  queryExpUsingVampyre exp 
      in
      Message.msg_string Message.Debug ("Vampyre says "^(if vamans then "true" else "false")^"\n");
      Message.msg_string Message.Debug ("Ics says "^(if icsans then "true" else "false")^"\n");
      if (not(vamans = icsans)) then 
	failwith ("TheoremProvers disagree on "^(Predicate.toString exp)^"!! Vampyre="^(if vamans then "true" else "false")^", Ics ="^(if icsans then "true" else "false"))
      else 
	vamans
    end
  else 
    if (!tp = "vampyre") then
      queryExpUsingVampyre exp
    else
      if (!tp = "simplify") then
	(VampyreStats.time "simplify querytime" (queryExpUsingSimplify) exp) 
      else
	begin
	  tp :=  Options.getValueOfString "s";
	  if (!tp = "ics" || !tp = "vampyre" || !tp = "simplify") then
	    queryExp exp
	  else
	    failwith ("TheoremProver: queryExp: Unknown theorem prover "^(!tp)^". Options are ics, simplify, and vampyre.")
	end
	
let queryExpContext query =
  if (!tp = "") then tp := "simplify";
  if (!tp = "simplify") then
    (VampyreStats.time "simplify querytime" (queryExpUsingSimplify) query) 
  else
    begin 
      let vam_query = Input.pushNeg (convertToVampyreSyntax query) in
      vampyre_prove_context vam_query
    end

(* get the hash table of theorems. For those theorems that are true,
   get proofs from Vampyre and put in another hash table. Finally Marshal
   the hashtable and write out to disk.
*)  
let generateProofsSimple htbl =
  let pfHt = Hashtbl.create 511 in
  let genPf thm ans =
    if (ans = true) then
      let pf = 
	try
	  let vampyreInput = convertToVampyreSyntax thm
	  in
	  try
	    vampyreProver "foo" (Input.pushNeg vampyreInput)
	  with Engine.Failure ->((Message.msg_string Message.Error "Bug in Vampyre.\n") ; failwith "Vampyre bug in proofgen")
	with
	  VampyreErrormsg.Error -> (failwith ("Bug in Vampyre ...Proof gen"))
      in
      Hashtbl.add pfHt thm pf
  in
  Hashtbl.iter genPf htbl ;
  let pfString = Marshal.to_string pfHt [] in
  Message.msg_string Message.Error ("Proof size is "^(string_of_int (String.length pfString)))


(***********************************************************************************)
(* Jhala: Adding the proof functions ... *)
(***********************************************************************************)


 let queryExpUsingVampyre_proof exp =
   Message.msg_string Message.Debug "in queryExpUsingVampyre_proof";
   let correct_ans = queryExpUsingSimplify exp in
   let vampyreInput = convertToVampyreSyntax exp in
   let vampyreInput_pushneg = (Input.pushNeg  vampyreInput) in  
     try
       let proof = vampyreProver "foo" (vampyreInput_pushneg)
       in
	 if (not correct_ans) then 
	   begin
	     Message.msg_string Message.Debug ("PROOF ERROR: TP MISMATCH + !: "^(Predicate.toString exp)); 
	     (*VampyrePretty.printf "Vampyre Saw: %a \n" Input.dp vampyreInput_pushneg;
	     VampyrePretty.printf "Vampyre Proof: %a \n" Proof.dp proof;*)
	     None
	end
      else Some(proof)
    with Engine.Failure -> 
      begin
	(Message.msg_string Message.Debug "Proof failed.\n") ; if (correct_ans) then 
	  begin
	    Message.msg_string Message.Debug ("PROOF ERROR: TP MISMATCH - !: "^(Predicate.toString exp)); 
            (* VampyrePretty.printf "Vampyre Saw: %a \n" Input.dp vampyreInput_pushneg; *)
	    Some (Proof.Unimplemented)
	  end
	else None
      end
      | VampyreErrormsg.Error -> 
	  begin
	    Message.msg_string Message.Debug ("PROOF ERROR: BUG IN VAMPYRE !: "^(Predicate.toString exp)); 
	    VampyrePretty.printf "Vampyre Saw: %a \n" Input.dp vampyreInput_pushneg;
	    let backup_ans = queryExpUsingSimplify exp in
	      if backup_ans then Some (Proof.Unimplemented) else None
	  end
 (* (failwith ("Bug in Vampyre ...")) *)

let queryExp_proof exp = 
  Message.msg_string Message.Debug ("Query _proof : "^(Predicate.toString exp)); 
  queryExpUsingVampyre_proof exp
  
let queryExpContext_proof query = 
  let vam_query = Input.pushNeg (convertToVampyreSyntax query) in
    vampyre_prove_context_proof vam_query
 
(****************************************************************************)
(************************  interface to UCLID *******************************)
(****************************************************************************)
let uclid_ground_term_list = ref [] 

let add_uclid_ground str = 
  if List.mem str (!uclid_ground_term_list) 
  then () 
  else uclid_ground_term_list := str::(!uclid_ground_term_list)
    
let reset_uclid_ground () = 
  uclid_ground_term_list := 
  ["zero : TERM;";
   "select : FUNC[2];"; 
   "select_dot : FUNC[2];";
   "guf : FUNC[2];";
   "gur : PRED[2];";
   "guf_plus : FUNC[2];";
   "guf_minus : FUNC[2];";
  ]
    
let convertToUclidSyntax exp =
(*Simplify has issues with variable names starting with "_", so we map all vars to v<num> *)
  let get_next_var () = next_var := !next_var + 1; !next_var in
  let _ = if (false) then (Hashtbl.clear varHashtbl;next_var :=0) in
  let rec convertExp e =
    match e with
	Expression.Binary (op, e1, e2) -> 
	  begin
	    match op with 
		Expression.Minus -> 
		  begin
		    match e2 with
	                Expression.Constant (Constant.Int 0) -> (convertExp e1)
		      |	Expression.Constant (Constant.Int i) -> 
			  if (i > 0) then Printf.sprintf "(pred^%s(%s))" (string_of_int i) (convertExp e1)
			  else Printf.sprintf "(succ^%s(%s))" (string_of_int (-i)) (convertExp e1)
		      | _ -> Printf.sprintf "guf_minus(%s,%s)" (convertExp e1) (convertExp e2)
		  end
	      |	Expression.Plus  -> 
		  begin
		    match e2 with
                        Expression.Constant (Constant.Int 0) -> (convertExp e1)
		      | Expression.Constant (Constant.Int i) -> 
			  if (i > 0) then Printf.sprintf "(succ^%s(%s))" (string_of_int i) (convertExp e1)
			  else Printf.sprintf "(pred^%s(%s))" (string_of_int (-i)) (convertExp e1)
		      | _ -> Printf.sprintf "guf_plus(%s,%s)" (convertExp e1) (convertExp e2)
		  end
              
	      |	Expression.Mul   
	      |	Expression.BitAnd 
	      |	Expression.BitOr 
	      |	Expression.Div 
	      |	Expression.LShift 
	      |	Expression.Rem 
	      |	Expression.RShift 
	      |	Expression.Xor 
	      |	Expression.Comma ->
	        Printf.sprintf "guf(%s,%s)" (convertExp e1) (convertExp e2)
	 
	      |	_ -> failwith ("convertExp: (Uclid) This operator should not be handled here "^(Expression.toString e))
	  end
      | Expression.Constant (c) -> 
	  begin
	    match c with
		Constant.Int i ->  
	          if (i = 0) then "zero" 
	          else
		    if (i < 0) then Printf.sprintf "(pred^%s(zero)" (string_of_int (-i)) 
		    else Printf.sprintf "succ^%s(zero)" (string_of_int i) 
	      | Constant.String s -> add_uclid_ground (s^" : TERM;");(s) 
	      | _ -> failwith ("unknown constant type")
	  end

      | Expression.Unary (uop, e) -> 
	  begin
	    match uop with 
		Expression.UnaryMinus -> failwith "cannot handle uminus"
	      | Expression.UnaryPlus -> failwith "cannot handle uplus"
	      | Expression.Address -> failwith "cannot handle address"
	      |	_ -> failwith "unknown unop" 
	  end
      | _ -> failwith ("convertExp: " ^ (Expression.toString e) ^ "not handled")

  and convertLval l =
    match l with
      | Expression.Symbol(str) -> 
	     begin
	     (if Misc.is_prefix "b_p" str then add_uclid_ground (str^" : TRUTH;") else add_uclid_ground (str^" : TERM;"));
	     str 
	     end
      | Expression.Access(Expression.Dot, e, s) -> 
	  begin
	    add_uclid_ground (s ^" : TERM;");
	    (Printf.sprintf "(select_dot(%s,%s))" (convertExp e) s)
	  end
      | Expression.Access(Expression.Arrow, e, s) ->
	  begin
	    add_uclid_ground (s ^" : TERM;"); 
	    (Printf.sprintf "(select_dot(%s,%s))" (convertExp e) s)
	  end
      | Expression.Dereference e -> 
	     begin
	       match e with
	         Expression.Binary(Expression.Plus,e1,e2) -> Printf.sprintf "(select(%s,%s))" (convertExp e1) (convertExp e2)
	       | _ -> Printf.sprintf "(select(%s,zero))" (convertExp e)
	     end
      | Expression.Indexed (e1, e2) ->
	     begin
	       Printf.sprintf "(select (%s, %s))" (convertExp e1) (convertExp e2)
	     end
  and convertAtom e =
    match e with
	Expression.Binary (op, e1, e2) -> 
	  begin
	    let opstring = 
	      match op with 
		  Expression.Eq  -> "="
		| Expression.Ne  -> "!="
		| Expression.Gt  -> ">"
		| Expression.Lt  -> "<"
		| Expression.Ge  -> ">="
		| Expression.Le  -> "<="
		| _   -> "" 
	    in
              if not (opstring = "") then
		Printf.sprintf "(%s %s %s)" (convertExp e1) opstring (convertExp e2)
	      else (convertExp e)
	  end
    | _ -> convertExp e 
  
and convertPred e =
    match e with
	Predicate.True -> "1"
      | Predicate.False -> "0"
      | Predicate.And plist -> 
	    List.fold_left 
	      (fun a -> fun b -> Printf.sprintf " (%s & %s) " (convertPred b) a) 
	      "1" (plist)
      | Predicate.Or plist -> 
	  List.fold_left 
	    (function a -> function b -> Printf.sprintf "(%s | %s)" (convertPred b) a )
	    "0" (plist)
      | Predicate.Implies (e1, e2) -> Printf.sprintf "(%s => %s)" (convertPred e1) (convertPred e2)
      (*| Predicate.Iff (e1, e2) -> Printf.sprintf "(%s <=> %s)" (convertPred e1) (convertPred e2)*)

      | Predicate.Not e1 -> Printf.sprintf "(~ %s)" (convertPred e1)
      | Predicate.Atom e ->  (convertAtom e)
      | _ -> failwith ("Cannot handle this predicate "^(Predicate.toString e))
  
      in
	(convertPred exp)


          (* this function checks if the formula given below is satisfiable -- 
	     right now it just dumps a trace into a file called blast-trace.ucl *)

let uclidSat fmla = 
   (* first we have to convert the formula and collect all the ground terms that appear in it *)
   reset_uclid_ground ();
   
   let block_counter = ref 0 in     
   let block_temp_list = ref [] in
   let formula_const _fmla = 
       block_counter := !block_counter + 1;
       block_temp_list := ("blast_temp_"^(string_of_int !block_counter))::(!block_temp_list);
       Printf.sprintf "blast_temp_%s := %s; \n" (string_of_int !block_counter) (convertToUclidSyntax _fmla)
   in
   let formulas = match fmla with Predicate.And l -> l | _ -> failwith ("bad call to uclidSat") in
   
   let def_string = String.concat "\n" (List.map formula_const formulas) in	     
   let root_string = " \n root := "^(String.concat " & " (!block_temp_list))^";\n" in
   let const_def_string = List.fold_left (function a -> function b -> Printf.sprintf "%s \n  %s" b a )
	  "" !uclid_ground_term_list
   in
   let uclid_string = def_string^root_string in
   let send_to_file = 
       Printf.sprintf "MODEL blasttrace \n CONTROL \n EXTVAR \n STOREVAR \n VAR \n CONST \n %s \n DEFINE %s \n EXEC \n decide(~root); \n" const_def_string uclid_string 
   in 
   Misc.write_to_file "blast-trace.ucl" send_to_file;
   let uclidret = Sys.command "runuclid blast-trace.ucl" in
     (*if uclidret = 127 then failwith "uclid is NOT in the path or something!"
     else*)
     begin
       Message.msg_string Message.Debug "running smvsat...";
       let smvsat_ret =  Sys.command "smvsat blastrace.cnf foobaz" in
	 if (smvsat_ret <> 1) then failwith "issues with smvsat" 
	 else
	   begin
	     Message.msg_string Message.Debug "...smvsat finished";
	     
	     Sys.command "grep \"b\_p\" blasttrace.map > tmp-blast-bpvars";
	     Sys.command "cut -d ' ' -f 1 ztrace.variables > tmp-blast-pfvars";
	     let useful_vars = List.map (List.hd) (Misc.grep_file "tmp-blast-bpvars" "tmp-blast-pfvars") in
	       useful_vars
	   end 
       end

let uclid_output modelname  fmla comment_string = 
  (* first we have to convert the formula and collect all the ground terms that appear in it *)
   reset_uclid_ground ();
   let block_counter = ref 0 in     
   let block_temp_list = ref [] in
   let formula_const _fmla = 
       block_counter := !block_counter + 1;
       block_temp_list := ("blast_temp_"^(string_of_int !block_counter))::(!block_temp_list);
       Printf.sprintf "blast_temp_%s := %s; \n" (string_of_int !block_counter) (convertToUclidSyntax _fmla)
   in
   let formulas = [fmla] (* match fmla with Predicate.And l -> l | _ -> failwith ("bad call to uclidSat:"^(Predicate.toString fmla)) *) in
   
   let def_string = String.concat "\n" (List.map formula_const formulas) in	     
   let root_string = " \n root := "^(String.concat " & " (!block_temp_list))^";\n" in
   let const_def_string = List.fold_left (function a -> function b -> Printf.sprintf "%s \n  %s" b a )
	  "" !uclid_ground_term_list
   in
   let uclid_string = def_string^root_string in
   let send_to_file = 
       Printf.sprintf "MODEL %s \n CONTROL \n EXTVAR \n STOREVAR \n VAR \n CONST \n %s \n DEFINE %s \n EXEC \n decide(~root); \n (* %s *)" modelname const_def_string uclid_string comment_string 
   in 
   Misc.write_to_file (modelname^".ucl") send_to_file

(*****************************************************************)
(** Implement a simple "predicate flow analysis" on the useful   *)
(*  predicates. This does a congruence closure on the variables  *)
(*  in the set of useful predicates and returns extra predicates *)
(*  that may be useful.                                          *)
(*****************************************************************)

let scope_pred_table = Hashtbl.create 37

let predicate_closure predlist = 
  let cong_classes = ref [] in
  let new_predicates = ref [] in
  let set_of_elt e = (* puts e in a new set if its not anywhere and if it is then removes it*)
    let _temp_set = List.filter (List.mem e) (!cong_classes) in
    let _cong_classes = List.filter (fun x -> not (List.mem e x)) (!cong_classes) in
    cong_classes := _cong_classes;
    if _temp_set = [] then ([e]) else List.hd _temp_set
  in
  
  let process_pred pred = 
    match pred with
      Predicate.Atom (Expression.Binary (Expression.Eq,e1,e2)) -> 
	let s1 = set_of_elt e1 in
	if (List.mem e2 s1) then cong_classes := s1:: !cong_classes
	else 
	  let u = Misc.union s1 (set_of_elt e2) in
	  cong_classes := u::(!cong_classes)
    | _ -> ()
  in
  let _ = List.iter process_pred predlist 
  in
    let scope_of_exp exp =
      match exp with
	Expression.Lval Expression.Symbol(s) -> 
	  begin
	    let names = Misc.chop s "@" in
	    if (List.length names = 1) then "__BLAST_GLOBAL" else (List.nth names 1)
	  end
      | _ -> "__BLAST_GLOBAL"
    in
    
    let process_cong_class cc = 
      let _ = print_string "cc: \n";List.iter (fun x -> print_string ((Expression.toString x)^"  \n")) cc in  
  (* this should return a list of predicates *)
      let _ = Hashtbl.clear scope_pred_table in
      let process_exp e = 
	let scp = scope_of_exp e in
	let _ = Pretty.printf "exp: %s   scope: %s" (Expression.toString e) scp in
	let present_set = try Hashtbl.find scope_pred_table scp with Not_found -> [] in
	Hashtbl.replace scope_pred_table scp (e::present_set)
      in
      let proc_scope_terms globals scp lst =
	if (not (scp = "__BLAST_GLOBAL")) then
	  begin
	    let pairs = Misc.cross_product lst globals (* Misc.make_all_pairs (lst@globals)*) in
	    new_predicates := (List.map (fun (x,y) -> Predicate.Atom (Expression.Binary(Expression.Eq,x,y))) pairs )@(!new_predicates)
	  end
	else 
	  ()
      in
      List.iter process_exp cc;
      let globals_cc = try Hashtbl.find scope_pred_table "__BLAST_GLOBAL" with Not_found -> [] in
      let _ = print_string "globals: \n";List.iter (fun x -> print_string ((Expression.toString x)^"  \n")) globals_cc    
      in      
      Hashtbl.iter (proc_scope_terms globals_cc) scope_pred_table 
    in  
    let _ = List.iter process_cong_class (!cong_classes) in
    !new_predicates
    (* very inefficient congruence closure algorithm *)
   
(**********************************************************)      
(* Get the list of useful predicates from the last proof. *)
(**********************************************************)
let getUsefulPredicates pred =
  let getPredsFromProof pred =
    if ("cvc" = Options.getValueOfString "s") 
    then
      getProofsFromCvc pred    
    else 
    if ("simplify" = Options.getValueOfString "s") then
      getProofsFromSimplify pred 
    else
      begin
        let _eToCheck = Predicate.implies pred (Predicate.False) in
        let ans = queryExpUsingVampyre _eToCheck
        in
        let _ = Message.msg_string Message.Debug "Got UsefulPred" in
        if ans = false
        then
          []
        else
          begin
            let _lst = Engine.getUsefulPredicates () in
            (*
	       let inferred_atoms = 
	       match (!proof) with 
	       Some(prf) -> 
	       Engine.getInferredAtoms (prf)
	       | None -> []
	       in
	     *)
	     List.map convertToPredicate _lst
          end
      end 
  in
  let plist = getPredsFromProof pred in
	  Message.msg_string Message.Debug "Leaf predicates.\n";
          List.iter (fun p -> Message.msg_string Message.Debug (Predicate.toString p)) (plist); 
      (* let inf_plist = List.map convertToPredicate inferred_atoms in *)
      if(Options.getValueOfBool "scope") then
	begin
	  let pred_closure_list = predicate_closure plist in
      
	  Message.msg_string Message.Debug "Closure predicates.\n";
   	  List.iter (fun p -> Message.msg_string Message.Debug (Predicate.toString p)) (pred_closure_list); 
	  plist @ pred_closure_list 
	end
      else
	plist

	       
(*************************************************************************************)
(*************************************************************************************)
(*                               Interface to FOCI                                   *)
(*************************************************************************************)
(*************************************************************************************)

module FociPredicate = FociAst.Predicate
module FociExpression = FociAst.Expression
module FociConstant = FociAst.Constant

(* the main function this needs to export is: 
getFociInterpolant : 
   (Predicate.predicate * Predicate.predicate) -> (Predicate.predicate * Predicate.predicate)
which could optionally return an exception if the conjunction is true *)

(* this function has been cut-and-pasted from the one above -- which converts to vampyre *)
(* hence I have stolen the listOfAddress variables *)

let fctr = ref 0 

let new_foci_var () = 
  fctr := !fctr + 1;
  "foci"^(string_of_int !fctr)

let listOfAddrVarsFoci = ref []
let chlval_table = Hashtbl.create 109
let chlval_map_table  = Hashtbl.create 109
let int_table_to = Hashtbl.create 23
let int_table_from = Hashtbl.create 23

let max_int = 100000 (* some RANDOM number I made up at 1am -- RJ *)
let num_offset = ref 100000
let make_new_num i =
  let s_i = string_of_int i in
  let s_i_c = Misc.k_suffix 2 s_i in
  let rv = int_of_string s_i_c in
    rv
  (*num_offset := !num_offset + 1; 
  !num_offset*)

let sign i = if i > 0 then 1 else if i = 0 then 0 else -1 

let int_map_to i =
  Message.msg_string Message.Debug ("int_map_to"^(string_of_int i));
 let abs_i = abs i in
 let ret_val = 
   (sign i) * 
   (if abs_i < max_int then abs_i
    else
      begin
	if (Hashtbl.mem int_table_to abs_i) 
	then 
	  (Hashtbl.find int_table_to abs_i)
	else
	  let key = make_new_num abs_i in
	  let _ = (Hashtbl.add int_table_to abs_i key) in
	  let _ = (Hashtbl.add int_table_from key abs_i) in
	    key
      end)
 in
   Message.msg_string Message.Debug ("returns: "^(string_of_int ret_val));
   ret_val

let int_map_from i = 
  Message.msg_string Message.Debug ("int_map_from"^(string_of_int i));
  let abs_i = abs i in
  let ret_val =
    (try
      Hashtbl.find int_table_from abs_i
    with Not_found -> abs_i) * (sign i)
  in
    Message.msg_string Message.Debug ("returns: "^(string_of_int ret_val));
    ret_val
 
(* in theory one could even use simplify or something else to "simplify" *)
let isConstant e = 
  match e with
      Expression.Constant (Constant.Int i) -> Some i
    | Expression.Unary (Expression.UnaryMinus,Expression.Constant (Constant.Int i)) -> 
	Some (-i)
    | Expression.Unary (Expression.UnaryPlus,Expression.Constant (Constant.Int i)) -> 
	Some (i)
    | _ -> None
	
let rec convertExpFoci e =
  let convert_times e1 e2 = 
    match (isConstant e1, isConstant e2) with
	(Some i1,Some i2) -> FociExpression.Constant (FociConstant.Int (i1 * i2))
      | (Some i1, None) -> FociExpression.Coeff (FociConstant.Int i1, (convertExpFoci e2))
      | (None, Some i2) -> FociExpression.Coeff (FociConstant.Int i2, (convertExpFoci e1))
      | (None, None) -> 
	  begin
	    Message.msg_string Message.Debug (Printf.sprintf "WARNING! convertExpFoci: times of %s * %s  " (Expression.toString e1) (Expression.toString e2));
	    FociExpression.Application("times",List.map convertExpFoci [e1;e2])
	  end
  in
  let foci_minus e' =
    match isConstant e' with
	Some i -> FociExpression.Constant (FociConstant.Int (int_map_to (-i)))
      | None -> FociExpression.Coeff (FociConstant.Int (-1),(convertExpFoci e'))
  in
    match e with
	Expression.Lval lv -> convertLvalFoci lv
      | Expression.Chlval (l, s) -> 
	  begin
	    try FociExpression.Variable (Hashtbl.find chlval_map_table e) 
	    with 
		Not_found ->
		  let s_e = new_foci_var () (* Expression.toString e*) in
		    Hashtbl.add chlval_table s_e e;
		    Hashtbl.add chlval_map_table e s_e;
		    FociExpression.Variable s_e
	  end
      (* So the chlval thing didnt work out so well. look at test/foci/ex3.foci 
	 so now we just convert it to string and for reconversion, stuff the result in a hashtbl!
	FociExpression.Application ("chlval", [ convertLvalFoci l; FociExpression.Variable ("_"^s) ])
      *)
  | Expression.Binary (op, e1, e2) -> 
      begin
	match op with 
	    Expression.Minus -> FociExpression.Sum [convertExpFoci e1; foci_minus e2]
	  | Expression.Plus  -> FociExpression.Sum[convertExpFoci e1; convertExpFoci e2]
	  | Expression.Mul   -> convert_times e1 e2
	  |	Expression.BitAnd 
	  |	Expression.BitOr 
	  |	Expression.Div 
	  |	Expression.LShift 
	  |	Expression.Rem 
	  |	Expression.RShift 
	  |	Expression.Xor 
	  |	Expression.Comma ->
		  FociExpression.Application(Expression.binaryOpToVanillaString op, List.map convertExpFoci [e1;e2])
	  |     Expression.FieldOffset ->

	     let v = convertExpFoci e1 in
	     
	      (* the second argument to field offset is a field id *)
	     let s =
	       match e2 with
	         Expression.Lval (Expression.Symbol s) -> s
	       | _ -> failwith ("convertExp : (FOCI) Second operator of field offset not a field id "^
				 (Expression.toString e))
	     in
	     FociExpression.Application("foffset", [v; FociExpression.Variable ("@field_"^s)])
	     
	  |	_ -> failwith ("convertExp (Foci): This operator should not be handled here "^(Expression.toString e))
      end
  | Expression.Constant (c) -> 
      begin
	match c with
	    Constant.Int i -> FociExpression.Constant (FociConstant.Int (int_map_to i))
	  | Constant.String s -> 
	      begin
		Message.msg_string Message.Debug ("convertExpFoci: String Constant! in"^(Expression.toString e)); 
		FociExpression.Variable ("_STRINGCONSTANT"^s)
	      end
	  | Constant.Float f ->
	     	    if (f -. (ceil f) = 0.0) then FociExpression.Constant (FociConstant.Int (int_of_float f))
	    else FociExpression.Variable "_FLOATINGPTCONSTANT"
	  | _ -> failwith ("Cannot handle constant "^(Constant.toString c))
      end
  | Expression.Unary (uop, e) -> 
      begin
	match uop with 
	    Expression.UnaryMinus -> foci_minus e
	  | Expression.UnaryPlus -> convertExpFoci e
	  | Expression.Address -> 
	    let v = convertExpFoci e in
	    if (List.mem v (!listOfAddrVarsFoci)) then () 
	    else
	      listOfAddrVarsFoci := v::(!listOfAddrVarsFoci);
	    FociExpression.Application("addrOf", [v])
	(*$| Expression.Deref -> failwith "Deref"*)
	  |	_ -> FociExpression.Application(Expression.unaryOpToVanillaString uop, [convertExpFoci e])
      end
  | _ -> failwith ("convertExpFoci: " ^ (Expression.toString e) ^ "not handled")
	
  and convertLvalFoci l =
    match l with
    | Expression.Symbol(str) -> FociExpression.Variable (str) 
    | Expression.Access(Expression.Dot, e, s) -> (* NEEDS DEBUGGING *) 
	let v = convertExpFoci e in (* hack to get around foci bug! *)
	FociExpression.Application("select@field_"^s, [v(*; FociExpression.Variable ("@field_"^s)*)])
    | Expression.Access(Expression.Arrow, e, s) ->
	(*
	  Input.Fun ("sel4", [convertExp e; Input.Var ("_"^(s))])
	*)
	failwith "foci doesnt like Arrow"
    | Expression.Dereference e -> 
	FociExpression.Application("select", [convertExpFoci e; FociAst.zero])
    | Expression.Indexed (e1, e2) -> 
	FociExpression.Application("select", List.map convertExpFoci [e1;e2])
	(* WARNING!! need to check this -- RJ: Whats the difference between sel4 and select ? *)

  and 
  convertAtomFoci e = (* this returns a focipredicate *)
  let make_eq e1' e2' = FociPredicate.Equality (convertExpFoci e1', convertExpFoci e2') in
  let make_leq e1' e2' = FociPredicate.Leq (convertExpFoci e1', convertExpFoci e2') in

    match e with
	Expression.Binary (op, e1, e2) -> 
	  begin
	    match op with 
		Expression.Eq  -> make_eq e1 e2 
	      |	Expression.Ne  -> FociPredicate.Not (make_eq e1 e2)
	      |	Expression.Le  -> make_leq e1 e2	      
	      |	Expression.Lt  -> FociPredicate.Not (make_leq e2 e1)
		  (*FociPredicate.And 
		  [(make_leq e1 e2); FociPredicate.Not (make_eq e1 e2)]*)
	      |	Expression.Gt  -> FociPredicate.Not (make_leq e1 e2)
		  (*FociPredicate.And 
		  [(make_leq e2 e1); FociPredicate.Not (make_eq e2 e1)]*)
	      |	Expression.Ge  -> make_leq e2 e1 
	      |	_   -> failwith ("convertAtomFoci: Cannot handle this expression: "^(Expression.toString e))
	end
    | _ -> failwith ("convertAtomFoci: "^(Expression.toString e)^ " Not a binary relation operator")
  
  and convertPredFoci e =
    match e with
      Predicate.True -> FociPredicate.True
    | Predicate.False -> FociPredicate.False
    | Predicate.And plist -> 
	FociPredicate.And (List.map convertPredFoci plist)
    | Predicate.Or plist -> 
	FociPredicate.Or (List.map convertPredFoci plist)
    | Predicate.Implies (e1, e2) -> FociPredicate.Implies (convertPredFoci e1, convertPredFoci e2)
    | Predicate.Not e1 -> FociPredicate.Not (convertPredFoci e1)
    | Predicate.Atom e ->  (convertAtomFoci e)
    | _ -> failwith ("Foci Cannot handle this predicate "^(Predicate.toString e))
 
and convertToFociSyntax (e: Predicate.predicate) =
(*  let prfnos = string_of_int (!prfno) in*) (* get rid of prfnos once reinit is fixed *)(*  prfno := !prfno + 1; *)
  listOfAddrVarsFoci := [];
  let v_exp = convertPredFoci e
  in
  let genAddrOfEq b =
    FociPredicate.Equality(b, FociExpression.Application("select", 
							 [FociExpression.Application("addrOf", [b]); FociAst.zero]))
  in
    begin
    let vv = FociPredicate.And (v_exp::(List.map genAddrOfEq !listOfAddrVarsFoci)) in
      vv
  end

(* oh no! now we have to write the convert from foci to BLAST. ugh. *)
let rec convertFociLval l = 
  (* this is weird ... since the arg: "l" is supposed to be a foci expression that is really an lval ! *)
  match convertFociExp l with
      Expression.Lval l' -> l'
    | _ -> failwith ("convertFociLval failed on argument"^(FociExpression.toString l))
    
and make_bin op e1 e2 = 
  Expression.Binary(op,convertFociExp e1, convertFociExp e2)
and make_eq e1 e2 = make_bin (Expression.Eq) e1 e2
and make_leq e1 e2 = make_bin (Expression.Le) e1 e2
and convertFociExp e = 
  match e with
	FociExpression.Constant (FociConstant.Int i) -> Expression.Constant (Constant.Int (int_map_from i))
    | FociExpression.Variable v -> 
	begin
	  match (Hashtbl.mem chlval_table v) with
	      true -> Hashtbl.find chlval_table v 
	    | false -> Expression.Lval (Expression.Symbol v)
	end
    | FociExpression.Application (f,el) ->
	begin
	  match f with
	      "chlval" -> 
		begin
		  (* hmm... this should never happen! *)
		  try
		    let [l;(FociExpression.Variable s)] = el in
		      Expression.Chlval (convertFociLval l, Misc.chop_after_prefix "_" s)
		  with _ -> (* which should be a match failure *)
		    failwith ("unknown chlval args in convertFociExp! "^(FociExpression.toString e))
		end
	      | "*" -> 
		  begin
		    try
		      let [e1;e2] = el in
			make_bin Expression.Mul e1 e2
		    with _ ->
		      failwith ("unknown * args in convertFociExp! "^(FociExpression.toString e))
		  end
		  (* Hopefully we wont have to deal with wacko spl cases here! *)
	      | "addrOf" ->
		  begin
		    try 
		      let [e1] = el in
			Expression.Unary(Expression.Address, convertFociExp e1) 
		    with _ ->
		      failwith ("unknown & args in convertFociExp! "^(FociExpression.toString e))
		  end
	      | "select" ->
		  begin
		    try 
		      let [e1;e2] = el in
			begin
			  match e2 with
			      FociExpression.Variable s ->
				begin
				  if (Misc.is_prefix "@field_" s)
				  then
				    Expression.Lval (Expression.Access(Expression.Dot, convertFociExp e1, (Misc.chop_after_prefix "@field_" s)))
				  else
				    Expression.Lval (Expression.Indexed(convertFociExp e1,convertFociExp e2))
				end

			    | _ ->
				if (e2 = FociAst.zero) then
				  Expression.Lval (
				  Expression.Dereference(convertFociExp e1))
				else
				  Expression.Lval (
				    Expression.Indexed(convertFociExp e1,convertFociExp e2))
			end
		    with _ -> 
		      failwith ("unknown select args in convertFociExp! "^(FociExpression.toString e))
				  
		  end
	      | _ -> 
		  begin
		    if (Misc.is_prefix "select@field_" f )
		    then
		      (* a field "access" thing *)
		      let field_name = Misc.chop_after_prefix "select@field_" f in
			try
			  let [e] = el in
			    Expression.Lval (Expression.Access(Expression.Dot, convertFociExp e, field_name))
			with _ -> failwith "multiple args to select@field_ ! ?"
		    else
		      begin
			Message.msg_string Message.Debug ("WARNING!: Unknown UIF in:"^(FociExpression.toString e));
			Expression.FunctionCall(Expression.Lval (Expression.Symbol f), List.map convertFociExp el)
		      end
		  end
	end

      | FociExpression.Sum el ->
	  begin
	    match el with
		[] -> Expression.Constant (Constant.Int 0)
	      | h::tl -> 
		  List.fold_left 
		  (fun a b -> Expression.Binary (Expression.Plus,a,convertFociExp b)) 
		  (convertFociExp h) tl
	  end
      | FociExpression.Coeff ((FociConstant.Int i),e') ->
	  if (i < 0) then 
	    if (i = (-1)) then
	      Expression.Unary (Expression.UnaryMinus, (convertFociExp e'))
	    else 
	      Expression.Binary(Expression.Mul, Expression.Constant (Constant.Int (int_map_from i)),
						   convertFociExp e')
	  else
	    if (i = 1) then convertFociExp e'
	    else
	      (Expression.Binary(Expression.Mul, Expression.Constant (Constant.Int (int_map_from i)), convertFociExp e'))
	      

and convertFociPred p =
  match p with
      FociPredicate.True -> Predicate.True
    | FociPredicate.False -> Predicate.False
    | FociPredicate.And plist -> Predicate.And (List.map convertFociPred plist)
    | FociPredicate.Or plist -> Predicate.Or (List.map convertFociPred plist)
    | FociPredicate.Not p' -> Predicate.Not (convertFociPred p')
    | FociPredicate.Implies (p1,p2) -> Predicate.Implies (convertFociPred p1,convertFociPred p2)
    | FociPredicate.Equality (e1,e2) -> Predicate.Atom (make_eq e1 e2)
    | FociPredicate.Leq (e1,e2) -> Predicate.Atom (make_leq e1 e2)
    | FociPredicate.Atom s -> failwith "Shouldn't have to deal with Atomic predicates!"
	(* jhala: actually I'm being lazy... whats wrong with atomic predicates! *)


(* now the somewhat non-trivial bit *)
(* cons_pre and cons_post are the two sets of constraints. This function returns the predicates
appearing in the fwd interpolant  *)

(*
let read_foci_query filename = ()

  let inchan = open_in filename in
  let _ = Errormsg.startFile filename in
  let lexbuf = Lexing.from_channel inchan in
    FociParse.pred2 FociLex.token lexbuf 
*)

let extract_foci_preds (cons_pre,cons_post) = 
  let _ = Hashtbl.clear chlval_table in (* reset the string -> chlval map *)
  let _ = Hashtbl.clear chlval_map_table in
  let cons_l = List.map (fun x -> Predicate.And x) [cons_pre;cons_post] in
  let _ = Message.msg_string Message.Debug "In extract_foci_preds:" in
  let _ = List.iter (fun x -> Message.msg_string Message.Debug (Predicate.toString x)) cons_l 
  in
  let [foci_c_pre; foci_c_post] = List.map convertToFociSyntax cons_l in
    try
      let _ = 
	Message.msg_string Message.Debug ">>>>>>>>>>>> calling foci";
	Message.msg_string Message.Debug (FociPredicate.toString foci_c_pre);
	Message.msg_string Message.Debug ";";
	Message.msg_string Message.Debug (FociPredicate.toString foci_c_post);
	Message.msg_string Message.Debug ">>>>>>>>>>>> foci done" 
      in
      let foci_file_string = Printf.sprintf "\n \n \n %s \n ; \n  %s \n \n \n" (FociPredicate.toString foci_c_pre) (FociPredicate.toString foci_c_post)
      in
      let _ = Misc.append_to_file "fociqs" foci_file_string in
      
      (*let (p1,p2) = read_foci_query "foci-query" in
      let _ = 
	match (p1,p2) with
	    (foci_c_pre,foci_c_post) -> Message.msg_string Message.Normal "Foci Query same"
	  | _ -> Message.msg_string Message.Normal "Foci query differs!!!!!!"
      in *)
  
    let fwd_ip = 
      try Foci.interpolate (foci_c_pre,foci_c_post) 
      with e -> 
	(Message.msg_string Message.Error ("Foci FAILS! raises exception!!"^ (Printexc.to_string e));
	 Message.msg_string Message.Error ("Foci FAILS! raises exception!!"^ (Printexc.to_string e));
	 Misc.append_to_file "foci_failed_q" foci_file_string;
	FociPredicate.True (* FociPredicate.And [foci_c_pre;foci_c_post ] *))
    in
      let _ = Message.msg_string Message.Debug ("Foci yields Interpolant: "^(FociPredicate.toString fwd_ip))
      in
      let fwd_ip_b = convertFociPred fwd_ip in
      let _ = Message.msg_string Message.Debug ("Foci yields b_Interpolant: "^(Predicate.toString fwd_ip_b))
      in
	Predicate.getAtoms fwd_ip_b
    with
	Foci.SAT -> (* should be that the thing was SAT *)
	  failwith "Foci raised SAT!"
      | e -> raise e(*failwith "Foci raises some other exception..."*)
	  
(******************************  End Interface to FOCI  ******************************)

(* WE DONT WANT ICS THANK YOU VERY MUCH.     
(*************************************************************************************)
      
(**
 * Interface to ICS.  This module contains the interface to the theorem prover.
   *)
      
(* convert an Expression.expression to Ics Input *)
let convertToIcsSyntax e =
  let rec convertExp e =
    match e with
      Expression.Binary (op, e1, e2) -> 
	begin
	  match op with 
	    Expression.Minus -> Ics.mk_minus (convertExp e1) (convertExp e2)
	  |	Expression.Plus  -> Ics.mk_plus2 (convertExp e1)  (convertExp e2)
	  |	Expression.Mul   -> Ics.mk_times2 (convertExp e1) (convertExp e2)
	  |	Expression.BitAnd 
	  |	Expression.BitOr 
	  |	Expression.Div 
	  |	Expression.LShift 
	  |	Expression.Rem 
	  |	Expression.RShift 
	  |	Expression.Xor 
	  |	Expression.Comma ->
	      Ics.mk_app (Ics.mk_var (Expression.binaryOpToVanillaString op)) [convertExp e1; convertExp e2]
	  |	_ -> failwith ("convertExp: (Ics) This operator should not be handled here"^(Expression.toString e))
	end
    | Expression.Constant (c) -> 
	begin
	  match c with
	    Constant.Int i -> Ics.mk_num (Ics.num_of_int i)
	  | Constant.String s -> Ics.mk_var ("CONSTANTSTRING"^s)
	  | _ -> failwith ("Cannot handle constant "^(Constant.toString c))
	end
    | Expression.Symbol(str) -> Ics.mk_var (str) 
    | Expression.Unary (uop, e) -> 
	begin
	  match uop with 
	    Expression.UnaryMinus -> Ics.mk_unary_minus (convertExp e)
	  | Expression.UnaryPlus -> convertExp e
	  | Expression.Deref -> Ics.mk_app (Ics.mk_var "sel4") ([convertExp e; Ics.mk_num (Ics.num_of_int 0)])
	  | Expression.Address -> 
	      let v = convertExp e in
	      Ics.mk_app (Ics.mk_var "addrOf") [v]
	  |	_ -> Ics.mk_app (Ics.mk_var (Expression.unaryOpToVanillaString uop)) ([convertExp e])
	end
    | Expression.Access(Expression.Dot, e, s) -> 
	let v = convertExp e in
	Ics.mk_app (Ics.mk_var "sel4") [v; Ics.mk_var s]
    | Expression.Access(Expression.Arrow, e, s) -> (* treat e->s as ( *e ).s *)
	let v = Ics.mk_app (Ics.mk_var "sel4") [convertExp e; Ics.mk_num (Ics.num_of_int 0)] 
	in
     	Ics.mk_app (Ics.mk_var "sel4") ([v; Ics.mk_var s])
    | _ -> failwith ("convertExp: " ^ (Expression.toString e) ^ "not handled")



  and convertAtom e =
    match e with
      Expression.Binary (op, e1, e2) -> 
	begin
	  match op with 
	    Expression.Eq  -> Ics.mk_equal (convertExp e1) (convertExp e2)
	  |	Expression.Ne  -> Ics.mk_diseq (convertExp e1) (convertExp e2)
	  |	Expression.Gt  -> Ics.mk_gt (convertExp e1) (convertExp e2)
	  |	Expression.Lt  -> Ics.mk_lt(convertExp e1) (convertExp e2)
	  |	Expression.Ge  -> Ics.mk_ge (convertExp e1) (convertExp e2)
	  |	Expression.Le  -> Ics.mk_le (convertExp e1) (convertExp e2)
	  |	_   -> failwith ("convertAtom: Cannot handle this expression: "^(Expression.toString e))
	end
    | _ -> failwith ("convertAtom: "^(Expression.toString e)^ " Not a binary relation operator")
  
  and convertPred e =
    match e with
      Predicate.True -> Ics.mk_true ()
    | Predicate.False -> Ics.mk_false ()
    | Predicate.And plist -> 
	List.fold_left (function a -> function b -> Ics.mk_and (convertPred b) a) 
	  (Ics.mk_true ()) (plist)
    | Predicate.Or plist -> 
	List.fold_left (function a -> function b -> Ics.mk_or (convertPred b) a) 
	  (Ics.mk_false ()) (plist)
    | Predicate.Implies (e1, e2) -> Ics.mk_imp (convertPred e1) (convertPred e2)
    | Predicate.Not e1 -> Ics.mk_not (convertPred e1)
    | Predicate.Atom e ->  (convertAtom e)
    | _ -> failwith ("Cannot handle this predicate "^(Predicate.toString e))
  in
  let v_term = convertPred e
  in
  v_term
      
      
WE DONT WANT ICS THANK YOU VERY MUCH *)      









