(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: wpdag.ml,v 1.5 2003/06/19 01:47:37 jhala Exp $
 *
 * This file is part of the BLAST Project.
 *)



(*****************************************************************************************)
(*****************************************************************************************)
(*****************************************************************************************)
(*                                                                                       *)
(*                                     The WP DAG MODULE                                 *)
(*                                                                                       *)
(*****************************************************************************************)
(*****************************************************************************************)
(*****************************************************************************************)

(**
 * This module implements a procedure that when given a DAG cfa, with a single sink,
  (and a single root ?) pulls the wp back from the sink to the root  .
 *)
 
(* here we go ... *)
module Symbol = Ast.Symbol
module Constant = Ast.Constant
module Expression = Ast.Expression
module Predicate = Ast.Predicate
module C_SD = BlastCSystemDescr.C_System_Descr
module Command = BlastCSystemDescr.C_Command

(*****************************************************************************************)
(************************************ MAP functions **************************************) 
(*****************************************************************************************)

(* first the functions that deal with the variable map i.e. the v: lval --> lval
   where the v(l) = the "current" value of l 
*)

let mapsize = 31
let loc_table_size = 31 

let ctr = ref 0
  
let new_map () = 
  Hashtbl.create mapsize

let copy_update_map v changed_lvals = 
  (* we assume the precondition that every table has one 1 binding per key *)
  (* we assume that the list changed_lvals does not have copies *)
  let new_table = new_map () in
  let _ = List.iter (fun x -> Hashtbl.replace new_table x (Expression.fresh_lval x)) changed_lvals 
  in
  let _ = Hashtbl.iter 
	    (fun x y -> 
	       if Hashtbl.mem new_table x then () 
	       else Hashtbl.replace new_table x y
	    ) v  
  in
    new_table

(** join v_l -> v,D where v_l is a list of maps and v is their "join" and D the variables that need to be updated *)
(** IMP: go over this to double check that all maps have single bindings *)

let join v_l = 
  let temp_table = new_map () in
  let diff = ref [] in
  let proc key data =
    match Hashtbl.find_all temp_table key with
	[] -> Hashtbl.replace temp_table key data
      | [lval] -> if not (lval = data) then diff := key::(!diff)
      | _ -> failwith "inconsistent temp_table"
  in
    List.iter (fun t -> Hashtbl.iter proc t) v_l;
    (copy_update_map temp_table !diff, !diff)

(*****************************************************************************************)
(***** a replace function, that takes an expression and computes e [v] for a substable v *)
(*****************************************************************************************)

let map_of_v v = 
  fun x -> 
    try Hashtbl.find v x
    with Not_found -> Expression.Lval x
  
let eqcons e1 e2 =
  Predicate.Atom (Expression.Binary(Expression.Eq,e1,e2))

(** Which variables in the domain of v are possibly aliased to lval *)
let aliases v lval = 
  List.filter 
    (fun x -> 
       AliasAnalyzer.queryAlias 
       (Expression.Lval x) 
       (Expression.Lval lval)) 
    (Misc.hashtbl_keys v) 

(*****************************************************************************************)
(************************************Different Pres **************************************) 
(*****************************************************************************************)

let pre_v_asn v lval =
  let changed_vars = aliases v lval in
    copy_update_map v changed_vars   

let pre_v_asm v lval_l =
  let v' = copy_update_map v [] in
    List.fold_right 
      (fun lval tbl -> 
	 if Hashtbl.mem tbl lval then () 
	 else Hashtbl.replace tbl lval (Expression.Lval lval);
	 tbl
      ) lval_l v'

let phi i = Predicate.Atom (Expression.Lval (Expression.Symbol (Printf.sprintf "ND%d" i)))

let pre (op,n) =
  let (n_id_fmla, n_v) = 
    match n with 
	(id,_,Some(v,_)) -> (phi id,v)
      | _ -> failwith "invalid data at successor node"
  in
  let acp asn (fmla,v) =
    match asn with
      | Command.Expr (Expression.Assignment(Expression.Assign,x,e)) -> 
	  let v' = pre_v_asn v x in	  
	  let mapv = map_of_v v in
	  let mapv' = map_of_v v' in
	  let e' = Expression.deep_alpha_convert (map_of_v v') e in
	  let lx = Expression.Lval x in
	  let con = 
	    (eqcons (Expression.deep_alpha_convert mapv lx) e')::
	    (List.map
	       (fun y ->
		  let (ly) = (Expression.Lval y) in
		  let addrcons = (AliasAnalyzer.generateAliasConstraints x y) in
		  let pos = 
		    Predicate.conjoinL 
		      [addrcons;eqcons (Expression.deep_alpha_convert mapv ly) e'] 
		  in
		  let neg = 
		    Predicate.conjoinL 
		      [Predicate.Not addrcons;
		       eqcons (Expression.deep_alpha_convert mapv ly) (Expression.deep_alpha_convert mapv' ly)]
		  in
		    Predicate.disjoinL [pos;neg]
	       )
	       (aliases v x))
	  in
(* Jhala: check if you should use "deep" alpha convert here or just the lookup in the hashtable *)
(* VVV IMP ! *)
	    ((* v(x) == e[v'] ::constraints*) con@fmla, v') (* TBD: constraints for aliased variables *)
      | _ -> failwith "invalid statement fed to acp"
  in
    match op.Command.code with
	Command.Pred p -> ((Predicate.deep_alpha_convert (map_of_v n_v) p)::[n_id_fmla], 
		     pre_v_asm n_v (Predicate.allVarExps_deep p)) (* returns lvals ? *)
      | Command.Block l -> List.fold_right acp l ([n_id_fmla],n_v)


(** diffcons v' d v gives the constraints connecting the changed variables of v' with their changed values in v *)
let diffcons v' d v =
  let mapv = map_of_v v in (* should we be using the map here ? *)
  let mapv' = map_of_v v' in
  let cons lv = eqcons 
		  (Expression.deep_alpha_convert mapv (Expression.Lval lv)) 
		  (Expression.deep_alpha_convert mapv' (Expression.Lval lv))
  in
    Predicate.conjoinL (List.map cons d)
  
(** These are the  the main function that is passed to the dag-iterator *)
(* the argument passed is a list of edge_labels, node_labels *)

(***************************************************************************************************)
(******************************** THe functions passed to the DAG iterator *************************)
(***************************************************************************************************)

let f n_label e_n_l =
  let boolean_dot l1 l2 = 
    if not (List.length l1 = List.length l2) then failwith "boolean_dot: bad args !"
    else Predicate.disjoinL (List.map2 (fun x y -> Predicate.conjoinL [x;y]) l1 l2)
  in
  let (pref_l,prev_l) = Misc.unzip (List.map pre e_n_l) in
  let (n_v,d) = join prev_l in
  let cons_l = List.map (fun v -> diffcons v d n_v) prev_l in
    (Misc.fst3 n_label,
     Misc.snd3 n_label,
     Some (n_v ,boolean_dot cons_l (List.map Predicate.conjoinL pref_l))  )
    

(* RJ: This needs to be changed to be True only if the leaf corresponds to error_location *)
let leaf_val error_coords (i,xy,_) = 
  (i,xy,Some(new_map (),if xy = error_coords then Predicate.True else Predicate.False))


let is_init_fn n =
  match BiDirectionalLabeledDag.get_node_label n with
      (_,_,Some(_)) -> true
    | (_,_,None) -> false


(************************************************************************************************)
(*************   The main function that returns the fmla for the wp of a dag   ******************)
(************************************************************************************************)


(* 1. hit the root of the dag with the above 3 functions 
   2. go over the dag and collect nodes, and take the fmla component
   3. and all the fmlas *)

let weakest_pre error_coords root  = 
  let _ = BiDirectionalLabeledDag.compute_and_update_root_val is_init_fn f (leaf_val error_coords)  root
  in
  let constr_node x = 
    match x with
	(_,_,Some(_,f)) -> f
      | _ -> Predicate.True
  in
    Predicate.conjoinL (BiDirectionalLabeledDag.dag_map constr_node root) 
  
(***********************************************************************************************)
(***************************All that remains is to get the unrolled DAG! ***********************)
(***********************************************************************************************)

(********************** Hmm ... now the somewhat more iffy code for the unrolling **************)

let dag_ctr = ref 0
let new_node_id () = dag_ctr := !dag_ctr + 1; !dag_ctr

(* as usual we will do everything backwards ... *)
let copy_map map = 
    let _t = Hashtbl.create loc_table_size in
    let _ = Hashtbl.iter (fun x y -> Hashtbl.replace _t x y) map in
      _t

let is_latest map n = 
  let (_,loc_id,_) = BiDirectionalLabeledDag.get_node_label n in
    (n = Hashtbl.find map loc_id)
    
let make_node (old_map,new_map) loc_id loc =
  (* we will NOT create a new node if the children of the old one are all "new" *)
  let (i1,i2) = loc_id in
  let _ = Message.msg_string Message.Debug (Printf.sprintf "in make_node %d %d" i1 i2) in
  let on = Hashtbl.find old_map loc_id in
  let oe = C_SD.get_outgoing_edges loc in  
  let children = BiDirectionalLabeledDag.get_children on in
  let eschew = (List.for_all (is_latest old_map) children) && (oe <> []) in
    match eschew with
      true -> Message.msg_string Message.Debug "No node made"
    | false ->
	begin
	  let _ = Message.msg_string Message.Debug "Making node" in
	  let nn = BiDirectionalLabeledDag.create_root (new_node_id (), loc_id, None) in
	  
	  let succ_l = 
	    List.map 
	      (fun x -> Hashtbl.find old_map (C_SD.location_coords (C_SD.get_target x)))
	      oe 
	  in
	  let succ_e_l = List.map C_SD.get_command oe in
	  let _ = List.map2 
		    (BiDirectionalLabeledDag.hookup_parent_child nn) 
		    succ_l succ_e_l 
	  in
	    Hashtbl.replace new_map loc_id nn
	end
	
let build_dag fname k = 
(* do I need to pass the " CFA" to this ?? *)
  (* let us first collect the list of locations *)
  let visited_table = Hashtbl.create loc_table_size in	       
  let rec traverse_location_and_edges loc =
    let (loc_id) = C_SD.location_coords loc in
      if (Hashtbl.mem visited_table loc_id) then
        ()
      else
        begin
	  let _ = Hashtbl.replace visited_table loc_id loc in
	  let outgoing_edges = List.rev (C_SD.get_outgoing_edges loc) in
	    (List.iter (fun e -> traverse_location_and_edges (C_SD.get_target e)) 
	       outgoing_edges)          
	end 
  in    
  let _ = traverse_location_and_edges (C_SD.lookup_entry_location fname) in 
    (* after this point visited_table : loc_id -> loc *)
  let init_map = Hashtbl.create loc_table_size in
  let _ = Hashtbl.iter 
	    (fun x y -> Hashtbl.replace init_map x 
	       (BiDirectionalLabeledDag.create_root (new_node_id (),x,None)))
	    visited_table 
  in
  let rec _ac_build_dag map _k = 
    Message.msg_string Message.Debug ("in _ac_build_dag depth: "^(string_of_int _k));
    if _k = 0 then map
    else
      begin
	let new_map = copy_map map in
	let _ = Hashtbl.iter (make_node (map,new_map)) visited_table in
	  _ac_build_dag new_map (_k - 1)
      end
  in
    _ac_build_dag init_map k


let prune_fun sink_list =
  let keep_table = Hashtbl.create loc_table_size in
  let reachable_ancestors = BiDirectionalLabeledDag.ancestors sink_list in
  let _ = List.iter (fun x -> Hashtbl.replace keep_table x true) reachable_ancestors in
    fun x -> not (Hashtbl.mem keep_table x)

(********************************************************************************************)
(****************************   Tying it all Together  **************************************)
(********************************************************************************************)

let dag_node_to_string n_label = 
			let (i,(l1,l2),_) = n_label in
			  string_of_int i

let dag_edge_to_string e_label = 
   Command.to_string e_label


let dump_wplist wplist node_coord_list = 
  Message.msg_string Message.Normal "***************************************************************";
  Message.msg_string Message.Normal "***************************************************************";
  Message.msg_string Message.Normal "Dumping WP formulas for the various roots of the DAG";
  let ac_dump (phi,(id1,id2)) = 
    Message.msg_string Message.Normal "***************************************************************";
    Message.msg_string Message.Normal (Printf.sprintf "Node: (%d, %d)" id1 id2);
    Message.msg_string Message.Normal (Predicate.toString phi);
    ()
  in
    try
      List.iter ac_dump (List.combine wplist node_coord_list) 
    with _ -> failwith "bad args to dump_wplist"
      
      
let main () = 
  Message.set_verbosity Message.Default;
  Stats.reset ();
  Options.buildOptionsDatabase ();
  let reachLabel = (Options.getValueOfString "L")
  and source_files = (Options.getSourceFiles ()) in
    if(reachLabel = "") then
      begin
        Printf.printf "No error label specified!\n" ;
        exit 1
      end
    else
      () ;
    if (source_files == [])
    then
      begin
        Message.msg_string Message.Error ("Error: no input source file specified") ;
        exit 1
      end ;
    Message.msg_string Message.Normal ("Parsing files: " ^ (Misc.strList source_files)) ;
    AliasAnalyzer.constructAliasDatabase () ;  
    Stats.time "Parse:" C_SD.initialize source_files ;
    let error_location =
      match (C_SD.get_locations_at_label reachLabel) with
          [l] -> l
        | []  ->
            Message.msg_string Message.Error ("Error: label '" ^ reachLabel ^
                                              "' was not found in the source file(s)") ;
            exit 1
        | _   ->
            Message.msg_string Message.Error ("Error: label '" ^ reachLabel ^
                                              "' appears multiple times in the source file(s)") ;
            exit 1
    in
    let error_coords = C_SD.location_coords error_location in
    let dag_root_table = build_dag (List.hd (Options.getValueOfStringList "main"))
			   (Options.getValueOfInt "depth") 
    in 
    let dag_root_list = List.map (Hashtbl.find dag_root_table) (Misc.hashtbl_keys dag_root_table) 
    in
    let reachable_list = BiDirectionalLabeledDag.descendants dag_root_list in
    let rsize1 = List.length reachable_list in
    let error_nodes = List.filter 
			(fun x -> let (_,c,_) = BiDirectionalLabeledDag.get_node_label x in c = error_coords) 
			reachable_list
    in
    let _ = BiDirectionalLabeledDag.prune (prune_fun error_nodes) dag_root_list in
    let rsize2 = List.length (BiDirectionalLabeledDag.descendants dag_root_list) in
    let _ = Printf.printf "Size1 = %d, Size2 = %d \n" rsize1 rsize2 in
    let _ = BiDirectionalLabeledDag.output_dag_dot 
	      dag_node_to_string dag_edge_to_string "dag.dot" dag_root_list in
      (* all the bogus nodes are now gone *)
    let wp_list = List.map (weakest_pre error_coords) dag_root_list in
      dump_wplist wp_list 
	(List.map (fun x -> Misc.snd3 (BiDirectionalLabeledDag.get_node_label x))
	   dag_root_list)
	(* hmm... will this work if the dag_root_list has size more than 1 ? since 
	   nodes will get state etc etc ? Shouldnt we have to "reset" the tables or 
	   something ? or NO! the Dynamic Programming deals with all of that!*)
	(* maybe one should print out the formulas or something ? *)
	   
let () = main ()   
