(*
 * Description:
 *
 * Copyright (c) 2003 by
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

(* 
   Module Special
   Adds special rewrite equalities to the system.
   Rewrite rules:
      fold - unfold for recursive types
      select (Mem, address x) = x
*)
module E = VampyreErrormsg
module I = Input
module P = Proof
module S = Satproc
open VampyrePretty
open Satproc
open Dag

(*
 * State Variables
 *)
let debug = ref false					(* set on command line *)
let spid = ref ""

type undoElem = 
	  Snapshot  (* marks a snapshot *)

let undoStack : undoElem list ref = ref [] 
(* 
 * Record a snapshot 
 *)
let snapshot () = 
	undoStack := Snapshot :: !undoStack

(* 
 * Discard all assertions since last snapshot 
 *)
let undo () = 
	let rec loop = function
		Snapshot :: rest -> undoStack := rest
		| [] -> E.s (E.bug "GT tried to pop an empty stack (too many undos?)")
	in 
		loop !undoStack


(* counter for making unique message ids *)
let counter = ref 0
let prfId () = 
	counter := !counter + 1 ;
	!counter


let makeMessage eq =
  if(!debug) then ignore(
    printf "New equality\n");
  { S.body = eq;
    S.owner = !spid;
    S.id = 0;
    S.goal = false;
  } 

(*
  [ expression / var] s
*)
let rec substitute s var expression =
  (if (!debug) then 
    ignore (printf "SP: substitute called\n" ));
  match var.op with
    Var vv ->
      (match s.op with
	Var v -> if (v=vv) then expression else s
      |	Int i -> s
      |	Plus -> newNode Plus (List.map (function ss -> substitute ss var expression) s.args) 
      |	Minus -> newNode Minus (List.map (function ss -> substitute ss var expression) s.args) 
      |	Times -> newNode Times (List.map (function ss -> substitute ss var expression) s.args) 
      |	Div i -> newNode (Div i) (List.map (function ss -> substitute ss var expression) s.args) 
      |	Mod i -> newNode (Mod i) (List.map (function ss -> substitute ss var expression) s.args) 
      |	Floor i -> newNode (Floor i) (List.map (function ss -> substitute ss var expression) s.args) 
      |	Ceil i -> newNode (Ceil i) (List.map (function ss -> substitute ss var expression) s.args )
      |	Fun fnname -> if fnname=vv then expression else
	            newNode (Fun fnname) (List.map (function ss -> substitute ss var expression) s.args )
      |	_ -> failwith "Not implemented"
	  )
  | Fun vv ->
      (match s.op with
	Var v -> if (v=vv) then expression else s
      |	Int i -> s
      |	Plus -> newNode Plus (List.map (function ss -> substitute ss var expression) s.args) 
      |	Minus -> newNode Minus (List.map (function ss -> substitute ss var expression) s.args) 
      |	Times -> newNode Times (List.map (function ss -> substitute ss var expression) s.args) 
      |	Div i -> newNode (Div i) (List.map (function ss -> substitute ss var expression) s.args) 
      |	Mod i -> newNode (Mod i) (List.map (function ss -> substitute ss var expression) s.args) 
      |	Floor i -> newNode (Floor i) (List.map (function ss -> substitute ss var expression) s.args) 
      |	Ceil i -> newNode (Ceil i) (List.map (function ss -> substitute ss var expression) s.args )
      |	Fun fnname -> 
	  if fnname=vv then expression 
	  else
	    newNode (Fun fnname) (List.map (function ss -> substitute ss var expression) s.args )
      |	_ -> failwith "Not implemented"
	    )

  | _ -> failwith "substitute: this should be a variable"



let rec getAllMuExpressions e =
  match e.op with
    Eq  -> (getAllMuExpressions (List.nth e.args 0)) @ (getAllMuExpressions (List.nth e.args 1)) 
  | Neq  -> (getAllMuExpressions (List.nth e.args 0)) @ (getAllMuExpressions (List.nth e.args 1)) 
  | Geq  -> (getAllMuExpressions (List.nth e.args 0)) @ (getAllMuExpressions (List.nth e.args 1)) 
  | Leq  -> (getAllMuExpressions (List.nth e.args 0)) @ (getAllMuExpressions (List.nth e.args 1)) 
  | Gt  -> (getAllMuExpressions (List.nth e.args 0)) @ (getAllMuExpressions (List.nth e.args 1)) 
  | Lt  -> (getAllMuExpressions (List.nth e.args 0)) @ (getAllMuExpressions (List.nth e.args 1)) 
  | BFun _ -> List.flatten (List.map (getAllMuExpressions) e.args)
  | NotBFun _ -> List.flatten (List.map (getAllMuExpressions) e.args)

  | Int _ -> []
  | Var var -> []
  | Plus -> []
  | Minus -> []
  | Times -> []
  | Div _ -> []
  | Mod _ -> []
  | Floor _ -> []
  | Ceil _ -> []
  | Fun "mu" -> e:: (List.flatten (List.map (getAllMuExpressions) e.args))
  | Fun _  -> (List.flatten (List.map (getAllMuExpressions) e.args))



let makeNewEquality muexp =
  let body = (List.nth muexp.args 1) and
      var = (List.nth muexp.args 0) in
  S.Equality (muexp, (substitute body var muexp))

(* 
 * Asserts a literal (the second argument). Accumulate the list of
 * equalities to the first argument  
 *
 * Input Params: 
 * 	ml : message list			  Accumulated equalities
 * 	msg : message						Asserted literal
 * 	retval : message list   New accumulated equalities
 *
 *  message = { body:msgbody , owner:ident , id:int, goal:bool } 
 *)
let assume msg_list msg =
  if(!debug) then ignore (printf "SP: assume called.\n");
  let assumeWorker msg node =
    let allmu = getAllMuExpressions node
    in
    List.map makeNewEquality allmu
  in
  if (!debug) then ignore 
      (printf "SP: assume %a\n" S.dm msg) ;
  if msg.owner <> !spid then
    let newEqList =
      match msg.body with
	Literal (l) ->  assumeWorker msg l 
      | Equality (a,b) -> []
    in
    msg_list @ (List.map makeMessage newEqList)
  else msg_list
;;


let makeExprFromDag n =
  let ne = nodeVal n in
  match ne with
    DExp (e) -> e
  | DAtom _ -> failwith "makeExprFromDag: tried to make expression of an atom node"


let proveEqual n1 n2 =
  (* check that n1 and n2 are of the form mu t.e and e[mu t.e / t] resp.
   *)
  match n1.op with
    Fun "mu" -> 
      begin
	let bdvar = List.hd n1.args and
	    body = List.nth n1.args 1
	in
	if (n2 = (substitute body bdvar n1)) then
	  let n1ex = makeExprFromDag n1 and
	      n2ex = makeExprFromDag n2
	  in
	  Proof.G("fold", [ P.E n1ex; P.E n2ex])
	else failwith "SP: Wrong proof"
      end
  | _ -> failwith "SP: proveEqual -- Term of unknown type"



(*
 * Prove a message generated earlier. 
 * 
 *	msg : message
 *	retval : P.proof
 *)
let proveIt msg =
	if !debug then
	  ignore (printf "%s: proveIt %a\n" !spid Satproc.dm msg);
	if msg.owner <> !spid then
		failwith ((!spid)^".proveIt: not my message");
	match msg.body with
	Equality (n1, n2) -> proveEqual n1 n2
	| _ -> failwith ((!spid)^".proveIt: Not an equality message")
	 
    
(* 
 * Called to initialize the decision procedure and to create a table of 
 * the entry points. 
 *
 * 	id : ident
 * 	retval : entries
 *)
let initialize id = 
  spid := id;
  { S.snapshot = snapshot;
    S.undo = undo;
    S.assume = assume; 
    S.proveIt = proveIt;
    S.debug = debug;
    S.ident  = id;
    S.reinitialize = function () -> ();
  } 
(* eof: special.ml *)




