(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
open VampyrePretty
module E = VampyreErrormsg
module I = Input
module S = Satproc
module Eng = Engine

 (* The top-level function of the prover *)
let failed  = ref false                 (* Something failed. Must exit with 
                                         * an error code *)

                                        (* These are the decision procedures. 
                                         * To avoid using some of them you 
                                         * can use the -exclude comand line 
                                         * argument *)
module CC : S.SATPROC = Equality 
module SX : S.SATPROC = Simplex 
module GT : S.SATPROC = Tactical 
module TYP: S.SATPROC = Typing
(* module SI : S.SATPROC = Sup_inf *)
module SP : S.SATPROC = Special

let allSatProcs = 
  [CC.initialize "CC";
    SX.initialize "SX";
(*    SI.initialize "SI"; *)
    SP.initialize "SP";
    TYP.initialize "TYP";
    GT.initialize "GT"]


let rec theMain () =
  (* circular dependency hack *)
  Proof.simp_and_subtract_is_zero := Sup_inf.simp_and_subtract_is_zero;
  Proof.simp_and_subtract_is_at_least_zero := Sup_inf.simp_and_subtract_is_at_least_zero;
  (* end circular dependency hack *)

  let usageMsg = "Usage: prover [options] lffiles" in
  let files : string list ref = ref [] in
  let recordFile fname = files := fname :: (!files) in
  let openLog lfile = 
    try E.logChannel := open_out lfile with _ -> 
      raise (Arg.Bad "Cannot open log file") in
  let proofFile fname = 
    try E.proofChannel := open_out fname with _ ->
      raise (Arg.Bad "Cannot open proof file") in
  let excludes : string list ref = ref [] in
  let excludeProc name = excludes := name :: !excludes in
 	(* [weimer] handle "-debug CC" and the like *)
 	let debugs : string list ref = ref [] in 
 	let debugProc name = debugs := name :: !debugs in 
  (*********** COMMAND LINE ARGUMENTS *****************)
  let argDescr = [
    "-debug", Arg.String debugProc, "name of satproc to print debug output (or ALL)";
    "-L", Arg.String openLog, "the name of the log file";
    "-o", Arg.String proofFile, "the name of the output proof file";
    "-exclude", Arg.String excludeProc,  "name of satproc to exclude";
  ] in
  begin
    VampyreStats.reset ();
    Arg.parse argDescr recordFile usageMsg;
 		(* [weimer] set the debug flags in all of the satprocs mentioned on
 		 * the command line. Note: if "-debug ALL" is specified, all satprocs
 		 * have their debug flags set. *)
 		List.iter (fun sp -> 
 			if (List.mem sp.S.ident !debugs ||
 					List.mem "ALL" !debugs) then sp.S.debug := true) allSatProcs ;
 		(* [weimer] if any "-debug" was specified, turn on E.debugFlag. *)
		if (List.mem "EG" !debugs) then Engine.debug := true; 
 		if (!debugs != []) then E.debugFlag := true ;
		Dag.debug := Equality.debug ; 
    (* Initialize the engine and construct the prover *)
    let satprocs = List.filter (fun sp -> 
                               not (List.mem sp.S.ident !excludes)) 
                   allSatProcs  in
    ignore (printf "Using satprocs: %a\n"
              (docList (chr ',') (fun sp -> text sp.S.ident)) satprocs);
    let (prover,_,_,_) = Eng.prove satprocs in
    List.iter (proveFile prover) (List.rev !files);
    VampyreStats.print stderr "Timings:\n"
  end

and proveFile prover file = 
                                        (* Do the parsing *)
  let doParse inchan = 
    let lexbuf = Lexing.from_channel inchan in
    let _ = E.theLexbuf := lexbuf  in
    Lfparse.main Lflex.token lexbuf
  in
  let processInput = function
      I.Goal(name, p) -> 
        (* 7/19/2000: apply pushNeg to expression before proving *)
	let proof = VampyreStats.time ("prove " ^ name) (prover name) (Input.pushNeg p) in
        ignore proof
    | _ -> ignore (printf "Ignoring input\n")
  in

  VampyreErrormsg.startFile (Filename.basename file);
  let inchan = 
    try open_in file with _ -> E.s (E.error "Cannot find file %s" file) in
  let inputs = VampyreStats.time "parse" doParse inchan in
  close_in inchan;
  
                                        (* Now process the inputs *)
  List.iter processInput inputs
;;
                                        (* Define a wrapper for main to 
                                         * intercept the exit *)

let main () = 
(*	Unix.alarm 600 ; 
*)

  let term = 
    try  (* [weimer] comented out so that we can actually see exceptions *)
      theMain ();
      fun () -> exit (if !failed then 1 else 0)
    with e -> raise e
    (* with e -> (fun () -> raise e) *) (* comment out to see exceptions *)
  in
  begin
    close_out (! E.logChannel);  
    close_out (! E.proofChannel);
    term ()
  end
;;

Printexc.catch main ()
;;
