(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
(* Dror : 664.2950 *)

(*
#cd "../obj"
#load "pretty.cmo"
#load "errormsg.cmo"
#load "input.cmo"
#load "unionFind.cmo"
#load "dag.cmo"
#load "proof.cmo"
#load "satproc.cmo"
*)

open Num
open VampyrePretty
open Input
open Dag
open Proof
open Satproc


(*
   ----------------------------------------------------------------
   DATA TYPES 
   ----------------------------------------------------------------
*)

(* TABLEAU *)

(* tableau vector id *)
type tab_vector_id = Dag_id of int | Const

(* int of ID function *)
let intId id = 
  match id with
    (Dag_id i) -> i;
  | (Const) -> -1  

(* leq function for id's *)
let id_leq_id id1 id2 =
  match (id1,id2) with
    (Const,_) -> true
  | (_,Const) -> false
  | (Dag_id i1, Dag_id i2) -> i1 <= i2

  
(* entries of the tableau vectors *)
type tab_entry = {mutable row: int; mutable col: int; mutable coef: num}

(* tableau vector *)
type tab_vector  = {
    mutable tab_id          : tab_vector_id;
    mutable star_rst        : bool;
    mutable plus_rst        : bool;
    mutable entries         : tab_entry ref list;
  } 

(* tableau *)
type tableau = {
    mutable rows : tab_vector list;
    mutable cols : tab_vector list;
  } 

(* our_tab : a global variable *)
let our_tab = {rows=[]; cols=[{tab_id = Const; star_rst = false; plus_rst = false; entries = []}]}  (* reinit *)

(* UNDO *)
type undoEntry = SnapMark | AddedPrivateNode of int | AddedID of tab_vector_id | StarID of tab_vector_id | PlusID of tab_vector_id
let undoStack = Stack.create ()  (* reinit *)

(* LINEAR COMBINATIONS *)
type lin_elem = {lin_id : tab_vector_id; lin_coef : num}
type lin_comb = lin_elem list

type int_lin_elem = {i_lin_id : tab_vector_id; i_lin_coef : int}
type int_lin_comb = int_lin_elem list

(* the list of pairs of starred id's with their negative linear combs *)
let star_combs_list = ref []   (* reinit *)

(* for compliance *)
let debug = ref false

(* global association list of private dags *)
let private_nodes = ref [] (* reinit *)

(* next_private_id *)
let next_private_id = ref (-1)  (* reinit *)

(* ourFindDag *)
let ourFindDag id =
  List.assoc id !private_nodes

(* from Jhala's code... *)
let dummyNode = Dag.dummyNode
let dummyMesg = Dag.dummyMesg


(* newPrivateNode *)
let newPrivateNode op args =
  let id = !next_private_id in
  next_private_id := (!next_private_id - 1);
  let rec node = { Dag.id = id ; Dag.op = op ; args = args ; root = node ; ac_use = []; use = []; eqClass = [id];isAsserted = false;mesg = dummyMesg} in
  private_nodes := (id,node) :: !private_nodes;
  Stack.push (AddedPrivateNode id) undoStack;
  node

(*
   getDag

   given a tab_id, returns the dag node or fails
*)
let getDag id = 
  match id with
    Dag_id(i) -> if (i < 0) then (ourFindDag i) else (findDag i)
  | Const -> failwith "Attempted to get dag of non-dag id"

(* the identifier of this module *)
let simplex_id = ref ""

(* pf_exp_list: ref to an associated list of functions that proof that an exp >= 0 *)
let pf_exp_list = ref []  (* reinit *)


(* ----------------------------------------------------------------
   FUNCTIONS FOR PRINTING
   ---------------------------------------------------------------- *)

(* print tab_vector_id *)
let print_tab_id id =
  match (id) with
    (Dag_id(int)) -> print_int int
  | (Const) -> print_string "CONST"

(* print tab entry *)
let print_tab_entry entry = 
  print_string "(";
  print_int entry.row;
  print_string ",";
  print_int entry.col;
  print_string ")";
  print_string " : ";
  print_string (string_of_num entry.coef);
  print_string "   "
    
(* print dag *)

let print_dag_op op = 
  match (op) with
    (Int(i)) -> print_int i;
  | (Var(v)) -> print_string v;
  | (Plus) -> print_string "+";
  | (Minus) -> print_string "-";
  | (Times) -> print_string "*";
  | (Eq) -> print_string "=";
  | (Lt) -> print_string "<";
  | (Gt) -> print_string ">";
  | (_) -> print_string "?"
	
let rec print_dag dag = 
  match (List.length (dag.args)) with
    2 ->
      let args = dag.args in
      print_string "(";
      print_dag (List.hd args);
      print_string ")";
      print_dag_op dag.op;
      print_string "(";
      print_dag (List.hd (List.tl args));
      print_string ")"
  | 1 ->
      let arg = (List.hd dag.args) in
      print_dag_op dag.op;
      print_string "[";
      print_dag arg;
      print_string "]"
  | 0 ->
      print_dag_op dag.op
  | _ ->
      failwith "I do not print a dag with more than two args"

(* print tableau vector *)
let print_tableau_vector vect =
  print_string "\n";
  print_string "[";
  print_tab_id vect.tab_id;
  print_string "] ";
  print_string "  ";
  if (vect.tab_id <> Const) then print_dag (getDag vect.tab_id);
  if (vect.star_rst) then print_string "  [*-rst]";
  if (vect.plus_rst) then print_string "  [+-rst]";
  print_string "\n";
  List.iter (function e -> (print_tab_entry !e)) vect.entries;
  print_string "\n"

(* print tableau *)
let print_tableau t =
  print_string "... ROWS ...\n";
  List.iter (function r -> (print_tableau_vector r)) t.rows;
  print_string "\n... COLS ...\n";
  List.iter (function c -> (print_tableau_vector c)) t.cols;
  print_string "\n\n"

(* print comb *)
let rec print_comb comb =
  match comb with
    [] -> ()
  | [head] ->
      print_string "(";
      print_string (string_of_num ((Num.Int 0) -/ head.lin_coef));
      print_string ")*[";
      print_dag (getDag head.lin_id);
      print_string "]"

  | (head::tail) ->
      print_string "(";
      print_string (string_of_num ((Num.Int 0) -/ head.lin_coef));
      print_string ")*[";
      print_dag (getDag head.lin_id);
      print_string "] + ";
      print_comb tail

(* print combs list *)
let print_combs_list () =
  print_string "... Starred Cols' Linear Combinations ...\n";
  List.iter (function id_comb ->
    print_dag (getDag (fst id_comb));
    print_string " = ";
    print_comb (snd (id_comb));
    print_string "\n") !star_combs_list

      
(******************************************************************************************************)


(* 
   ----------------------------------------------------------------
   RANDOM HELPER FUNCTIONS
   ----------------------------------------------------------------
*)

(* 
   getListIndex:

   returns the index of the 1st element in the list 'l' that
   satisfies the predicate 'p', or raises a Not_found exception. 
*)
let rec getListIndex p l =
    match l with
      [] -> raise Not_found
    | head :: tail -> if (p head) then 0 else ((getListIndex p tail) + 1)

(* 
   getPartListIndex:

   returns the index of the 1st element starting from 'start' in the
   list 'l' that satisfies the predicate 'p', or raises a Not_found
   exception.
*)
let rec getPartListIndex p l start =
  if start = 0 then getListIndex p l
  else
    match l with
      [] -> raise Not_found
    | head :: tail -> (getPartListIndex p tail (start-1)) + 1


(* 
   deleteListIndex:
   
   return the given list after the 'n'th item was the deleted
   from it. Raises fialure if n>=length of list or if n<0.
*)
let rec deleteListIndex n lst =
  match lst with
    [] -> failwith "delteListIndex: input index too large"
  | head :: tail ->
      if n>0 then
 	head :: deleteListIndex (n-1) tail
      else if n=0 then
	tail
      else
	failwith "delteListIndex: negative index as input"

(*
   ----------------------------------------------------------------
   MANIPULATING THE TABLEAU
   ----------------------------------------------------------------
*)

(* 
   getRowIndex, getColIndex:

   given an id, returns the index of the row/col with that id or the
   number of rows/columns if it's not in the list
*)
let getRowIndex id =
  getListIndex (function x -> (x.tab_id = id)) our_tab.rows

let getColIndex id =
  getListIndex (function x -> (x.tab_id = id)) our_tab.cols

(* 
   findRow, Col:

   given an id, returns a row with that id or ()
*)
let findRow id = 
  List.find (function x -> (x.tab_id = id)) our_tab.rows
  
let findCol id =
  List.find (function x -> (x.tab_id = id)) our_tab.cols

(*
   getCol,getRow:

   returns the given row or column from our_tab
*)
let getCol n = List.nth our_tab.cols n
let getRow n = List.nth our_tab.rows n

(* 
   insertEntryToCol

   Inserts an entry to its right place in a list of one column entries
   if an entry with same row number already exists, replace it.
*)
let rec insertEntryToCol ent col_ents =
  match col_ents with
    [] -> [ent]
  | head :: tail -> if (!ent).row < (!head).row then ent :: col_ents else
    if (!ent).row = (!head).row then ent :: tail else
    head :: insertEntryToCol ent tail

(* 
   deleteEntryFromCol

   deletes the entry with of 'row_num' from 'col_ents'.
*)
let rec deleteEntryFromCol row_num col_ents =
  match col_ents with
    [] -> []
  | head :: tail ->
      if row_num > (!head).row then
 	head :: (deleteEntryFromCol row_num tail)
      else if row_num = (!head).row then
	(* this is the deleted entry - ignore it*)
	tail
      else
	(* the list doesn't contain that index *)
	col_ents

(* 
   deleteEntryFromRow

   deletes the entry with of 'col_num' from 'row_ents'.
*)
let rec deleteEntryFromRow col_num row_ents =
  match row_ents with
    [] -> []
  | head :: tail ->
      if col_num > (!head).col then
 	head :: (deleteEntryFromRow col_num tail)
      else if col_num = (!head).col then
	(* this is the deleted entry - ignore it*)
	tail
      else
	(* the list doesn't contain that index *)
	row_ents

(* 
   addEntryToCol 
   
   adds an entry to its column -- assumes that this is the last entry
   in the column.  
*)    
let addEntryToCol entry =
  let col = getCol (!entry).col in
  col.entries <- insertEntryToCol entry col.entries
  
(* 
   getRowEntry n entry-lst
  
   returns the first entry in entry-lst which has row n or a 0 entry
*)
let getRowEntry n lst =
  try List.find (function e -> ((!e).row = n)) lst
  with Not_found -> ref {row =  n; col = (-1); coef = Num.Int 0}

(* 
   getColEntry n entry-lst

   returns the first entry in entry-lst which has col n or a 0 entry
*)
let getColEntry n lst = 
  try List.find (function e -> ((!e).col = n)) lst
  with Not_found -> ref {row = (-1); col = n; coef = Num.Int 0}

(*
   ----------------------------------------------------------------
   COMBINING ROWS OF THE TABLEAU
   ----------------------------------------------------------------
*)

(*
   copyEntries:

   given a list of entry refs, returns a list of copies of those entry refs
*)
let copyEntries lst =
  List.map (function r -> ref {row=(-1); col=(!r).col; coef=(!r).coef}) lst

(* 
   sumEntries:

   given two lists of entries, merges the two lists together.
*)
let rec sumEntries lst1 lst2 =
  match (lst1, lst2) with
    (* two empty lists returns nothing *)
    ([],[]) -> []
    (* lst1 has elements but lst2 is empty *)
  | (head :: tail, []) -> copyEntries lst1  
    (* lst2 has elements but lst1 is empty *)
  | ([], head::tail) -> copyEntries lst2
    (* both lists have elements *)
  | (head1::tail1, head2::tail2) ->
      (* compare the columns *)
      match (((!head1).col < (!head2).col), 
	     ((!head1).col > (!head2).col)) with
	(* column1 < column2 *)
	(true, false) -> 
	  ref {row=(-1); col=(!head1).col; coef=(!head1).coef} :: 
	  (sumEntries (List.tl lst1) lst2)
	(* column1 > column2 *)
      |	(false, true) ->
	  ref {row=(-1); col=(!head2).col; coef=(!head2).coef} :: 
	  (sumEntries lst1 (List.tl lst2))
	(* column2 = column1 *)
      |	(false, false) ->
	  ref {row=(-1); col=(!head1).col; coef=(!head1).coef+/(!head2).coef} ::
	  (sumEntries (List.tl lst1) (List.tl lst2))
        (* just to avoid warning *)
      |	(true, true) -> []
	     
(*
   negateEntries:

   negates a list of entries
*)
let negateEntries entries = 
  List.map (function e -> ref {row = (-1); col = (!e).col; coef = (Num.Int 0) -/ (!e).coef}) entries

(*
   multEntries:

   multiplies a list of entries by a given constant
*)
let multEntries m entries = 
  List.map (function e -> ref {row = (-1); col = (!e).col; coef = m */ (!e).coef}) entries

(* 
   ----------------------------------------------------------------
   FUNCTIONS FOR PIVOTING
   ----------------------------------------------------------------
*)

(* 
   findColToPivot row_to_increase start_col

   returns a column number whose pivoting will increase the const of
   row_to_increase. searches only columns whose index is start_col or higher.
   raises Not_found if no such column
*)
let findColToPivot row_to_increase start_col =
  try
    (* try to find an unrestricted col with non-zero entry in row_to_inc *)
    getPartListIndex (function col -> ((not col.plus_rst) && ((!(getRowEntry row_to_increase col.entries)).coef <>/ Num.Int 0))) our_tab.cols start_col
  with Not_found ->
    (* try to find a non-starred col with positive entry in row_to_inc *)
    getPartListIndex (function col -> ((not col.star_rst) && ((!(getRowEntry row_to_increase col.entries)).coef >/ Num.Int 0))) our_tab.cols start_col


(* colsStates records all the sets of cols we had during one session
   of pivoting in order to avoid loops *)
let colsStates = ref [] (* reinit *)

(*
   resetColsState

   sets the states list as an empty list
*)
let resetColsStates () =
  colsStates := []

(*
   addCurColsState

   adds the current set of cols to the states list
*)
let addCurColsState () =
  let new_state = Sort.list id_leq_id (List.map (function c -> c.tab_id) our_tab.cols) in
  colsStates := new_state :: !colsStates

(*
   stateAfterPivotExists

   returns true if pivoting at 'row_num' and 'col_num' will get us to a state
   we were already in during this pivoting session
*)
let stateAfterPivotExists row_num col_num =
  let row_id = (getRow row_num).tab_id in
  let col_id = (getCol col_num).tab_id in
  let new_state = Sort.list id_leq_id (List.map (function c -> if c.tab_id = col_id then row_id else c.tab_id) our_tab.cols) in
  List.exists (function x -> x=new_state) !colsStates


(* type to record the row that goes tight first *)
type champ_t = {mutable row_num : int; mutable score : num}

(* 
   findRowToPivot row_to_increase col_num

   returns a row number whose pivoting with col_num will be a legal a pivot
   that will increase the const of row_to_increase.
   'inc_dir' should be positive when pivoting to incease const, and negative
   when pivoting a feasible tableau in order to change a col to a row.
   takes into account not to return to a previous state.
   assumes the entry at row_to_increase,col_num is non-zero
   returns -1 if no such row (all possible ones return to a previous state)
*)
let findRowToPivot row_to_increase col_num inc_dir =
  let col = getCol col_num in
  let row = getRow row_to_increase in
  let a = (!(getRowEntry row_to_increase col.entries)).coef in
  let const = (!(getColEntry 0 row.entries).coef) in
  let sign = if (a </ Num.Int 0) then Num.Int (-1) else if (a >/ Num.Int 0) then Num.Int 1 else (failwith "findRowToPivot: zero in col to pivot") in
  (* find the row that will go tight first *)
  let champ = {row_num=row_to_increase; score = Num.Int (-1)} in

  (* define a function that for a given entry checks if it the
     one that will go tight first from the ones checked so far *)
  let updateChamp entry =
    let cur_row_num = (!entry).row in
    let cur_row = getRow cur_row_num in
    let cur_a = (!entry).coef in
    let cur_const = (!(getColEntry 0 cur_row.entries).coef) in

    (* for an entry to go tight first:
       - it has to be plus restricted
       - it has to have the opposite sign than in row_to_increase
       - it is either the first one checked, or it must have the lowest
       score so far, or it has the same as lowest score before it and
       the champ before it is an illegal move *)
    if cur_row.plus_rst && (sign */ cur_a */ inc_dir) </ (Num.Int 0) &&
      ((champ.score=Num.Int (-1)) ||
       ((abs_num (cur_const // cur_a)) </ champ.score) ||
       (((abs_num (cur_const // cur_a)) =/ champ.score) &&
        champ.row_num=(-1))) then

      (* update the score, and the row_num (-1 if illegal) *)
      (champ.score <- abs_num (cur_const // cur_a);
       if (stateAfterPivotExists cur_row_num col_num) then
	 champ.row_num <- (-1)
       else  
	 champ.row_num <- cur_row_num)
    else
      () in

  List.iter updateChamp col.entries;
  champ.row_num

(* pivot point data type *)
type pivot_point = {row_num : int; col_num : int}

(* 
   findPivot row_to_increase start_col

   returns a pivot point whose pivoting will increase the const of
   row_to_increase. searches only columns whose index is start_col or higher.
   raises Not_found if no such point
*)
let rec findPivot row_to_increase start_col =
  let col_num = findColToPivot row_to_increase start_col in
  let row_num = findRowToPivot row_to_increase col_num (Num.Int 1) in
  if row_num = -1 then
    findPivot row_to_increase (start_col+1)
  else
    {row_num = row_num; col_num = col_num}

(*
   pivot row column
*)
let pivot pivot_p = 
  (* sanity check *)
  if pivot_p.col_num < 1 then
    failwith "tried to pivot the const row";

  (* get the actual row and column *)
  let row = getRow pivot_p.row_num in
  let col = getCol pivot_p.col_num in
        
  (* extract a values from the paper *)
  let a = !(getColEntry pivot_p.col_num row.entries).coef in
  (* sanity check *)
  if a =/ (Num.Int 0) then
    failwith "tried to pivot at a zero coefficient";
  
  (* set each entry to d-bc/a *)
  List.iter 
    (function r ->
      (* make list of entries which are -bc/a *)
      let cur_row_num = getRowIndex r.tab_id in
      if cur_row_num <> pivot_p.row_num then (
      	let c_val = !(getRowEntry cur_row_num col.entries).coef // a in
        (* if c != 0 *)
      	if (c_val <>/ Num.Int 0) then (
	  let entry_lst = negateEntries (multEntries c_val row.entries) in 
          (* add those entries to the current row *)
	  r.entries <- sumEntries r.entries entry_lst;
   	  (* set the pivot column entry *)
	  !(getColEntry pivot_p.col_num r.entries).coef <- c_val;
  	  List.iter (function e -> (!e).row <- cur_row_num) r.entries;
          (* add all entries to the columns *)
  	  List.iter addEntryToCol r.entries
	)
      )
    )
    our_tab.rows;
  
  (* set the pivot row entries *)
  List.iter (function e -> (!e).coef <- if (!e).col=pivot_p.col_num then
    (Num.Int 1)//a  else  (Num.Int 0)-/((!e).coef//a)) row.entries;
        
  (* finally, swap owners and flags *)
  let t = row.tab_id in
  row.tab_id <- col.tab_id;
  col.tab_id <- t;
  let t = row.plus_rst in
  row.plus_rst <- col.plus_rst;
  col.plus_rst <- t;
  let t = row.star_rst in
  row.star_rst <- col.star_rst;
  col.star_rst <- t

(* 
   addRowToTab

   adds a row to our_tab and adds all the entries to the
   correct columns returns the new row 
*)  
let addRowToTab entries id =  
  (* current row number is the length of the list *)
  let row_num = (List.length our_tab.rows) in
  (* set the correct row number in all entries *)
  List.iter (function e -> (!e).row <- row_num) entries;
  (* add all entries to the column *)
  List.iter addEntryToCol entries;
  (* add the row to our_tab.rows *)
  let row = { tab_id=id; star_rst=false; plus_rst=false; entries=entries } in
  our_tab.rows <- List.rev (row :: List.rev our_tab.rows);
  Stack.push (AddedID id) undoStack;
  row
  					       
(* 
   addColToTab :

   adds an empty column with the given id
*)
let addColToTab id = 
  our_tab.cols <- List.rev ({tab_id = id; star_rst=false; plus_rst=false; entries = []} :: (List.rev our_tab.cols));
  Stack.push (AddedID(id)) undoStack

(* 
   removeRowFromTab

   reomves a row from our_tab, deletes all its entries from the
   correct columns, and update the rest of the entries to reflect
   the index changes.
*)  
let removeRowFromTab row_num =
  (* remove this row entries from the columns *)
  List.iter (function x -> x.entries <- (deleteEntryFromCol row_num x.entries)) our_tab.cols;
  (* remove this row from the list of rows *)
  our_tab.rows <- deleteListIndex row_num our_tab.rows;
  (* update the entries of the higher rows - decrease the row index by 1*)
  for i=row_num to ((List.length our_tab.rows)-1) do
    List.iter (function x -> (!x).row <- ((!x).row-1)) (getRow i).entries
  done

(* 
   removeColFromTab

   reomves a column from our_tab, deletes all its entries from the
   correct rows, and update the rest of the entries to reflect
   the index changes.
   will actually do this only if the column is all zeros, otherwise, pivots
   and deletes the resulting row.
*)
let removeColFromTab col_num =
  try
    let row_to_inc = (!(List.find (function x -> (!x).coef <>/ Num.Int 0) (getCol col_num).entries)).row in
    let pivot_row = findRowToPivot row_to_inc col_num (Num.Int (-1)) in
    pivot {row_num=pivot_row; col_num=col_num};
    removeRowFromTab pivot_row
  with Not_found ->
    (* col has only zero entries - remove it as is *)
     (* remove this col entries from the rows *)
    (List.iter (function x -> x.entries <- (deleteEntryFromRow col_num x.entries)) our_tab.rows;
     (* remove this col from the list of cols *)
     our_tab.cols <- deleteListIndex col_num our_tab.cols;
     (* update the entries of the higher cols - decrease the col index by 1*)
     for i=col_num to ((List.length our_tab.cols)-1) do
       List.iter (function x -> (!x).col <- ((!x).col-1)) (getCol i).entries
     done)

(* vars_list_ref: a global ref to the list of all variables seen so far *)
let vars_list_ref = ref []  (* reinit *)

(*
   addExpression:

   given an expression in a dag (like 3x+4y), returns a list of
   entries representing that expression.  note that this function may
   add rows to the tableau if needed to represent the expression
   (including a row for the expression itself)
*)
let rec addExpression dag = 
(*  if !debug then
    ignore (printf "%s: addExpression %d %a\n" !simplex_id dag.Dag.id Dag.dn dag);
*)
  let id = Dag_id dag.Dag.id in
  try 
    (findRow id).entries
  with Not_found ->
    try 
      [ref {row=(-1); col=getColIndex id; coef=Num.Int 1}]
    with Not_found ->

      (* otherwise, we match on the operator *)
      match (dag.op) with
        (* if it's plus, sum the entries of the arguments *)
	Plus  -> 
	  (addRowToTab 
	     (sumEntries (addExpression (List.nth dag.args 0)) (addExpression (List.nth dag.args 1))) id).entries

        (* if it's minus, subtract the second argument from the first *)
      | Minus -> 
	  (addRowToTab 
	     (sumEntries (addExpression (List.nth dag.args 0)) (negateEntries (addExpression (List.nth dag.args 1)))) id).entries

        (* if it's times, check the kinds of arguments *)
      | Times ->
	  (match ((List.nth dag.args 0).op, (List.nth dag.args 1).op) with
	    (* constant x, whatever *)
	    (Dag.Int(i), _) -> 
	      ignore (addExpression (List.nth dag.args 0));
	      (addRowToTab (multEntries (Num.Int i) (addExpression (List.nth dag.args 1))) id).entries
        	(* whatever x, constant *)
	  | (_, Dag.Int(i)) -> 
	      ignore (addExpression (List.nth dag.args 1));
	      (addRowToTab (multEntries (Num.Int i) (addExpression (List.nth dag.args 0))) id).entries
            (* anything else *)
	  | (_, _) ->  
	      ignore (addExpression (List.nth dag.args 0));
	      ignore (addExpression (List.nth dag.args 1));
	      ignore (addColToTab id);
	      [ref {row=(-1); col=getColIndex id; coef=Num.Int 1}])

        (* if it's a function, add the arguments and then add a column and return the dummy entry *)
      | Fun _ | Div _ | Mod _ | Floor _ | Ceil _  ->
	  let args = dag.args in
	  List.iter (function x -> ignore (addExpression x)) args;
	  ignore (addColToTab id);
	  [ref {row=(-1); col=getColIndex id; coef=Num.Int 1}]

        (* if it's a variable, add a new column and return a dummy entry of that column *)
      | Var _ ->
	  vars_list_ref := id :: !vars_list_ref;
	  ignore (addColToTab id);
	  [ref {row=(-1); col=getColIndex id; coef=Num.Int 1}]

        (* if it's a constant, add a new row with that constant *)
      | Int i ->
	  (addRowToTab [ref {row=(-1); col=0; coef=Num.Int i}] id).entries

      |	_ -> failwith "Tried to add a non-expression to the simplex tableau"

(*  
    setPlusRest, setStarRest :

    set the plus/star restriction of 'id' to 'b'.
    raises Not_found if id not in tableau
*)
let setPlusRest id b =
  try
    let row = findRow id in
    row.plus_rst <- b
  with Not_found ->
    let col = findCol id in
    col.plus_rst <- b

let setStarRest id b =
  try
    let row = findRow id in
    row.star_rst <- b
  with Not_found ->
    let col = findCol id in
    col.star_rst <- b

(*
   ----------------------------------------------------------------
   FOR PROVING EQUALITIES AND CONTRADICTIONS
   ----------------------------------------------------------------
*)


(*
   addToComb :

   adds a new element to a lin. comb. the elem's id is as the col id of the
   entry, its value is the coef of that entry divided by 'fact'.
   'dir' decides which entries are taken as are and which are multiplied
   by their col's comb. (when 'dir' > 0, the positive ones are taken as are)
*)
let addToComb ent comb fact dir =
  if (!ent).col = 0 then
    (* we never add the const col *)
    comb
  else
    let col_id = (getCol (!ent).col).tab_id in
    let c = abs_num ((!ent).coef // fact) in
    let sign = (sign_num (!ent).coef) * dir in
    if (sign > 0) then
      (* add this coef to comb *)
      {lin_id = col_id; lin_coef = c} :: comb
    else if (sign < 0) then
      (* this is a starred col, use it's comb list *)
      List.append (List.map (function elem -> {lin_id = elem.lin_id; lin_coef = c */ elem.lin_coef}) (List.assoc col_id !star_combs_list)) comb
    else
      comb


(*
   simpComb

   returns a simplified linear combination
*)
let simpComb comb =
  (* sort the combination by the id's *)
  let sorted_comb = Sort.list (function elem1 -> function elem2 -> id_leq_id elem1.lin_id elem2.lin_id) comb in
  (* merge the coefs of identical id's *)
  let rec mergeComb com =
    match com with
      [] -> []
    | [head] -> [head]
    | hd1 :: (hd2 :: tail) ->
	if hd1.lin_id = hd2.lin_id then
	  mergeComb ({lin_id = hd1.lin_id; lin_coef = hd1.lin_coef +/ hd2.lin_coef} :: tail)
	else
	  hd1 :: mergeComb (hd2 :: tail) in
  mergeComb sorted_comb


(*
   starCol

   star restrict the col with the given id.
   'row' is the row that caused it to be star rest.
   add the linear combination for that id to the list of linear combs.
*)
let starCol col_id row =
  let col_num = getColIndex col_id in
  (* recall 'a', the factor to divide all coefs by *)
  let a = abs_num (!(getColEntry col_num row.entries)).coef in
  (* create a lin. comb. which includes only the row id *)
  let l_comb = [{lin_id = row.tab_id; lin_coef = (Num.Int 1)//a}] in
  (* add all the other col's coefs, and simplify it *)
  let l_comb = simpComb (List.fold_left (function comb -> function ent -> if (!ent).col <> col_num then addToComb ent comb a (-1) else comb) l_comb row.entries) in
  (* add the new lin. comb. to the list of starred cols *)
  star_combs_list := (col_id,l_comb) :: !star_combs_list;
  (* mark col as star restricted and recall it for future undo *)
  setStarRest col_id true;
  Stack.push (StarID col_id) undoStack


(*
   leastCommonMult and greatestCommonDiv

   list versions of our favorite number theoretic functions
*)
let leastCommonMult lst =
  List.fold_left
    (function x -> function y -> (Big_int.div_big_int
(Big_int.mult_big_int x y) (Big_int.gcd_big_int x y)))
    Big_int.unit_big_int
    lst

let greatCommonDiv lst =
  List.fold_left
    (function x -> function y -> Big_int.gcd_big_int x y)
    (List.hd lst)
    lst

(*
   makeIntComb

   turns a lin_comb into an int_lin_comb
*)
let makeIntComb comb =
  (* get the coefficients *)
  let coef_lst = List.map (function e -> (Num.ratio_of_num e.lin_coef))
comb in
  (* find the lcm *)
  let lcm = leastCommonMult (List.map (function c -> Ratio.denominator_ratio c) coef_lst) in
  (* find integer coefficients *)
  let int_coefs = List.map (function c -> (Ratio.big_int_of_ratio (Ratio.mult_big_int_ratio lcm c))) coef_lst in
  (* find the gcd of integer coefficients *)
  let gcd = greatCommonDiv int_coefs in
  (* divide integer coefficients by their gcd *)
  let best_coefs = List.map (function c -> let ans = Big_int.div_big_int c gcd in if (Big_int.is_int_big_int ans) then Big_int.int_of_big_int ans else failwith "makeIntComb: big_int too large") int_coefs in
  (* return linear combination with best coefficients *)
  List.map2 (function i -> function c -> {i_lin_id = i.lin_id; i_lin_coef = c}) comb best_coefs

(*
   expOfDag

*)
let expOfDag node =
  match nodeVal node with
    DExp (e) -> e
  | DAtom _ -> failwith "expOfDag: tried to get expression of an atom node"

let proveID id =
(*  if !debug then
    ignore (printf "%s: proveID %d\n" !simplex_id (intId id));
*)
  (List.assoc (intId id) !pf_exp_list) ()

(*
   makeCombProof

   builds the proof nested geq*>= necessary in the two kinds of proof
*)
let makeCombProof i_comb =
(*  if !debug then
    ignore (printf "%s: makeCombProof\n" !simplex_id);
*)
  List.fold_left
    (function pf -> function elem -> Proof.G ("geq*>=", [I elem.i_lin_coef; P (proveID elem.i_lin_id); P pf]))
    (Proof.G ("geq0",[])) i_comb

(* 
   expOfComb

   returns the expression of a given combination
   argi=0 means the to take left hand side of each inequality
   argi=1 means the right hand side
*)
let expOfComb comb arg_i = 
  List.fold_left (function exp -> function elem -> Input.Plus (Input.Times ((Input.Int elem.i_lin_coef), expOfDag (List.nth (getDag elem.i_lin_id).args arg_i)), exp)) (Input.Int 0) comb


(*
   proveContra
   
   proves a contradiction
*)
let proveContra row =
(*  if !debug then
    ignore (printf "%s: proveContra %d\n" !simplex_id (intId row.tab_id));
*)
  (* create a comb with 1 times the row *)
  let l_comb = [{lin_id = row.tab_id; lin_coef = Num.Int 1}] in
  (* add all the col's coefs to the comb, and simplify it *)
  let l_comb = simpComb (List.fold_left (function comb -> function ent -> addToComb ent comb (Num.Int 1) (-1)) l_comb row.entries) in
  (* make into an int comb *)
  let i_comb = makeIntComb l_comb in
  (* prove the ineq which is the lin comb of ineqs *)
  let geq_proof = makeCombProof i_comb in
  (* prove that the 0 >= comb + 1 *)
  let lt_proof = Proof.G ("geqCheck", [E (expOfComb i_comb 1); E (Input.Plus (expOfComb i_comb 0, Input.Int 1))]) in
  (* prove the contradiction *)
  let contra_proof = Proof.G ("geqLt_FI", [P geq_proof; P lt_proof]) in
  raise (Contradiction contra_proof)

(*
   provePositiveDifference

   given row1 and row2, proves that row1 > row2
*)

let provePositiveDifference row1 row2 = 
  (* find difference of the two rows *)
  let entries = sumEntries row1.entries (negateEntries row2.entries) in
  (* make integer linear combination of row1 - row2 *)
  let l_comb = simpComb (List.fold_left (function comb -> function ent -> addToComb ent comb (Num.Int 1) 1) [] entries) in

  (* if comb is empty these expressions are equivalent *)
  if (List.length l_comb) = 0 then
    Proof.G ("geqCheck", [E (expOfDag (getDag row1.tab_id)); E (expOfDag (getDag row2.tab_id))])
  else
    let i_comb = makeIntComb l_comb in
    (* create proof that s1 > 0 *)
    let sum_geq_proof = makeCombProof i_comb in
    (* find the factor *)
    let d = int_of_num ((Num.Int (List.hd i_comb).i_lin_coef) // (List.hd l_comb).lin_coef) in
    (* show that d*(r1-r2) = s1 *)
    let eq_prf = Proof.G ("eqCheck", [ E (Input.Plus 
					    (Input.Times 
					       (Input.Int d, 
					      	Input.Minus ((expOfDag (getDag row1.tab_id)),expOfDag (getDag row2.tab_id))),
					     (Input.Minus ((Input.Int 1),(Input.Int 1)))));
				       E (Input.Minus
					    (expOfComb i_comb 0, expOfComb i_comb 1)) ]) in
    (* return value *)
    Proof.G ("geq*i",[I 1; I d; P eq_prf; P sum_geq_proof])
    
(* 
   proveEquality
   
   proves that row1=row2
*)
    
let proveEquality id1 id2 =
  (* find row1 or make a dummy row *)
  let row1 = 
    try 
      findRow id1
    with Not_found ->
      {tab_id = id1; star_rst = false; plus_rst = false; 
       entries=[ref {row=(-1); col=getColIndex id1; coef=Num.Int 1}]} in
  (* find row2 or make a dummy row *)
  let row2 = 
    try 
      findRow id2
    with Not_found ->
      {tab_id = id2; star_rst = false; plus_rst = false;
       entries = [ref {row=(-1); col=getColIndex id2; coef=Num.Int 1}]} in
  (* make the proof *)
  Proof.G("eqi",[P (provePositiveDifference row1 row2); P (provePositiveDifference row2 row1)])
    
(*  
    makeRowNonNegative

    restrict a row of our_tab to be non-negative
    makes all the needed pivoting
    raises contradiction if can not make it non-negative
*)
let makeRowNonNegative row_num =
(*  if !debug then
    ignore (printf "%s: makeRowNonNegative %d\n" !simplex_id row_num);
*)
  let row = getRow row_num in
  row.plus_rst <- true;
  Stack.push (PlusID row.tab_id) undoStack;
  let const = ref (!(getColEntry 0 row.entries)).coef in
  let cont = ref true in
  while row.plus_rst && (!const <=/ Num.Int 0) && !cont do
    addCurColsState (); (* save the current state, to not return to it *)
    try
      let pivot_pt = findPivot row_num 1 in
      if (!const </ Num.Int 0) || (pivot_pt.row_num <> row_num) || ((not (getCol pivot_pt.col_num).plus_rst)) then
	(pivot pivot_pt; const := (!(getColEntry 0 row.entries)).coef)
      else
	cont := false (* no need to pivot, can increase const to inf *)
    with Not_found ->
      (cont := false;
       resetColsStates (); (* this is here for the case we raise a contra *)

       (* sanity check - that all neg are plus_rst, all pos are star_rst *)
       if List.exists (function c -> let coef = (!(getRowEntry row_num c.entries)).coef in (c.tab_id <> Const) && (((coef >/ (Num.Int 0)) && (not c.star_rst)) || ((coef </ (Num.Int 0)) && (not c.plus_rst)))) our_tab.cols then
	 failwith "no pivot, but row is not maximized";

       if !const =/ (Num.Int 0) then
	 (* maximized at 0 - star all the negative collumns *)
	 List.iter (function c -> if (not c.star_rst) && (!(getRowEntry row_num c.entries)).coef </ (Num.Int 0) then starCol c.tab_id row) our_tab.cols
       else
	 (* maximized at negative - contra *)
	 proveContra row)
  done;
  resetColsStates ()

(*  
    makeIDNonNegative

    restrict an id of our_tab to be non-negative
    makes all the needed pivoting
    raises contradiction if can not make it non-negative
*)
let makeIDNonNegative id =
(*  if !debug then
    ignore (printf "%s: makeIDNonNegative %d\n" !simplex_id (intId id));
*)
  try
    let col_num = getColIndex id in
    if (getCol col_num).plus_rst then
      () (* col is already plus restricted, do nothing *)
    else
      (* if col has non-zeros, pivot into a row and restrict row *)
      try
    	let row_to_inc = (!(List.find (function x -> (!x).coef <>/ Num.Int 0) (getCol col_num).entries)).row in
    	let pivot_row = findRowToPivot row_to_inc col_num (Num.Int (-1)) in
	(* pivot into a row *)
	pivot {row_num=pivot_row; col_num=col_num};
	(* restrict the row *)
	makeRowNonNegative pivot_row;
      with Not_found ->
	(* an empty col, restrict it as is *)
  	(setPlusRest id true;
  	 Stack.push (PlusID id) undoStack)
  with Not_found ->
    (* this is a row *)
    let row_num = getRowIndex id in    
    if (getRow row_num).plus_rst then
      () (* row is already plus restricted, do nothing *)
    else
      makeRowNonNegative row_num

(* 
   ----------------------------------------------------------------
   REPORTING EQUALITIES
   ----------------------------------------------------------------
*)

(*
   isVariable

   returns whether a given vector is owned by a variable Dag
*)
let isVariable v = List.exists (function i -> i = v.tab_id) !vars_list_ref

(*
   isZeroInNonStarRestrictedColumns 

   returns whether a given set of entries is ...
*)
let isZeroInNonStarRestrictedColumns entries =
  if (List.exists (function e -> (((!e).coef <>/ Num.Int 0) && (not (getCol (!e).col).star_rst))) entries)
  then false 
  else true
    
(* 
   findEqualityWithRow
   
   given a row and list of other rows, returns a list of those rows which are equal to the first
*)
let findEqualityWithRow row other_rows =
  row :: List.find_all
          (* function returns true when row.entries - other_row.entries = 0 in non star restricted columns *)
	  (function v -> (isZeroInNonStarRestrictedColumns (sumEntries row.entries (negateEntries v.entries))))
	  (* applied to other_rows *)
	  other_rows
    
(*
   findEqualitiesBetweenRows

   given a list of rows, returns a list of lists of ids which are equal
*)
let rec findEqualitiesBetweenAllRows row_lst found_equalities = 
  match row_lst with
    [] -> []
  | head::tail -> 
      (* if the first row is already in an equality, skip it *)
      if (List.exists (function i -> i.tab_id=head.tab_id) found_equalities) then
	findEqualitiesBetweenAllRows tail found_equalities
      else
	(* get the list of rows which are equal to the first row *)
	let equilist = findEqualityWithRow head tail in
	(* if there are no equalities, just recurse *)
	if (equilist = [head]) then findEqualitiesBetweenAllRows tail found_equalities
	(* otherwise, return them and recurse *)
	else equilist :: findEqualitiesBetweenAllRows tail (List.concat [equilist; found_equalities])

(*
   equalityPairs

   given a list of lists of tab vectors which are equal, returns a list of all pairs of equal dags
*)
let rec makeAllPairs lst =
  match lst with
    [] -> []
  | head::tail ->
      List.concat [ 
      (List.map (function e -> [head; e]) tail); 
      (makeAllPairs tail) ]
	
(* 
   makeEqualityMessage

   given a pair of vectors, returns a message which contains a dag asserting the equality of those vectors
*)
let makeEqualityMessage p = 
  let dag1 = getDag (List.nth p 0).tab_id in
  let dag2 = getDag (List.nth p 1).tab_id in
  {body=Equality(dag1,dag2); owner= !simplex_id; id = 1; goal=false}

(* 
   isPublic
*)
let isPublic v = (intId v.tab_id) >= 0
    
(* 
   findEqualitiesBetweenRowsAndCols

   given a list of rows, they are equal to a column if they only 
   have one entry in a non-starred column and its coefficient is 1
*)
let findEqualitiesBetweenRowsAndCols rows =
  List.find_all
    (function e -> e <> [])
    (List.map 
       (function r ->
	 let relevant_entries = (List.find_all (function e -> (not (getCol (!e).col).star_rst) && ((!e).coef <>/ (Num.Int 0))) r.entries) in
	 if (List.length relevant_entries = 1) then
	   let e = !(List.hd relevant_entries) in
	   if (e.col <> 0) && (e.coef =/ Num.Int 1) && (isPublic (getCol e.col)) then
	     [r; (getCol e.col)]
	   else []
	 else [])
       rows)

(*
   equalityMsgEq

   checks if two equality messages represent the same thing (checks commutativity)
*)
let equalityMsgEq eq_msg1 eq_msg2 =
  match (eq_msg1.body, eq_msg2.body) with
    (Equality (a,b), Equality (c,d)) -> (a=c && b=d) || (a=d && b=c)
  | (_,_) -> false

(*
  isNewEquality

  returns true if the given msg is of a new equality
*)
let isNewEquality msg msg_list =
  if match msg.body with
    Equality (a,b) -> not (Dag.knownEq a b)
  | _ -> failwith "bug in Simplex.isNewEquality"
  then
    not (List.exists (function o -> (equalityMsgEq o msg)) msg_list)
  else
    false
(*  
    reportEqualities

    report all new equalities
*)
let reportEqualities msg_list =
  let public_rows = List.find_all (function v -> (isPublic v)) our_tab.rows in
  let public_cols = List.find_all (function v -> (isPublic v)) our_tab.cols in
  let equality_lst = List.concat [
    (* first add appropriate star restricted columns to the equality list *)
    (List.find_all (function v -> (v.star_rst = true)) public_cols);
    (* then add appropriate rows *)
    (List.find_all (function v -> (isZeroInNonStarRestrictedColumns v.entries)) public_rows)
  ] in
  (* find equalities between rows *)
  let row_equalities = 
    findEqualitiesBetweenAllRows public_rows equality_lst in
  (* find equalities between rows and columns *)
  let row_col_equalities =
    findEqualitiesBetweenRowsAndCols public_rows in
  (* return all the pairs of equal vectors *)
  let equality_pairs = 
    List.concat 
      (List.map (function eq_lst -> makeAllPairs eq_lst) (List.concat [equality_lst::row_equalities; row_col_equalities])) in
  let equality_messages = 
    List.map (function p -> makeEqualityMessage p) equality_pairs in
  (* filter out messages of already known equalities *)
  List.find_all (function m -> isNewEquality m msg_list) equality_messages
    

(********************************************************************************************)

(* a global node for the 1 const *)
let one_node = newPrivateNode (Int 1) []
let zero_node = newPrivateNode (Int 0) []
(*
   our_assume

   input is a messgae list of equalities, a Dag node which is a new atomic
   formula, and a proof of the atomic formula.
   Adds the inequality to the tableau.
   returns an updated message list of equalities.
*)
let rec our_assume msg_list atom pf_f =
(*  if !debug then
    ignore (printf "%s: our_assume %a\n" !simplex_id Dag.dn atom);
*)
  match atom.op with
    Geq ->
      (* make an expression which is >= 0 *)
      let minus_node = newPrivateNode Minus atom.args in
      (pf_exp_list := (minus_node.Dag.id, pf_f) :: !pf_exp_list;
       ignore (addExpression minus_node);
       makeIDNonNegative (Dag_id minus_node.Dag.id);
       msg_list @ (reportEqualities msg_list))

  | Gt ->
      let new_atom = newPrivateNode Geq [List.nth atom.args 0; newPrivateNode Plus [List.nth atom.args 1; one_node]] in
      our_assume msg_list new_atom (function () -> Proof.G ("gtgeq", [P (pf_f())]))

  | Leq ->
      let new_atom = newPrivateNode Geq [List.nth atom.args 1; List.nth atom.args 0] in
      our_assume msg_list new_atom (function () -> Proof.G ("leqgeq", [P (pf_f())]))

  | Lt ->
      let new_atom = newPrivateNode Geq [List.nth atom.args 1; newPrivateNode Plus [List.nth atom.args 0; one_node]] in
      our_assume msg_list new_atom (function () -> Proof.G ("ltgeq", [P (pf_f())]))

  | Eq ->
      let new_atom = newPrivateNode Geq atom.args in
      let new_msg_list = our_assume msg_list new_atom (function () -> Proof.G ("eqgeq", [P (pf_f())])) in
      let new_atom = newPrivateNode Geq [List.nth atom.args 1; List.nth atom.args 0] in
      our_assume new_msg_list new_atom (function () -> Proof.G ("eqgeqr", [P (pf_f())]))

  | Neq ->
      (ignore (addExpression (List.nth atom.args 0));
       ignore (addExpression (List.nth atom.args 1));
      msg_list @ (reportEqualities msg_list))

  | BFun _ ->
      (List.iter (function exp -> ignore (addExpression exp)) atom.args;
       msg_list @ (reportEqualities msg_list))

  | _ -> msg_list

(*
   assume

   the interface to the outside world:
   gets a list of eq messages and a new message. add the contents of the new
   message to our data structures. raise a contradiction if new message
   contradicts the existing data. return the list of eq messages appended with
   new implied equalities.
*)
exception PredicateEq
let assume msg_list msg =
  if !debug then
    ignore (printf "%s: assume %a\n" !simplex_id Satproc.dm msg); 
  if msg.owner <> !simplex_id then
    try
      let atom = match msg.body with
      	Literal (l) -> l
      | Equality (n1,n2) ->
	  if (Dag.isExp n1) && (Dag.isExp n2) then
	    newPrivateNode Eq [n1;n2]
	  else
	    raise PredicateEq
      in
      let real_atom = if msg.goal then
      	match (atom.op,atom.args) with
      	  (Geq, [n1;n2]) -> newPrivateNode Lt [n1;n2]
      	| (Gt, [n1;n2])  -> newPrivateNode Leq [n1;n2]
      	| (Leq, [n1;n2]) -> newPrivateNode Gt [n1;n2]
      	| (Lt, [n1;n2])  -> newPrivateNode Geq [n1;n2]
      	| (Eq, [n1;n2])  -> newPrivateNode Neq [n1;n2]
      	| (Neq, [n1;n2]) -> newPrivateNode Eq [n1;n2]
      	| _ -> atom
      else
      	atom
      in
      our_assume msg_list real_atom (function () -> !(Satproc.proveIt) msg)
    with PredicateEq -> msg_list
  else
    msg_list

(* prove_it :

   return the proof of msg, where msg is an equality reported by simplex
   !to be implemented!
*)
let proveIt msg =
  if !debug then
    ignore (printf "%s: proveIt %a\n" !simplex_id Satproc.dm msg);
  if msg.owner <> !simplex_id then
    failwith "SX.proveIt: not a simplex message";
  match msg.body with
    Equality (n1,n2) -> proveEquality (Dag_id n1.Dag.id) (Dag_id n2.Dag.id)
  | _ -> failwith "SX.proveIt: not an equality message"


(* 
   snapshot

   puts a marker on the undo stack to recall current state
*)
let snapshot () =
  if !debug then
    ignore (printf "%s: snapshot\n" !simplex_id);
  Stack.push SnapMark undoStack
    

(*
   undo

   restore state to one of last snapshot, or the initial state if
   no snapshot marks on the stack
*)
let rec undo () =
  if !debug then
    ignore (printf "%s: undo\n" !simplex_id);
  try
    match (Stack.pop undoStack) with
      SnapMark -> ()
    | (AddedPrivateNode id) ->
	(private_nodes := List.remove_assoc id !private_nodes;
	 undo ())
    | (AddedID id) -> 
	(* in case it's a var, delete from vars list *)
	(try
	  vars_list_ref := deleteListIndex (getListIndex (function x -> x = id) !vars_list_ref) !vars_list_ref;
	with Not_found ->
	  (););
	(try
	  removeColFromTab (getColIndex id);
	  undo()
	with Not_found ->
	  (removeRowFromTab (getRowIndex id);
	   undo()))
    | (PlusID id) ->
	(setPlusRest id false;
	 (* remove the proof of this expression being >= 0 *)
	 pf_exp_list := List.remove_assoc (intId id) !pf_exp_list;
	 undo())
    | (StarID id) ->
 	(setStarRest id false;
	 (* remove the lin. comb. of this col from the list of combs of stars*)
	 if fst (List.hd !star_combs_list) <> id then failwith "undo star";
	 star_combs_list := List.tl !star_combs_list;
	 undo())
  with Stack.Empty -> ()


let reinitialize () =
    (* our_tab *)
  our_tab.rows <-[]; 
  our_tab.cols <-[{tab_id = Const; star_rst = false; plus_rst = false; entries = []}];
  (* *)
  Stack.clear undoStack;
  star_combs_list := [];
  private_nodes := [] ;
  next_private_id := -1 ;
  pf_exp_list := [];
  colsStates := [];
  vars_list_ref := []; ()


(*
   initialize:

   for the main process to initialize simplex
*)
let initialize id = 
  simplex_id := id;
  { Satproc.snapshot = snapshot;
    Satproc.undo = undo;
    Satproc.assume = assume;
    Satproc.proveIt = proveIt;
    Satproc.debug = debug;
    Satproc.ident  = id;
    Satproc.reinitialize = reinitialize;
  } 




(* a couple rows to test things out *)
(*  
let r1 = {tab_id=Dag_id 1; star_rst=false; plus_rst=false; 
	  entries= [ref {row=0; col=0; coef=Num.Int 1}; ref {row=0; col=1; coef=Num.Int 1}]}
let r2 = {tab_id=Dag_id 2; star_rst=false; plus_rst=false; 
	  entries= [ref {row=1; col=1; coef=Num.Int 1}; ref {row=1; col=2; coef=Num.Int 1}]}
*)

(* dags for testing *)
(* dags that form the expression 3x+2y, and more...*)
(*
let d_x = newNode (Var "x") []
let d_y = newNode (Var "y") []
let d_3 = newNode (Dag.Int 3) []
let d_2 = newNode (Dag.Int 2) []
let d_5 = newNode (Dag.Int 5) []
let d_14 = newNode (Dag.Int 14) []
let d_3x = newNode Times [d_3; d_x]
let d_2y = newNode Times [d_2; d_y]
let d_3x_plus_2y = newNode Plus [d_3x; d_2y]
let d_3x_times_14 = newNode Times [d_3x; d_14]
let d_5_plus_y = newNode Plus [d_5; d_y]
let d_x_lt_y = newNode Lt [d_x; d_y]
let d_x_gt_5_plus_y = newNode Gt [d_x; d_5_plus_y]
let d_3x_times_14_eq_5_plus_y = newNode Eq [d_3x_times_14; d_5_plus_y]

let d_f_x_y_3 = newNode (Fun "f") [d_x; d_y; d_3]
let d_mod_2_14 = newNode (Mod 2) [d_14]
	      
let msg1 = {body=Literal d_x_lt_y; owner = "other"; id=0; goal = false}
let msg2 = {body=Literal d_x_gt_5_plus_y; owner = "other"; id=1; goal = false}

let d_z = newNode (Var "z") []
let d_3z = newNode Times [d_3; d_z]
let d_3x_gt_2y = newNode Gt [d_3x; d_2y]
let d_2y_gt_3z = newNode Gt [d_2y; d_3z]
let d_z_lt_x   = newNode Lt [d_z;d_x]
let d_x_gt_y   = newNode Gt [d_x; d_y]
let d_y_gt_z   = newNode Gt [d_y; d_z]
let d_z_gt_x   = newNode Gt [d_z;d_x]
    
let msg3 = {body = Literal d_3x_gt_2y; owner = "other"; id=2; goal=false}
let msg4 = {body = Literal d_2y_gt_3z; owner = "other"; id=3; goal=false}
let msg5 = {body = Literal d_z_gt_x; owner = "other"; id=4; goal=false}
    
let msg6 = {body = Literal d_x_gt_y; owner = "other"; id = 5; goal=false}
let msg7 = {body = Literal d_y_gt_z; owner = "other"; id = 5; goal=false}
*)

let pt () = print_tableau our_tab

let rec assumeMsgs lst =
  List.iter 
    (function m ->
      print_string "adding   : ";
      (match m.body with
	Literal l -> print_dag l
      |	_ -> ());
      print_string "\n";
      let msgs = assume [] m in
      List.iter 
	(function n -> 
	  print_string "reported : ";
	  (match n.body with
	    Literal l -> print_dag l
	  | _ -> ());
	  print_string "\n"
	)
	msgs
    )
    lst

